/* _______              __ 
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::ifstream;
using std::ios;

#include "FABView.H"

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"

#include "AMR.H"
#include "AMRLevel.H"
#include "AMRLevelPolytropicGasFactory.H"
#include "AMRLevelPolytropicGas.H"

#include "PatchPolytropicGas.H"

#include "RampIBC.H"
#include "ExplosionIBC.H"
#include "GaussianIBC.H"
#include "SineIBC.H"
#include "WaveIBC.H"


// Available sample problems
#define PROBLEM_RAMP      0
#define PROBLEM_EXPLOSION 1
#define PROBLEM_GAUSSIAN  2
#define PROBLEM_SINE      3
#define PROBLEM_WAVE      4

// Possible pressure relationships for the initial condition
#define PRESSURE_ISENTROPIC 0
#define PRESSURE_CONSTANT   1

// amrGodunov is a function (as opposed to inline in main()) to get
// around MPI scoping problems
void amrGodunov();

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile);

// One more function for MPI
void dumpmemoryatexit();

int main(int a_argc, char* a_argv[])
{
#ifdef MPI
  // Start MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  // Check for an input file
  char* inFile = NULL;

  if (a_argc > 1)
  {
    inFile = a_argv[1];
  }
  else
  {
    pout() << "Usage:  amrGodunov...ex <inputfile>" << endl;
    pout() << "No input file specified" << endl;
    return -1;
  }

  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  // Run amrGodunov, i.e., do the computation
  amrGodunov();

#ifdef MPI
  // Exit MPI 
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void amrGodunov()
{
  // Read inputs that are prefixed with "godonouv."
  ParmParse ppgodunov("godunov");

  // Determine the sample problem specified
  int problem = -1;
  std::string problemString;

  if (ppgodunov.contains("problem"))
    {
      ppgodunov.query("problem",problemString);
      if (problemString == "ramp")
        {
          problem = PROBLEM_RAMP;
        }
      else if (problemString == "explosion")
        {
          problem = PROBLEM_EXPLOSION;
        }
      else if (problemString == "gaussian")
        {
          problem = PROBLEM_GAUSSIAN;
        }
      else if (problemString == "sine")
        {
          problem = PROBLEM_SINE;
        }
      else if (problemString == "wave")
        {
          problem = PROBLEM_WAVE;
        }

      // The sample problem given isn't valid
      if (problem == -1)
        {
          pout() << "Invalid problem, \"" << problemString << "\", specified in input file" << endl << endl; 
          return;
        }
    }
  else
    {
      // A sample problem must be specified
      pout() << "\"godunov.problem\" not specified in input file" << endl << endl; 
      return;
    }

  // This determines the amount of diagnositic output generated
  int verbosity = 0;
  ppgodunov.query("verbosity",verbosity);
  assert(verbosity >= 0);

  // For all gas dynamics
  Real gamma = 1.4;
  ppgodunov.query("gamma",gamma);

  // For "ramp"
  Real alpha = 30.0;
  Real xcorner = 0.1;

  // For "ramp" and "explosion"
  Real ms = 10.0;

  // For "explosion" and "gaussian"
  Real size = 0.25;

  // For "explosion", "gaussian", "sine", and "wave"
  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  vector<Real> velocitypp(SpaceDim,0.0);
  RealVect velocity;

  // For "gaussian", "sine", and "wave"
  Real ambientDensity = 1.4;
  Real deltaDensity = 0.014;
  int pressure = -1;
  std::string pressureString;

  // For "wave"
  vector<int> waveNumberpp(SpaceDim,0);
  waveNumberpp[0] = 1;
  IntVect waveNumber;

  // Parameters specific to different sample problems
  switch (problem)
    {
    case PROBLEM_RAMP:
      {
        ppgodunov.query("angle_deg",alpha);
        ppgodunov.query("shock_mach",ms);
        ppgodunov.query("xcorner",xcorner);
        break;
      }

    case PROBLEM_EXPLOSION:
      {
        ppgodunov.query("shock_mach",ms);

        ppgodunov.queryarr("center",centerpp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

        ppgodunov.query("size",size);

        ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

        break;
      }

    case PROBLEM_GAUSSIAN:
      {
        ppgodunov.query("ambient_density",ambientDensity);
        ppgodunov.query("delta_density",deltaDensity);

        if (ppgodunov.contains("pressure"))
          {
            ppgodunov.query("pressure",pressureString);
            if (pressureString == "isentropic")
              {
                pressure = PRESSURE_ISENTROPIC;
              }
            else if (pressureString == "constant")
              {
                pressure = PRESSURE_CONSTANT;
              }

            if (pressure == -1)
              {
                pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
                return;
              }
          }
        else
          {
            pressure = PRESSURE_ISENTROPIC;
          }

        ppgodunov.queryarr("center",centerpp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

        ppgodunov.query("size",size);

        ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

        break;
      }
    
    case PROBLEM_SINE:
      {
        ppgodunov.query("ambient_density",ambientDensity);
        ppgodunov.query("delta_density",deltaDensity);

        if (ppgodunov.contains("pressure"))
          {
            ppgodunov.query("pressure",pressureString);
            if (pressureString == "isentropic")
              {
                pressure = PRESSURE_ISENTROPIC;
              }
            else if (pressureString == "constant")
              {
                pressure = PRESSURE_CONSTANT;
              }

            if (pressure == -1)
              {
                pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
                return;
              }
          }
        else
          {
            pressure = PRESSURE_ISENTROPIC;
          }

        ppgodunov.queryarr("center",centerpp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

        ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

        break;
      }
    
    case PROBLEM_WAVE:
      {
        ppgodunov.query("ambient_density",ambientDensity);
        ppgodunov.query("delta_density",deltaDensity);

        if (ppgodunov.contains("pressure"))
          {
            ppgodunov.query("pressure",pressureString);
            if (pressureString == "isentropic")
              {
                pressure = PRESSURE_ISENTROPIC;
              }
            else if (pressureString == "constant")
              {
                pressure = PRESSURE_CONSTANT;
              }

            if (pressure == -1)
              {
                pout() << "Invalid pressure, \"" << pressureString << "\", specified in input file" << endl << endl; 
                return;
              }
          }
        else
          {
            pressure = PRESSURE_ISENTROPIC;
          }

        ppgodunov.queryarr("wave_number",waveNumberpp,0,SpaceDim);
        int norm2 = 0;
        for (int i = 0; i < SpaceDim; i++)
          {
            waveNumber[i] = waveNumberpp[i];
            norm2 += waveNumber[i]*waveNumber[i];
          }

        if (norm2 == 0)
          {
            pout() << "One component of the wave number must be non-zero" << endl << endl;
            return;
          }

        ppgodunov.queryarr("center",centerpp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) center[i] = centerpp[i];

        ppgodunov.queryarr("velocity",velocitypp,0,SpaceDim);
        for (int i = 0; i < SpaceDim; i++) velocity[i] = velocitypp[i];

        break;
      }
    
    default:
      {
        MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
        break;
      }
    }

  // Stop after this number of steps
  int nstop = 0;
  ppgodunov.query("max_step",nstop);

  // Stop when the simulation time get here
  Real stopTime = 0.0;
  ppgodunov.query("max_time",stopTime);

  // Set the physical size of the longest dimension of the domain
  Real domainLength = 1.0;
  ppgodunov.query("domain_length",domainLength);

  // Set the resolution of the coarsest level 
  vector<int> numCells(SpaceDim);
  for (int i = 0; i < SpaceDim; ++i) numCells[i]=0;
  ppgodunov.queryarr("num_cells",numCells,0,SpaceDim);

  assert(D_TERM(   (numCells[0] > 0),
                && (numCells[1] > 0),
                && (numCells[2] > 0)));
  assert(D_TERM(   (numCells[0] % 2 == 0),
                && (numCells[1] % 2 == 0),
                && (numCells[2] % 2 == 0)));

  // Determine which spatial directions are periodic
  vector<int> isPeriodica(SpaceDim,0);
  bool isPeriodic[SpaceDim];
  
  ppgodunov.queryarr("is_periodic",isPeriodica,0,SpaceDim);
  // convert periodic from int->bool
  for (int dim=0; dim<SpaceDim; dim++) 
    {
      isPeriodic[dim] = (isPeriodica[dim] == 1);
      if (isPeriodic[dim] && verbosity >= 2 && procID() == 0)
        pout() << "Using Periodic BCs in direction: " << dim << endl;
    }

  // Maximum AMR level limit
  int maxLevel = 0;
  ppgodunov.query("max_level",maxLevel);
  int numReadLevels = Max(maxLevel,1);

  // Refinement ratios between levels
  std::vector<int> refRatios;
  // Note: this requires a refRatio to be defined for the finest level
  // (even though it will never be used)
  ppgodunov.queryarr("ref_ratio",refRatios,0,numReadLevels+1);

  // Number of coarse time steps from one regridding to the next
  std::vector<int> regridIntervals;
  ppgodunov.queryarr("regrid_interval",regridIntervals,0,numReadLevels);

  // How far to extend refinement from cells newly tagged for refinement
  int tagBufferSize = 3;
  ppgodunov.query("tag_buffer_size",tagBufferSize);

  // Threshold that triggers refinement
  Real refineThresh = 0.3;
  ppgodunov.query ("refine_thresh",refineThresh);

  // Minimum dimension of a grid
  int blockFactor = 1;
  ppgodunov.query("block_factor",blockFactor);

  // Maximum dimension of a grid
  int maxGridSize = 32;
  ppgodunov.query("max_grid_size",maxGridSize);

  Real fillRatio = 0.75;
  ppgodunov.query("fill_ratio",fillRatio);

  // Do 4th order (1) or 2nd order (0) slope computations
  int inFourthOrderSlopes = 1;
  bool fourthOrderSlopes;
  ppgodunov.query("fourth_order_slopes",inFourthOrderSlopes);
  fourthOrderSlopes = (inFourthOrderSlopes == 1);

  // Do slope flattening - only valid with 4th order slopes
  int inFlattening = 1;
  bool flattening;
  ppgodunov.query("flattening",inFlattening);
  flattening = (inFlattening == 1);

  // Do slope limiting
  int inLimiting = 1;
  bool limiting;
  ppgodunov.query("limiting",inLimiting);
  limiting = (inLimiting == 1);

  assert(fourthOrderSlopes || !flattening);

  // Apply artificial viscosity
  int inUseArtificialViscosity = 1;
  bool useArtificialViscosity;
  ppgodunov.query("use_artificial_viscosity",inUseArtificialViscosity);
  useArtificialViscosity = (inUseArtificialViscosity == 1);

  // Artificial viscosity coefficient/multiplier
  Real artificialViscosity = 0.1;
  ppgodunov.query("artificial_viscosity",artificialViscosity);

  // Set up checkpointing
  int checkpointInterval = 0;
  ppgodunov.query("checkpoint_interval",checkpointInterval);

  // Set up plot file writing
  int plotInterval = 0;
  ppgodunov.query("plot_interval",plotInterval);

  // CFL multiplier
  Real cfl = 0.8;
  ppgodunov.query("cfl",cfl);

  // Initial CFL multiplier
  Real initialCFL = 0.1;
  ppgodunov.query("initial_cfl",initialCFL);

  // Determine if a fixed or variable time step will be used
  Real fixedDt = -1;
  ppgodunov.query("fixed_dt",fixedDt);

  // Limit the time step growth
  Real maxDtGrowth = 1.1;
  ppgodunov.query("max_dt_growth",maxDtGrowth);

  // Let the time step grow by this factor above the "maximum" before
  // reducing it
  Real dtToleranceFactor = 1.1;
  ppgodunov.query("dt_tolerance_factor",dtToleranceFactor);

  // Print the parameters
  if ( verbosity >= 2 )
    {
      pout() << "problem = " << problemString << endl;
      pout() << "gamma = " << gamma << endl;

      // display inputs
      switch (problem)
        {
        case PROBLEM_RAMP:
          {
            pout() << "alpha = " << alpha << endl;
            pout() << "shock mach = " << ms << endl;
            pout() << "xcorner = " << xcorner << endl;
            break;
          }

        case PROBLEM_EXPLOSION:
          {
            pout() << "shock mach = " << ms << endl;
            pout() << "center = " << D_TERM(center[0] << "  " <<,
                                            center[1] << "  " <<,
                                            center[2] << ) endl;
            pout() << "size = " << size << endl;
            pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                              velocity[1] << "  " <<,
                                              velocity[2] << ) endl;
            break;
          }

        case PROBLEM_GAUSSIAN:
          {
            pout() << "ambient density = " << ambientDensity << endl;
            pout() << "delta density = " << deltaDensity << endl;
            pout() << "pressure = " << pressureString << endl;
            pout() << "center = " << D_TERM(center[0] << "  " <<,
                                            center[1] << "  " <<,
                                            center[2] << ) endl;
            pout() << "size = " << size << endl;
            pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                              velocity[1] << "  " <<,
                                              velocity[2] << ) endl;
            break;
          }

        case PROBLEM_SINE:
          {
            pout() << "ambient density = " << ambientDensity << endl;
            pout() << "delta density = " << deltaDensity << endl;
            pout() << "pressure = " << pressureString << endl;
            pout() << "center = " << D_TERM(center[0] << "  " <<,
                                            center[1] << "  " <<,
                                            center[2] << ) endl;
            pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                              velocity[1] << "  " <<,
                                              velocity[2] << ) endl;
            break;
          }
      
        case PROBLEM_WAVE:
          {
            pout() << "ambient density = " << ambientDensity << endl;
            pout() << "delta density = " << deltaDensity << endl;
            pout() << "pressure = " << pressureString << endl;
            pout() << "wave number = " << D_TERM(waveNumber[0] << "  " <<,
                                                 waveNumber[1] << "  " <<,
                                                 waveNumber[2] << ) endl;
            pout() << "center = " << D_TERM(center[0] << "  " <<,
                                            center[1] << "  " <<,
                                            center[2] << ) endl;
            pout() << "velocity = " << D_TERM(velocity[0] << "  " <<,
                                              velocity[1] << "  " <<,
                                              velocity[2] << ) endl;
            break;
          }
      
        default:
          {
            MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
            break;
          }
        }

      pout() << "maximum step = " << nstop << endl;
      pout() << "maximum time = " << stopTime << endl;

      pout() << "number of cells = " << D_TERM(numCells[0] << "  " <<,
                                               numCells[1] << "  " <<,
                                               numCells[2] << ) endl;

      pout() << "maximum level = " << maxLevel << endl;

      pout() << "refinement ratio = ";
      for (int i = 0; i < refRatios.size(); ++i) pout() << refRatios[i] << " ";
      pout() << endl;

      pout() << "regrid interval = ";
      for (int i = 0; i < regridIntervals.size(); ++i) pout() << regridIntervals[i] << " ";
      pout() << endl;

      pout() << "refinement threshold = " << refineThresh << endl;

      pout() << "blocking factor = " << blockFactor << endl;
      pout() << "max grid size = " << maxGridSize << endl;
      pout() << "fill ratio = " << fillRatio << endl;

      pout() << "checkpoint interval = " << checkpointInterval << endl;
      pout() << "plot interval = " << plotInterval << endl;
      pout() << "CFL = " << cfl << endl;
      pout() << "initial CFL = " << initialCFL << endl;
      if (fixedDt > 0)
        {
          pout() << "fixed dt = " << fixedDt << endl;
        }
      pout() << "maximum dt growth = " << maxDtGrowth << endl;
      pout() << "dt tolerance factor = " << dtToleranceFactor << endl;
    }

  // Create initial and boundary condition (IBC) object and initialize
  PhysIBC* ibc;

  switch (problem)
    {
    case PROBLEM_RAMP:
      {
        RampIBC* rampibc = new RampIBC;
        rampibc->setFortranCommon(gamma,
                                  alpha,
                                  ms,
                                  xcorner,
                                  artificialViscosity);
        ibc = rampibc;
        break;
      }

    case PROBLEM_EXPLOSION:
      {
        ExplosionIBC* explosionibc = new ExplosionIBC;
        explosionibc->setFortranCommon(gamma,
                                       ms,
                                       center,
                                       size,
                                       velocity,
                                       artificialViscosity);
        ibc = explosionibc;
        break;
      }

    case PROBLEM_GAUSSIAN:
      {
        GaussianIBC* gaussianibc = new GaussianIBC;
        gaussianibc->setFortranCommon(gamma,
                                      ambientDensity,
                                      deltaDensity,
                                      pressure,
                                      center,
                                      size,
                                      velocity,
                                      artificialViscosity);
        ibc = gaussianibc;
        break;
      }

    case PROBLEM_SINE:
      {
        SineIBC* sineibc = new SineIBC;
        sineibc->setFortranCommon(gamma,
                                  ambientDensity,
                                  deltaDensity,
                                  pressure,
                                  center,
                                  velocity,
                                  artificialViscosity);
        ibc = sineibc;
        break;
      }
    
    case PROBLEM_WAVE:
      {
        WaveIBC* waveibc = new WaveIBC;
        waveibc->setFortranCommon(gamma,
                                  ambientDensity,
                                  deltaDensity,
                                  pressure,
                                  waveNumber,
                                  center,
                                  velocity,
                                  artificialViscosity);
        ibc = waveibc;
        break;
      }
    
    default:
      {
        MayDay::Error("amrGodunov \"problem\" (used for initial and boundary conditions) is invalid");
        break;
      }
    }

  ProblemDomain probDomain (IntVect::Zero,
                            IntVect(D_DECL(numCells[0]-1,
                                           numCells[1]-1,
                                           numCells[2]-1)),
                            isPeriodic);

  AMRLevelPolytropicGasFactory amrGodFact;

  RefCountedPtr<PatchGodunov> patchPolytropicGas = static_cast<PatchGodunov*>
                                                   (new PatchPolytropicGas());

  // Set up the patch integrator for polytropic gas dynamics
  patchPolytropicGas->setPhysIBC(ibc);
  patchPolytropicGas->setSlopeParameters(fourthOrderSlopes,
                                         flattening,
                                         limiting);
  patchPolytropicGas->setArtificialViscosity(useArtificialViscosity,
                                             artificialViscosity);

  // Set up the AMRLevel... factory
  amrGodFact.CFL(cfl);
  amrGodFact.domainLength(domainLength);
  amrGodFact.refinementThreshold(refineThresh);
  amrGodFact.tagBufferSize(tagBufferSize);
  amrGodFact.patchGodunov(patchPolytropicGas);
  amrGodFact.verbosity(verbosity);
  amrGodFact.initialDtMultiplier(initialCFL);

  AMR amr;

  // Set up the AMR object
  amr.define(maxLevel,refRatios,probDomain,&amrGodFact);

  if (fixedDt > 0)
    {
      amr.fixedDt(fixedDt);
    }

  // Set grid generation parameters
  amr.maxGridSize(maxGridSize);
  amr.blockFactor(blockFactor);
  amr.fillRatio(fillRatio);

  // The hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // Set output parameters
  amr.checkpointInterval(checkpointInterval);
  amr.plotInterval(plotInterval);
  amr.regridIntervals(regridIntervals);
  amr.maxDtGrow(maxDtGrowth);
  amr.dtToleranceFactor(dtToleranceFactor);

  // Set up output files
  if (ppgodunov.contains("plot_prefix"))
    {
      std::string prefix;
      ppgodunov.query("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (ppgodunov.contains("chk_prefix"))
    {
      std::string prefix;
      ppgodunov.query("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  // Set up input files
  if (!ppgodunov.contains("restart_file"))
    {
      if (!ppgodunov.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string gridFile;
          ppgodunov.query("fixed_hierarchy",gridFile);

          // initialize from a list of grids in "gridFile"
          Vector<Vector<Box> > amrGrids(maxLevel+1);
          setupFixedGrids(amrGrids,
                          probDomain,
                          maxLevel,
                          maxGridSize,
                          blockFactor,
                          verbosity,
                          gridFile);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restartFile;
      ppgodunov.query("restart_file",restartFile);

#ifdef HDF5
      HDF5Handle handle(restartFile,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // Run the computation
  amr.run(stopTime,nstop);

  // Output the last plot file and statistics
  amr.conclude();
}

// setupFixedGrids allows fixed grids to be read in and used in this AMR
// computation example
void setupFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                     const ProblemDomain&  a_domain,
                     int                   a_maxLevel,
                     int                   a_maxGridSize,
                     int                   a_blockFactor,
                     int                   a_verbosity,
                     std::string           a_gridFile)
{
  // Run this task on one processor
  if (procID() == uniqueProc(SerialTask::compute))
  {
    a_amrGrids.push_back(Vector<Box>(1,a_domain.domainBox()));

    // Read in predefined grids
    ifstream is(a_gridFile.c_str(), ios::in);

    if (is.fail())
    {
      MayDay::Error("Cannot open grids file");
    }
  
    // Format of file:
    //   number of levels, then for each level (starting with level 1):
    //   number of grids on level, list of boxes

    int inNumLevels;
    is >> inNumLevels;

    assert (inNumLevels <= a_maxLevel+1);

    if (a_verbosity >= 3)
    {
      pout() << "numLevels = " << inNumLevels << endl;
    }

    while (is.get() != '\n');

    a_amrGrids.resize(inNumLevels);

    // Check to see if coarsest level needs to be broken up
    domainSplit(a_domain,a_amrGrids[0],a_maxGridSize,a_blockFactor);

    if (a_verbosity >= 3)
    {
      pout() << "level 0: ";
      for (int n = 0; n < a_amrGrids[0].size(); n++)
      {
        pout() << a_amrGrids[0][0] << endl;
      }
    }

    // Now loop over levels, starting with level 1
    int ngrid;
    for (int lev = 1; lev < inNumLevels; lev++)
    {
      is >> ngrid;

      if (a_verbosity >= 3)
      {
        pout() << "level " << lev << " numGrids = " << ngrid << endl;
        pout() << "Grids: ";
      }

      while (is.get() != '\n');

      a_amrGrids[lev].resize(ngrid);

      for (int i = 0; i < ngrid; i++)
      {
        Box bx;
        is >> bx;

        while (is.get() != '\n');

        // Quick check on box size
        Box bxRef(bx);

        if (bxRef.longside() > a_maxGridSize)
        {
          pout() << "Grid " << bx << " too large" << endl;
          MayDay::Error();
        }

        if (a_verbosity >= 3) 
        {
          pout() << bx << endl;
        }

        a_amrGrids[lev][i] = bx;
      } // End loop over boxes on this level
    } // End loop over levels
  }

  // Broadcast results to all the processors
  broadcast(a_amrGrids,uniqueProc(SerialTask::compute));
}
