/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeDotProduct.cpp
// petermc, 11 June 2003

#include "NodeDotProduct.H"
#include "NodeSetOperations.H"
#include "NodeDotProdF_F.H"

// ---------------------------------------------------------
Real
DotProductNodes(const BoxLayoutData<NodeFArrayBox>& a_dataOne,
                const BoxLayoutData<NodeFArrayBox>& a_dataTwo,
                const BoxLayout& a_dblIn)
{
  Interval comps = a_dataOne.interval();
  return DotProductNodes(a_dataOne, a_dataTwo, a_dblIn, comps);
}


// ---------------------------------------------------------
Real
DotProductNodes(const BoxLayoutData<NodeFArrayBox>& a_dataOne,
                const BoxLayoutData<NodeFArrayBox>& a_dataTwo,
                const BoxLayout& a_dblIn,
                const Interval& a_comps)
{
  const int startComp = a_comps.begin();
  const int endComp = a_comps.end();
  assert(a_dataOne.nComp() == a_dataTwo.nComp());
  Real rhodot = 0.0;
  //calculate the single-processor dot product
  DataIterator dit = a_dataOne.dataIterator();
  for (dit.reset(); dit.ok(); ++dit)
    {
      Box fabbox(surroundingNodes(a_dblIn.get(dit())));
      const FArrayBox& onefab = a_dataOne[dit()].getFab();
      const FArrayBox& twofab = a_dataTwo[dit()].getFab();
      assert(onefab.box().contains(fabbox));
      assert(twofab.box().contains(fabbox));

      // Weights for trapezoidal rule.
      FArrayBox weight(fabbox, 1);
      Real dv = 1.;
      FORT_TRAPWEIGHTS(CHF_FRA1(weight, 0),
                       CHF_BOX(fabbox),
                       CHF_CONST_REAL(dv));
                       
      Real dotgrid = 0;
      // petermc, 5 Apr 2002, changed DOTPROD to DOTPRODUCT
      // petermc, 15 Apr 2002 changed parameters
      // petermc, 22 Oct 2002, corrected bug that was looping over
      // all components
      // petermc, 22 Oct 2002, changed to NODEDOTPRODUCT
      FORT_NODEDOTPRODUCT(CHF_REAL(dotgrid),
                          CHF_CONST_FRA(onefab),
                          CHF_CONST_FRA(twofab),
                          CHF_CONST_FRA1(weight, 0),
                          CHF_BOX(fabbox),
                          CHF_CONST_INT(startComp),
                          CHF_CONST_INT(endComp));

      rhodot += dotgrid;
    }

  // now for the multi-processor fandango

  //gather all the rhodots onto a vector and add them up
  int baseProc = 0;
  Vector<Real> dotVec;
  gather(dotVec, rhodot, baseProc);

  Real rhodotTot = 0.0;
  if (procID() == baseProc)
    {
      assert(dotVec.size() == numProc());
      for(int ivec = 0; ivec < dotVec.size(); ivec++)
        {
          rhodotTot += dotVec[ivec];
        }
    }
  //broadcast the sum to all processors.
  broadcast(rhodotTot, baseProc);

  //return the total
  return rhodotTot;
}


// ---------------------------------------------------------
Real
DotProductNodes(const LevelData<NodeFArrayBox>& a_dataOne,
                const LevelData<NodeFArrayBox>& a_dataTwo,
                const Box& a_domain,
                const LayoutData< Vector<IntVectSet> >& a_IVSVext,
                const Interval& a_comps)
{
  ProblemDomain probdomain(a_domain);
  return DotProductNodes(a_dataOne, a_dataTwo, probdomain, a_IVSVext, a_comps);
}


// ---------------------------------------------------------
Real
DotProductNodes(const LevelData<NodeFArrayBox>& a_dataOne,
                const LevelData<NodeFArrayBox>& a_dataTwo,
                const ProblemDomain& a_domain,
                const LayoutData< Vector<IntVectSet> >& a_IVSVext,
                const Interval& a_comps)
{
  // a_dataOne and a_dataTwo must have the same layout.

  // Idea:  copy a_dataOne to tempOne, then zero out tempOne on
  // exterior nodes of grids at this level.
  // Do the same with a_dataTwo to tempTwo.
  int ncomps = a_comps.size();
  const DisjointBoxLayout& grids = a_dataOne.getBoxes();
  LevelData<NodeFArrayBox> tempOne(grids, ncomps);
  LevelData<NodeFArrayBox> tempTwo(grids, ncomps);
  
  // Copy a_dataOne to tempOne, a_dataTwo to tempTwo.
  for (DataIterator dit(grids.dataIterator()); dit.ok(); ++dit)
    {
      tempOne[dit()].copy(a_dataOne[dit()]);
      tempTwo[dit()].copy(a_dataTwo[dit()]);
    }

  // Zero out tempOne and tempTwo on exterior nodes.
  zeroBoundaryNodes(tempOne, a_IVSVext);
  zeroBoundaryNodes(tempTwo, a_IVSVext);

  Real thisdot = DotProductNodes(tempOne, tempTwo, grids, a_comps);
  
  return thisdot;
}


// ---------------------------------------------------------
Real
DotProductNodes(const LevelData<NodeFArrayBox>& a_dataOne,
                const LevelData<NodeFArrayBox>& a_dataTwo,
                const Box& a_domain,
                const Interval& a_comps)
{
  ProblemDomain probdomain(a_domain);
  return DotProductNodes(a_dataOne, a_dataTwo, probdomain, a_comps);
}


// ---------------------------------------------------------
Real
DotProductNodes(const LevelData<NodeFArrayBox>& a_dataOne,
                const LevelData<NodeFArrayBox>& a_dataTwo,
                const ProblemDomain& a_domain,
                const Interval& a_comps)
{
  // a_dataOne and a_dataTwo must have the same layout.

  // Idea:  copy a_dataOne to tempOne, then zero out tempOne on
  // exterior nodes of grids at this level.
  // Do the same with a_dataTwo to tempTwo.
  const DisjointBoxLayout& grids = a_dataOne.getBoxes();

  LayoutData< Vector<IntVectSet> > IVSVint, IVSVext;
  interiorBoundaryNodes(IVSVint, grids, a_domain);
  exteriorBoundaryNodes(IVSVext, IVSVint, grids);

  Real thisdot = DotProductNodes(a_dataOne, a_dataTwo, a_domain,
                                 IVSVext, a_comps);

  return thisdot;
}
