/*  _______              __
   / ___/ /  ___  __ _  / /  ___
  / /__/ _ \/ _ \/  ' \/ _ \/ _ \
  \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// NodeQuadCFInterp.cpp
// petermc, Mon, Apr 23, 2001
// petermc, 1 Nov 2001, redid this so that it uses NodeQuadCFInterp2
#include "NodeQuadCFInterp.H"

// ---------------------------------------------------------
void  
NodeQuadCFInterp::setDefaultValues()
{
  m_isDefined = false;
  m_isBCDefined = false;
  m_verbose = false;
}


// ---------------------------------------------------------
NodeQuadCFInterp::NodeQuadCFInterp()
{
  setDefaultValues();
}


// ---------------------------------------------------------
NodeQuadCFInterp::NodeQuadCFInterp(const DisjointBoxLayout& a_grids,
                                   Real a_dx,
                                   const Box& a_domain,
                                   const LayoutData<NodeCFIVS>* const a_loCFIVS,
                                   const LayoutData<NodeCFIVS>* const a_hiCFIVS,
                                   int a_refToCoarse,
                                   int a_interpolationDegree,
                                   int a_ncomp,
                                   bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  setDefaultValues();
  define(a_grids, a_dx, probdomain,
         a_loCFIVS, a_hiCFIVS,
         a_refToCoarse, a_interpolationDegree, a_ncomp, a_verbose);
}


// ---------------------------------------------------------
NodeQuadCFInterp::NodeQuadCFInterp(const DisjointBoxLayout& a_grids,
                                   Real a_dx,
                                   const ProblemDomain& a_domain,
                                   const LayoutData<NodeCFIVS>* const a_loCFIVS,
                                   const LayoutData<NodeCFIVS>* const a_hiCFIVS,
                                   int a_refToCoarse,
                                   int a_interpolationDegree,
                                   int a_ncomp,
                                   bool a_verbose)
{
  setDefaultValues();
  define(a_grids, a_dx, a_domain,
         a_loCFIVS, a_hiCFIVS,
         a_refToCoarse, a_interpolationDegree, a_ncomp, a_verbose);
}


// ---------------------------------------------------------
NodeQuadCFInterp::~NodeQuadCFInterp() 
{
  clearMemory();
}



// ---------------------------------------------------------
bool 
NodeQuadCFInterp::isDefined() const
{
  return(m_isDefined);
}


// ---------------------------------------------------------
void
NodeQuadCFInterp::define(const DisjointBoxLayout& a_grids,
                         Real a_dx,
                         const Box& a_domain,
                         const LayoutData<NodeCFIVS>* const a_loCFIVS,
                         const LayoutData<NodeCFIVS>* const a_hiCFIVS,
                         int a_refToCoarse,
                         int a_interpolationDegree,
                         int a_ncomp,
                         bool a_verbose)
{
  ProblemDomain probdomain(a_domain);
  define(a_grids, a_dx, probdomain,
         a_loCFIVS, a_hiCFIVS,
         a_refToCoarse, a_interpolationDegree, a_ncomp, a_verbose);
}


// ---------------------------------------------------------
void
NodeQuadCFInterp::define(const DisjointBoxLayout& a_grids,
                         Real a_dx,
                         const ProblemDomain& a_domain,
                         const LayoutData<NodeCFIVS>* const a_loCFIVS,
                         const LayoutData<NodeCFIVS>* const a_hiCFIVS,
                         int a_refToCoarse,
                         int a_interpolationDegree,
                         int a_ncomp,
                         bool a_verbose)
{
  assert(a_refToCoarse >= 2);
  m_grids = a_grids;
  m_dx = a_dx;
  m_verbose = a_verbose;

  // m_coarsenings == log2(a_refToCoarse);
  m_coarsenings = 0;
  for (int interRatio = a_refToCoarse; interRatio >= 2; interRatio /= 2)
    m_coarsenings++;

  m_qcfi2.resize(m_coarsenings);
  m_inter.resize(m_coarsenings-1);
  m_loCFIVScoarser.resize(m_coarsenings-1);
  m_hiCFIVScoarser.resize(m_coarsenings-1);
  // for i=1:m_c-2, m_qcfi2[i] interpolates m_inter[i] from m_inter[i-1].

  if (m_coarsenings == 1)
    {
      m_qcfi2[0] = new NodeQuadCFInterp2(a_grids, a_domain,
                                         a_loCFIVS, a_hiCFIVS,
                                         false, a_interpolationDegree, a_ncomp);
    }
  else
    {
      // if m_coarsenings == 2:
      // m_qcfi2[1] = new NodeQuadCFInterp2(a_grids, true);
      // m_qcfi2[0] = new NodeQuadCFInterp2(a_grids.coarsen(2), false);
      // m_inter[0] = new LevelData(a_grids.coarsen(2));

      ProblemDomain domainLevel(a_domain);
      Real dxLevel = m_dx;
      DisjointBoxLayout gridsInterFine(a_grids);
      DisjointBoxLayout gridsInterCoarse;
      // m_qcfi2[m_coarsenings-1] interpolates from a 2-coarsening of
      // the fine grids onto the fine grids (a_grids).
      // Interpolation is from the interface only.
      m_qcfi2[m_coarsenings-1] =
        new NodeQuadCFInterp2(a_grids, domainLevel,
                              a_loCFIVS, a_hiCFIVS,
                              true, a_interpolationDegree, a_ncomp);
      for (int interlev = m_coarsenings - 2; interlev >= 0; interlev--)
        {
          // m_coarsenings >= 2, so this will be executed at least once.
          coarsen(gridsInterCoarse, gridsInterFine, 2);
          domainLevel.coarsen(2);
          dxLevel *= 2.;
          bool interfaceOnly = (interlev > 0);
          // m_qcfi2[interlev] interpolates from some coarsening of
          // the fine grids onto gridsInterCoarse.
          // Except for the first time (interlev == 0),
          // interpolation is from the interface only.

          // Define objects containing the nodes of the coarse/fine interfaces.
          m_loCFIVScoarser[interlev] = new LayoutData<NodeCFIVS>[SpaceDim];
          m_hiCFIVScoarser[interlev] = new LayoutData<NodeCFIVS>[SpaceDim];
          for (int idir = 0; idir < SpaceDim; idir++)
            {
              LayoutData<NodeCFIVS>& loCFIVS =
                m_loCFIVScoarser[interlev][idir];
              LayoutData<NodeCFIVS>& hiCFIVS =
                m_hiCFIVScoarser[interlev][idir];

              loCFIVS.define(gridsInterCoarse);
              hiCFIVS.define(gridsInterCoarse);
              for (DataIterator dit = gridsInterCoarse.dataIterator(); dit.ok(); ++dit)
                {
                  const Box& bx = gridsInterCoarse.get(dit());
                  loCFIVS[dit()].define(domainLevel, bx, gridsInterCoarse,
                                        idir, Side::Lo);
                  hiCFIVS[dit()].define(domainLevel, bx, gridsInterCoarse,
                                        idir, Side::Hi);
                }
            }

          m_qcfi2[interlev] =
            new NodeQuadCFInterp2(gridsInterCoarse, domainLevel,
                                  m_loCFIVScoarser[interlev],
                                  m_hiCFIVScoarser[interlev],
                                  interfaceOnly, a_interpolationDegree, a_ncomp);

          m_inter[interlev] = 
            new LevelData<NodeFArrayBox>(gridsInterCoarse, a_ncomp, IntVect::Zero);
          gridsInterFine = gridsInterCoarse;
        }
      m_domainPenultimate = domainLevel;
      m_dxPenultimate = dxLevel;
    }

  m_ncomp = a_ncomp;

  m_isDefined = true;
}


// ---------------------------------------------------------
void
NodeQuadCFInterp::clearMemory()
{
  if (isDefined())
    {
      for (int interlev = 0; interlev < m_coarsenings; interlev++)
        {
          delete m_qcfi2[interlev];
        }
      for (int interlev = 0; interlev < m_coarsenings-1; interlev++)
        {
          delete m_inter[interlev];
          delete [] m_loCFIVScoarser[interlev];
          delete [] m_hiCFIVScoarser[interlev];
        }
    }
  m_refToCoarse = -1;
  m_isDefined = false;
}



// ---------------------------------------------------------
void
NodeQuadCFInterp::setDomainNodeBC(const DomainNodeBC& a_dombcIn)
{
  m_dombc = a_dombcIn;
  m_isBCDefined = true;
}



// ---------------------------------------------------------
void
NodeQuadCFInterp::setVerbose(bool a_verbose)
{
  m_verbose = a_verbose;
}



// ---------------------------------------------------------
void
NodeQuadCFInterp::coarseFineInterp(LevelData<NodeFArrayBox>& a_phiFine,
                                   const LevelData<NodeFArrayBox>& a_phiCoarse,
                                   bool a_inhomogeneous)
{
  assert(isDefined());
  assert(a_phiFine.nComp() == m_ncomp);
  assert(a_phiCoarse.nComp() == m_ncomp);

  if (m_coarsenings == 1)
    {
      m_qcfi2[0]->coarseFineInterp(a_phiFine, a_phiCoarse);
    }
  else // m_coarsenings >= 2
    {
      // if m_coarsenings == 2:
      // m_qcfi2[1] = new NodeQuadCFInterp2(a_grids, true);
      // m_qcfi2[0] = new NodeQuadCFInterp2(a_grids.coarsen(2), false);
      // m_inter[0] = new LevelData(a_grids.coarsen(2));
      
      // m_qcfi2[0] interpolates m_inter[0] from a_phiCoarse on all of it.
      // m_qcfi2[1] interpolates a_phiFine from m_inter[0] on interface only.

      m_qcfi2[0]->coarseFineInterp(*m_inter[0], a_phiCoarse);
      m_inter[0]->exchange(m_inter[0]->interval());

      // Set physical boundary conditions on *m_inter[0].  Need m_dombc.
      assert(m_isBCDefined);
      // Set domain and dx for coarsest level refined by 2.
      ProblemDomain domain(m_domainPenultimate);
      Real dx = m_dxPenultimate;
      if (a_inhomogeneous)
        for (DataIterator dit = m_inter[0]->dataIterator(); dit.ok(); ++dit)
          m_dombc.applyInhomogeneousBCs(m_inter[0]->operator[](dit()), domain, dx);
      else
        for (DataIterator dit = m_inter[0]->dataIterator(); dit.ok(); ++dit)
          m_dombc.applyHomogeneousBCs(m_inter[0]->operator[](dit()), domain, dx);


      for (int interlev = 1; interlev < m_coarsenings - 1; interlev++)
        {
          m_qcfi2[interlev]->coarseFineInterp(*m_inter[interlev],
                                              *m_inter[interlev-1]);


          m_inter[interlev]->exchange(m_inter[interlev]->interval());

          domain.refine(2);
          dx *= 0.5;
          if (a_inhomogeneous)
            for (DataIterator dit = m_inter[interlev]->dataIterator(); dit.ok(); ++dit)
              m_dombc.applyInhomogeneousBCs(m_inter[interlev]->operator[](dit()), domain, dx);
          else
            for (DataIterator dit = m_inter[interlev]->dataIterator(); dit.ok(); ++dit)
              m_dombc.applyHomogeneousBCs(m_inter[interlev]->operator[](dit()), domain, dx);
        }
      
      m_qcfi2[m_coarsenings-1]->coarseFineInterp(a_phiFine,
                                                 *m_inter[m_coarsenings-2]);
      // Don't need to set boundary conditions for a_phiFine
      // because the calling function will do that.
    }
}
