/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"

#include "EBExplosionIBCFactory.H"
#include "EBPatchPolytropicFactory.H"

#include "ArrayView.H"
#include "EBPatchPolytropic.H"

#include "EBAMRGodunovFactory.H"
#include "EBAMRGodunov.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "BaseIVFactory.H"
#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "VoFIterator.H"
#include "EBArith.H"
#include "AllRegularService.H"
#include "EBArrayView.H"
#include "EBLevelRedist.H"
#include "RedistStencil.H"
#include "SlabService.H"
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "EBAMRIO.H"
#include "GodunovGeom.H"
 



void amrGodunov(const Box&  coarsestDomain,
                const Real& dx);


void dumpmemoryatexit();

/***************/
/***************/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  char* inFile = NULL;

  if (a_argc > 1)
    {
      inFile = a_argv[1];
    }
  else
    {
      pout() << "Usage: <executable name> <inputfile>" << endl;
      pout() << "No input file specified" << endl;
      return -1;
    }
  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  Box coarsestDomain;
  Real dx;
  // run amrGodunov
  godunovGeometry(coarsestDomain, dx);
  amrGodunov(coarsestDomain, dx);

#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}
/***************/

void amrGodunov(const Box& a_domain, 
                const Real& a_dx)
{
  // read inputs
  ParmParse ppgodunov;

  int verbosity;
  ppgodunov.get("verbosity",verbosity);
  assert(verbosity >= 0);

  Real gamma = 1.4;
  ppgodunov.get("gamma",gamma);


  int nstop = 0;
  ppgodunov.get("max_step",nstop);

  int redistRad = 0;
  ppgodunov.get("redist_radius",redistRad);

  Real stopTime = 0.0;
  ppgodunov.get("max_time",stopTime);

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  int max_level = 0;
  ppgodunov.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> ref_ratios; // (num_read_levels,1);
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  ppgodunov.getarr("ref_ratio",ref_ratios,0,num_read_levels+1);

  std::vector<int> regrid_intervals; // (num_read_levels,1);
  ppgodunov.getarr("regrid_interval",regrid_intervals,0,num_read_levels);

  Real refineThresh = 0.3;
  ppgodunov.get ("refine_thresh",refineThresh);

  int block_factor = 1;
  ppgodunov.get("block_factor",block_factor);

  int tagBufferSize = 3;
  ppgodunov.get("tag_buffer_size",tagBufferSize);

  int max_grid_size = 32;
  ppgodunov.get("max_grid_size",max_grid_size);

  Real fill_ratio = 0.75;
  ppgodunov.get("fill_ratio",fill_ratio);

  int checkpoint_interval = 0;
  ppgodunov.get("checkpoint_interval",checkpoint_interval);

  int plot_interval = 0;
  ppgodunov.get("plot_interval",plot_interval);

  Real cfl = 0.8;
  ppgodunov.get("cfl",cfl);

  Real initialCFL = 0.1;
  ppgodunov.get("initial_cfl",initialCFL);

  Real fixed_dt = -1;
  ppgodunov.get("fixed_dt",fixed_dt);

  Real max_dt_growth = 1.1;
  ppgodunov.get("max_dt_growth",max_dt_growth);

  Real dt_tolerance_factor = 1.1;
  ppgodunov.get("dt_tolerance_factor",dt_tolerance_factor);
  Real domainLength;
  ppgodunov.get("domain_length",domainLength);

  int ifourth, iflatten, iartvisc;
  ppgodunov.get("use_fourth_order_slopes", ifourth);
  ppgodunov.get("use_flattening"         , iflatten);
  ppgodunov.get("use_art_visc"           , iartvisc);
  bool useFourthOrderSlopes = (ifourth  ==1);
  bool useFlattening        = (iflatten ==1);
  bool useArtificialVisc    = (iartvisc ==1);

  ProblemDomain prob_domain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  pout() << "explosion initial conditions" << endl;
  pout() << "solid wall boundary conditions" << endl;

  Real p0, r0, p1, r1;
  ppgodunov.get("explosion_p0", p0);
  ppgodunov.get("explosion_r0", r0);
  ppgodunov.get("explosion_p1", p1);
  ppgodunov.get("explosion_r1", r1);
  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  ppgodunov.getarr("explosion_center",centerpp,0,SpaceDim);
  for (int i = 0; i < SpaceDim; i++) 
    center[i] = centerpp[i];
      
  Real size = 0.25;
  ppgodunov.get("explosion_size",size);
      
  //force these for now.
  bool doSmushing = true;
  bool doRZCoords = true;;
  bool hasSourceTerm = true;;
  bool useLimiting = true;

  EBExplosionIBCFactory bcfactory(gamma, size, p0, r0, p1, r1, center, doRZCoords); 

  //create patch integrator
  EBPatchPolytropicFactory patchGamma(&bcfactory,
                                      gamma, 
                                      useFourthOrderSlopes,
                                      useFlattening,
                                      useArtificialVisc,
                                      useLimiting,
                                      doRZCoords);
  int iusemassredist;
  ppgodunov.get("use_mass_redist", iusemassredist);
  bool useMassRedist    = (iusemassredist ==1);

  
  EBAMRGodunovFactory amrg_fact(initialCFL,
                                cfl,
                                redistRad,
                                domainLength,
                                refineThresh,
                                tagBufferSize,
                                verbosity,
                                useMassRedist,
                                doSmushing,
                                doRZCoords,
                                hasSourceTerm,
                                &patchGamma);
  
  AMR amr;
  amr.define(max_level,ref_ratios,prob_domain,&amrg_fact);

  if (fixed_dt > 0)
    {
      amr.fixedDt(fixed_dt);
    }

  // set grid generation parameters
  amr.maxGridSize(max_grid_size);
  amr.blockFactor(block_factor);
  amr.fillRatio(fill_ratio);

  // the hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // set output parameters
  amr.checkpointInterval(checkpoint_interval);
  amr.plotInterval(plot_interval);
  amr.regridIntervals(regrid_intervals);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  if (ppgodunov.contains("plot_prefix"))
    {
      std::string prefix;
      ppgodunov.get("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (ppgodunov.contains("chk_prefix"))
    {
      std::string prefix;
      ppgodunov.get("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  if (!ppgodunov.contains("restart_file"))
    {
      if (!ppgodunov.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string grid_file;
          ppgodunov.get("fixed_hierarchy",grid_file);

          // initialize from a list of grids in "grid_file"
          Vector<Vector<Box> > amrGrids(max_level+1);
          godunovFixedGrids(amrGrids,
                            prob_domain,
                            max_level,
                            max_grid_size,
                            block_factor,
                            verbosity,
                            grid_file);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restart_file;
      ppgodunov.get("restart_file",restart_file);
      pout() << "restarting from " << restart_file << endl;

#ifdef HDF5
      HDF5Handle handle(restart_file,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // run 
  amr.run(stopTime,nstop);

  // output last pltfile and statistics
  amr.conclude();
}

