/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>

#include "ParmParse.H"
#include "CH_HDF5.H"
#include "parstream.H"

#include "EBExplosionIBCFactory.H"
#include "EBPatchPolytropicFactory.H"

#include "ArrayView.H"
#include "EBPatchPolytropic.H"

#include "EBAMRGodunovFactory.H"
#include "EBAMRGodunov.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "BaseIVFactory.H"
#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "VoFIterator.H"
#include "EBArith.H"
#include "AllRegularService.H"
#include "EBArrayView.H"
#include "EBLevelRedist.H"
#include "RedistStencil.H"
#include "SlabService.H"
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "EBAMRIO.H"
#include "GodunovGeom.H"
#include "UnionIF.H"
#include "PlaneIF.H"
#include "IntersectionIF.H"
#include "ImplicitFunctionBL.H"
#include "ComplementIF.H"
#include "LatheIF.H"
 

IntVect ivdebug(D_DECL(10,54,0));
VolIndex vofdebug(ivdebug, 0);
/***************/

void amrGodunov(const Box& a_domain, 
                const Real& a_domainLength,
                const Real& a_dx)
{
  // read inputs
  ParmParse pp;

  int verbosity;
  pp.get("verbosity",verbosity);
  assert(verbosity >= 0);

  Real gamma = 1.4;
  pp.get("gamma",gamma);


  int nstop = 0;
  pp.get("max_step",nstop);

  int redistRad = 0;
  pp.get("redist_radius",redistRad);

  Real stopTime = 0.0;
  pp.get("max_time",stopTime);

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  int max_level = 0;
  pp.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> ref_ratios; // (num_read_levels,1);
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  pp.getarr("ref_ratio",ref_ratios,0,num_read_levels+1);

  std::vector<int> regrid_intervals; // (num_read_levels,1);
  pp.getarr("regrid_interval",regrid_intervals,0,num_read_levels);

  Real refineThresh = 0.3;
  pp.get ("refine_thresh",refineThresh);

  int block_factor = 1;
  pp.get("block_factor",block_factor);

  int tagBufferSize = 3;
  pp.get("tag_buffer_size",tagBufferSize);

  int max_grid_size = 32;
  pp.get("max_grid_size",max_grid_size);

  Real fill_ratio = 0.75;
  pp.get("fill_ratio",fill_ratio);

  int checkpoint_interval = 0;
  pp.get("checkpoint_interval",checkpoint_interval);

  int plot_interval = 0;
  pp.get("plot_interval",plot_interval);

  Real cfl = 0.8;
  pp.get("cfl",cfl);

  Real initialCFL = 0.1;
  pp.get("initial_cfl",initialCFL);

  Real fixed_dt = -1;
  pp.get("fixed_dt",fixed_dt);

  Real max_dt_growth = 1.1;
  pp.get("max_dt_growth",max_dt_growth);

  Real dt_tolerance_factor = 1.1;
  pp.get("dt_tolerance_factor",dt_tolerance_factor);

  int ifourth, iflatten, iartvisc;
  pp.get("use_fourth_order_slopes", ifourth);
  pp.get("use_flattening"         , iflatten);
  pp.get("use_art_visc"           , iartvisc);
  bool useFourthOrderSlopes = (ifourth  ==1);
  bool useFlattening        = (iflatten ==1);
  bool useArtificialVisc    = (iartvisc ==1);

  ProblemDomain prob_domain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  pout() << "explosion initial conditions" << endl;
  pout() << "solid wall boundary conditions" << endl;

  Real p0, r0, p1, r1;
  pp.get("explosion_p0", p0);
  pp.get("explosion_r0", r0);
  pp.get("explosion_p1", p1);
  pp.get("explosion_r1", r1);
  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  pp.getarr("explosion_center",centerpp,0,SpaceDim);
  for (int i = 0; i < SpaceDim; i++) 
    center[i] = centerpp[i];
      
  Real size = 0.25;
  pp.get("explosion_size",size);
      
  int dosmushing;
  pp.get("do_smushing", dosmushing);
  bool doSmushing = (dosmushing==1);
  //force these for now.
  bool doRZCoords = true;;
  bool hasSourceTerm = true;;
  bool useLimiting = true;

  EBExplosionIBCFactory bcfactory(gamma, size, p0, r0, p1, r1, center, doRZCoords); 

  //create patch integrator
  EBPatchPolytropicFactory patchGamma(&bcfactory,
                                      gamma, 
                                      useFourthOrderSlopes,
                                      useFlattening,
                                      useArtificialVisc,
                                      useLimiting,
                                      doRZCoords);
  int iusemassredist;
  pp.get("use_mass_redist", iusemassredist);
  bool useMassRedist    = (iusemassredist ==1);

  
  EBAMRGodunovFactory amrg_fact(initialCFL,
                                cfl,
                                redistRad,
                                a_domainLength,
                                refineThresh,
                                tagBufferSize,
                                verbosity,
                                useMassRedist,
                                doSmushing,
                                doRZCoords,
                                hasSourceTerm,
                                &patchGamma);
  
  AMR amr;
  amr.define(max_level,ref_ratios,prob_domain,&amrg_fact);

  if (fixed_dt > 0)
    {
      amr.fixedDt(fixed_dt);
    }

  // set grid generation parameters
  amr.maxGridSize(max_grid_size);
  amr.blockFactor(block_factor);
  amr.fillRatio(fill_ratio);

  // the hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // set output parameters
  amr.checkpointInterval(checkpoint_interval);
  amr.plotInterval(plot_interval);
  amr.regridIntervals(regrid_intervals);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  if (pp.contains("plot_prefix"))
    {
      std::string prefix;
      pp.get("plot_prefix",prefix);
      amr.plotPrefix(prefix);
    }

  if (pp.contains("chk_prefix"))
    {
      std::string prefix;
      pp.get("chk_prefix",prefix);
      amr.checkpointPrefix(prefix);
    }

  amr.verbosity(verbosity);

  if (!pp.contains("restart_file"))
    {
      if (!pp.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string grid_file;
          pp.get("fixed_hierarchy",grid_file);

          // initialize from a list of grids in "grid_file"
          Vector<Vector<Box> > amrGrids(max_level+1);
          godunovFixedGrids(amrGrids,
                            prob_domain,
                            max_level,
                            max_grid_size,
                            block_factor,
                            verbosity,
                            grid_file);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restart_file;
      pp.get("restart_file",restart_file);
      pout() << "restarting from " << restart_file << endl;

#ifdef HDF5
      HDF5Handle handle(restart_file,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // run 
  amr.run(stopTime,nstop);

  // output last pltfile and statistics
  amr.conclude();
}
/***************/
UnionIF* makeCrossSection(const Vector<Vector<RealVect> >& a_polygons)
{
  // The final result
  UnionIF* retval;

  // Get the number of polygons and make this inside of the domain
  // the inside of the polygons
  int numPolys = a_polygons.size();
  bool inside = true;

  // A list of all the polygons as implicit functions
  Vector<BaseImplicitFunction*> polytopes;
  polytopes.resize(0);

  // Process each polygon
  for (int p = 0; p < numPolys; p++)
  {
    // All the half planes/spaces used to make a polygon
    Vector<BaseImplicitFunction*> planes;
    planes.resize(0);

    // Get the current polygon (as a vector of points)
    const Vector<RealVect>& polygon = a_polygons[p];

    // Get the number of points in the polygon
    int numPts = polygon.size();

    // Process each pair of points
    for (int n = 0; n < numPts; n++)
    {
      // The normal and point is space used to specify each half plane/space
      RealVect normal(RealVect::Zero);
      RealVect point;

      // Set the normal remembering that the last point connects to the first
      // point.
      normal[0] = -(polygon[(n+1) % numPts][1] - polygon[n][1]);
      normal[1] =  (polygon[(n+1) % numPts][0] - polygon[n][0]);

      point = polygon[n];

      // Generate the appropriate half plane/space (as an implicit function)
      PlaneIF* plane;
      plane = new PlaneIF(normal,point,inside);

      // Save the result
      planes.push_back(plane);
    }

    // Intersect all the half planes/spaces to create an implicit function
    // that represents the polygon
    IntersectionIF* polygonIF = new IntersectionIF(planes);

    polytopes.push_back(polygonIF);
  }

  // Union all the polygon implicit functions to get the implicit function
  // returned
  retval = new UnionIF(polytopes);

  return retval;
}
/***************/
void makeGeometry(Box&      a_coarsestDomain,
                  Real&     a_domainLength,
                  Real&     a_dxCoarsest)
{
  bool     insideRegular;

  BaseImplicitFunction* retval;

  // parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);

  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;

  for (int ivec = 0; ivec < SpaceDim; ivec++)
  {
    if (n_cell[ivec] <= 0)
    {
      pout() << "Bogus number of cells input = " << n_cell[ivec];
      exit(1);
    }

    hi[ivec] = n_cell[ivec] - 1;
  }

  a_coarsestDomain.setSmall(lo);
  a_coarsestDomain.setBig(hi);
  //parse input file
  int max_level = 0;
  std::vector<int> refRatios; // (num_read_levels,1);
  // note this requires a refRatio to be defined for the 
  // finest level (even though it will never be used)

  pp.get("max_level",max_level);
  int num_read_levels = Max(max_level,1);
  pp.getarr("ref_ratio",refRatios,0,num_read_levels+1);

  Box finestDomain = a_coarsestDomain;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      finestDomain.refine(refRatios[ilev]);
    }

  Vector<Real> prob_lo(SpaceDim,0.0);
  Vector<Real> prob_hi(SpaceDim,1.0);

  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.getarr("prob_hi",prob_hi,0,SpaceDim);

  a_domainLength = (prob_hi[0]-prob_lo[0]);
  a_dxCoarsest = (prob_hi[0]-prob_lo[0])/n_cell[0];
  Real fineDx = a_dxCoarsest;
  int ebMaxCoarsen = 0;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      fineDx /= refRatios[ilev];
      ebMaxCoarsen += refRatios[ilev]/2;
    }

  RealVect origin;
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    origin[idir] = prob_lo[idir];
  }

  int numPolygonBody;
  pp.get("num_polygons_in_body", numPolygonBody);
  // Data for polygons making up nozzle
  Vector<Vector<RealVect> > polygonsBody(numPolygonBody);
  std::vector<int> pointsInPolygonBody;
  pp.getarr("points_in_body_polygons",pointsInPolygonBody,0,numPolygonBody);

  for(int ipoly = 0; ipoly < numPolygonBody; ipoly++)
    {
      Vector<RealVect>& thisPolygon = polygonsBody[ipoly];
      thisPolygon.resize(pointsInPolygonBody[ipoly]);
      for(int ipt = 0; ipt < pointsInPolygonBody[ipoly]; ipt++)
        {
          RealVect& thisPt = thisPolygon[ipt];
          std::vector<Real> vecPt(SpaceDim);
          char pointname[100];
          sprintf(pointname,"bodypt%d%d", ipoly,ipt);
          pp.getarr(pointname, vecPt, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              thisPt[idir] = vecPt[idir];
            }
        }
    }

  int numPolygonsPoppet;
  pp.get("num_polygons_in_poppet", numPolygonsPoppet);
  // Data for polygons making up nozzle
  Vector<Vector<RealVect> > polygonsPoppet(numPolygonsPoppet);
  std::vector<int> pointsInPolygonPoppet;
  pp.getarr("points_in_poppet_polygons",pointsInPolygonPoppet,0,numPolygonsPoppet);
  for(int ipoly = 0; ipoly < numPolygonsPoppet; ipoly++)
    {
      Vector<RealVect>& thisPolygon = polygonsPoppet[ipoly];
      thisPolygon.resize(pointsInPolygonPoppet[ipoly]);
      for(int ipt = 0; ipt < pointsInPolygonPoppet[ipoly]; ipt++)
        {
          RealVect& thisPt = thisPolygon[ipt];
          std::vector<Real> vecPt(SpaceDim);
          char pointname[100];
          sprintf(pointname,"poppetpt%d%d", ipoly,ipt);
          pp.getarr(pointname, vecPt, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              thisPt[idir] = vecPt[idir];
            }
        }
    }


  // Make the vector of (convex) polygons (vectors of points) into a union
  // of convex polygons, each made from the intersection of a set of half
  // planes/spaces - all represented by implicit functions.
  Vector<Vector<RealVect> > polygons = polygonsBody;
  polygons.append(polygonsPoppet);

  UnionIF* crossSection = makeCrossSection(polygons);

  ImplicitFunctionBL *myGeom;

  if (SpaceDim == 2)
  {
    // In 2D use "as is"
    ComplementIF insideOut(*crossSection);
    retval = insideOut.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(insideOut);
  }
  else
  {
    // In 3D rotate about the z-axis
    LatheIF implicit(*crossSection,insideRegular);
    retval = implicit.newImplicitFunction();
    myGeom = new ImplicitFunctionBL(implicit);
  }

  GeometryShop workshop(*myGeom);

  // This generates the new EBIS
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(finestDomain,origin,fineDx,workshop);

  delete myGeom;
  delete crossSection;
}

/***************/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  setChomboMPIErrorHandler();
#endif

  char* inFile = NULL;

  if (a_argc > 1)
    {
      inFile = a_argv[1];
    }
  else
    {
      pout() << "Usage: <executable name> <inputfile>" << endl;
      pout() << "No input file specified" << endl;
      return -1;
    }
  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  Box coarsestDomain;
  Real dx, domainLength;
  // run amrGodunov
  makeGeometry(coarsestDomain, domainLength, dx);
  amrGodunov(coarsestDomain, domainLength, dx);

#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

