/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "EBExplosionIBC.H"
#include "EBISLayout.H"
#include "EBLoHiCenter.H"
#include "VoFIterator.H"
#include "EBLGIntegrator.H"
/****************************/
/****************************/
EBExplosionIBC::
EBExplosionIBC(const Real&     a_gamma,
               const Real&     a_size,
               const Real&     a_p0,
               const Real&     a_r0,
               const Real&     a_p1,
               const Real&     a_r1,
               const RealVect& a_center,
               const bool&     a_doRZCoords)
  :EBPhysIBC()
{
  /**/
  FORT_SETEXPLOSION(CHF_CONST_REAL(a_gamma),
                    CHF_CONST_REALVECT(a_center),
                    CHF_CONST_REAL(a_size),
                    CHF_CONST_REAL(a_p0),
                    CHF_CONST_REAL(a_r0),
                    CHF_CONST_REAL(a_p1),
                    CHF_CONST_REAL(a_r1));
  /**/
      
  m_doRZCoords = a_doRZCoords;
  m_gamma    = a_gamma;
  m_p0       = a_p0;
  m_r0       = a_r0;
  m_p1       = a_p1;
  m_r1       = a_r1;
  m_center   = a_center;
  m_size     = a_size;
  m_isFortranCommonSet = true;
  m_isDefined = false;
}
/****************************/
/****************************/
void 
EBExplosionIBC::
setBndrySlopes(EBCellFAB&       a_deltaPrim,
               const EBCellFAB& a_primState,
               const EBISBox&   a_ebisBox,
               const Box&       a_box,
               const int&       a_dir)
{
  // don't want to deal with the periodic case
  assert(!m_domain.isPeriodic(a_dir));
  assert(m_isDefined);
  assert(m_isFortranCommonSet);

  Box loBox,hiBox,centerBox,domain;
  int hasLo,hasHi;
  Box slopeBox = a_deltaPrim.getRegion()&m_domain;
  int numSlop = a_deltaPrim.nComp();
  // Generate the domain boundary boxes, loBox and hiBox, if there are
  // domain boundarys there
  loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
             slopeBox,m_domain,a_dir);

  // Set the boundary slopes if necessary
  if ((hasLo != 0) || (hasHi != 0))
    {
      BaseFab<Real>& regDeltaPrim = a_deltaPrim.getSingleValuedFAB();
      const BaseFab<Real>& regPrimState = a_primState.getSingleValuedFAB();
      /**/
      FORT_SLOPEBCS(CHF_FRA(regDeltaPrim),
                    CHF_CONST_FRA(regPrimState),
                    CHF_CONST_REAL(m_dx),
                    CHF_CONST_INT(a_dir),
                    CHF_BOX(loBox),
                    CHF_CONST_INT(hasLo),
                    CHF_BOX(hiBox),
                    CHF_CONST_INT(hasHi));
      /**/
    }
  for(SideIterator sit; sit.ok(); ++sit)
    {
      bool doThisSide;
      Box thisBox;
      if(sit() == Side::Lo)
        {
          doThisSide = (hasLo != 0);
          thisBox = loBox;
        }
      else
        {
          doThisSide = (hasHi != 0);
          thisBox = hiBox;
        }
      if(doThisSide)
        {

          // the cells for which the regular calculation
          //are incorrect are the grown set of the multivalued
          //cells intersected with the appropriate bndry box
          //and added to the irregular cells on the boundary.
          Box boxGrown = thisBox;
          boxGrown.grow(a_dir, 1);
          boxGrown &= m_domain;
          IntVectSet ivs = a_ebisBox.getMultiCells(boxGrown);
          ivs &= loBox;
          IntVectSet ivsIrreg = a_ebisBox.getIrregIVS(thisBox);
          ivs |= ivsIrreg;
          for(VoFIterator vofit(ivs, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
            {
              const VolIndex& vof = vofit();
              //all slopes at boundary get set to zero 
              //except normal velocity.  just following the
              //fortran here
              int inormVelVar = a_dir + QVELX;
              for(int ivar = 0; ivar < numSlop; ivar++)
                {
                  if(ivar != inormVelVar)
                    {
                      a_deltaPrim(vof, ivar) = 0.0;
                    }
                }
              
              //for normal velocity
              //do strangely limited slope
              //just lifted from the fortran.
              Side::LoHiSide otherSide = flip(sit());
              Vector<FaceIndex> faces = 
                a_ebisBox.getFaces(vof, a_dir, otherSide);

              Real thisVel = a_primState(vof, inormVelVar);
              Real otherVel = 0.;
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  VolIndex otherVoF = faces[iface].getVoF(otherSide);
                  otherVel += a_primState(otherVoF,inormVelVar);
                }
              if(faces.size() > 0)
                {
                  otherVel /= Real(faces.size());
                  Real slope;
                  if(sit() == Side::Lo)
                    {
                      slope = otherVel - thisVel;
                    }
                  else
                    {
                      slope = thisVel  - otherVel;
                    }
                  //trick to make sure state will not change sign
                  if(slope*thisVel < 0) 
                    {
                      a_deltaPrim(vof, inormVelVar) = 0.0;
                    }
                  else
                    { //slope and vel same sign
                      Real rsign = 1.0;
                      if(thisVel < 0.0)
                        rsign = -1.0;
                      //told you this was odd.  
                      Real dvmin = Min(Abs(slope), Abs(2.*thisVel));
                      a_deltaPrim(vof, inormVelVar) = rsign*dvmin;
                    }
                }
              else //no faces on high side of low boundary vof
                {
                  a_deltaPrim(vof, inormVelVar) = 0.0;
                }
            } //end loop over irregular vofs at boundary 
        }
    } //end loop over boundary sides
}
/****************************/
/****************************/
void 
EBExplosionIBC::
fluxBC(EBFaceFAB&            a_flux,
       const EBCellFAB&      a_primCenter,
       const EBCellFAB&      a_primExtrap,
       const Side::LoHiSide& a_side,
       const Real&           a_time,
       const EBISBox&        a_ebisBox,
       const Box&            a_box,
       const int&            a_dir)
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  assert(!m_domain.isPeriodic(a_dir));

  Box FBox = a_flux.getSingleValuedFAB().box();
  Box cellBox = FBox;
  int numFlux = a_flux.nComp();

  // Determine which side and thus shifting directions
  int isign = sign(a_side);
  cellBox.shiftHalf(a_dir,isign);

  // Is there a domain boundary next to this grid
  if (!m_domain.contains(cellBox))
    {
      cellBox &= m_domain;
      // Find the strip of cells next to the domain boundary
      Box boundaryBox = bdryBox(cellBox, a_dir, a_side, 1);

      // Shift things to all line up correctly
      boundaryBox.shiftHalf(a_dir,-isign);
      BaseFab<Real>& regFlux = a_flux.getSingleValuedFAB();
      const BaseFab<Real>& regPrimExtrap = a_primExtrap.getSingleValuedFAB();
      regFlux.shiftHalf(a_dir,-isign);

      // Set the boundary fluxes
      /**/
      if(m_doRZCoords)
        {
          FORT_SOLIDBCRZ(CHF_FRA(regFlux),
                         CHF_CONST_FRA(regPrimExtrap),
                         CHF_CONST_INT(isign),
                         CHF_CONST_REAL(m_dx),
                         CHF_CONST_INT(a_dir),
                         CHF_BOX(boundaryBox));
        }
      else
        {
          FORT_SOLIDBC(CHF_FRA(regFlux),
                       CHF_CONST_FRA(regPrimExtrap),
                       CHF_CONST_INT(isign),
                       CHF_CONST_REAL(m_dx),
                       CHF_CONST_INT(a_dir),
                       CHF_BOX(boundaryBox));
        }

      // Shift returned fluxes to be face centered
      regFlux.shiftHalf(a_dir,isign);
      int inormMomVar = CMOMX + a_dir;
      int inormVelVar = QVELX + a_dir;
      //now for the multivalued cells.  Since it is pointwise,
      //the regular calc is correct for all single-valued cells.
      IntVectSet ivs = a_ebisBox.getMultiCells(boundaryBox);
      for(VoFIterator vofit(ivs, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          //set all fluxes except normal momentum  to zero.
          //set normal momemtum flux to sign(normal)*pressure
          Vector<FaceIndex> bndryFaces = a_ebisBox.getFaces(vof, a_dir, a_side);
          for(int iface= 0; iface < bndryFaces.size(); iface++)
            {
              const FaceIndex& face = bndryFaces[iface];
              //set all fluxes to zero then fix momentum flux
              for(int ivar = 0; ivar < numFlux; ivar++)
                {
                  a_flux(face, ivar) = 0;
                }
              Real press = a_primExtrap(vof, QPRES);
              Real dense = a_primExtrap(vof, QRHO);
              Real unorm = a_primExtrap(vof, inormVelVar);
              Real speed = sqrt(m_gamma*press/dense);

              if(m_doRZCoords)
                {
                  if(a_dir == 0)
                    {
                      a_flux(face,  CPRES) = press;
                      //at axis, all non pressure fluxes are zero
                      if(a_side == Side::Hi)
                        {
                          a_flux(face, inormMomVar) = 
                            isign*dense*unorm*speed;
                        }
                    }
                  else
                    {
                      a_flux(face,  CPRES) = press;
                      a_flux(face, inormMomVar) = 
                        isign*dense*unorm*speed;
                    }
                }
              else
                {

                  //blessed cartesian coords.
                  a_flux(face, inormMomVar) = 
                    press + isign*dense*unorm*speed;
                }
            }
        }
    }
}
/****************************/
/****************************/
void 
EBExplosionIBC::define(const ProblemDomain&  a_domain,
                       const Real& a_dx)
{
  m_domain = a_domain;
  m_dx = a_dx;
  m_isDefined = true;
}
/****************************/
/****************************/
void 
EBExplosionIBC::
initialize(LevelData<EBCellFAB>& a_conState,
           const EBISLayout& a_ebisl) const
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  
  // Iterator of all grids in this level
  for(DataIterator dit = a_conState.dataIterator(); dit.ok(); ++dit)
  {
    const EBISBox& ebisBox = a_ebisl[dit()];
    // Storage for current grid
    EBCellFAB& conFAB = a_conState[dit()];

    BaseFab<Real>& regConFAB = conFAB.getSingleValuedFAB();
    // Box of current grid
    Box uBox = regConFAB.box();
    uBox &= m_domain;
    // Set up initial condition in this grid
    /**/
    FORT_EXPLOSIONINIT(CHF_CONST_FRA(regConFAB),
                       CHF_CONST_REAL(m_dx),
                       CHF_BOX(uBox));
    /**/

    //now for the multivalued cells.  Since it is pointwise,
    //the regular calc is correct for all single-valued cells.
    IntVectSet ivs = ebisBox.getMultiCells(uBox);
    for(VoFIterator vofit(ivs, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        const IntVect& iv = vof.gridIndex();
        RealVect momentum;
        Real energy, density;
        /**/
        FORT_POINTEXPLOSIONINIT(CHF_REAL(density),
                                CHF_REALVECT(momentum),
                                CHF_REAL(energy),
                                CHF_CONST_INTVECT(iv),
                                CHF_CONST_REAL(m_dx));
        /**/
        conFAB(vof, CRHO) = density;
        conFAB(vof, CENG) = energy;
        for(int idir = 0; idir < SpaceDim; idir++)
          {
            conFAB(vof, CMOMX+idir) = momentum[idir];
          }
      }//end loop over multivalued cells
  } //end loop over boxes
}
/****************************/
/****************************/
EBExplosionIBC::
~EBExplosionIBC()
{
}
/****************************/
/****************************/
