/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental, 
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
#include "EBDebugOut.H"
#include "DebugOut.H"
#include "EBPatchPolytropic.H"
#include "EBPatchPolytropicF_F.H"
#include "EBPatchGodunovF_F.H"
#include "DatasetClient.H"
#include "EBLGIntegrator.H"
#include "VoFIterator.H"
#include "FaceIterator.H"
#include "EBLoHiCenter.H"
#include "EBPatchGodunov.H"
#include "Stencils.H"
#include "EBArith.H"
#include "EBAMRIO.H"
#include "PolyGeom.H"
#include "ParmParse.H"
#include <cstdio>
/*****************************/
void
EBPatchPolytropic::
doRZCoords(bool a_doRZCoords)
{
  m_doRZCoords = a_doRZCoords;
}
/*****************************/
void 
EBPatchPolytropic::
setSource(EBCellFAB&       a_source,
            const EBCellFAB& a_consState,
            const Box&       a_box)
{
  if(m_doRZCoords)
    {
      setRZSource(a_source,
                  a_consState,
                  a_box);
    }
}
void 
EBPatchPolytropic::
setRZSource(EBCellFAB&       a_source,
            const EBCellFAB& a_consState,
            const Box&       a_box)
{
  a_source.setVal(0.);
  const BaseFab<Real>& regConsState = a_consState.getSingleValuedFAB();
  BaseFab<Real>& regSource          =    a_source.getSingleValuedFAB();
  FORT_SETSOURCERZ(CHF_FRA(regSource),
                   CHF_CONST_FRA(regConsState),
                   CHF_CONST_REAL(m_dx),
                   CHF_BOX(a_box));

  //pointwise operation so just have to do the multi valued cells
  //as an irregular iteration
  IntVectSet multiIVS = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(multiIVS, m_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      Real dense = a_consState(vof, CRHO);
      Real energy = a_consState(vof, CENG);
      RealVect momen;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          momen[idir] = a_consState(vof, CMOMX + idir);
        }
      Real radius = (Real(iv[0]) + 0.5)*m_dx;
      Real densitySource, pressureSource;
      FORT_POINTSETSOURCERZ(CHF_REAL(densitySource),
                            CHF_REAL(pressureSource),
                            CHF_CONST_REAL(dense),
                            CHF_CONST_REALVECT(momen),
                            CHF_CONST_REAL(energy),
                            CHF_CONST_REAL(radius));
      //only the density and energy equations have non-zero source
      //so set the others to zero
      for(int ivar = 0; ivar < numConserved(); ivar++)
        {
          a_source(vof, ivar) =  0.0;
        }
      a_source(vof, QRHO) = densitySource;
      a_source(vof, QPRES) = pressureSource;
    }
}
/*****************************/
void
EBPatchPolytropic::
assembleFluxReg(EBFaceFAB&       a_fluxRegFlux,
                const EBFaceFAB& a_godunovFlux,
                const int&       a_idir, 
                const Box&       a_cellBox)
{
  assert(a_idir==a_fluxRegFlux.direction());
  assert(a_idir==a_godunovFlux.direction());
  assert(a_fluxRegFlux.getCellRegion().contains(a_cellBox));
  assert(a_godunovFlux.getCellRegion().contains(a_cellBox));
  assert(a_godunovFlux.nComp() == numFluxes());
  assert(a_fluxRegFlux.nComp() == numConserved());
  assert(SpaceDim==2);

  int rdir = 0;
  BaseFab<Real>& regFluxRegFlux       = a_fluxRegFlux.getSingleValuedFAB();
  const BaseFab<Real>& regGodunovFlux = a_godunovFlux.getSingleValuedFAB();
  const Box& faceRegion = regGodunovFlux.box();
  assert(regFluxRegFlux.box().contains(faceRegion));

  FORT_FLUXASSEMBLE(CHF_FRA(regFluxRegFlux),
                    CHF_CONST_FRA(regGodunovFlux),
                    CHF_CONST_REAL(m_dx),
                    CHF_CONST_INT(a_idir),
                    CHF_BOX(faceRegion));

  IntVectSet ivsirreg = m_ebisBox.getIrregIVS(a_cellBox);
  FaceStop::WhichFaces facestop = FaceStop::SurroundingWithBoundary;
  for(FaceIterator faceit(ivsirreg, m_ebisBox.getEBGraph(), a_idir, facestop); 
      faceit.ok(); ++faceit)
    {
      const FaceIndex& face = faceit();
      Real rind = face.gridIndex(Side::Hi)[rdir];
      Real faceRad;
      if(a_idir == rdir)
        {
          faceRad = m_dx*rind;
        }
      else
        {
          Real rcent = m_ebisBox.centroid(face)[rdir];
          faceRad = m_dx*(rind + 0.5 + rcent);
        }

      for(int ivar=0; ivar < numConserved(); ivar++)
        {
          Real flux = faceRad*a_godunovFlux(face, ivar);
          //add in pressure term
          if(((ivar == CMOMX) && (a_idir==0)) || 
             ((ivar == CMOMY) && (a_idir==1))) 
            {
              flux += faceRad*a_godunovFlux(face, CPRES);
            }
          a_fluxRegFlux(face, ivar) = flux;
        }
    }
}
/*****************************/
void
EBPatchPolytropic::
assembleFluxIrr(BaseIFFAB<Real>&       a_fluxRegFlux,
                const BaseIFFAB<Real>& a_godunovFlux,
                const int&             a_idir, 
                const Box&             a_cellBox,
                const IntVectSet&      a_set)
{
  int rdir = 0;
  IntVectSet ivsirreg = m_ebisBox.getIrregIVS(a_cellBox);
  FaceStop::WhichFaces facestop = FaceStop::SurroundingWithBoundary;
  for(FaceIterator faceit(ivsirreg, m_ebisBox.getEBGraph(), a_idir, facestop); 
      faceit.ok(); ++faceit)
    {
      const FaceIndex& face = faceit();
      Real rind = face.gridIndex(Side::Hi)[rdir];
      Real faceRad;
      if(a_idir == rdir)
        {
          faceRad = m_dx*rind;
        }
      else
        {
          Real rcent = m_ebisBox.centroid(face)[rdir];
          faceRad = m_dx*(rind + 0.5 + rcent);
        }
      for(int ivar=0; ivar < numConserved(); ivar++)
        {
          Real flux = faceRad*a_godunovFlux(face, ivar);
          //add in pressure term
          if(((ivar == CMOMX) && (a_idir==0)) || 
             ((ivar == CMOMY) && (a_idir==1))) 
            {
              flux += faceRad*a_godunovFlux(face, CPRES);
            }
          a_fluxRegFlux(face, ivar) = flux;
        }
    }
}
/*****************************/
void
EBPatchPolytropic::
nonconservativeDivergenRZ(EBCellFAB&             a_divF, 
                          const EBFluxFAB&       a_flux, 
                          const BaseIVFAB<Real>  a_coveredFluxMinu[SpaceDim], 
                          const BaseIVFAB<Real>  a_coveredFluxPlus[SpaceDim], 
                          const Vector<VolIndex> a_coveredFaceMinu[SpaceDim], 
                          const Vector<VolIndex> a_coveredFacePlus[SpaceDim], 
                          const Box&             a_box)
{
  assert(SpaceDim==2);
  assert(isDefined());
  assert(a_divF.nComp() >= numConserved());
  assert(a_flux.nComp() >= numConserved());
  assert(a_flux.getRegion().contains(a_box));
  assert(a_divF.getRegion().contains(a_box));

  //set the divergence initially to zero
  //then loop through directions and increment the divergence 
  //with each directions flux difference.
  int ncons = numConserved();
  a_divF.setVal(0.0);
  BaseFab<Real>&       regDivF = a_divF.getSingleValuedFAB();
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //update for the regular vofs in the nonconservative 
      //case  works for all single valued vofs.
      const EBFaceFAB& fluxDir = a_flux[idir];
      
      const BaseFab<Real>& regFluxDir = fluxDir.getSingleValuedFAB();

      /* do the regular vofs */
      /**/          
      FORT_DIVERGERZ( CHF_BOX(a_box), 
                      CHF_FRA(regDivF), 
                      CHF_CONST_FRA(regFluxDir), 
                      CHF_CONST_INT(idir), 
                      CHF_CONST_INT(ncons), 
                      CHF_CONST_REAL(m_dx));
      /**/
            
    } // end loop over directions

  //update the irregular vofs
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //divergence was set in regular update.  we reset it
      // to zero and recalc.
      for(int ivar = 0; ivar < ncons; ivar++)
        {
          //r diffs
          Real rupdate = 0.0;
          {
            int rdir = 0;
            const EBFaceFAB& rflux = a_flux[rdir];
            for(SideIterator sit; sit.ok(); ++sit)
              {
                int isign = sign(sit());
                Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, rdir, sit());
                int rindex   =  vof.gridIndex()[rdir];
                Real cellRad = m_dx*(Real(rindex) + 0.5);
                Real faceRad = cellRad + 0.5*isign*m_dx;
                Real faceupdate = 0.0;
                for(int iface = 0; iface < faces.size(); iface++)
                  {
                    const FaceIndex& face = faces[iface];
                    Real flux = rflux(face, ivar);
                    faceupdate += isign*flux*faceRad/(cellRad*m_dx);
                  }
                if(faces.size() > 1)
                  {
                    faceupdate /= Real(faces.size());
                  }
                rupdate += faceupdate;
                //add in pressure terms
                if(ivar == CMOMX)
                  {
                    Real pupdate = 0.0;
                    for(int iface = 0; iface < faces.size(); iface++)
                      {
                        const FaceIndex& face = faces[iface];
                        Real flux = rflux(face, CPRES);
                        pupdate += isign*flux/m_dx;
                      }
                    if(faces.size() > 1)
                      {
                        pupdate /= Real(faces.size());
                      }
                    rupdate += pupdate;
                  }
              }
          }

          //z diffs
          Real zupdate = 0.0;
          {
            int zdir = 1;
            const EBFaceFAB& zflux = a_flux[zdir];
            for(SideIterator sit; sit.ok(); ++sit)
              {
                int isign = sign(sit());
                Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, zdir, sit());
                Real faceupdate = 0.0;
                for(int iface = 0; iface < faces.size(); iface++)
                  {
                    const FaceIndex& face = faces[iface];
                    Real flux = zflux(face, ivar);
                    faceupdate += isign*flux/m_dx;
                  }
                if(faces.size() > 1)
                  {
                    faceupdate /= Real(faces.size());
                  }
                zupdate += faceupdate;
                //add in pressure terms
                if(ivar == CMOMY)
                  {
                    Real pupdate = 0.0;
                    for(int iface = 0; iface < faces.size(); iface++)
                      {
                        const FaceIndex& face = faces[iface];
                        Real flux = zflux(face, CPRES);
                        pupdate += isign*flux/m_dx;
                      }
                    if(faces.size() > 1)
                      {
                        pupdate /= Real(faces.size());
                      }
                    zupdate += pupdate;
                  }
              } //end loop over sides
          }
          a_divF(vof, ivar) = rupdate + zupdate;
        }//end loop over variables
    }//end loop over irreg vofs

  //now correct for the covered fluxes
  for(SideIterator sit; sit.ok(); ++sit)
    {
      int isign = sign(sit());
      const BaseIVFAB<Real>*  rcoveredFluxPtr;
      const Vector<VolIndex>* rcoveredFacePtr;
      const BaseIVFAB<Real>*  zcoveredFluxPtr;
      const Vector<VolIndex>* zcoveredFacePtr;
      int rdir = 0;
      int zdir = 1;
      if(sit() == Side::Lo)
        {
          rcoveredFluxPtr = &a_coveredFluxMinu[rdir];
          rcoveredFacePtr = &a_coveredFaceMinu[rdir];
          zcoveredFluxPtr = &a_coveredFluxMinu[zdir];
          zcoveredFacePtr = &a_coveredFaceMinu[zdir];
        }
      else
        {
          rcoveredFluxPtr = &a_coveredFluxPlus[rdir];
          rcoveredFacePtr = &a_coveredFacePlus[rdir];
          zcoveredFluxPtr = &a_coveredFluxPlus[zdir];
          zcoveredFacePtr = &a_coveredFacePlus[zdir];
        }
      //r update
      {
        const BaseIVFAB<Real>&  rflux = *rcoveredFluxPtr;
        const Vector<VolIndex>& rface = *rcoveredFacePtr;
        for(int ivof = 0; ivof < rface.size(); ivof++)
          {
            const VolIndex& vof = rface[ivof];
            int rindex   =  vof.gridIndex()[rdir];
            Real cellRad = m_dx*(Real(rindex) + 0.5);
            Real faceRad = cellRad + 0.5*isign*m_dx;
          
            if(a_box.contains(vof.gridIndex()))
              {
                //face on this side is covered.  use covered flux.
                for(int ivar = 0; ivar < ncons; ivar++)
                  {
                    Real flux =  rflux(vof, ivar);
                    Real update = isign*flux*faceRad/(cellRad*m_dx);
                    //add in pressure terms
                    if(ivar == CMOMX)
                      {
                        flux = rflux(vof, CPRES);
                        Real pupdate  = isign*flux/m_dx;
                        update += pupdate;
                      }
                    a_divF(vof, ivar) += update;
                  }
              }
          }
      }

      //z update
      {
        const BaseIVFAB<Real>&  zflux = *zcoveredFluxPtr;
        const Vector<VolIndex>& zface = *zcoveredFacePtr;
        for(int ivof = 0; ivof < zface.size(); ivof++)
          {
            const VolIndex& vof = zface[ivof];
            if(a_box.contains(vof.gridIndex()))
              {
                //face on this side is covered.  use covered flux.
                for(int ivar = 0; ivar < ncons; ivar++)
                  {
                    Real flux =  zflux(vof, ivar);
                    Real update = isign*flux/(m_dx);
                    //add in pressure terms
                    if(ivar == CMOMY)
                      {
                        flux = zflux(vof, CPRES);
                        Real pupdate = isign*flux/m_dx;
                        update += pupdate;
                      }
                    a_divF(vof, ivar) += update;
                  }
              }
          }
      }
    }
}
/*****************************/
void
EBPatchPolytropic::
consUndividedDivergenRZ(BaseIVFAB<Real>&       a_divF, 
                        const BaseIFFAB<Real>  a_centroidFlux[SpaceDim], 
                        const BaseIVFAB<Real>& a_ebIrregFlux,
                        const IntVectSet&      a_ivs)
{
  assert(SpaceDim==2);
  assert(isDefined());
  assert(a_divF.nComp() >= numConserved());

  int ncons = numConserved();
  a_divF.setVal(0.0);
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();

      Real cellVol, kvol;
      EBArith::getKVolRZ(kvol, cellVol, m_ebisBox, m_dx, vof);
      assert(kvol >= 0.0);
      assert(kvol <= 1.0);
      //      int rindex   =  vof.gridIndex()[0];
      //     Real cellRad = m_dx*(Real(rindex) + 0.5);
      Real volFrac = m_ebisBox.volFrac(vof);

      Real bndryArea = m_ebisBox.bndryArea(vof);
      Real vofFactor =m_dx*volFrac/(kvol * cellVol);
      
      int rdir = 0;
      for(int ivar = 0; ivar < ncons ; ivar++)
        {
          //do r diffs
          Real rupdate = 0.;
          {
            const BaseIFFAB<Real>& rflux = a_centroidFlux[rdir];
            Real faceupdate = 0.0; 
            for(SideIterator sit; sit.ok(); ++sit)
              {
                int isign = sign(sit());
                Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, rdir, sit());
                for(int iface = 0; iface < faces.size(); iface++)
                  {
                    const FaceIndex& face = faces[iface];
                    Real areaFrac = m_ebisBox.areaFrac(face);
                    Real faceFlux = rflux(face, ivar);
                    int faceInd =  face.gridIndex(Side::Hi)[rdir];
                    Real faceRad    = m_dx*Real(faceInd);
                    Real faceFactor = faceRad*areaFrac;

                    faceupdate += isign*vofFactor*faceFactor*faceFlux;

                  }
              }
            rupdate += faceupdate;
            //lets not forget the pressure
            if(ivar == CMOMX)
              {
                Real pupdate = 0.0;
                for(SideIterator sit; sit.ok(); ++sit)
                  {
                    int isign = sign(sit());
                    Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, rdir, sit());
                    for(int iface = 0; iface < faces.size(); iface++)
                      {
                        const FaceIndex& face = faces[iface];
                        Real pflux = rflux(face, CPRES);
                        Real areaFrac = m_ebisBox.areaFrac(face);
                        pupdate += isign*areaFrac*pflux/m_dx;
                      }
                  }
                rupdate += pupdate;
              }
          }
          //zdiffs look like cartesian 
          Real zupdate = 0.;
          {
            int zdir = 1;
            const BaseIFFAB<Real>& zflux = a_centroidFlux[zdir];
            Real faceupdate = 0.0;
            for(SideIterator sit; sit.ok(); ++sit)
              {
                int isign = sign(sit());
                Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, zdir, sit());
                for(int iface = 0; iface < faces.size(); iface++)
                  {
                    const FaceIndex& face = faces[iface];
                    Real areaFrac = m_ebisBox.areaFrac(face);
                    Real faceFlux = zflux(face, ivar);
                    int faceInd =  face.gridIndex(Side::Hi)[rdir];
                    RealVect faceCent = m_ebisBox.centroid(face);
                    Real faceRad    = m_dx*(Real(faceInd) + faceCent[0]);
                    Real faceFactor = faceRad*areaFrac;

                    faceupdate += isign*vofFactor*faceFactor*faceFlux;
                  }
              }
            zupdate += faceupdate;
            if(ivar == CMOMY)
              {

                Real pupdate = 0.0;
                for(SideIterator sit; sit.ok(); ++sit)
                  {
                    int isign = sign(sit());
                    Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, zdir, sit());
                    for(int iface = 0; iface < faces.size(); iface++)
                      {
                        const FaceIndex& face = faces[iface];
                        Real areaFrac = m_ebisBox.areaFrac(face);
                        Real pflux = zflux(face, CPRES);
                        pupdate += isign*areaFrac*pflux/m_dx;
                      }
                  }
                zupdate += pupdate;
              }
          }
          
          //add in the stinking boundary flux.  
          //only pressure terms here so always diff as if cartesian.
          //the rmomentum term in the ebirregflux fab is left in the rmom
          //slot (so we dont have to futz with the CPRES crap)
          //since the pressure term is all there is.
          Real pflux = a_ebIrregFlux(vof, ivar);
          Real bupdate =  pflux*bndryArea/m_dx;               

          a_divF(vof, ivar) = rupdate + zupdate + bupdate;
        } //end loop over variables
    } //end loop over vofs
}
/*****************************/

/*****************************/
void 
EBPatchPolytropic::getCoveredValuesCons(Vector<Real>& a_covValues)
{
  a_covValues.resize(numConserved());
  a_covValues[CRHO] = 1.0;
  a_covValues[CENG] = 1.0;
  a_covValues[CMOMX] = 0.0;
  a_covValues[CMOMY] = 0.0;
#if CH_SPACEDIM==3
  a_covValues[CMOMZ] = 0.0;
#endif
}
/*****************************/
void 
EBPatchPolytropic::getCoveredValuesPrim(Vector<Real>& a_covValues)
{
  a_covValues.resize(numPrimitives());
  a_covValues[QRHO] = 1.0;
  a_covValues[QVELX] = 0.0;
  a_covValues[QVELY] = 0.0;
  a_covValues[QPRES] = 1.0;
  a_covValues[QENTR] = 0.0;
  a_covValues[QC] = 1.0;
#if CH_SPACEDIM==3
  a_covValues[QVELZ] = 0.0;
#endif
}
/*****************************/
void 
EBPatchPolytropic::
setGamma(const Real& a_gamma)
{
  m_gamma = a_gamma;
  m_isGammaSet = true;
}
/*****************************/
Real
EBPatchPolytropic::
getGamma() const
{
  assert(m_isGammaSet);
  return m_gamma;
}
/******/
bool 
EBPatchPolytropic::
usesFlattening() const
{
  assert(m_isSlopeSet);

  return m_useFlattening;
}
/******/
bool 
EBPatchPolytropic::
usesArtificialViscosity() const
{
  return m_useArtificialVisc;
}
/******/
bool 
EBPatchPolytropic::
usesFourthOrderSlopes() const
{
  assert(m_isSlopeSet);
  return m_useFourthOrderSlopes;
}
/******/
Vector<string>
EBPatchPolytropic::
stateNames()
{
  Vector<string> retval;

  retval.push_back("mass-density");
  if(m_doRZCoords)
    {
      retval.push_back("r-momentum");
      if (SpaceDim >= 2)
        {
          retval.push_back("z-momentum");
        }
      if (SpaceDim >= 3)
        {
          MayDay::Error();
        }
    }
  else
    {
      retval.push_back("x-momentum");
      if (SpaceDim >= 2)
        {
          retval.push_back("y-momentum");
        }

      if (SpaceDim >= 3)
        {
          retval.push_back("z-momentum");
        }
    }
  retval.push_back("energy-density");
  return retval; 
}
/******/
Vector<string>
EBPatchPolytropic::
primNames()
{
  Vector<string> retval;

  retval.push_back("log10density");
  if(m_doRZCoords)
    {
      retval.push_back("r-velocity");
      retval.push_back("z-velocity");
    }
  else
    {
      retval.push_back("x-velocity");
      retval.push_back("y-velocity");
    }

#if CH_SPACEDIM==3
  retval.push_back("z-velocity");
#endif

  retval.push_back("log10pressure");
  retval.push_back("log10entropy");
  retval.push_back("soundspeed");

#ifdef MODIANO_PROBLEM
  retval.push_back("modiano-velocity-axial");
  retval.push_back("modiano-velocity-tangent0");

#if CH_SPACEDIM==3
  retval.push_back("modiano-velocity-tangent1");
#endif
#endif

  return retval; 
}

/******/
Vector<string>
EBPatchPolytropic::
primNamesNoLog()
{
  Vector<string> retval;

  retval.push_back("density");
  if(m_doRZCoords)
    {
      retval.push_back("r-velocity");
      retval.push_back("z-velocity");
    }
  else
    {
      retval.push_back("x-velocity");
      retval.push_back("y-velocity");
    }

#if CH_SPACEDIM==3
  retval.push_back("z-velocity");
#endif

  retval.push_back("pressure");
  retval.push_back("entropy");
  retval.push_back("soundspeed");

#ifdef MODIANO_PROBLEM
  retval.push_back("modiano-velocity-axial");
  retval.push_back("modiano-velocity-tangent0");

#if CH_SPACEDIM==3
  retval.push_back("modiano-velocity-tangent1");
#endif
#endif

  return retval; 
}
/******/
int 
EBPatchPolytropic::
numPrimitives() const
{
  return QNUM;
}
/******/
int 
EBPatchPolytropic::
numFluxes() const
{
  return FNUM;
}
/******/
int 
EBPatchPolytropic::
numConserved() const
{
  return CNUM;
}
/******/
int 
EBPatchPolytropic::
numSlopes() const
{
  return QSLOPE;
}
/******/
Interval 
EBPatchPolytropic::
velocityInterval() const
{
#if CH_SPACEDIM==2
  Interval retval(QVELX, QVELY);
#elif CH_SPACEDIM==3
  Interval retval(QVELX, QVELZ);
#else 
  bogus_spacedim();
#endif
  return retval;
}
/******/
int 
EBPatchPolytropic::
pressureIndex() const
{
  return QPRES;
}
/******/
int 
EBPatchPolytropic::
densityIndex() const
{
  return QRHO;
}
/******/
int 
EBPatchPolytropic::
bulkModulusIndex() const
{
  //phil said this was OK 1-2-2002
  //  MayDay::Warning("returning pressure for bulk modulus");
  return QPRES;
}
/******/
Real 
EBPatchPolytropic::
artificialViscosityCoefficient() const
{
  //phil said the following number is the one always used ---tjl
  //return 0.1;
  ParmParse pp;
  Real retval;
  pp.get("artificial_viscosity", retval);
  return retval;
}
/******/
Real 
EBPatchPolytropic::
getMaxWaveSpeed(const EBCellFAB& a_consState, 
                const Box& a_box)
{
  assert(m_isDefined && m_isBoxSet && m_isGammaSet);
  Real speed = 0.0;
  const EBISBox& ebisBox = a_consState.getEBISBox();
  IntVectSet ivs(a_box);
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  //cannot just send to fortran because covered cell values
  //would get in and multiply-valued cells would use bogus
  //values that could also get in.  The could be avoided by
  //either masks or some clever manipulation of the underlying
  //basefab 
  for(VoFIterator vofit(ivs, ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //      const  IntVect& iv = vof.gridIndex();
      Real dense = Max(smallr, a_consState(vof, CRHO));
      Real eng   = Max(small,  a_consState(vof, CENG));
      Real vmax = 0.0;
      Real kinetic = 0.0;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          Real momen = a_consState(vof, CMOMX + idir);
          Real vel = Abs(momen/dense);
          kinetic += 0.5*vel*vel;
          vmax = Max(vmax, vel);
        }
      Real intEng = eng/dense - kinetic;
      intEng = Max(intEng, 0.001*small);
      Real sqrtarg = m_gamma*(m_gamma -1.0)*intEng;
      assert(sqrtarg >= 0.0);
      Real cspeed = sqrt(sqrtarg);
      speed = Max(speed, cspeed + vmax);
    }
  return speed;
      
}
/******/
void 
EBPatchPolytropic::
consToPrim(EBCellFAB&       a_primState, 
           const EBCellFAB& a_consState,
           const Box&       a_box,
                 int        a_logflag)
{ 
  assert(m_isDefined && m_isBoxSet);
  assert(m_isGammaSet);
  assert(a_consState.getRegion().contains(a_box));
  //have to set the covered cell vals so need the cast
  //does not change real data
  //  EBCellFAB& conData = (EBCellFAB&) a_consState;
  //  setCoveredConsVals(conData);

  const BaseFab<Real>&   regCons = a_consState.getSingleValuedFAB();
  BaseFab<Real>&   regPrim = a_primState.getSingleValuedFAB();

      
  FORT_CONS2PRM(CHF_BOX(a_box), 
                CHF_CONST_FRA(regCons), 
                CHF_FRA(regPrim),
                CHF_CONST_INT(a_logflag));
      
  
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();

      Real dense = a_consState(vof, CRHO);
      Real energy = a_consState(vof, CENG);
      RealVect momen;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          momen[idir] = a_consState(vof, CMOMX + idir);
        }
      Real press, entropy;
      RealVect vel;
      RealVect modianovel;
      Real soundspeed;
      FORT_POINTCONS2PRM(CHF_REAL(dense), 
                         CHF_REALVECT(vel), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy),
                         CHF_REALVECT(modianovel), 
                         CHF_REAL(entropy),
                         CHF_REAL(soundspeed),
                         CHF_CONST_INT(a_logflag));
          
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_primState(vof, QVELX+idir)  =     vel[idir];
#ifdef MODIANO_PROBLEM
          a_primState(vof, QMVAX+idir)=modianovel[idir];
#endif
        }
      a_primState(vof, QRHO) = dense;
      a_primState(vof, QPRES)= press;
      a_primState(vof, QENTR)= entropy;
      a_primState(vof, QC)= soundspeed;
    }
  //  setCoveredPrimVals(a_primState);
}
/******/
void 
EBPatchPolytropic::
consToPrim(BaseIVFAB<Real>&  a_primState, 
           const EBCellFAB&  a_consState,
           const IntVectSet& a_ivs)
{ 
  assert(isDefined());
  int logflag = 0;
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();

      Real dense = a_consState(vof, CRHO);
      Real energy = a_consState(vof, CENG);
      RealVect momen;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          momen[idir] = a_consState(vof, CMOMX + idir);
        }
      Real press, entropy;
      RealVect vel, modianovel;
      Real soundspeed;
      FORT_POINTCONS2PRM(CHF_REAL(dense), 
                         CHF_REALVECT(vel), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy),
                         CHF_REALVECT(modianovel), 
                         CHF_REAL(entropy),
                         CHF_REAL(soundspeed),
                         CHF_CONST_INT(logflag)
                         );
          
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_primState(vof, QVELX+idir)  = vel[idir];
#ifdef MODIANO_PROBLEM
          a_primState(vof, QMVAX+idir)=modianovel[idir];
#endif
        }
      a_primState(vof, QRHO) = dense;
      a_primState(vof, QPRES)= press;
      a_primState(vof, QENTR)= entropy;
      a_primState(vof, QC)= soundspeed;
    }
}
/******/
void 
EBPatchPolytropic::
consToPrim(BaseIVFAB<Real>&  a_primState, 
           const BaseIVFAB<Real>&  a_consState,
           const IntVectSet& a_ivs)
{ 
  assert(isDefined());
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();

      Real dense = a_consState(vof, CRHO);
      Real energy = a_consState(vof, CENG);
      RealVect momen;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          momen[idir] = a_consState(vof, CMOMX + idir);
        }
      Real press, entropy;
      RealVect vel, modianovel;
      Real soundspeed;
      int logflag = 0;
      FORT_POINTCONS2PRM(CHF_REAL(dense), 
                         CHF_REALVECT(vel), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy),
                         CHF_REALVECT(modianovel), 
                         CHF_REAL(entropy),
                         CHF_REAL(soundspeed),
                         CHF_CONST_INT(logflag)
                         );
          
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_primState(vof, QVELX+idir)  = vel[idir];
#ifdef MODIANO_PROBLEM
          a_primState(vof, QMVAX+idir)=modianovel[idir];
#endif
        }
      a_primState(vof, QRHO) = dense;
      a_primState(vof, QPRES)= press;
      a_primState(vof, QENTR)= entropy;
      a_primState(vof, QC)= soundspeed;
    }
}
/******/
void 
EBPatchPolytropic::
primToCons(EBCellFAB&       a_consState, 
           const EBCellFAB& a_primState,
           const Box&       a_box)
{ 
  assert(isDefined());
  const Box& regBox = a_box;
  assert(a_primState.getRegion().contains(regBox));
  //set covered vals. does not change real data.
  //  EBCellFAB& primData = (EBCellFAB&) a_primState;
  //  setCoveredPrimVals(primData);

  BaseFab<Real>&   regCons = a_consState.getSingleValuedFAB();
  const BaseFab<Real>&  regPrim = a_primState.getSingleValuedFAB();

      
  FORT_PRM2CONS(CHF_BOX(regBox), 
                CHF_FRA(regCons), 
                CHF_CONST_FRA(regPrim));
      
  
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(regBox);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();

      Real dense = a_primState(vof, QRHO);
      Real press = a_primState(vof, QPRES);
      RealVect vel;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          vel[idir] = a_primState(vof, QVELX + idir);
        }
      Real energy;
      RealVect momen;
          
      FORT_POINTPRM2CONS(CHF_REAL(dense), 
                         CHF_REALVECT(vel), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy));
          
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_consState(vof, CMOMX+idir)  = momen[idir];
        }
      a_consState(vof, CRHO) = dense;
      a_consState(vof, CENG) = energy;
    }
  //  setCoveredConsVals(a_consState);
}
/******/
void 
EBPatchPolytropic::
primToCons(BaseIVFAB<Real>&       a_consState, 
           const BaseIVFAB<Real>& a_primState,
           const IntVectSet&      a_ivs)
{ 
  assert(isDefined());
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();

      Real dense = a_primState(vof, QRHO);
      Real press = a_primState(vof, QPRES);
      RealVect vel;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          vel[idir] = a_primState(vof, QVELX + idir);
        }
      Real energy;
      RealVect momen;
          
      FORT_POINTPRM2CONS(CHF_REAL(dense), 
                         CHF_REALVECT(vel), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy));
          
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_consState(vof, CMOMX+idir)  = momen[idir];
        }
      a_consState(vof, CRHO) = dense;
      a_consState(vof, CENG) = energy;
    }
}
/*****************************/
void EBPatchPolytropic:: 
normalPred(EBCellFAB&       a_primLo, 
           EBCellFAB&       a_primHi, 
           const EBCellFAB& a_primState, 
           const EBCellFAB& a_slopePrim, 
           const Real&      a_dtbydx, 
           const int&       a_dir, 
           const Box&       a_box)
{
  int nslope =  numSlopes();
  assert(isDefined());
  assert(a_primLo.nComp()    >= nslope);
  assert(a_primHi.nComp()    >= nslope);
  assert(a_primState.nComp() >= nslope);
  assert(a_slopePrim.nComp() == nslope);

  Real dtbydx = a_dtbydx;
  const BaseFab<Real>& regState = a_primState.getSingleValuedFAB();
  const BaseFab<Real>& regSlope = a_slopePrim.getSingleValuedFAB();
  BaseFab<Real>& regPrimLo = a_primLo.getSingleValuedFAB();
  BaseFab<Real>& regPrimHi = a_primHi.getSingleValuedFAB();

      
  FORT_PRED(CHF_BOX(a_box), 
            CHF_CONST_FRA(regState), 
            CHF_CONST_FRA(regSlope), 
            CHF_FRA(regPrimLo), 
            CHF_FRA(regPrimHi), 
            CHF_CONST_INT(a_dir), 
            CHF_CONST_REAL(dtbydx));

  IntVectSet ivsMulti = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();
      RealVect veloc, vello, velhi, velslope;
      Real dense, denlo, denhi, denslope;
      Real press, prelo, prehi, preslope;

      dense    = a_primState(vof, QRHO);
      denslope = a_slopePrim(vof, QRHO);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          veloc[idir]    = a_primState(vof, QVELX + idir);
          velslope[idir] = a_slopePrim(vof, QVELX + idir);
        }
      press    = a_primState(vof, QPRES);
      preslope = a_slopePrim(vof, QPRES);

          
      FORT_POINTPRED(CHF_REAL(dense), 
                     CHF_REAL(denlo), 
                     CHF_REAL(denhi), 
                     CHF_REAL(denslope), 
                     CHF_REALVECT(veloc), 
                     CHF_REALVECT(vello), 
                     CHF_REALVECT(velhi), 
                     CHF_REALVECT(velslope), 
                     CHF_REAL(press), 
                     CHF_REAL(prelo), 
                     CHF_REAL(prehi), 
                     CHF_REAL(preslope), 
                     CHF_CONST_INT(a_dir), 
                     CHF_CONST_REAL(dtbydx));
          
      a_primLo(vof, QRHO) = denlo;
      a_primHi(vof, QRHO) = denhi;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_primLo(vof, QVELX + idir) = vello[idir];
          a_primHi(vof, QVELX + idir) = velhi[idir];
        }
      a_primLo(vof, QPRES) = prelo;
      a_primHi(vof, QPRES) = prehi;

    }
}
/******/
void 
EBPatchPolytropic::
riemann(EBFaceFAB&       a_flux, 
        const EBCellFAB& a_primLeft, 
        const EBCellFAB& a_primRigh, 
        const int&       a_dir, 
        const Box&       a_box)
{
  int nPrim = numPrimitives();
  Box cellBox = enclosedCells(a_box);
  EBFaceFAB primGdnv(m_ebisBox, cellBox, a_dir, nPrim);
  
  //explicitly cast the left and right states to modifiable references
  //because we need to shift them to faces and then shift them back.
  BaseFab<Real>& regPrimRigh = (BaseFab<Real>&)a_primRigh.getSingleValuedFAB();
  BaseFab<Real>& regPrimLeft = (BaseFab<Real>&)a_primLeft.getSingleValuedFAB();
  BaseFab<Real>& regPrimGdnv = primGdnv.getSingleValuedFAB();
  //shift the cell centered stuff to edges
  regPrimRigh.shiftHalf(a_dir, -1);
  regPrimLeft.shiftHalf(a_dir,  1);

  assert(regPrimRigh.box().contains(a_box));
  assert(regPrimLeft.box().contains(a_box));
  assert(regPrimGdnv.box().contains(a_box));
  
  //find the regular part of qgdnv.  
      
  FORT_RIEMANN(CHF_BOX(a_box), 
               CHF_CONST_FRA(regPrimLeft), 
               CHF_CONST_FRA(regPrimRigh), 
               CHF_FRA(regPrimGdnv), 
               CHF_CONST_INT(a_dir));
      

  //shift the cell centered stuff back to cells
  regPrimRigh.shiftHalf(a_dir,  1);
  regPrimLeft.shiftHalf(a_dir, -1);

  //the box sent into this is face-centered.
  //we need to use the cell-centered one it surrounds.
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(cellBox);
  FaceStop::WhichFaces stopCrit = FaceStop::SurroundingNoBoundary;
  for(FaceIterator faceit(ivsMulti, m_ebisBox.getEBGraph(), a_dir, stopCrit);
      faceit.ok(); ++faceit)
    { 
      const FaceIndex& face = faceit();
      VolIndex vofl = face.getVoF(Side::Lo);
      VolIndex vofr = face.getVoF(Side::Hi);

      RealVect velocl, velocr, velgod;
      Real     densel, denser, dengod;
      Real     pressl, pressr, pregod;

      densel = a_primLeft(vofl, QRHO);
      denser = a_primRigh(vofr, QRHO);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velocl[idir] = a_primLeft(vofl, QVELX + idir);
          velocr[idir] = a_primRigh(vofr, QVELX + idir);
        }
      pressl = a_primLeft(vofl, QPRES);
      pressr = a_primRigh(vofr, QPRES);

          
      FORT_POINTRIEMANN(CHF_REAL(densel),   
                        CHF_REAL(denser),   
                        CHF_REAL(dengod),   
                        CHF_REALVECT(velocl), 
                        CHF_REALVECT(velocr), 
                        CHF_REALVECT(velgod), 
                        CHF_REAL(pressl),   
                        CHF_REAL(pressr),   
                        CHF_REAL(pregod),   
                        CHF_CONST_INT(a_dir));
          

      primGdnv(face, QRHO) = dengod;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          primGdnv(face, QVELX + idir) = velgod[idir];
        }
      primGdnv(face, QPRES) = pregod;
    }
  //from the godunov state, get the flux

  getFlux(a_flux, primGdnv, a_dir, a_box);
}
/*****************************/
void 
EBPatchPolytropic::
riemann(BaseIVFAB<Real>&        a_coveredFlux, 
        const BaseIVFAB<Real>&  a_exteState, 
        const EBCellFAB&        a_primState, 
        const Vector<VolIndex>& a_vofset, 
        const int&              a_dir, 
        const Side::LoHiSide&   a_sd)
{
  for(int ivof = 0; ivof < a_vofset.size(); ivof++)
    {
      const VolIndex& vof = a_vofset[ivof];
      RealVect velocl, velocr, velgod, velext, velvof;
      Real     densel, denser, dengod, denext, denvof;
      Real     pressl, pressr, pregod, preext, prevof;

      denvof = a_primState(vof, QRHO);
      denext = a_exteState(vof, QRHO);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velvof[idir] = a_primState(vof, QVELX + idir);
          velext[idir] = a_exteState(vof, QVELX + idir);
        }
      prevof = a_primState(vof, QPRES);
      preext = a_exteState(vof, QPRES);
      if(a_sd == Side::Hi)
        {
          densel = denvof ;
          denser = denext ;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              velocl[idir] = velvof[idir];
              velocr[idir] = velext[idir];
            }
          pressl = prevof;
          pressr = preext;
        }
      else
        {
          denser = denvof ;
          densel = denext ;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              velocr[idir] = velvof[idir];
              velocl[idir] = velext[idir];
            }
          pressr = prevof;
          pressl = preext;
        }

          
      FORT_POINTRIEMANN(CHF_REAL(densel),   
                        CHF_REAL(denser),   
                        CHF_REAL(dengod),   
                        CHF_REALVECT(velocl), 
                        CHF_REALVECT(velocr), 
                        CHF_REALVECT(velgod), 
                        CHF_REAL(pressl),   
                        CHF_REAL(pressr),   
                        CHF_REAL(pregod),   
                        CHF_CONST_INT(a_dir));
          

      RealVect momenflux;
      Real     denseflux;
      Real     energflux;

      int  iloc = vof.gridIndex()[0];
      //only takes effect if idir == 0
      if(!m_doRZCoords)
        {
          FORT_POINTGETFLUX(CHF_REAL(dengod),      
                              CHF_REAL(pregod),      
                              CHF_REALVECT(velgod),  
                              CHF_REAL(denseflux),    
                              CHF_REAL(energflux),    
                              CHF_REALVECT(momenflux), 
                            CHF_CONST_INT(a_dir));
        }
      else
        {
          Real radius = iloc*m_dx;
          FORT_POINTGETFLUXRZ(CHF_REAL(dengod),      
                              CHF_REAL(pregod),      
                              CHF_REALVECT(velgod),  
                              CHF_REAL(denseflux),    
                              CHF_REAL(energflux),    
                              CHF_REALVECT(momenflux), 
                              CHF_CONST_INT(a_dir),
                              CHF_CONST_REAL(radius));
        }

      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_coveredFlux(vof, CMOMX + idir) = momenflux[idir];
        }
      a_coveredFlux(vof, CRHO) = denseflux;
      a_coveredFlux(vof, CENG) = energflux;
      
      if(m_doRZCoords)
        {
          a_coveredFlux(vof, CPRES) = pregod;
        }
    }
}
/******/
void 
EBPatchPolytropic::
getFlux(EBFaceFAB&       a_flux, 
        const EBFaceFAB& a_prim, 
        const int&       a_dir, 
        const Box&       a_box)
{
  assert(a_flux.direction() == a_dir);
  assert(a_flux.getRegion().contains(a_box)); 
  assert(a_prim.getRegion().contains(a_box)); 
  assert(a_prim.direction() == a_dir);
  
  const BaseFab<Real>& regPrim = a_prim.getSingleValuedFAB();
  BaseFab<Real>& regFlux = a_flux.getSingleValuedFAB();

  if(!m_doRZCoords)
    {
      FORT_GETFLUX(CHF_BOX(a_box), 
                   CHF_CONST_FRA(regPrim), 
                   CHF_CONST_INT(a_dir), 
                   CHF_FRA(regFlux),
                   CHF_CONST_REAL(m_dx)); 
    }
  else
    {
      FORT_GETFLUXRZ(CHF_BOX(a_box), 
                     CHF_CONST_FRA(regPrim), 
                     CHF_CONST_INT(a_dir), 
                     CHF_FRA(regFlux),
                     CHF_CONST_REAL(m_dx)); 
    }
      
  //box sent in is  face-centered
  Box cellBox = enclosedCells(a_box);
  cellBox.grow(a_dir, 1);
  cellBox &= m_domain;
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(cellBox);
  FaceStop::WhichFaces stopCrit = FaceStop::SurroundingNoBoundary;
  for(FaceIterator faceit(ivsMulti, m_ebisBox.getEBGraph(), a_dir, stopCrit);
      faceit.ok(); ++faceit)
    { 
      const FaceIndex& face = faceit();
      if(a_box.contains(face.gridIndex(Side::Hi)))
        {

          RealVect velgdnv, momenflux;
          Real     dengdnv, denseflux;
          Real     pregdnv, energflux;

          for(int idir = 0; idir < SpaceDim; idir++)
            {
              velgdnv[idir] = a_prim(face, QVELX + idir);
            }
          dengdnv = a_prim(face, QRHO);
          pregdnv = a_prim(face, QPRES);

          int iloc = face.gridIndex(Side::Hi)[0];
          if(!m_doRZCoords)
            {
              FORT_POINTGETFLUX(CHF_REAL(dengdnv),      
                                CHF_REAL(pregdnv),      
                                CHF_REALVECT(velgdnv),  
                                CHF_REAL(denseflux),    
                                CHF_REAL(energflux),    
                                CHF_REALVECT(momenflux), 
                                CHF_CONST_INT(a_dir));
            }
          else
            {
              Real radius = iloc*m_dx;
              FORT_POINTGETFLUXRZ(CHF_REAL(dengdnv),      
                                CHF_REAL(pregdnv),      
                                CHF_REALVECT(velgdnv),  
                                CHF_REAL(denseflux),    
                                CHF_REAL(energflux),    
                                CHF_REALVECT(momenflux), 
                                CHF_CONST_INT(a_dir),
                                CHF_CONST_REAL(radius));
            }
          

          for(int idir = 0; idir < SpaceDim; idir++)
            {
              a_flux(face, CMOMX + idir) = momenflux[idir];
            }
          a_flux(face, CRHO) = denseflux;
          a_flux(face, CENG) = energflux;
          if(m_doRZCoords)
            {
              a_flux(face, CPRES) = pregdnv;
            }
        }
    }
}
/******/
void 
EBPatchPolytropic::
updateCons(EBCellFAB&            a_consState, 
           const EBFaceFAB&        a_flux, 
           const BaseIVFAB<Real>&  a_coveredFluxMinu, 
           const BaseIVFAB<Real>&  a_coveredFluxPlus, 
           const Vector<VolIndex>& a_coveredFaceMinu, 
           const Vector<VolIndex>& a_coveredFacePlus, 
           const int&              a_dir, 
           const Box&              a_box, 
           const Real&             a_scale)           
{
  if(m_doRZCoords)
    {
      updateConsRZ( a_consState,       
                    a_flux,            
                    a_coveredFluxMinu, 
                    a_coveredFluxPlus, 
                    a_coveredFaceMinu, 
                    a_coveredFacePlus, 
                    a_dir,             
                    a_box,             
                    a_scale);          
    } 
  else
    {
      EBPatchGodunov::updateCons( a_consState,       
                                  a_flux,            
                                  a_coveredFluxMinu, 
                                  a_coveredFluxPlus, 
                                  a_coveredFaceMinu, 
                                  a_coveredFacePlus, 
                                  a_dir,             
                                  a_box,             
                                  a_scale);           
    }

}
void 
EBPatchPolytropic::
updateConsRZ(EBCellFAB&              a_consState, 
             const EBFaceFAB&        a_flux, 
             const BaseIVFAB<Real>&  a_coveredFluxMinu, 
             const BaseIVFAB<Real>&  a_coveredFluxPlus, 
             const Vector<VolIndex>& a_coveredFaceMinu, 
             const Vector<VolIndex>& a_coveredFacePlus, 
             const int&              a_dir, 
             const Box&              a_box, 
             const Real&             a_scale) 
{
  assert(m_doRZCoords);
  assert(isDefined());
  assert(a_consState.nComp() <= a_flux.nComp());
  assert(a_consState.nComp() == numConserved() );
  assert((a_dir >= 0) && (a_dir < SpaceDim));
  assert(a_flux.direction() == a_dir);
  assert(a_flux.getRegion().contains(surroundingNodes(a_box, a_dir)));
  assert(a_consState.getRegion().contains(a_box));

  //update the regular vofs.  
  //first we have to copy the original state 
  //so that the irregular stuff can be updated later.
  EBCellFAB inputConsState(m_ebisBox, 
                           a_consState.getRegion(), 
                           a_consState.nComp());
  //  setCoveredConsVals(a_consState);
  Interval wholeInterv(0, a_consState.nComp()-1);
  inputConsState.copy(a_consState.getRegion(),  wholeInterv, 
                      a_consState.getRegion(),  a_consState, wholeInterv);

  int ncons = numConserved();
  BaseFab<Real>&       consReg = a_consState.getSingleValuedFAB();
  const BaseFab<Real>& fluxReg = a_flux.getSingleValuedFAB();


  FORT_UPDATERZ( CHF_BOX(a_box), 
                 CHF_FRA(consReg), 
                 CHF_CONST_FRA(fluxReg), 
                 CHF_CONST_INT(a_dir), 
                 CHF_CONST_INT(ncons), 
                 CHF_CONST_REAL(a_scale));

  //update the irregular vofs
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //      const IntVect& iv = vof.gridIndex();
      for(int ivar = 0; ivar < a_consState.nComp(); ivar++)
        {
          Real update = 0.0;
          //state changed in regular update. 
          //need to overwrite this.
          Real originalState= inputConsState(vof, ivar);
          for(SideIterator sit; sit.ok(); ++sit)
            {
              int isign = sign(sit());
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, a_dir, sit());
              //if there are faces, use the flux at them
              Real aggFlux = 0.0;
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  Real faceFlux = a_flux(face, ivar);
                  if(m_doRZCoords)
                    {
                      //add in pressure term if we have taken it out
                      if((ivar==CMOMX) && (a_dir==0) || 
                         (ivar==CMOMY) && (a_dir==1))
                        {
                          faceFlux += a_flux(face,CPRES);
                        }
                    }
                  aggFlux += faceFlux;
                }
              if(faces.size() > 1)
                aggFlux /= Real(faces.size());

              update += isign*aggFlux;
            }
          //dx is already divided out (part of scale)
          update *= -a_scale;
          a_consState(vof, ivar) = originalState + update;
        }
    }
  //add in covered face fluxes
  for(SideIterator sit; sit.ok(); ++sit)
    {
      int isign = sign(sit());
      const BaseIVFAB<Real>*  coveredFluxPtr;
      const Vector<VolIndex>* coveredFacePtr;
      if(sit() == Side::Lo)
        {
          coveredFluxPtr = &(a_coveredFluxMinu);
          coveredFacePtr = &(a_coveredFaceMinu);
        }                                     
      else                                    
        {                                     
          coveredFluxPtr = &(a_coveredFluxPlus);
          coveredFacePtr = &(a_coveredFacePlus);
        }
      const BaseIVFAB<Real>&  coveredFlux = *coveredFluxPtr;
      const Vector<VolIndex>& coveredFace = *coveredFacePtr;
      for(int ivof = 0; ivof < coveredFace.size(); ivof++)
        {
          const VolIndex& vof = coveredFace[ivof];
          //          const IntVect& iv = vof.gridIndex();
          if(a_box.contains(vof.gridIndex()))
            {
              for(int ivar = 0; ivar < a_consState.nComp(); ivar++)
                {
                  Real covFlux = coveredFlux(vof, ivar);
                  if(m_doRZCoords)
                    {
                      //add in pressure term if we have taken it out
                      if((ivar==CMOMX) && (a_dir==0) || 
                         (ivar==CMOMY) && (a_dir==1))
                        {
                          covFlux += coveredFlux(vof, CPRES);
                        }
                    }
                  Real update =   isign*covFlux;
                  //dx is already divided out (part of scale)
                  update *= -a_scale;
                  Real state = a_consState(vof, ivar);
                  a_consState(vof, ivar) = state + update;
                }
            }
        }
    }

  floorConserved(a_consState, a_box);
}
/******/
void 
EBPatchPolytropic::floorPrimitives(EBCellFAB&  a_primState,
                                   const Box&  a_box)
{
  BaseFab<Real>&       primReg = a_primState.getSingleValuedFAB();
      
  FORT_FLOORPRIM( CHF_BOX(a_box), 
                  CHF_FRA(primReg));
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      a_primState(vof, QRHO)  = Max(a_primState(vof, QRHO) ,  smallr);
      a_primState(vof, QPRES) = Max(a_primState(vof, QPRES),  smallp);
    }
  
}
/******/
void 
EBPatchPolytropic::floorConserved(EBCellFAB&  a_consState,
                                  const Box&  a_box)
{
  BaseFab<Real>&       consReg = a_consState.getSingleValuedFAB();
      
  FORT_FLOORCONS( CHF_BOX(a_box), 
                  CHF_FRA(consReg));
  
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      a_consState(vof, CRHO) = Max(a_consState(vof, CRHO), smallr);
      a_consState(vof, CENG) = Max(a_consState(vof, CENG), small );
//      Real rmom =       a_consState(vof, CMOMX) ;
//      Real zmom =       a_consState(vof, CMOMY) ;
//      Real ratio = Abs(rmom/zmom);
//      if((ratio > 1.0) && (Abs(zmom) > 1.0e-3))
//        {
//          cout << vof << " has wacked ratio =" << ratio << endl;
//        }
    }
}
/******/
void 
EBPatchPolytropic::floorConserved(BaseIVFAB<Real>& a_consState,
                                  const IntVectSet& a_ivsIrreg)
{
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  for(VoFIterator vofit(a_ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      a_consState(vof, CRHO) = Max(a_consState(vof, CRHO), smallr);
      a_consState(vof, CENG) = Max(a_consState(vof, CENG), small );
    }

}
/******/
void 
EBPatchPolytropic::floorPrimitives(BaseIVFAB<Real>& a_primState,
                                   const IntVectSet& a_ivsIrreg)
{
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  for(VoFIterator vofit(a_ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //floors
      a_primState(vof, QRHO)  = Max(a_primState(vof, QRHO) ,  smallr);
      a_primState(vof, QPRES) = Max(a_primState(vof, QPRES),  smallp);
    }
}
/*****************************/
EBPatchPolytropic::
EBPatchPolytropic():EBPatchGodunov()
{
  m_isGammaSet    = false;
  m_isArtViscSet  = false;
  m_doRZCoords    = false;
}
/******/
bool 
EBPatchPolytropic::
isDefined() const
{
  bool retval = EBPatchGodunov::isDefined() &&  m_isGammaSet;
  return retval;
}

/*****************************/
EBPatchPolytropic::
~EBPatchPolytropic()
{
}
/*****************************/
void
EBPatchPolytropic::
computeEBIrregFlux(BaseIVFAB<Real>&  a_ebIrregFlux,
                   const EBCellFAB&  a_primState,
                   const EBCellFAB&  a_nonConsDivF,
                   const EBCellFAB   a_slopePrim[SpaceDim],
                   const IntVectSet& a_irregIVS,
                   const EBCellFAB&  a_source)
{
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  int numCons = numConserved();
  int numPrim = numPrimitives();
  assert(isDefined());
  assert(a_ebIrregFlux.nComp() == numCons);
  assert(a_nonConsDivF.nComp() == numCons);
  assert(a_primState.nComp() == numPrim);
  assert(a_ebIrregFlux.getIVS().contains(a_irregIVS));
  
  for(VoFIterator vofit(a_irregIVS, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      RealVect centroid = m_ebisBox.bndryCentroid(vof);

      assert(a_primState(vof, QRHO) > 0.0);

      Real     dense  = a_primState(vof, QRHO);
      Real     press  = a_primState(vof, QPRES);
      Real     sound  = sqrt(m_gamma*press/dense);
      RealVect veloc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          veloc[idir] = a_primState(vof, QVELX+idir);
        }
      Real     deltaDense = 0.0;
      Real     deltaPress = 0.0;
      RealVect deltaVeloc = RealVect::Zero;
      Real dtOver2H = 0.5*m_dt/m_dx;
      Real    rhoc2 = dense*sound*sound;
      for(int extrapDir = 0; extrapDir < SpaceDim; extrapDir++)
        {
          Tuple<int, CH_SPACEDIM-1> tanDirs = PolyGeom::computeTanDirs(extrapDir);
          Real centDir = centroid[extrapDir];
          const EBCellFAB& slopePrimDir =  a_slopePrim[extrapDir];

          Real     slopeDense = slopePrimDir(vof, QRHO);
          Real     slopePress = slopePrimDir(vof, QPRES);
          RealVect slopeVeloc;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              slopeVeloc[idir]= slopePrimDir(vof, QVELX + idir);
            }
          Real slopeUNorm = slopeVeloc[extrapDir];
          Real      velNo =      veloc[extrapDir];

          deltaDense += centDir*slopeDense - dtOver2H*(velNo*slopeDense + dense*slopeUNorm);
          deltaPress += centDir*slopePress - dtOver2H*(velNo*slopePress + rhoc2*slopeUNorm);
          
          deltaVeloc[extrapDir] += centDir*slopeUNorm - dtOver2H*(velNo*slopeUNorm + slopePress/dense);
            

          for(int itan = 0; itan < SpaceDim-1; itan++)
            {
              int tanDir = tanDirs[itan];
              Real slopeUTang = slopeVeloc[tanDir];

              deltaVeloc[tanDir] += centDir*slopeUTang - dtOver2H*(velNo*slopeUTang);
            }
        }

      Real xstate[QNUM];
      xstate[QRHO]  = dense + deltaDense;
      xstate[QPRES] = press + deltaPress;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xstate[QVELX + idir] = veloc[idir] + deltaVeloc[idir];
        }

      if(a_source.isDefined())
        {
          for(int ivar = 0; ivar < QNUM; ivar++)
            {
              Real sourceval = a_source(vof, ivar);
              xstate[ivar] += 0.5*m_dt*sourceval;
            }
        }
             

      //enforce positivity
      Real pext = xstate[QPRES];
      Real rext = xstate[QRHO];
      if((pext < 0.0) || (rext < 0.0))
        {
          for(int ivar = 0; ivar < numPrim; ivar++)
            {
              xstate[ivar] = a_primState(vof, ivar);
            }
        }

      //now compute the boundary condition assuming
      //solid walls at the embedded boundary
      RealVect velocity;
      Real     density, pressure;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velocity[idir] = xstate[QVELX + idir];
        }

      density  = xstate[QRHO];
      pressure = xstate[QPRES];
      RealVect normal = m_ebisBox.normal(vof);
      Real normalvel = PolyGeom::dot(normal, velocity);

      assert(density > 0.0);
      pressure = Max(pressure, smallp);
      Real soundspeed = sqrt(m_gamma*pressure/density);
      Real pstar = pressure - density*normalvel*soundspeed;
      //modification for artificial viscosity
#if 1
      if(usesArtificialViscosity())
        {
          Real coeff = artificialViscosityCoefficient();
          //approximation to the velocity divergence
          Real divu = 0.0;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const EBCellFAB& slopeDir = a_slopePrim[idir];
              int ivel = QVELX + idir;
              divu += slopeDir(vof, ivel);
            }
          Real dp = -2.0*density*coeff*normalvel*abs(divu);
          pstar = pstar + dp;
        }
#endif
      //debug
      pstar = Max(pstar, smallp);
      //end debug

      //fluxes as per modiano colella
      a_ebIrregFlux(vof, CRHO) = 0.0;
      a_ebIrregFlux(vof, CENG) = 0.0;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          int imom = CMOMX + idir;
          //negative because our normal points into the fluid
          //these are not separated out for RZ since the pressure flux
          //is the only flux at the EB
          a_ebIrregFlux(vof, imom) = -pstar*normal[idir];
        }
    }
}
/********/
/********/
void
EBPatchPolytropic::
computeBoundaryPress(BaseIVFAB<Real>&  a_pstar,
                     BaseIVFAB<Real>&  a_vnorm,
                     BaseIVFAB<Real>&  a_normDotModiano,
                     BaseIVFAB<Real>&  a_normError,
                     BaseIVFAB<Real>&  a_pstarNoVNorm,
                     BaseIVFAB<Real>&  a_vDotTrueNorm,
                     const RealVect&   a_modianoAxis,
                     const RealVect&   a_modianoCorner,
                     const EBCellFAB&  a_primState,
                     const EBCellFAB&  a_nonConsDivF,
                     const EBCellFAB   a_slopePrim[SpaceDim],
                     const IntVectSet& a_irregIVS,
                     const EBCellFAB&  a_source)
{
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  int numCons = numConserved();
  int numPrim = numPrimitives();
  assert(isDefined());
  assert(a_nonConsDivF.nComp() == numCons);
  assert(a_primState.nComp() == numPrim);
  assert(a_pstar.getIVS().contains(a_irregIVS));
  
  for(VoFIterator vofit(a_irregIVS, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      RealVect centroid = m_ebisBox.bndryCentroid(vof);

      assert(a_primState(vof, QRHO) > 0.0);

      Real     dense  = a_primState(vof, QRHO);
      Real     press  = a_primState(vof, QPRES);
      Real     sound  = sqrt(m_gamma*press/dense);
      RealVect veloc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          veloc[idir] = a_primState(vof, QVELX+idir);
        }
      Real     deltaDense = 0.0;
      Real     deltaPress = 0.0;
      RealVect deltaVeloc = RealVect::Zero;
      Real dtOver2H = 0.5*m_dt/m_dx;
      Real    rhoc2 = dense*sound*sound;
      for(int extrapDir = 0; extrapDir < SpaceDim; extrapDir++)
        {
          Tuple<int, CH_SPACEDIM-1> tanDirs = PolyGeom::computeTanDirs(extrapDir);
          Real centDir = centroid[extrapDir];
          const EBCellFAB& slopePrimDir =  a_slopePrim[extrapDir];

          Real     slopeDense = slopePrimDir(vof, QRHO);
          Real     slopePress = slopePrimDir(vof, QPRES);
          RealVect slopeVeloc;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              slopeVeloc[idir]= slopePrimDir(vof, QVELX + idir);
            }
          Real slopeUNorm = slopeVeloc[extrapDir];
          Real      velNo =      veloc[extrapDir];

          deltaDense += centDir*slopeDense - dtOver2H*(velNo*slopeDense + dense*slopeUNorm);
          deltaPress += centDir*slopePress - dtOver2H*(velNo*slopePress + rhoc2*slopeUNorm);
          
          deltaVeloc[extrapDir] += centDir*slopeUNorm - dtOver2H*(velNo*slopeUNorm + slopePress/dense);
            

          for(int itan = 0; itan < SpaceDim-1; itan++)
            {
              int tanDir = tanDirs[itan];
              Real slopeUTang = slopeVeloc[tanDir];

              deltaVeloc[tanDir] += centDir*slopeUTang - dtOver2H*(velNo*slopeUTang);
            }
        }
      Real xstate[QNUM];
      xstate[QRHO]  = dense + deltaDense;
      xstate[QPRES] = press + deltaPress;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xstate[QVELX + idir] = veloc[idir] + deltaVeloc[idir];
        }

      if(a_source.isDefined())
        {
          for(int ivar = 0; ivar < QNUM; ivar++)
            {
              Real sourceval = a_source(vof, ivar);
              xstate[ivar] += 0.5*m_dt*sourceval;
            }
        }
             

      //enforce positivity
      Real pext = xstate[QPRES];
      Real rext = xstate[QRHO];
      if((pext < 0.0) || (rext < 0.0))
        {
          for(int ivar = 0; ivar < numPrim; ivar++)
            {
              xstate[ivar] = a_primState(vof, ivar);
            }
        }

      //now compute the boundary condition assuming
      //solid walls at the embedded boundary
      RealVect velocity;
      Real     density, pressure;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velocity[idir] = xstate[QVELX + idir];
        }

      density  = xstate[QRHO];
      pressure = xstate[QPRES];
      RealVect normal = m_ebisBox.normal(vof);
      Real normalvel = PolyGeom::dot(normal, velocity);

      assert(density > 0.0);
      pressure = Max(pressure, smallp);
      Real soundspeed = sqrt(m_gamma*pressure/density);
      Real pstar = pressure - density*normalvel*soundspeed;
      //modification for artificial viscosity
      if(usesArtificialViscosity())
        {
          Real coeff = artificialViscosityCoefficient();
          //approximation to the velocity divergence
          Real divu = 0.0;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const EBCellFAB& slopeDir = a_slopePrim[idir];
              int ivel = QVELX + idir;
              divu += slopeDir(vof, ivel);
            }
          Real dp = -2.0*coeff*normalvel*Max(-divu, 0.0);
          pstar = pstar + dp;
        }
      a_pstar(vof, 0) = pstar;
      a_vnorm(vof, 0) = normalvel;
      Real normdotmod = PolyGeom::dot(normal, a_modianoAxis);

      a_normDotModiano(vof, 0) = normdotmod;
      
      a_pstarNoVNorm(vof, 0)  = pressure;

      const IntVect& iv = vof.gridIndex();
      RealVect centroidPt;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          centroidPt[idir] = Real(iv[idir]) + 0.5 + centroid[idir];
        }
      centroidPt *= m_dx;
      RealVect trueNorm, closestPt;
      PolyGeom::pointToLine(closestPt, trueNorm, centroidPt,
                            a_modianoCorner, a_modianoAxis);
      
      Real vdottruenorm = PolyGeom::dot(velocity, trueNorm);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_normError(vof, idir) = normal[idir] - trueNorm[idir];
        }
      a_vDotTrueNorm(vof, 0) = vdottruenorm;
      
    }
}
/*****************************/
/********/
void
EBPatchPolytropic::
computeBoundaryWDivU(BaseIVFAB<Real>&  a_pstar,
                     BaseIVFAB<Real>&  a_vnorm,
                     BaseIVFAB<Real>&  a_normDotModiano,
                     const RealVect&   a_modianoAxis,
                     const EBCellFAB&  a_primState,
                     const EBCellFAB&  a_nonConsDivF,
                     const EBCellFAB   a_slopePrim[SpaceDim],
                     const IntVectSet& a_irregIVS,
                     const EBCellFAB&  a_source)
{
  Real small, smallp, smallu, smallr;
  FORT_GETSMALL(CHF_REAL(small),
                CHF_REAL(smallp),
                CHF_REAL(smallu),
                CHF_REAL(smallr));
  int numCons = numConserved();
  int numPrim = numPrimitives();
  assert(isDefined());
  assert(a_nonConsDivF.nComp() == numCons);
  assert(a_primState.nComp() == numPrim);
  assert(a_pstar.getIVS().contains(a_irregIVS));
  
  for(VoFIterator vofit(a_irregIVS, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      //extrapolate in space
      const VolIndex& vof = vofit();
      RealVect centroid = m_ebisBox.bndryCentroid(vof);

      assert(a_primState(vof, QRHO) > 0.0);

      Real xstatePrim[QNUM];
      for(int ivar = 0; ivar < numPrim; ivar++)
        {
          xstatePrim[ivar] = a_primState(vof,ivar);
        }
      for(int ivar = 0; ivar < numPrim; ivar++)
        {
          for(int extrapDir = 0; extrapDir < SpaceDim; extrapDir++)
            {
              Real centDir = centroid[extrapDir];
              const EBCellFAB& slopePrimDir =  a_slopePrim[extrapDir];
          
              xstatePrim[ivar] += centDir*slopePrimDir(vof, ivar);
            }
        }
      if(a_source.isDefined())
        {
          for(int ivar = 0; ivar < QNUM; ivar++)
            {
              Real sourceval = a_source(vof, ivar);
              xstatePrim[ivar] += 0.5*m_dt*sourceval;
            }
        }


      Real     dense  = xstatePrim[QRHO];
      Real     press  = xstatePrim[QPRES];
      RealVect veloc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          veloc[idir] = xstatePrim[QVELX+idir];
        }

      Real energy;
      RealVect momen;
      //convert to conserved
      FORT_POINTPRM2CONS(CHF_REAL(dense), 
                         CHF_REALVECT(veloc), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy));
      Real xstateCons[CNUM];
      xstateCons[CRHO] = dense;
      xstateCons[CENG] = energy;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xstateCons[CMOMX+idir] = momen[idir];
        }

      //extrapolate in time
      for(int ivar = 0; ivar <  CNUM; ivar++)
        {
          Real divf = a_nonConsDivF(vof, ivar);
          xstateCons[ivar] -= 0.5*m_dt*divf;
        }
          
      //back into primitives
      dense = xstateCons[CRHO];
      energy = xstateCons[CENG];
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          momen[idir] = xstateCons[CMOMX+idir];
        }
      RealVect modianovel;
      Real entropy, soundspeed;
      int logflag = 0;
      FORT_POINTCONS2PRM(CHF_REAL(dense), 
                         CHF_REALVECT(veloc), 
                         CHF_REAL(press), 
                         CHF_REALVECT(momen), 
                         CHF_REAL(energy),
                         CHF_REALVECT(modianovel), 
                         CHF_REAL(entropy),
                         CHF_REAL(soundspeed),
                         CHF_CONST_INT(logflag)
                         );
             
      xstatePrim[CRHO] = dense;
      xstatePrim[CENG] = energy;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xstatePrim[QVELX+idir]=  veloc[idir]; 
        }

      //enforce positivity
      Real pext = xstatePrim[QPRES];
      Real rext = xstatePrim[QRHO];
      if((pext < 0.0) || (rext < 0.0))
        {
          for(int ivar = 0; ivar < numPrim; ivar++)
            {
              xstatePrim[ivar] = a_primState(vof, ivar);
            }
        }


      //now compute the boundary condition assuming
      //solid walls at the embedded boundary
      RealVect velocity;
      Real     density, pressure;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velocity[idir] = xstatePrim[QVELX + idir];
        }

      density  = xstatePrim[QRHO];
      pressure = xstatePrim[QPRES];
      RealVect normal = m_ebisBox.normal(vof);
      Real normalvel = PolyGeom::dot(normal, velocity);

      assert(density > 0.0);
      pressure = Max(pressure, smallp);
      Real pstar = pressure - density*normalvel*soundspeed;
      //modification for artificial viscosity
      if(usesArtificialViscosity())
        {
          Real coeff = artificialViscosityCoefficient();
          //approximation to the velocity divergence
          Real divu = 0.0;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const EBCellFAB& slopeDir = a_slopePrim[idir];
              int ivel = QVELX + idir;
              divu += slopeDir(vof, ivel);
            }
          Real dp = -2.0*coeff*normalvel*Max(-divu, 0.0);
          pstar = pstar + dp;
        }
      a_pstar(vof, 0) = pstar;
      a_vnorm(vof, 0) = normalvel;
      Real normdotmod = PolyGeom::dot(normal, a_modianoAxis);

      a_normDotModiano(vof, 0) = normdotmod;
    }
}
/*****************************/
/*****************************/
void 
EBPatchPolytropic::
setCoveredConsVals(EBCellFAB& a_consState)
{
  a_consState.setInvalidData(1.0, CRHO);
  a_consState.setInvalidData(0.0, CMOMX);
  a_consState.setInvalidData(0.0, CMOMY);
#if CH_SPACEDIM==3                        
  a_consState.setInvalidData(0.0, CMOMZ);
#endif                                    
  a_consState.setInvalidData(1.0, CENG);
}
/**************/
//this is the stable, non-conservative estimate of the solution update
void
EBPatchPolytropic::
nonconservativeDivergence(EBCellFAB&             a_divF, 
                          const EBFluxFAB&       a_flux, 
                          const BaseIVFAB<Real>  a_coveredFluxMinu[SpaceDim], 
                          const BaseIVFAB<Real>  a_coveredFluxPlus[SpaceDim], 
                          const Vector<VolIndex> a_coveredFaceMinu[SpaceDim], 
                          const Vector<VolIndex> a_coveredFacePlus[SpaceDim], 
                          const Box&             a_box)
{
  if(m_doRZCoords)
    {
      nonconservativeDivergenRZ(a_divF, a_flux, 
                                a_coveredFluxMinu, 
                                a_coveredFluxPlus, 
                                a_coveredFaceMinu, 
                                a_coveredFacePlus, 
                                a_box);
    }
  else
    {
      EBPatchGodunov::nonconservativeDivergence(a_divF, a_flux, 
                                                a_coveredFluxMinu, 
                                                a_coveredFluxPlus, 
                                                a_coveredFaceMinu, 
                                                a_coveredFacePlus, 
                                                a_box);
    }
}
/*****************************/
void
EBPatchPolytropic::
consUndividedDivergence(BaseIVFAB<Real>&       a_divF, 
                        const BaseIFFAB<Real>  a_centroidFlux[SpaceDim], 
                        const BaseIVFAB<Real>& a_ebIrregFlux,
                        const IntVectSet&      a_ivs)
{
  if(m_doRZCoords)
    {
      consUndividedDivergenRZ(a_divF, a_centroidFlux, a_ebIrregFlux, a_ivs);
    }
  else
    {
      EBPatchGodunov::consUndividedDivergence(a_divF, a_centroidFlux, a_ebIrregFlux, a_ivs);
    }
}
/*****************************/
