/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _GODUNOVGEOM__
#define _GODUNOVGEOM__

#include "Box.H"
#include "REAL.H"
extern "C"
{
  typedef void (*IrregErrorFcn)(BaseIVFAB<Real>&  a_error, 
                                const IntVectSet& a_ivsIrreg, 
                                const EBISBox&    a_ebisBox, 
                                const RealVect&   a_cylinderAxis,
                                const Real& a_dx);
}
extern "C"
{
  RealVect getCentroidPt(const EBISBox&  a_ebisBox,
                         const VolIndex& a_vof,
                         const Real&     a_dx)
  {
    RealVect centroid = a_ebisBox.bndryCentroid(a_vof);

    const IntVect&   iv = a_vof.gridIndex();
    RealVect centroidPt;
    for(int idir = 0; idir < SpaceDim; idir++)
      {
        centroidPt[idir] = Real(iv[idir]) + 0.5 + centroid[idir];
      }
    centroidPt *= a_dx;
    return centroidPt;
  }
  /************/
  void
  getNormalDotAxis(BaseIVFAB<Real>&  a_error, 
                   const IntVectSet& a_ivsIrreg, 
                   const EBISBox&    a_ebisBox, 
                   const RealVect&   a_cylinderAxis,
                   const Real& a_dx)
  {
    //dot normal with the cylinder axis.  right answer == 0
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        RealVect normal = a_ebisBox.normal(vof);
        Real error = PolyGeom::dot(a_cylinderAxis, normal);
        a_error(vof, 0) = error;
      }
  }

  /************/
  void
  getVolFrac(BaseIVFAB<Real>&  a_error, 
             const IntVectSet& a_ivsIrreg, 
             const EBISBox&    a_ebisBox, 
             const RealVect&   a_cylinderAxis,
             const Real& a_dx)
  {
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        Real volfrac = a_ebisBox.volFrac(vof);
        a_error(vof, 0) = volfrac;
      }
  }
  /************/
  void
  getNormMinTrueNormDotAxis(BaseIVFAB<Real>&  a_error, 
                            const IntVectSet& a_ivsIrreg, 
                            const EBISBox&    a_ebisBox, 
                            const RealVect&   a_cylinderAxis,
                            const Real&       a_dx)
  {
    //dot normal with truenormal.  right answer == 1 so we subtract 1
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        RealVect centroidPt = getCentroidPt(a_ebisBox, vof, a_dx);

        RealVect trueNorm, closestPt;
        RealVect cylinderCorner = RealVect::Zero;
        PolyGeom::pointToLine(closestPt, trueNorm, centroidPt,
                              cylinderCorner, a_cylinderAxis);

        RealVect normal = a_ebisBox.normal(vof);
        RealVect diff = normal;
        diff -= trueNorm;

        Real error = PolyGeom::dot(a_cylinderAxis, diff);
        a_error(vof, 0) = error ;
      }
  }
  /************/
  void
  getNormMinTrueNorm(BaseIVFAB<Real>&  a_error, 
                     const IntVectSet& a_ivsIrreg, 
                     const EBISBox&    a_ebisBox, 
                     const RealVect&   a_cylinderAxis,
                     const Real&       a_dx)
  {
    //dot normal with truenormal.  right answer == 1 so we subtract 1
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        RealVect centroidPt = getCentroidPt(a_ebisBox, vof, a_dx);

        RealVect trueNorm, closestPt;
        RealVect cylinderCorner = RealVect::Zero;
        PolyGeom::pointToLine(closestPt, trueNorm, centroidPt,
                              cylinderCorner, a_cylinderAxis);

        RealVect normal = a_ebisBox.normal(vof);
        Real error = 0.0;
        for(int idir = 0; idir < SpaceDim; idir++)
          {
            Real diff = normal[idir] - trueNorm[idir];
            error += diff*diff;
          }
        a_error(vof, 0) = sqrt(error);
      }
  }

  /*****/
  void
  getNormalDotTrueNormM1(BaseIVFAB<Real>&  a_error, 
                         const IntVectSet& a_ivsIrreg, 
                         const EBISBox&    a_ebisBox, 
                         const RealVect&   a_cylinderAxis,
                         const Real&       a_dx)
  {
    //dot normal with truenormal.  right answer == 1 so we subtract 1
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        RealVect centroidPt = getCentroidPt(a_ebisBox, vof, a_dx);

        RealVect trueNorm, closestPt;
        RealVect cylinderCorner = RealVect::Zero;
        PolyGeom::pointToLine(closestPt, trueNorm, centroidPt,
                              cylinderCorner, a_cylinderAxis);

        RealVect normal = a_ebisBox.normal(vof);
        Real dotProd = PolyGeom::dot(trueNorm, normal);
        Real error = Abs(dotProd) - 1.0;

        a_error(vof, 0) = error ;
      }
  }
  /************/
  void
  getNormalDotCrossVec(BaseIVFAB<Real>&  a_error, 
                       const IntVectSet& a_ivsIrreg, 
                       const EBISBox&    a_ebisBox, 
                       const RealVect&   a_cylinderAxis,
                       const Real&       a_dx)
  {
    //dot normal with axis cross truenormal.  right answer == 0
    for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        RealVect centroidPt = getCentroidPt(a_ebisBox, vof, a_dx);

        RealVect trueNorm, closestPt;
        RealVect cylinderCorner = RealVect::Zero;
        PolyGeom::pointToLine(closestPt, trueNorm, centroidPt,
                              cylinderCorner, a_cylinderAxis);
        RealVect crossVec = PolyGeom::cross(trueNorm, a_cylinderAxis);

        RealVect normal = a_ebisBox.normal(vof);
        Real error = PolyGeom::dot(crossVec, normal);
        a_error(vof, 0) = error;
      }
  }
}
/****/
extern  void
godunovFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                  const ProblemDomain&  a_coarsestDomain,
                  int                   a_max_level,
                  int                   a_max_grid_size,
                  int                   a_block_factor,
                  int                   a_verbosity,
                  std::string           a_grid_file);

/****/
extern void
godunovGeometry(Box& a_coarsestDomain, 
                Real& a_dx);
/****/
#endif
