/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef  _EB_MODIANOIBC_H_
#define  _EB_MODIANOIBC_H_

#include  <iostream>

#include "LevelData.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RealVect.H"
#include "AMRIO.H"

#include "EBPhysIBC.H"
#include "EBFluxFAB.H"
///
/**
   IBC for doing the modiano problem.
 */
class ModianoIBC : public EBPhysIBC
{
public:

  virtual ~ModianoIBC();

  ///
  /**
     waveAmp is the amplitude of the disturbance in the density
     (alpha in Modiano and Colella) . \\
     waveWidth is the width of the wave (w in Modiano Colella) \\
     waveDir is the direction of  the wave.  The wave runs perpendicular to
     the direction of the walls. \\
     gamma is of course the ratio of specific heats for the gas.
   */
  ModianoIBC(const Real&     a_gamma,
             const Real&     a_waveAmp,
             const Real&     a_waveWidth,
             const RealVect& a_center,
             const RealVect& a_waveDir,
             bool  a_freeStreamOnly,
             bool a_negativeWave = false);

  ///
  void define(const ProblemDomain&  a_domain,
              const Real& a_dx);

  void  setToExact(LevelData<EBCellFAB>& a_conState,
                   const EBISLayout& a_ebisl,
                   const Real&       a_time) const;

  void  setToExact(EBCellFAB&     a_conState,
                   const EBISBox& a_ebisl,
                   const Real&    a_time) const;


  void  setToExactConsAndPrim(LevelData<EBCellFAB>& a_conState,
                              const EBISLayout& a_ebisl,
                              const Real&       a_time) const;

  void  setToExactConsAndPrim(EBCellFAB&     a_conState,
                              const EBISBox& a_ebisl,
                              const Real&    a_time) const;

  ///  For every box in this level, this function is called.
  void fluxBC(EBFaceFAB&            a_flux,
              const EBCellFAB&      a_Wcenter,
              const EBCellFAB&      a_Wextrap,
              const Side::LoHiSide& a_sd,
              const Real&           a_time,
              const EBISBox&   a_ebisBox,
              const Box&       a_box,
              const int&            a_dir);

  /// Initialize
  void initialize(LevelData<EBCellFAB>& a_conState,
                  const EBISLayout& a_ebisl) const;
  
  ///
  void initialize(EBCellFAB& a_conState,
                  const EBISBox& a_ebisl) const;

  ///
  void      
  computeExactFluxes(EBFluxFAB& cflux,
                     BaseIVFAB<Real>  coveredCfluxMinu[SpaceDim], 
                     BaseIVFAB<Real>  coveredCfluxPlus[SpaceDim], 
                     Vector<VolIndex> coveredFacesMinu[SpaceDim], 
                     Vector<VolIndex> coveredFacesPlus[SpaceDim], 
                     IntVectSet coveredSetsMinu[SpaceDim], 
                     IntVectSet coveredSetsPlus[SpaceDim], 
                     BaseIVFAB<Real>& boundaryPress,
                     IntVectSet&      irregIVS,
                     const EBISBox& ebisBox,
                     const Box& a_region,
                     const Box& domainBox,
                     const Real& time);

  /// Set boundary slopes
  /**
     The boundary slopes in a_dW are already set to one sided difference
     approximations.  If this function doesn't change them they will be
     used for the slopes at the boundaries.
  */
  void setBndrySlopes(EBCellFAB&       a_deltaPrim,
                      const EBCellFAB& a_primState,
                      const EBISBox&   a_ebisBox,
                      const Box&       a_box,
                      const int&       a_dir);

protected:

  bool     m_isFortranCommonSet;
  bool     m_isDefined;
  Real     m_gamma;
  Real     m_waveAmp;
  Real     m_waveWidth;
  RealVect m_waveDir;
  RealVect m_center;
  bool m_freeStreamOnly;

  Real m_dx;
  ProblemDomain  m_domain;

  int m_iNegWave;
  int m_idoFreeStreamOnly;
private:
  //forcing strong construction 
  ModianoIBC()
  {
    MayDay::Error("invalid operator");
  }
  //disallowed for all the usual reasons
  void operator=(const ModianoIBC& a_input)
  {
    MayDay::Error("invalid operator");
  }
  ModianoIBC(const ModianoIBC& a_input)
  {
    MayDay::Error("invalid operator");
  }


};

extern RealVect 
mibcGetNormalVector(const RealVect& channelNormal);

#endif
