/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cerr;

#include "ParmParse.H"
#include "LoadBalance.H"
#include "BRMeshRefine.H"

#include "MultiGrid.H"

#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "EBCellFactory.H"
#include "AllRegularService.H"

#include "GeometryShop.H"
#include "ImplicitFunctionBL.H"
#include "PlaneIF.H"
#include "SphereIF.H"
#include "PolarIF.H"
#include "LatheIF.H"
#include "TransformIF.H"

#include "EBFABView.H"
#include "EBDebugDump.H"

#include "EBAMRPoissonOp.H"
#include "EBLevelDataOps.H"
#include "BaseBCValue.H"
#include "BaseBCFlux.H"
#include "BaseDomainBC.H"
#include "NeumannDomainBC.H"
#include "DirichletDomainBC.H"
#include "BaseEBBC.H"
#include "DirichletEBBC.H"
#include "NeumannEBBC.H"

#include "EBSimpleSolver.H"
#include "functionsF_F.H"
#include "TrigBCValue.H"
#include "TrigBCFlux.H"

#define BCNEUMANN       0   // Neumann boundary conditions
#define BCDIRICHLET1    1   // Dirichlet boundary conditions (1st order)
#define BCDIRICHLET2    2   // Dirichlet boundary conditions (2nd order)

struct testParams
{
  int      verbose;
  RealVect trig;
  IntVect  nCells;
  int      mgDepth;
  int      maxBoxSize;
  int      blockFactor;
  int      mgRelaxType;
  int      mgSmoothDown;
  int      mgSmoothUp;
  int      mgCycles;
  Real     mgResidReduc;
  int      mgMaxIter;
  RealVect probLo;
  RealVect probHi;
  RealVect dx;
  int      orderEB;
  int      ebBcType;
  int      domBcType;
  int      geometry;
  RealVect point;
  RealVect normal;
  Real     radius;
  RealVect center;
  int      interior;
  Real     baseRadius;
  RealVect polarCenter;
  Real     perturbation;
  int      frequency;
  int      polarInterior;
};

void multigrid(struct testParams& a_params);

void setPhi(LevelData<EBCellFAB>&    a_phi,
            const struct testParams& a_params);

void setKappaLOfPhi(LevelData<EBCellFAB>&    a_kappaLOfPhi,
                    const struct testParams& a_params);

void printNorms(Real&                       a_errorMaxNorm,
                const Real&                 a_lastErrorMaxNorm,
                const int&                  a_nCells,
                const int&                  a_iter,
                const LevelData<EBCellFAB>& a_resid,
                const LevelData<EBCellFAB>& a_error,
                EBLevelDataOps&             a_levelDataOps,
                const ProblemDomain&        a_domain);

int main(int argc, char* argv[])
{
  // Parameters
  struct testParams params;

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  // Scoping trick
  {
#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
    if (argc < 2)
    {
      cerr << " usage " << argv[0] << " <input_file_name> " << endl;
      exit(0);
    }

    char* inFile = argv[1];
    ParmParse pp(argc-2,argv+2,NULL,inFile);

    ParmParse ppMain("main");

    ppMain.get("verbose",params.verbose);

    std::vector<Real> trig(SpaceDim);
    ppMain.getarr("trig",trig,0,SpaceDim);

    ppMain.get("orderEB",params.orderEB);
    if (params.orderEB < 0 || params.orderEB > 1)
    {
      MayDay::Error("EB order not 0 or 1");
    }

    ppMain.get("ebBcType",params.ebBcType);
    if (params.ebBcType < 0 || params.ebBcType > 2)
    {
      MayDay::Error("BC type not 0, 1, or 2");
    }

    ppMain.get("domBcType",params.domBcType);
    if (params.domBcType < 0 || params.domBcType > 1)
    {
      MayDay::Error("BC type not 0 or 1");
    }

    std::vector<int> nCellsArray(SpaceDim);
    ppMain.getarr("nCells",nCellsArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.nCells[idir] = nCellsArray[idir];
    }

    params.mgDepth = 0;
    ppMain.query("mgDepth",params.mgDepth);

    ppMain.get("maxBoxSize",params.maxBoxSize);
    ppMain.get("blockFactor",params.blockFactor);

    ppMain.get("mgRelaxType",params.mgRelaxType);
    if (params.mgRelaxType < 0 || params.mgRelaxType > 1)
    {
      MayDay::Error("Relaxation type not 0 or 1");
    }

    ppMain.get("mgSmoothDown",params.mgSmoothDown);
    if (params.mgSmoothDown < 0)
    {
      params.mgSmoothDown = 0;
    }
    ppMain.get("mgSmoothUp",params.mgSmoothUp);
    if (params.mgSmoothUp < 0)
    {
      params.mgSmoothUp = 0;
    }

    ppMain.get("mgCycles",params.mgCycles);
    if (params.mgCycles < 1)
    {
      params.mgCycles = 1;
    }

    ppMain.get("mgResidReduc",params.mgResidReduc);
    ppMain.get("mgMaxIter",params.mgMaxIter);

    std::vector<Real> probLoArray(SpaceDim);
    ppMain.getarr("probLo",probLoArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.probLo[idir] = probLoArray[idir];
    }

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.trig[idir] = trig[idir];
    }

    std::vector<Real> probHiArray(SpaceDim);
    ppMain.getarr("probHi",probHiArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.probHi[idir] = probHiArray[idir];
    }

    ppMain.get("geometry",params.geometry);

    std::vector<Real> pointArray(SpaceDim);
    ppMain.getarr("point",pointArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.point[idir] = pointArray[idir];
    }

    std::vector<Real> normalArray(SpaceDim);
    ppMain.getarr("normal",normalArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.normal[idir] = normalArray[idir];
    }

    ppMain.get("radius",params.radius);

    std::vector<Real> centerArray(SpaceDim);
    ppMain.getarr("center",centerArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.center[idir] = centerArray[idir];
    }

    ppMain.get("interior",params.interior);

    ppMain.get("baseRadius",params.baseRadius);

    ppMain.getarr("polarCenter",centerArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.polarCenter[idir] = centerArray[idir];
    }

    ppMain.get("perturbation",params.perturbation);

    ppMain.get("frequency",params.frequency);

    ppMain.get("polarInterior",params.polarInterior);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.dx[idir] = (params.probHi[idir] - params.probLo[idir]) / params.nCells[idir];
    }

    multigrid(params);
  }
  // End scoping trick

#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void multigrid(struct testParams& a_params)
{
  int nComps = 1;

  int maxDepth = 1;
  IntVect cells = a_params.nCells;

  while (1)
  {
    IntVect newCells = cells;
    newCells /= 2;
    newCells *= 2;

    if (newCells == cells)
    {
      maxDepth++;
      cells /= 2;
    }
    else
    {
      break;
    }
  }

  if (a_params.mgDepth == 0)
  {
    a_params.mgDepth = maxDepth;
  }

  if (a_params.mgDepth > maxDepth)
  {
    MayDay::Error("Too many multigrid levels for domain");
  }

  bool isPeriodic[SpaceDim];
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    isPeriodic[idir] = false;
  }

  IntVect ivLo = IntVect::TheZeroVector();
  IntVect ivHi;

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    ivHi[idir] = a_params.nCells[idir] - 1;
  }

  Box domainBox(ivLo,ivHi);
  ProblemDomain domain(domainBox,isPeriodic);

  Vector<Box> boxes;
  domainSplit(domain,boxes,a_params.maxBoxSize,a_params.blockFactor);

  Vector<int> procIDs(boxes.size());
  LoadBalance(procIDs,boxes);

  DisjointBoxLayout dbl(boxes,procIDs,domain);

  BaseImplicitFunction *baseIF;

  if (a_params.geometry == 0)
  {
    RealVect point = a_params.point;
    point += a_params.probLo;

    bool isInterior = (a_params.interior == 1);

    baseIF = new PlaneIF(a_params.normal,point,isInterior);
  }
  else
  if (a_params.geometry == 1)
  {
    RealVect center = a_params.center;
    center += a_params.probLo;

    bool isInterior = (a_params.interior == 1);

    baseIF = new SphereIF(a_params.radius,center,isInterior);
  }
  else
  if (a_params.geometry == 2)
  {
    RealVect center = a_params.polarCenter;
    center += a_params.probLo;

    bool isInterior = (a_params.polarInterior == 1);

    if (SpaceDim == 2)
    {
      PolarIF polar(a_params.baseRadius,a_params.perturbation,
                    a_params.frequency,isInterior);

      TransformIF* moveit = new TransformIF(polar);
      moveit->translate(center);

      baseIF = moveit;
    }
    else
    {
      bool interior = true;
      PolarIF polar(a_params.baseRadius,a_params.perturbation,
                    a_params.frequency,interior);

      LatheIF rotate(polar,isInterior);

      TransformIF* moveit = new TransformIF(rotate);
      moveit->translate(center);

      baseIF = moveit;
    }
  }
  else
  {
    MayDay::Error("Unknown geometry specified");
  }

  ImplicitFunctionBL geometry(*baseIF);
  GeometryShop geometryShop(geometry);

  EBIndexSpace* space = Chombo_EBIS::instance();
  space->define(domainBox,a_params.probLo,a_params.dx[0],geometryShop,
                a_params.nCells[0]);

  RefCountedPtr<BaseDomainBC> baseDomainBC = NULL;

  if (a_params.domBcType == 0)
  {
    TrigBCFlux* trigFluxPtr = new TrigBCFlux();
    trigFluxPtr->define(a_params.trig);

    RefCountedPtr<BaseBCFlux> baseTrigFluxPtr = trigFluxPtr;

    NeumannDomainBC* domainBC = new NeumannDomainBC();
    domainBC->setFunction(baseTrigFluxPtr);

    baseDomainBC = domainBC;
  }
  else if (a_params.domBcType == 1)
  {
    TrigBCValue* trigValuePtr = new TrigBCValue();
    trigValuePtr->define(a_params.trig);

    RefCountedPtr<BaseBCValue> baseTrigValuePtr = trigValuePtr;

    DirichletDomainBC* domainBC = new DirichletDomainBC();
    domainBC->setFunction(baseTrigValuePtr);

    baseDomainBC = domainBC;
  }
  else
  {
    MayDay::Error("Unknown domain BC type");
  }

  EBISLayout layout;
  RefCountedPtr<BaseEBBCFactory> baseEBBCFactory = NULL;
  int numGhostCells;

  if (a_params.ebBcType == 0)
  {
    TrigBCFlux* trigFluxPtr = new TrigBCFlux();
    trigFluxPtr->define(a_params.trig);

    RefCountedPtr<BaseBCFlux> baseTrigFluxPtr = trigFluxPtr;

    numGhostCells = 1;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    NeumannEBBCFactory* ebBcFactory = new NeumannEBBCFactory();
    ebBcFactory->setFunction(baseTrigFluxPtr);

    baseEBBCFactory = ebBcFactory;
  }
  else if (a_params.ebBcType == 1)
  {
    TrigBCValue* trigValuePtr = new TrigBCValue();
    trigValuePtr->define(a_params.trig);

    RefCountedPtr<BaseBCValue> baseTrigValuePtr = trigValuePtr;

    numGhostCells = 1;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    DirichletEBBCFactory* ebBcFactory = new DirichletEBBCFactory();
    ebBcFactory->setOrder(1);
    ebBcFactory->setFunction(baseTrigValuePtr);

    baseEBBCFactory = ebBcFactory;
  }
  else if (a_params.ebBcType == 2)
  {
    TrigBCValue* trigValuePtr = new TrigBCValue();
    trigValuePtr->define(a_params.trig);

    RefCountedPtr<BaseBCValue> baseTrigValuePtr = trigValuePtr;

    numGhostCells = 3;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    DirichletEBBCFactory* ebBcFactory = new DirichletEBBCFactory();
    ebBcFactory->setOrder(2);
    ebBcFactory->setFunction(baseTrigValuePtr);

    baseEBBCFactory = ebBcFactory;
  }
  else
  {
    MayDay::Error("Unknown EB BC type");
  }

  RefCountedPtr<DataFactory<EBCellFAB> > rcpFactory(new EBCellFactory(layout));

  EBAMRPoissonOpFactory levelOpFactory;

  levelOpFactory.define(domain,dbl,space,numGhostCells,
                        a_params.mgDepth,a_params.probLo,a_params.dx,
                        a_params.orderEB,a_params.mgRelaxType,
                        baseDomainBC,baseEBBCFactory);

  int level = 0;
  EBAMRPoissonOp* levelOp = levelOpFactory.MGnewOp(domain,level);

  EBSimpleSolver bottomSolver;
  bottomSolver.setNumSmooths(a_params.mgSmoothDown);

  MultiGrid<LevelData<EBCellFAB> > solver;

  solver.define(levelOpFactory,a_params.mgDepth,&bottomSolver,domain);

  solver.m_pre  = a_params.mgSmoothDown;
  solver.m_post = a_params.mgSmoothUp;

  solver.m_cycle = a_params.mgCycles;

  solver.m_homogeneous = false;

  EBLevelDataOps levelDataOps;

  LevelData<EBCellFAB> phi(dbl,nComps,numGhostCells*IntVect::TheUnitVector(),*rcpFactory);
  LevelData<EBCellFAB> rhs(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),*rcpFactory);
  LevelData<EBCellFAB> phiExact(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),*rcpFactory);

  bool homogeneous = true;

  setPhi(phi,a_params);
  levelOp->applyOp(rhs,phi,!homogeneous);

  levelDataOps.setVal(phi,0.0);

  setPhi(phiExact,a_params);

  LevelData<EBCellFAB> resid(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),*rcpFactory);
  LevelData<EBCellFAB> error(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),*rcpFactory);

  Real errorMaxNorm;

  Real a =  1.0;
  Real b = -1.0;

  Real lastErrorMaxNorm = -1.0;

  int iter = 0;
  do {
    levelOp->residual(resid,phi,rhs,!homogeneous);
    levelOp->axby(error,phi,phiExact,a,b);

    printNorms(errorMaxNorm,lastErrorMaxNorm,a_params.nCells[0],iter,
               resid,error,levelDataOps,domain);

    if (lastErrorMaxNorm >= 0.0 && errorMaxNorm >= lastErrorMaxNorm)
    {
      break;
    }

    lastErrorMaxNorm = errorMaxNorm;

    solver.solve(phi,rhs);

    iter++;
  } while (iter < a_params.mgMaxIter);

  if (iter == a_params.mgMaxIter)
  {
    levelOp->residual(resid,phi,rhs,!homogeneous);
    levelOp->axby(error,phi,phiExact,a,b);

    lastErrorMaxNorm = 0.0;

    printNorms(errorMaxNorm,lastErrorMaxNorm,a_params.nCells[0],iter,
               resid,error,levelDataOps,domain);
  }

  pout() << endl;

  // writeEBLevelname(&phi,"phi.hdf5");
  // writeEBLevelname(&rhs,"rhs.hdf5");
  // writeEBLevelname(&resid,"resid.hdf5");
  // writeEBLevelname(&phiExact,"phiExact.hdf5");
#ifdef HDF5
  writeEBLevelname(&error,"error.hdf5");
#endif
  solver.reclaim(levelOpFactory);

  delete levelOp;
}

void setPhi(LevelData<EBCellFAB>&    a_phi,
            const struct testParams& a_params)
{
  assert(a_phi.nComp() == 1);

  int comp = 0;

  for (DataIterator dit = a_phi.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& curPhiEBCellFAB = a_phi[dit()];
    Box curPhiBox = curPhiEBCellFAB.getRegion();
    FArrayBox& curPhiFAB = curPhiEBCellFAB.getFArrayBox();

    FORT_GETPHI(CHF_FRA1(curPhiFAB,comp),
                CHF_CONST_REALVECT(a_params.trig),
                CHF_CONST_REALVECT(a_params.dx),
                CHF_CONST_REALVECT(a_params.probLo),
                CHF_CONST_REALVECT(a_params.probHi),
                CHF_BOX(curPhiBox));

    const EBISBox& curPhiEBISBox = curPhiEBCellFAB.getEBISBox();
    const EBGraph& curPhiEBGraph = curPhiEBISBox.getEBGraph();

    IntVectSet notRegular;

    notRegular |= curPhiEBISBox.getIrregIVS(curPhiBox);
    notRegular |= curPhiEBISBox.getMultiCells(curPhiBox);

    for (VoFIterator vofit(notRegular,curPhiEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      RealVect x;

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        x[idir] = a_params.dx[idir] * (iv[idir] + 0.5)
                + a_params.probLo[idir];
      }

      Real onePoint;

      FORT_GETPHIPOINT(CHF_REAL(onePoint),
                       CHF_CONST_REALVECT(a_params.trig),
                       CHF_CONST_REALVECT(x));

      curPhiEBCellFAB(vof,comp) = onePoint;
    }
  }
}

void setKappaLOfPhi(LevelData<EBCellFAB>&    a_kappaLOfPhi,
                    const struct testParams& a_params)
{
  assert(a_kappaLOfPhi.nComp() == 1);

  int comp = 0;

  for (DataIterator dit = a_kappaLOfPhi.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& curKappaLOfPhiEBCellFAB = a_kappaLOfPhi[dit()];
    Box curKappaLOfPhiBox = curKappaLOfPhiEBCellFAB.getRegion();
    FArrayBox& curKappaLOfPhiFAB = curKappaLOfPhiEBCellFAB.getFArrayBox();

    FORT_GETLOFPHI(CHF_FRA1(curKappaLOfPhiFAB,comp),
                   CHF_CONST_REALVECT(a_params.trig),
                   CHF_CONST_REALVECT(a_params.dx),
                   CHF_CONST_REALVECT(a_params.probLo),
                   CHF_CONST_REALVECT(a_params.probHi),
                   CHF_BOX(curKappaLOfPhiBox));

    const EBISBox& curKappaLOfPhiEBISBox = curKappaLOfPhiEBCellFAB.getEBISBox();

    const EBGraph& curKappaLOfPhiEBGraph = curKappaLOfPhiEBISBox.getEBGraph();

    IntVectSet notRegular;

    notRegular |= curKappaLOfPhiEBISBox.getIrregIVS(curKappaLOfPhiBox);
    notRegular |= curKappaLOfPhiEBISBox.getMultiCells(curKappaLOfPhiBox);

    for (VoFIterator vofit(notRegular,curKappaLOfPhiEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      RealVect x;
      // const RealVect& centroid = curKappaLOfPhiEBISBox.centroid(vof);

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        x[idir] = a_params.dx[idir] * (iv[idir] + 0.5)
                + a_params.probLo[idir];
      }

      Real onePoint;

      FORT_GETLOFPHIPOINT(CHF_REAL(onePoint),
                          CHF_CONST_REALVECT(a_params.trig),
                          CHF_CONST_REALVECT(x));

      Real volfrac = curKappaLOfPhiEBISBox.volFrac(vof);
      curKappaLOfPhiEBCellFAB(vof,comp) = volfrac * onePoint;
    }
  }
}

void printNorms(Real&                       a_errorMaxNorm,
                const Real&                 a_lastErrorMaxNorm,
                const int&                  a_nCells,
                const int&                  a_iter,
                const LevelData<EBCellFAB>& a_resid,
                const LevelData<EBCellFAB>& a_error,
                EBLevelDataOps&             a_levelDataOps,
                const ProblemDomain&        a_domain)
{
  Real residMaxNorm;
  Real residOneNorm;
  Real residTwoNorm;

  Real errorMaxNorm;
  Real errorOneNorm;
  Real errorTwoNorm;

  Real errorEBMaxNorm;

  Real volume;

  residMaxNorm = a_levelDataOps.noKappaNorm(volume,a_resid,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,0);
  residOneNorm = a_levelDataOps.noKappaNorm(volume,a_resid,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,1);
  residTwoNorm = a_levelDataOps.noKappaNorm(volume,a_resid,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,2);

  errorMaxNorm = a_levelDataOps.noKappaNorm(volume,a_error,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,0);
  errorOneNorm = a_levelDataOps.noKappaNorm(volume,a_error,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,1);
  errorTwoNorm = a_levelDataOps.noKappaNorm(volume,a_error,
                                            EBLEVELDATAOPS_ALLVOFS,a_domain,2);

  errorEBMaxNorm = a_levelDataOps.noKappaNorm(volume,a_error,
                                              EBLEVELDATAOPS_INTERIORIRREGVOFS,a_domain,0);

  a_errorMaxNorm = errorMaxNorm;

  pout() << setw(4) << a_nCells << "  ";
  pout() << setw(2) << a_iter << ": ";

  pout() << "Resid norms: ";

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << residMaxNorm;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << residOneNorm;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << residTwoNorm;

  pout() << "  Error norms: ";

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << errorMaxNorm;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << errorOneNorm;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << errorTwoNorm;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << errorEBMaxNorm;

  if (a_lastErrorMaxNorm >= 0.0 && a_errorMaxNorm >= a_lastErrorMaxNorm)
  {
    pout() << " (final)" << endl;
  }
  else
  {
    pout() << endl;
  }
}
