/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cerr;

#include "ParmParse.H"
#include "LoadBalance.H"
#include "BRMeshRefine.H"

#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "EBCellFactory.H"
#include "AllRegularService.H"

#include "GeometryShop.H"
#include "ImplicitFunctionBL.H"
#include "PlaneIF.H"
#include "SphereIF.H"
#include "PolarIF.H"
#include "LatheIF.H"
#include "TransformIF.H"

#include "EBFABView.H"
#include "EBDebugDump.H"

#include "EBAMRPoissonOp.H"
#include "EBLevelDataOps.H"
#include "BaseBCValue.H"
#include "BaseBCFlux.H"
#include "BaseDomainBC.H"
#include "NeumannDomainBC.H"
#include "DirichletDomainBC.H"
#include "BaseEBBC.H"
#include "DirichletEBBC.H"
#include "NeumannEBBC.H"

#include "functionsF_F.H"
#include "JohanBCValue.H"
#include "JohanBCFlux.H"

#define BCNEUMANN       0   // Neumann boundary conditions
#define BCDIRICHLET1    1   // Dirichlet boundary conditions (1st order)
#define BCDIRICHLET2    2   // Dirichlet boundary conditions (2nd order)

struct testParams
{
  int      verbose;
  int      johanExpo;
  int      johanFreq;
  IntVect  nCells;
  int      maxBoxSize;
  int      blockFactor;
  RealVect probLo;
  RealVect probHi;
  RealVect dx;
  int      orderEB;
  int      ebBcType;
  int      domBcType;
  int      geometry;
  RealVect point;
  RealVect normal;
  Real     radius;
  RealVect center;
  int      interior;
  Real     baseRadius;
  RealVect polarCenter;
  Real     perturbation;
  int      frequency;
  int      polarInterior;
};

void truncOp(struct testParams& a_params);

void setPhi(LevelData<EBCellFAB>&    a_phi,
            const struct testParams& a_params);

void setKappaLOfPhi(LevelData<EBCellFAB>&    a_kappaLOfPhi,
                    const struct testParams& a_params);

int main(int argc, char* argv[])
{
  // Parameters
  struct testParams params;

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  // Scoping trick
  {
#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
    if (argc < 2)
    {
      cerr << " usage " << argv[0] << " <input_file_name> " << endl;
      exit(0);
    }

    char* inFile = argv[1];
    ParmParse pp(argc-2,argv+2,NULL,inFile);

    ParmParse ppMain("main");

    ppMain.get("verbose",params.verbose);

    ppMain.get("johanExpo",params.johanExpo);
    ppMain.get("johanFreq",params.johanFreq);

    ppMain.get("orderEB",params.orderEB);
    if (params.orderEB < 0 || params.orderEB > 1)
    {
      MayDay::Error("EB order not 0 or 1");
    }

    ppMain.get("ebBcType",params.ebBcType);
    if (params.ebBcType < 0 || params.ebBcType > 2)
    {
      MayDay::Error("BC type not 0, 1, or 2");
    }

    ppMain.get("domBcType",params.domBcType);
    if (params.domBcType < 0 || params.domBcType > 1)
    {
      MayDay::Error("BC type not 0 or 1");
    }

    std::vector<int> nCellsArray(SpaceDim);
    ppMain.getarr("nCells",nCellsArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.nCells[idir] = nCellsArray[idir];
    }

    ppMain.get("maxBoxSize",params.maxBoxSize);
    ppMain.get("blockFactor",params.blockFactor);

    std::vector<Real> probLoArray(SpaceDim);
    ppMain.getarr("probLo",probLoArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.probLo[idir] = probLoArray[idir];
    }

    std::vector<Real> probHiArray(SpaceDim);
    ppMain.getarr("probHi",probHiArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.probHi[idir] = probHiArray[idir];
    }

    ppMain.get("geometry",params.geometry);

    std::vector<Real> pointArray(SpaceDim);
    ppMain.getarr("point",pointArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.point[idir] = pointArray[idir];
    }

    std::vector<Real> normalArray(SpaceDim);
    ppMain.getarr("normal",normalArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.normal[idir] = normalArray[idir];
    }

    ppMain.get("radius",params.radius);

    std::vector<Real> centerArray(SpaceDim);
    ppMain.getarr("center",centerArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.center[idir] = centerArray[idir];
    }

    ppMain.get("interior",params.interior);

    ppMain.get("baseRadius",params.baseRadius);

    ppMain.getarr("polarCenter",centerArray,0,SpaceDim);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.polarCenter[idir] = centerArray[idir];
    }

    ppMain.get("perturbation",params.perturbation);

    ppMain.get("frequency",params.frequency);

    ppMain.get("polarInterior",params.polarInterior);

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      params.dx[idir] = (params.probHi[idir] - params.probLo[idir]) / params.nCells[idir];
    }

    truncOp(params);
  }
  // End scoping trick

#ifdef MPI
  dumpmemoryatexit();
  MPI_Finalize();
#endif
}

void truncOp(struct testParams& a_params)
{
  int nComps = 1;

  bool isPeriodic[SpaceDim];
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    isPeriodic[idir] = false;
  }

  IntVect ivLo = IntVect::TheZeroVector();
  IntVect ivHi;

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    ivHi[idir] = a_params.nCells[idir] - 1;
  }

  Box domainBox(ivLo,ivHi);
  ProblemDomain domain(domainBox,isPeriodic);

  Vector<Box> boxes;
  domainSplit(domain,boxes,a_params.maxBoxSize,a_params.blockFactor);

  Vector<int> procIDs(boxes.size());
  LoadBalance(procIDs,boxes);

  DisjointBoxLayout dbl(boxes,procIDs,domain);

  BaseImplicitFunction *baseIF;

  if (a_params.geometry == 0)
  {
    RealVect point = a_params.point;
    point += a_params.probLo;

    bool isInterior = (a_params.interior == 1);

    baseIF = new PlaneIF(a_params.normal,point,isInterior);
  }
  else
  if (a_params.geometry == 1)
  {
    RealVect center = a_params.center;
    center += a_params.probLo;

    bool isInterior = (a_params.interior == 1);

    baseIF = new SphereIF(a_params.radius,center,isInterior);
  }
  else
  if (a_params.geometry == 2)
  {
    RealVect center = a_params.polarCenter;
    center += a_params.probLo;

    bool isInterior = (a_params.polarInterior == 1);

    if (SpaceDim == 2)
    {
      PolarIF polar(a_params.baseRadius,a_params.perturbation,
                    a_params.frequency,isInterior);

      TransformIF* moveit = new TransformIF(polar);
      moveit->translate(center);

      baseIF = moveit;
    }
    else
    {
      bool interior = true;
      PolarIF polar(a_params.baseRadius,a_params.perturbation,
                    a_params.frequency,interior);

      LatheIF rotate(polar,isInterior);

      TransformIF* moveit = new TransformIF(rotate);
      moveit->translate(center);

      baseIF = moveit;
    }
  }
  else
  {
    MayDay::Error("Unknown geometry specified");
  }

  ImplicitFunctionBL geometry(*baseIF);
  GeometryShop geometryShop(geometry);

  EBIndexSpace* space = Chombo_EBIS::instance();
  space->define(domainBox,a_params.probLo,a_params.dx[0],geometryShop,
                a_params.nCells[0]);

  RefCountedPtr<BaseDomainBC> baseDomainBC = NULL;

  if (a_params.domBcType == 0)
  {
    JohanBCFlux* johanFluxPtr = new JohanBCFlux();
    johanFluxPtr->define(a_params.johanExpo,a_params.johanFreq);

    RefCountedPtr<BaseBCFlux> baseJohanFluxPtr = johanFluxPtr;

    NeumannDomainBC* domainBC = new NeumannDomainBC();
    domainBC->setFunction(baseJohanFluxPtr);

    baseDomainBC = domainBC;
  }
  else if (a_params.domBcType == 1)
  {
    JohanBCValue* johanValuePtr = new JohanBCValue();
    johanValuePtr->define(a_params.johanExpo,a_params.johanFreq);

    RefCountedPtr<BaseBCValue> baseJohanValuePtr = johanValuePtr;

    DirichletDomainBC* domainBC = new DirichletDomainBC();
    domainBC->setFunction(baseJohanValuePtr);

    baseDomainBC = domainBC;
  }
  else
  {
    MayDay::Error("Unknown domain BC type");
  }

  EBISLayout layout;
  RefCountedPtr<BaseEBBC> baseEBBC = NULL;
  int numGhostCells;

  if (a_params.ebBcType == 0)
  {
    JohanBCFlux* johanFluxPtr = new JohanBCFlux();
    johanFluxPtr->define(a_params.johanExpo,a_params.johanFreq);

    RefCountedPtr<BaseBCFlux> baseJohanFluxPtr = johanFluxPtr;

    numGhostCells = 1;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    NeumannEBBC* ebBC = new NeumannEBBC(domain,layout,a_params.dx);
    ebBC->setFunction(baseJohanFluxPtr);

    baseEBBC = ebBC;
  }
  else if (a_params.ebBcType == 1)
  {
    JohanBCValue* johanValuePtr = new JohanBCValue();
    johanValuePtr->define(a_params.johanExpo,a_params.johanFreq);

    RefCountedPtr<BaseBCValue> baseJohanValuePtr = johanValuePtr;

    numGhostCells = 1;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    DirichletEBBC* ebBC = new DirichletEBBC(domain,layout,a_params.dx);
    ebBC->setOrder(1);
    ebBC->setFunction(baseJohanValuePtr);

    baseEBBC = ebBC;
  }
  else if (a_params.ebBcType == 2)
  {
    JohanBCValue* johanValuePtr = new JohanBCValue();
    johanValuePtr->define(a_params.johanExpo,a_params.johanFreq);

    RefCountedPtr<BaseBCValue> baseJohanValuePtr = johanValuePtr;

    numGhostCells = 3;
    space->fillEBISLayout(layout,dbl,domainBox,numGhostCells);

    DirichletEBBC* ebBC = new DirichletEBBC(domain,layout,a_params.dx);
    ebBC->setOrder(2);
    ebBC->setFunction(baseJohanValuePtr);

    baseEBBC = ebBC;
  }
  else
  {
    MayDay::Error("Unknown EB BC type");
  }

  RefCountedPtr<DataFactory<EBCellFAB> > rcpFactory(new EBCellFactory(layout));
  LevelData<EBCellFAB> phi(dbl,nComps,numGhostCells*IntVect::TheUnitVector(),
                           *rcpFactory);

  setPhi(phi,a_params);
  // writeEBLevelname(&phi,"phi.hdf5");

  LevelData<EBCellFAB> kappaLOfPhi(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),
                                   *rcpFactory);

  EBAMRPoissonOp levelOp;
  bool isHomogeneous = false;

  levelOp.define(domain,layout,a_params.probLo,a_params.dx,
                 a_params.orderEB,baseDomainBC,baseEBBC,rcpFactory);
  levelOp.applyOp(kappaLOfPhi,phi,isHomogeneous);
  // writeEBLevelname(&kappaLOfPhi,"lphic.hdf5");

  LevelData<EBCellFAB> error(dbl,nComps,numGhostCells*IntVect::TheZeroVector(),
                             *rcpFactory);
  EBLevelDataOps levelDataOps;

  setKappaLOfPhi(error,a_params);
  // writeEBLevelname(&error,"lphie.hdf5");
  levelDataOps.incr(error,kappaLOfPhi,-1.0);
#ifdef HDF5
  writeEBLevelname(&error,"error.hdf5");
#endif
  Real volume;

  Real maxNormAll;
  maxNormAll = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_ALLVOFS          ,domain,0);

  Real maxNormEB;
  maxNormEB  = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS |
                                                     EBLEVELDATAOPS_INTERIORIRREGVOFS,domain,0);

  Real maxNormInt;
  maxNormInt = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS  ,domain,0);

  Real oneNormAll;
  oneNormAll = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_ALLVOFS          ,domain,1);

  Real oneNormEB;
  oneNormEB  = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS |
                                                     EBLEVELDATAOPS_INTERIORIRREGVOFS,domain,1);

  Real oneNormInt;
  oneNormInt = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS  ,domain,1);

  Real twoNormAll;
  twoNormAll = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_ALLVOFS          ,domain,2);

  Real twoNormEB;
  twoNormEB  = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS |
                                                     EBLEVELDATAOPS_INTERIORIRREGVOFS,domain,2);

  Real twoNormInt;
  twoNormInt = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORREGVOFS  ,domain,2);

  Real oneNormOnlyEB;
  oneNormOnlyEB  = levelDataOps.noKappaNorm(volume,error,EBLEVELDATAOPS_INTERIORIRREGVOFS,domain,1);

  pout() << setw(4) << a_params.nCells[0];

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << maxNormAll;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << maxNormEB;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << maxNormInt;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << oneNormAll;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << oneNormEB;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << oneNormInt;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << twoNormAll;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << twoNormEB;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << twoNormInt;

  pout() << "  "
         << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific)
         << oneNormOnlyEB;

  pout() << endl;
  pout() << endl;
}

void setPhi(LevelData<EBCellFAB>&    a_phi,
            const struct testParams& a_params)
{
  assert(a_phi.nComp() == 1);

  int comp = 0;

  for (DataIterator dit = a_phi.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& curPhiEBCellFAB = a_phi[dit()];
    Box curPhiBox = curPhiEBCellFAB.getRegion();
    FArrayBox& curPhiFAB = curPhiEBCellFAB.getFArrayBox();

    FORT_GETPHI(CHF_FRA1(curPhiFAB,comp),
                CHF_CONST_INT(a_params.johanExpo),
                CHF_CONST_INT(a_params.johanFreq),
                CHF_CONST_REALVECT(a_params.dx),
                CHF_CONST_REALVECT(a_params.probLo),
                CHF_CONST_REALVECT(a_params.probHi),
                CHF_BOX(curPhiBox));

    const EBISBox& curPhiEBISBox = curPhiEBCellFAB.getEBISBox();
    const EBGraph& curPhiEBGraph = curPhiEBISBox.getEBGraph();

    IntVectSet notRegular;

    notRegular |= curPhiEBISBox.getIrregIVS(curPhiBox);
    notRegular |= curPhiEBISBox.getMultiCells(curPhiBox);

    for (VoFIterator vofit(notRegular,curPhiEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      RealVect x;

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        x[idir] = a_params.dx[idir] * (iv[idir] + 0.5)
                + a_params.probLo[idir];
      }

      Real onePoint;

      FORT_GETPHIPOINT(CHF_REAL(onePoint),
                       CHF_CONST_INT(a_params.johanExpo),
                       CHF_CONST_INT(a_params.johanFreq),
                       CHF_CONST_REALVECT(x));

      curPhiEBCellFAB(vof,comp) = onePoint;
    }
  }
}

void setKappaLOfPhi(LevelData<EBCellFAB>&    a_kappaLOfPhi,
                    const struct testParams& a_params)
{
  assert(a_kappaLOfPhi.nComp() == 1);

  int comp = 0;

  for (DataIterator dit = a_kappaLOfPhi.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& curKappaLOfPhiEBCellFAB = a_kappaLOfPhi[dit()];
    Box curKappaLOfPhiBox = curKappaLOfPhiEBCellFAB.getRegion();
    FArrayBox& curKappaLOfPhiFAB = curKappaLOfPhiEBCellFAB.getFArrayBox();

    FORT_GETLOFPHI(CHF_FRA1(curKappaLOfPhiFAB,comp),
                   CHF_CONST_INT(a_params.johanExpo),
                   CHF_CONST_INT(a_params.johanFreq),
                   CHF_CONST_REALVECT(a_params.dx),
                   CHF_CONST_REALVECT(a_params.probLo),
                   CHF_CONST_REALVECT(a_params.probHi),
                   CHF_BOX(curKappaLOfPhiBox));

    const EBISBox& curKappaLOfPhiEBISBox = curKappaLOfPhiEBCellFAB.getEBISBox();

    const EBGraph& curKappaLOfPhiEBGraph = curKappaLOfPhiEBISBox.getEBGraph();

    IntVectSet notRegular;

    notRegular |= curKappaLOfPhiEBISBox.getIrregIVS(curKappaLOfPhiBox);
    notRegular |= curKappaLOfPhiEBISBox.getMultiCells(curKappaLOfPhiBox);

    for (VoFIterator vofit(notRegular,curKappaLOfPhiEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      RealVect x;
      // const RealVect& centroid = curKappaLOfPhiEBISBox.centroid(vof);

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        x[idir] = a_params.dx[idir] * (iv[idir] + 0.5 /* + centroid[idir] */)
                + a_params.probLo[idir];
      }

      Real onePoint;

      FORT_GETLOFPHIPOINT(CHF_REAL(onePoint),
                          CHF_CONST_INT(a_params.johanExpo),
                          CHF_CONST_INT(a_params.johanFreq),
                          CHF_CONST_REALVECT(x));

      Real volfrac = curKappaLOfPhiEBISBox.volFrac(vof);
      curKappaLOfPhiEBCellFAB(vof,comp) = volfrac * onePoint;
    }
  }
}
