/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//Chombo includes
#include "CH_HDF5.H"
#include "ParmParse.H"
#include "parstream.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "ParmParse.H"
#include "LevelData.H"
#include "EBArith.H"
#include "memtrack.H"

//EBChombo includes
#include "EBIndexSpace.H"
#include "GeometryShop.H"
#include "SphereBL.H"
#include "EBCellFactory.H"
#include "EBAMRIO.H"

//application-specific includes
#include "TiltedCylinderIF.H"
#include "ImplicitFunctionBL.H"
#include "EBBumpIBCFactory.H"
#include "EBPatchScalarAdvect.H"
#include "EBPatchScalarAdvectFactory.H"
#include "EBAMRScalarAdvectFactory.H"
#include "EBAMRScalarAdvect.H"
#include "GodunovGeom.H"
#include "ConstVel.H"

//debug includes
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "ArrayView.H"
#include "EBArrayView.H"


RealVect g_constVel;
/***************/
void
makeFinestDomain(Box& a_domain, 
                 Real& a_dx)
{
  //parse input file.  single level
  ParmParse pp;
  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);
 
  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain = Box(lo, hi);
  
  Real prob_hi;
  int numOpen = n_cell[0];
  pp.get("domain_length",prob_hi);
  a_dx = prob_hi/numOpen;
}
/************/
void
makeGeometry(const Box& a_domain, 
             const Real& a_dx)
{

  RealVect origin = RealVect::Zero;
  ParmParse pp;
  Real channelRadius;
  pp.get("cylinder_radius", channelRadius);

  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  int biggridsize = 2048;
  pp.get("max_grid_size", biggridsize);
  pout() << "using a tilted cylinder" << endl;
  vector<Real>  cylinderAxisVect(SpaceDim);
  pp.getarr("cylinder_axis",cylinderAxisVect, 0, SpaceDim);
  RealVect cylinderAxis;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      cylinderAxis[idir] = cylinderAxisVect[idir];
    }
  Real sum;
  PolyGeom::unifyVector(cylinderAxis, sum);
  g_constVel = cylinderAxis;
  RealVect corner = RealVect::Zero;
  bool negativeInside = true;
  TiltedCylinderIF tunnel(channelRadius, cylinderAxis, corner, negativeInside);
  ImplicitFunctionBL func(tunnel);
  GeometryShop workshop(func);
  ebisPtr->define(a_domain, origin, a_dx, workshop, biggridsize);
}
/***************/
void
compareError(const EBCellFAB& a_errorFine,
             const EBISBox& a_ebisBoxFine, 
             const EBCellFAB& a_errorCoar,
             const EBISBox& a_ebisBoxCoar)
{
  Box gridFine = a_errorFine.getRegion();
  Box gridCoar = a_errorCoar.getRegion();
  pout() << "==============================================" << endl;
  for(int comp = 0; comp < a_errorFine.nComp(); comp++)
    {
      pout() << "Comparing error in variable  " << comp << endl;
      pout() << "==============================================" << endl;
      for(int itype = 0; itype < 3; itype++)
        {
          EBNormType::NormMode normtype;
          if(itype == 0)
            {
              normtype = EBNormType::OverBoth;
              pout() << endl << "Using all uncovered cells." << endl;
            }
          else if(itype == 1)
            {
              normtype = EBNormType::OverOnlyRegular;
              pout() << endl << "Using only regular cells." << endl;
            }
          else
            {
              normtype = EBNormType::OverOnlyIrregular;
              pout() << endl << "Using only irregular cells." << endl;
            }
          for(int inorm = 0; inorm < 3; inorm++)
            {

              if(inorm == 0)
                {
                  pout() << endl << "Using max norm." << endl;
                }
              else
                {
                  pout() << endl << "Using L-" << inorm << "norm." << endl;
                }
              Real coarnorm = EBArith::norm(a_errorCoar, 
                                            gridCoar, a_ebisBoxCoar, 
                                            comp, inorm, normtype);
              Real finenorm = EBArith::norm(a_errorFine, 
                                            gridFine, a_ebisBoxFine, 
                                            comp, inorm, normtype);

              pout() << "Coarse Error Norm = " << coarnorm << endl;
              pout() << "Fine   Error Norm = " << finenorm << endl;
              if((Abs(finenorm) > 1.0e-8) && (Abs(coarnorm) > 1.0e-8))
                {
                  Real order = log(coarnorm/finenorm)/log(2.0);
                  pout() << "Order of scheme = " << order << endl;
                }
            }
        }
      pout() << "==============================================" << endl;
    }
}
/************/
Real
exactUdelr(const RealVect& a_xval)
{
  Real retval = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real x = a_xval[idir];
      //cubic density
      retval += 3.0*x*x*g_constVel[idir];
    }

  //debug
  //  Real x0 = a_xval[0];
  //  retval = 3.0*x0*x0*g_constVel[0];
  //retval = g_constVel[0];
  //end debug

  return retval;
}
/************/
Real
exactDens(const RealVect& a_xval)
{
  Real retval = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real x = a_xval[idir];
      retval +=  x*x*x;
    }
  //debug
  //Real x0 = a_xval[0];
  //  retval = x0*x0*x0;
  //retval = x0;
  //end debug

  return retval;
}
/************/
void
setToExactUdelr(EBCellFAB&     a_exactUdelr, 
                const EBISBox& a_ebisBox, 
                const Real&    a_dx,
                const Box&     a_region)
{
  a_exactUdelr.setVal(0.);
  IntVectSet ivsregion(a_region);
  for(VoFIterator vofit(ivsregion, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      RealVect xval;
      IntVect iv = vof.gridIndex();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
        }
      Real udelrrv = exactUdelr(xval);
      a_exactUdelr(vof, 0) = udelrrv;
    }
}
/************/
void
setToExactDens(EBFluxFAB&                        a_dens,
               Vector< BaseIVFAB<Real> * >&      a_coveredDensMinu, 
               Vector< BaseIVFAB<Real> * >&      a_coveredDensPlus, 
               const Vector< Vector<VolIndex> >& a_coveredFaceMinu,  
               const Vector< Vector<VolIndex> >& a_coveredFacePlus,  
               const EBISBox&                    a_ebisBox, 
               const Box&                        a_region,
               const Real&                       a_dx)
{
  IntVectSet ivsregion(a_region);
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      EBFaceFAB& densDir = a_dens[faceDir];
      densDir.setVal(0.);
      for(FaceIterator faceit(ivsregion, a_ebisBox.getEBGraph(), faceDir, FaceStop::SurroundingWithBoundary); 
          faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          RealVect xval;
          IntVect iv = face.gridIndex(Side::Hi);
      
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              if(idir != faceDir)
                {
                  xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
                }
              else
                {
                  xval[idir] = (Real(iv[idir]))*a_dx;
                }
            }
          Real densrv = exactDens(xval);
          densDir(face, 0) = densrv;
          
        } //end loop over faces
      
      const Vector<VolIndex>& covFaceMinu = a_coveredFaceMinu[faceDir];
      BaseIVFAB<Real>*  covDensMinuPtr = a_coveredDensMinu[faceDir];
      BaseIVFAB<Real>&  covDensMinu = *covDensMinuPtr;
      for(int ivof = 0; ivof < covFaceMinu.size(); ivof++)
        {
          const VolIndex& vof = covFaceMinu[ivof];
          RealVect xval;
          IntVect iv = vof.gridIndex();
      
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              if(idir != faceDir)
                {
                  xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
                }
              else
                {
                  xval[idir] = (Real(iv[idir]))*a_dx;
                }
            }
          Real densrv = exactDens(xval);
          covDensMinu(vof, 0) = densrv;
        }
          
      const Vector<VolIndex>& covFacePlus = a_coveredFacePlus[faceDir];
      BaseIVFAB<Real>&  covDensPlus = *a_coveredDensPlus[faceDir];
      for(int ivof = 0; ivof < covFacePlus.size(); ivof++)
        {
          const VolIndex& vof = covFacePlus[ivof];
          RealVect xval;
          IntVect iv = vof.gridIndex();
      
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              if(idir != faceDir)
                {
                  xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
                }
              else
                {
                  xval[idir] = (Real(iv[idir]) + 1.0)*a_dx;
                }
            }
          Real densrv = exactDens(xval);
          covDensPlus(vof, 0) = densrv;
        }
    } //end loop over face directions
}
/************/
void 
divUAutopsy()
{
  //make layouts == domain
  Box domainBoxFine, domainBoxCoar;
  Real dxFine, dxCoar;
  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  makeFinestDomain(domainBoxFine, dxFine);
  domainBoxCoar = coarsen(domainBoxFine, 2);
  dxCoar = 2.*dxFine;

  Vector<Box> boxFine(1, domainBoxFine);
  Vector<Box> boxCoar(1, domainBoxCoar);
  Vector<int> proc(1, 0);

  DisjointBoxLayout dblFine(boxFine, proc);
  DisjointBoxLayout dblCoar;
  coarsen(dblCoar, dblFine, 2);

  //make ebislayouts
  //no need for ghost since we are using the domain
  //for the box

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  makeGeometry(domainBoxFine, dxFine);
  pout() << "making fine geometry" << endl;
  EBISLayout ebislFine, ebislCoar;
  ebisPtr->fillEBISLayout(ebislFine, dblFine, domainBoxFine, 0);

  //debug
  dxFine= 1.0;
  dxCoar = 2.*dxFine;
  //end debug

  pout() << "making coarse geometry" << endl;
  makeGeometry(domainBoxCoar, dxCoar);
  ebisPtr->fillEBISLayout(ebislCoar, dblCoar, domainBoxCoar, 0);
  
  for(DataIterator dit = dblFine.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBoxFine = ebislFine[dit()];
      const EBISBox& ebisBoxCoar = ebislCoar[dit()];

      EBCellFAB errorFine(ebisBoxFine, domainBoxFine, 1);
      EBCellFAB errorCoar(ebisBoxCoar, domainBoxCoar, 1);
      EBCellFAB exactFine(ebisBoxFine, domainBoxFine, 1);
      EBCellFAB exactCoar(ebisBoxCoar, domainBoxCoar, 1);
      EBCellFAB udelrFine(ebisBoxFine, domainBoxFine, 1);
      EBCellFAB udelrCoar(ebisBoxCoar, domainBoxCoar, 1);

      errorFine.setVal(0.);
      errorCoar.setVal(0.);
      exactFine.setVal(0.);
      exactCoar.setVal(0.);
      udelrFine.setVal(0.);
      udelrCoar.setVal(0.);


      setToExactUdelr(exactFine, ebisBoxFine, dxFine, domainBoxFine);
      setToExactUdelr(exactCoar, ebisBoxCoar, dxCoar, domainBoxCoar);

      //computed covered fluxes
      Vector< BaseIVFAB<Real> * >  coveredDensMinuFine(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredDensPlusFine(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredDensMinuCoar(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredDensPlusCoar(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredVeloMinuFine(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredVeloPlusFine(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredVeloMinuCoar(SpaceDim, NULL);
      Vector< BaseIVFAB<Real> * >  coveredVeloPlusCoar(SpaceDim, NULL);

      //vofs over which covered vofs live
      Vector< Vector<VolIndex>  > coveredFaceMinuFine(SpaceDim);
      Vector< Vector<VolIndex>  > coveredFacePlusFine(SpaceDim);
      Vector< Vector<VolIndex>  > coveredFaceMinuCoar(SpaceDim);
      Vector< Vector<VolIndex>  > coveredFacePlusCoar(SpaceDim);

      Vector<IntVectSet> coveredSetsMinuFine(SpaceDim);
      Vector<IntVectSet> coveredSetsPlusFine(SpaceDim);
      Vector<IntVectSet> coveredSetsMinuCoar(SpaceDim);
      Vector<IntVectSet> coveredSetsPlusCoar(SpaceDim);

      EBPatchScalarAdvect patchCoar, patchFine;
      ProblemDomain domainFine(domainBoxFine);
      ProblemDomain domainCoar(domainBoxCoar);
      patchCoar.define(domainCoar, dxCoar);
      patchFine.define(domainFine, dxFine);
      IntVectSet cfivs;
      Real time = 0.;
      patchCoar.setValidBox(domainBoxCoar, ebisBoxCoar, cfivs, time, time);
      patchFine.setValidBox(domainBoxFine, ebisBoxFine, cfivs, time, time);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          IntVectSet ivs;
          patchFine.computeCoveredFaces(coveredFacePlusFine[idir], 
                                        coveredSetsPlusFine[idir], ivs,
                                        idir, Side::Hi, domainBoxFine);
          patchFine.computeCoveredFaces(coveredFaceMinuFine[idir], 
                                        coveredSetsMinuFine[idir], ivs,
                                        idir, Side::Lo, domainBoxFine);
          patchCoar.computeCoveredFaces(coveredFacePlusCoar[idir], 
                                        coveredSetsPlusCoar[idir], ivs,
                                        idir, Side::Hi, domainBoxCoar);
          patchCoar.computeCoveredFaces(coveredFaceMinuCoar[idir], 
                                        coveredSetsMinuCoar[idir], ivs,
                                        idir, Side::Lo, domainBoxCoar);

          coveredDensPlusFine[idir]= new BaseIVFAB<Real>(coveredSetsPlusFine[idir], ebisBoxFine.getEBGraph(), 1);
          coveredDensMinuFine[idir]= new BaseIVFAB<Real>(coveredSetsMinuFine[idir], ebisBoxFine.getEBGraph(), 1);

          coveredDensPlusCoar[idir]= new BaseIVFAB<Real>(coveredSetsPlusCoar[idir], ebisBoxCoar.getEBGraph(), 1);
          coveredDensMinuCoar[idir]= new BaseIVFAB<Real>(coveredSetsMinuCoar[idir], ebisBoxCoar.getEBGraph(), 1);

          coveredVeloPlusFine[idir]= new BaseIVFAB<Real>(coveredSetsPlusFine[idir], ebisBoxFine.getEBGraph(), 1);
          coveredVeloMinuFine[idir]= new BaseIVFAB<Real>(coveredSetsMinuFine[idir], ebisBoxFine.getEBGraph(), 1);

          coveredVeloPlusCoar[idir]= new BaseIVFAB<Real>(coveredSetsPlusCoar[idir], ebisBoxCoar.getEBGraph(), 1);
          coveredVeloMinuCoar[idir]= new BaseIVFAB<Real>(coveredSetsMinuCoar[idir], ebisBoxCoar.getEBGraph(), 1);

          coveredDensPlusFine[idir]->setVal(0.);
          coveredDensMinuFine[idir]->setVal(0.);
          coveredDensPlusCoar[idir]->setVal(0.);
          coveredDensMinuCoar[idir]->setVal(0.);

        }

  
      EBFluxFAB densFine(ebisBoxFine, domainBoxFine, 1);
      EBFluxFAB densCoar(ebisBoxCoar, domainBoxCoar, 1);
      densFine.setVal(0.);
      densCoar.setVal(0.);

      EBFluxFAB veloFine(ebisBoxFine, domainBoxFine, 1);
      EBFluxFAB veloCoar(ebisBoxCoar, domainBoxCoar, 1);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          veloFine[idir].setVal(g_constVel[idir]);
          veloCoar[idir].setVal(g_constVel[idir]);

          coveredVeloPlusFine[idir]->setVal(g_constVel[idir]);
          coveredVeloMinuFine[idir]->setVal(g_constVel[idir]);
          coveredVeloPlusCoar[idir]->setVal(g_constVel[idir]);
          coveredVeloMinuCoar[idir]->setVal(g_constVel[idir]);
        }

      setToExactDens(densFine,
                     coveredDensMinuFine, coveredDensPlusFine, 
                     coveredFaceMinuFine, coveredFacePlusFine, 
                     ebisBoxFine, domainBoxFine, dxFine);

      setToExactDens(densCoar,
                     coveredDensMinuCoar, coveredDensPlusCoar, 
                     coveredFaceMinuCoar, coveredFacePlusCoar, 
                     ebisBoxCoar, domainBoxCoar, dxCoar);

      patchFine.advectiveDerivative(udelrFine, densFine, veloFine,
                                    coveredDensMinuFine, coveredDensPlusFine, 
                                    coveredVeloMinuFine, coveredVeloPlusFine, 
                                    coveredFaceMinuFine, coveredFacePlusFine, 
                                    domainBoxFine);

      patchCoar.advectiveDerivative(udelrCoar, densCoar, veloCoar,
                                    coveredDensMinuCoar, coveredDensPlusCoar, 
                                    coveredVeloMinuCoar, coveredVeloPlusCoar, 
                                    coveredFaceMinuCoar, coveredFacePlusCoar, 
                                    domainBoxCoar);
 
      Interval interv(0, 0);
      errorFine.copy(domainBoxFine, interv, domainBoxFine, exactFine, interv);
      errorFine -=   udelrFine;

      errorCoar.copy(domainBoxCoar, interv, domainBoxCoar, exactCoar, interv);
      errorCoar -=   udelrCoar;
      
      //do the whole convergence test thing.
      compareError(errorFine, ebisBoxFine,
                   errorCoar, ebisBoxCoar);

      //delete the news
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          delete coveredDensPlusFine[idir];
          delete coveredDensMinuFine[idir];
          delete coveredDensPlusCoar[idir];
          delete coveredDensMinuCoar[idir];
          delete coveredVeloPlusFine[idir];
          delete coveredVeloMinuFine[idir];
          delete coveredVeloPlusCoar[idir];
          delete coveredVeloMinuCoar[idir];
        }
    }
  
}
/************/
/************/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  {
    setChomboMPIErrorHandler();
#endif
    // Check for an input file
    char* inFile = NULL;

    if (a_argc > 1)
      {
        inFile = a_argv[1];
      }
    else
      {
        pout() << "Usage: <executable name> <inputfile>" << endl;
        pout() << "No input file specified" << endl;
        return -1;
      }
    ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

    divUAutopsy();

#ifdef MPI
    dumpmemoryatexit();
  }
  MPI_Finalize();
#endif
}

