/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//Chombo includes
#include "CH_HDF5.H"
#include "ParmParse.H"
#include "parstream.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "ParmParse.H"
#include "LevelData.H"
#include "EBArith.H"
#include "memtrack.H"

//EBChombo includes
#include "EBIndexSpace.H"
#include "GeometryShop.H"
#include "SphereBL.H"
#include "EBCellFactory.H"
#include "EBAMRIO.H"

//application-specific includes
#include "TiltedCylinderIF.H"
#include "ImplicitFunctionBL.H"
#include "EBBumpIBCFactory.H"
#include "EBPatchScalarAdvect.H"
#include "EBPatchScalarAdvectFactory.H"
#include "EBAMRScalarAdvectFactory.H"
#include "EBAMRScalarAdvect.H"
#include "GodunovGeom.H"
#include "ConstVel.H"

//debug includes
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "ArrayView.H"
#include "EBArrayView.H"

/***************/
Vector<string> getNames()
{
  EBPatchScalarAdvect patchInt;
  Vector<string> consNames  = patchInt.stateNames();
  return consNames;
}
/***************/
void 
generateCalculatedData(LevelData<EBCellFAB>& a_calcData,
                       Real&                 a_dt,
                       Real&                 a_finalTime,
                       DisjointBoxLayout&    a_grids,
                       EBISLayout&           a_ebisl,
                       const Box&            a_domain,
                       const Real&           a_dx,
                       const bool&           a_thisCalcCoar)
{
  // read inputs
  ParmParse pp;

  int verbosity;
  pp.get("verbosity",verbosity);
  assert(verbosity >= 0);

  int nstop = 0;
  pp.get("max_step",nstop);

  if(a_thisCalcCoar)
    {
      nstop /= 2;
    }
  int redistRad = 0;
  pp.get("redist_radius",redistRad);

  Real stopTime = 0.0;
  pp.get("max_time",stopTime);

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  //single level test
  int max_level = 0;
  std::vector<int> ref_ratios(1,1);
  std::vector<int> regrid_intervals(1,-1);
  Real refineThresh = 0.3;
  int block_factor = 1;
  int tagBufferSize = 3;
  int max_grid_size = 32;
  Real fill_ratio = 0.75;

  pp.get("max_grid_size",max_grid_size);
  int checkpoint_interval = 0;
  pp.get("checkpoint_interval",checkpoint_interval);

  int plot_interval = 0;
  pp.get("plot_interval",plot_interval);

  Real cfl = 0.8;
  pp.get("cfl",cfl);
  Real initialCFL = cfl;

  //this app always has fixed dt
  Real fixed_dt = -1;
  Real max_dt_growth = 1.1;

  Real dt_tolerance_factor = 1.1;
  Real domainLength;
  pp.get("domain_length",domainLength);

  ProblemDomain prob_domain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  
  // create initial/boundary condition object
  
  pout() << "blob initial conditions " << endl;

  vector<Real> centerpp(SpaceDim,0.5);
  vector<Real> axispp(SpaceDim,0.5);
  RealVect center, axis;
  pp.getarr("bump_center",centerpp,0,SpaceDim);
  pp.getarr("cylinder_axis",axispp,0,SpaceDim);
  Real sum;
  PolyGeom::unifyVector(axis, sum);
  for (int i = 0; i < SpaceDim; i++) 
    {
      center[i] = centerpp[i];
      axis[i]   = axispp[i];
    }
  Real size, rnot, rone;
  pp.get("bump_size",size);
  pp.get("bump_rnot",rnot);
  pp.get("bump_rone",rone);
  Real rotVel;
  pp.get("velocity_constant", rotVel);
  RealVect velocity = axis;
  axis *= rotVel;

  EBBumpIBCFactory bcfactory(center, axis, velocity, size, rnot, rone);

  int idosmushing;
  pp.get("do_smushing", idosmushing);
  bool doSmushing    = (idosmushing ==1);

  //create patch integrator
  bool useLimiting = false;
  EBPatchScalarAdvectFactory   patchGamma(&bcfactory, useLimiting);
  pout() << "using constant velocity function" << endl;

  ConstVelFactory velFactory(axis, rotVel);
  EBAMRScalarAdvectFactory fact(initialCFL,
                                cfl,
                                redistRad,
                                domainLength,
                                refineThresh,
                                tagBufferSize,
                                verbosity,
                                doSmushing,
                                &velFactory,
                                &patchGamma);
  

  AMR amr;
  amr.define(max_level,ref_ratios,prob_domain,&fact);

  if (fixed_dt > 0)
    {
      amr.fixedDt(fixed_dt);
    }

  // set grid generation parameters
  amr.maxGridSize(max_grid_size);
  amr.blockFactor(block_factor);
  amr.fillRatio(fill_ratio);
  //  int gridBufferSize =1;

  // set output parameters
  amr.checkpointInterval(checkpoint_interval);
  amr.plotInterval(plot_interval);
  amr.regridIntervals(regrid_intervals);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  std::string pltprefix;
  std::string chkprefix;
  if(a_thisCalcCoar)
    {
      pltprefix =  std::string("pltCoar");
      chkprefix =  std::string("chkCoar");
    }
  else
    {
      pltprefix =  std::string("pltFine");
      chkprefix =  std::string("chkFine");
    }
  amr.plotPrefix(pltprefix);
  amr.checkpointPrefix(chkprefix);

  amr.verbosity(verbosity);

  amr.setupForNewAMRRun();

  // run 
  amr.run(stopTime,nstop);

  // output last pltfile and statistics
  //cleanup
  amr.conclude();

  Vector<AMRLevel*> amrLevelData = amr.getAMRLevels();
  assert(amrLevelData.size() == 1);
 
  EBAMRScalarAdvect* amrScalarAdvectData = dynamic_cast<EBAMRScalarAdvect*>(amrLevelData[0]);
  LevelData<EBCellFAB>& amrData = amrScalarAdvectData->getStateNew();

  a_grids = amrData.disjointBoxLayout();
  a_finalTime = amr.getCurrentTime();
  a_dt = amrScalarAdvectData->getDt();
  a_ebisl = amrScalarAdvectData->getEBISLayout();

  EBPatchScalarAdvect patch;
  int ncomp = patch.numConserved();
  Interval interv(0, ncomp-1);
  EBCellFactory ebcellFact(a_ebisl);
  a_calcData.define(a_grids, ncomp, IntVect::Zero, ebcellFact);
  amrData.copyTo(interv, a_calcData,  interv);
}
/************/
void
makeGeometry(const Box& a_domain, 
             const Real& a_dx)
{

  RealVect origin = RealVect::Zero;
  ParmParse pp;
  Real channelRadius;
  pp.get("cylinder_radius", channelRadius);

  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  int biggridsize = 2048;
  pp.get("max_grid_size", biggridsize);
  pout() << "using a tilted cylinder" << endl;
  vector<Real>  cylinderAxisVect(SpaceDim);
  pp.getarr("cylinder_axis",cylinderAxisVect, 0, SpaceDim);
  RealVect cylinderAxis;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      cylinderAxis[idir] = cylinderAxisVect[idir];
    }
  Real sum;
  PolyGeom::unifyVector(cylinderAxis, sum);

  RealVect corner = RealVect::Zero;
  bool negativeInside = true;
  TiltedCylinderIF tunnel(channelRadius, cylinderAxis, corner, negativeInside);
  ImplicitFunctionBL func(tunnel);
  GeometryShop workshop(func);
  ebisPtr->define(a_domain, origin, a_dx, workshop, biggridsize);
}
/***************/
void 
generateExactData(LevelData<EBCellFAB>& a_exactData,
                  const DisjointBoxLayout& a_grids,
                  const EBISLayout&        a_ebisl,
                  const Box& a_domain, 
                  const Real& a_dx,
                  const Real& a_finalTime,
                  const Real& a_dt,
                  const int&  a_nvar)
{
  ParmParse pp;
  EBCellFactory ebcf(a_ebisl);
  a_exactData.define(a_grids, a_nvar, IntVect::Zero, ebcf);


  vector<Real> centerpp(SpaceDim,0.5);
  vector<Real> axispp(SpaceDim,0.5);
  RealVect center, axis;
  pp.getarr("bump_center",centerpp,0,SpaceDim);
  pp.getarr("cylinder_axis",axispp,0,SpaceDim);
  Real sum;
  PolyGeom::unifyVector(axis, sum);
  for (int i = 0; i < SpaceDim; i++) 
    {
      center[i] = centerpp[i];
      axis[i]   = axispp[i];
    }
  Real size, rnot, rone;
  pp.get("bump_size",size);
  pp.get("bump_rnot",rnot);
  pp.get("bump_rone",rone);
  Real rotVel;
  pp.get("velocity_constant", rotVel);
  RealVect velocity = axis;
  axis *= rotVel;

  EBBumpIBC  bcfactory(center, axis, velocity, size, rnot, rone);
  bcfactory.define(a_domain, a_dx);
  bcfactory.setToExact(a_exactData, a_ebisl, a_finalTime);
}
/***************/
void
generateError(LevelData<EBCellFAB>& a_error,
              DisjointBoxLayout&    a_grids,
              EBISLayout&           a_ebisl,
              const Box&            a_domain, 
              const Real&           a_dx,
              const bool&           a_thisCalcCoar)
{
  LevelData<EBCellFAB> calc;
  LevelData<EBCellFAB> exac;

  Real finalTime, dt;
  generateCalculatedData(calc, dt,
                         finalTime, a_grids, a_ebisl,
                         a_domain, a_dx, a_thisCalcCoar);


  EBPatchScalarAdvect patchInt;
  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }
  ProblemDomain probdomain(a_domain.smallEnd(), a_domain.bigEnd(), is_periodic);
  patchInt.define(probdomain, a_dx);

  int nCons = patchInt.numConserved();
  generateExactData( exac, a_grids, a_ebisl, 
                     a_domain, a_dx, finalTime, dt, nCons);


  EBCellFactory ebcf(a_ebisl);

  a_error.define(a_grids, nCons, IntVect::Zero, ebcf);

  Interval interv(0, nCons-1);
  for(DataIterator dit = a_grids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& box = a_grids.get(dit());
      //error is exact-calculated data 
      a_error[dit()].copy(box, interv, box, exac[dit()], interv);
      a_error[dit()] -= calc[dit()];

      for(int icomp = 0; icomp < a_error[dit()].nComp(); icomp++)
        {
          a_error[dit()].setCoveredCellVal(0.0, icomp);
        }
    }
}
/***************/
int iscript(int icomp, int inorm, int ncomp)
{
  return icomp + inorm*ncomp;
}
/***************/
void
compareError(const LevelData<EBCellFAB>& a_errorFine,
             const EBISLayout& a_ebislFine, 
             const DisjointBoxLayout& a_gridsFine, 
             const Box& a_domainFine,
             const LevelData<EBCellFAB>& a_errorCoar,
             const EBISLayout& a_ebislCoar, 
             const DisjointBoxLayout& a_gridsCoar, 
             const Box& a_domainCoar)
{
  Vector<string> names  = getNames();

  const int ncomp = names.size();
  const int nnorm = 3;
  Real* normsCoar = new Real[ncomp*nnorm];
  Real* normsFine = new Real[ncomp*nnorm];
  Real* orders    = new Real[ncomp*nnorm];
  for(int icomp = 0; icomp < ncomp; icomp++)
    {
      orders[icomp] = 0.0;
      for(int inorm = 0; inorm < nnorm; inorm++)
        {
          normsCoar[iscript(icomp, inorm, ncomp)] = 0;
          normsFine[iscript(icomp, inorm, ncomp)] = 0;
        }
    }

  EBCellFactory factFine(a_ebislFine);
  EBCellFactory factCoar(a_ebislCoar);
  IntVect ivghost = IntVect::Zero;
  ParmParse pp;
  pout() << "==============================================" << endl;
  for(int comp = 0; comp < names.size(); comp++)
    {
      pout() << "Comparing error in variable  " << names[comp] << endl;
      pout() << "==============================================" << endl;
      for(int itype = 0; itype < 3; itype++)
        {
          EBNormType::NormMode normtype;
          if(itype == 0)
            {
              normtype = EBNormType::OverBoth;
              pout() << endl << "Using all uncovered cells." << endl  ;
            }
          else if(itype == 1)
            {
              normtype = EBNormType::OverOnlyRegular;
              pout() << endl << "Using only regular cells." << endl ;
            }
          else
            {
              normtype = EBNormType::OverOnlyIrregular;
              pout() << endl << "Using only irregular cells." << endl;
            }
          
          for(int inorm = 0; inorm <= 2; inorm++)
            {

              if(inorm == 0)
                {
                  pout() << endl << "Using max norm." << endl;
                }
              else
                {
                  pout() << endl << "Using L-" << inorm << "norm." << endl;
                }
              Real coarnorm = EBArith::norm(a_errorCoar, 
                                            a_gridsCoar, a_ebislCoar, 
                                            comp, inorm, normtype);
              Real finenorm = EBArith::norm(a_errorFine, 
                                            a_gridsFine, a_ebislFine, 
                                            comp, inorm, normtype);
              pout() << "Coarse Error Norm = " << coarnorm << endl;
              pout() << "Fine   Error Norm = " << finenorm << endl;
              if(itype == 0)
                {
                  normsCoar[iscript(comp,inorm,ncomp)] = coarnorm;
                  normsFine[iscript(comp,inorm,ncomp)] = finenorm;
                }
              if((Abs(finenorm) > 1.0e-12) && (Abs(coarnorm) > 1.0e-12)) 
                {
                  Real order = log(Abs(coarnorm/finenorm))/log(2.0);
                  pout() << "Order of scheme = " << order << endl;
                  if(itype == 0)
                    {
                      orders[iscript(comp,inorm,ncomp)] = order;
                    }
                }
            }
        }
      pout() << "==============================================" << endl ;;
    }


  //output in latex format to be safe
  int nfine = a_domainFine.size(0);
  pout() << setw(12)
         << setprecision(6)
         << setiosflags(ios::showpoint)
         << setiosflags(ios::scientific) ;
  for (int inorm = 0; inorm <= 2; inorm++)
    {
      pout() << "\\begin{table}[p]" << endl;
      pout() << "\\begin{center}" << endl;
      pout() << "\\begin{tabular}{|c|c|c|c|} \\hline" << endl;
      pout() << "Variable & Coarse Error & Fine Error & Order\\\\" << endl;;
      pout() << "\\hline \\hline " << endl;
      for(int icomp = 0; icomp < ncomp; icomp++)
        {
          int iindex = iscript(icomp,inorm,ncomp);
          pout() << names[icomp] << " &    \t "
                 << setw(12)
                 << setprecision(6)
                 << setiosflags(ios::showpoint)
                 << setiosflags(ios::scientific) 
                 << normsCoar[iindex]  << " & "
                 << setw(12)
                 << setprecision(6)
                 << setiosflags(ios::showpoint)
                 << setiosflags(ios::scientific) 
                 << normsFine[iindex] << " & "
                 << setw(12)
                 << setprecision(6)
                 << setiosflags(ios::showpoint)
                 << setiosflags(ios::scientific) 
                 << orders[iindex];
          pout() << " \\\\ " << endl <<   "\\hline"  <<  endl;
        }
      pout() << "\\end{tabular}" << endl;
      pout() << "\\end{center}" << endl;
      pout() << "\\caption{Solution error convergence rates using L-" << inorm << " norm. " << endl;
      pout() << "$h_f = \\frac{1}{" << nfine << "}$ and $h_c = 2 h_f$, $D = " << SpaceDim << "$ }"  << endl;
      pout() << "\\end{table}" << endl;
      pout() << endl << endl;
    }
  delete[] normsCoar;
  delete[] normsFine;
  delete[] orders   ;
}
/***************/
void
outputError(const LevelData<EBCellFAB>& a_errorFine,
            const EBISLayout& a_ebislFine, 
            const DisjointBoxLayout& a_gridsFine, 
            const Box& a_domainFine,
            const LevelData<EBCellFAB>& a_errorCoar,
            const EBISLayout& a_ebislCoar, 
            const DisjointBoxLayout& a_gridsCoar, 
            const Box& a_domainCoar)
{
#ifdef HDF5
#if CH_SPACEDIM==2
  string fileFine("pltFineError.2d.hdf5");
  string fileCoar("pltCoarError.2d.hdf5");
#else
  string fileFine("pltFineError.3d.hdf5");
  string fileCoar("pltCoarError.3d.hdf5");
#endif
  EBPatchScalarAdvect patchInt;
  Vector<string> names  = getNames();
  bool rCov = true;
  int nCons = patchInt.numConserved();
  Vector<Real> cVal(nCons, 0.0);
  //values that don't matter in output file
  Real dxFine = 1.0;
  Real dxCoar = 2.0;
  Real time   = 1.0;
  Real dt     = 1.0;

  writeEBHDF5(fileFine, a_gridsFine, a_errorFine, names,
              a_domainFine, dxFine, dt, time,
              rCov, cVal);

  writeEBHDF5(fileCoar, a_gridsCoar, a_errorCoar, names,
              a_domainCoar, dxCoar, dt, time,
              rCov, cVal);
#endif
}
/***************/
void
makeFinestDomain(Box& a_domain, 
                 Real& a_dx)
{
  //parse input file.  single level
  ParmParse pp;
  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);
 
  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain = Box(lo, hi);
  
  Real prob_hi;
  int numOpen = n_cell[0];
  pp.get("domain_length",prob_hi);
  a_dx = prob_hi/numOpen;
}
/***/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  {
    setChomboMPIErrorHandler();
#endif

    // Check for an input file
    char* inFile = NULL;

    if (a_argc > 1)
      {
        inFile = a_argv[1];
      }
    else
      {
        pout() << "Usage: <executable name> <inputfile>" << endl;
        pout() << "No input file specified" << endl;
        return -1;
      }
    ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

    Box domainFine, domainCoar;
    Real dxFine, dxCoar;
    makeFinestDomain(domainFine, dxFine);
    domainCoar = coarsen(domainFine, 2);
    dxCoar = 2.0*dxFine;

    LevelData<EBCellFAB>     errorFine, errorCoar;
    DisjointBoxLayout gridsFine, gridsCoar;
    EBISLayout        ebislFine, ebislCoar;
    pout() << "generating fine domain" << endl;

    makeGeometry(domainFine, dxFine);

    pout() << "generating fine error" << endl;

    generateError(errorFine, 
                   gridsFine, ebislFine, domainFine, dxFine, false);

    pout() << "generating coarse domain" << endl;

    makeGeometry(domainCoar, dxCoar);

    pout() << "generating coarse error" << endl;

    generateError(errorCoar, 
                  gridsCoar, ebislCoar, domainCoar, dxCoar, true);

    int fileout;
    pp.get("do_erroroutput", fileout);
    
    if(fileout == 1)
      {
        pout() << "Outputting error to file" << endl;

        outputError(errorFine, 
                    ebislFine, gridsFine, domainFine,
                    errorCoar, 
                    ebislCoar, gridsCoar, domainCoar);
      }


      pout() << "Comparing Error in Solution" << endl;

    //compute convergence rates and output to stdout
    compareError(errorFine, ebislFine, gridsFine, domainFine,
                 errorCoar, ebislCoar, gridsCoar, domainCoar);

#ifdef MPI
  MPI_Finalize();
#endif
}


