/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

//Chombo includes
#include "CH_HDF5.H"
#include "ParmParse.H"
#include "parstream.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "ParmParse.H"
#include "LevelData.H"
#include "EBArith.H"
#include "memtrack.H"

//EBChombo includes
#include "EBIndexSpace.H"
#include "GeometryShop.H"
#include "SphereBL.H"

//application-specific includes
#include "EBBlobIBCFactory.H"
#include "EBPatchScalarAdvectFactory.H"
#include "EBAMRScalarAdvectFactory.H"
#include "EBAMRScalarAdvect.H"
#include "GodunovGeom.H"
#include "RotVel.H"

//debug includes
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "ArrayView.H"
#include "EBArrayView.H"

/***************/
void amrAdvect(const Box& a_domain, 
                const Real& a_dx)
{
  // read inputs
  ParmParse pp;

  int verbosity;
  pp.get("verbosity",verbosity);
  assert(verbosity >= 0);

  int nstop = 0;
  pp.get("max_step",nstop);

  int redistRad = 0;
  pp.get("redist_radius",redistRad);

  Real stopTime = 0.0;
  pp.get("max_time",stopTime);

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  int max_level = 0;
  pp.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> ref_ratios; // (num_read_levels,1);
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  pp.getarr("ref_ratio",ref_ratios,0,num_read_levels+1);

  std::vector<int> regrid_intervals; // (num_read_levels,1);
  pp.getarr("regrid_interval",regrid_intervals,0,num_read_levels);

  Real refineThresh = 0.3;
  pp.get ("refine_thresh",refineThresh);

  int block_factor = 1;
  pp.get("block_factor",block_factor);

  int tagBufferSize = 3;
  pp.get("tag_buffer_size",tagBufferSize);

  int max_grid_size = 32;
  pp.get("max_grid_size",max_grid_size);

  Real fill_ratio = 0.75;
  pp.get("fill_ratio",fill_ratio);

  int checkpoint_interval = 0;
  pp.get("checkpoint_interval",checkpoint_interval);

  int plot_interval = 0;
  pp.get("plot_interval",plot_interval);

  Real cfl = 0.8;
  pp.get("cfl",cfl);

  Real initialCFL = 0.1;
  pp.get("initial_cfl",initialCFL);

  Real fixed_dt = -1;
  pp.get("fixed_dt",fixed_dt);

  Real max_dt_growth = 1.1;
  pp.get("max_dt_growth",max_dt_growth);

  Real dt_tolerance_factor = 1.1;
  pp.get("dt_tolerance_factor",dt_tolerance_factor);
  Real domainLength;
  pp.get("domain_length",domainLength);

  ProblemDomain prob_domain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  
  // create initial/boundary condition object
  
  pout() << "blob initial conditions " << endl;

  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  pp.getarr("blob_center",centerpp,0,SpaceDim);
  for (int i = 0; i < SpaceDim; i++) 
    center[i] = centerpp[i];
  Real size = 0.25;
  Real r0, r1;
  pp.get("blob_size",size);
  pp.get("blob_r0", r0);
  pp.get("blob_r1", r1);

  EBBlobIBCFactory bcfactory(size, r0, r1, center);  

  int idosmushing;
  pp.get("do_smushing", idosmushing);
  bool doSmushing    = (idosmushing ==1);

  //create patch integrator
  bool useLimiting = true;
  EBPatchScalarAdvectFactory   patchGamma(&bcfactory, useLimiting);
  pout() << "using solid body rotation as velocity function" << endl;

  vector<Real> sphere_center(SpaceDim);
  pp.getarr("sphere_center",sphere_center, 0, SpaceDim);
  RealVect sphereCenter;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      sphereCenter[idir] = sphere_center[idir];
    }
  RealVect rotAxis = RealVect::Zero;
  if(SpaceDim==3)
    {
      vector<Real> rot_axis(SpaceDim);
      pp.getarr("rotational_axis",rot_axis, 0, SpaceDim);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          rotAxis[idir] = rot_axis[idir];
        }
    }
  Real rotVel;
  pp.get("rotational_constant", rotVel);

  SolidRotVelFactory rotFactory(sphereCenter, rotAxis, rotVel);
  EBAMRScalarAdvectFactory fact(initialCFL,
                                cfl,
                                redistRad,
                                domainLength,
                                refineThresh,
                                tagBufferSize,
                                verbosity,
                                doSmushing,
                                &rotFactory,
                                &patchGamma);
  

  AMR amr;
  amr.define(max_level,ref_ratios,prob_domain,&fact);

  if (fixed_dt > 0)
    {
      amr.fixedDt(fixed_dt);
    }

  // set grid generation parameters
  amr.maxGridSize(max_grid_size);
  amr.blockFactor(block_factor);
  amr.fillRatio(fill_ratio);

  // the hyperbolic codes use a grid buffer of 1
  int gridBufferSize;
  pp.get("grid_buffer_size",gridBufferSize);
  amr.gridBufferSize(gridBufferSize);

  // set output parameters
  amr.checkpointInterval(checkpoint_interval);
  amr.plotInterval(plot_interval);
  amr.regridIntervals(regrid_intervals);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  std::string prefix;
  pp.get("plot_prefix",prefix);
  amr.plotPrefix(prefix);
  pp.get("chk_prefix",prefix);
  amr.checkpointPrefix(prefix);

  amr.verbosity(verbosity);

  if (!pp.contains("restart_file"))
    {
      if (!pp.contains("fixed_hierarchy"))
        {
          // initialize from scratch for AMR run
          // initialize hierarchy of levels
          amr.setupForNewAMRRun();
        }
      else
        {
          std::string grid_file;
          pp.get("fixed_hierarchy",grid_file);

          // initialize from a list of grids in "grid_file"
          Vector<Vector<Box> > amrGrids(max_level+1);
          godunovFixedGrids(amrGrids,
                            prob_domain,
                            max_level,
                            max_grid_size,
                            block_factor,
                            verbosity,
                            grid_file);
          amr.setupForFixedHierarchyRun(amrGrids,1);
        }
    }
  else
    {
      std::string restart_file;
      pp.get("restart_file",restart_file);
      pout() << " restarting from file " << restart_file << endl;

#ifdef HDF5
      HDF5Handle handle(restart_file,HDF5Handle::OPEN_RDONLY);
      // read from checkpoint file
      amr.setupForRestart(handle);
      handle.close();
#else
      MayDay::Error("amrGodunov restart only defined with hdf5");
#endif
    }

  // run 
  amr.run(stopTime,nstop);

  // output last pltfile and statistics
  //cleanup
  amr.conclude();

}
/************/
void
sphereGeometry(Box& a_coarsestDomain, 
               Real& a_dx)
{
  ParmParse ppgodunov;
  //parse input file
  int max_level = 0;
  ppgodunov.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> refRatios; // (num_read_levels,1);
  // note this requires a refRatio to be defined for the 
  // finest level (even though it will never be used)

  ppgodunov.getarr("ref_ratio",refRatios,0,num_read_levels+1);
  ParmParse pp;
  RealVect origin = RealVect::Zero;
  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);
 
  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_coarsestDomain = Box(lo, hi);
  Box finestDomain = a_coarsestDomain;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      finestDomain.refine(refRatios[ilev]);
    }
  
  Real prob_hi;
  int numOpen = n_cell[0];
  pp.get("domain_length",prob_hi);
  a_dx = prob_hi/numOpen;
  Real fineDx = a_dx;
  int ebMaxCoarsen = 0;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      fineDx /= refRatios[ilev];
      ebMaxCoarsen += refRatios[ilev]/2;
    }
  int ebMaxSize;
  ppgodunov.get("max_grid_size", ebMaxSize);
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  pout() << "sphere geometry" << endl;
  vector<Real> sphere_center(SpaceDim);
  pp.getarr("sphere_center",sphere_center, 0, SpaceDim);
  RealVect sphereCenter;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      sphereCenter[idir] = sphere_center[idir];
    }
  Real sphereRadius;
  pp.get("sphere_radius", sphereRadius);
  int inside_fluid;
  pp.get("sphere_fluid_inside", inside_fluid);
  bool insideFluid = (inside_fluid==1);
  SphereBL sphere(sphereRadius, sphereCenter, insideFluid, fineDx);
  GeometryShop workshop(sphere);
  //this generates the new EBIS
  ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize);
}
/***************/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  MPI_Barrier(Chombo_MPI::comm);
  setChomboMPIErrorHandler();
#endif

 // Check for an input file
  char* inFile = NULL;

  if (a_argc > 1)
    {
      inFile = a_argv[1];
    }
  else
    {
      pout() << "Usage: <executable name> <inputfile>" << endl;
      pout() << "No input file specified" << endl;
      return -1;
    }
  // Parse the command line and the input file (if any)
  ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

  Box coarsestDomain;
  Real dx;
  // run amrGodunov
  sphereGeometry(coarsestDomain, dx);

  amrAdvect(coarsestDomain, dx);

#ifdef MPI
  MPI_Finalize();
#endif
}


