/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _EBAMRSCALARADVECT_H_
#define _EBAMRSCALARADVECT_H_

#include "AMRLevel.H"
#include "BaseVelocity.H"
#include "EBAMRGodunov.H"
#include "EBCellFAB.H"
#include "BaseIVFAB.H"
#include "LevelData.H"
#include "EBLevelRedist.H"
#include "EBCoarToCoarRedist.H"
#include "EBCoarToFineRedist.H"
#include "EBFineToCoarRedist.H"
#include "EBCoarseAverage.H"
#include "EBPWLFineInterp.H"
#include "EBFluxRegister.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "DisjointBoxLayout.H"
#include "EBLevelAdvect.H"

/// 
/**
 */
class EBAMRScalarAdvect : public EBAMRGodunov
{
public:

  EBAMRScalarAdvect();

  virtual ~EBAMRScalarAdvect();

  void define(AMRLevel*  a_coarser_level_ptr,
              const Box& a_problem_domain,
              int        a_level,
              int        a_ref_ratio);

  void define(AMRLevel*  a_coarser_level_ptr,
              const ProblemDomain& a_problem_domain,
              int        a_level,
              int        a_ref_ratio);


  /// advance by one timestep
  Real advance();

  Real advectAdvance(LevelData<EBCellFAB>&         a_consState,
                     LevelData<BaseIVFAB<Real> >&  a_massDiff,
                     EBFluxRegister&               a_fineFluxRegister,
                     EBFluxRegister&               a_coarFluxRegister,
                     const LevelData<EBCellFAB>&   a_consStateCoarseOld,
                     const LevelData<EBCellFAB>&   a_consStateCoarseNew,
                     const Real&                   a_time,
                     const Real&                   a_coarTimeOld,
                     const Real&                   a_coarTimeNew,
                     const Real&                   a_dt);

  void 
  setVelocityFunction(const BaseVelocityFactory& a_velFact);


  void updatePatchReg(EBFluxFAB&        a_flux, 
                      const DataIndex&  a_datInd);

  void updatePatchIrreg(BaseIFFAB<Real>   a_centroidFlux[SpaceDim],
                        const DataIndex&  a_datInd);

  /// write plotfile header
  void writePlotHeader(HDF5Handle& a_handle) const;

  /// write plot file data for this level
  void writePlotLevel(HDF5Handle& a_handle) const;

  void extrapolateCoveredVeloc();

  void clear();
  Real computeDt();
  Real computeInitialDt();
private:
  EBAMRScalarAdvect* getCoarserLevel() const;
  EBAMRScalarAdvect* getFinerLevel() const;
  void levelSetup();


  bool                 m_isVelSet;
  bool                 m_isLevelSetUp;
  BaseVelocity*        m_velocityFunc;

  //these are not grown by one.
  LayoutData<IntVectSet> m_irregSetsSmall;
  //these are grown by one in the directions != idir
  LayoutData<IntVectSet>      m_cfIVS;
  LayoutData<IntVectSet> m_irregSetsGrown[SpaceDim];

  LevelData<EBFluxFAB>           m_extrapDense;
  LevelData<EBFluxFAB>           m_advectVeloc;
  LevelData<EBCellFAB>           m_normalVeloc;

  //covered density and velocity are defined this way
  //because it is convenient to send them to EBLevelAdvect.
  //this makes definition more awkward so the
  //law of conservation of misery still applies
  LayoutData< Vector< BaseIVFAB<Real> * > > m_coveredDenseLo;
  LayoutData< Vector< BaseIVFAB<Real> * > > m_coveredDenseHi; 
  LayoutData< Vector< BaseIVFAB<Real> * > > m_coveredVelocLo;
  LayoutData< Vector< BaseIVFAB<Real> * > > m_coveredVelocHi;
  
  LayoutData< Vector< Vector<VolIndex> > >  m_coveredFaceLitLo; 
  LayoutData< Vector< Vector<VolIndex> > >  m_coveredFaceLitHi;
  LayoutData< Vector< IntVectSet > >        m_coveredSetsLitLo; 
  LayoutData< Vector< IntVectSet > >        m_coveredSetsLitHi;

  LayoutData< Vector< Vector<VolIndex> > >  m_coveredFaceBigLo; 
  LayoutData< Vector< Vector<VolIndex> > >  m_coveredFaceBigHi;
  LayoutData< Vector< IntVectSet > >        m_coveredSetsBigLo; 
  LayoutData< Vector< IntVectSet > >        m_coveredSetsBigHi;

  LevelData<BaseIFFAB<Real> >    m_fluxInterpol[SpaceDim];
  LevelData<BaseIVFAB<Real> >    m_nonConsDivergence;
  LevelData<BaseIVFAB<Real> >    m_ebIrregFaceFlux;
  EBLevelAdvect                  m_ebLevelAdvect;
  EBLevelRedist                  m_ebLevelRedist;
  
  Real m_maxVel; 

  void fillAdvectionVel();
  void levelAdvance();
  RealVect getFaceLoc(const FaceIndex& a_face);
  RealVect getCoveredFaceLoc(const VolIndex&      a_vof,
                             const int&           a_idir,
                             const Side::LoHiSide a_sd);

private:
  //disallowed for all the usual reasons
  void operator=(const EBAMRScalarAdvect& a_input)
  {
    MayDay::Error("invalid operator");
  }
  EBAMRScalarAdvect(const EBAMRScalarAdvect& a_input)
  {
    MayDay::Error("invalid operator");
  }
    
};

#endif
