/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>
#include "parstream.H"
#include "PolyGeom.H"
#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "EBCellFAB.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "CH_HDF5.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"

#include "AMRLevel.H"
#include "EBAMRScalarAdvect.H"
#include "EBCellFactory.H"
#include "BaseIVFactory.H"
#include "VoFIterator.H"
#include "EBPatchAdvectF_F.H"
#include "EBIndexSpace.H"
#include "EBArith.H"
#include "EBFluxFactory.H"
#include "BaseIFFactory.H"
#include "EBPatchScalarAdvect.H"
#include "EBMACProjector.H"
/***************************/
EBAMRScalarAdvect::EBAMRScalarAdvect():EBAMRGodunov()
{
  m_isVelSet = false;
  m_isLevelSetUp = false;
  m_maxVel = -1.0;
  m_velocityFunc = NULL;
}
/***************************/
EBAMRScalarAdvect::~EBAMRScalarAdvect()
{
  if (m_velocityFunc != NULL)
    {
      delete m_velocityFunc;
    }
}
/***************************/
void EBAMRScalarAdvect::define(AMRLevel*  a_coarser_level_ptr,
                               const Box& a_problem_domain,
                               int        a_level,
                               int        a_ref_ratio)
{
  MayDay::Error("EBAMRScalarAdvect::define -\n\tShould never be called with a Box for a problem domain");
}
/***************************/
void EBAMRScalarAdvect::define(AMRLevel*  a_coarser_level_ptr,
                               const ProblemDomain& a_problem_domain,
                               int        a_level,
                               int        a_ref_ratio)
{
  if(s_verbosity >= 3)
    {
      pout() << "EBAMRScalarAdvect::define, level=" << a_level << endl;
    }

  EBAMRGodunov::define(a_coarser_level_ptr,
                       a_problem_domain,
                       a_level,
                       a_ref_ratio);
}
/*****************************/
void 
EBAMRScalarAdvect::
setVelocityFunction(const BaseVelocityFactory& a_velFact)
{
  // Delete old velocity object - if any
  if (m_velocityFunc != NULL)
    {
      delete m_velocityFunc;
    }
  m_velocityFunc  = a_velFact.create();
  m_isVelSet = true;
}
/***************************/
EBAMRScalarAdvect* 
EBAMRScalarAdvect::getCoarserLevel() const
{
  EBAMRScalarAdvect* retval = NULL;
  if(m_coarser_level_ptr != NULL)
    {
      retval = dynamic_cast <EBAMRScalarAdvect*> (m_coarser_level_ptr);

      if (retval == NULL)
        {
          pout() << "EBAMRG::getCoarserLevel: dynamic cast failed"
                 << endl;
          MayDay::Error();
        }
    }
  return retval;
}

/***************************/
EBAMRScalarAdvect* 
EBAMRScalarAdvect::getFinerLevel() const
{
  EBAMRScalarAdvect* retval = NULL;
  if(m_finer_level_ptr != NULL)
    {
      retval = dynamic_cast <EBAMRScalarAdvect*> (m_finer_level_ptr);

      if (retval == NULL)
        {
          pout() << "EBAMRG::getFinerLevel: dynamic cast failed"
                 << endl;
          MayDay::Error();
        }
    }
  return retval;
}
/***************************/
void 
EBAMRScalarAdvect::clear()
{
  //rather than check all the nulls
  if(m_isLevelSetUp)
    {
      for(DataIterator dit = m_coveredDenseLo.dataIterator(); dit.ok(); ++dit)
        {
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              delete(m_coveredDenseLo[dit()][idir]);
              delete(m_coveredDenseHi[dit()][idir]);
              delete(m_coveredVelocLo[dit()][idir]);
              delete(m_coveredVelocHi[dit()][idir]);

              m_coveredDenseLo[dit()][idir] = NULL;
              m_coveredDenseHi[dit()][idir] = NULL;
              m_coveredVelocLo[dit()][idir] = NULL;
              m_coveredVelocHi[dit()][idir] = NULL;
            }
        }
    }
  m_isLevelSetUp = false;
}
/***************************/
void 
EBAMRScalarAdvect::levelSetup()
{
  clear();
  m_isLevelSetUp = true;
  //do not call EBAMRGodunov::levelSetup because our level operator is different
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRScalarAdvect levelSetup for level " << m_level << endl;
    }

  EBAMRScalarAdvect* coarPtr = getCoarserLevel();
  EBAMRScalarAdvect* finePtr = getFinerLevel();
  
  m_hasCoarser = (coarPtr != NULL);
  m_hasFiner   = (finePtr != NULL);
      
  //define redistribution object for this level
  //for now set to volume weighting
  m_ebLevelRedist.define(m_grids, m_ebisl,
                         m_domainBox, m_nComp);

  if (m_hasCoarser)
    {
      int nRefCrse = m_coarser_level_ptr->refRatio();
      const DisjointBoxLayout& coarser_m_grids = coarPtr->m_grids;
      const EBISLayout& coarser_m_ebisl = coarPtr->m_ebisl;
      const Box& domainCoar = coarPtr->m_domainBox;
      
      
      m_ebCoarseAverage.define(m_grids,
                               coarser_m_grids,
                               m_ebisl,
                               coarser_m_ebisl, 
                               domainCoar,
                               nRefCrse,
                               m_nComp);
      m_ebFineInterp.define(m_grids,
                            coarser_m_grids,
                            m_ebisl,
                            coarser_m_ebisl, 
                            domainCoar,
                            nRefCrse,
                            m_nComp);

      // maintain levelgodunov
      m_ebLevelAdvect.define(m_grids,
                             coarser_m_grids,
                             m_ebisl,
                             coarser_m_ebisl,
                             m_problem_domain,
                             nRefCrse,
                             m_dx,
                             m_hasCoarser,
                             m_hasFiner,
                             m_ebPatchGodunovFactory);

      //define fine to coarse redistribution object
      //for now set to volume weighting
      m_ebFineToCoarRedist.define(m_grids, coarser_m_grids,
                                  m_ebisl, coarser_m_ebisl,
                                  domainCoar, nRefCrse, m_nComp);
      m_ebFineToCoarRedist.setToZero();

      coarPtr->syncWithFineLevel();
    }
  else
    {
      m_ebLevelAdvect.define(m_grids,
                             DisjointBoxLayout(),
                             m_ebisl,
                             EBISLayout(),
                             m_problem_domain,
                             m_ref_ratio,
                             m_dx,
                             m_hasCoarser,
                             m_hasFiner,
                             m_ebPatchGodunovFactory);
    }

  //set up mass redistribution array
  Box domainBox = m_problem_domain.domainBox();
  m_sets.define(m_grids);
  for(DataIterator dit = m_grids.dataIterator(); 
      dit.ok(); ++dit)
    {
      Box thisBox = m_grids.get(dit());
      m_sets[dit()] = m_ebisl[dit()].getIrregIVS(thisBox);
    }
  BaseIVFactory<Real> bivfact(m_ebisl, m_sets);
  //the ghost cells on the mass redistribution array
  //are tied to the redistribution radius
  int nghost = 2*m_redistRad;
  IntVect ivGhost = nghost*IntVect::Unit;
  m_massDiff.define(m_grids, m_nComp, ivGhost, bivfact);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      m_massDiff[dit()].setVal(0.0);
    }
  //define other obscure stuff
  EBFluxFactory ebfluxfact(m_ebisl);
  EBCellFactory ebcellfact(m_ebisl);
  m_extrapDense.define(m_grids,        1, ivGhost, ebfluxfact);
  m_normalVeloc.define(m_grids, SpaceDim, 4*IntVect::Unit, ebcellfact);
  
  //yes this is ugly.  it saves ugliness later
  m_coveredDenseLo.define(m_grids);
  m_coveredDenseHi.define(m_grids);
  m_coveredVelocLo.define(m_grids);
  m_coveredVelocHi.define(m_grids);

  m_coveredFaceLitLo. define(m_grids);
  m_coveredFaceLitHi. define(m_grids);
  m_coveredSetsLitLo. define(m_grids);
  m_coveredSetsLitHi. define(m_grids);

  m_coveredFaceBigLo. define(m_grids);
  m_coveredFaceBigHi. define(m_grids);
  m_coveredSetsBigLo. define(m_grids);
  m_coveredSetsBigHi. define(m_grids);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      Box litBox = m_grids.get(dit());
      Box bigBox = grow(litBox, 4);
      bigBox &= m_problem_domain;
      const EBGraph& ebgraph = m_ebisl[dit()].getEBGraph();
      m_coveredDenseLo[dit()].resize(SpaceDim, NULL);
      m_coveredDenseHi[dit()].resize(SpaceDim, NULL);
      m_coveredVelocLo[dit()].resize(SpaceDim, NULL);
      m_coveredVelocHi[dit()].resize(SpaceDim, NULL);

      m_coveredFaceLitLo[dit()].resize(SpaceDim);
      m_coveredFaceLitHi[dit()].resize(SpaceDim);
      m_coveredSetsLitLo[dit()].resize(SpaceDim);
      m_coveredSetsLitHi[dit()].resize(SpaceDim);

      m_coveredFaceBigLo[dit()].resize(SpaceDim);
      m_coveredFaceBigHi[dit()].resize(SpaceDim);
      m_coveredSetsBigLo[dit()].resize(SpaceDim);
      m_coveredSetsBigHi[dit()].resize(SpaceDim);

      for(int idir = 0; idir < SpaceDim; idir++)
        {
          
          IntVectSet irregIVSPlus, irregIVSMinu;
          //get the covered sets and faces
          EBArith::computeCoveredFaces(m_coveredFaceBigHi[dit()][idir], 
                                       m_coveredSetsBigHi[dit()][idir], 
                                       irregIVSPlus,
                                       idir, Side::Hi, 
                                       m_ebisl[dit()], bigBox);
          EBArith::computeCoveredFaces(m_coveredFaceBigLo[dit()][idir], 
                                       m_coveredSetsBigLo[dit()][idir], 
                                       irregIVSMinu,
                                       idir, Side::Lo, 
                                       m_ebisl[dit()], bigBox);
          
          EBArith::computeCoveredFaces(m_coveredFaceLitHi[dit()][idir], 
                                       m_coveredSetsLitHi[dit()][idir], 
                                       irregIVSPlus,
                                       idir, Side::Hi, 
                                       m_ebisl[dit()], litBox);
          EBArith::computeCoveredFaces(m_coveredFaceLitLo[dit()][idir], 
                                       m_coveredSetsLitLo[dit()][idir], 
                                       irregIVSMinu,
                                       idir, Side::Lo, 
                                       m_ebisl[dit()], litBox);
          
          //the density lives on a smaller box than the velocity because the velocity
          //gets used in all the integration steps.
          m_coveredDenseLo[dit()][idir]= new BaseIVFAB<Real>(m_coveredSetsLitLo[dit()][idir], ebgraph, m_nComp);
          m_coveredDenseHi[dit()][idir]= new BaseIVFAB<Real>(m_coveredSetsLitHi[dit()][idir], ebgraph, m_nComp);
          m_coveredVelocLo[dit()][idir]= new BaseIVFAB<Real>(m_coveredSetsBigLo[dit()][idir], ebgraph, m_nComp);
          m_coveredVelocHi[dit()][idir]= new BaseIVFAB<Real>(m_coveredSetsBigHi[dit()][idir], ebgraph, m_nComp);

          m_coveredDenseLo[dit()][idir]->setVal(0.);
          m_coveredDenseHi[dit()][idir]->setVal(0.);
          m_coveredVelocLo[dit()][idir]->setVal(0.);
          m_coveredVelocHi[dit()][idir]->setVal(0.);
        }
    }
  
  m_irregSetsSmall.define(m_grids);
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      m_irregSetsGrown[faceDir].define(m_grids);
    }
  for(DataIterator dit = m_grids.dataIterator(); 
      dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = m_ebisl[dit()];
      if(!ebisBox.isAllCovered())
        {
          const Box&     thisBox = m_grids.get(dit());
          IntVectSet ivsIrreg = ebisBox.getIrregIVS(thisBox);
          m_irregSetsSmall[dit()] = ivsIrreg;
      
          for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
            {
              IntVectSet& grownIVS = m_irregSetsGrown[faceDir][dit()];
              //need to do this in case there is an irregular cells
              //just outside the box.
              Box grownBox2 = grow(thisBox, 2);
              Box grownBox1 = grow(thisBox,1);
              grownBox2 &= domainBox;
              grownBox1 &= domainBox;
              IntVectSet irregivs = ebisBox.getIrregIVS(grownBox2);
              grownIVS = irregivs;
              //grow in all directions != faceDir
              for(int jdir = 0; jdir < SpaceDim; jdir++)
                {
                  if (jdir != faceDir)
                    {
                      grownIVS.grow(jdir, 1);
                    }
                }
              //restrict to domain
              grownIVS &= grownBox1;
            }
        }
    }

  BaseIVFactory<Real> cellFactorySmall(m_ebisl, m_irregSetsSmall);
  m_nonConsDivergence.define(m_grids, m_nComp, 
                             IntVect::Zero, cellFactorySmall);
  //define flux interpolant stuff
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      BaseIFFactory<Real> faceFactory(m_ebisl, m_irregSetsGrown[faceDir], faceDir);
      m_fluxInterpol[faceDir].define(m_grids, m_nComp, 
                                     IntVect::Unit, faceFactory);
    }
  //ebpatch needs coarse-fine IVS to know where to drop order for interpolation
  m_cfIVS.define(m_grids);
  for(DataIterator ditOuter = m_grids.dataIterator(); 
      ditOuter.ok(); ++ditOuter)
    {
      const EBISBox& ebisBox = m_ebisl[ditOuter()];
      if(!ebisBox.isAllCovered())
        {
          // make a grown box of the grid and then subtract off each grid in the
          //domain to make the coarse-fine IVS 
          // CFIVS = grow(b, 1) - sum(bi)
          Box grownBox = grow(m_grids.get(ditOuter()), 1);
          grownBox &= domainBox;
          IntVectSet complementIVS(grownBox);

          for(LayoutIterator litInner = m_grids.layoutIterator(); 
              litInner.ok(); ++litInner)
            {
              Box innerBox = m_grids.get(litInner());
              complementIVS -= innerBox;
            }
      
          m_cfIVS[ditOuter()] = complementIVS;
        }
    }

  //set up advection velocity
  fillAdvectionVel();

  m_ebLevelAdvect.computeNormalVel(m_normalVeloc,
                                   m_advectVeloc,
                                   m_coveredVelocLo,
                                   m_coveredVelocHi,
                                   m_coveredFaceBigLo,
                                   m_coveredFaceBigHi);
}
/***************************/
void 
EBAMRScalarAdvect::
fillAdvectionVel()
{
  assert(m_isVelSet);
  int nghostvel = 6;
  EBFluxFactory fluxfact(m_ebisl);
  IntVect fluxGhost = nghostvel*IntVect::Unit;
  m_advectVeloc.define(m_grids,m_nComp,fluxGhost, fluxfact);
  FaceStop::WhichFaces facestop = FaceStop::SurroundingWithBoundary;
  m_maxVel = 1.0e-10;
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      EBFluxFAB& fluxVelFAB  = m_advectVeloc[dit()];
      Box grownbox = grow(m_grids.get(dit()), nghostvel);
      grownbox &= m_problem_domain;
      IntVectSet ivs(grownbox);
      const EBGraph& ebgraph = m_ebisl[dit()].getEBGraph();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          EBFaceFAB& normalVel = fluxVelFAB[idir];
          for(FaceIterator faceit(ivs, ebgraph, idir, facestop);
              faceit.ok(); ++faceit)
            {
              const FaceIndex& face = faceit();
              RealVect loc = getFaceLoc(face);
              RealVect vel = m_velocityFunc->getVel(loc);
              normalVel(face, 0) = vel[idir];
              if(Abs(vel[idir]) > m_maxVel)
                {
                  m_maxVel = Abs(vel[idir]);
                }
            }
        }
    }
  Vector<Real> all_mv;
  gather(all_mv,m_maxVel,uniqueProc(SerialTask::compute));
  if (procID() == uniqueProc(SerialTask::compute))
    {
      m_maxVel = all_mv[0];
      for (int i = 1; i < all_mv.size(); ++i)
        {
          m_maxVel = Min(m_maxVel,all_mv[i]);
        }
    }
  broadcast(m_maxVel,uniqueProc(SerialTask::compute));

//  //project the velocities
//  EBLevelMACProjector projectinator;
//  projectinator.define(m_grids, m_ebisl, m_domainBox, m_dx);
//  LevelData<EBFluxFAB> gradient(m_grids,m_nComp,fluxGhost, fluxfact);
//  //this  corrects the velocities on the open faces
//  projectinator.project(m_advectVeloc, gradient);

  //now need to extrapolate the (now discretely divergence-free)
  // advection  velocity to the covered faces.   
  // Just use linear 1-D extrapolation.   This is a bad idea for
  //stuff more complicated than simple advection
  extrapolateCoveredVeloc();
  
}
/***********************/
void 
EBAMRScalarAdvect::
extrapolateCoveredVeloc()
{
  //now need to extrapolate the (now discretely divergence-free)
  // advection  velocity to the covered faces.   
  // Just use linear 1-D extrapolation.   This is a bad idea for
  //stuff more complicated than advection with a simple vel field
  //that is constant in time
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = m_ebisl[dit()];
      for(int velDir = 0; velDir < SpaceDim; velDir++)
        {
          const EBFaceFAB& faceVel = m_advectVeloc[dit()][velDir];
          assert(faceVel.nComp() == 1);
          for(SideIterator sit; sit.ok(); ++sit)
            {
              const Vector<VolIndex>* coveredFacesPtr = NULL;
              BaseIVFAB<Real>*        coveredVelocPtr = NULL;
              if(sit() == Side::Lo)
                {
                  coveredFacesPtr =   &m_coveredFaceBigLo[dit()][velDir];
                  coveredVelocPtr =      m_coveredVelocLo[dit()][velDir];
                }
              else
                {
                  coveredFacesPtr =   &m_coveredFaceBigHi[dit()][velDir];
                  coveredVelocPtr =      m_coveredVelocHi[dit()][velDir];
                }

              const Vector<VolIndex>& coveredFaces = *coveredFacesPtr;
              BaseIVFAB<Real>&        coveredVeloc = *coveredVelocPtr;
              for(int ivof = 0; ivof < coveredFaces.size(); ivof++)
                {
                  const VolIndex& vof = coveredFaces[ivof];
                  bool hasFaceOpp = false;
                  bool hasFaceFar = false;
                  Vector<FaceIndex> facesOpp, facesFar;
                  facesOpp   = ebisBox.getFaces(vof, velDir, flip(sit()));
                  hasFaceOpp = ((facesOpp.size() == 1) && (!facesOpp[0].isBoundary()));
                  if(hasFaceOpp)
                    {
                      const VolIndex& otherVoF =  facesOpp[0].getVoF(flip(sit()));
                      facesFar = ebisBox.getFaces(otherVoF,  velDir, flip(sit()));
                      hasFaceFar = ((facesFar.size() == 1) && (!facesFar[0].isBoundary()));
                    }
                  
                  Real velOpp, velFar, velExtrap;
                  if(hasFaceOpp)
                    {
                      velOpp = faceVel(facesOpp[0], 0);
                    }
                  if(hasFaceFar)
                    {
                      velFar = faceVel(facesFar[0], 0);
                    }
                  if(hasFaceOpp && hasFaceFar)
                    {
                      velExtrap = 2.*velOpp - velFar;
                    }
                  else if(hasFaceOpp)
                    {
                      velExtrap = velOpp;
                    }
                  else
                    {
                      velExtrap = 0.0;
                    }
                  coveredVeloc(vof, 0) = velExtrap;
                }
            }
        }
    }
}
/*****/
RealVect 
EBAMRScalarAdvect::getCoveredFaceLoc(const VolIndex&      a_vof,
                                     const int&           a_faceDir,
                                     const Side::LoHiSide a_sd)
{
  RealVect retval = RealVect::Zero;
  const IntVect&  hiiv = a_vof.gridIndex();
  Real cellOffset = 0.0;
  if(a_sd == Side::Hi)
    {
      cellOffset = 1.0;
    }
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real loc;
      if(idir == a_faceDir)
        {
          loc = m_dx*(hiiv[idir] + cellOffset);
        }
      else
        {
          loc = m_dx*(hiiv[idir] + 0.5);
        }
      retval[idir]=loc;
    }
  return retval;
  
}
/*****/
RealVect 
EBAMRScalarAdvect::getFaceLoc(const FaceIndex& a_face)
{
  RealVect retval = RealVect::Zero;
  const IntVect&  hiiv = a_face.gridIndex(Side::Lo);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real loc;
      if(idir == a_face.direction())
        {
          loc = m_dx*hiiv[idir];
        }
      else
        {
          loc = m_dx*(hiiv[idir] + 0.5);
        }
      retval[idir]=loc;
    }
  return retval;
}
/***************************/
Real EBAMRScalarAdvect::advance()
{
  if(s_verbosity >= 3)
    {
      pout() << " in EBAMRScalarAdvect advance for level " << m_level << endl;
    }
  m_stateNew.copyTo(m_stateNew.interval(),
                    m_stateOld,
                    m_stateOld.interval());


  //undefined leveldata in case we need it
  const LevelData<EBCellFAB> ld;

  //set arguments to dummy arguments and
  //then fix if they are available
  const LevelData<EBCellFAB>* coarDataOld = &ld;
  const LevelData<EBCellFAB>* coarDataNew = &ld;
    
  Real told = 0.0;
  Real tnew = 0.0;
  if(m_hasCoarser)
    {
      EBAMRScalarAdvect* coarPtr = getCoarserLevel();
      coarDataOld = &coarPtr->m_stateOld;
      coarDataNew = &coarPtr->m_stateNew;
      tnew = coarPtr->m_time;
      told = tnew - coarPtr->m_dt;
      //time should never be greater than the newest coarse 
      //time.  time might be very slightly smaller than
      //told because of the above subtraction.
      Real eps = 1.0e-10;
      if((m_time > tnew) || (m_time < (told - eps)))
        {
          MayDay::Error("out of bounds time input to AMRGodunov");
        }
      //correct for said floating-point nastiness
      m_time = Max(m_time, told);
    }


  m_ebLevelAdvect.advectToFaces(m_extrapDense, 
                                m_coveredDenseLo, 
                                m_coveredDenseHi, 
                                m_coveredFaceLitLo, 
                                m_coveredFaceLitHi, 
                                m_coveredSetsLitLo, 
                                m_coveredSetsLitHi, 
                                m_stateNew, 
                                m_normalVeloc,
                                m_advectVeloc,
                                coarDataOld, 
                                coarDataNew,
                                told, tnew, m_time, m_dt);
                  
  //a lot of the stuff that happens in old levelgodunov happens here
  //flux registers and all that
  levelAdvance();

  Interval interv(0, m_nComp-1); 
  // increment redistribution register between this level
  // and next coarser level
  if(m_hasCoarser)
    {
      for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
        {
          m_ebFineToCoarRedist.increment(m_massDiff[dit()], dit(), interv);
        }
    }

  //initialize redistirbution register between this level
  // and next finer level
  //this includes re-redistribution registers
  if(m_hasFiner)
    {

      m_ebCoarToFineRedist.setToZero();
      m_ebCoarToCoarRedist.setToZero();
      for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
        {
          BaseIVFAB<Real>& massDiffFAB = m_massDiff[dit()];
          m_ebCoarToFineRedist.increment(massDiffFAB, dit(), interv);
          m_ebCoarToCoarRedist.increment(massDiffFAB, dit(), interv);
        }
    }

  //deal with time and time step
  m_time += m_dt;
  Real dtnew = computeDt();

  return dtnew;
}
/****************************/
void EBAMRScalarAdvect::levelAdvance()
{
  m_ebLevelRedist.setToZero();
  EBFluxRegister* coarFR = NULL;
  if(m_hasCoarser)
    {
      EBAMRScalarAdvect* coarPtr = getCoarserLevel();
      coarFR = &(coarPtr->m_ebFluxRegister);
    }
  //remember this level is the coarse level to the fine FR
  if(m_hasFiner)
    {
      m_ebFluxRegister.setToZero();
    }
  //update stuff
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {

      const Box& cellBox = m_grids.get(dit());
      const EBISBox& ebisBox = m_ebisl[dit()];
      if(!ebisBox.isAllCovered())
        {
          
          EBFluxFAB flux(ebisBox, cellBox, m_nComp);
          updatePatchReg(flux, dit());

          //do fluxregister cha-cha
          /*
            Coarse flux register is flux register with the coarse level.
            Fine flux register is the flux register with the fine level.
            To the finer level FR, this level is the coarse level.
            To the coarser level FR, this level is the fine level.
          */
          Interval consInterv(0, m_nComp-1);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Real scale = m_dt;

              if(m_hasFiner)
                {
                  m_ebFluxRegister.incrementCoarseRegular(flux[idir], scale,dit(),
                                                          consInterv, idir);
                }

              if(m_hasCoarser)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      {
                        coarFR->incrementFineRegular(flux[idir],scale, dit(), 
                                                     consInterv, idir,sit());
                      }
                    }
                }
            }
          //copy fluxes into sparse interpolant
          for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
            {
              IntVectSet ivsIrregGrown = m_irregSetsGrown[faceDir][dit()];
              ivsIrregGrown &= cellBox;
              FaceStop::WhichFaces stopCrit = FaceStop::SurroundingWithBoundary;

              BaseIFFAB<Real>& interpol = m_fluxInterpol[faceDir][dit()];
              EBFaceFAB& fluxDir = flux[faceDir];
              for(FaceIterator faceit(ivsIrregGrown, ebisBox.getEBGraph(), faceDir, stopCrit); faceit.ok(); ++faceit)
                {
                  for(int ivar = 0; ivar < m_nComp; ivar++)
                    {
                      interpol(faceit(), ivar) = fluxDir(faceit(), ivar);
                    }
                }
            }  
        }
    } // end loop over grids
  Interval fluxInterv(0, m_nComp-1);
  //exchange ghost cell data for flux interpolant
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      m_fluxInterpol[faceDir].exchange(fluxInterv);
    }

  //now for the irregular update 
  int ibox = 0;
  Interval consInterv(0, m_nComp-1);
  for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit, ibox++)
    {
      const EBISBox& ebisBox = m_ebisl[dit()];
      if(!ebisBox.isAllCovered())
        {
          BaseIFFAB<Real> centroidFlux[SpaceDim];
          updatePatchIrreg(centroidFlux, dit());
          
          //increment redistribution register.
          m_ebLevelRedist.increment(m_massDiff[dit()], dit(), consInterv);

          //do fluxregister mambo
          /*
            Coarse flux register is flux register with the coarse level.
            Fine flux register is the flux register with the fine level.
            To the finer level FR, this level is the coarse level.
            To the coarser level FR, this level is the fine level.
          */
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              Real scale = m_dt;
              if(m_hasFiner)
                {
                  m_ebFluxRegister.incrementCoarseIrregular(centroidFlux[idir], 
                                                            scale,dit(),
                                                            consInterv, idir);
                }

              if(m_hasCoarser)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      coarFR->incrementFineIrregular(centroidFlux[idir],
                                                     scale, dit(), 
                                                     consInterv, idir,sit());
                    }
                }
            }
        }
    }// end of loop over grids.

  if(m_doSmushing)
    {
      //redistribute mass at this level.  Redistribution to other
      //levels (including reredist) is handled in EBAMRGodunov::postTimeStep
      Interval consInterv(0, m_nComp-1);
      m_ebLevelRedist.redistribute(m_stateNew, consInterv);
    }
}
/****************************/
void 
EBAMRScalarAdvect::updatePatchReg(EBFluxFAB&        a_flux, 
                                  const DataIndex&  a_datInd)
{
  //extract the relevant data for the patch
  const EBFluxFAB& faceVeloc = m_advectVeloc[a_datInd];
  const EBFluxFAB& faceDense = m_extrapDense[a_datInd];
  const Box&       cellBox   = m_grids.get(a_datInd);
  const EBISBox&   ebisBox   = m_ebisl[a_datInd];

  EBCellFAB& stateFAB = m_stateNew[a_datInd];

  a_flux.setVal(0.);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      EBFaceFAB& fluxFAB = a_flux[idir];
      const EBFaceFAB& velocFAB = faceVeloc[idir];
      const EBFaceFAB& denseFAB = faceDense[idir];
      fluxFAB +=   denseFAB;
      fluxFAB *=   velocFAB;
    }

  //set up our integrator.
  const IntVectSet& cfivs = m_cfIVS[a_datInd];
  m_ebPatchGodunov->setValidBox(cellBox, ebisBox, cfivs, m_time, m_dt);

  //compute nonconservative divergences of fluxes
  EBCellFAB nonConsDivF(ebisBox, cellBox, m_nComp);

  EBPatchScalarAdvect* patchAdvect = dynamic_cast<EBPatchScalarAdvect*>(m_ebPatchGodunov);
  assert(patchAdvect != NULL);
  //this is the stable, non-conservative estimate of the solution update
  patchAdvect->advectiveDerivative(nonConsDivF, 
                                   m_extrapDense[a_datInd], 
                                   m_advectVeloc[a_datInd],
                                   m_coveredDenseLo[a_datInd], 
                                   m_coveredDenseHi[a_datInd], 
                                   m_coveredVelocLo[a_datInd], 
                                   m_coveredVelocHi[a_datInd], 
                                   m_coveredFaceBigLo[a_datInd],  
                                   m_coveredFaceBigHi[a_datInd], 
                                   cellBox);

  BaseIVFAB<Real>& sparseNCDF = m_nonConsDivergence[a_datInd];
  //copy nonconservative div f into sparse output thingy
  const IntVectSet& ivsIrreg = m_irregSetsSmall[a_datInd];
  for(VoFIterator vofit(ivsIrreg, ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < m_nComp; ivar++)
        {
          sparseNCDF(vof, ivar) = nonConsDivF(vof, ivar);
        }
    }

  //update the state with nonconservative divergence
  nonConsDivF *= -m_dt;

  //add to the state
  stateFAB += nonConsDivF;;

}
/****************************/
void 
EBAMRScalarAdvect::updatePatchIrreg(BaseIFFAB<Real> centroidFlux[SpaceDim],
                                    const DataIndex&  a_datInd)

{
  EBCellFAB& denseFAB = m_stateNew[a_datInd];
  const EBISBox& ebisBox = m_ebisl[a_datInd];
  const Box& cellBox = m_grids.get(a_datInd);
  const IntVectSet& cfivs = m_cfIVS[a_datInd];

  m_ebPatchGodunov->setValidBox(cellBox, ebisBox, cfivs, m_time, m_dt);

  const BaseIFFAB<Real>* interpolantGrid[SpaceDim];
  const IntVectSet ivsIrregSmall = m_irregSetsSmall[a_datInd];
  int ncons = 1;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      const BaseIFFAB<Real>& interpol = m_fluxInterpol[idir][a_datInd];
      interpolantGrid[idir] = &interpol;
      BaseIFFAB<Real>& fluxDir= centroidFlux[idir];
      fluxDir.define(ivsIrregSmall, ebisBox.getEBGraph(), idir, ncons);
    }
  m_ebPatchGodunov->interpolateFluxToCentroids(centroidFlux,
                                               interpolantGrid,
                                               ivsIrregSmall);

  BaseIVFAB<Real>  ebFlux(ivsIrregSmall, ebisBox.getEBGraph(), ncons);
  BaseIVFAB<Real> consDiv(ivsIrregSmall, ebisBox.getEBGraph(), ncons);
  ebFlux.setVal(0.0);
  consDiv.setVal(0.0);
  m_ebPatchGodunov->consUndividedDivergence(consDiv, 
                                            centroidFlux, ebFlux, 
                                            ivsIrregSmall);
  
  // Update conserved variables to the next time step using the final
  // flux in each direction.  As directions progress, increment a_massDiff
  BaseIVFAB<Real>& massDiff   = m_massDiff[a_datInd];
  BaseIVFAB<Real>& nonConsDiv = m_nonConsDivergence[a_datInd];

  m_ebPatchGodunov->finalUpdate(denseFAB, massDiff, nonConsDiv, 
                                consDiv,  ivsIrregSmall);

}
      
/***************************/
Real 
EBAMRScalarAdvect::
computeDt()
{
  Real newDT = m_cfl*m_dx/m_maxVel;

  return newDT;
}
/***************************/
Real 
EBAMRScalarAdvect::
computeInitialDt()
{
  Real newDT = m_cfl*m_dx/m_maxVel;

  return newDT;
}
#ifdef HDF5
/***************************/
void 
EBAMRScalarAdvect::
writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
    {
      pout() << "EBAMRGodunov::writePlotHeader" << endl;
    }

  HDF5HeaderData header;
  int nCons = 1;
  int nPrim = SpaceDim;
  int consAndPrim = nCons + nPrim;

  int indexVolFrac = consAndPrim;
  int indexAreaFrac = indexVolFrac+1;
  int indexNormal = indexAreaFrac+ 2*SpaceDim;
  int indexDist = indexNormal+SpaceDim;
  int nCompTotal = indexDist+1;
  assert(nCompTotal == consAndPrim + 3*SpaceDim+2);

  Vector<string> names(nCompTotal);
  
  names[0] = string("scalar_density");
  for (int i = 0; i < SpaceDim; i++)
    {
      char velname[40];
      sprintf(velname, "velocity%d", i);
      names[1 + i] = string(velname);
    }

  string volFracName("fraction-0");
  Vector<string> normName(3);
  Vector<string> areaName(6);
  string distName("distance-0");

  normName[0] = "xnormal-0";
  normName[1] = "ynormal-0";
  normName[2] = "znormal-0";

  areaName[0] = "xAreafractionLo-0";
  areaName[1] = "xAreafractionHi-0";
  areaName[2] = "yAreafractionLo-0";
  areaName[3] = "yAreafractionHi-0";
  areaName[4] = "zAreafractionLo-0";
  areaName[5] = "zAreafractionHi-0";

  names[indexVolFrac] = volFracName;

  for (int i=0; i < 2*SpaceDim; i++)
    {
      names[indexAreaFrac+i] = areaName[i];
    }

  for (int i=0; i < SpaceDim; i++)
    {
      names[indexNormal+i] = normName[i];
    }

  names[indexDist] = distName;

  //now output this into the hdf5 handle
  header.m_int["num_components"] = nCompTotal;
  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < nCompTotal; ++comp)
    {
      sprintf(compStr,"component_%d",comp);
      header.m_string[compStr] = names[comp];
    }

  // Write the header to the file
  header.writeToFile(a_handle);

  if (s_verbosity >= 4)
    {
      pout() << header << endl;
    }
}
/***************************/
void 
EBAMRScalarAdvect::
writePlotLevel(HDF5Handle& a_handle) const
{

  if (s_verbosity >= 3)
    {
      pout() << "EBAMRGodunov::writePlotLevel" << endl;
    }
  //fill output data including a bunch of geometric crud
  int nCons = 1;
  int nPrim = SpaceDim;
  int consAndPrim = nCons + nPrim;
  int indexVolFrac = consAndPrim;
  int indexAreaFrac = indexVolFrac+1;
  int indexNormal = indexAreaFrac+ 2*SpaceDim;
  int indexDist = indexNormal+SpaceDim;
  int nCompTotal = indexDist+1;
  assert(nCompTotal == consAndPrim + 3*SpaceDim+2);

  Vector<Real> coveredValuesCons(nCons, -1.0);
  Vector<Real> coveredValuesPrim(nPrim, -1.0);

  Vector<Real> coveredValues;
  coveredValues.append(coveredValuesCons);
  coveredValues.append(coveredValuesPrim);

  LevelData<FArrayBox> fabData(m_grids, nCompTotal, IntVect::Zero);

#ifdef MPI
  MPI_Barrier(Chombo_MPI::comm);
#endif
  EBCellFactory ebcellfact(m_ebisl);
  LevelData<EBCellFAB> normalVel(m_grids, SpaceDim, 4*IntVect::Unit, ebcellfact);
  m_ebLevelAdvect.computeNormalVel(normalVel,
                                   m_advectVeloc,
                                   m_coveredVelocLo,
                                   m_coveredVelocHi,
                                   m_coveredFaceBigLo,
                                   m_coveredFaceBigHi);
  for (DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisbox = m_ebisl[dit()];
      Box  grid = m_grids.get(dit());
      const EBCellFAB& consfab = m_stateNew[dit()];

      EBCellFAB& primfab =  normalVel[dit()];
      FArrayBox& currentFab = fabData[dit()];

      // copy regular data
      currentFab.copy(consfab.getSingleValuedFAB(),0,0,nCons);
      currentFab.copy(primfab.getSingleValuedFAB(),0,nCons,nPrim);

      // set default volume fraction
      currentFab.setVal(1.0,indexVolFrac);

      // set default area fractions
      for (int i=0; i < 2*SpaceDim; i++)
        {
          currentFab.setVal(1.0,indexAreaFrac+i);
        }

      // set default normal
      for (int i=0; i < SpaceDim; i++)
        {
          currentFab.setVal(0.0,indexNormal+i);
        }

      // set default distance of EB from corner
      currentFab.setVal(0.0,indexDist);

      // set special values
      // iterate through the current grid
      // NOTE:  this is probably an inefficient way to do this
      for (BoxIterator bit(grid); bit.ok(); ++bit)
        {
          const IntVect& iv = bit();
          // set special values for covered cells
          if (ebisbox.isCovered(iv))
            {
              for (int icomp = 0; icomp < consAndPrim; icomp++)
                {
                  Real cval = coveredValues[icomp];

                  currentFab(iv,icomp) = cval;
                }
              // volume fraction is zero
              currentFab(iv,indexVolFrac) = 0.0;

              // area fractions are zero
              for (int i=0; i < 2*SpaceDim; i++)
                {
                  currentFab(iv,indexAreaFrac+i) = 0.0;
                }
            }

          // set special values for irregular cells
          if (ebisbox.isIrregular(iv))
            {
              Vector<VolIndex> vofs = ebisbox.getVoFs(iv);
              Real volFrac = ebisbox.volFrac(vofs[0]);
              RealVect normal = ebisbox.normal(vofs[0]);

              // set volume fraction
              currentFab(iv,indexVolFrac) = volFrac;

              // set area fractions--use only the first face you find
              for (int i=0; i < SpaceDim; i++)
                {
                  Vector<FaceIndex> faces;
                  
                  faces = ebisbox.getFaces(vofs[0],i,Side::Lo);
                  if (faces.size() == 0)
                    {
                      currentFab(iv,indexAreaFrac+2*i) = 0.0;
                    }
                  else
                    {
                      currentFab(iv,indexAreaFrac+2*i) = 
                        ebisbox.areaFrac(faces[0]);
                    }

                  faces = ebisbox.getFaces(vofs[0],i,Side::Hi);
                  if (faces.size() == 0)
                    {
                      currentFab(iv,indexAreaFrac+2*i+1) = 0.0;
                    }
                  else
                    {
                      currentFab(iv,indexAreaFrac+2*i+1) = 
                        ebisbox.areaFrac(faces[0]);
                    }
                }

              // set normal
              for (int i=0; i < SpaceDim; i++)
                {
                  currentFab(iv,indexNormal+i) = normal[i];
                }

              // set distance unless the length of the normal is zero
              Real length = PolyGeom::dot(normal,normal);

              if (length > 0)
                {
                  Real dist = PolyGeom::computeAlpha(volFrac,normal)*m_dx;
                  currentFab(iv,indexDist) = -dist;
                }
            } //end if(isIrregular)
        }//end loop over cells
    }//end loop over grids


#ifdef MPI
    MPI_Barrier(Chombo_MPI::comm);
#endif
  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]   = m_ref_ratio;
  header.m_real["dx"]          = m_dx;
  header.m_real["dt"]          = m_dt;
  header.m_box ["prob_domain"] = m_problem_domain.domainBox();

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 4)
    {
      pout() << header << endl;
    }

  // Write the data for this level
  write(a_handle,fabData.boxLayout());
  write(a_handle,fabData,"data");
}

#endif
