/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "EBBlobIBC.H"
#include "EBISLayout.H"
#include "EBLoHiCenter.H"
#include "VoFIterator.H"
/****************************/
EBBlobIBC::
EBBlobIBC(const Real&     a_size,
          const Real&     a_r0,
          const Real&     a_r1,
          const RealVect& a_center)
  :EBPhysIBC()
{
  FORT_SETBLOB(CHF_CONST_REALVECT(a_center),
               CHF_CONST_REAL(a_size),
               CHF_CONST_REAL(a_r0),
               CHF_CONST_REAL(a_r1));
      
  m_r0       = a_r0;
  m_r1       = a_r1;
  m_center   = a_center;
  m_size     = a_size;
  m_isFortranCommonSet = true;
  m_isDefined = false;
}
/****************************/
void 
EBBlobIBC::
setBndrySlopes(EBCellFAB&       a_deltaPrim,
               const EBCellFAB& a_primState,
               const EBISBox&   a_ebisBox,
               const Box&       a_box,
               const int&       a_dir)
{
}
/****************************/
void 
EBBlobIBC::
fluxBC(EBFaceFAB&            a_primGdnv,
       const EBCellFAB&      a_primCenter,
       const EBCellFAB&      a_primExtrap,
       const Side::LoHiSide& a_side,
       const Real&           a_time,
       const EBISBox&        a_ebisBox,
       const Box&            a_box,
       const int&            a_dir)
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  assert(!m_domain.isPeriodic(a_dir));

  Box FBox = a_primGdnv.getSingleValuedFAB().box();
  Box cellBox = FBox;
  int numFlux = a_primGdnv.nComp();

  // Determine which side and thus shifting directions
  int isign = sign(a_side);
  cellBox.shiftHalf(a_dir,isign);

  // Is there a domain boundary next to this grid
  if (!m_domain.contains(cellBox))
    {
      cellBox &= m_domain;
      // Find the strip of cells next to the domain boundary
      Box bndryBox = adjCellBox(cellBox, a_dir, a_side, 1);

      // Shift things to all line up correctly
      bndryBox.shift(a_dir,-isign);

      IntVectSet ivs(bndryBox);
      for(VoFIterator vofit(ivs, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          
          Vector<FaceIndex> bndryFaces = a_ebisBox.getFaces(vof, a_dir, a_side);
          for(int iface= 0; iface < bndryFaces.size(); iface++)
            {
              const FaceIndex& face = bndryFaces[iface];
              //set all fluxes to zero then fix momentum flux
              for(int ivar = 0; ivar < numFlux; ivar++)
                {
                  a_primGdnv(face, ivar) = a_primExtrap(vof,ivar);
                }
            }
        }
    }
}
/****************************/
void 
EBBlobIBC::define(const ProblemDomain&  a_domain,
                  const Real& a_dx)
{
  m_domain = a_domain;
  m_dx = a_dx;
  m_isDefined = true;
}
/****************************/
void 
EBBlobIBC::
initialize(LevelData<EBCellFAB>& a_conState,
           const EBISLayout& a_ebisl) const
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  
  // Iterator of all grids in this level
  for(DataIterator dit = a_conState.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = a_ebisl[dit()];
      // Storage for current grid
      EBCellFAB& conFAB = a_conState[dit()];

      BaseFab<Real>& regConFAB = conFAB.getSingleValuedFAB();
      // Box of current grid
      Box uBox = regConFAB.box();
      uBox &= m_domain;
      // Set up initial condition in this grid
      FORT_BLOBINIT(CHF_CONST_FRA1(regConFAB, 0),
                    CHF_CONST_REAL(m_dx),
                    CHF_BOX(uBox));

      //now for the multivalued cells.  Since it is pointwise,
      //the regular calc is correct for all single-valued cells.
      IntVectSet ivs = ebisBox.getMultiCells(uBox);
      for(VoFIterator vofit(ivs, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect& iv = vof.gridIndex();
          Real density;

          FORT_POINTBLOBINIT(CHF_REAL(density),
                             CHF_CONST_INTVECT(iv),
                             CHF_CONST_REAL(m_dx));

          conFAB(vof, 0) = density;
        }//end loop over multivalued cells
    } //end loop over boxes
}
/****************************/
EBBlobIBC::
~EBBlobIBC()
{
}
/****************************/
