/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "EBMACProjector.H"
#include "EBMACProjectorF_F.H"
#include "NeumannEBBC.H"
#include "NeumannDomainBC.H"
#include "EBFluxFactory.H"
#include "EBAMRPoissonOp.H"
#include "EBArith.H"
#include "EBPatchGodunovF_F.H"
/*****/
void 
EBLevelMACProjector::
setDefaultValues()
{
  m_numSmoothUp    = 4;
  m_numSmoothDown  = 4;
  m_cycles         = 1;
  m_maxIterations  = 33;
  m_tolerance      = 1.0e-12;
}

/*****/
void 
EBLevelMACProjector::
define(const DisjointBoxLayout& a_grids,
       const EBISLayout&        a_ebisl,
       const Box&               a_domain,
       const Real&              a_dx)
{
  m_isDefined = true;
  m_grids  = a_grids;
  m_ebisl  = a_ebisl;
  m_domain = a_domain;
  m_dx     = a_dx;

  //solid wall bcs assumed for now
  NeumannEBBCFactory* neumEBBC  = new NeumannEBBCFactory();
  NeumannDomainBC*    neumDomBC = new NeumannDomainBC();
  neumDomBC->setValue(0.);
  neumEBBC->setValue(0.);
  RefCountedPtr<BaseEBBCFactory> baseEBBC  = (BaseEBBCFactory*)  neumEBBC;
  RefCountedPtr<BaseDomainBC>    baseDomBC = (BaseDomainBC *  )  neumDomBC;

  EBIndexSpace* space = Chombo_EBIS::instance();


  int mgDepth = computeMGDepth(m_grids);
  RealVect probLo = RealVect::Zero;
  int orderEB = 1;
  int mgRelaxType = 1;
  EBAMRPoissonOpFactory levelOpFactory;
  ProblemDomain probDomain(m_domain);
  RealVect dxVect = m_dx*RealVect::Unit;
  int numghost = 1;
  m_levelOpFactory.define(probDomain,m_grids,space, numghost, mgDepth,
                          probLo,dxVect,orderEB,mgRelaxType,
                          baseDomBC, baseEBBC);
                        

  m_bottomSolver.setNumSmooths(m_numSmoothDown);
  m_solver.m_pre   = m_numSmoothDown;
  m_solver.m_post  = m_numSmoothUp;
  m_solver.m_cycle = m_cycles;

  m_solver.m_homogeneous = true;
  m_solver.define(m_levelOpFactory,mgDepth,&m_bottomSolver,probDomain);
}
/*****/
int 
EBLevelMACProjector::
computeMGDepth(const DisjointBoxLayout& a_dbl)
{
  int depth = 0;
  bool coarsenMore = true;
  int refRat = 2;
  while(coarsenMore)
    {
      if(a_dbl.coarsenable(refRat))
        {
          depth++;
          refRat *= 2;
        }
      else
        {
          coarsenMore = false;
        }
    }
  return depth;
}
/*****/
void 
EBLevelMACProjector::
project(LevelData<EBFluxFAB>&  a_velocity,
        LevelData<EBFluxFAB>&  a_gradPhi)
{
  assert(m_isDefined);
  assert(a_velocity.nComp() == 1);
  assert(a_gradPhi.nComp() == 1);

  EBFluxFactory fluxfact(m_ebisl);
  EBCellFactory cellfact(m_ebisl);
  LevelData<EBFluxFAB> centroidVelocity(m_grids, 1, IntVect::Zero, fluxfact);

  //interpolate velocity to face centroids
  EBArith::interpolateFluxToCentroids(centroidVelocity, a_velocity, 
                                      m_grids, m_ebisl, m_domain);

  LevelData<EBCellFAB> phi(m_grids, 1, IntVect::Unit, cellfact);
  LevelData<EBCellFAB> rhs(m_grids, 1, IntVect::Zero, cellfact);

  //set rhs = kappa*div(u)
  kappaDivergence(rhs, centroidVelocity);

  //solve kappa lapl phi = kappa div(u)
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      phi[dit()].setVal(0.);
    }
  
  solve(phi, rhs);
  Interval phiinterv(0,0);
  phi.exchange(phiinterv);
  //compute gradient and subtract off of velocity
  gradient(a_gradPhi, phi); 
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_velocity[dit()][idir] -= a_gradPhi[dit()][idir];
        }
    }
}
/*****/
void 
EBLevelMACProjector::
solve(LevelData<EBCellFAB>&       a_phi,
      const LevelData<EBCellFAB>& a_rhs)
{
  EBCellFactory cellfact(m_ebisl);
  LevelData<EBCellFAB> resid(m_grids, 1, IntVect::Zero, cellfact);
  bool isHomogeneous  = true;
  int level = 0;
  EBAMRPoissonOp* levelOp = m_levelOpFactory.MGnewOp(m_domain,level);
  levelOp->residual(resid, a_phi, a_rhs, isHomogeneous);

  int icomp = 0;
  int inorm = 0;
  Real errorno = EBArith::norm(resid, m_grids, m_ebisl, icomp, inorm);
  Real error = errorno;
  Real compval = m_tolerance*errorno;
  compval = Max(compval, 1.0e-16);
  int iter = 0;
  pout() << " EBLevelMACProjector::solve" << endl;
  while((error > compval) && (error > m_tolerance*errorno) && (iter < m_maxIterations))
    {
      pout() << " iter  = " << iter  << ",  error = " << error << endl;
      m_solver.solve(a_phi,  a_rhs);
      levelOp->residual(resid, a_phi, a_rhs, isHomogeneous);
      error = EBArith::norm(resid, m_grids, m_ebisl, icomp, inorm);
      iter++;
    }
}
void
EBLevelMACProjector::
gradient(LevelData<EBFluxFAB>&        a_gradPhi,
         const LevelData<EBCellFAB>&  a_phi)
{
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      gradient(a_gradPhi[dit()], a_phi[dit()], 
               m_ebisl[dit()], m_grids.get(dit()));
    }
}
/****/
void
EBLevelMACProjector::
gradient(EBFluxFAB&        a_gradPhi,
         const EBCellFAB&  a_phi,
         const EBISBox&    a_ebisBox,
         const Box&        a_box)
{
  a_gradPhi.setVal(0.);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      EBFaceFAB&      gradPhiFAB = a_gradPhi[idir];

      Box interiorBox = a_box;
      
      Box faceBox = surroundingNodes(interiorBox, idir);
      BaseFab<Real>&   regGradPhi = gradPhiFAB.getSingleValuedFAB();
      const BaseFab<Real>& regPhi = a_phi.getSingleValuedFAB();

      FORT_MACGRADPHI(CHF_FRA1(regGradPhi, 0),
                      CHF_CONST_FRA1(regPhi, 0),
                      CHF_CONST_INT(idir),
                      CHF_CONST_REAL(m_dx),
                      CHF_BOX(faceBox));

      FaceStop::WhichFaces stopCrit = FaceStop::SurroundingNoBoundary;
      IntVectSet irregIVS = a_ebisBox.getIrregIVS(interiorBox);
      for(FaceIterator faceit(irregIVS, a_ebisBox.getEBGraph(), idir, stopCrit);
          faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          Real faceGrad = 0.;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              faceGrad += sign(sit())*a_phi(face.getVoF(sit()), 0);
            }
          gradPhiFAB(face, 0) = faceGrad;
        }
    }
}
/****/
void
EBLevelMACProjector::
kappaDivergence(LevelData<EBCellFAB>&        a_divVel,
                const LevelData<EBFluxFAB>&  a_velocity)
{
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      kappaDivergence(a_divVel[dit()], a_velocity[dit()], 
                      m_ebisl[dit()],  m_grids.get(dit()));
    }
}
/*****/
void
EBLevelMACProjector::
kappaDivergence(EBCellFAB&        a_divVel,
                const EBFluxFAB&  a_velo,
                const EBISBox&    a_ebisBox,
                const Box&        a_box)
{
  //set the divergence initially to zero
  //then loop through directions and increment the divergence 
  //with each directions flux difference.
  a_divVel.setVal(0.0);
  BaseFab<Real>&       regDivF = a_divVel.getSingleValuedFAB();
  regDivF.setVal(0.);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //update for the regular vofs in the nonconservative 
      //case  works for all single valued vofs.
      /* do the regular vofs */
      /**/
      const EBFaceFAB& veloDir = a_velo[idir];
      const BaseFab<Real>& regVeloDir = veloDir.getSingleValuedFAB();
      int ncons = 1;
      FORT_DIVERGEF( CHF_BOX(a_box), 
                     CHF_FRA(regDivF), 
                     CHF_CONST_FRA(regVeloDir), 
                     CHF_CONST_INT(idir), 
                     CHF_CONST_INT(ncons), 
                     CHF_CONST_REAL(m_dx));
      /**/
    }
  //update the irregular vofs using conservative diff
  IntVectSet ivsIrreg = a_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //divergence was set in regular update.  we reset it
      // to zero and recalc.
      Real update = 0.;
      for( int idir = 0; idir < SpaceDim; idir++)
        {
          const EBFaceFAB& veloDir = a_velo[idir];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              int isign = sign(sit());
              Vector<FaceIndex> faces = 
                a_ebisBox.getFaces(vof, idir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  Real areaFrac = a_ebisBox.areaFrac(face);
                  Real faceVelo =veloDir(face, 0);
                  update += isign*areaFrac*faceVelo;
                      
                }
            }
        }
      a_divVel(vof, 0) = update;
    }
  //EB Flux assumed zero
}
/*****/
EBLevelMACProjector::
~EBLevelMACProjector()
{
  m_solver.reclaim(m_levelOpFactory);
}
/*****/
EBLevelMACProjector::
EBLevelMACProjector()
{
  setDefaultValues();
  m_isDefined      = false;
}
/*****/
