/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
#include <fstream>
#include <stdio.h>

#include "Box.H"
#include "ParmParse.H"
#include "EBIndexSpace.H"
#include "Vector.H"
#include "OvoidBL.H"
#include "MultiSphereBL.H"
#include "SphereBL.H"
#include "DuellingParabolas.H"
#include "ParabolicMirror.H"
#include "AllRegularService.H"
#include "RampBL.H"
#include "SlabService.H"
#include "GeometryShop.H"
#include "TiltedCylinderIF.H"
#include "ImplicitFunctionBL.H"
#include "BRMeshRefine.H"
#include "parstream.H"
#include "PolyGeom.H"

using std::ifstream;
using std::ios;

/************/
void
godunovGeometry(Box& a_coarsestDomain, 
                Real& a_dx)
{
  ParmParse ppgodunov;
  //parse input file
  int max_level = 0;
  ppgodunov.get("max_level",max_level);

  int num_read_levels = Max(max_level,1);
  std::vector<int> refRatios; // (num_read_levels,1);
  // note this requires a refRatio to be defined for the 
  // finest level (even though it will never be used)

  ppgodunov.getarr("ref_ratio",refRatios,0,num_read_levels+1);
  ParmParse pp;
  RealVect origin = RealVect::Zero;
  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);
 
  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_coarsestDomain = Box(lo, hi);
  Box finestDomain = a_coarsestDomain;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      finestDomain.refine(refRatios[ilev]);
    }
  
  Real prob_hi;
  int numOpen = n_cell[0];
  pp.get("domain_length",prob_hi);
  a_dx = prob_hi/numOpen;
  Real fineDx = a_dx;
  int ebMaxCoarsen = 0;
  for(int ilev = 0; ilev < max_level; ilev++)
    {
      fineDx /= refRatios[ilev];
      ebMaxCoarsen += refRatios[ilev]/2;
    }
  int whichgeom;
  pp.get("which_geom",whichgeom);
  int ebMaxSize;
  ppgodunov.get("max_grid_size", ebMaxSize);
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  if (!pp.contains("ebis_file"))
    {
      if(whichgeom == 0)
        {
          //allregular
          pout() << "all regular geometry" << endl;

          AllRegularService regserv;
          ebisPtr->define(finestDomain, origin, fineDx, regserv, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 1)
        {
          pout() << "ramp geometry" << endl;
          RealVect rampNormal;
          vector<Real>  rampNormalVect(SpaceDim);
          pp.getarr("ramp_normal",rampNormalVect, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              rampNormal[idir] = rampNormalVect[idir];
            }

          Real rampAlpha;
          pp.get("ramp_alpha",rampAlpha);
          RampBL ramp(rampNormal, rampAlpha);
          GeometryShop workshop(ramp);
          //this generates the new EBIS
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 4)
        {
  
          RealVect cylinderAxis;
          vector<Real>  cylinderAxisVect(SpaceDim);
          pp.getarr("cylinder_axis",cylinderAxisVect, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              cylinderAxis[idir] = cylinderAxisVect[idir];
            }
          Real cylinderRadius;

          pp.get("cylinder_radius",cylinderRadius);
          pout() << "using a tilted cylinder implicit function" << endl;
          RealVect corner = RealVect::Zero;
          bool negativeInside = true;
          TiltedCylinderIF tunnel(cylinderRadius, cylinderAxis, corner, negativeInside);
          ImplicitFunctionBL func(tunnel);
          GeometryShop workshop(func);
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 2)
        {
          pout() << "slab geometry" << endl;
          vector<int> slab_lo(SpaceDim);
          pp.getarr("slab_lo",slab_lo,0,SpaceDim);
          vector<int> slab_hi(SpaceDim);
          pp.getarr("slab_hi",slab_hi,0,SpaceDim);
          IntVect lo, hi;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              lo[idir] = slab_lo[idir];
              hi[idir] = slab_hi[idir];
            }
          Box coveredBox(lo,hi);
          SlabService slab(coveredBox);
          //this generates the new EBIS
          RealVect origin = RealVect::Zero;
          ebisPtr->define(finestDomain, origin, fineDx, slab, ebMaxSize, ebMaxCoarsen);
        }      
      else if(whichgeom == 5)
        {
          pout() << "sphere geometry" << endl;
          vector<Real> sphere_center(SpaceDim);
          pp.getarr("sphere_center",sphere_center, 0, SpaceDim);
          RealVect sphereCenter;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              sphereCenter[idir] = sphere_center[idir];
            }
          Real sphereRadius;
          pp.get("sphere_radius", sphereRadius);
          SphereBL sphere(sphereRadius, sphereCenter, false, fineDx);
          GeometryShop workshop(sphere);
          //this generates the new EBIS
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 6)
        {
          pout() << "multisphere geometry" << endl;
          int numSpheres;
          pp.get("num_spheres", numSpheres);
          Vector<Real>     radius(numSpheres);
          Vector<RealVect> center(numSpheres);
          for(int isphere = 0; isphere < numSpheres; isphere++)
            {
              char radiusString[80];
              char centerString[80];
              sprintf(radiusString, "sphere_radius_%d", isphere);
              sprintf(centerString, "sphere_center_%d", isphere);
              vector<Real> sphere_center(SpaceDim);
              Real sphereRadius;
              pp.get(radiusString, sphereRadius);
              pp.getarr(centerString,sphere_center, 0, SpaceDim);
              RealVect sphereCenter;
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  sphereCenter[idir] = sphere_center[idir];
                }
              center[isphere] = sphereCenter;
              radius[isphere] = sphereRadius;
            }
          MultiSphereBL spheres(radius, center, fineDx);
          GeometryShop workshop(spheres);
          //this generates the new EBIS
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 7)
        {
          pout() << "multiparabola geometry" << endl;
          int numSpheres;
          pp.get("num_parabolas", numSpheres);
          int updir;
          pp.get("parabola_updir", updir);
          Vector<Real>     radius(numSpheres);
          Vector<RealVect> center(numSpheres);
          for(int isphere = 0; isphere < numSpheres; isphere++)
            {
              char radiusString[80];
              char centerString[80];
              sprintf(radiusString, "parabola_amplitude_%d", isphere);
              sprintf(centerString, "parabola_center_%d", isphere);
              vector<Real> sphere_center(SpaceDim);
              Real sphereRadius;
              pp.get(radiusString, sphereRadius);
              pp.getarr(centerString,sphere_center, 0, SpaceDim);
              RealVect sphereCenter;
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  sphereCenter[idir] = sphere_center[idir];
                }
              center[isphere] = sphereCenter;
              radius[isphere] = sphereRadius;
            }
          DuellingParabolas parabolas(radius, center, fineDx, updir);
          GeometryShop workshop(parabolas);
          //this generates the new EBIS
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 8)
        {
          pout() << "parabolic mirror geometry" << endl;
          Real amplitude;
          RealVect center;
          vector<Real> centervec;
          int updir;
          pp.get("mirror_updir", updir);
          pp.get("mirror_amplitude", amplitude);
          pp.getarr("mirror_center",centervec, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              center[idir] = centervec[idir];
            }
          ParabolicMirror mirror(amplitude, center, updir, fineDx);
          GeometryShop workshop(mirror);
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else if(whichgeom == 9)
        {
          pout() << "ovoid geometry" << endl;
          Real radius, axialdist;
          RealVect center;
          vector<Real> centervec;
          pp.get("ovoid_radius", radius);
          pp.get("ovoid_axial_dist", axialdist);
          int axis;
          pp.get("ovoid_axis", axis);
          pp.getarr("ovoid_center_lo",centervec, 0, SpaceDim);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              center[idir] = centervec[idir];
            }
          OvoidBL ovoid(radius, center, axialdist, axis, fineDx);
          GeometryShop workshop(ovoid);
          ebisPtr->define(finestDomain, origin, fineDx, workshop, ebMaxSize, ebMaxCoarsen);
        }
      else
        {
          //bogus which_geom
          pout() << " bogus which_geom input = " 
                 << whichgeom << endl;
          MayDay::Error();
        }

    }
  else
    {
      std::string ebis_file;
      pp.get("ebis_file",ebis_file);
      pout() << " recreating  geometry from file " << ebis_file << endl;
      //define ebis anew from file input
#ifdef HDF5
      HDF5Handle handleIn(ebis_file, HDF5Handle::OPEN_RDONLY);
      ebisPtr->define(handleIn);
      handleIn.close();
#endif
    }
}
/***************/
void godunovFixedGrids(Vector<Vector<Box> >& a_amrGrids,
                       const ProblemDomain&  a_domain,
                       int                   a_max_level,
                       int                   a_max_grid_size,
                       int                   a_block_factor,
                       int                   a_verbosity,
                       std::string           a_grid_file)
{
  if (procID() == uniqueProc(SerialTask::compute))
    {
      a_amrGrids.push_back(Vector<Box>(1,a_domain.domainBox()));
      // read in predefined grids
      ifstream is(a_grid_file.c_str(), ios::in);

      if (is.fail())
        {
          MayDay::Error("Cannot open grids file");
        }
    
      // format of file -- number of levels, then for each level starting
      // with level 1, number of grids on level, list of boxes

      int in_numLevels;
      is >> in_numLevels;

      assert (in_numLevels <= a_max_level+1);

      pout() << "numLevels = " << in_numLevels << endl;

      while (is.get() != '\n');

      a_amrGrids.resize(in_numLevels);

      // check to see if coarsest level needs to be broken up
      domainSplit(a_domain,a_amrGrids[0],a_max_grid_size,a_block_factor);
    
      if (a_verbosity >= 3)
        {
          pout() << "level 0: ";
          for (int n = 0; n < a_amrGrids[0].size(); n++)
            {
              pout() << a_amrGrids[0][0] << endl;
            }
        }

      // now loop over levels, starting with level 1
      int ngrid;
      for (int lev = 1; lev < in_numLevels; lev++)
        {
          is >> ngrid;

          if (a_verbosity >= 3)
            {
              pout() << "level " << lev << " numGrids = " << ngrid << endl;
              pout() << "Grids: ";
            }

          while (is.get() != '\n');

          a_amrGrids[lev].resize(ngrid);

          for (int i = 0; i < ngrid; i++)
            {
              Box bx;
              is >> bx;

              while (is.get() != '\n');

              // quick check on box size
              Box bxRef(bx);

              if (bxRef.longside() > a_max_grid_size)
                {
                  pout() << "Grid " << bx << " too large" << endl;
                  MayDay::Error();
                }

              if (a_verbosity >= 3) {
                pout() << bx << endl;
              }

              a_amrGrids[lev][i] = bx;
            } // end loop over boxes on this level
        } // end loop over levels
    }

  broadcast(a_amrGrids,uniqueProc(SerialTask::compute));
}
