/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//

// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <stdio.h>

#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "DebugDump.H"

#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "PolyGeom.H"
#include "GeometryShop.H"

#include "EBFABView.H"

#include "ImplicitFunctionBL.H"
#include "SphereIF.H"
#include "TiltedCylinderIF.H"
#include "UnionIF.H"
#include "IntersectionIF.H"
#include "ComplementIF.H"

/***************/
// Define an EBIS.
/***************/
BaseImplicitFunction* makeGeometry(Box&      a_domain,
                                   RealVect& a_origin,
                                   Real&     a_dx);

/***************/
// Get the distance from a point to a line
/***************/
Real getDistance(const RealVect& a_point,
                 const RealVect& a_lineDirection,
                 const RealVect& a_linePoint);

/***************/
/// Define grids by splitting up domain
/***************/
void makeLayout(DisjointBoxLayout& a_dbl1,
                const Box&         a_domain);

/***************/
// Make the corresponding layout
/***************/
void makeEBISL(EBISLayout&              a_ebisl,
               const DisjointBoxLayout& a_grids,
               const Box&               a_domain,
               const int&               a_nghost);

/***************/
// Put data in the layout
/***************/
void fillData(LevelData<EBCellFAB>&       a_level,
              const RealVect&             a_origin,
              const Real&                 a_dx,
              const BaseImplicitFunction& a_implicit);

int main (int argc, char** argv)
{
#ifdef MPI
  MPI_Init(&argc,&argv);
#endif

  // Begin forever present scoping trick
  {
    const char* in_file = "channelWithSpheres.inputs";

    // Parse input file
    ParmParse pp(0,NULL,NULL,in_file);

    Box domain;
    RealVect origin;
    Real dx;

    BaseImplicitFunction* implicit;

    // Make geometry
    implicit = makeGeometry(domain,origin,dx);

    // Make grids
    DisjointBoxLayout grids;
    makeLayout(grids,domain);

    // Create ebislayout
    int nghost = 2;
    EBISLayout ebisl;
    makeEBISL(ebisl,grids,domain,nghost);

    // Make a LevelData
    int nComps = 1;

    IntVect ghost = IntVect::TheUnitVector();
    ghost *= nghost;

    RefCountedPtr<DataFactory<EBCellFAB> > rcpFactory(new EBCellFactory(ebisl));
    LevelData<EBCellFAB> level(grids,nComps,ghost,*rcpFactory);

    // Put some data in the data holders
    fillData(level,origin,dx,*implicit);

    // Done with this object
    delete implicit;

    // Write the data and the EB out
    char* basename = "channelWithSpheres";

    char  name[1000];
    sprintf(name,"%s%dd.hdf5",basename,SpaceDim);

#ifdef HDF5
    writeEBLevelname(&level,name);
#endif
  } // End scoping trick

  // Clean up
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->clear();

#ifdef MPI
  MPI_Finalize();
#endif

  return 0;
}

BaseImplicitFunction* makeGeometry(Box&      a_domain,
                                   RealVect& a_origin,
                                   Real&     a_dx)
{
  RealVect cylDirection;
  RealVect cylPoint;
  Real     cylRadius;

  int      sphNumber;
  Real     sphMinRadius;
  Real     sphMaxRadius;

  bool     insideRegular;

  // Initialize random number generator
  srand48(time(0));

  // parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);

  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;

  for (int ivec = 0; ivec < SpaceDim; ivec++)
  {
    if (n_cell[ivec] <= 0)
    {
      pout() << "Bogus number of cells input = " << n_cell[ivec];
      exit(1);
    }

    hi[ivec] = n_cell[ivec] - 1;
  }

  a_domain.setSmall(lo);
  a_domain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim,1.0);
  Real prob_hi;

  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.get("prob_hi",prob_hi);

  a_dx = (prob_hi-prob_lo[0])/n_cell[0];

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    a_origin[idir] = prob_lo[idir];
  }

  // ParmParse doesn't get RealVects, so work-around with Vector<Real>
  Vector<Real> vectorDirection;
  pp.getarr("cylDirection",vectorDirection,0,SpaceDim);

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    cylDirection[idir] = vectorDirection[idir];
  }

  Vector<Real> vectorPoint;
  pp.getarr("cylPoint",vectorPoint,0,SpaceDim);

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    cylPoint[idir] = vectorPoint[idir];
  }

  pp.get("cylRadius",cylRadius);

  pp.get("sphNumber",sphNumber);

  pp.get("sphMinRadius",sphMinRadius);
  pp.get("sphMaxRadius",sphMaxRadius);

  // Parm Parse doesn't get bools, so work-around with int
  int intInsideRegular;
  pp.get("insideRegular",intInsideRegular);

  if (intInsideRegular != 0) insideRegular = true;
  if (intInsideRegular == 0) insideRegular = false;

  // Stay inside until the end
  bool inside = true;

  // The vector of spheres
  Vector<BaseImplicitFunction*> spheres;
  spheres.resize(sphNumber);

  // Place random spheres inside the cylinder
  int i = 0;

  while (i < sphNumber) {
    RealVect center;
    Real     radius;

    // Get a random center and radius for a sphere
    for (int idir = 0; idir < SpaceDim; idir++)
    {
      center[idir] = (prob_hi - prob_lo[idir])*drand48() + prob_lo[idir];
    }

    radius = (sphMaxRadius - sphMinRadius)*drand48() + sphMinRadius;

    // Get the distance from the center of the sphere to the axis of the
    // cylinder
    Real dist = getDistance(center,cylDirection,cylPoint);

    // If the sphere is inside cylinder, use it
    if (dist + radius <= cylRadius) {
      spheres[i] = new SphereIF(radius,center,inside);
      i++;
    }
  }

  // Take the union of the insides of the spheres
  UnionIF insideSpheres(spheres);

  // Complement to get the outside of the spheres
  ComplementIF outsideSpheres(insideSpheres,true);

  // Define the cylinder
  TiltedCylinderIF cylinder(cylRadius,cylDirection,cylPoint,inside);

  // Intersect the inside of the cylinder with the outside of the spheres
  IntersectionIF implicit(cylinder,outsideSpheres);

  // Complement if necessary
  ComplementIF insideOut(implicit,!insideRegular);

  ImplicitFunctionBL myGeom(insideOut);
  GeometryShop workshop(myGeom);

  // This generates the new EBIS
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(a_domain,a_origin,a_dx,workshop);

  return insideOut.newImplicitFunction();
}

Real getDistance(const RealVect& a_point,
                 const RealVect& a_lineDirection,
                 const RealVect& a_linePoint)
{
  RealVect diff = a_linePoint;
  diff -= a_point;

  Real normDiff2 = 0.0;
  for (int idir = 0; idir < SpaceDim; idir++) {
    normDiff2 += diff[idir] * diff[idir];
  }

  Real normDirection2 = 0.0;
  for (int idir = 0; idir < SpaceDim; idir++) {
    normDirection2 += a_lineDirection[idir] * a_lineDirection[idir];
  }

  Real dotDiffDirection2 = 0.0;
  for (int idir = 0; idir < SpaceDim; idir++) {
    dotDiffDirection2 += diff[idir] * a_lineDirection[idir];
  }
  dotDiffDirection2 = dotDiffDirection2 * dotDiffDirection2;

  return sqrt(normDiff2 - dotDiffDirection2 / normDirection2);
}

void makeLayout(DisjointBoxLayout& a_dbl,
                const Box&         a_domain)
{
  ParmParse pp;

  int ipieces;
  ipieces = Max(ipieces,1);

  int maxsize;
  Vector<Box> vbox(1,a_domain);

  pp.get("maxboxsize",maxsize);

  domainSplit(a_domain,vbox,maxsize);

  Vector<int> procAssign;
  LoadBalance(procAssign,vbox);

  a_dbl.define(vbox,procAssign);
}

void makeEBISL(EBISLayout&              a_ebisl,
               const DisjointBoxLayout& a_grids,
               const Box&               a_domain,
               const int&               a_nghost)
{
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());

  ebisPtr->fillEBISLayout(a_ebisl,a_grids,a_domain,a_nghost);
}

void fillData(LevelData<EBCellFAB>&       a_level,
              const RealVect&             a_origin,
              const Real&                 a_dx,
              const BaseImplicitFunction& a_implicit)
{
  // Set the data in each VoF to the value of the implicit function at the
  // cell center of the cell containing the VoF
  for (DataIterator dit = a_level.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& ebcell = a_level[dit()];
    const EBISBox& ebisbox = ebcell.getEBISBox();
    FArrayBox& data = ebcell.getFArrayBox();
    const Box& box = data.box();

    for (BoxIterator bit(box); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();

      RealVect coord(a_origin);
      for (int idir = 0; idir < SpaceDim; idir++)
      {
        coord[idir] += a_dx * (iv[idir] + 0.5);
      }

      Real value = a_implicit.value(coord);

      Vector<VolIndex> vofs = ebisbox.getVoFs(iv);
      int size = vofs.size();

      if (size == 0)
      {
        data(iv,0) = value;
      }
      else
      {
        for (int i = 0; i < size; i++)
        {
          ebcell(vofs[i],0) = value;
        }
      }
    }
  }
}
