/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _GODUNOV_PHYSICS_H_
#define _GODUNOV_PHYSICS_H_

#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "PhysIBC.H"

#include <string>
using std::string;

///
/**
   The base class GodunovPhysics provides the physics-dependent components 
   for a higher-order Godunov method for a single patch: characteristic 
   analysis, Riemann solver, quasilinear update, conservative update,
   and transformations between conserved, primitive, and flux variables.
   This class is essentially pure virtual, i.e. all of its member functions 
   are virtual, and the ones that have default implementations are ones
   which are optionally defined, i.e. the default definition is to send
   an error message. Physics-dependent versions of this class that are 
   required in real applications are derived from this class by inheritance.
*/
class GodunovPhysics
{
public:
  /// Constructor
  /**
   */
  GodunovPhysics();

  /// Get the initial and boundary condition object
  /**
   */
  PhysIBC* getPhysIBC() const;

  /// Set the initial and boundary condition object
  /**
   */
  void setPhysIBC(PhysIBC* a_bc);

  /// Destructor
  /**
   */
  virtual ~GodunovPhysics();

  /// Define the object
  /**
   */
  virtual void define(const ProblemDomain& a_domain,
                      const Real&          a_dx);
  /// Compute the maximum wave speed 
  /**
   */
  virtual Real getMaxWaveSpeed(const FArrayBox& a_U,
                               const Box&       a_box) = 0;

  virtual GodunovPhysics* new_godunovPhysics() const = 0;

  /// Number of conserved variables
  /**
     Return the number of conserved variables.
  */
  virtual int numConserved() = 0;

  virtual void getFlux(FArrayBox&       a_flux,
                       const FArrayBox& a_whalf,
                       const Box&       a_box,
                       const int&       a_idir) = 0; 

  /// Names of the conserved variables
  /**
     Return the names of the conserved variables.  A default implementation is
     provided that puts in generic names (i.e., "variable#" which "#" ranges
     for 0 to numConserved()-1.
  */
  virtual Vector<string> stateNames() = 0;

  /// Number of flux variables
  /**
     Return the  number of flux variables.  This can be greater than the number
     of conserved variables if addition fluxes/face-centered quantities are
     computed.
  */
  virtual int numFluxes() = 0;

  /// Is the object completely defined
  /**
     Return true if the object is completely defined.
  */
  virtual bool isDefined() const;

  /// Number of primitive variables
  /**
     Return the number of primitive variables.  This may be greater than the
     number of conserved variables if derived/redundant quantities are also
     stored for convenience.
  */
  virtual int numPrimitives() = 0;

  
  /// Compute the characteristic decomposition of dW.
  /**
     On input, a_dW contains the increments of the primitive variables. On
     output, it contains the increments in the characteristic variables.
  */
  virtual void charAnalysis(FArrayBox&       a_dW,
                            const FArrayBox& a_W,
                            const int&       a_dir,
                            const Box&       a_box) = 0;

  /// Compute the sum of the right eigenvectors times the weights dW.
  /**
     On input, a_dW contains the increments of the characteristic variables.
     On output, it contains the increments in the primitive variables.
  */
  
  virtual void charSynthesis(FArrayBox&       a_dW,
                             const FArrayBox& a_W,
                             const int&       a_dir,
                             const Box&       a_box) = 0;

  /// Compute the eigenvalues.
  /**
   */
  
  virtual void charValues(FArrayBox&       a_lambda,
                          const FArrayBox& a_W,
                          const int&       a_dir,
                          const Box&       a_box) = 0;

  /// Compute the solution to the Riemann problem.
  /**
     Given input left and right states in a direction, a_dir, compute a
     Riemann problem and generate fluxes at the faces within a_box.
  */
  virtual void riemann(FArrayBox&       a_WStar,
                       const FArrayBox& a_WLeft,
                       const FArrayBox& a_WRight,
                       const FArrayBox& a_W,
                       const Real&      a_time,
                       const int&       a_dir,
                       const Box&       a_box) = 0;

  /// Post-normal predictor calculation.
  /**
     Add increment to normal predictor to account e.g. for source terms due to
     spatially-varying coefficients.
  */
  virtual void postNormalPredictor(FArrayBox& a_dWLow,
                                   FArrayBox& a_dWHigh,
                                   const FArrayBox& a_W,
                                   const int&       a_dir,
                                   const Box&       a_box) = 0;

  /// Compute the quasilinear update A*dW/dx.
  /**
   */
  virtual void quasilinearUpdate(FArrayBox& a_AdWdx,
                                 const FArrayBox& a_wHalf,
                                 const FArrayBox& a_W,
                                 const Real&      a_scale,
                                 const int&       a_dir,
                                 const Box&       a_box) = 0;

  /// Compute primitive variables from conserved variables.
  /**
   */
  virtual void consToPrim(FArrayBox& a_W,
                          const FArrayBox& a_U,
                          const Box&       a_box) = 0;


  /// Interval within the primitive variables corresponding to the velocities
  /**
     Return the interval of component indices within the primitive variable 
     of the velocities.  Used for slope flattening (slope computation) and
     computing the divergence of the velocity (artificial viscosity).
   */
  virtual Interval velocityInterval() = 0;

  /// Component index within the primitive variables of the pressure
  /**
     Return the component index withn the primitive variables for the
     pressure.  Used for slope flattening (slope computation).
   */
  virtual int pressureIndex() = 0;

  /// Component index within the primitive variables of the bulk modulus
  /**
     Return the component index withn the primitive variables for the
     bulk modulus.  Used for slope flattening (slope computation) used
     as a normalization to measure shock strength. 
   */
  virtual int bulkModulusIndex() = 0;

protected:

  // Has define() been called
  bool m_isDefined;

  // Problem domain and grid spacing
  ProblemDomain m_domain;
  Real          m_dx;

  // Initial and boundary condition object and has it been set
  PhysIBC* m_bc;
  bool m_isBCSet;

private:
  // Disallowed for all the usual reasons
  void operator=(const GodunovPhysics& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  GodunovPhysics(const GodunovPhysics& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
