/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _GODUNOV_UTILITIES_H_
#define _GODUNOV_UTILITIES_H_

#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "REAL.H"
#include "FArrayBox.H"
#include "GodunovPhysics.H"

///
/**
   Utility class for higher-order Godunov methods: slopes, parabolic 
   interpolants, limiters. Contains no physics-dependent methods, but 
   one of the member functions (PPMFaceLimiter) may require a pointer 
   to a physics class analysis class in order to perform limiting in 
   characteristic variables.
*/
class GodunovUtilities
{
public:
  /// Constructor
  /**
   */
  GodunovUtilities();

  /// Destructor
  /**
   */
  ~GodunovUtilities();

  /// Define the object
  /**
   */
  void define(ProblemDomain& a_domain,
              const Real&    a_dx);
  
  /// Compute the slope flattening coefficients
  /**
     Compute the slope flattening coefficients, a_flattening, using the
     primitive variables, a_W, within a_box.
  */
  
  void computeFlattening(FArrayBox&       a_flattening,
                         const FArrayBox& a_W,
                         const Interval&  a_velInt,
                         const int&       a_presInd,
                         const int&       a_bulkModulusInd,
                         const Box&       a_box);

/// apply the flattening to slopes.

  void applyFlattening(  FArrayBox&       a_dW,
                         const FArrayBox& a_flat,
			 const Box&       a_box);

  /// Compute componentwise van Leer slopes.
  /**
     Given cell averages W, compute van Leer slopes dW on a 
     component-by-component basis.
  */
  void vanLeerSlopes(FArrayBox&       a_dW,
                     const FArrayBox& a_W,
                     const int&       a_numSlopes,
                     const int&       a_dir,
                     const Box&       a_box);


  /// Compute fourth-order slopes.
  /**
     Given cell averages W and van Leer slopes dWvL, compute fourth-order
     slopes dW4. Limiting is performed in a separate pass.
  */
  void fourthOrderSlopes(FArrayBox&       a_dW4,
                         const FArrayBox& a_W,
                         const FArrayBox& a_dWvL,
                         const int&       a_numSlopes,
                         const int&       a_dir,
                         const Box&       a_box);


  ///
  void oneSidedDifferences(FArrayBox&       a_dWMinu,
                           FArrayBox&       a_dWPlus,
                           const FArrayBox& a_W,
                           const int&       a_dir,
                           const Box&       a_box);
  
  void slopes(FArrayBox&       a_dWCent,
              FArrayBox&       a_dWMinu,
              FArrayBox&       a_dWPlus,
              const FArrayBox& a_W,
              const int&       a_numSlopes,
              const int&       a_dir,
              const Box&       a_loBox,
              const Box&       a_hiBox,
              const Box&       a_centerBox,
              const Box&       a_entireBox,
              const int&       a_hasLo,
              const int&       a_hasHi);

  /// van Leer slope limiter.
  /**
     On input, dW contains the centered, unlimited slopes, and
     dW{Low,High} contain the one-sided slopes from the low, high sides.
     On output, dW contains the limited slopes.
     slopes dW4. Limiting is performed in a separate pass.
  */
  void slopeLimiter(FArrayBox&       a_dW,
                    const FArrayBox& a_dWLow,
                    const FArrayBox& a_dWHigh,
                    const int&       a_numSlopes,
                    const int&       a_dir,
                    const Box&       a_box);

  /// PLM normal predictor.
  /**
     Compute the increments in the characteristic amplitudes
  */
  void PLMNormalPredictor(FArrayBox&       a_dWCharLow,
                          FArrayBox&       a_dWCharHigh,
                          const FArrayBox& a_dWChar,
                          const FArrayBox& a_Lambda,
                          const Real&      a_dtbydx,
                          const Box&       a_box);


  /// PPM face-centered interpolant.
  /**
     Given the cell average W and the van Leer slope dWvL, compute fourth-order
     accurate face-centered values WFace given by differentiating the indefinite 
     integral. Limiting is performed in a separate pass.
  */
  void PPMFaceValues(FArrayBox&       a_WFace,
                     const FArrayBox& a_W,
                     const int&       a_numSlopes,
                     const int&       a_dir,
                     const Box&       a_box,
		     const GodunovPhysics* a_physPtr = NULL);

  /// PPM Limiter.
  /**
     On input, dWLow and dWHigh are the differences between the face 
     values on the low and high sides of cells and the average in the cell. 
     The PPM limiter is applied to these values to obtain a monotone
     interpolant in the cell.
  */
  
  void PPMLimiter(FArrayBox& a_dWLow,
                  FArrayBox& a_dWHigh,
                  const int& a_numSlopes,
                  const int& a_dir,
                  const Box& a_box);

  /// PPM normal predictor.
  /**
     On input, dW{Low,High}, contain the characteristic 
     expansions of the differences between the {low, high} face values
     and the cell average. On output, dW{Low,High} contain the 
     characteristic amplitudes of the corrections required to compute
     the normal predictor.
  */
  void PPMNormalPredictor(FArrayBox&       a_dWLow,
                          FArrayBox&       a_dWHigh,
                          const FArrayBox& a_Lambda,
                          const Real&      a_dtbydx,
                          const int&       a_numSlopes,
                          const Box&       a_box);
  
  /// Compute face-centered velocity divergence.
  /**
     Returns face-centered velocity divergence on for faces in the
     direction a_dir. The velocities are the components a_velInterval
     of a_W.
  */
  
  void divVel(FArrayBox&       a_divVel,
              const FArrayBox& a_W,
              const Interval&  a_velInt,
              const int&       a_dir,
              const Box&       a_box);

  /// Compute face-centered artificial viscosity flux.
  /**
     Increments face-centered flux in the a_dir direction with quadratic 
     artificial viscosity.
  */
  void artificialViscosity(FArrayBox&       a_Flux,
                           const FArrayBox& a_U,
                           const FArrayBox& a_divVel,
                           const Real&      a_scale,
                           const int&       a_dir,
                           const Box&       a_box);


protected:

  // Problem domain and grid spacing
  ProblemDomain m_domain;
  Real m_dx;
  bool m_isDefined;
private:
  // We may allow copying and assignment later.
  // Disallowed for all the usual reasons
  void operator=(const GodunovUtilities& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  GodunovUtilities(const GodunovUtilities& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
