/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "PatchGodunov.H"
#include "PatchGodunovF_F.H"
#include "LoHiCenter.H"
#include "LoHiSide.H"

#include "PatchGodunovF_F.H"

// Flag everything as not defined or set
PatchGodunov::PatchGodunov()
{
  m_isDefined = false;
  m_isSlopeSet = false;
  m_isArtViscSet = false;
  m_isCurrentTimeSet = false;
  m_isCurrentBoxSet = false;
  m_gdnvPhysics = NULL;
}

PatchGodunov::~PatchGodunov()
{
  // Delete the initial/boundary condition object - if it exists
  if(m_gdnvPhysics != NULL)
    {
      delete m_gdnvPhysics;
    }
}

// Define this object and the boundary condition object
void PatchGodunov::define(const ProblemDomain&           a_domain,
                          Real                           a_dx,
                          bool                           a_useArtificialViscosity,
                          Real                           a_artificialViscosity,
                          bool                           a_useFlattening,
                          bool                           a_limitSlopes,
                          bool                           a_useCharLimiting,
                          const GodunovPhysics* const    a_physPtr)
{
  // Store the domain and grid spacing
  m_domain = a_domain;
  m_dx = a_dx;
  if(m_gdnvPhysics != NULL)
    {
      delete m_gdnvPhysics;
    }
  m_gdnvPhysics = a_physPtr->new_godunovPhysics();
  m_gdnvPhysics->define(m_domain, m_dx);
  m_util.define(m_domain, m_dx);

  // Artificial viscosity coefficient must be greater than zero
  assert(!a_useArtificialViscosity || (a_artificialViscosity >= 0.0));

  // Store the artificial viscosity flag and coefficient
  m_useArtificialViscosity = a_useArtificialViscosity;
  m_artificialViscosity = a_artificialViscosity;
  m_limitSlopes     = a_limitSlopes;
  m_useFlattening   = a_useFlattening;
  m_useCharLimiting = a_useCharLimiting;
  m_isSlopeSet = true;
  m_isArtViscSet = true;

  m_isDefined = true;
}
// Set the current box.
void PatchGodunov::setCurrentBox(const Box& a_currentBox)
{
  m_currentBox = a_currentBox;
  m_isCurrentBoxSet = true;
}
// Set the current box.
void PatchGodunov::setCurrentTime(const Real& a_time)
{
  m_time = a_time;
  m_isCurrentTimeSet = true;
}
 
void PatchGodunov::updateState(FArrayBox&       a_U,
                               FArrayBox        a_F[CH_SPACEDIM],
                               Real&            a_maxWaveSpeed,
                               const FArrayBox& a_S,
                               const Real&      a_dt,
                               const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  int nprim = m_gdnvPhysics->numPrimitives();
  int nflux = m_gdnvPhysics->numFluxes();
  
  FArrayBox whalf[CH_SPACEDIM];
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Box faceBox = surroundingNodes(a_box, idir);
      whalf[idir].resize(faceBox, nprim);
      a_F[idir].resize( faceBox,  nflux);
      whalf[idir].setVal(0.);
      a_F[idir].setVal(0.);
    }

  computeWHalf(whalf, a_U, a_S, a_dt, a_box);
  
  FArrayBox dU(a_U.box(),a_U.nComp());

  computeUpdate(dU, a_F, a_U, whalf, a_dt, a_box);

  a_U += dU;

  // Get and return the maximum wave speed on this patch/grid
  a_maxWaveSpeed = m_gdnvPhysics->getMaxWaveSpeed(a_U, m_currentBox);

}

///////////
void PatchGodunov::computeUpdate(FArrayBox&       a_dU,
                                 FArrayBox        a_F[CH_SPACEDIM],
                                 const FArrayBox& a_U,
                                 const FArrayBox  a_wHalf[CH_SPACEDIM],
                                 const Real&      a_dt,
                                 const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  a_dU.setVal(0.);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //get flux from whalf
      m_gdnvPhysics->getFlux(a_F[idir], a_wHalf[idir], a_F[idir].box(), idir);
      FArrayBox divu(a_F[idir].box(),1);
      int numPrim = m_gdnvPhysics->numPrimitives();
      FArrayBox W(a_U.box(),numPrim);
      m_gdnvPhysics->consToPrim(W,a_U,a_U.box());
      Interval velInt = m_gdnvPhysics->velocityInterval();
      m_util.divVel(divu,W,velInt,idir,a_F[idir].box());
      Box faceBox = enclosedCells(a_F[idir].box());
      faceBox.grow(idir,1);
      faceBox &= m_domain;
      faceBox.grow(idir,-1);
      faceBox.surroundingNodes(idir);
//CLEANUP:: faceBox should be related to a_box.
      m_util.artificialViscosity(a_F[idir],a_U,
            divu,m_artificialViscosity,idir,faceBox);
//CLEANUP:: time should be passed as an argument.
      Real time = 0.;
      m_gdnvPhysics->getPhysIBC()->artViscBC(a_F[idir],a_U,divu,idir,time);

      //compute flux difference fhi-flo
      FArrayBox diff(a_dU.box(), a_dU.nComp());
      diff.setVal(0.);
      
      FORT_FLUXDIFF(CHF_FRA(diff),
                    CHF_CONST_FRA(a_F[idir]),
                    CHF_CONST_INT(idir),
                    CHF_BOX(a_box));
      
      //add flux difference to dU
      a_dU += diff;
    }

  //multiply du by dt/dx because that is what the output expects
  Real dtbydx = -a_dt/m_dx;
  a_dU *= dtbydx;

}
// Compute the time-centered values of the primitive variable on the 
// interior cell faces of the cell-centered box a_box. 
void PatchGodunov::computeWHalf(FArrayBox        a_WHalf[CH_SPACEDIM],
                                const FArrayBox& a_U,
                                const FArrayBox& a_S,
                                const Real&      a_dt,
                                const Box&       a_box)
{
  assert(isDefined());
  assert(a_box == m_currentBox);

  // Get the number of various variables
  int numPrim = m_gdnvPhysics->numPrimitives();
  int numFlux = m_gdnvPhysics->numFluxes();

  // The current box
  Box curBox = m_currentBox;

  // Boxes for face centered state - used for the riemann() and
  // artificialViscosity() calls
  Box faceBox[SpaceDim];

  // Boxes for face centered fluxes
  Box fluxBox[SpaceDim];

  // Boxes for cell centered state - used for the updatePrim() calls
  Box ccBox[SpaceDim];

  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
    {
      // This box is face centered in direction "dir1", is one bigger than the
      // input box in all directions except "dir1" and stays one cell away from
      // the domain boundary in "dir1"
      faceBox[dir1] = curBox;
      faceBox[dir1].grow(1);
      faceBox[dir1] &= m_domain;
      faceBox[dir1].grow(dir1,-1);
      faceBox[dir1].surroundingNodes(dir1);

      // This box is face centered in direction "dir1", is one bigger than the
      // input box in all directions except "dir1"
      fluxBox[dir1] = curBox;
      fluxBox[dir1].grow(1);
      fluxBox[dir1].grow(dir1,-1);
      fluxBox[dir1] &= m_domain;
      fluxBox[dir1].surroundingNodes(dir1);

      // This is the cell centered analog of "fluxBox"
      ccBox[dir1] = curBox;
      ccBox[dir1].grow(1);
      ccBox[dir1].grow(dir1,-1);
      ccBox[dir1] &= m_domain;
    }

  // Primitive variables
  Box WBox = a_U.box();
  WBox &= m_domain;

  FArrayBox W(WBox,numPrim);

  // Create a temp storage for a_U - RS 06/10/02
  FArrayBox Uold(a_U.box(),a_U.nComp());
  Uold.copy(a_U);

  // Calculate the primitive variables from the conserved variables
  m_gdnvPhysics->consToPrim(W,a_U,WBox);

  // Define the box where slopes will be needed (one larger than the final
  // update box)
  Box slopeBox = curBox;
  slopeBox.grow(1);
  slopeBox &= m_domain;

  // Compute flattening once for all slopes if needed
  FArrayBox flattening(slopeBox,1);

  if (m_useFlattening)
    {
      Interval velInt= m_gdnvPhysics->velocityInterval();
      int presIndex  = m_gdnvPhysics->pressureIndex();
      int bulkIndex  = m_gdnvPhysics->bulkModulusIndex();
      m_util.computeFlattening(flattening, W,
                               velInt,
                               presIndex, 
                               bulkIndex, 
                               slopeBox);
    }
  
  // Intermediate, extrapolated primitive variables
  FArrayBox WMinu[SpaceDim];
  FArrayBox WPlus[SpaceDim];

  // Initial fluxes
  FArrayBox WHalf1[SpaceDim];
  FArrayBox Flux1[SpaceDim];

  // Compute initial fluxes
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
    {
      // Size the intermediate, extrapolated primitive variables
      WMinu[dir1].resize(slopeBox,numPrim);
      WPlus[dir1].resize(slopeBox,numPrim);
      
      Real dtbydx = a_dt/m_dx;
      // Compute predictor step to obtain extrapolated primitive variables
      PPMNormalPred(WMinu[dir1],    
                    WPlus[dir1],
                    dtbydx,
                    W,
                    flattening,
                    dir1,
                    slopeBox);

      // If the source term is valid add it to the primitive quantities
      if (!a_S.box().isEmpty())
        {
          FArrayBox diff(a_S.box(), a_S.nComp());
          diff.copy(a_S);
          diff *= 0.5*a_dt;
          
          WMinu[dir1] += diff;
          WPlus[dir1] += diff;
        }

      // Solve the Riemann problem and get fluxes
      WHalf1[dir1].resize(fluxBox[dir1],numFlux);
      Flux1[dir1].resize(fluxBox[dir1],numFlux);
      m_gdnvPhysics->riemann(WHalf1[dir1],
                             WPlus[dir1],WMinu[dir1], W,
                             m_time, dir1,faceBox[dir1]);
      m_gdnvPhysics->getFlux(Flux1[dir1], WHalf1[dir1], Flux1[dir1].box(), dir1);

    }

#if (CH_SPACEDIM == 3)
  // In 3D, compute some additional intermediate fluxes
  //
  // NOTE:  The diagonal entries of this array of fluxes are not
  // used and will not be defined.
  FArrayBox WHalf2[SpaceDim][SpaceDim];
  
  // Compute the intermediate, corrected fluxes in each direction
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
    {
      // Correct fluxes using fluxes from a different direction
      for (int dir2 = 0; dir2 < SpaceDim; dir2++)
        {
          // A different direction has been found
          if (dir2 != dir1)
            {
              // Temporary primitive variables
              FArrayBox WMinuTemp(WMinu[dir1].box(),numPrim);
              FArrayBox WPlusTemp (WPlus [dir1].box(),numPrim);
              FArrayBox AdWdx(WPlus[dir1].box(),numPrim);

	      // preventing uninitialized memory reads which cause
	      // FPE's on some machines
	      AdWdx.setVal(666.666);

              // Copy data for in place modification
              WMinuTemp.copy(WMinu[dir1]);
              WPlusTemp.copy(WPlus[dir1]);

              // Update the current, extrapolated primitive variable using a flux
              // in a different direction
 
              m_gdnvPhysics->quasilinearUpdate(AdWdx,
                  WHalf1[dir2],WMinuTemp,-(1.0/3.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WMinuTemp += AdWdx;
              m_gdnvPhysics->quasilinearUpdate(AdWdx,
                  WHalf1[dir2],WMinuTemp,-(1.0/3.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WPlusTemp += AdWdx;

              // Solve the Riemann problem.
       
              WHalf2[dir1][dir2].resize(fluxBox[dir1],numPrim);
              m_gdnvPhysics->riemann(WHalf2[dir1][dir2],
                                     WPlusTemp,WMinuTemp,W, m_time, dir1,faceBox[dir1]);
            }
        }
    }
#endif

  // faceBox and fluxBox are now a bit smaller for the final corrections
  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
    {
      faceBox[dir1] = curBox;
      faceBox[dir1].grow(dir1,1);
      faceBox[dir1] &= m_domain;
      faceBox[dir1].grow(dir1,-1);
      faceBox[dir1].surroundingNodes(dir1);
      
      fluxBox[dir1] = curBox;
      fluxBox[dir1].surroundingNodes(dir1);
    }

  // Do the final corrections to the fluxes
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
    {
      // Correct the flux using fluxes in the remaining direction(s)
      for (int dir2 = 0; dir2 < SpaceDim; dir2++)
        {
          // A different direction has been found
          if (dir2 != dir1)
            {
#if (CH_SPACEDIM == 2)
              // In 2D, the current primitive state is updated by a flux in
              // the other direction
              FArrayBox AdWdx(WPlus[dir1].box(),numPrim);
        
              m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf1[dir2],WPlus[dir1],
                                   -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WPlus[dir1] += AdWdx;
              m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf1[dir2],WMinu[dir1],
                                   -(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WMinu[dir1] += AdWdx;

#elif (CH_SPACEDIM == 3)
              // In 3D, find a direction different from the two above
              int dir3 = 3 - dir1 - dir2;

              // Update the conservative state using both corrected fluxes in
              // the other two directions
              FArrayBox AdWdx(WPlus[dir1].box(),numPrim);
              m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf2[dir2][dir3],
                    WPlus[dir1],-(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WPlus[dir1].plus(AdWdx,0,0,numPrim);
              m_gdnvPhysics->quasilinearUpdate(AdWdx,WHalf2[dir2][dir3],
                    WMinu[dir1],-(1.0/2.0)*a_dt/m_dx,dir2,ccBox[dir2]);
              WMinu[dir1].plus(AdWdx,0,0,numPrim);
#else
              // Only 2D and 3D should be possible
              MayDay::Error("LevelGodunov::step - CH_SPACEDIM not 2 or 3!");
#endif
            }
        }

      // Solve the Riemann problem to obtain time-centered face values.
      // be returned
      FArrayBox& WHalf = a_WHalf[dir1];
      m_gdnvPhysics->riemann(WHalf,WPlus[dir1],WMinu[dir1],W,m_time,dir1,faceBox[dir1]);
    }
}

// Return true if everything is defined and setup
bool PatchGodunov::isDefined() const
{
  return 
    m_isDefined        &&
    m_isSlopeSet       &&
    m_isArtViscSet     &&
    m_isCurrentTimeSet &&
    m_isCurrentBoxSet;
}
/////////
void
PatchGodunov::PLMNormalPred(FArrayBox&       a_WMinu,
                            FArrayBox&       a_WPlus,
                            const Real&      a_dtbydx,
                            const FArrayBox& a_W,
                            const FArrayBox& a_flat,
                            const int&       a_dir,
                            const Box&       a_box)
{

  // Preliminary van Leer slopes with componentwise limiting (inputs to 
  // fourth-order slope calculation.

  int numprim = m_gdnvPhysics->numPrimitives();

  Box boxVL = a_box;
  boxVL.grow(a_dir,1);
  boxVL &= m_domain;
  Real time = 0.;
 
  FArrayBox dWvL(boxVL, numprim);
  m_util.vanLeerSlopes(dWvL, a_W, numprim, a_dir,boxVL);
  m_gdnvPhysics->getPhysIBC()->setBdrySlopes(dWvL,a_W,a_dir,time);

  // Compute fourth-order slopes, without limiting.

  FArrayBox dW4(a_box,numprim);
  m_util.fourthOrderSlopes(dW4,a_W,dWvL, numprim, a_dir,a_box);
  
  // Set up for applying limiter. To save on storage, we use the input
  // values as temporaries.

  FArrayBox& dWMinu = a_WMinu;
  FArrayBox& dWPlus = a_WPlus;
  dWMinu.setVal(0.);
  dWPlus.setVal(0.);

  // Compute one-sided differences as inputs for limiting.
  m_util.oneSidedDifferences(dWMinu,dWPlus,a_W,a_dir,a_box);
    
  FArrayBox lambda(a_box, numprim);
  m_gdnvPhysics->charValues(lambda, a_W, a_dir,a_box);

  // Apply limiter on components or on characteristic variables. Either
  // way, must end up with characteristic variables to pass to to the
  // normal predictor utility.

  if (m_useCharLimiting) 
   {
      m_gdnvPhysics->charAnalysis(dWMinu,a_W, a_dir,a_box);
      m_gdnvPhysics->charAnalysis(dWPlus,a_W, a_dir,a_box);
      m_gdnvPhysics->charAnalysis(dW4,   a_W, a_dir,a_box);
      m_util.slopeLimiter(dW4,dWMinu,dWPlus, numprim,a_dir,a_box);
    }
  else
    {
      m_util.slopeLimiter(dW4,dWMinu,dWPlus,numprim,a_dir,a_box);
      m_gdnvPhysics->charAnalysis(dW4,a_W,a_dir,a_box);
    }
  if (m_useFlattening) m_util.applyFlattening(dW4,a_flat,a_box);
  m_util.PLMNormalPredictor(dWMinu,dWPlus,dW4,lambda,a_dtbydx,a_box);

  // Construct the increments to the primitive variables.
  m_gdnvPhysics->charSynthesis(dWMinu,a_W,a_dir,a_box);
  m_gdnvPhysics->charSynthesis(dWPlus,a_W,a_dir,a_box);
    
  // Apply a physics-dependent post-normal predictor step, for example, to 
  // compute source terms that depend on the spatially varying
  // coefficients.
  m_gdnvPhysics->postNormalPredictor(dWMinu,dWPlus,a_W,a_dir,a_box);

  a_WMinu += a_W;
  a_WPlus += a_W;
}
/////////
void
PatchGodunov::PPMNormalPred(FArrayBox&       a_WMinu,
                            FArrayBox&       a_WPlus,
                            const Real&      a_dtbydx,
                            const FArrayBox& a_W,
                            const FArrayBox& a_flat,
                            const int&       a_dir,
                            const Box&       a_box)
{

  // Preliminary van Leer slopes with componentwise limiting (inputs to 
  // fourth-order slope calculation.

  int numprim = m_gdnvPhysics->numPrimitives();

  Box faceBox = a_box;
  faceBox.surroundingNodes(a_dir);
  FArrayBox WFace(faceBox,numprim);
 
  m_util.PPMFaceValues(WFace, a_W, numprim,a_dir,faceBox,m_gdnvPhysics);

  FArrayBox& dWMinu = a_WMinu;
  FArrayBox& dWPlus = a_WPlus;
  dWMinu.setVal(0.);
  dWPlus.setVal(0.);
  dWMinu -= a_W;
  dWPlus -= a_W;
  WFace.shiftHalf(a_dir,1);
  dWMinu += WFace;
  WFace.shift(a_dir,-1);
  dWPlus += WFace;
  
  FArrayBox lambda(a_box, numprim);
  m_gdnvPhysics->charValues(lambda, a_W, a_dir,a_box);

  // Apply limiter on components.

  m_util.PPMLimiter(dWMinu,dWPlus,numprim,a_dir,a_box);
  m_gdnvPhysics->charAnalysis(dWMinu,a_W,a_dir,a_box);
  m_gdnvPhysics->charAnalysis(dWPlus,a_W,a_dir,a_box);

  if (m_useFlattening) 
  {
    m_util.applyFlattening(dWMinu,a_flat,a_box);
    m_util.applyFlattening(dWPlus,a_flat,a_box);
  }
  m_util.PPMNormalPredictor(dWMinu,dWPlus,lambda,a_dtbydx,numprim,a_box);

  // Construct the increments to the primitive variables.
  m_gdnvPhysics->charSynthesis(dWMinu,a_W,a_dir,a_box);
  m_gdnvPhysics->charSynthesis(dWPlus,a_W,a_dir,a_box);
    
  // Apply a physics-dependent post-normal predictor step, for example, to 
  // compute source terms that depend on the spatially varying
  // coefficients.
  m_gdnvPhysics->postNormalPredictor(dWMinu,dWPlus,a_W,a_dir,a_box);

  a_WMinu += a_W;
  a_WPlus += a_W;
}
