/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _POLYTROPIC_PHYSICS_H_
#define _POLYTROPIC_PHYSICS_H_

#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "GodunovPhysics.H"

#include <string>
using std::string;

///
/**
   The base class PolytropicPhysics provides the physics-dependent components 
   for a higher-order Polytropic method for a single patch: characteristic 
   analysis, Riemann solver, quasilinear update, conservative update,
   and transformations between conserved, primitive, and flux variables.
   This class is essentially pure  , i.e. all of its member functions 
   are  , and the ones that have default implementations are ones
   which are optionally defined, i.e. the default definition is to send
   an error message. Physics-dependent versions of this class that are 
   required in real applications are derived from this class by inheritance.
*/
class PolytropicPhysics: public GodunovPhysics
{
public:
  /// Constructor
  /**
   */
  PolytropicPhysics();

  /// Destructor
  /**
   */
  ~PolytropicPhysics();

  /// Define the object
  /**
   */
  void define(ProblemDomain& a_domain);

  ///
  virtual GodunovPhysics* new_godunovPhysics() const;


  /// Compute the maximum wave speed 
  /**
   */
  Real getMaxWaveSpeed(const FArrayBox& a_U,
                       const Box&       a_box) ;

  /// Number of conserved variables
  /**
     Return the number of conserved variables.
  */
  int numConserved() ;

  /// Names of the conserved variables
  /**
     Return the names of the conserved variables.  A default implementation is
     provided that puts in generic names (i.e., "variable#" which "#" ranges
     for 0 to numConserved()-1.
  */
  Vector<string> stateNames();

  /// Number of flux variables
  /**
     Return the  number of flux variables.  This can be greater than the number
     of conserved variables if addition fluxes/face-centered quantities are
     computed.
  */
  int numFluxes() ;

  void getFlux(FArrayBox&       a_flux,
               const FArrayBox& a_whalf,
               const Box&       a_box,
               const int&       a_idir); 

  virtual void postNormalPredictor(FArrayBox& a_dWLow,
                                   FArrayBox& a_dWHigh,
                                   const FArrayBox& a_W,
                                   const int&       a_dir,
                                   const Box&       a_box);

  /// Number of primitive variables
  /**
     Return the number of primitive variables.  This may be greater than the
     number of conserved variables if derived/redundant quantities are also
     stored for convenience.
  */
  int numPrimitives() ;

  /// Perform any preliminary setup needed for the characteristic analysis.
  /**
     This class is primarily here to permit performance optimization by 
     precomputing some of the values required for doing the characteristic 
     analysis. It has a default implementation that consists of doing nothing.
  */
  void charSetup(const FArrayBox& a_W,
                 const int&       a_dir,
                 const Box&       a_box);
  
  bool hasCharSetup();
  
  /// Compute the characteristic decomposition of dW.
  /**
     On input, a_dW contains the increments of the primitive variables. On
     output, it contains the increments in the characteristic variables.
  */
  void charAnalysis(FArrayBox&       a_dW,
                    const FArrayBox& a_W,
                    const int&       a_dir,
                    const Box&       a_box) ;

  /// Compute the sum of the right eigenvectors times the weights dW.
  /**
     On input, a_dW contains the increments of the characteristic variables.
     On output, it contains the increments in the primitive variables.
  */
  
  void charSynthesis(FArrayBox&       a_dW,
                     const FArrayBox& a_W,
                     const int&       a_dir,
                     const Box&       a_box) ;

  /// Compute the eigenvalues.
  /**
   */
  
  void charValues(FArrayBox&       a_lambda,
                  const FArrayBox& a_W,
                  const int&       a_dir,
                  const Box&       a_box) ;

  /// Compute the solution to the Riemann problem.
  /**
     Given input left and right states in a direction, a_dir, compute a
     Riemann problem and generate fluxes at the faces within a_box.
  */
  void riemann(FArrayBox&       a_WStar,
               const FArrayBox& a_WLeft,
               const FArrayBox& a_WRight,
               const FArrayBox& a_W,
               const Real&      a_time,
               const int&       a_dir,
               const Box&       a_box) ;

  /// Post-normal predictor calculation.
  /**
     Add increment to normal predictor to account e.g. for source terms due to
     spatially-varying coefficients.
  */
  void postNormalPredictor(FArrayBox&       a_AdWdx,
                           const FArrayBox& a_wHalf,
                           const Real&      a_scale,
                           const int&       a_dir,
                           const Box&       a_box) ;

  /// Compute primitive variables from conserved variables.
  /**
   */
  void consToPrim(FArrayBox&       a_W,
                  const FArrayBox& a_U,
                  const Box&       a_box) ;


  void quasilinearUpdate(FArrayBox&       a_dWdx,
                         const FArrayBox& a_WHalf,
                         const FArrayBox& a_W,
                         const Real&      a_scale,
                         const int&       a_dir,
                         const Box&       a_box);


  /// Interval within the primitive variables corresponding to the velocities
  /**
     Return the interval of component indices within the primitive variable 
     of the velocities.  Used for slope flattening (slope computation) and
     computing the divergence of the velocity (artificial viscosity).
   */
  virtual Interval velocityInterval();

  /// Component index within the primitive variables of the pressure
  /**
     Return the component index withn the primitive variables for the
     pressure.  Used for slope flattening (slope computation).
   */
  virtual int pressureIndex();

  /// Component index within the primitive variables of the bulk modulus
  /**
     Return the component index withn the primitive variables for the
     bulk modulus.  Used for slope flattening (slope computation) used
     as a normalization to measure shock strength. 
   */
  virtual int bulkModulusIndex();

private:
  // Disallowed for all the usual reasons
  void operator=(const PolytropicPhysics& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  PolytropicPhysics(const PolytropicPhysics& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
