/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "ModianoIBC.H"
#include "ModianoF_F.H"
#include "PatchPolytropicGasF_F.H"
#include "SolidF_F.H"
#include "LoHiCenter.H"
///
/**
   IBC for doing the modiano problem.
   The lower left corner of the domain corresponds to the bottom corner
   of the channel, so it sort of looks like this:
*/
ModianoIBC::~ModianoIBC()
{
}

PhysIBC* 
ModianoIBC::new_physIBC()
{
  ModianoIBC* retval = new ModianoIBC(m_gamma,          
                                      m_waveAmp,        
                                      m_waveWidth,      
                                      m_center,         
                                      m_waveDir,        
                                      m_freeStreamOnly, 
                                      m_useNegativeWave);
  
  return static_cast<PhysIBC*>(retval);
}

void 
unifyVectorModiano(RealVect& a_normal, Real& a_sumSquare)
{
  Real sum = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      sum += a_normal[idir]*a_normal[idir];
    }
  sum = sqrt(sum);
  //divide both by the sum of the square
  assert(sum > 1.0e-12);
  a_normal /= sum;
  a_sumSquare = sum;
}

void 
posifyVectorModiano(RealVect& a_vect, IntVect& a_sign)
{
  a_sign = IntVect::Unit;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(a_vect[idir] < 0.0)
        {
          a_sign[idir] = -1;
          a_vect[idir] = Abs(a_vect[idir]);
        }
    }  
}
void 
sortVectorModiano(RealVect& a_vect, IntVect& a_ivmap)
{
  for(int idir = 0; idir < SpaceDim; idir++)
    a_ivmap[idir] = idir;

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      assert(a_vect[idir] >= 0.0);
      for(int jdir = 1; jdir < SpaceDim; jdir++)
        {
          //if bigger normal is first, switch them
          if(a_vect[jdir-1]  > a_vect[jdir])
            {
              {
                Real temp = a_vect[jdir-1];
                a_vect[jdir-1] = a_vect[jdir];
                a_vect[jdir]   = temp;
              }
              {
                int itemp = a_ivmap[jdir-1];
                a_ivmap[jdir-1] = a_ivmap[jdir];
                a_ivmap[jdir] = itemp;
              }
            }
        }
    }
}
ModianoIBC::ModianoIBC(const Real&     a_gamma,
                       const Real&     a_waveAmp,
                       const Real&     a_waveWidth,
                       const RealVect& a_center,
                       const RealVect& a_waveDir,
                       bool            a_freeStreamOnly,
                       bool            a_useNegativeWave)
  :PhysIBC()
{
  m_gamma       = a_gamma       ;
  m_waveAmp     = a_waveAmp     ;
  m_waveWidth   = a_waveWidth   ;
  m_waveDir     = a_waveDir     ;

  m_iNegWave    = 0;
  m_useNegativeWave = a_useNegativeWave;
  if(a_useNegativeWave)
    m_iNegWave  = 1;

  Real sumSquare;
  unifyVectorModiano(m_waveDir, sumSquare);
  m_center      = a_center      ;
  m_idoFreeStreamOnly = 0;
  if(a_freeStreamOnly)
    m_idoFreeStreamOnly = 1;

  /**/
  FORT_SETMODIANO(CHF_CONST_REAL(m_gamma),
                  CHF_CONST_REAL(m_waveAmp),
                  CHF_CONST_REAL(m_waveWidth),
                  CHF_REALVECT(m_center),
                  CHF_REALVECT(m_waveDir),
                  CHF_CONST_INT(m_idoFreeStreamOnly),
                  CHF_CONST_INT(m_iNegWave));
  /**/
  m_freeStreamOnly = a_freeStreamOnly;
  m_isFortranCommonSet = true;
  m_isDefined = false;
}

void 
ModianoIBC::define(const ProblemDomain&  a_domain,
                   const Real&           a_dx)
{
  m_isDefined = true;;
  m_domain = a_domain;
  m_dx = a_dx;
}

void 
ModianoIBC::fluxBC(FArrayBox&            a_flux,
                   const FArrayBox&      a_primCenter,
                   const FArrayBox&      a_primExtrap,
                   const int&            a_dir,
                   const Side::LoHiSide& a_side,
                   const Real&           a_time)
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  assert(!m_domain.isPeriodic(a_dir));

  Box FBox = a_flux.box();
  Box cellBox = FBox;

  // Determine which side and thus shift directions
  int isign = ::sign(a_side);
  cellBox.shiftHalf(a_dir,isign);

  // Is there a domain boundary next to this grid
  if (!m_domain.contains(cellBox))
    {
      cellBox &= m_domain;
      // Find the strip of cells next to the domain boundary
      Box boundaryBox = bdryBox(cellBox, a_dir, a_side, 1);

      // Shift things to all line up correctly
      boundaryBox.shiftHalf(a_dir,-isign);
      a_flux.shiftHalf(a_dir,-isign);

      // Set the boundary fluxes
      /**/
      FORT_MODIANOBC(CHF_FRA(a_flux),
                     CHF_CONST_FRA(a_primExtrap),
                     CHF_CONST_REAL(m_dx),
                     CHF_CONST_INT(a_dir),
                     CHF_BOX(boundaryBox));
      /**/

      // Shift returned fluxes to be face centered
      a_flux.shiftHalf(a_dir,isign);

    }
}

void 
ModianoIBC::initialize(LevelData<FArrayBox>& a_conState)
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  
  // Iterator of all grids in this level
  for(DataIterator dit = a_conState.dataIterator(); dit.ok(); ++dit)
    {
      // Storage for current grid
      FArrayBox& conFAB = a_conState[dit()];

      initialize(conFAB);
    }
}

void 
ModianoIBC::initialize(FArrayBox& a_conFAB) const
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  
  // Box of current grid
  Box uBox = a_conFAB.box();
  // Set up initial condition in this grid
  /**/
  FORT_MODIANOINIT(CHF_CONST_FRA(a_conFAB),
                   CHF_CONST_REAL(m_dx),
                   CHF_BOX(uBox));
  /**/
}

void 
ModianoIBC::setToExact(LevelData<FArrayBox>& a_conState,
                       const Real&       a_time) const
{
  assert(m_isDefined);
  assert(m_isFortranCommonSet);
  
  // Iterator of all grids in this level
  for(DataIterator dit = a_conState.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& conFAB = a_conState[dit()];

      // Box of current grid
      Box uBox = conFAB.box();
      // Set up initial condition in this grid
      /**/
      FORT_MODIANOEXACT(CHF_CONST_FRA(conFAB),
                        CHF_CONST_REAL(m_dx),
                        CHF_CONST_REAL(a_time),
                        CHF_BOX(uBox));
      /**/

    } //end loop over boxes
}
// Set boundary slopes:
//   The boundary slopes in a_dW are already set to one sided difference
//   approximations.  If this function doesn't change them they will be
//   used for the slopes at the boundaries.
void ModianoIBC::setBdrySlopes(FArrayBox&       a_dW,
                                 const FArrayBox& a_W,
                                 const int&       a_dir,
                                 const Real&      a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  // In periodic case, this doesn't do anything
  if (!m_domain.isPeriodic(a_dir))
    {
      Box loBox,hiBox,centerBox,domain;
      int hasLo,hasHi;
      Box slopeBox = a_dW.box()&m_domain;

      // Generate the domain boundary boxes, loBox and hiBox, if there are
      // domain boundarys there
      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
                 slopeBox,m_domain,a_dir);

      // Set the boundary slopes if necessary
      if ((hasLo != 0) || (hasHi != 0))
        {
          FORT_SLOPEBCS(CHF_FRA(a_dW),
                        CHF_CONST_FRA(a_W),
                        CHF_CONST_REAL(m_dx),
                        CHF_CONST_INT(a_dir),
                        CHF_BOX(loBox),
                        CHF_CONST_INT(hasLo),
                        CHF_BOX(hiBox),
                        CHF_CONST_INT(hasHi));
        }
    }
}
