/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>

#include "ParmParse.H"
#include "HDF5.H"
#include "parstream.H"
#include "ModianoIBC.H"

#include "ArrayView.H"
#include "AMRLevel.H"
#include "AMR.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "LevelData.H"
#include "UGIO.H"
#include "DebugDump.H"
#include "ArrayView.H"
#include "PatchGodunov.H"
#include "PatchPolytropicGas.H"
#include "AMRLevelPolytropicGasFactory.H"
#include "AMRLevelPolytropicGas.H"
#include "AMR.H"
 
#if CH_SPACEDIM==2
IntVect ivdebug(7,1);
#elif CH_SPACEDIM==3
IntVect ivdebug(1, 1, 0);
#endif

//#define TRAP_FPE  //(should be off by default) (does not work on all machines!)
#ifdef TRAP_FPE
extern "C" {
#include <fpu_control.h>
}
/* IM: Invalid operation mask
 * DM: Denormalized operand mask 
 * ZM: Zero-divide mask
 * OM: Overflow mask 
 * UM: Underflow mask 
  ---um  flag both uninitialized memory and 
     but will also get many false positives
 * PM: Precision (inexact result) mask 
  ---(pm is kinda stupid)
*/
static void __attribute__ ((constructor)) trapfpe(void)
{
  fpu_control_t cw = 
    _FPU_DEFAULT & ~(_FPU_MASK_IM | _FPU_MASK_ZM | _FPU_MASK_OM | _FPU_MASK_DM | _FPU_MASK_UM);

  _FPU_SETCW(cw);
}
#endif


void 
generateCalculatedData(LevelData<FArrayBox>& a_calcData,
                       Real&                 a_finalTime,
                       DisjointBoxLayout&    a_grids,
                       const Box&            domain,
                       const Real&           dx,
                       const bool&           thisCalcCoarse);


void 
generateExactData(LevelData<FArrayBox>& a_exactData,
                  const DisjointBoxLayout& a_grids,
                  const Box& a_domain, 
                  const Real& a_dx,
                  const Real& a_finalTime,
                  const int& a_ncomp);

void makeGeometry(Box& a_domain, 
                  Real& a_dx);

void dumpmemoryatexit();

void
generateError(LevelData<FArrayBox>& a_error,
              DisjointBoxLayout&    a_grids,
              const Box& a_domain, 
              const Real& a_dx,
              const bool& a_thisCalcCoar);
/***************/
/***************/
void
outputError(const LevelData<FArrayBox>& a_errorFine,
            const DisjointBoxLayout& a_gridsFine, 
            const Box& a_domainFine,
            const LevelData<FArrayBox>& a_errorCoar,
            const DisjointBoxLayout& a_gridsCoar, 
            const Box& a_domainCoar);

/***************/
/***************/
void
compareError(const LevelData<FArrayBox>& a_fineError,
             const DisjointBoxLayout& a_gridsFine, 
             const Box& a_domainFine,
             const LevelData<FArrayBox>& a_coarError,
             const DisjointBoxLayout& a_gridsCoar, 
             const Box& a_domainCoar);
/***************/
/***************/
RealVect getNormalVector(const RealVect& rampNormal);
/***************/
/***************/
int 
main(int a_argc, char* a_argv[])
{
#ifdef MPI
  MPI_Init(&a_argc,&a_argv);
  {
    setChomboMPIErrorHandler();
#endif

    // Check for an input file
    char* inFile = NULL;

#ifndef NDEBUG
#ifdef TRAP_FPE
    trapfpe();
#endif
#endif

    if (a_argc > 1)
      {
        inFile = a_argv[1];
      }
    else
      {
        pout() << "Usage: <executable name> <inputfile>" << endl;
        pout() << "No input file specified" << endl;
        return -1;
      }
    ParmParse pp(a_argc-2,a_argv+2,NULL,inFile);

    Box domainFine, domainCoar;
    Real dxFine, dxCoar;
    makeGeometry(domainFine, dxFine);
    domainCoar = coarsen(domainFine, 2);
    dxCoar = 2.0*dxFine;

    LevelData<FArrayBox>     errorFine, errorCoar;
    DisjointBoxLayout gridsFine, gridsCoar;
    //generate fine error
    //generate coarse error
    generateError(errorFine, gridsFine, domainFine, dxFine, false);
    generateError(errorCoar, gridsCoar, domainCoar, dxCoar, true);

    //output both errors to file
    outputError(errorFine,  gridsFine, domainFine,
                errorCoar,  gridsCoar, domainCoar);


    //compute convergence rates and output to stdout
    compareError(errorFine,  gridsFine, domainFine,
                 errorCoar,  gridsCoar, domainCoar);

#ifdef MPI
    dumpmemoryatexit();
  }
  MPI_Finalize();
#endif
}
/***************/
/***************/
void
generateError(LevelData<FArrayBox>& a_error,
              DisjointBoxLayout&    a_grids,
              const Box&            a_domain, 
              const Real&           a_dx,
              const bool&           a_thisCalcCoar)
{
  LevelData<FArrayBox> calcData;
  Real finalTime;
  generateCalculatedData(calcData,  finalTime, a_grids, 
                         a_domain, a_dx, a_thisCalcCoar);

  LevelData<FArrayBox> exactData;

  int nvar = calcData.nComp();
  generateExactData(exactData, a_grids, 
                    a_domain, a_dx, finalTime, nvar);

  a_error.define(a_grids, nvar, IntVect::Zero);

  Interval interv(0, nvar-1);
  //truncation error is calc-exact/dt
  for(DataIterator dit = a_grids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& box = a_grids.get(dit());
      a_error[dit()].copy(box, interv, box, 
                          exactData[dit()], interv);
      a_error[dit()] -= calcData[dit()];
      a_error[dit()] /= finalTime;
    }

  string fileFine("pltFineExact.hdf5");
  string fileCoar("pltCoarExact.hdf5");

  string filename;
  if(a_thisCalcCoar)
    filename = fileCoar;
  else
    filename = fileFine;

  WriteUGHDF5(filename, a_grids, exactData, a_domain);
}

/***************/
/***************/
void
outputError(const LevelData<FArrayBox>& a_errorFine,
            const DisjointBoxLayout& a_gridsFine, 
            const Box& a_domainFine,
            const LevelData<FArrayBox>& a_errorCoar,
            const DisjointBoxLayout& a_gridsCoar, 
            const Box& a_domainCoar)
{
  string fileFine("pltFineError.hdf5");
  string fileCoar("pltCoarError.hdf5");

  WriteUGHDF5(fileFine, a_gridsFine, a_errorFine, a_domainFine);

  WriteUGHDF5(fileCoar, a_gridsCoar, a_errorCoar, a_domainCoar);
}
/***************/
/***************/
void 
generateExactData(LevelData<FArrayBox>& a_exactData,
                  const DisjointBoxLayout& a_grids,
                  const Box& a_domain, 
                  const Real& a_dx,
                  const Real& a_finalTime,
                  const int&  a_nvar)
{
  a_exactData.define(a_grids, a_nvar, IntVect::Zero);

  RealVect modianoAxis, center;
  Real waveAmp, waveWidth, gamma;
  {
    ParmParse pp;
    pp.get("wave_amplitude", waveAmp);
    pp.get("wave_width", waveWidth);
    pp.get("gamma",gamma);

    RealVect channelNormal;
    vector<Real>  channelNormalVect(SpaceDim);
    vector<Real> centerpp(SpaceDim,0.5);
    pp.getarr("channel_normal",channelNormalVect, 0, SpaceDim);
    pp.getarr("wave_center",centerpp,0,SpaceDim);

    for(int idir = 0; idir < SpaceDim; idir++)
      {
        channelNormal[idir] = channelNormalVect[idir];
        center[idir] = centerpp[idir];
      }
    modianoAxis = getNormalVector(channelNormal);;
  }

  ModianoIBC mibc(gamma, waveAmp, waveWidth,
                  center, modianoAxis, false);

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }
  ProblemDomain probdomain(a_domain.smallEnd(), a_domain.bigEnd(), is_periodic);
  mibc.define(probdomain, a_dx);
  mibc.setToExact(a_exactData,  a_finalTime);
}
/***************/
/***************/
void
compareError(const LevelData<FArrayBox>& a_errorFine,
             const DisjointBoxLayout& a_gridsFine, 
             const Box& a_domainFine,
             const LevelData<FArrayBox>& a_errorCoar,
             const DisjointBoxLayout& a_gridsCoar, 
             const Box& a_domainCoar)
{
  IntVect ivghost = IntVect::Zero;
  Real dxCoar = 2;
  Real dxFine = 1;
  pout() << "==============================================" << endl;
  for(int comp = 0; comp < a_errorFine.nComp(); comp++)
    {
      pout() << "Comparing error in variable  " << comp << endl;
      pout() << "==============================================" << endl;
      for(int inorm = 0; inorm < 3; inorm++)
        {
          
          if(inorm == 0)
            {
              pout() << endl << "Using max norm." << endl;
            }
          else
            {
              pout() << endl << "Using L-" << inorm << "norm." << endl;
            }
          Interval interv(comp, comp);
          Real coarnorm = norm(a_errorCoar, interv, inorm);
          Real finenorm = norm(a_errorFine, interv, inorm);

          //fix up norm because boxtools norm is screwy
          if(inorm != 0)
            {
              int ipower = SpaceDim/inorm;
              Real factorFine = pow(dxFine, ipower);
              Real factorCoar = pow(dxCoar, ipower);
              coarnorm *= factorCoar;
              finenorm *= factorFine;
            }
          pout() << "Coarse Error Norm = " << coarnorm << endl;
          pout() << "Fine   Error Norm = " << finenorm << endl;
          if(Abs(finenorm) > 1.0e-8)
            {
              Real order = log(coarnorm/finenorm)/log(2.0);
              pout() << "Order of scheme = " << order << endl;
            }
        }
      pout() << "==============================================" << endl;
    }

}
/***************/
/***************/
void
makeGeometry(Box& a_domain, 
             Real& a_dx)
{
  //parse input file.  single level
  ParmParse pp;
  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);
 
  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain = Box(lo, hi);
  
  Real prob_hi;
  int numOpen = n_cell[0];
  pp.get("domain_length",prob_hi);
  a_dx = prob_hi/numOpen;
}
void 
unifyVector(RealVect& a_normal, Real& a_sumSquare)
{
  Real sum = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      sum += a_normal[idir]*a_normal[idir];
    }
  sum = sqrt(sum);
  //divide both by the sum of the square
  assert(sum > 1.0e-12);
  a_normal /= sum;
  a_sumSquare = sum;
}

void 
posifyVector(RealVect& a_vect, IntVect& a_sign)
{
  a_sign = IntVect::Unit;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(a_vect[idir] < 0.0)
        {
          a_sign[idir] = -1;
          a_vect[idir] = Abs(a_vect[idir]);
        }
    }  
}
void 
sortVector(RealVect& a_vect, IntVect& a_ivmap)
{
  for(int idir = 0; idir < SpaceDim; idir++)
    a_ivmap[idir] = idir;

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      assert(a_vect[idir] >= 0.0);
      for(int jdir = 1; jdir < SpaceDim; jdir++)
        {
          //if bigger normal is first, switch them
          if(a_vect[jdir-1]  > a_vect[jdir])
            {
              {
                Real temp = a_vect[jdir-1];
                a_vect[jdir-1] = a_vect[jdir];
                a_vect[jdir]   = temp;
              }
              {
                int itemp = a_ivmap[jdir-1];
                a_ivmap[jdir-1] = a_ivmap[jdir];
                a_ivmap[jdir] = itemp;
              }
            }
        }
    }
}
/***************/
/***************/
RealVect getNormalVector(const RealVect& channelNormal)
{
  RealVect channelSlope;
  //make slope perp to normal
  RealVect channelSortNormal = channelNormal;
  IntVect ivsort, ivsigns;
  posifyVector(channelSortNormal, ivsigns);
  sortVector(channelSortNormal, ivsort);
  //divide sorted normal by the largest one
  channelSortNormal /= channelSortNormal[SpaceDim-1];
#if CH_SPACEDIM==2
  if(Abs(channelSortNormal[0]) < 1.0e-12)
    {
      //really a one-d channel.  make the slope perp 
      //to the normal
      channelSlope = BASISREALV(ivsort[0]);
    }
  else
    {
      //we have a genuine 2D channel.
      channelSlope = RealVect::Unit;
      channelSlope[ivsort[0]] =-1.0/channelSortNormal[0];
      channelSlope[ivsort[0]] *= ivsigns[ivsort[0]];
    }
#elif CH_SPACEDIM==3
  if(Abs(channelSortNormal[0]) < 1.0e-12 &&
     Abs(channelSortNormal[1]) < 1.0e-12)
    {
      //really a one-d channel.  make the slope perp 
      //to the normal
      channelSlope = BASISREALV(ivsort[0]);
    }
  else if(Abs(channelSortNormal[0]) < 1.0e-12)
    {
      //really a 2-d channel.  make the slope perp 
      //to the normal
      //further confuse the issue by dividing out the biggest one
      //this way there is only one non-unity number in the slope
      //channelSortNormal /= channelSortNormal[1];
      channelSlope = RealVect::Unit;
      channelSlope[ivsort[0]] = 0.0;
      channelSlope[ivsort[1]] = -1.0/channelSortNormal[1];
      channelSlope[ivsort[1]] *= ivsigns[ivsort[1]];
    }
  else
    {
      //genuine 3d channel.
      channelSlope = RealVect::Unit;
      channelSlope[ivsort[0]] = -0.5/channelSortNormal[0];
      channelSlope[ivsort[0]] *= ivsigns[ivsort[0]];
      channelSlope[ivsort[1]] = -0.5/channelSortNormal[1];
      channelSlope[ivsort[1]] *= ivsigns[ivsort[1]];
    }
#else
  bogus_ch_spacedim_macro();
#endif

  Real dotprod = 0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      dotprod += channelSlope[idir]*channelNormal[idir];
    }
  if(Abs(dotprod > 1.0e-12))
    {
      MayDay::Error("getnormalvector failed");
    }
  return channelSlope;
}
/************/
/************/
void 
generateCalculatedData(LevelData<FArrayBox>& a_calcData,
                       Real& a_finalTime,
                       DisjointBoxLayout& a_grids,
                       const Box&  a_domain,
                       const Real& a_dx,
                       const bool& a_thisCalcCoar)
{
  // read inputs
  ParmParse pp;

  int verbosity;
  pp.get("verbosity",verbosity);
  assert(verbosity >= 0);

  Real gamma = 1.4;
  pp.get("gamma",gamma);


  int nstop = 1;
  //only running one step
  Real stopTime = 1.0;

  bool is_periodic[SpaceDim];
  for (int i = 0; i < SpaceDim; ++i) 
    {
      is_periodic[i] = false;
    }

  //single level calc
  int maxLevel = 0;
  // note this requires a ref_ratio to be defined for the 
  // finest level (even though it will never be used)
  Vector<int> refRatios(1,1);

  Real cfl = 0.8;
  pp.get("cfl",cfl);

  //only running one time step
  Real initialCFL = cfl;

  Real fixedDt = -1;
  pp.get("fixed_dt",fixedDt);
  if(a_thisCalcCoar && fixedDt > 0)
    {
      fixedDt *= 2.0;
    }

  //not relevant because only running one step
  Real max_dt_growth = 1.1;
  Real dt_tolerance_factor = 1.1;

  Real domainLength;
  pp.get("domain_length",domainLength);

  int ifourth, iflatten;
  pp.get("use_fourth_order_slopes", ifourth);
  pp.get("use_flattening"         , iflatten);
  bool useFourthOrderSlopes = (ifourth  ==1);
  bool useFlattening        = (iflatten ==1);

  ProblemDomain probDomain(a_domain.smallEnd(),
                            a_domain.bigEnd(),
                            is_periodic);

  // create initial/boundary condition object
  pout() << "Modiano initial and boundary conditions" << endl;
      
  RealVect modianoAxis;
  RealVect channelNormal;
  vector<Real>  channelNormalVect(SpaceDim);
  pp.getarr("channel_normal",channelNormalVect, 0, SpaceDim);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      channelNormal[idir] = channelNormalVect[idir];
    }
  modianoAxis = getNormalVector(channelNormal);;
  Real waveAmp, waveWidth;
  pp.get("wave_amplitude", waveAmp);
  pp.get("wave_width", waveWidth);
      
  vector<Real> centerpp(SpaceDim,0.5);
  RealVect center;
  pp.getarr("wave_center",centerpp,0,SpaceDim);
  for (int i = 0; i < SpaceDim; i++) 
    center[i] = centerpp[i];
      
  bool freeStreamOnly = false;
      
  ModianoIBC ibc(gamma, waveAmp, waveWidth,
                 center, modianoAxis, freeStreamOnly);

  //dummies 
  Real refineThresh = 0.7;
  int tagBufferSize = 1;

  AMRLevelPolytropicGasFactory amrGodFact;

  RefCountedPtr<PatchGodunov> patchPolytropicGas = static_cast<PatchGodunov*>
                                                   (new PatchPolytropicGas());

  // Set up the patch integrator for polytropic gas dynamics
  patchPolytropicGas->setPhysIBC(&ibc);
  patchPolytropicGas->setSlopeParameters(useFourthOrderSlopes,
                                         useFlattening);
  patchPolytropicGas->setArtificialViscosity(false, 1.0);

  // Set up the AMRLevel... factory
  amrGodFact.CFL(cfl);
  amrGodFact.domainLength(domainLength);
  amrGodFact.refinementThreshold(refineThresh);
  amrGodFact.tagBufferSize(tagBufferSize);
  amrGodFact.patchGodunov(patchPolytropicGas);
  amrGodFact.verbosity(verbosity);
  amrGodFact.initialDtMultiplier(initialCFL);

  AMR amr;

  // Set up the AMR object
  amr.define(maxLevel,refRatios,probDomain,&amrGodFact);

  if (fixedDt > 0)
    {
      amr.fixedDt(fixedDt);
    }

  // Set grid generation parameters
  int maxGridSize;
  int blockFactor = 2;
  Real fillRatio = 0.8;
  pp.get("max_grid_size", maxGridSize);
  amr.maxGridSize(maxGridSize);
  amr.blockFactor(blockFactor);
  amr.fillRatio(fillRatio);

  // The hyperbolic codes use a grid buffer of 1
  amr.gridBufferSize(1);

  // Set output parameters
  amr.checkpointInterval(-1);
  amr.plotInterval(1);
  amr.maxDtGrow(max_dt_growth);
  amr.dtToleranceFactor(dt_tolerance_factor);

  if(a_thisCalcCoar)
    {
      std::string prefix("pltCoar");
      amr.plotPrefix(prefix);
    }
  else
    {
      std::string prefix("pltFine");
      amr.plotPrefix(prefix);
    }

  amr.verbosity(verbosity);

  // initialize hierarchy of levels
  amr.setupForNewAMRRun();


  // run 
  amr.run(stopTime,nstop);
  // output last pltfile and statistics
  amr.conclude();
 
  Vector<AMRLevel*> amrLevelData = amr.getAMRLevels();
  assert(amrLevelData.size() == 1);
 
  AMRLevelPolytropicGas* amrGodunovData = dynamic_cast<AMRLevelPolytropicGas*>(amrLevelData[0]);
  
  const LevelData<FArrayBox>& amrData = amrGodunovData->getStateNew();
  a_grids = amrData.disjointBoxLayout();
  int ncomp = amrData.nComp();
  a_calcData.define(a_grids, ncomp, IntVect::Zero);
  Interval interv(0, ncomp-1);

  amrData.copyTo(interv, a_calcData, interv);
  a_finalTime = amr.getCurrentTime();
}


