/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// LevelHelmholtzSolver.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "LevelHelmholtzSolver.H"

// default constructor
LevelHelmholtzSolver::LevelHelmholtzSolver()
{
  setDefaultValues();
}

// destructor
LevelHelmholtzSolver::~LevelHelmholtzSolver()
{
  clearMemory();
}

LevelHelmholtzSolver::LevelHelmholtzSolver(const DisjointBoxLayout& a_grids,
					   const DisjointBoxLayout* a_baseGrids,
					   const ProblemDomain& a_domain,
					   Real a_dxLevel, 
					   int a_nRefCrse,
					   const BaseHelmholtzOp* const a_opin,
                                           int a_ncomp)
{

  setDefaultValues();
  LevelSolver::define(a_grids, a_baseGrids, a_domain, a_dxLevel, 
		      a_nRefCrse, a_opin, a_ncomp);
}


LevelHelmholtzSolver::LevelHelmholtzSolver(const DisjointBoxLayout& a_grids,
					   const DisjointBoxLayout* a_baseGrids,
					   const Box& a_domain,
					   Real a_dxLevel, 
					   int a_nRefCrse,
					   const BaseHelmholtzOp* const a_opin,
                                           int a_ncomp)
{

  setDefaultValues();
  ProblemDomain physdomain(a_domain);
  LevelSolver::define(a_grids, a_baseGrids, physdomain, a_dxLevel, 
		      a_nRefCrse, a_opin, a_ncomp);
}


void
LevelHelmholtzSolver::define(const DisjointBoxLayout& a_grids, 
			     const DisjointBoxLayout* a_baseGrids,
			     const Box& a_domain, 
			     Real a_dxLevel,
			     int a_nRefCrse,
			     const BaseHelmholtzOp* const a_opin,
                             int a_ncomp)
{
  ProblemDomain physdomain(a_domain);
  LevelSolver::define(a_grids, a_baseGrids, physdomain, a_dxLevel,
		      a_nRefCrse, a_opin, a_ncomp);
}


void
LevelHelmholtzSolver::define(const DisjointBoxLayout& a_grids, 
			     const DisjointBoxLayout* a_baseGrids,
			     const ProblemDomain& a_domain, 
			     Real a_dxLevel,
			     int a_nRefCrse,
			     const BaseHelmholtzOp* const a_opin,
                             int a_ncomp)
{
  LevelSolver::define(a_grids, a_baseGrids, a_domain, a_dxLevel,
		      a_nRefCrse, a_opin, a_ncomp);
}


// this is just here to ensure that we don't try to get things going
// with a non-helmholtzOp
void
LevelHelmholtzSolver::define(const DisjointBoxLayout& a_grids, 
			     const DisjointBoxLayout* a_baseGrids,
			     const Box& a_domain, 
			     Real a_dxLevel,
			     int a_nRefCrse,
			     const LevelOp* const a_opin,
                             int a_ncomp)
{
  MayDay::Error("Tried to define a LevelHelmholtzSolver with a non-BaseHelmholtzOp");

}

// this is just here to ensure that we don't try to get things going
// with a non-helmholtzOp
void
LevelHelmholtzSolver::define(const DisjointBoxLayout& a_grids, 
			     const DisjointBoxLayout* a_baseGrids,
			     const ProblemDomain& a_domain, 
			     Real a_dxLevel,
			     int a_nRefCrse,
			     const LevelOp* const a_opin,
                             int a_ncomp)
{
  MayDay::Error("Tried to define a LevelHelmholtzSolver with a non-BaseHelmholtzOp");

}


// allows coefficient to be changed on the fly -- multiply beta by a_scale
void
LevelHelmholtzSolver::scaleBeta(Real a_scale)
{
  
  assert (m_levelOpPtr != NULL);

  BaseHelmholtzOp* helmOpPtr = (BaseHelmholtzOp*) m_levelOpPtr;

  if (helmOpPtr != NULL) {
    helmOpPtr->scaleBeta(a_scale);
  } else {
    MayDay::Error("Bad levelOp->helmholtzOp conversion!");
  }

  LevelMG* levelmgPtr = &m_levelMG;

  while (levelmgPtr != NULL) {
    helmOpPtr = (BaseHelmholtzOp*) levelmgPtr->levelOpPtr();
    if (helmOpPtr != NULL) {
      helmOpPtr->scaleBeta(a_scale);
    } else {
      MayDay::Error("Bad levelOp->helmholtzOp conversion!");
    }

    levelmgPtr = levelmgPtr->lCoarsePtr();
  }
}
  

/// allows BC to be changed on the fly
void
LevelHelmholtzSolver::setBC(const DomainGhostBC& a_BC) 
{
  assert (m_levelOpPtr != NULL);
  
  BaseHelmholtzOp* helmOpPtr = (BaseHelmholtzOp*) m_levelOpPtr;

  if (helmOpPtr != NULL) {
    helmOpPtr->setDomainGhostBC(a_BC);
  } else {
    MayDay::Error("Bad levelOp->helmholtzOp conversion!");
  }

  LevelMG* levelmgPtr = &m_levelMG;

  while (levelmgPtr != NULL) {
    helmOpPtr = (BaseHelmholtzOp*) levelmgPtr->levelOpPtr();
    if (helmOpPtr != NULL) {
      helmOpPtr->setDomainGhostBC(a_BC);
    } else {
      MayDay::Error("Bad levelOp->helmholtzOp conversion!");
    }

    levelmgPtr = levelmgPtr->lCoarsePtr();
  }
}
  


