/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMR_LEVEL_IDEAL_MHD_H_
#define _AMR_LEVEL_IDEAL_MHD_H_

#include "FArrayBox.H"
#include "LevelData.H"
#include "AMRLevel.H"
#include "CoarseAverage.H"
#include "FineInterp.H"
#include "LevelFluxRegister.H"
#include "LevelSolver.H"
#include "Poisson.H"

#include "LevelGodunov.H"
#include "PhysIBC.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "DisjointBoxLayout.H"

/// AMR Godunov
/**
 */
class AMRLevelIdealMHD : public AMRLevel
{
public:
  /// Constructor
  /**
   */
  AMRLevelIdealMHD();

  /// Destructor
  /**
   */
  virtual ~AMRLevelIdealMHD();

  /// This instance should never get called - historical
  /**
   */
  virtual void define(AMRLevel*  a_coarserLevelPtr,
                      const Box& a_problemDomain,
                      int        a_level,
                      int        a_refRatio);

  /// Define new AMR level
  /**
   */
  virtual void define(AMRLevel*            a_coarserLevelPtr,
                      const ProblemDomain& a_problemDomain,
                      int                  a_level,
                      int                  a_refRatio);


  /// Advance by one timestep
  /**
   */
  virtual Real advance();

  /// Things to do after a timestep
  /**
   */
  virtual void postTimeStep();

  /// Create tags for regridding
  /**
   */
  virtual void tagCells(IntVectSet& a_tags);

  /// Create tags at initialization
  /**
   */
  virtual void tagCellsInit(IntVectSet& a_tags);

  /// Set up data on this level after regridding
  /**
   */
  virtual void regrid(const Vector<Box>& a_newGrids);

  /// Initialize grids
  /**
   */
  virtual void initialGrid(const Vector<Box>& a_newGrids);

  /// Initialize data
  /**
   */
  virtual void initialData();

  /// Things to do after initialization
  /**
   */
  virtual void postInitialize();

  /// Create a local PatchGodunov factory using the argument as a factory
  /**
   */
  virtual void patchGodunov(const PatchGodunov* const a_patchGodunov);

  /// Create a local PoissonOp factory using the argument as a factory
  /**
   */
  virtual void poissonOp(const PoissonOp* const a_poissonOp);

#ifdef HDF5
  /// Write checkpoint header
  /**
   */
  virtual void writeCheckpointHeader(HDF5Handle& a_handle) const;

  /// Write checkpoint data for this level
  /**
   */
  virtual void writeCheckpointLevel(HDF5Handle& a_handle) const;

  /// Read checkpoint header
  /**
   */
  virtual void readCheckpointHeader(HDF5Handle& a_handle);

  /// Read checkpoint data for this level
  /**
   */
  virtual void readCheckpointLevel(HDF5Handle& a_handle);

  /// Write plotfile header
  /**
   */
  virtual void writePlotHeader(HDF5Handle& a_handle) const;

  /// Write plotfile data for this level
  /**
   */
  virtual void writePlotLevel(HDF5Handle& a_handle) const;
#endif

  /// Returns the dt computed earlier for this level
  /**
   */
  virtual Real computeDt();

  /// Compute dt using initial data
  /**
   */
  virtual Real computeInitialDt();

  /// Set the CFL number
  /**
   */
  virtual void CFL(Real a_cfl);

  /// Set the physical dimension of the longest side of the domain
  /**
   */
  virtual void domainLength(Real a_domainLength);

  /// Set the refinement threshold
  /**
   */
  virtual void refinementThreshold(Real a_refineThresh);

  /// Set the tag buffer size
  /**
   */
  void tagBufferSize(int a_tagBufferSize);

  //Phi interpolation in time
  void InterpolatePhiInTime(LevelData<FArrayBox> &a_Phi,
				       const LevelData<FArrayBox> &a_PhiNew,
				       const LevelData<FArrayBox> &a_PhiOld,
				       const DisjointBoxLayout & a_grids,
				       const Real& alpha);


protected:
  // Create a load-balanced DisjointBoxLayout from a collection of Boxes
  DisjointBoxLayout loadBalance(const Vector<Box>& a_grids);

  // Setup menagerie of data structures
  void levelSetup();

  // Get the next coarser level
  AMRLevelIdealMHD* getCoarserLevel() const;

  // Get the next finer level
  AMRLevelIdealMHD* getFinerLevel() const;

  // Conserved state, U, at old and new time
  LevelData<FArrayBox> m_UOld,m_UNew;

  // Phi for projection at old and new time
  LevelData<FArrayBox> m_phiOld,m_phiNew;
  // rhs=(divB) for projection
  LevelData<FArrayBox> m_rhs;
  //Hyperbolic Fluxes
  LevelData<FArrayBox> m_flux[CH_SPACEDIM];
  //  LevelData<FArrayBox> m_xflux;
  //  LevelData<FArrayBox> m_yflux;

  // CFL number
  Real m_cfl;

  // Grid spacing
  Real m_dx;

  // Interpolation from fine to coarse level
  FineInterp m_fineInterp;
  // Interpolation from fine to coarse level of Phi
  FineInterp m_fineInterpPhi;

  // Averaging from coarse to fine level
  CoarseAverage m_coarseAverage;

  // Averaging from coarse to fine level for Phi
  CoarseAverage m_coarseAveragePhi;

  // New time step
  Real m_dtNew;

  // Number of converved states
  int m_numStates;

  // Number of fluxes
  int m_numFluxes;

  // Names of conserved states
  Vector<string> m_stateNames;

  // Names of plotting variables
  Vector<string> m_plotNames;

  // Number of ghost cells (in each direction)
  int m_numGhost;

  // Physical dimension of the longest side of the domain
  Real m_domainLength;

  // Level integrator
  LevelGodunov m_levelGodunov;

  // Level solver for elliptic problems
  LevelSolver m_levelSolver;

  //  PoissonOp m_levelop;

  // Flux register
  LevelFluxRegister m_fluxRegister;

  // Patch integrator (factory and object)
  PatchGodunov* m_patchGodunovFactory;
  PatchGodunov* m_patchGodunov;

  // Poisson operator (factory and object)
  PoissonOp* m_poissonOpFactory;
  PoissonOp* m_poissonOp;

  // Refinement threshold for gradient
  Real m_refineThresh;

  // Tag buffer size
  int m_tagBufferSize;

  // Flag coarser and finer levels
  bool m_hasCoarser;
  bool m_hasFiner;

  DisjointBoxLayout m_grids;

private:
  // Disallowed for all the usual reasons
  void operator=(const AMRLevelIdealMHD& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  AMRLevelIdealMHD(const AMRLevelIdealMHD& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
