/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LoHiSide.H"

#include "BenchmarkIBC.H"
#include "BenchmarkF_F.H"

#include "LoHiCenter.H"

// Null constructor
BenchmarkIBC::BenchmarkIBC()
{
    m_isFortranCommonSet = false;
}

// Constructor which defines parameters used by Fortran routines
BenchmarkIBC::BenchmarkIBC(const Real&     a_gamma,
                           const Real&     a_energy,
                           const Real&     a_rho0,
                           const Real&     a_p0,
                           const RealVect& a_center,
                           const Real&     a_size,
                           const RealVect& a_velocity,
                           const Real&     a_artvisc)
{
    setFortranCommon(a_gamma,
                     a_energy,
                     a_rho0,
                     a_p0,
                     a_center,
                     a_size,
                     a_velocity,
                     a_artvisc);
}

// Sets parameters in a common block used by Fortran routines:
//   a_gamma    - Gamma for polytropic, gamma-law gas
//   a_ms       - Mach shock number of the discontinuity
//   a_center   - Center of the explosion
//   a_size     - Initial radius of the explosion
//   a_velocity - Initial velocity of the gas
//   a_artvisc  - Artificial viscosity coefficient
void BenchmarkIBC::setFortranCommon(const Real&     a_gamma,
                                    const Real&     a_energy,
                                    const Real&     a_rho0,
                                    const Real&     a_p0,
                                    const RealVect& a_center,
                                    const Real&     a_size,
                                    const RealVect& a_velocity,
                                    const Real&     a_artvisc)
{
    assert(m_isFortranCommonSet == false);

    FORT_SETBENCHMARK(CHF_CONST_REAL(a_gamma),
                      CHF_CONST_REAL(a_energy),
                      CHF_CONST_REAL(a_rho0),
                      CHF_CONST_REAL(a_p0),
                      CHF_CONST_REALVECT(a_center),
                      CHF_CONST_REAL(a_size),
                      CHF_CONST_REALVECT(a_velocity),
                      CHF_CONST_REAL(a_artvisc));

    m_isFortranCommonSet = true;
}

// Set the flag m_isFortranCommonSet to true so that new IBCs made with
// new_physIBC() will have this flag set without calling setFortranCommon()
// (this is a clumsy design and should be improved).
void BenchmarkIBC::setFortranCommonSet()
{
  m_isFortranCommonSet = true;
}

// Factory method - this object is its own factory:
//   Return a pointer to a new PhysIBC object with m_isDefined = false (i.e.,
//   its define() must be called before it is used) and m_isFortranCommonSet
//   set to value of m_isFortranCommonset in the current (factory) object.
PhysIBC* BenchmarkIBC::new_physIBC()
{
  BenchmarkIBC* retval = new BenchmarkIBC();

  if (m_isFortranCommonSet == true)
  {
    retval->setFortranCommonSet();
  }

  return static_cast<PhysIBC*>(retval);
}

// Set boundary fluxes
void BenchmarkIBC::fluxBC(FArrayBox&            a_F,
                          const FArrayBox&      a_W,
                          const FArrayBox&      a_Wextrap,
                          const int&            a_dir,
                          const Side::LoHiSide& a_side,
                          const Real&           a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  // In periodic case, this doesn't do anything
  if (!m_domain.isPeriodic(a_dir))
  {
    int sign;
    Box FBox = a_F.box();
    Box tmp = FBox;

    // Determine which side and thus shifting directions
    if (a_side == Side::Lo)
    {
      sign = -1;
    }
    else
    {
      sign = 1;
    }

    tmp.shiftHalf(a_dir,sign);

    // Is there a domain boundary next to this grid
    if (!m_domain.contains(tmp))
    {
      tmp &= m_domain;

      Box boundaryBox;

      // Find the strip of cells next to the domain boundary
      if (a_side == Side::Lo)
      {
        boundaryBox = bdryLo(tmp,a_dir);
      }
      else
      {
        boundaryBox = bdryHi(tmp,a_dir);
      }

      // Shift things to all line up correctly
      boundaryBox.shiftHalf(a_dir,-sign);
      a_F.shiftHalf(a_dir,-sign);

      // Set the boundary fluxes
      FORT_SOLIDBC(CHF_FRA(a_F),
                   CHF_CONST_FRA(a_Wextrap),
                   CHF_CONST_INT(sign),
                   CHF_CONST_REAL(m_dx),
                   CHF_CONST_INT(a_dir),
                   CHF_BOX(boundaryBox));

      // Shift returned fluxes to be face centered
      a_F.shiftHalf(a_dir,sign);
    }
  }
}

// Set boundary slopes:
//   The boundary slopes in a_dW are already set to one sided difference
//   approximations.  If this function doesn't change them they will be
//   used for the slopes at the boundaries.
void BenchmarkIBC::setBdrySlopes(FArrayBox&       a_dW,
                                 const FArrayBox& a_W,
                                 const int&       a_dir,
                                 const Real&      a_time)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  // In periodic case, this doesn't do anything
  if (!m_domain.isPeriodic(a_dir))
    {
      Box loBox,hiBox,centerBox,domain;
      int hasLo,hasHi;
      Box slopeBox = a_dW.box()&m_domain;

      // Generate the domain boundary boxes, loBox and hiBox, if there are
      // domain boundarys there
      loHiCenter(loBox,hasLo,hiBox,hasHi,centerBox,domain,
                 slopeBox,m_domain,a_dir);

      // Set the boundary slopes if necessary
      if ((hasLo != 0) || (hasHi != 0))
        {
          FORT_SLOPEBCS(CHF_FRA(a_dW),
                        CHF_CONST_FRA(a_W),
                        CHF_CONST_REAL(m_dx),
                        CHF_CONST_INT(a_dir),
                        CHF_BOX(loBox),
                        CHF_CONST_INT(hasLo),
                        CHF_BOX(hiBox),
                        CHF_CONST_INT(hasHi));
        }
    }
}

// Set up initial conditions
void BenchmarkIBC::initialize(LevelData<FArrayBox>& a_U)
{
  assert(m_isFortranCommonSet == true);
  assert(m_isDefined == true);

  DataIterator dit = a_U.boxLayout().dataIterator();
  
  Real sumScaledEnergy = 0.0;
  // Iterator of all grids in this level
  for (dit.begin(); dit.ok(); ++dit)
  {
    // Storage for current grid
    FArrayBox& U = a_U[dit()];

    // Box of current grid
    Box uBox = U.box();
    uBox &= m_domain;
    
    // Set up initial condition in this grid
    FORT_BENCHMARKINIT(CHF_FRA(U),
                       CHF_REAL(sumScaledEnergy),
                       CHF_CONST_REAL(m_dx),
                       CHF_BOX(uBox));
  }

  
#ifdef MPI
  // do broadcast/gather thing here
  Real recv;
  int result = MPI_Allreduce(&sumScaledEnergy, &recv, 1, MPI_CH_REAL,
                             MPI_SUM, Chombo_MPI::comm);
  if (result != MPI_SUCCESS)
    {
      MayDay::Error("Sorry, communication error in BenchmarkIBC::initialize");
    }

  sumScaledEnergy = recv;
#endif


  // now may need to do rescaling to ensure that the sum of the 
  // explosion energy is correct; sumScaledEnergy is the ratio of the
  // actual explosion energy to what we wanted.
  // don't do this if no cells were within the explosion
  if (sumScaledEnergy != 0.0)
    {
      // compute correction factor
      Real scalingFactor = 1.0/sumScaledEnergy;
      sumScaledEnergy = 0.0;

      for (dit.begin(); dit.ok(); ++dit)
        {
          FArrayBox& U = a_U[dit()];
          Box uBox = U.box();
          uBox &= m_domain;

          FORT_BENCHMARKINITCORRECT(CHF_FRA(U),
                                    CHF_REAL(scalingFactor),
                                    CHF_REAL(sumScaledEnergy),
                                    CHF_CONST_REAL(m_dx),
                                    CHF_BOX(uBox));
        }
    }

}
