/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Mon, July 19, 1999

#include "GhostBC.H"
#include "GhostBCF_F.H"

BoxGhostBC::BoxGhostBC(int             a_dir,
                       Side::LoHiSide  a_sd,
                       const Interval& a_comps)
  : m_components(-1,-1)
{
  define(a_dir, a_sd, a_comps);
}

BoxGhostBC::BoxGhostBC(int            a_dir,
                       Side::LoHiSide a_sd)
  : m_components(-1,-1)
{
  Interval comps(0,0);
  define(a_dir,a_sd, comps);
}

void BoxGhostBC::define(int            a_dir,
                        Side::LoHiSide a_sd)
{
  Interval comps(0,0);
  define(a_dir, a_sd, comps);
}

void BoxGhostBC::define(int             a_dir,
                        Side::LoHiSide  a_sd,
                        const Interval& a_comps)
{
  m_direction = a_dir;
  m_side = a_sd;
  m_components = a_comps;
}

void BoxGhostBC::applyInhomogeneousBCs(FArrayBox& a_state,
                                       const Box& a_domain,
                                       Real       a_dx) const
{
  ProblemDomain probdomain(a_domain);
  applyInhomogeneousBCs(a_state, probdomain, a_dx);
}

void BoxGhostBC::applyInhomogeneousBCs(FArrayBox&           a_state,
                                       const ProblemDomain& a_domain,
                                       Real                 a_dx) const
{
  // apply ghost bc to regular data at box boundary
  const Box& bx= a_state.box();

  // Set bcBox to be the ghost cells of a_state's box that lie
  // outside a_domain, adjacent to the face associated with this object.
  Box bcBox;
  bool isbc;
  int idir = m_direction;

  // periodic BC's trump all other BCs -- do nothing in that case
  if (!a_domain.isPeriodic(m_direction))
  {
    const Box& domBox = a_domain.domainBox();
    if (m_side == Side::Lo)
      {
        isbc = (bx.smallEnd(idir) < domBox.smallEnd(idir));
        if (isbc)
          {
            int ichop = domBox.smallEnd(m_direction);
            bcBox = a_state.box();
            bcBox.chop(m_direction, ichop);
          }
      }
    else
    if (m_side == Side::Hi)
      {
        isbc = (bx.bigEnd(idir) > domBox.bigEnd(idir));
        if (isbc)
          {
            int ichop = domBox.bigEnd(m_direction)+1;
            Box chopBox = a_state.box();
            bcBox = chopBox.chop(m_direction,ichop);
          }
      }
    else
      {
        cerr << "DomainGhostBC::applyghostbc: bogus side" << endl;
        abort();
      }

    if (isbc)
      {
        int ncomps = a_state.nComp();
        FArrayBox neumfac(bcBox, ncomps);
        FArrayBox dircfac(bcBox, ncomps);
        FArrayBox inhmval(bcBox, ncomps);

        fillBCValues(neumfac, dircfac, inhmval, a_dx, domBox);

        applyBCs(bcBox, a_state, neumfac, dircfac, inhmval, a_dx);
      }
  } // end if not periodic
}

void BoxGhostBC::applyHomogeneousBCs(FArrayBox& a_state,
                                     const Box& a_domain,
                                     Real       a_dx) const
{
  ProblemDomain probdomain(a_domain);
  applyHomogeneousBCs(a_state, probdomain, a_dx);
}

void BoxGhostBC::applyHomogeneousBCs(FArrayBox&           a_state,
                                     const ProblemDomain& a_domain,
                                     Real                 a_dx) const
{
  // apply ghost bc to regular data at box boundary
  const Box& bx= a_state.box();

  // Set bcBox to be the ghost cells of a_state's box that lie
  // outside a_domain, adjacent to the face associated with this object.
  Box bcBox;
  bool isbc;
  int idir = m_direction;

  if (!a_domain.isPeriodic(m_direction))
    {
      const Box& domBox = a_domain.domainBox();
      if (m_side == Side::Lo)
        {
          isbc = (bx.smallEnd(idir) <domBox.smallEnd(idir));
          if (isbc)
            {
              int ichop = domBox.smallEnd(m_direction);
              bcBox = a_state.box();
              bcBox.chop(m_direction, ichop);
            }
        }
      else
      if (m_side == Side::Hi)
        {
          isbc = (bx.bigEnd(idir) >domBox.bigEnd(idir));
          if (isbc)
            {
              int ichop = domBox.bigEnd(m_direction)+1;
              Box chopBox = a_state.box();
              bcBox = chopBox.chop(m_direction,ichop);
            }
        }
      else
        {
          cerr << "DomainGhostBC::applyghostbc: bogus side" << endl;
          abort();
        }

      if (isbc)
        {
          int ncomps = a_state.nComp();
          FArrayBox neumfac(bcBox, ncomps);
          FArrayBox dircfac(bcBox, ncomps);
          FArrayBox inhmval(bcBox, ncomps);

          fillBCValues(neumfac, dircfac, inhmval, a_dx, domBox);
          inhmval.setVal(0.0);

          applyBCs(bcBox, a_state, neumfac, dircfac, inhmval, a_dx);
        }
    }
}

void BoxGhostBC::applyBCs(const Box&       a_bcBox,
                          FArrayBox&       a_state,
                          const FArrayBox& a_neumfac,
                          const FArrayBox& a_dircfac,
                          const FArrayBox& a_inhmval,
                          Real             a_dx) const
{
  int iSide = sign(m_side);
  int iDir = m_direction;

  assert((m_side == Side::Hi) || (m_side == Side::Lo));

  Box biggerbox = a_bcBox;

  // state has to contain one cell on side of ghost cell
  if (m_side == Side::Lo)
    {
      biggerbox.growHi(iDir, 1);
    }
  else
    {
      biggerbox.growLo(iDir, 1);
    }

  // in order to deal with nghost > 1, need some measure of
  // where interior domain is. for now, just look at cells
  // adjacent to bcBox
  Box interiorBox;
  if (m_side == Side::Lo)
    {
      interiorBox = adjCellHi(a_bcBox,iDir,2);
    }
  else
    {
      interiorBox = adjCellLo(a_bcBox,iDir,2);
    }

  assert(a_state.box().contains(biggerbox));
  assert(a_dircfac.box().contains(a_bcBox));
  assert(a_neumfac.box().contains(a_bcBox));
  assert(a_inhmval.box().contains(a_bcBox));
  assert(m_components.begin() >= 0);
  assert(m_components.end() < a_state.nComp());

  int compbegin = m_components.begin();
  int compend   = m_components.end();

  FORT_BOXGHOSTBC(CHF_FRA(a_state),
                  CHF_CONST_FRA(a_neumfac),
                  CHF_CONST_FRA(a_dircfac),
                  CHF_CONST_FRA(a_inhmval),
                  CHF_BOX(a_bcBox),
                  CHF_BOX(interiorBox),
                  CHF_CONST_INT(iDir),
                  CHF_CONST_INT(iSide),
                  CHF_CONST_REAL(a_dx),
                  CHF_CONST_INT(compbegin),
                  CHF_CONST_INT(compend));
}


DomainGhostBC::DomainGhostBC()
{
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loGhostBC[idir] = NULL;
      m_hiGhostBC[idir] = NULL;
    }
}

DomainGhostBC::~DomainGhostBC()
{
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      if (m_loGhostBC[idir] != NULL)
        {
          delete m_loGhostBC[idir];
        }

      m_loGhostBC[idir] = NULL;

      if (m_hiGhostBC[idir] != NULL)
        {
          delete m_hiGhostBC[idir];
        }

      m_hiGhostBC[idir] = NULL;
    }
}

void
DomainGhostBC::setBoxGhostBC(const BoxGhostBC& a_ghostBC)
{
  int direction = a_ghostBC.m_direction;
  Interval components = a_ghostBC.m_components;
  Side::LoHiSide side = a_ghostBC.m_side;

  assert ((side == Side::Lo) || (side == Side::Hi));

  switch (side)
    {
      case Side::Lo:
        {
          if (m_loGhostBC[direction] != NULL)
            {
              delete m_loGhostBC[direction];
            }

          m_loGhostBC[direction] = a_ghostBC.new_boxghostbc();
          m_loGhostBC[direction]->define(direction,side, components);

          break;
        }

      case Side::Hi:
        {
          if (m_hiGhostBC[direction] != NULL)
            {
              delete m_hiGhostBC[direction];
            }

          m_hiGhostBC[direction] = a_ghostBC.new_boxghostbc();
          m_hiGhostBC[direction]->define(direction,side, components);

          break;
        }

      default:
        {
          MayDay::Error("DomainGhostBC::setboxghostbc: bogus side");
        }
    }
}

const BoxGhostBC& DomainGhostBC::operator()(int            a_direction,
                                            Side::LoHiSide a_side) const
{
  assert ((a_side == Side::Lo) || (a_side == Side::Hi));

  BoxGhostBC* retPtr = NULL;

  if (a_side == Side::Lo)
    {
      assert(m_loGhostBC[a_direction] != NULL);

      retPtr = m_loGhostBC[a_direction];
    }
  else
  if (a_side == Side::Hi)
    {
      assert(m_hiGhostBC[a_direction] != NULL);

      retPtr = m_hiGhostBC[a_direction];
    }
  else
    {
      MayDay::Error("DomainGhostBC::operator(): bogus side");
    }

  return *retPtr;
}

void DomainGhostBC::applyHomogeneousBCs(FArrayBox& a_state,
                                        const Box& a_domain,
                                        Real       a_dx) const
{
  ProblemDomain probdomain(a_domain);

  applyHomogeneousBCs(a_state, probdomain, a_dx);
}

void DomainGhostBC::applyHomogeneousBCs(FArrayBox&           a_state,
                                        const ProblemDomain& a_domain,
                                        Real                 a_dx) const
{
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loGhostBC[idir]->applyHomogeneousBCs(a_state, a_domain, a_dx);
      m_hiGhostBC[idir]->applyHomogeneousBCs(a_state, a_domain, a_dx);
    }
}

void DomainGhostBC::applyInhomogeneousBCs(FArrayBox& a_state,
                                          const Box& a_domain,
                                          Real       a_dx) const
{
  ProblemDomain probdomain(a_domain);

  applyInhomogeneousBCs(a_state, probdomain,a_dx);
}

void DomainGhostBC::applyInhomogeneousBCs(FArrayBox&           a_state,
                                          const ProblemDomain& a_domain,
                                          Real                 a_dx) const
{
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loGhostBC[idir]->applyInhomogeneousBCs(a_state, a_domain, a_dx);
      m_hiGhostBC[idir]->applyInhomogeneousBCs(a_state, a_domain, a_dx);
    }
}

DomainGhostBC::DomainGhostBC(const DomainGhostBC& a_dgbcin)
{

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      m_loGhostBC[idir] = NULL;
      m_hiGhostBC[idir] = NULL;
    }

  *this = a_dgbcin;
}

DomainGhostBC& DomainGhostBC::operator=(const DomainGhostBC& a_dgbcin)
{
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      SideIterator sit;
      for (sit.begin(); sit.ok(); sit.next())
        {
          if (a_dgbcin.isBCDefined(idir,sit()))
            {
              setBoxGhostBC(a_dgbcin(idir,sit()));
            }
          else
            {
              // set pointer to NULL
              resetBoxGhostBC(idir,sit());
            }
        }
    }

  return *this;
}

bool DomainGhostBC::isBCDefined(const int            a_dir,
                                const Side::LoHiSide a_side) const
{
  bool isdefined;

  assert ((a_side == Side::Lo) || (a_side == Side::Hi));

  if (a_side == Side::Lo)
    {
      isdefined = (m_loGhostBC[a_dir] != NULL);
    }
  else
  if (a_side == Side::Hi)
    {
      isdefined = (m_hiGhostBC[a_dir] != NULL);
    }
  else
    {
      MayDay::Error("DomainGhostBC::isDefined(): bogus side");
    }

  return isdefined;
}

void DomainGhostBC::resetBoxGhostBC(const int            a_dir,
                                    const Side::LoHiSide a_side)
{
  assert ((a_side == Side::Lo) || (a_side == Side::Hi));

  if (a_side == Side::Lo)
    {
      if (m_loGhostBC[a_dir] != NULL)
        {
          delete m_loGhostBC[a_dir];
        }

      m_loGhostBC[a_dir] = NULL;
    }
  else
  if (a_side == Side::Hi)
    {
      if (m_hiGhostBC[a_dir] != NULL)
        {
          delete m_hiGhostBC[a_dir];
        }

      m_hiGhostBC[a_dir] = NULL;
    }
  else
    {
      MayDay::Error("DomainGhostBC::resetBoxGhostBC(): bogus side");
    }
}
