/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "BaseFab.H"
#include "REAL.H"
#include "DataIterator.H"
#include "LevelMGF_F.H"

#include "MGInterp.H"

MGInterp::MGInterp()
  :m_isDefined(false)
{
}

MGInterp::~MGInterp()
{
}

MGInterp::MGInterp(const DisjointBoxLayout& a_fineDomain,
                   int                      a_numcomps,
                   int                      a_refRatio,
                   const Box&               a_problemDomain)
  :m_isDefined(false)
{
  ProblemDomain finePhysdomain(a_problemDomain);

  define(a_fineDomain, a_numcomps, a_refRatio, finePhysdomain);
}

MGInterp::MGInterp(const DisjointBoxLayout& a_fineDomain,
                   int                      a_numcomps,
                   int                      a_refRatio,
                   const ProblemDomain&     a_problemDomain)
  :m_isDefined(false)
{
  define(a_fineDomain, a_numcomps, a_refRatio, a_problemDomain);
}

void MGInterp::define(const DisjointBoxLayout& a_fineDomain,
                      int                      a_numcomps,
                      int                      a_refRatio,
                      const Box&               a_problemDomain)
{
  ProblemDomain finePhysdomain(a_problemDomain);

  define(a_fineDomain, a_numcomps, a_refRatio, finePhysdomain);
}

void MGInterp::define(const DisjointBoxLayout& a_fineDomain,
                      int                      a_numcomps,
                      int                      a_refRatio,
                      const ProblemDomain&     a_problemDomain)
{
  m_refRatio = a_refRatio;
  m_problemDomain = a_problemDomain;
  m_grids = a_fineDomain;

  // create the work array 
  DisjointBoxLayout coarsenedFineDomain;

  coarsen(coarsenedFineDomain, a_fineDomain, m_refRatio);

  m_coarsenedFineData.define(coarsenedFineDomain, 
                             a_numcomps, 
                             IntVect::TheZeroVector());

  m_isDefined = true;
}

// interpolate from coarse level to fine level
void MGInterp::interpToFine(LevelData<FArrayBox>&       a_fineData,
                            const LevelData<FArrayBox>& a_coarseData)
{
  assert(m_isDefined);
  a_coarseData.copyTo(a_coarseData.interval(),
                      m_coarsenedFineData,
                      m_coarsenedFineData.interval());

  DataIterator dit = m_grids.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      FArrayBox& fine = a_fineData[dit()];
      const FArrayBox& coarsenedFine = m_coarsenedFineData[dit()];

      Box crseBox = coarsenedFine.box();
      Box fineBox = m_grids.get(dit());

      fineBox.coarsen(m_refRatio);

      assert(fineBox == crseBox);

      Box nrefbox(IntVect::TheZeroVector(), 
                  (m_refRatio-1)*IntVect::TheUnitVector());

      FORT_INTERPMG(CHF_FRA(fine),
                    CHF_FRA(coarsenedFine),
                    CHF_BOX(crseBox),
                    CHF_CONST_INT(m_refRatio),
                    CHF_BOX(nrefbox));
    }
}

bool MGInterp::isDefined() const
{
  return m_isDefined;
}
