/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LevelFluxRegister.H"
#include "LayoutIterator.H"
#include "DebugOut.H"
#include "Copier.H"
#include "parstream.H"



LevelFluxRegister::LevelFluxRegister(const DisjointBoxLayout& a_dblFine,
                                     const DisjointBoxLayout& a_dblCoar,
                                     const Box& a_dProblem,
                                     int a_nRefine,
                                     int a_nComp)
{

  ProblemDomain physDomain(a_dProblem);
  define(a_dblFine, a_dblCoar, physDomain, a_nRefine, a_nComp);
}


LevelFluxRegister::LevelFluxRegister(const DisjointBoxLayout& a_dblFine,
                                     const DisjointBoxLayout& a_dblCoar,
                                     const ProblemDomain& a_dProblem,
                                     int a_nRefine,
                                     int a_nComp)
{

  define(a_dblFine, a_dblCoar, a_dProblem, a_nRefine, a_nComp);
}

void 
LevelFluxRegister::define(const DisjointBoxLayout& a_dbl,
                          const DisjointBoxLayout& a_dblCoarse,
                          const Box& a_dProblem,
                          int a_nRefine,
                          int a_nComp)
{
  ProblemDomain physDomain(a_dProblem);
  define(a_dbl, a_dblCoarse, physDomain, a_nRefine, a_nComp);
}



void 
LevelFluxRegister::define(const DisjointBoxLayout& a_dbl,
                          const DisjointBoxLayout& a_dblCoarse,
                          const ProblemDomain& a_dProblem,
                          int a_nRefine,
                          int a_nComp)
{
  m_isDefined = true;
  m_nRefine   = a_nRefine;
  m_coarFlux.define( a_dblCoarse, a_nComp);
  m_domain = a_dProblem;
  DisjointBoxLayout coarsenedFine;
  coarsen(coarsenedFine, a_dbl, a_nRefine);

  m_fineFlux.define( coarsenedFine, a_nComp, IntVect::TheUnitVector());

  m_reverseCopier.ghostDefine(coarsenedFine, a_dblCoarse, 
                              a_dProblem, IntVect::TheUnitVector());

  Box domainBox = a_dProblem.domainBox();
  for(int i=0; i<CH_SPACEDIM; i++)
    {
      m_coarseLocations[i].define(a_dblCoarse);
      m_coarseLocations[i+CH_SPACEDIM].define(a_dblCoarse);
    }
  // OK, now the hard part of the code.  building the IntVectSet for
  // the different directions of the coarse grid.  Fortunately, this
  // part can be done in parallel now, as it is just the local
  // IntVectSets that we need to worry about.

  DataIterator dC   = a_dblCoarse.dataIterator();
  LayoutIterator dF = coarsenedFine.layoutIterator();

  for(dC.begin(); dC.ok(); ++dC)
    {
      const Box& cBox = a_dblCoarse.get(dC);

      for(dF.begin(); dF.ok(); ++dF)
        {
          const Box& fBox  = coarsenedFine.get(dF);

          if(fBox.bigEnd(0)+1 < cBox.smallEnd(0))
            {
              //can skip this box since they cannot intersect, due to sorting
            }
          else if (fBox.smallEnd(0)-1 > cBox.bigEnd(0))
            {
              //skip to end, since all the rest of boxes will not intersect either
              dF.end();
            }
          else
            {
  
              for(int i=0; i<CH_SPACEDIM; i++)
                {
                  Vector<Box>& lo = m_coarseLocations[i][dC];
                  Vector<Box>& hi = m_coarseLocations[i+CH_SPACEDIM][dC];
                  
                  Box loBox = adjCellLo(fBox, i, 1);
                  Box hiBox = adjCellHi(fBox, i, 1);
                  if(cBox.intersectsNotEmpty(loBox)) lo.push_back(loBox & cBox);
                  if(cBox.intersectsNotEmpty(hiBox)) hi.push_back(hiBox & cBox);
                }
            }
        }
    }


  if(a_dProblem.isPeriodic())
    {
      Vector<Box> periodicBoxes[2*CH_SPACEDIM];
      for(dF.begin(); dF.ok(); ++dF)
        {
          const Box& fBox  = coarsenedFine.get(dF);
          for(int i=0; i<CH_SPACEDIM; i++)
            {
              if(a_dProblem.isPeriodic(i))
                {
                  if(fBox.smallEnd(i) == domainBox.smallEnd(i))
                    periodicBoxes[i].push_back(adjCellLo(fBox, i, 1));
                  if(fBox.bigEnd(i) == domainBox.bigEnd(i))
                    periodicBoxes[i+CH_SPACEDIM].push_back(adjCellHi(fBox, i, 1));
                }
            }
        }
      for(int i=0; i<CH_SPACEDIM; i++)
        {
          Vector<Box>& loV = periodicBoxes[i];
          Vector<Box>& hiV = periodicBoxes[i+CH_SPACEDIM];
          int size = domainBox.size(i);
          for(int j=0; j<loV.size(); j++) loV[j].shift(i, size);
          for(int j=0; j<hiV.size(); j++) hiV[j].shift(i, -size);
        }
      for(dC.begin(); dC.ok(); ++dC)
        {
          const Box& cBox = a_dblCoarse.get(dC);
          for(int i=0; i<CH_SPACEDIM; i++)
            if(a_dProblem.isPeriodic(i))
              {
                Vector<Box>& loV = periodicBoxes[i];
                Vector<Box>& hiV = periodicBoxes[i+CH_SPACEDIM];
                /*
                if(cBox.smallEnd(i) == domainBox.smallEnd(i) )
                  {
                    IntVectSet& lo = m_coarseLocations[i][dC];
                    for(int j=0; j<hiV.size(); j++) lo |= hiV[j];
                  }
                if(cBox.bigEnd(i) == domainBox.bigEnd(i) )
                  {
                    IntVectSet& hi = m_coarseLocations[i+CH_SPACEDIM][dC];
                    for(int j=0; j<loV.size(); j++) hi |= loV[j];
                  }
                */
                if(cBox.smallEnd(i) == domainBox.smallEnd(i) )
                  {
                    Vector<Box>& lo = m_coarseLocations[i][dC];
                    for(int j=0; j<hiV.size(); j++)
                      {
                        if(cBox.intersectsNotEmpty(hiV[j])) lo.push_back(cBox & hiV[j]);
                      }
                  }
                if(cBox.bigEnd(i) == domainBox.bigEnd(i) )
                  {
                    Vector<Box>& hi = m_coarseLocations[i+CH_SPACEDIM][dC];
                    for(int j=0; j<loV.size(); j++)
                      {
                        if(cBox.intersectsNotEmpty(loV[j])) hi.push_back(cBox & loV[j]);
                      }
                  }
              }
        }
    }
            
}


LevelFluxRegister::LevelFluxRegister()
{
  m_isDefined = false;
}

LevelFluxRegister::~LevelFluxRegister()
{
}

bool 
LevelFluxRegister::isDefined() const
{
  return m_isDefined;
}

void
LevelFluxRegister::undefine()
{
  m_isDefined = false;
}

void 
LevelFluxRegister::setToZero()
{
  DataIterator d =  m_coarFlux.dataIterator();
  for(d.begin(); d.ok(); ++d)
    {
      m_coarFlux[d].setVal(0.0);
    }
  d =   m_fineFlux.dataIterator();
  for(d.begin(); d.ok(); ++d)
    {
      m_fineFlux[d].setVal(0.0);
    }

} 

void 
LevelFluxRegister::incrementCoarse(FArrayBox& a_coarseFlux,
                                   Real a_scale,
                                   const DataIndex& a_coarseDataIndex,
                                   const Interval& a_srcInterval,
                                   const Interval& a_dstInterval,
                                   int a_dir)

{
  assert(isDefined());
  incrementCoarse(a_coarseFlux, a_scale, a_coarseDataIndex, a_srcInterval,
                  a_dstInterval, a_dir, Side::Lo);
  incrementCoarse(a_coarseFlux, a_scale, a_coarseDataIndex, a_srcInterval,
                  a_dstInterval, a_dir, Side::Hi);

}

void 
LevelFluxRegister::incrementCoarse(FArrayBox& a_coarseFlux,
                                   Real a_scale,
                                   const DataIndex& a_coarseDataIndex,
                                   const Interval& a_srcInterval,
                                   const Interval& a_dstInterval,
                                   int a_dir,
                                   Side::LoHiSide a_sd)
{
  assert(isDefined());
  const Vector<Box>& intersect =  
    m_coarseLocations[a_dir+a_sd*CH_SPACEDIM][a_coarseDataIndex];

  FArrayBox& coarse = m_coarFlux[a_coarseDataIndex];

  a_coarseFlux.shiftHalf(a_dir, sign(a_sd));
  Real scale = -sign(a_sd)*a_scale;
  int s = a_srcInterval.begin();
  int d = a_dstInterval.begin();
  int size = a_srcInterval.size();

  for(int b=0; b<intersect.size(); ++b)
    {
      const Box& box = intersect[b];
      coarse.plus(a_coarseFlux, box, box, scale, s, d, size);
    }


  a_coarseFlux.shiftHalf(a_dir, - sign(a_sd));
}

void 
LevelFluxRegister::incrementFine(FArrayBox& a_fineFlux,
                                 Real a_scale,
                                 const DataIndex& a_fineDataIndex,
                                 const Interval& a_srcInterval,
                                 const Interval& a_dstInterval,
                                 int a_dir,
                                 Side::LoHiSide a_sd)
{
  assert(isDefined());

  a_fineFlux.shiftHalf(a_dir, sign(a_sd));
  Real denom = D_TERM(1, *m_nRefine, *m_nRefine);

  Real scale = sign(a_sd)*a_scale/denom;
  FArrayBox& cFine = m_fineFlux[a_fineDataIndex];
  Box clipBox = m_fineFlux.box(a_fineDataIndex);
  clipBox.refine(m_nRefine);
  clipBox.grow(a_dir, 1);
  Box fineBox;
  if(a_sd == Side::Lo)
    {
      fineBox = adjCellLo(a_fineFlux.box(), a_dir, 1);
      fineBox.shift(a_dir, 1);
      fineBox&=clipBox;
    }
  else
    {
      fineBox = adjCellHi(a_fineFlux.box(), a_dir, 1);
      fineBox.shift(a_dir, -1);
      fineBox&=clipBox;
    }


  for(BoxIterator b(fineBox); b.ok(); ++b)
    {
      int s = a_srcInterval.begin();
      int d = a_dstInterval.begin();
      for(;s<=a_srcInterval.end(); ++s, ++d)
        {
          cFine(coarsen(b(), m_nRefine), d) += scale*a_fineFlux(b(), s);
        }
    }
 
 

    
  a_fineFlux.shiftHalf(a_dir, - sign(a_sd));
 
}

void 
LevelFluxRegister::reflux(LevelData<FArrayBox>& a_uCoarse,
                          Real a_scale)
{
  Interval interval(0, a_uCoarse.nComp()-1);
  reflux(a_uCoarse, interval, interval, a_scale);
}


class AddOp : public LDOperator<FArrayBox>
{
public:
  Real scale;
  AddOp():scale(1.0){;}

  virtual void linearIn(FArrayBox& arg,  void* buf, const Box& R, 
                        const Interval& comps) const
  {
    Real* buffer = (Real*)buf;
    if(scale != 1.0){
      ForAllXBNN(Real, arg, R, comps.begin(), comps.size())
        {
          argR+=*buffer * scale;
          buffer++;
        } EndFor
            }
    else
      {
        ForAllXBNN(Real, arg, R, comps.begin(), comps.size())
          {
            argR+=*buffer;
            buffer++;
          } EndFor
              }
  }
  void op(FArrayBox& dest, 
          const Box& RegionFrom, 
          const Interval& Cdest, 
          const Box& RegionTo,
          const FArrayBox& src, 
          const Interval& Csrc) const
  {
    if(scale != 1.0)
      dest.plus(src, RegionFrom, RegionTo, scale, Csrc.begin(), Cdest.begin(), Cdest.size());
    else
      dest.plus(src, RegionFrom, RegionTo, Csrc.begin(), Cdest.begin(), Cdest.size());
  }
};

void LevelFluxRegister::reflux( 
                               LevelData<FArrayBox>& a_uCoarse,
                               const Interval&       a_coarse_interval,
                               const Interval&       a_flux_interval,
                               Real                  a_scale )
{
  for(DataIterator dit(a_uCoarse.dataIterator()); dit.ok(); ++dit)
    {
      FArrayBox& u = a_uCoarse[dit];
      FArrayBox& coarse = m_coarFlux[dit];
      u.plus(coarse, -a_scale);
    }
  AddOp op;
  op.scale = -a_scale;
  m_fineFlux.copyTo(a_flux_interval, a_uCoarse, a_coarse_interval, m_reverseCopier, op);
  
}

