/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "LevelFluxRegister.H"
#include "LayoutIterator.H"
#include "DebugOut.H"
#include "Copier.H"
#include "parstream.H"

void
LevelFluxRegister::setDefaultValues()
{
  m_isDefined = false;
  m_nComp = -1;
  m_nRefine = -1;
}

int LevelFluxRegister::index(int dir, Side::LoHiSide side)
{
  return side*SpaceDim+dir;
}

LevelFluxRegister::LevelFluxRegister(const DisjointBoxLayout& a_dblFine,
                                     const DisjointBoxLayout& a_dblCoar,
                                     const Box& a_dProblem,
                                     int a_nRefine,
                                     int a_nComp)
{
  setDefaultValues();
  ProblemDomain physDomain(a_dProblem);
  define(a_dblFine, a_dblCoar, physDomain, a_nRefine, a_nComp);
}


LevelFluxRegister::LevelFluxRegister(const DisjointBoxLayout& a_dblFine,
                                     const DisjointBoxLayout& a_dblCoar,
                                     const ProblemDomain& a_dProblem,
                                     int a_nRefine,
                                     int a_nComp)
{
  setDefaultValues();
  define(a_dblFine, a_dblCoar, a_dProblem, a_nRefine, a_nComp);
}

void 
LevelFluxRegister::define(const DisjointBoxLayout& a_dbl,
                          const DisjointBoxLayout& a_dblCoarse,
                          const Box& a_dProblem,
                          int a_nRefine,
                          int a_nComp)
{
  setDefaultValues();
  ProblemDomain physDomain(a_dProblem);
  define(a_dbl, a_dblCoarse, physDomain, a_nRefine, a_nComp);
}

// This define function started showing up in timing profiles when
// running larger problems and a couple of critical loops were
// uncovered (labeled as CRITICAL LOOP below in code).  A temporary
// performance improvement was made in the first one (and most
// time-consuming one) by computing as much as possible outside of the
// loop.  However, we may want to consider rewriting entire
// function. (ndk)

void 
LevelFluxRegister::define(const DisjointBoxLayout& a_dbl,
                          const DisjointBoxLayout& a_dblCoarse,
                          const ProblemDomain& a_dProblem,
                          int a_nRefine,
                          int a_nComp)
{
  m_isDefined = true;
  m_gridsFine = a_dbl;
  m_gridsCoar = a_dblCoarse;
  assert(a_nRefine > 0);
  assert(a_nComp > 0);
  assert(!a_dProblem.isEmpty());
  assert (a_dbl.checkPeriodic(a_dProblem));
  m_nComp = a_nComp;
  m_nRefine = a_nRefine;
  m_domainCoarse = coarsen(a_dProblem, a_nRefine);
  assert (a_dblCoarse.checkPeriodic(m_domainCoarse));
  
  // allocate copiers
  crseCopiers.resize(SpaceDim*2);

  // create periodic test box which will contain all cells which
  // are _not_ adjacent to a periodic boundary.
  Box periodicTestBox(m_domainCoarse.domainBox());
  if (m_domainCoarse.isPeriodic())
    {
      for (int idir=0; idir<SpaceDim; idir++) 
        {
          if (m_domainCoarse.isPeriodic(idir))
            periodicTestBox.grow(idir,-1);
        }
    }
  
  
  SideIterator side;
  
  for(int idir=0 ; idir<SpaceDim; ++idir)
    {
      for(side.begin(); side.ok(); ++side)
        {
          // step one, build fineBoxes, flux register boxes
          // indexed by the fine level but in the coarse index
          // space
          DisjointBoxLayout fineBoxes, tmp;
          // first create coarsened dbl, then compute flux register boxes
          // adjacent to coarsened fine boxes            
          coarsen(tmp,a_dbl,a_nRefine);
          if (side() == Side::Lo) 
            {
              adjCellLo(fineBoxes,tmp,idir,1);
            }
          else 
            {
              adjCellHi(fineBoxes,tmp,idir,1);
            }
          // now define the FArrayBoxes of fabFine on this DisjointBoxLayout
          fabFine[index(idir, side())].define(fineBoxes, a_nComp);

          // now, for every fineBox, we need an IntVectSet that indicates
          // which parts are ACTUALLY coarse-fine interfaces, and which
          // parts are covered by another fine box.
          Vector<IntVectSet> coarseFineIVS(fineBoxes.size());
          LayoutIterator it = fineBoxes.layoutIterator();
          LayoutIterator it2 = fineBoxes.layoutIterator();

          // optimization here.          
          // Try to do as much of the inner-inner loop stuff out here... (ndk)
          Vector<Box> vectcoarsenedFineBoxsave;
          vector<bool> vectContainTest;
          IntVectSet constMinusIVS;
          bool hit;
          for(it.begin(); it.ok(); ++it) 
            {
              Box coarsenedFineBox(a_dbl[it()]);
              coarsenedFineBox.coarsen(a_nRefine);
              vectcoarsenedFineBoxsave.push_back(coarsenedFineBox);
              
              constMinusIVS&=coarsenedFineBox;
              
              hit = false;
              if (!periodicTestBox.contains(coarsenedFineBox)) 
                {
                  hit=true;
                }
              vectContainTest.push_back(hit);
            }  

          int i = 0;
          for(it2.begin(); it2.ok();++i, ++it2)
            {
              // optimization here -- move the following test out of the inner-inner loop
              bool containsTest2 = false;
              if (!periodicTestBox.contains(fineBoxes.get(it2()))) 
                {
                  containsTest2 = true;
                }

              IntVectSet& ivs = coarseFineIVS[i];
              // this needs to be done differently depending on 
              // whether we're looking at a box in the computational
              // domain or one which will become a periodic image
              const Box& compDomBox = m_domainCoarse.domainBox();
              if (compDomBox.contains(fineBoxes.get(it2())))
                {
                  ivs.define(fineBoxes.get(it2()));
                } 
              else
                {
                  // look for appropriate periodic image here
                  // in partially periodic case, only do this if
                  // box is contained in periodically extended domain
                  if (m_domainCoarse.contains(fineBoxes.get(it2())))
                    {
                      ShiftIterator shiftIt = m_domainCoarse.shiftIterator();
                      IntVect shiftMult(compDomBox.size());
                      Box regBox = fineBoxes.get(it2());
                      for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
                        {
                          IntVect shiftVect(shiftIt()*shiftMult);
                          regBox.shift(shiftVect);
                          if (compDomBox.contains(regBox)) 
                            {
                              // should only hit this once.
                              assert(ivs.isEmpty());
                              ivs.define(regBox);
                            }
                          regBox.shift(-shiftVect);
                        }
                    } // end if is periodic
                  else 
                    {
                      // revert to "normal" behavior
                      ivs.define(fineBoxes.get(it2()));
                    } // end catch if not periodic
                } // end if register is not contained in computational domain
              // sanity check
              assert (!ivs.isEmpty());

              // optimization.  use precomputed values.
              // This is what i'm referring to as the inner-inner loop --
              // this is executed an enormous amount of times -- especially as the
              // problem size grows.  Use the values we've already computed above.
              ivs -= constMinusIVS; 
              int iii=0;
              if (m_domainCoarse.isPeriodic() && containsTest2) 
                {  
                  for (vector<bool>::iterator vbi=vectContainTest.begin();
                       vbi != vectContainTest.end(); ++iii, ++vbi)       // CRITICAL LOOP
                    {
                      if (*vbi) {  // never true for Dan's INS example...
                        ShiftIterator shiftIt = m_domainCoarse.shiftIterator();
                        IntVect shiftMult(m_domainCoarse.domainBox().size());
                        for (shiftIt.begin(); shiftIt.ok(); ++shiftIt) 
                          {
                            IntVect shiftVect = shiftMult*shiftIt();
                            vectcoarsenedFineBoxsave[iii].shift(shiftVect);
                            ivs -= vectcoarsenedFineBoxsave[iii];
                            vectcoarsenedFineBoxsave[iii].shift(-shiftVect);
                          } // end iteration over periodic shifts
                      } // end if is periodic
                    }
                }
            }
          
          
          // These IntVectSets represent all the coarseFine locations, 
          // plus some locations hanging off the edge of the domain.
          // that's OK, since they will get discarded in the intersection
          // with the coarse domain step.
      
          LayoutData<Vector<IntVectSet> >& ivsetsV 
            = refluxLocations[index(idir, side())];
          ivsetsV.define(a_dblCoarse);
      
          LayoutData<Vector<DataIndex> >& mapsV = 
            coarToCoarMap[index(idir, side())];
          mapsV.define(a_dblCoarse);
      
          DisjointBoxLayout coarseBoxes;
          LayoutIterator dit = a_dblCoarse.layoutIterator();
          for(dit.begin(); dit.ok(); ++dit)
            {
              const Box& coarseBox = a_dblCoarse[dit()];
              i=0;
              int count = 0;
              for(it.begin(); it.ok(); ++it, ++i)
                {
                  // since we want the coarse cell that we will reflux
                  // into, need to take the potential of periodic
                  // images into account
                  Box regBox = fineBoxes.get(it());
                  if(regBox.intersectsNotEmpty(coarseBox))
                    {
                      regBox &= coarseBox;
                      ++count;
                      unsigned int proc = a_dblCoarse.procID(dit());
                      DataIndex index = coarseBoxes.addBox(regBox,proc);
                      if(proc == procID())
                        {
                          mapsV[DataIndex(dit())].push_back(index);
                          ivsetsV[DataIndex(dit())].push_back(coarseFineIVS[i]);
                          //need to modify this IntVectSet for just
                          // the part in this coarse box.
                          IntVectSet* reducedIVS = &(ivsetsV[DataIndex(dit())][count-1]);
                          (*reducedIVS) &= regBox;
                        }
                    } 
                  // as usual, only check for periodic images if both
                  // boxes adjoin a periodic boundary
                  else if (m_domainCoarse.isPeriodic()
                           && !periodicTestBox.contains(regBox)
                           && !periodicTestBox.contains(coarseBox))
                    // check for intersection with periodic images
                    // this is different, because we want to add the 
                    // shifted image (inside the domainBox) rather than
                    // the unshifted one
                    {
                      ShiftIterator shiftIt = m_domainCoarse.shiftIterator();
                      IntVect shiftMult(m_domainCoarse.domainBox().size());
                      for (shiftIt.begin(); shiftIt.ok(); ++shiftIt)     // CRITICAL LOOP
                        {
                          IntVect shiftVect = shiftMult*shiftIt();
                          regBox.shift(shiftVect);
                          if (regBox.intersectsNotEmpty(coarseBox))
                            {
                              regBox &= coarseBox;
                              ++count;
                              unsigned int proc = a_dblCoarse.procID(dit());
                              DataIndex index = coarseBoxes.addBox(regBox,proc);
                              if (proc == procID())
                                {
                                  mapsV[DataIndex(dit())].push_back(index);
                                  ivsetsV[DataIndex(dit())].push_back(coarseFineIVS[i]);
                                  // need to modify this IntVectSet for just the 
                                  // part in this coarse box
                                  IntVectSet* reducedIVS = 
                                    &(ivsetsV[DataIndex(dit())][count-1]);
                                  //reducedIVS->shift(shiftVect);
                                  (*reducedIVS) &= regBox;
                                  //reducedIVS->shift(-shiftVect);
                                } // end if procID's match
                            } // end if regBox intersects this coarse box
                          regBox.shift(-shiftVect);
                        } // end loop over periodic shifts
                    }
                }
            }
          coarseBoxes.close();
          fabCoarse[index(idir, side())].define(coarseBoxes, a_nComp);

      
          // last thing to do here -- define copiers
          crseCopiers[index(idir, side())].define(fineBoxes, coarseBoxes,
                                                  IntVect::Zero);

        }
    }
}


LevelFluxRegister::LevelFluxRegister()
{
  setDefaultValues();
}

LevelFluxRegister::~LevelFluxRegister()
{
}

bool 
LevelFluxRegister::isDefined() const
{
  return m_isDefined;
}

void
LevelFluxRegister::undefine()
{
  m_isDefined = false;
}

void 
LevelFluxRegister::setToZero()
{
  SideIterator side;
  for(int idir=0 ; idir<SpaceDim; ++idir)
    {
      for(side.begin(); side.ok(); ++side)
        {
          LevelData<FArrayBox>& fineReg = fabFine[index(idir, side())];
          LevelData<FArrayBox>& coarReg = fabCoarse[index(idir, side())];

          for(DataIterator dit = fineReg.dataIterator(); dit.ok(); ++dit)
            fineReg[dit()].setVal(0.0);

          for(DataIterator dit = coarReg.dataIterator(); dit.ok(); ++dit)
            coarReg[dit()].setVal(0.0);

        }
    }
} 

void 
LevelFluxRegister::incrementCoarse(FArrayBox& a_coarseFlux,
                                   Real a_scale,
                                   const DataIndex& a_coarseDataIndex,
                                   const Interval& a_srcInterval,
                                   const Interval& a_dstInterval,
                                   int a_dir)

{
  assert(isDefined());
  assert(!a_coarseFlux.box().isEmpty());
  assert(a_dir >= 0);
  assert(a_dir < SpaceDim);

  assert(a_srcInterval.size() == a_dstInterval.size());
  assert(a_srcInterval.begin() >= 0);
  assert(a_srcInterval.end() < a_coarseFlux.nComp());
  assert(a_dstInterval.begin() >= 0);
  assert(a_dstInterval.end() < m_nComp);

  SideIterator side;
  for(side.begin(); side.ok(); ++side)
    {
        incrementCoarse(a_coarseFlux,a_scale,a_coarseDataIndex,
                        a_srcInterval,a_dstInterval,a_dir,side());
    }
}

void 
LevelFluxRegister::incrementCoarse(FArrayBox& a_coarseFlux,
                                   Real a_scale,
                                   const DataIndex& a_coarseDataIndex,
                                   const Interval& a_srcInterval,
                                   const Interval& a_dstInterval,
                                   int a_dir,
                                   Side::LoHiSide a_sd)
{
  assert(isDefined());
  assert(!a_coarseFlux.box().isEmpty());
  assert(a_dir >= 0);
  assert(a_dir < SpaceDim);

  assert(a_srcInterval.size() == a_dstInterval.size());
  assert(a_srcInterval.begin() >= 0);
  assert(a_srcInterval.end() < a_coarseFlux.nComp());
  assert(a_dstInterval.begin() >= 0);
  assert(a_dstInterval.end() < m_nComp);


  Vector<DataIndex>& registerIndex = 
    coarToCoarMap[index(a_dir, a_sd)][a_coarseDataIndex];
  LevelData<FArrayBox>&  registers = 
    fabCoarse[index(a_dir, a_sd)];

  a_coarseFlux.shiftHalf(a_dir, sign(a_sd));

  for(int i=0; i<registerIndex.size(); ++i)
    {
      FArrayBox& reg = registers[registerIndex[i]];
      reg.plus(a_coarseFlux, -a_scale, a_srcInterval.begin(),
               a_dstInterval.begin(), a_srcInterval.size());

    }
  a_coarseFlux.shiftHalf(a_dir, - sign(a_sd));
}

void 
LevelFluxRegister::incrementFine(FArrayBox& a_fineFlux,
                                 Real a_scale,
                                 const DataIndex& a_fineDataIndex,
                                 const Interval& a_srcInterval,
                                 const Interval& a_dstInterval,
                                 int a_dir,
                                 Side::LoHiSide a_sd)
{
  assert(isDefined());
  assert(!a_fineFlux.box().isEmpty());
  assert(a_srcInterval.size() == a_dstInterval.size());
  assert(a_srcInterval.begin() >= 0);
  assert(a_srcInterval.end() < a_fineFlux.nComp());
  assert(a_dstInterval.begin() >= 0);
  assert(a_dstInterval.end() < m_nComp);

  assert(a_dir >= 0);
  assert(a_dir < SpaceDim);
  assert((a_sd == Side::Lo)||(a_sd == Side::Hi));
  //
  //
  //denom is the number of fine faces per coarse face
  //this is intrinsically dimension-dependent 
  //yes, i could do some lame loop here, but that would
  //be more confusing than this
#if (CH_SPACEDIM == 2)
  Real denom = m_nRefine;                
#elif (CH_SPACEDIM == 3)
  Real denom = m_nRefine*m_nRefine;
#else
  bogus spacedim;
#endif

  Real scale = a_scale/denom;

  FArrayBox& reg = fabFine[index(a_dir, a_sd)][a_fineDataIndex];

  a_fineFlux.shiftHalf(a_dir, sign(a_sd));

  BoxIterator regIt(m_domainCoarse & reg.box());
  for(regIt.begin(); regIt.ok(); ++regIt)
    {
      const IntVect& coarseIndex = regIt();
      Box box(coarseIndex, coarseIndex);
      box.refine(m_nRefine);
      if(a_sd == Side::Lo) box.growLo(a_dir,-(m_nRefine-1));
      else                 box.growHi(a_dir,-(m_nRefine-1));
      BoxIterator fluxIt(box);
      for(fluxIt.begin(); fluxIt.ok(); ++fluxIt)
        {
          // this is for debugging purposes
          //          IntVect thisIV = fluxIt();
          int src = a_srcInterval.begin();
          int dest =  a_dstInterval.begin();
          for( ; src <=a_srcInterval.end(); ++src,++dest)
            reg(coarseIndex, dest) += scale*a_fineFlux(fluxIt(), src);
        }
    }
  a_fineFlux.shiftHalf(a_dir, -sign(a_sd));
}

void 
LevelFluxRegister::reflux(LevelData<FArrayBox>& a_uCoarse,
                          Real a_scale)
{
  assert(isDefined());
  assert(a_uCoarse.nComp() == m_nComp);

  SideIterator side;
  for(int idir=0 ; idir<SpaceDim; ++idir)
    {
      for(side.begin(); side.ok(); ++side)
        {
          LevelData<FArrayBox>& fineReg = fabFine[index(idir, side())];
          LevelData<FArrayBox>& coarReg = fabCoarse[index(idir, side())];

          // first, we need to create a temp LevelData<FArrayBox>
          // to make a local copy in the coarse layout space of
          // the fine register increments

          LevelData<FArrayBox> fineRegLocal(coarReg.getBoxes(), m_nComp);
          //Copier copier(fineReg.getBoxes(), fineRegLocal.getBoxes(), fineRegLocal.ghostVect());
          // pout() << copier<<endl;
          fineReg.copyTo(fineReg.interval(), fineRegLocal, fineRegLocal.interval(),
                         crseCopiers[index(idir,side())]);

          for(DataIterator it = a_uCoarse.dataIterator(); it.ok(); ++it)
            {
              FArrayBox& U = a_uCoarse[it()];
              Vector<IntVectSet>& ivsV = refluxLocations[index(idir, side())][it()];
              Vector<DataIndex>&  indexV = coarToCoarMap[index(idir, side())][it()];
              IVSIterator iv;
              for(int i=0; i<ivsV.size(); ++i)
                {
                  iv.define(ivsV[i]);
                  const FArrayBox& coar = coarReg[indexV[i]];
                  const FArrayBox& fine = fineRegLocal[indexV[i]];
                  for(iv.begin(); iv.ok(); ++iv)
                    for(int comp=0; comp <m_nComp; ++comp)
                      U(iv(), comp) -= sign(side())*a_scale*(coar(iv(), comp) +
                                                             fine(iv(), comp));
                }
            }
        }
    }  
}


void LevelFluxRegister::reflux( 
                               LevelData<FArrayBox>& a_uCoarse,
                               const Interval&       a_coarse_interval,
                               const Interval&       a_flux_interval,
                               Real                  a_scale )
{
  assert(isDefined());
  assert(a_coarse_interval.size() == a_flux_interval.size());
  assert(a_coarse_interval.begin() >= 0);
  assert(a_coarse_interval.end() < a_uCoarse.nComp());
  assert(a_flux_interval.begin() >= 0);
  assert(a_flux_interval.end() < m_nComp );

  int this_nComp = a_coarse_interval.size();
  int fro = a_flux_interval.begin();
  int cdo = a_coarse_interval.begin();

  SideIterator side;
  for( int idir=0 ; idir<SpaceDim; ++idir )
    {
      for( side.begin(); side.ok(); ++side )
        {
          LevelData<FArrayBox>& fineReg = fabFine[index(idir, side())];
          LevelData<FArrayBox>& coarReg = fabCoarse[index(idir, side())];

          // first, we need to create a temp LevelData<FArrayBox>
          // to make a local copy in the coarse layout space of
          // the fine register increments

          LevelData<FArrayBox> fineRegLocal( coarReg.getBoxes(), m_nComp);
          //Copier copier(fineReg.getBoxes(), fineRegLocal.getBoxes(),
          //fineRegLocal.ghostVect());
          fineReg.copyTo( fineReg.interval(), fineRegLocal,
                          fineRegLocal.interval(), crseCopiers[index(idir,side())]);

          for( DataIterator it = a_uCoarse.dataIterator(); it.ok(); ++it ) 
            {
              FArrayBox& U = a_uCoarse[it()];
              Vector<IntVectSet>& ivsV = refluxLocations[
                                                         index(idir, side())][it()];
              Vector<DataIndex>&  indexV = coarToCoarMap[
                                                         index(idir, side())][it()];
              IVSIterator iv;
              for( int i=0; i<ivsV.size(); ++i ) 
                {
                  iv.define(ivsV[i]);
                  const FArrayBox& coar = coarReg[indexV[i]];
                  const FArrayBox& fine = fineRegLocal[indexV[i]];
                  for( iv.begin(); iv.ok(); ++iv )
                    for( int comp=0; comp <this_nComp; ++comp ) 
                      {
                        int cvc = comp + cdo;
                        int frc = comp + fro;
                        U(iv(), cvc) -= sign(side())*a_scale*(coar(iv(), frc)
                                                              + fine(iv(), frc));
                      }
                }
            }
        }
    }

}

//spew the contents of all the registers
void
LevelFluxRegister::dump()
{
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      dumpLoCoar(idir);
      dumpLoFine(idir);
      dumpHiCoar(idir);
      dumpHiFine(idir);
    }
}
void
LevelFluxRegister::dumpLoCoar(int a_idir)
{
  const LevelData<FArrayBox>& loFabCoar = fabCoarse[index(a_idir, Side::Lo)];
  cout << "LoFabCoar::" << a_idir << endl;
  dumpLDF(&loFabCoar);
}
void
LevelFluxRegister::dumpHiCoar(int a_idir)
{
  const LevelData<FArrayBox>& hiFabCoar = fabCoarse[index(a_idir, Side::Hi)];
  cout << "HiFabCoar::" << a_idir << endl;
  dumpLDF(&hiFabCoar);
}

void
LevelFluxRegister::dumpLoFine(int a_idir)
{
  const LevelData<FArrayBox>& loFabFine = fabFine[index(a_idir, Side::Lo)];
  cout << "LoFabFine::" << a_idir << endl;
  dumpLDF(&loFabFine);
}
void
LevelFluxRegister::dumpHiFine(int a_idir)
{
  const LevelData<FArrayBox>& hiFabFine = fabFine[index(a_idir, Side::Hi)];
  cout << "HiFabFine::" << a_idir << endl;
  dumpLDF(&hiFabFine);
}
