/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _DATAITERATOR_H_
#define _DATAITERATOR_H_


#include "DataIndex.H"
#include "BoxLayout.H"
#include "SPMD.H"


#ifdef MPI

///An Iterator based on a BoxLayout object.
/**
   An Iterator based on a BoxLayout object.  It does not
   support a dereferencing operation, since it is intended
   to work with all of BoxLayouts, DisjointBoxLayouts, BoxLayoutDatas
   LevelDatas, and any object that is built on top
   of a BoxLayout object.  

   DataIterator accesses the data in a BoxLayout-based
   object in a <b> data-parallel </b> manner. 
   This means that it skips over entries for Boxes not
   assigned to this processor.  The order of access is not defined.

   In serial execution mode, there is no difference between
   DataIterator and LayoutIterator.
*/
class DataIterator
{
  //Note: parallel version doesn't inherit LayoutIterator, serial version does
public:

  /// a null constructed DataIterator will return false on ok()
  DataIterator();

  DataIterator(const BoxLayout& a_layout){*this = a_layout.dataIterator();}

  ~DataIterator() {;}

  /// return the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  inline const DataIndex& operator()() const ;
  /// return a copy of the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  DataIndex i() const { return this->operator()();}

  /// move the iterator to the next index in the layout
  inline void operator++();
  /// move the iterator to the next index in the layout
  void incr() { ++(*this);}

  /// return true if this iterator is still in the layout
  bool ok() const;

  /// initialize this iterator to the first index in the layout
  void begin();

  /// same as begin()
  void reset();

  /// move this iterator to after the last index in the layout
  /** The iterator will be !ok() afterwards. */
  void end();

private:

  friend class BoxLayout;
  friend class DisjointBoxLayout;

  DataIterator(const BoxLayout& boxlayout, const int* layoutID);

  BoxLayout m_layout;

  unsigned int m_index;

  DataIndex m_current;

  unsigned int m_procID;

};


inline DataIterator::DataIterator()
  : m_index(0), m_procID(0)
{}


inline const DataIndex& DataIterator::operator()() const
{
  assert(ok());
  return m_current;
}

inline void DataIterator::operator++()
{
  const Entry* box;
  while(++m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
    }

}

inline bool DataIterator::ok() const
{
  return m_index < m_layout.size();
}



inline void DataIterator::reset()
{
  begin();
}

inline void DataIterator::begin()
{
  m_index = 0;
  const Entry* box;
  while(m_index < m_layout.size())
    {
      box = &(*(m_layout.m_boxes))[m_index];
      if(box->m_procID == m_procID)
        {
          m_current.m_index = box->index;
          return;
        }
      ++m_index;
    }
}

#else

// serial version
#include "LayoutIterator.H"

class DataIterator : public LayoutIterator
{
public:
  DataIterator(){;}
  DataIterator(const BoxLayout& a_layout){*this = a_layout.dataIterator();}

  /// return the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  const DataIndex& operator()() const
  {
    return (const DataIndex&)(LayoutIterator::operator()());
  };

  /// return a copy of the index that this iterator is at
  /** Aborts if the iterator is not ok() */
  DataIndex i() const {return this->operator()();}

private:
  friend class BoxLayout;
  friend class DisjointBoxLayout;

  DataIterator(const BoxLayout& boxlayout, const int* layoutID)
	: LayoutIterator(boxlayout, layoutID) {;}
};

#endif  /*MPI*/


#define DATAITERATOR(CLASS, BOXLAYOUT) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN1(CLASS, DataIndex, di)

#define ENDITERATOR(CLASS)  \
        MT_END1(CLASS, DataIndex, di) \
        }

#define DATAITERATOR1(CLASS, BOXLAYOUT, TYPE1, VAL1) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN2(CLASS, TYPE1, VAL1, DataIndex, di)

#define ENDITERATOR1(CLASS, TYPE1, VAL1)  \
        MT_END2(CLASS, TYPE1, VAL1, DataIndex, di) \
        }

#define DATAITERATOR2(CLASS, BOXLAYOUT, TYPE1, VAL1, TYPE2, VAL2) \
        DataIterator dit = BOXLAYOUT .dataIterator(); \
        for(dit.begin(); dit.ok(); ++dit) {     \
        DataIndex di = dit();                   \
        MT_BEGIN3(CLASS, TYPE1, VAL1, TYPE2, VAL2, DataIndex, di)

#define ENDITERATOR2(CLASS, TYPE1, VAL1, TYPE2, VAL2)  \
        MT_END3(CLASS, TYPE1, VAL1, TYPE2, VAL2, DataIndex, di) \
        }

#endif
