/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

// FluxBox.cpp

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// Dan Martin, Fri, Jan 14, 2000


#include "FluxBox.H"

// first do simple access functions
// ---------------------------------------------------------
int 
FluxBox::nComp() const 
{ 
  return m_nvar;
}

// ---------------------------------------------------------
const Box& 
FluxBox::box() const 
{
  return m_bx;
}

// ---------------------------------------------------------
FArrayBox& 
FluxBox::getFlux(const int dir)
{
  assert(m_nvar >0);
  assert(dir < SpaceDim);
  assert(m_fluxes[dir] != NULL);

  return *m_fluxes[dir];
}


// ---------------------------------------------------------
const FArrayBox& 
FluxBox::getFlux(const int dir) const
{
  assert(m_nvar >0);
  assert(dir < SpaceDim);
  assert(m_fluxes[dir] != NULL);

  return *m_fluxes[dir];
}


// ---------------------------------------------------------
FArrayBox&
FluxBox::operator[] (const int dir) 
{
  assert(m_nvar >0);
  assert(dir < SpaceDim);
  assert(m_fluxes[dir] != NULL);

  return *m_fluxes[dir];
}  


// ---------------------------------------------------------
const FArrayBox&
FluxBox::operator[] (const int dir)  const 
{
  assert(m_nvar >0);
  assert(dir < SpaceDim);
  assert(m_fluxes[dir] != NULL);

  return *m_fluxes[dir];
}  


// ---------------------------------------------------------
// constructors and destructors
// ---------------------------------------------------------
FluxBox::FluxBox() : m_fluxes(SpaceDim,NULL)
{
  m_nvar = -1;
}

// ---------------------------------------------------------
FluxBox::FluxBox(const Box& a_bx, int a_nComp)  : m_fluxes(SpaceDim,NULL)
{
  define(a_bx, a_nComp);
}

// ---------------------------------------------------------
FluxBox::~FluxBox()
{

  clear();
}

// ---------------------------------------------------------
void 
FluxBox::clear() 
{
  // first delete storage
  for (int dir=0; dir < SpaceDim; dir++) 
    {
      if (m_fluxes[dir] != NULL) 
        {
          delete m_fluxes[dir];
          m_fluxes[dir] = NULL;
        }
    }

  // now reset all other variables
  m_nvar = -1;
  // set the box to the empty box...
  m_bx = Box();
}
    


// ---------------------------------------------------------
// define function
void
FluxBox::define(const Box& a_bx, int a_nComp)
{
  assert(a_nComp > 0);

  m_bx = a_bx;
  m_nvar = a_nComp;

  m_fluxes.resize(SpaceDim);

  for (int dir=0; dir<SpaceDim; dir++) 
    {
      Box edgeBox(surroundingNodes(m_bx,dir));
      FArrayBox* newFabPtr = new FArrayBox(edgeBox,m_nvar);
      m_fluxes[dir] = newFabPtr;
    }
   
}
  
// ---------------------------------------------------------
// should resize fluxes in space (could be faster than re-allocating
// storage)
void
FluxBox::resize(const Box& a_bx, int a_nComp)
{
  // if this object has not already been defined, call define fn.
  if (m_nvar < 0) 
    {
      define(a_bx, a_nComp);
    } 
  else 
    {
      assert(a_nComp > 0);
      m_bx = a_bx;
      m_nvar = a_nComp;

      for (int dir=0; dir<SpaceDim; dir++) 
        {
          Box edgeBox(surroundingNodes(m_bx, dir));
          if (m_fluxes[dir] != NULL) 
            {
              m_fluxes[dir]->resize(edgeBox,m_nvar);
            } 
          else 
            {
              FArrayBox* newFabPtr = new FArrayBox(edgeBox, m_nvar);
              m_fluxes[dir] = newFabPtr;
            }    
        }

    }

}      


// ---------------------------------------------------------
void
FluxBox::setVal(const Real val)
{
  assert(m_nvar > 0);

  for (int dir = 0; dir < SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != NULL);
      m_fluxes[dir]->setVal(val);
    }
}

// ---------------------------------------------------------
void
FluxBox::setVal(const Real val, const int dir)
{
  assert(dir < SpaceDim);
  
  assert(m_fluxes[dir] != NULL);
  m_fluxes[dir]->setVal(val);
}



// ---------------------------------------------------------
void
FluxBox::setVal(const Real val, const int dir, const int startComp,
                const int nComp)
{
  assert(startComp >-1);
  assert(startComp + nComp < m_nvar);
  assert(dir < SpaceDim);
  assert(m_fluxes[dir] != NULL);

  for (int comp=startComp; comp < startComp+nComp; comp++) 
    {
      m_fluxes[dir]->setVal(val,comp);
    }

}


// ---------------------------------------------------------
void 
FluxBox::setVal(const Real val, const Box& bx) 
{
  assert(m_bx.contains(bx));
  
  for (int dir = 0; dir<SpaceDim; dir++) {
    assert(m_fluxes[dir] != NULL);
    // move cell-centered box to appropriate edge
    Box edgeBox(surroundingNodes(bx, dir));
    m_fluxes[dir]->setVal(val,edgeBox,0,m_nvar);
  }

}


// ---------------------------------------------------------
void
FluxBox::setVal(const Real val, const Box& bx, const int dir,
                const int startComp, const int nComp)
{
  assert(m_bx.contains(bx));
  assert(m_fluxes[dir] != NULL);
  assert(startComp > -1);
  assert(startComp +nComp <=m_nvar);

  Box edgeBox(surroundingNodes(bx, dir));
  m_fluxes[dir]->setVal(val, edgeBox, startComp, nComp);
}


// ---------------------------------------------------------
void 
FluxBox::copy(const FluxBox& src)
{
  assert(src.box() == m_bx);
  assert(src.nComp() == m_nvar);

  for (int dir=0; dir < SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != NULL);
      m_fluxes[dir]->copy(src[dir]);
    }

}

// ---------------------------------------------------------
void 
FluxBox::copy(const FluxBox& src, const int srcComp,
              const int destComp, const int numComp)
{
  // to ensure that neither comp is negative
  assert(srcComp*destComp > -1);
  assert(srcComp+numComp <= src.nComp());
  assert(destComp+numComp <= m_nvar);

  for (int dir=0; dir<SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != 0);
      const FArrayBox& srcFab = src[dir];
      m_fluxes[dir]->copy(srcFab,srcComp, destComp, numComp);
    }
}

// ---------------------------------------------------------
void 
FluxBox::copy(const FluxBox& src, const int dir, 
              const int srcComp, const int destComp,
              const int numComp)
{
  // to ensure that neither comp is negative
  assert(srcComp*destComp > -1);
  assert(srcComp+numComp <= src.nComp());
  assert(destComp+numComp <= m_nvar);
  assert(dir < SpaceDim);
  assert(dir > -1);
  assert(m_fluxes[dir] != NULL);
  
  const FArrayBox& srcFab = src[dir];
  m_fluxes[dir]->copy(srcFab,srcComp, destComp, numComp);
}


// ---------------------------------------------------------
void 
FluxBox::copy(const Box& R, const Interval& Cdest, const FluxBox& src,
              const Interval& Csrc)
{
  
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != NULL);
      Box Redge(R);
      Redge.surroundingNodes(dir);
      const FArrayBox& srcFab = src[dir];
      // all this intersecting is necessary due to the face-centered 
      // nature of things
      Redge &= srcFab.box();
      Redge &= m_fluxes[dir]->box();
      if (!Redge.isEmpty()) 
        {
          //this is probably wrong in periodic---dtg
          m_fluxes[dir]->copy(Redge,Cdest,Redge,srcFab,Csrc);
        }
    }
  
}


void
FluxBox::copy(const Box& srcbox,
              const Interval& srccomps,
              const Box& destbox,
              const FluxBox& src,
              const Interval& destcomps)
{
  
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      assert (m_fluxes[dir] != NULL);
      Box srcEdgeBox(srcbox);
      srcEdgeBox.surroundingNodes(dir);
      Box destEdgeBox(destbox);
      destEdgeBox.surroundingNodes(dir);
      const FArrayBox& srcFab = src[dir];
      // safety check -- due to edge-centered nature of things,
      // destbox may not be contained in m_fluxes[dir]
      destEdgeBox &= (m_fluxes[dir]->box());
      m_fluxes[dir]->copy(srcEdgeBox, srccomps,
                          destEdgeBox, srcFab, destcomps);
    } 

}


// ---------------------------------------------------------
FluxBox&
FluxBox::shift(const IntVect& iv) 
{
  m_bx.shift(iv);

  for (int dir=0; dir<SpaceDim; dir++) 
    {
      m_fluxes[dir]->shift(iv);
    }

  return *this;

}


// ---------------------------------------------------------
int 
FluxBox::size(const Box& bx, const Interval& comps) const
{
  int totalSize = 0;


  FArrayBox tempFab;
  for (int dir =0; dir<SpaceDim; dir++) 
    {
      // why did i put this line in???
      //    assert(m_fluxes[dir] != NULL);
      Box edgeBox(surroundingNodes(bx,dir));
      const int dirSize = tempFab.size(edgeBox,comps);
      totalSize += dirSize;
    }

  return totalSize;
}


// ---------------------------------------------------------
void
FluxBox::linearOut(void* buf, const Box& R, const Interval& comps) const
{
  Real* buffer = (Real*) buf;
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != NULL);
      Box dirBox(surroundingNodes(R,dir));
      int dirSize = m_fluxes[dir]->size(dirBox, comps);
      m_fluxes[dir]->linearOut(buffer, dirBox, comps);
      buffer += dirSize/sizeof(Real);
    }
}
  

// ---------------------------------------------------------
void 
FluxBox::linearIn(void* buf, const Box& R, const Interval& comps)
{
  Real* buffer = (Real*) buf;
  for (int dir=0; dir<SpaceDim; dir++) 
    {
      assert(m_fluxes[dir] != NULL);
      Box dirBox(surroundingNodes(R,dir));
      int dirSize = m_fluxes[dir]->size(dirBox, comps);
      m_fluxes[dir]->linearIn(buffer,dirBox,comps);
      buffer += dirSize/sizeof(Real);
    }
}


