/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef POOL_H
#define POOL_H




#include <cstdlib>
#include "Vector.H"
#include <list>
 
class Pool;
typedef std::list<Pool*> PoolList;


///  Pool is a class to optimize memory allocation.
/**
 Pool is a class to optimize memory allocation.  It is specialized to
 allocate fixed size chunks of memory specified by ptrSize in the
 constructor.  Its operation is analogous to malloc, not new.  It
 does not initialize the memory in any way.  The constructor can
 optionally specify an initial pool size, and memory alignment.  The
 pool size will grow as needed by calling ::new.  The pool size never
 shrinks.  Memory will be reclaimed at ~Pool().  The units of poolSize
 are number-of-ptrSize-chunks.  The units of alignment are bytes.

 Pool can only be used for objects of fixed size.  Typically used in
 situations where a class or struct is being constantly constructed and
 deleted.  Objects returned by value from functions, elements in a database, 
 etc.  Pool's tend to be allocated statically.

from Copier.cpp
\code

// static data member s_motionIemPool getting constructed
Pool Copier::s_motionItemPool(sizeof(MotionItem), "Copier::MotionItem");



// note the use of a 'placement new'
MotionItem* item = new (s_motionItemPool.getPtr()) MotionItem(fromdi, todi, box);


// if your object does requires it's destructor to be called.
item->~MotionItem();

// then return the memory chunk to your pool
s_motionItemPool.returnPtr(item)

\endcode

*/
class Pool {
public:
 
  /// 
  /**
	 @param ptrSize Size of fixed memory allocations needed

	 @param poolSize Size of allocations (ptrSize*poolSize) made
	 of the operating system.  This controls the granularity of the
	 Pool allocations.  Smaller values result in less wasted memory,
	 at the cost of more calls to the operating system for heap memory.

	 @param name optional name for this Pool.  Used in reporting memory 
	 by the memory tracking system in Chombo.

  */
  Pool(int ptrSize, const char* name = "unnamed", 
       int poolSize = 100, int alignment = sizeof(int));
  ///
  ~Pool();
  /// request a section of memory of  ptrSize_  contiguous bytes.
  void* getPtr();
  /// return memory previous acquired with the getPtr() function.
  void returnPtr(void* ptr);
  /// report how much memory this Pool is currently using. 
  /** memUsage for a Pool only grows until Pool destruction. The Pool
	  object has no knowledge of what pieces of memory it has parcelled out
	  to a user, so it keeps it all available. The user is responsible for
	  not leaking Pool memory.
  */
  long memUsage() const;


  /** undocumented function, call this at own risk.  You must
   be absolutely positive that you have returned all the
   ptr's you've asked for, or you can have major seg faults.
  */
  void clear();
 

  // not for public consumption. used in memory tracking code.
  static PoolList* poolList_;
  char name_[64];

protected:
private:

  ///
  Vector<char*> pool_;
  ///
  int ptrSize_;
  ///
  int poolSize_;
  ///
  int alignment_;
  ///
  void* next_; 
  ///
  void* getMoreMemory();
  ///
  /// Not implemented.  Compiler will not let you copy a Pool
  Pool(const Pool& rhs);  
  /// Not implemented. Compiler will not let you copy a Pool
  const Pool& operator=(const Pool& rhs);  

};


#endif
