/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "BoxIterator.H"

#include "DirichletDomainBC.H"
#include "DirichletDomainBCF_F.H"

DirichletDomainBC::DirichletDomainBC()
{
  m_value = 12345.6789;
  m_func = NULL;

  m_onlyHomogeneous = true;
  m_isFunctional = false;

  m_ebOrder = 2;
}

DirichletDomainBC::~DirichletDomainBC()
{
}

void DirichletDomainBC::setValue(Real a_value)
{
  m_value = a_value;
  m_func = NULL;

  m_onlyHomogeneous = false;
  m_isFunctional = false;
}

void DirichletDomainBC::setFunction(RefCountedPtr<BaseBCValue> a_func)
{
  m_value = 12345.6789;
  m_func = a_func;

  m_onlyHomogeneous = false;
  m_isFunctional = true;
}

void DirichletDomainBC::setEBOrder(int a_ebOrder)
{
  assert(a_ebOrder >= 1);
  assert(a_ebOrder <= 2);

  m_ebOrder = a_ebOrder;
}

void DirichletDomainBC::getFaceFlux(FArrayBox&       a_faceFlux,
                                    const FArrayBox& a_phi,
                                    const RealVect&  a_probLo,
                                    const RealVect&  a_dx,
                                    int              a_idir,
                                    Side::LoHiSide   a_side,
                                    bool             a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);
  int comp = 0;

  const Box& box = a_faceFlux.box();

  int iside;

  if (a_side == Side::Lo)
  {
    iside = 1;
  }
  else
  {
    iside = -1;
  }

  if (a_useHomogeneous)
  {
    Real value = 0.0;

    FORT_SETDIRICHLETFACEFLUX(CHF_FRA1(a_faceFlux,comp),
                              CHF_CONST_FRA1(a_phi,comp),
                              CHF_CONST_REAL(value),
                              CHF_CONST_REALVECT(a_dx),
                              CHF_CONST_INT(a_idir),
                              CHF_CONST_INT(iside),
                              CHF_BOX(box));
  }
  else
  {
    if (m_isFunctional)
    {
      Real ihdx;

      ihdx = 2.0 / a_dx[a_idir];

      BoxIterator bit(box);

      for (bit.begin(); bit.ok(); ++bit)
      {
        const IntVect& iv = bit();

        RealVect point = RealVect::Unit;

        point *= 0.5;
        point[a_idir] -= iside * 0.5;
        point += iv;

        point *= a_dx;
        point += a_probLo;

        Real value = m_func->value(point);

        a_faceFlux(iv,comp) = iside * ihdx * (a_phi(iv,comp) - value);
      }
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("DirichletDomainBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      Real value = m_value;

      FORT_SETDIRICHLETFACEFLUX(CHF_FRA1(a_faceFlux,comp),
                                CHF_CONST_FRA1(a_phi,comp),
                                CHF_CONST_REAL(value),
                                CHF_CONST_REALVECT(a_dx),
                                CHF_CONST_INT(a_idir),
                                CHF_CONST_INT(iside),
                                CHF_BOX(box));
    }
  }
}

void DirichletDomainBC::getFaceFlux(Real&            a_faceFlux,
                                    const VolIndex&  a_vof,
                                    const EBCellFAB& a_phi,
                                    const RealVect&  a_probLo,
                                    const RealVect&  a_dx,
                                    int              a_idir,
                                    Side::LoHiSide   a_side,
                                    bool             a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);
  int comp = 0;

  int iside;

  if (a_side == Side::Lo)
  {
    iside = 1;
  }
  else
  {
    iside = -1;
  }

  if (a_useHomogeneous)
  {
    Real value = 0.0;

    Real ihdx;

    ihdx = 2.0 / a_dx[a_idir];

    const EBISBox& ebisBox = a_phi.getEBISBox();

    const IntVect& iv = a_vof.gridIndex();

    RealVect point = RealVect::Unit;

    point *= 0.5;
    point[a_idir] -= iside * 0.5;
    point += iv;

    point *= a_dx;
    point += a_probLo;

    a_faceFlux = iside * ihdx * (a_phi(a_vof,comp) - value);

    Real faceAreaFrac = 0.0;

    Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
    for (int i = 0; i < faces.size(); i++)
    {
      faceAreaFrac += ebisBox.areaFrac(faces[i]);
    }

    a_faceFlux *= faceAreaFrac;
  }
  else
  {
    if (m_isFunctional)
    {
      Real ihdx;

      ihdx = 2.0 / a_dx[a_idir];

      const EBISBox& ebisBox = a_phi.getEBISBox();

      const IntVect& iv = a_vof.gridIndex();

      RealVect point = RealVect::Unit;

      point *= 0.5;
      point[a_idir] -= iside * 0.5;
      point += iv;

      point *= a_dx;
      point += a_probLo;

      Real value = m_func->value(point);

      a_faceFlux = iside * ihdx * (a_phi(a_vof,comp) - value);

      Real faceAreaFrac = 0.0;

      Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
      for (int i = 0; i < faces.size(); i++)
      {
        faceAreaFrac += ebisBox.areaFrac(faces[i]);
      }

      a_faceFlux *= faceAreaFrac;
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("DirichletDomainBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      Real value = m_value;

      Real ihdx;

      ihdx = 2.0 / a_dx[a_idir];

      const EBISBox& ebisBox = a_phi.getEBISBox();

      const IntVect& iv = a_vof.gridIndex();

      RealVect point = RealVect::Unit;

      point *= 0.5;
      point[a_idir] -= iside * 0.5;
      point += iv;

      point *= a_dx;
      point += a_probLo;

      a_faceFlux = iside * ihdx * (a_phi(a_vof,comp) - value);

      Real faceAreaFrac = 0.0;

      Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
      for (int i = 0; i < faces.size(); i++)
      {
        faceAreaFrac += ebisBox.areaFrac(faces[i]);
      }

      a_faceFlux *= faceAreaFrac;
    }
  }
}
