/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "BoxIterator.H"

#include "EBArith.H"

#include "DirichletEBBC.H"

DirichletEBBC::DirichletEBBC(const ProblemDomain& a_domain,
                             const EBISLayout&    a_layout,
                             const RealVect&      a_dx)
{
  m_value = 12345.6789;
  m_func = NULL;

  m_order = 1;

  m_onlyHomogeneous = true;
  m_isFunction = false;

  m_domain = a_domain;
  m_layout = a_layout;

  m_dx = a_dx;

  m_isDefined = false;
}

DirichletEBBC::~DirichletEBBC()
{
}

void DirichletEBBC::define()
{
  const DisjointBoxLayout& dbl = m_layout.getDisjointLayout();

  m_fluxStencil.define(dbl);
  m_fluxWeight.define(dbl);

  for (DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
  {
    const Box& curBox = dbl[dit()];
    const EBISBox& curEBISBox = m_layout[dit()];
    const EBGraph& curEBGraph = curEBISBox.getEBGraph();

    IntVectSet notRegular;
    int nComps = 1;

    notRegular |= curEBISBox.getIrregIVS  (curBox);
    notRegular |= curEBISBox.getMultiCells(curBox);

    BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[dit()];
    BaseIVFAB<Real>&       curWeightBaseIVFAB  = m_fluxWeight[dit()];

    curStencilBaseIVFAB.define(notRegular,curEBGraph,nComps);
    curWeightBaseIVFAB.define(notRegular,curEBGraph,nComps);

    for (VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();

      VoFStencil& curStencil = curStencilBaseIVFAB(vof,0);
      Real&       curWeight  = curWeightBaseIVFAB(vof,0);

      if (m_order == 1)
      {
        getFirstOrderStencil(curStencil,curWeight,vof,curEBISBox,m_dx);
      }
      else
      if (m_order == 2)
      {
        getSecondOrderStencil(curStencil,curWeight,vof,curEBISBox,m_dx);
      }
      else
      {
        MayDay::Error("DirichletEBBC::define stencil order not 1 or 2");
      }
    }
  }

  m_isDefined = true;
}

void DirichletEBBC::setOrder(int a_order)
{
  assert(a_order >= 1 && a_order <= 2);

  if (m_order != a_order)
  {
    m_isDefined = false;
  }

  m_order = a_order;
}

void DirichletEBBC::setValue(Real a_value)
{
  m_value = a_value;
  m_func = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}

void DirichletEBBC::setFunction(RefCountedPtr<BaseBCValue> a_func)
{
  m_value = 12345.6789;
  m_func = a_func;

  m_onlyHomogeneous = false;
  m_isFunction = true;
}

void DirichletEBBC::getEBFlux(Real&                       a_flux,
                              const VolIndex&             a_vof,
                              const LevelData<EBCellFAB>& a_phi,
                              const DataIndex             a_dit,
                              const RealVect&             a_probLo,
                              const RealVect&             a_dx,
                              bool                        a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);

  const EBCellFAB& curPhi = a_phi[a_dit];
  const EBISBox    curEBISBox = curPhi.getEBISBox();

  int comp = 0;

  if (!m_isDefined)
  {
    define();
  }

  Real value;

  if (a_useHomogeneous)
  {
    value = 0.0;
  }
  else
  {
    if (m_isFunction)
    {
      const IntVect& iv = a_vof.gridIndex();

      RealVect point = RealVect::Unit;
      point *= 0.5;
      point += iv;

      point += curEBISBox.bndryCentroid(a_vof);
      point *= a_dx;

      point += a_probLo;

      value = m_func->value(point);
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("NeumannEBBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      value = m_value;
    }
  }

  a_flux = 0.0;

  const BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_fluxStencil[a_dit];
  const VoFStencil& curStencil = curStencilBaseIVFAB(a_vof,comp);
  for (int i = 0; i < curStencil.size(); i++)
  {
    VolIndex curVoF = curStencil.vof(i);
    Real weight = curStencil.weight(i);

    a_flux += weight * curPhi(curVoF,comp);
  }

  const BaseIVFAB<Real>& curWeightBaseIVFAB = m_fluxWeight[a_dit];
  const Real& curWeight = curWeightBaseIVFAB(a_vof,comp);

  a_flux += curWeight * value;

  const Real& areaFrac = curEBISBox.bndryArea(a_vof);
  a_flux *= areaFrac;
}

void DirichletEBBC::getFirstOrderStencil(VoFStencil&     a_stencil,
                                         Real&           a_weight,
                                         const VolIndex& a_vof,
                                         const EBISBox&  a_ebisBox,
                                         const RealVect& a_dx)
{
  RealVect normal   = a_ebisBox.normal(a_vof);
  RealVect centroid = a_ebisBox.bndryCentroid(a_vof);

  IntVect iv0 = a_vof.gridIndex();

#if   CH_SPACEDIM == 2
  int stenSize = 3;
#elif CH_SPACEDIM == 3
  int stenSize = 7;
#else
  tjlCompileTimeError;
#endif

  IntVect sign;
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    if (normal[idir] < 0)
    {
      sign[idir]= -1;
    }
    else
    {
      sign[idir] = 1;
    }
  }

  Vector<IntVect> ivSten(stenSize);

  ivSten[0] = iv0 + sign[0]*BASISV(0)                                        ;
  ivSten[1] = iv0                     + sign[1]*BASISV(1)                    ;
  ivSten[2] = iv0 + sign[0]*BASISV(0) + sign[1]*BASISV(1)                    ;
#if CH_SPACEDIM == 3
  ivSten[3] = iv0                                         + sign[2]*BASISV(2);
  ivSten[4] = iv0 + sign[0]*BASISV(0)                     + sign[2]*BASISV(2);
  ivSten[5] = iv0                     + sign[1]*BASISV(1) + sign[2]*BASISV(2);
  ivSten[6] = iv0 + sign[0]*BASISV(0) + sign[1]*BASISV(1) + sign[2]*BASISV(2);
#endif

  bool dropOrder = false;

  Vector<VolIndex> volSten(stenSize);
  for (int isten = 0; isten < stenSize; isten++)
  {
    if (a_ebisBox.getDomain().contains(ivSten[isten]))
    {
      if (a_ebisBox.getVoFs(ivSten[isten]).size() > 0)
      {
        volSten[isten] = a_ebisBox.getVoFs(ivSten[isten])[0];
      }
      else
      {
        dropOrder = true;
        break;
      }
    }
    else
    {
      dropOrder = true;
      break;
    }
  }

  std::vector<bool> monotonePath(stenSize);

  for (int isten = 0; isten < stenSize; isten++)
  {
    monotonePath[isten] = EBArith::monotonePathVoFToCellVoF(volSten[isten],
                                                            a_vof,
                                                            ivSten[isten],
                                                            a_ebisBox);
    if (!monotonePath[isten])
    {
      dropOrder = true;
    }
  }

  if (!dropOrder)
  {
    RealVect x0;
    for (int idir = 0; idir < SpaceDim; idir++)
    {
      x0[idir] = a_dx[idir] * (0.5 + centroid[idir] + iv0[idir]);
    }

    Vector<RealVect> xp(stenSize);
    for (int isten = 0; isten < stenSize; isten++)
    {
      for (int idir = 0; idir < SpaceDim; idir++)
      {
        xp[isten][idir] = a_dx[idir] * (0.5 + ivSten[isten][idir]);
      }
    }

    Vector<RealVect> deltaX = xp;
    for (int isten = 0; isten < stenSize; isten++)
    {
      deltaX[isten] -= x0;
    }

    Vector<RealVect> aTransA(SpaceDim), invATransA(SpaceDim);
    for (int idir = 0; idir < SpaceDim; idir++)
    {
      for (int jdir = 0; jdir < SpaceDim; jdir++)
      {
        for (int isten = 0; isten < stenSize; isten++)
        {
          aTransA[idir][jdir] = aTransA[idir][jdir]
                              + deltaX[isten][idir]*deltaX[isten][jdir];
        }
      }
    }

    Real det;
#if   CH_SPACEDIM == 2
    det = aTransA[0][0] * aTransA[1][1] - aTransA[0][1] * aTransA[1][0];

    invATransA[0][0] =  aTransA[1][1] / det;
    invATransA[0][1] = -aTransA[0][1] / det;
    invATransA[1][0] = -aTransA[1][0] / det;
    invATransA[1][1] =  aTransA[0][0] / det;
#elif CH_SPACEDIM == 3
    det = aTransA[0][0] * ( aTransA[1][1] * aTransA[2][2]
                          - aTransA[1][2] * aTransA[2][1])
        + aTransA[0][1] * ( aTransA[1][2] * aTransA[2][0]
                          - aTransA[1][0] * aTransA[2][2])
        + aTransA[0][2] * ( aTransA[1][0] * aTransA[2][1]
                          - aTransA[1][1] * aTransA[2][0]);

    invATransA[0][0] = ( aTransA[1][1] * aTransA[2][2]
                       - aTransA[1][2] * aTransA[2][1]) / det;
    invATransA[0][1] = ( aTransA[1][2] * aTransA[2][0]
                       - aTransA[1][0] * aTransA[2][2]) / det;
    invATransA[0][2] = ( aTransA[1][0] * aTransA[2][1]
                       - aTransA[1][1] * aTransA[2][0]) / det;
    invATransA[1][0] = ( aTransA[2][1] * aTransA[0][2]
                       - aTransA[2][2] * aTransA[0][1]) / det;
    invATransA[1][1] = ( aTransA[2][2] * aTransA[0][0]
                       - aTransA[2][0] * aTransA[0][2]) / det;
    invATransA[1][2] = ( aTransA[2][0] * aTransA[0][1]
                       - aTransA[2][1] * aTransA[0][0]) / det;
    invATransA[2][0] = ( aTransA[0][1] * aTransA[1][2]
                       - aTransA[0][2] * aTransA[1][1]) / det;
    invATransA[2][1] = ( aTransA[0][2] * aTransA[1][0]
                       - aTransA[0][0] * aTransA[1][2]) / det;
    invATransA[2][2] = ( aTransA[0][0] * aTransA[1][1]
                       - aTransA[0][1] * aTransA[1][0]) / det;
#else
    tjlCompileTimeError;
#endif

    Vector<RealVect> invATransAdeltaX(stenSize);
    for (int idir = 0; idir < SpaceDim; idir++)
    {
      for (int isten = 0; isten < stenSize; isten++)
      {
        for (int jdir = 0; jdir < SpaceDim; jdir++)
        {
          invATransAdeltaX[isten][idir] = invATransAdeltaX[isten][idir]
                                        + invATransA[idir][jdir] * deltaX[isten][jdir];
        }
      }
    }

    a_stencil.clear();
    a_weight = 0.0;

    for (int isten = 0; isten < stenSize; isten++)
    {
      Real dphidnWeight = 0.0;
      for (int idir = 0; idir < SpaceDim; idir++)
      {
        dphidnWeight -= invATransAdeltaX[isten][idir] * normal[idir];
      }

      a_stencil.add(volSten[isten],dphidnWeight);
      a_weight -= dphidnWeight;
    }
  }
  else
  {
    a_stencil.clear();
    a_weight = 0.0;
  }
}

void DirichletEBBC::getSecondOrderStencil(VoFStencil&     a_stencil,
                                          Real&           a_weight,
                                          const VolIndex& a_vof,
                                          const EBISBox&  a_ebisBox,
                                          const RealVect& a_dx)
{
  bool droppedOrder = false;
  IntVect iv = a_vof.gridIndex();

  // Do Johansen-Colella
  RealVect bdCentroid = a_ebisBox.bndryCentroid(a_vof);
  RealVect normal     = a_ebisBox.normal(a_vof);

  // More complicated to prepare for anisotropic case
  RealVect vectDx = a_dx;

  // Find InterpDirection and hiLo
  Real nMax = 0.0;
  int nMaxDir;

  for (int idir = 0; idir < SpaceDim; ++idir)
  {
    if (Abs(normal[idir]) > nMax)
    {
      nMax = Abs(normal[idir]);
      nMaxDir = idir;
    }
  }

  int hiLo = 1;
  if (normal[nMaxDir] < 0.0)
  {
    hiLo = -1;
  }

  // Find Intersection of planes and ray
  RealVect intersection[2];

  // We want the intersection in a coordinate system with dx = dy = dz = 1 and
  // the the cell center of a_vof is the origin

  // We know the planes are defined by nMaxDir = pm (1 or 2)
  intersection[0][nMaxDir] = 1.0*hiLo;
  intersection[1][nMaxDir] = 2.0*hiLo;

  // ScaledCentroid = centroid in an anisotropic coordinate system with
  // the the cell center of a_vof is the origin
  RealVect scaledCentroid;
  for (int idir = 0; idir < SpaceDim; ++idir)
  {
    scaledCentroid[idir] = vectDx[idir]*bdCentroid[idir];
  }

  // Get intersection points(relative to VectDx)
  for (int iPlane = 0; iPlane < 2; ++iPlane)
  {
    for (int idir = 0; idir < SpaceDim; ++idir)
    {
      if (idir != nMaxDir)
      {
        // We get intersection[iPlane][idir] in an anisotropic
        // coordinate system with the cell center of a_vof = origin
        intersection[iPlane][idir] = scaledCentroid[idir]
          + (normal[idir] * ( vectDx[nMaxDir]*intersection[iPlane][nMaxDir]
                            - scaledCentroid[nMaxDir]) / normal[nMaxDir]);

        // We make intersection[iPlane][idir] in a coordinate system with
        // dx = dy = dz = 1 and the the cell center of a_vof is the origin
        intersection[iPlane][idir] /= vectDx[idir];
      }
    }
  }

  // Find interpCen
  IntVect interpCen[2];
  interpCen[0] = iv;
  interpCen[1] = iv;

  // Get ivs for cell containing intersection point
  for (int iPlane = 0; iPlane < 2; ++iPlane)
  {
    for (int idir = 0; idir < SpaceDim; ++idir)
    {
      interpCen[iPlane][idir] += (int)(rint(intersection[iPlane][idir]));
    }
  }

  // planeDir[0] and planeDir[1] are the directions != nMaxDir
  int planeDir[SpaceDim-1];
  planeDir[0] = (nMaxDir + 1) % SpaceDim;
#if CH_SPACEDIM == 3
  planeDir[1] = (planeDir[0] + 1) % SpaceDim;
#endif

  // Check do we have enough cells for a stencil
  johanDropOrder(droppedOrder,interpCen,planeDir,a_ebisBox);

  if (!droppedOrder)
  {
    // We have enough cells, so interpolate

    //vofStencil for each point in iPlane
    VoFStencil iPlaneStencil[2];
    for (int iPlane = 0; iPlane < 2; ++iPlane)
    {
      VolIndex colVofs[3];

#if   CH_SPACEDIM == 3
      VoFStencil colStencil[3];

      for (int iRow = -1; iRow < 2; ++iRow)
      {
        IntVect center = interpCen[iPlane] + iRow*BASISV(planeDir[0]);
        IntVect low    = center - BASISV(planeDir[1]);
        IntVect high   = center + BASISV(planeDir[1]);

        VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
        VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
        VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

        colVofs[1] = centerVof;
        colVofs[0] = lowVof;
        colVofs[2] = highVof;

        // Distance (relative to dx)
        Real distance = intersection[iPlane][planeDir[1]] -
          (interpCen[iPlane][planeDir[1]] - iv[planeDir[1]]);

        // BuildStencil
        colStencil[iRow+1] = BuildEQIStencil(colVofs,distance);
      }
#elif CH_SPACEDIM == 2
      IntVect center = interpCen[iPlane];
      IntVect low    = center - BASISV(planeDir[0]);
      IntVect high   = center + BASISV(planeDir[0]);

      VolIndex centerVof = a_ebisBox.getVoFs(center)[0];
      VolIndex lowVof    = a_ebisBox.getVoFs(low)[0];
      VolIndex highVof   = a_ebisBox.getVoFs(high)[0];

      colVofs[1] = centerVof;
      colVofs[0] = lowVof;
      colVofs[2] = highVof;
#else
      tjlCompileTimeError;
#endif

      //distance (relative to dx)
      Real distance = intersection[iPlane][planeDir[0]] -
        (interpCen[iPlane][planeDir[0]] - iv[planeDir[0]]);

      // Build stencil
#if   CH_SPACEDIM == 3
      iPlaneStencil[iPlane] = BuildEQIStencil(colStencil,distance);
#elif CH_SPACEDIM == 2
      iPlaneStencil[iPlane] = BuildEQIStencil(colVofs,distance);
#else
      tjlCompileTimeError;
#endif
    }

    Real d1 = 0.0;
    Real d2 = 0.0;

    for (int idir = 0; idir < SpaceDim; ++idir)
    {
      d1 += (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir])
          * (intersection[0][idir]*vectDx[idir] - scaledCentroid[idir]);

      d2 += (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir])
          * (intersection[1][idir]*vectDx[idir]- scaledCentroid[idir]);
    }

    d1 = sqrt(d1);
    d2 = sqrt(d2);

    // Build stencil
    a_stencil.clear();
    a_stencil = ivofStencil(iPlaneStencil,d1,d2);

    a_weight = 1/(d2-d1) * (d2/d1 - d1/d2);
  }
  else
  {
    getFirstOrderStencil(a_stencil,a_weight,a_vof,a_ebisBox,a_dx);
  }
}

void DirichletEBBC::johanDropOrder(bool&          a_droppedOrder,
                                   const IntVect  a_interpCen[2],
                                   const int      a_planeDir[SpaceDim-1],
                                   const EBISBox& a_ebisBox)
{
  // Return a_droppedOrder = true or false;
  a_droppedOrder = false;
  Box region = a_ebisBox.getRegion();

  Vector <VolIndex> interpVof[2];
  VolIndex dummyVof;

  // Check do we have enough cells for a stencil
  for (int iPlane = 0; iPlane < 2; ++iPlane)
  {
    if (!region.contains(a_interpCen[iPlane]))
    {
      a_droppedOrder = true;
      return;
    }

    interpVof[iPlane] = a_ebisBox.getVoFs(a_interpCen[iPlane]);

    if (interpVof[iPlane].size() != 1)
    {
      a_droppedOrder = true;
      return;
    }
    else
    {
      int radius = 1;
      Vector<VolIndex> vofsStencil;
      IntVect timesMoved = IntVect::Zero;
      IntVect pathSign   = IntVect::Zero;
      EBArith::getAllVoFsInMonotonePath(vofsStencil,
                                        timesMoved,
                                        pathSign,
                                        interpVof[iPlane][0],
                                        a_ebisBox,
                                        radius);

      // Recall planeDir[0] and planeDir[1] are the directions != nMaxDir

      // In 2D check if 2 neighboring cells are in vofsStencil
      // In 3D check if all cells (!= interpCen) in a 3X3 neighborhood are in
      //                vofsStencil
      for (int iRow = -1; iRow < 2; ++iRow)
      {
        int numCols = SpaceDim - 2;
        for (int iColumn = -numCols; iColumn < numCols + 1; ++iColumn)
        {
          IntVect cell = a_interpCen[iPlane] + iRow * BASISV(a_planeDir[0]);
#if CH_SPACEDIM == 3
          cell += iColumn * BASISV(a_planeDir[1]);
#endif
          if (cell != a_interpCen[iPlane])
          {
            bool isVof = EBArith::isVoFHere(dummyVof,vofsStencil,cell);
            if (!isVof)
            {
              a_droppedOrder = true;
              return;
            }
          }
        }
      }
    }
  }

  return;
}

VoFStencil DirichletEBBC::BuildEQIStencil(const VolIndex a_vofs[3],
                                          const Real&    a_distance)
{
  VoFStencil retval;

  Real x  = a_distance;
  Real x2 = x * x;

  retval.add(a_vofs[0], (x2 - x) / 2.0);
  retval.add(a_vofs[1],-(x2 - 1)      );
  retval.add(a_vofs[2], (x2 + x) / 2.0);

  return retval;
}

VoFStencil DirichletEBBC::BuildEQIStencil(VoFStencil a_vofs[3],
                                          const Real& a_distance)
{
  VoFStencil retval;

  Real x  = a_distance;
  Real x2 = x * x;

  a_vofs[0] *=  (x2 - x) / 2.0;
  retval += a_vofs[0];

  a_vofs[1] *= -(x2 - 1);
  retval += a_vofs[1];

  a_vofs[2] *=  (x2 + x) / 2.0;
  retval += a_vofs[2];

  return retval;
}

VoFStencil DirichletEBBC::ivofStencil(const VoFStencil a_iPlaneStencil[2],
                                      const Real&      a_distance1,
                                      const Real&      a_distance2)
{
  VoFStencil retvalTemp;
  VoFStencil retval;

  Real d1 = a_distance1;
  Real d2 = a_distance2;

  retvalTemp = a_iPlaneStencil[0];
  retvalTemp *= 1/(d2-d1) * (-d2/d1);
  retval += retvalTemp;
  
  retvalTemp = a_iPlaneStencil[1];
  retvalTemp *= 1/(d2-d1) * (d1/d2);
  retval += retvalTemp;

  return retval;
}


DirichletEBBCFactory::DirichletEBBCFactory()
{
  m_value = 12345.6789;
  m_func = NULL;

  m_order = 1;

  m_onlyHomogeneous = true;
  m_isFunction = false;
}

DirichletEBBCFactory::~DirichletEBBCFactory()
{
}

void DirichletEBBCFactory::setOrder(int a_order)
{
  assert(a_order >= 1 && a_order <= 2);
  m_order = a_order;
}

void DirichletEBBCFactory::setValue(Real a_value)
{
  m_value = a_value;
  m_func = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}

void DirichletEBBCFactory::setFunction(RefCountedPtr<BaseBCValue> a_func)
{
  m_value = 12345.6789;
  m_func = a_func;

  m_onlyHomogeneous = false;
  m_isFunction = true;
}

DirichletEBBC* DirichletEBBCFactory::newEBBC(const ProblemDomain& a_domain,
                                             const EBISLayout&    a_layout,
                                             const RealVect&      a_dx)
{
  DirichletEBBC* fresh = new DirichletEBBC(a_domain,a_layout,a_dx);

  fresh->setOrder(m_order);

  if (!m_onlyHomogeneous)
  {
    if (!m_isFunction)
    {
      fresh->setValue(m_value);
    }
    else
    {
      fresh->setFunction(m_func);
    }
  }

  return fresh;
}
