/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// TJLigocki, August 26, 2003

#ifndef _EBAMRPOISSONOP_H_
#define _EBAMRPOISSONOP_H_

#include "REAL.H"
#include "Box.H"
#include "FArrayBox.H"
#include "Vector.H"
#include "RefCountedPtr.H"

#include "AMRMultiGrid.H"

#include "EBIndexSpace.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"

#include "EBLevelDataOps.H"
#include "BaseEBBC.H"
#include "BaseDomainBC.H"

class EBAMRPoissonOp: public AMRLevelOp<LevelData<EBCellFAB> >
{
public:
  EBAMRPoissonOp();

  ~EBAMRPoissonOp();

  void define(const ProblemDomain&                   a_domain,
              const EBISLayout&                      a_layout,
              const RealVect&                        a_probLo,
              const RealVect&                        a_dx,
              const int&                             a_orderEB,
              RefCountedPtr<BaseDomainBC>            a_domainBC,
              RefCountedPtr<BaseEBBC>                a_ebBC,
              RefCountedPtr<DataFactory<EBCellFAB> > a_factory);

  void define(const ProblemDomain&                   a_domain,
              const EBISLayout&                      a_layout,
              const RealVect&                        a_probLo,
              const RealVect&                        a_dx,
              const int&                             a_orderEB,
              RefCountedPtr<BaseDomainBC>            a_domainBC,
              RefCountedPtr<BaseEBBC>                a_ebBC,
              RefCountedPtr<DataFactory<EBCellFAB> > a_factory,
              const int&                             a_relaxType,
              const DisjointBoxLayout&               a_dbl,
              RefCountedPtr<DataFactory<EBCellFAB> > a_factoryCoarse,
              RefCountedPtr<DataFactory<EBCellFAB> > a_factoryFine,
              const bool&                            a_coarseFlag);

  /**
     \name Linear operator functions */
  /*@{*/

  virtual void residual(LevelData<EBCellFAB>&       a_residual,
                        const LevelData<EBCellFAB>& a_phi,
                        const LevelData<EBCellFAB>& a_rhs,
                        bool                        a_homogeneous=false);

  virtual void residual(LevelData<EBCellFAB>&       a_residual,
                        const LevelData<EBCellFAB>& a_phi,
                        const LevelData<EBCellFAB>& a_rhs,
                        bool                        a_homogeneous,
                        const IntVect&              a_parity);

  virtual void preCond(LevelData<EBCellFAB>&       a_opPhi,
                       const LevelData<EBCellFAB>& a_phi);

  virtual void applyOp(LevelData<EBCellFAB>&       a_opPhi,
                       const LevelData<EBCellFAB>& a_phi,
                       bool                        a_homogeneous=false);

  virtual void create(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  virtual void assign(LevelData<EBCellFAB>&       a_lhs,
                      const LevelData<EBCellFAB>& a_rhs);

  virtual Real dotProduct(const LevelData<EBCellFAB>& a_1,
                          const LevelData<EBCellFAB>& a_2);

  virtual void incr(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    Real                        a_scale);

  virtual void axby(LevelData<EBCellFAB>&       a_lhs,
                    const LevelData<EBCellFAB>& a_x,
                    const LevelData<EBCellFAB>& a_y,
                    Real                        a_a,
                    Real                        a_b);

  virtual void scale(LevelData<EBCellFAB>& a_lhs,
                     const Real&           a_scale);

  virtual Real norm(const LevelData<EBCellFAB>& a_rhs,
                    int                         a_ord);

  virtual void setToZero(LevelData<EBCellFAB>& a_lhs);

  /*@}*/

  /**
     \name MGLevelOp functions */
  /*@{*/

  virtual void createCoarser(LevelData<EBCellFAB>&       a_coarse,
                             const LevelData<EBCellFAB>& a_fine,
                             bool                        a_ghosted);

  virtual void relax(LevelData<EBCellFAB>&       a_e,
                     const LevelData<EBCellFAB>& a_residual,
                     int                         a_iterations);

  /**
     Calculate restricted residual:
       a_resCoarse[2h] = I[h->2h] (a_rhsFine[h] - L[h](a_phiFine[h]))
  */
  virtual void restrictResidual(LevelData<EBCellFAB>&       a_resCoarse,
                                LevelData<EBCellFAB>&       a_phiFine,
                                const LevelData<EBCellFAB>& a_rhsFine);

  /**
     Correct the fine solution based on coarse correction:
       a_phiThisLevel += I[2h->h] (a_correctCoarse)
  */
  virtual void prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                                const LevelData<EBCellFAB>& a_correctCoarse);

  /*@}*/

  /**
     \name AMRLevelOp functions */
  /*@{*/

  /** Returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToCoarser();

  /** Generate approximate solution to L(a_correction) = a_res */
  virtual void AMRRelaxLevel(LevelData<EBCellFAB>& a_correction,
                             const LevelData<EBCellFAB>& a_res);

  /** a_residual = a_rhs - L(a_phi, a_phiFine, a_phiCoarse) */
  virtual void AMRResidual(LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_phiFine,
                           const LevelData<EBCellFAB>& a_phi,
                           const LevelData<EBCellFAB>& a_phiCoarse,
                           const LevelData<EBCellFAB>& a_rhs) ;

  /** a_residual = a_rhs - L(a_phi, a_phiCoarse)  */
  virtual void AMRResidualNF(LevelData<EBCellFAB>& a_residual,
                             const LevelData<EBCellFAB>& a_phi,
                             const LevelData<EBCellFAB>& a_phiCoarse,
                             const LevelData<EBCellFAB>& a_rhs);

  /** a_resCoarse = I[h-2h] (a_residual - L(a_correction, a_coarseCorrection)) */
  virtual void AMRRestrict(LevelData<EBCellFAB>& a_resCoarse,
                           const LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_correction,
                           const LevelData<EBCellFAB>& a_coarseCorrection);

  /** a_correction += I[2h->h](a_coarseCorrection) */
  virtual void AMRProlong(LevelData<EBCellFAB>&       a_correction,
                          const LevelData<EBCellFAB>& a_coarseCorrection);

  /** a_residual = a_residual - L(a_correction, a_coarseCorrection) */
  virtual void AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
                                 const LevelData<EBCellFAB>& a_correction,
                                 const LevelData<EBCellFAB>& a_coarseCorrection);

  /*@}*/

protected:
  bool           m_isDefined;

  bool           m_coarseFlag;
  ProblemDomain  m_domain;
  RealVect       m_probLo;
  RealVect       m_dx;

  int            m_orderEB;
  int            m_relaxType;

  RefCountedPtr<BaseDomainBC> m_domainBC;
  RefCountedPtr<BaseEBBC>     m_ebBC;

  RefCountedPtr<DataFactory<EBCellFAB> > m_factory;

  DisjointBoxLayout                      m_dblCoarse;
  RefCountedPtr<DataFactory<EBCellFAB> > m_factoryCoarse;
  RefCountedPtr<DataFactory<EBCellFAB> > m_factoryFine;

  EBLevelDataOps m_levelOps;

  LayoutData<BaseIVFAB<Real> >       m_weight;
  LayoutData<BaseIVFAB<VoFStencil> > m_opStencil;

  IntVect m_noParity;

  void getOpVoFStencil(VoFStencil&     a_stencil,
                       const EBISBox&  a_ebisbox,
                       const VolIndex& a_vof);

  void getOpFaceStencil(VoFStencil&             a_stencil,
                        const Vector<VolIndex>& a_allMonotoneVofs,
                        const EBISBox&          a_ebisbox,
                        const VolIndex&         a_vof,
                        int                     a_dir,
                        const Side::LoHiSide&   a_side,
                        const FaceIndex&        a_face);

  void levelMulticolorGS(LevelData<EBCellFAB>&       a_phi,
                         const LevelData<EBCellFAB>& a_rhs);

  void levelJacobi(LevelData<EBCellFAB>&       a_phi,
                   const LevelData<EBCellFAB>& a_rhs);

  void average(LevelData<EBCellFAB>&       a_coarse,
               const LevelData<EBCellFAB>& a_fine);

  void privateProlongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                               const LevelData<EBCellFAB>& a_correctCoarse);
};


class EBAMRPoissonOpFactory: public AMRLevelOpFactory<LevelData<EBCellFAB> >
{
public:
  EBAMRPoissonOpFactory();

  virtual ~EBAMRPoissonOpFactory();

  void define(const ProblemDomain&           a_domain,
              const DisjointBoxLayout&       a_dbl,
              const EBIndexSpace*            a_space,
              const int&                     a_numGhostCells,
              const int&                     a_numLevels,
              const RealVect&                a_probLo,
              const RealVect&                a_dx,
              const int&                     a_orderEB,
              const int&                     a_relaxType,
              RefCountedPtr<BaseDomainBC>    a_domainBC,
              RefCountedPtr<BaseEBBCFactory> a_ebBcFactory);

  /**
     \name MGLevelOpFactory functions */
  /*@{*/

  virtual EBAMRPoissonOp*
  MGnewOp(const ProblemDomain& a_FineindexSpace,
                int            a_depth,
                bool           a_homoOnly = true);

  virtual void reclaim(MGLevelOp<LevelData<EBCellFAB> >* a_reclaim);

  /*@}*/

  /**
     \name AMRLevelOpFactory functions */
  /*@{*/

  virtual EBAMRPoissonOp*
  AMRnewOp(const ProblemDomain& a_FineindexSpace);

  virtual void AMRreclaim(EBAMRPoissonOp* a_reclaim);

  /*@}*/

protected:
  bool m_isDefined;

  RealVect m_probLo;
  int      m_orderEB;
  int      m_relaxType;

  int      m_numLevels;

  Vector<int>           m_coarseFlag;
  Vector<ProblemDomain> m_domain;
  Vector<EBISLayout>    m_layout;
  Vector<RealVect>      m_dx;

  RefCountedPtr<BaseDomainBC>    m_domainBC;
  RefCountedPtr<BaseEBBCFactory> m_ebBcFactory;

  Vector<DisjointBoxLayout >                      m_dbl;
  Vector<RefCountedPtr<DataFactory<EBCellFAB> > > m_factory;
  Vector<RefCountedPtr<DataFactory<EBCellFAB> > > m_factoryFine;

  EBLevelDataOps m_levelOps;
};

#endif
