/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// TJLigocki, August 26, 2003

#include "LoadBalance.H"
#include "EBArith.H"

#include "EBAMRPoissonOp.H"
#include "EBEllipticLoHiCenter.H"

#include "EBAMRPoissonOpF_F.H"

EBAMRPoissonOp::EBAMRPoissonOp()
{
  m_isDefined = false;

  m_noParity  = IntVect::TheZeroVector();
  m_noParity -= IntVect::TheUnitVector();
}

EBAMRPoissonOp::~EBAMRPoissonOp()
{
}

void EBAMRPoissonOp::define(const ProblemDomain&                   a_domain,
                            const EBISLayout&                      a_layout,
                            const RealVect&                        a_probLo,
                            const RealVect&                        a_dx,
                            const int&                             a_orderEB,
                            RefCountedPtr<BaseDomainBC>            a_domainBC,
                            RefCountedPtr<BaseEBBC>                a_ebBC,
                            RefCountedPtr<DataFactory<EBCellFAB> > a_factory)
{
  int badRelaxType = -1;
  DisjointBoxLayout dblNull;
  RefCountedPtr<DataFactory<EBCellFAB> > factoryNull = NULL;
  bool coarserLevel = false;

  define(a_domain,a_layout,a_probLo,a_dx,
         a_orderEB,a_domainBC,a_ebBC,a_factory,
         badRelaxType,dblNull,factoryNull,factoryNull,coarserLevel);
}

void EBAMRPoissonOp::define(const ProblemDomain&                   a_domain,
                            const EBISLayout&                      a_layout,
                            const RealVect&                        a_probLo,
                            const RealVect&                        a_dx,
                            const int&                             a_orderEB,
                            RefCountedPtr<BaseDomainBC>            a_domainBC,
                            RefCountedPtr<BaseEBBC>                a_ebBC,
                            RefCountedPtr<DataFactory<EBCellFAB> > a_factory,
                            const int&                             a_relaxType,
                            const DisjointBoxLayout&               a_dblCoarse,
                            RefCountedPtr<DataFactory<EBCellFAB> > a_factoryCoarse,
                            RefCountedPtr<DataFactory<EBCellFAB> > a_factoryFine,
                            const bool&                            a_coarseFlag)
{
  assert(!m_isDefined);

  m_domain = a_domain;

  m_probLo = a_probLo;
  m_dx = a_dx;

  if (a_orderEB < 0 || a_orderEB > 1)
  {
    MayDay::Error("EBAMRPoissonOp::define - EB order must be 0 or 1");
  }
  m_orderEB = a_orderEB;
  m_relaxType = a_relaxType;

  m_domainBC = a_domainBC;
  m_ebBC     = a_ebBC;

  m_factory       = a_factory;

  m_dblCoarse     = a_dblCoarse;
  m_factoryCoarse = a_factoryCoarse;
  m_factoryFine   = a_factoryFine;

  m_coarseFlag = a_coarseFlag;

  m_levelOps.define(m_factory);

  const DisjointBoxLayout& dbl = a_layout.getDisjointLayout();

  m_opStencil.define(dbl);
  m_weight.define(dbl);

  for (DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
  {
    const Box& curBox = dbl[dit()];
    const EBISBox& curEBISBox = a_layout[dit()];
    const EBGraph& curEBGraph = curEBISBox.getEBGraph();

    IntVectSet notRegular;
    int nComps = 1;

    notRegular |= curEBISBox.getIrregIVS  (curBox);
    notRegular |= curEBISBox.getMultiCells(curBox);

    BaseIVFAB<VoFStencil>& curStencilBaseIVFAB = m_opStencil[dit()];
    BaseIVFAB<Real>&       curWeightBaseIVFAB  = m_weight[dit()];

    curStencilBaseIVFAB.define(notRegular,curEBGraph,nComps);
    curWeightBaseIVFAB.define(notRegular,curEBGraph,nComps);

    for (VoFIterator vofit(notRegular,curEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();

      VoFStencil& curStencil = curStencilBaseIVFAB(vof,0);
      getOpVoFStencil(curStencil,curEBISBox,vof);

      Real& curWeight  = curWeightBaseIVFAB(vof,0);
      curWeight = 0.0;

      for (int i = 0; i < curStencil.size(); i++)
      {
        VolIndex curVoF = curStencil.vof(i);

        if (curVoF == vof)
        {
          curWeight = curStencil.weight(i) / (m_dx[0]*m_dx[0]);
          break;
        }
      }

      const IntVect& iv = vof.gridIndex();

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        Box loSide = bdryLo(m_domain,idir);
        loSide.shiftHalf(idir,1);
        if (loSide.contains(iv))
        {
          Real faceAreaFrac = 0.0;
          Vector<FaceIndex> faces = curEBISBox.getFaces(vof,idir,Side::Lo);
          for (int i = 0; i < faces.size(); i++)
          {
            faceAreaFrac += curEBISBox.areaFrac(faces[i]);
          }

          curWeight += -1.0 * faceAreaFrac / (m_dx[0]*m_dx[0]);
        }

        Box hiSide = bdryHi(m_domain,idir);
        hiSide.shiftHalf(idir,-1);
        if (hiSide.contains(iv))
        {
          Real faceAreaFrac = 0.0;
          Vector<FaceIndex> faces = curEBISBox.getFaces(vof,idir,Side::Hi);
          for (int i = 0; i < faces.size(); i++)
          {
            faceAreaFrac += curEBISBox.areaFrac(faces[i]);
          }

          curWeight += -1.0 * faceAreaFrac / (m_dx[0]*m_dx[0]);
        }
      }

      if (curWeight == 0.0)
      {
        curWeight = -1.0;
      }
    }
  }

  m_isDefined = true;
}

// Linear operator functions

void EBAMRPoissonOp::residual(LevelData<EBCellFAB>&       a_residual,
                              const LevelData<EBCellFAB>& a_phi,
                              const LevelData<EBCellFAB>& a_rhs,
                              bool                        a_homogeneous)
{
  assert(m_isDefined);

  residual(a_residual,a_phi,a_rhs,a_homogeneous,m_noParity);
}

void EBAMRPoissonOp::residual(LevelData<EBCellFAB>&       a_residual,
                              const LevelData<EBCellFAB>& a_phi,
                              const LevelData<EBCellFAB>& a_rhs,
                              bool                        a_homogeneous,
                              const IntVect&              a_parity)
{
  assert(m_isDefined);

  assert(a_residual.nComp() == 1);
  assert(a_phi.nComp() == 1);
  assert(a_rhs.nComp() == 1);

  if (a_parity == m_noParity)
  {
    applyOp(a_residual,a_phi,a_homogeneous);
    scale(a_residual,-1.0);
    incr(a_residual,a_rhs,1.0);
  }
  else
  {
    MayDay::Error("EBAMRPoissonOp::residual - parity not implemented");
  }
}

void EBAMRPoissonOp::preCond(LevelData<EBCellFAB>&       a_lhs,
                             const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  m_levelOps.assign(a_lhs,a_rhs);
}

void EBAMRPoissonOp::applyOp(LevelData<EBCellFAB>&       a_opPhi,
                             const LevelData<EBCellFAB>& a_phi,
                             bool                        a_homogeneous)
{
  assert(m_isDefined);

  assert(a_phi.nComp() == 1);
  assert(a_opPhi.nComp() == 1);

  LevelData<EBCellFAB>& phi = (LevelData<EBCellFAB>&)a_phi;

  int nComps = 1;
  int comp = 0;

  phi.exchange(phi.interval());

  Real invDx  = 1.0/m_dx[0];
  Real invDx2 = invDx*invDx;

  for (DataIterator dit = phi.dataIterator(); dit.ok(); ++dit)
  {
    const EBCellFAB& curPhiEBCellFAB = phi[dit()];
    Box curPhiBox = curPhiEBCellFAB.box();
    const FArrayBox& curPhiFAB = curPhiEBCellFAB.getFArrayBox();

    EBCellFAB& curOpPhiEBCellFAB = a_opPhi[dit()];
    Box curOpPhiBox = curOpPhiEBCellFAB.box();
    FArrayBox& curOpPhiFAB = curOpPhiEBCellFAB.getFArrayBox();

    curOpPhiEBCellFAB.setVal(0.0);

    Box loBox[SpaceDim],hiBox[SpaceDim];
    int hasLo[SpaceDim],hasHi[SpaceDim];

    for (int idir = 0; idir < SpaceDim; idir++)
    {
      Box centerBox,entireBox;

      ebEllipticLoHiCenter(loBox[idir],hasLo[idir],hiBox[idir],hasHi[idir],
                           centerBox,entireBox,curOpPhiBox,curPhiBox,
                           m_domain,idir);

      FArrayBox opPhiDirFAB(curOpPhiBox,nComps);

      opPhiDirFAB.setVal(0.0);

      FORT_REGGET1DLAPLACIAN(CHF_FRA1(opPhiDirFAB,comp),
                             CHF_CONST_FRA1(curPhiFAB,comp),
                             CHF_CONST_REALVECT(m_dx),
                             CHF_CONST_INT(idir),
                             CHF_BOX(loBox[idir]),
                             CHF_CONST_INT(hasLo[idir]),
                             CHF_BOX(hiBox[idir]),
                             CHF_CONST_INT(hasHi[idir]),
                             CHF_BOX(centerBox));

      if (hasLo[idir] == 1)
      {
        FArrayBox loFaceFlux(loBox[idir],nComps);
        int side = -1;

        assert(m_domainBC != NULL);
        m_domainBC->getFaceFlux(loFaceFlux,curPhiFAB,m_probLo,m_dx,
                                idir,Side::Lo,a_homogeneous);

        FORT_REGAPPLYDOMAINFLUX(CHF_FRA1(opPhiDirFAB,comp),
                                CHF_CONST_FRA1(loFaceFlux,comp),
                                CHF_CONST_REALVECT(m_dx),
                                CHF_CONST_INT(side),
                                CHF_BOX(loBox[idir]));
      }

      if (hasHi[idir] == 1)
      {
        FArrayBox hiFaceFlux(hiBox[idir],nComps);
        int side = 1;

        assert(m_domainBC != NULL);
        m_domainBC->getFaceFlux(hiFaceFlux,curPhiFAB,m_probLo,m_dx,
                                idir,Side::Hi,a_homogeneous);

        FORT_REGAPPLYDOMAINFLUX(CHF_FRA1(opPhiDirFAB,comp),
                                CHF_CONST_FRA1(hiFaceFlux,comp),
                                CHF_CONST_REALVECT(m_dx),
                                CHF_CONST_INT(side),
                                CHF_BOX(hiBox[idir]));
      }

      FORT_REGSUMLAPLACIAN(CHF_FRA1(curOpPhiFAB,comp),
                           CHF_CONST_FRA1(opPhiDirFAB,comp),
                           CHF_BOX(curOpPhiBox));
    }

    const EBISBox& curOpPhiEBISBox = curOpPhiEBCellFAB.getEBISBox();
    const EBGraph& curOpPhiEBGraph = curOpPhiEBISBox.getEBGraph();
    const BaseIVFAB<VoFStencil>& curBaseIVFAB = m_opStencil[dit()];

    const IntVectSet& curIrreg = curBaseIVFAB.getIVS();

    for (VoFIterator vofit(curIrreg,curOpPhiEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();

      int comp = 0;

      const VoFStencil& curStencil = curBaseIVFAB(vof,comp);
      curOpPhiEBCellFAB(vof,comp) = 0.0;

      for (int i = 0; i < curStencil.size(); i++)
      {
        VolIndex curVoF = curStencil.vof(i);
        Real weight = curStencil.weight(i);

        curOpPhiEBCellFAB(vof,comp) += weight * curPhiEBCellFAB(curVoF,comp);
      }

      curOpPhiEBCellFAB(vof,comp) *= invDx2;

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        if (hasLo[idir] == 1)
        {
          if (loBox[idir].contains(iv))
          {
            Real flux;
            m_domainBC->getFaceFlux(flux,vof,curPhiEBCellFAB,
                                    m_probLo,m_dx,idir,Side::Lo,a_homogeneous);

            curOpPhiEBCellFAB(vof,comp) -= flux * invDx;
          }
        }

        if (hasHi[idir] == 1)
        {
          if (hiBox[idir].contains(iv))
          {
            Real flux;
            m_domainBC->getFaceFlux(flux,vof,curPhiEBCellFAB,
                                    m_probLo,m_dx,idir,Side::Hi,a_homogeneous);

            curOpPhiEBCellFAB(vof,comp) += flux * invDx;
          }
        }
      }

      Real ebFlux;
      m_ebBC->getEBFlux(ebFlux,vof,phi,dit(),m_probLo,m_dx,a_homogeneous);

      curOpPhiEBCellFAB(vof,comp) += ebFlux * invDx;
    }
  }
}

void EBAMRPoissonOp::create(LevelData<EBCellFAB>&       a_lhs,
                            const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  m_levelOps.create(a_lhs,a_rhs);
}

void EBAMRPoissonOp::assign(LevelData<EBCellFAB>&       a_lhs,
                            const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  m_levelOps.assign(a_lhs,a_rhs);
}

Real EBAMRPoissonOp::dotProduct(const LevelData<EBCellFAB>& a_1,
                                const LevelData<EBCellFAB>& a_2)
{
  assert(m_isDefined);

  ProblemDomain domain;
  Real volume;

  return m_levelOps.kappaDotProduct(volume,a_1,a_2,EBLEVELDATAOPS_ALLVOFS,domain);
}

void EBAMRPoissonOp::incr(LevelData<EBCellFAB>&       a_lhs,
                          const LevelData<EBCellFAB>& a_x,
                          Real                        a_scale)
{
  assert(m_isDefined);

  m_levelOps.incr(a_lhs,a_x,a_scale);
}

void EBAMRPoissonOp::axby(LevelData<EBCellFAB>&       a_lhs,
                          const LevelData<EBCellFAB>& a_x,
                          const LevelData<EBCellFAB>& a_y,
                          Real                        a_a,
                          Real                        a_b)
{
  assert(m_isDefined);

  m_levelOps.axby(a_lhs,a_x,a_y,a_a,a_b);
}

void EBAMRPoissonOp::scale(LevelData<EBCellFAB>& a_lhs,
                           const Real&           a_scale)
{
  assert(m_isDefined);

  m_levelOps.scale(a_lhs,a_scale);
}

Real EBAMRPoissonOp::norm(const LevelData<EBCellFAB>& a_rhs,
                          int                         a_ord)
{
  assert(m_isDefined);

  ProblemDomain domain;
  Real volume;

  return m_levelOps.kappaNorm(volume,a_rhs,EBLEVELDATAOPS_ALLVOFS,domain,a_ord);
}

void EBAMRPoissonOp::setToZero(LevelData<EBCellFAB>& a_lhs)
{
  assert(m_isDefined);

  m_levelOps.setToZero(a_lhs);
}

// MGLevelOp functions

void EBAMRPoissonOp::createCoarser(LevelData<EBCellFAB>&       a_coarse,
                                   const LevelData<EBCellFAB>& a_fine,
                                   bool                        a_ghosted)
{
  assert(m_isDefined);

  assert(a_fine.nComp() == 1);
  assert(m_factoryCoarse != NULL);

  if (m_coarseFlag)
  {
    DisjointBoxLayout dbl;

    // This is multigrid, so coarsen by 2
    assert(a_fine.disjointBoxLayout().coarsenable(2));
    coarsen(dbl,a_fine.disjointBoxLayout(),2);

    a_coarse.define(dbl,a_fine.nComp(),a_fine.ghostVect(),*m_factoryCoarse);
  }
  else
  {
    a_coarse.define(m_dblCoarse,a_fine.nComp(),a_fine.ghostVect(),*m_factoryCoarse);
  }
}

void EBAMRPoissonOp::relax(LevelData<EBCellFAB>&       a_e,
                           const LevelData<EBCellFAB>& a_residual,
                           int                         a_iterations)
{
  assert(m_isDefined);

  assert(a_e.nComp() == 1);
  assert(a_residual.nComp() == 1);

  if (m_relaxType == 0)
  {
    for (int i = 0; i < a_iterations; i++)
    {
      levelJacobi(a_e,a_residual);
    }
  }
  else
  if (m_relaxType == 1)
  {
    for (int i = 0; i < a_iterations; i++)
    {
      levelMulticolorGS(a_e,a_residual);
    }
  }
  else
  {
    MayDay::Error("EBAMRPoissonOp::relax - invalid relaxation type");
  }
}

void EBAMRPoissonOp::restrictResidual(LevelData<EBCellFAB>&       a_resCoarse,
                                      LevelData<EBCellFAB>&       a_phiFine,
                                      const LevelData<EBCellFAB>& a_rhsFine)
{
  assert(m_isDefined);

  assert(a_resCoarse.nComp() == 1);
  assert(a_phiFine.nComp() == 1);
  assert(a_rhsFine.nComp() == 1);

  LevelData<EBCellFAB> resFine;
  bool homogeneous = true;

  m_levelOps.create(resFine,a_rhsFine);

  // Get the residual on the fine grid
  residual(resFine,a_phiFine,a_rhsFine,homogeneous,m_noParity);

  // Average the residual from the fine grid to the coarse grid
  if (m_coarseFlag)
  {
    average(a_resCoarse,resFine);
  }
  else
  {
    DisjointBoxLayout dblFine;
    refine(dblFine,a_resCoarse.disjointBoxLayout(),2);

    LevelData<EBCellFAB> resCopyFine;
    resCopyFine.define(dblFine,resFine.nComp(),resFine.ghostVect(),*m_factoryFine);

    Interval interval(0,0);
    resFine.copyTo(interval,resCopyFine,interval);

    average(a_resCoarse,resCopyFine);
  }
}

void EBAMRPoissonOp::prolongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                                      const LevelData<EBCellFAB>& a_correctCoarse)
{
  assert(m_isDefined);

  assert(a_phiThisLevel.nComp() == 1);
  assert(a_correctCoarse.nComp() == 1);

  if (m_coarseFlag)
  {
    privateProlongIncrement(a_phiThisLevel,a_correctCoarse);
  }
  else
  {
    DisjointBoxLayout dblFine;
    refine(dblFine,a_correctCoarse.disjointBoxLayout(),2);

    LevelData<EBCellFAB> phiCopy;
    phiCopy.define(dblFine,a_phiThisLevel.nComp(),a_phiThisLevel.ghostVect(),*m_factoryFine);

    Interval interval(0,0);
    a_phiThisLevel.copyTo(interval,phiCopy,interval);

    privateProlongIncrement(phiCopy,a_correctCoarse);

    phiCopy.copyTo(interval,a_phiThisLevel,interval);
  }
}

// AMRLevelOp functions

int EBAMRPoissonOp::refToCoarser()
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::refToCoarser not implemented");

  return 1;
}

void EBAMRPoissonOp::AMRRelaxLevel(LevelData<EBCellFAB>& a_correction,
                                   const LevelData<EBCellFAB>& a_res)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRRelaxLevel not implemented");
}

void EBAMRPoissonOp::AMRResidual(LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_phiFine,
                           const LevelData<EBCellFAB>& a_phi,
                           const LevelData<EBCellFAB>& a_phiCoarse,
                           const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRRedisual not implemented");
}

void EBAMRPoissonOp::AMRResidualNF(LevelData<EBCellFAB>& a_residual,
                             const LevelData<EBCellFAB>& a_phi,
                             const LevelData<EBCellFAB>& a_phiCoarse,
                             const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRRedisualNF not implemented");
}

void EBAMRPoissonOp::AMRRestrict(LevelData<EBCellFAB>& a_resCoarse,
                           const LevelData<EBCellFAB>& a_residual,
                           const LevelData<EBCellFAB>& a_correction,
                           const LevelData<EBCellFAB>& a_coarseCorrection)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRRestrict not implemented");
}

void EBAMRPoissonOp::AMRProlong(LevelData<EBCellFAB>&       a_correction,
                          const LevelData<EBCellFAB>& a_coarseCorrection)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRProlong not implemented");
}

void EBAMRPoissonOp::AMRUpdateResidual(LevelData<EBCellFAB>&       a_residual,
                                 const LevelData<EBCellFAB>& a_correction,
                                 const LevelData<EBCellFAB>& a_coarseCorrection)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOp::AMRUpdateRedisual not implemented");
}

// Protected functions

void EBAMRPoissonOp::getOpVoFStencil(VoFStencil&     a_stencil,
                                     const EBISBox&  a_ebisbox,
                                     const VolIndex& a_vof)
{
  a_stencil.clear();

  Vector<VolIndex> allMonotoneVoFs;
  int radius = 1;
  IntVect timesMoved = IntVect::TheZeroVector();
  IntVect pathSign   = IntVect::TheZeroVector();

  EBArith::getAllVoFsInMonotonePath(allMonotoneVoFs,timesMoved,pathSign,
                                    a_vof,a_ebisbox,radius);

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    for (SideIterator sit; sit.ok(); ++sit)
    {
      Side::LoHiSide side = sit();
      Vector<FaceIndex> faces;

      faces = a_ebisbox.getFaces(a_vof,idir,side);

      for (int iface = 0; iface < faces.size(); iface++)
      {
        FaceIndex face = faces[iface];
        VoFStencil faceStencil;

        getOpFaceStencil(faceStencil,allMonotoneVoFs,a_ebisbox,a_vof,
                         idir,side,face);

        a_stencil += faceStencil;
      }
    }
  }
}

void EBAMRPoissonOp::getOpFaceStencil(VoFStencil&             a_stencil,
                                      const Vector<VolIndex>& a_allMonotoneVoFs,
                                      const EBISBox&          a_ebisbox,
                                      const VolIndex&         a_vof,
                                      int                     a_dir,
                                      const Side::LoHiSide&   a_side,
                                      const FaceIndex&        a_face)
{
  a_stencil.clear();

  const RealVect& faceCentroid = a_ebisbox.centroid(a_face);
  IntVect dirs = IntVect::TheZeroVector();

  for (int idir = 0; idir < SpaceDim; idir++)
  {
    if (idir != a_dir)
    {
      if (faceCentroid[idir] < 0.0)
      {
        dirs[idir] = -1;
      }
      else
      {
        dirs[idir] = 1;
      }
    }
  }

  bool orderOne = true;

  if (m_orderEB == 0)
  {
    orderOne = false;
  }
  else
  {
    IntVect loVect = dirs;
    loVect.min(IntVect::TheZeroVector());

    IntVect hiVect = dirs;
    hiVect.max(IntVect::TheZeroVector());

    IntVect origin = a_vof.gridIndex();

    Box fluxBox(loVect,hiVect);
    IntVectSet fluxIVS(fluxBox);

    IVSIterator ivsit(fluxIVS);
    for (ivsit.begin(); ivsit.ok(); ++ivsit)
    {
      bool curOkay;

      IntVect ivDelta = ivsit();

      IntVect iv1 = ivDelta;
      iv1 += origin;
      
      VolIndex vof1;

      curOkay = EBArith::isVoFHere(vof1,a_allMonotoneVoFs,iv1);

      IntVect iv2 = iv1;
      iv2[a_dir] += sign(a_side);

      VolIndex vof2;

      curOkay = curOkay && EBArith::isVoFHere(vof2,a_allMonotoneVoFs,iv2);

      orderOne = orderOne && curOkay;

      if (curOkay)
      {
        VoFStencil curPair;

        curPair.add(vof1,-1.0);
        curPair.add(vof2, 1.0);

        for (int idir = 0; idir < SpaceDim; idir++)
        {
          if (idir != a_dir)
          {
            if (ivDelta[idir] == 0)
            {
              curPair *= 1 - abs(faceCentroid[idir]);
            }
            else
            {
              curPair *= abs(faceCentroid[idir]);
            }
          }
        }

        a_stencil += curPair;
      }
    }
  }

  if (!orderOne)
  {
    VolIndex vof1 = a_vof;
    VolIndex vof2 = a_face.getVoF(a_side);

    if (a_ebisbox.getRegion().contains(vof2.gridIndex()))
    {
      a_stencil.clear();

      a_stencil.add(vof1,-1.0);
      a_stencil.add(vof2, 1.0);
    }
  }

  a_stencil *= a_ebisbox.areaFrac(a_face);
}

void EBAMRPoissonOp::levelMulticolorGS(LevelData<EBCellFAB>&       a_phi,
                                       const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  int comp = 0;

  LevelData<EBCellFAB> resid;
  bool homogeneous = true;

  m_levelOps.create(resid,a_rhs);

  IntVect color = IntVect::TheZeroVector();
  IntVect limit = IntVect::TheUnitVector();

  const DisjointBoxLayout& dblPhi = a_phi.disjointBoxLayout();

  // Loop over all possibilities (in all dimensions)
  while (color[SpaceDim-1] <= limit[SpaceDim-1])
  {
    // Get the residual everywhere
    residual(resid,a_phi,a_rhs,homogeneous,m_noParity);

    // Relax the current color/parity
    for (DataIterator dit = a_phi.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& curPhiEBCellFAB = a_phi[dit()];

      const EBISBox& curPhiEBISBox = curPhiEBCellFAB.getEBISBox();
      const Box& curPhiBox         = dblPhi[dit()];
      const int& curPhiNComp       = curPhiEBCellFAB.nComp();

      EBCellFAB curSavePhiEBCellFAB(curPhiEBISBox,
                                    curPhiBox,
                                    curPhiNComp);
      curSavePhiEBCellFAB.copy(curPhiEBCellFAB);

      const EBCellFAB& curResidEBCellFAB = resid[dit()];

      // Do the regular cells
      FArrayBox& curPhiFAB     = curPhiEBCellFAB.getFArrayBox();
      FArrayBox& curSavePhiFAB = curSavePhiEBCellFAB.getFArrayBox();
      const FArrayBox& curResidFAB = curResidEBCellFAB.getFArrayBox();

      Real weight = 0.0;

      IntVect loBox = curPhiBox.smallEnd();
      IntVect hiBox = curPhiBox.bigEnd();

      for (int idir = 0; idir < SpaceDim; idir++)
      {
        weight += -2.0 / (m_dx[idir]*m_dx[idir]);

        if (loBox[idir] % 2 != color[idir])
        {
          loBox[idir]++;
        }
      }

      weight = 1.0 / weight;

      if (loBox <= hiBox)
      {
        Box coloredBox(loBox,hiBox);

        FORT_REGMULTICOLORGS(CHF_FRA1(curPhiFAB,comp),
                             CHF_CONST_FRA1(curSavePhiFAB,comp),
                             CHF_CONST_REAL(weight),
                             CHF_CONST_FRA1(curResidFAB,comp),
                             CHF_BOX(coloredBox));
      }

      // Do the irregular cells
      const EBGraph& curPhiEBGraph = curPhiEBISBox.getEBGraph();
      const BaseIVFAB<Real>& curWeightBaseIVFAB = m_weight[dit()];

      const IntVectSet& curIrreg = curWeightBaseIVFAB.getIVS();

      for (VoFIterator vofit(curIrreg,curPhiEBGraph); vofit.ok(); ++vofit)
      {
        const VolIndex& vof = vofit();
        const IntVect& iv = vof.gridIndex();

        bool ok = true;
        for (int idir = 0; idir < SpaceDim; idir++)
        {
          if (iv[idir] % 2 != color[idir])
          {
            ok = false;
            break;
          }
        }

        if (ok)
        {
          const Real& weight = curWeightBaseIVFAB(vof,comp);

          curPhiEBCellFAB(vof,comp) = curSavePhiEBCellFAB(vof,comp)
                                    + 1.0/weight * curResidEBCellFAB(vof,comp);
        }
      }
    }

    color[0]++;

    // If greater that the maximum it can be...
    if (color[0] > limit[0])
    {
      // For all dimensions...
      for (int i = 0; i < SpaceDim-1; i++)
      {
        color[i] = 0;
        color[i+1]++;

        // If the next dimension's power is okay go on
        if (color[i+1] <= limit[i+1])
        {
          break;
        }
      }
    }
  }
}

void EBAMRPoissonOp::levelJacobi(LevelData<EBCellFAB>&       a_phi,
                                 const LevelData<EBCellFAB>& a_rhs)
{
  assert(m_isDefined);

  LevelData<EBCellFAB> resid;
  bool homogeneous = true;

  m_levelOps.create(resid,a_rhs);

  // Get the residual
  residual(resid,a_phi,a_rhs,homogeneous,m_noParity);

  // Create the weight
  Real weight = 0;
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    weight += -2.0 / (m_dx[idir]*m_dx[idir]);
  }
  // Do the Jacobi relaxation
  Real one = 1.0;
  axby(a_phi,a_phi,resid,one,0.5/weight);
}

void EBAMRPoissonOp::average(LevelData<EBCellFAB>&       a_coarse,
                             const LevelData<EBCellFAB>& a_fine)
{
  assert(m_isDefined);

  assert(a_coarse.nComp() == 1);
  assert(a_fine.nComp() == 1);

  int comp = 0;

  Real invNumFinePerCoarse = 1.0 / pow(2.0,SpaceDim);

  const DisjointBoxLayout& dblFine   = a_fine.disjointBoxLayout();
  const DisjointBoxLayout& dblCoarse = a_coarse.disjointBoxLayout();

  for (DataIterator ditCoarse = a_coarse.dataIterator(); ditCoarse.ok(); ++ditCoarse)
  {
    EBCellFAB& curCoarseEBCellFAB = a_coarse[ditCoarse()];
    FArrayBox& curCoarseFAB = curCoarseEBCellFAB.getFArrayBox();
    const EBISBox& curCoarseEBISBox = curCoarseEBCellFAB.getEBISBox();
    const Box& curCoarseBox = dblCoarse[ditCoarse()];

    const EBCellFAB& curFineEBCellFAB = a_fine[ditCoarse()];
    const FArrayBox& curFineFAB = curFineEBCellFAB.getFArrayBox();
    const Box& curFineBox = dblFine[ditCoarse()];

    curCoarseFAB.setVal(0.0);

    FORT_REGAVERAGE(CHF_FRA1(curCoarseFAB,comp),
                    CHF_CONST_FRA1(curFineFAB,comp),
                    CHF_BOX(curFineBox));

    const EBGraph& curCoarseEBGraph  = curCoarseEBISBox.getEBGraph();
    const IntVectSet& curCoarseIrreg = curCoarseEBISBox.getIrregIVS(curCoarseBox);

    for (VoFIterator vofit(curCoarseIrreg,curCoarseEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& coarseVof = vofit();

      curCoarseEBCellFAB(coarseVof,comp) = 0.0;

      Vector<VolIndex> fineVofs = curCoarseEBISBox.refine(coarseVof);
      assert(fineVofs.size() > 0);
      for (int i = 0; i < fineVofs.size(); i++)
      {
        const VolIndex& fineVof = fineVofs[i];
        curCoarseEBCellFAB(coarseVof,comp) += invNumFinePerCoarse
                                            * curFineEBCellFAB(fineVof,comp);
      }
    }
  }
}

void EBAMRPoissonOp::privateProlongIncrement(LevelData<EBCellFAB>&       a_phiThisLevel,
                                             const LevelData<EBCellFAB>& a_correctCoarse)
{
  assert(m_isDefined);

  assert(a_phiThisLevel.nComp() == 1);
  assert(a_correctCoarse.nComp() == 1);

  int comp = 0;

  const DisjointBoxLayout& dblPhiThisLevel  = a_phiThisLevel.disjointBoxLayout();
  const DisjointBoxLayout& dblCorrectCoarse = a_correctCoarse.disjointBoxLayout();

  for (DataIterator dit = a_phiThisLevel.dataIterator(); dit.ok(); ++dit)
  {
    EBCellFAB& curPhiThisLevelEBCellFAB   = a_phiThisLevel[dit()];

    const EBISBox& curPhiThisLevelEBISBox = curPhiThisLevelEBCellFAB.getEBISBox();
    const Box& curPhiThisLevelBox         = dblPhiThisLevel[dit()];
    const int& curPhiThisLevelNComp       = curPhiThisLevelEBCellFAB.nComp();

    EBCellFAB curSavePhiThisLevelEBCellFAB(curPhiThisLevelEBISBox,
                                           curPhiThisLevelBox,
                                           curPhiThisLevelNComp);
    curSavePhiThisLevelEBCellFAB.copy(curPhiThisLevelEBCellFAB);

    const EBCellFAB& curCorrectCoarseEBCellFAB = a_correctCoarse[dit()];

    const EBISBox& curCorrectCoarseEBISBox     = curCorrectCoarseEBCellFAB.getEBISBox();
    const Box& curCorrectCoarseBox             = dblCorrectCoarse[dit()];

    assert(curCorrectCoarseBox == coarsen(curPhiThisLevelBox,2));

    FArrayBox& curPhiThisLevelFAB        = curPhiThisLevelEBCellFAB.getFArrayBox();
    const FArrayBox& curCorrectCoarseFAB = curCorrectCoarseEBCellFAB.getFArrayBox();

    FORT_REGPROLONG(CHF_FRA1(curPhiThisLevelFAB,comp),
                    CHF_CONST_FRA1(curCorrectCoarseFAB,comp),
                    CHF_BOX(curPhiThisLevelBox));

    const EBGraph& curCorrectCoarseEBGraph  = curCorrectCoarseEBISBox.getEBGraph();
    const IntVectSet& curCorrectCoarseIrreg = curCorrectCoarseEBISBox.getIrregIVS(curCorrectCoarseBox);

    for (VoFIterator vofit(curCorrectCoarseIrreg,curCorrectCoarseEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& coarseVof = vofit();

      Vector<VolIndex> fineVofs = curCorrectCoarseEBISBox.refine(coarseVof);

      for (int i = 0; i < fineVofs.size(); i++)
      {
        const VolIndex& fineVof = fineVofs[i];

        curPhiThisLevelEBCellFAB(fineVof,comp) =
            curSavePhiThisLevelEBCellFAB(fineVof  ,comp)
          + curCorrectCoarseEBCellFAB   (coarseVof,comp);
      }
    }
  }
}

EBAMRPoissonOpFactory::EBAMRPoissonOpFactory()
{
  m_isDefined = false;
}

EBAMRPoissonOpFactory::~EBAMRPoissonOpFactory()
{
}

void
EBAMRPoissonOpFactory::define(const ProblemDomain&           a_domain,
                              const DisjointBoxLayout&       a_dbl,
                              const EBIndexSpace*            a_space,
                              const int&                     a_numGhostCells,
                              const int&                     a_numLevels,
                              const RealVect&                a_probLo,
                              const RealVect&                a_dx,
                              const int&                     a_orderEB,
                              const int&                     a_relaxType,
                              RefCountedPtr<BaseDomainBC>    a_domainBC,
                              RefCountedPtr<BaseEBBCFactory> a_ebBcFactory)
{
  assert(!m_isDefined);

  m_probLo = a_probLo;

  m_orderEB = a_orderEB;
  m_relaxType = a_relaxType;

  m_domainBC    = a_domainBC;
  m_ebBcFactory = a_ebBcFactory;

  {
    m_coarseFlag.resize(0);

    m_domain.resize(0);
    m_domain.push_back(a_domain);

    m_dbl.push_back(a_dbl);

    EBISLayout layout;
    a_space->fillEBISLayout(layout,a_dbl,a_domain.domainBox(),a_numGhostCells);

    m_layout.resize(0);
    m_layout.push_back(layout);

    RefCountedPtr<DataFactory<EBCellFAB> > factory(new EBCellFactory(layout));
    m_factory.resize(0);
    m_factory.push_back(factory);

    m_factoryFine.push_back(NULL);

    m_dx.resize(0);
    m_dx.push_back(a_dx);
  }

  m_levelOps.define(m_factory[0]);

  int i;
  for (i = 1; i < a_numLevels; i++)
  {
    if (!m_layout[i-1].getDisjointLayout().coarsenable(2))
    {
      break;
    }

    m_coarseFlag.push_back(1);

    ProblemDomain domain = m_domain[i-1];
    domain.coarsen(2);

    m_domain.push_back(domain);

    DisjointBoxLayout dbl;
    coarsen_dbl(dbl,m_layout[i-1].getDisjointLayout(),2);

    m_dbl.push_back(dbl);

    EBISLayout layout;
    a_space->fillEBISLayout(layout,dbl,domain.domainBox(),a_numGhostCells);
    m_layout.push_back(layout);

    RefCountedPtr<DataFactory<EBCellFAB> > factory(new EBCellFactory(layout));
    m_factory.push_back(factory);

    m_factoryFine.push_back(m_factory[i-1]);

    RealVect dx = m_dx[i-1];
    dx *= 2;
    m_dx.push_back(dx);
  }

  if (i < a_numLevels)
  {
    {
      m_coarseFlag.push_back(0);

      ProblemDomain domain = m_domain[i-1];
      domain.coarsen(2);

      m_domain.push_back(domain);

      Vector<Box> boxes;
      boxes.push_back(domain.domainBox());

      Vector<int> procIDs(boxes.size());
      LoadBalance(procIDs,boxes);

      DisjointBoxLayout dbl(boxes,procIDs,domain);

      m_dbl.push_back(dbl);

      EBISLayout layout;
      a_space->fillEBISLayout(layout,dbl,domain.domainBox(),a_numGhostCells);
      m_layout.push_back(layout);

      RefCountedPtr<DataFactory<EBCellFAB> > factory(new EBCellFactory(layout));
      m_factory.push_back(factory);

      DisjointBoxLayout dblFine;
      refine(dblFine,dbl,2);

      EBISLayout layoutFine;
      a_space->fillEBISLayout(layoutFine,dblFine,m_domain[i-1].domainBox(),a_numGhostCells);

      RefCountedPtr<DataFactory<EBCellFAB> > factoryFine(new EBCellFactory(layoutFine));
      m_factoryFine.push_back(factoryFine);

      RealVect dx = m_dx[i-1];
      dx *= 2;
      m_dx.push_back(dx);

      for (i++; i < a_numLevels; i++)
      {
        if (!m_layout[i-1].getDisjointLayout().coarsenable(2))
        {
          break;
        }

        m_coarseFlag.push_back(1);

        ProblemDomain domain = m_domain[i-1];
        domain.coarsen(2);

        m_domain.push_back(domain);

        DisjointBoxLayout dbl;
        coarsen_dbl(dbl,m_layout[i-1].getDisjointLayout(),2);

        m_dbl.push_back(dbl);
      
        EBISLayout layout;
        a_space->fillEBISLayout(layout,dbl,domain.domainBox(),a_numGhostCells);
        m_layout.push_back(layout);

        RefCountedPtr<DataFactory<EBCellFAB> > factory(new EBCellFactory(layout));
        m_factory.push_back(factory);

        m_factoryFine.push_back(m_factory[i-1]);

        RealVect dx = m_dx[i-1];
        dx *= 2;
        m_dx.push_back(dx);
      }
    }
  }

  m_numLevels = a_numLevels;
  m_isDefined = true;
}

EBAMRPoissonOp*
  EBAMRPoissonOpFactory::MGnewOp(const ProblemDomain& a_FineindexSpace,
                                 int                  a_depth,
                                 bool                 a_homoOnly)
{
  assert(m_isDefined);
  assert(a_depth >= 0);
  assert(a_depth <  m_numLevels);

  EBAMRPoissonOp* op = new EBAMRPoissonOp();

  RefCountedPtr<BaseEBBC> ebBC = m_ebBcFactory->newEBBC(m_domain[a_depth],
                                                        m_layout[a_depth],
                                                        m_dx[a_depth]);

  bool coarseFlag;
  if (a_depth < m_numLevels-1)
  {
    coarseFlag = (m_coarseFlag[a_depth] != 0);

    op->define(m_domain[a_depth],m_layout[a_depth],m_probLo,m_dx[a_depth],
               m_orderEB,m_domainBC,ebBC,m_factory[a_depth],
               m_relaxType,m_dbl[a_depth+1],
               m_factory[a_depth+1],m_factoryFine[a_depth+1],coarseFlag);
  }
  else
  if (a_depth == m_numLevels-1)
  {
    coarseFlag = false;
    DisjointBoxLayout dblNull;

    op->define(m_domain[a_depth],m_layout[a_depth],m_probLo,m_dx[a_depth],
               m_orderEB,m_domainBC,ebBC,m_factory[a_depth],
               m_relaxType,dblNull,NULL,NULL,coarseFlag);
  }
  else
  {
    MayDay::Error("EBAMRPoissonOpFactory::MGnewOp - depth greater than number of levels");
  }

  return op;
}

void
  EBAMRPoissonOpFactory::reclaim(MGLevelOp<LevelData<EBCellFAB> >* a_reclaim)
{
  assert(m_isDefined);

  delete a_reclaim;
}

EBAMRPoissonOp*
  EBAMRPoissonOpFactory::AMRnewOp(const ProblemDomain& a_FineindexSpace)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOpFactory::AMRnewOp not implemented");

  return NULL;
}

void
  EBAMRPoissonOpFactory::AMRreclaim(EBAMRPoissonOp* a_reclaim)
{
  assert(m_isDefined);

  MayDay::Error("EBAMRPoissonOpFactory::AMRreclaim not implemented");
}
