/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "SPMD.H"

#include "EBLevelDataOps.H"

#define ITER(a) for(DataIterator dit = a.dataIterator(); dit.ok(); ++dit) { \
                   DataIndex d = dit();
#define ENDFOR(a) }

void EBLevelDataOps::setVal(LevelData<EBCellFAB>& a_result,
                            Real                  a_value)
{
  ITER(a_result)
    EBCellFAB& result = a_result[d];

    result.setVal(a_value);
  ENDFOR(a_result);
}

void EBLevelDataOps::sum(LevelData<EBCellFAB>&       a_result,
                         const LevelData<EBCellFAB>& a_in1,
                         const LevelData<EBCellFAB>& a_in2)
{
  sum(a_result,a_in1,a_in2,m_noParity);
}

void EBLevelDataOps::sum(LevelData<EBCellFAB>&       a_result,
                         const LevelData<EBCellFAB>& a_in1,
                         const LevelData<EBCellFAB>& a_in2,
                         const IntVect&              a_parity)
{
  if (a_parity == m_noParity)
  {
    ITER(a_result)
      EBCellFAB& result = a_result[d];

      const EBCellFAB& in1 = a_in1[d];
      const EBCellFAB& in2 = a_in2[d];

      result.copy(in1);
      result += in2;
    ENDFOR(a_result);
  }
  else
  {
    MayDay::Error("EBLevelDataOps::sum - parity not implemented");
  }
}

void EBLevelDataOps::addConstant(LevelData<EBCellFAB>& a_data,
                                 Real                  a_constant)
{
  ITER(a_data)
    EBCellFAB& data = a_data[d];

    data += a_constant;
  ENDFOR(a_data);
}

void EBLevelDataOps::product(LevelData<EBCellFAB>&       a_result,
                             const LevelData<EBCellFAB>& a_in1,
                             const LevelData<EBCellFAB>& a_in2)
{
  product(a_result,a_in1,a_in2,m_noParity);
}

void EBLevelDataOps::product(LevelData<EBCellFAB>&       a_result,
                             const LevelData<EBCellFAB>& a_in1,
                             const LevelData<EBCellFAB>& a_in2,
                             const IntVect&              a_parity)
{
  if (a_parity == m_noParity)
  {
    ITER(a_result)
      EBCellFAB& result = a_result[d];

      const EBCellFAB& in1 = a_in1[d];
      const EBCellFAB& in2 = a_in2[d];

      result.copy(in1);
      result *= in2;
    ENDFOR(a_result);
  }
  else
  {
    MayDay::Error("EBLevelDataOps::product - parity not implemented");
  }
}

void EBLevelDataOps::kappaWeight(LevelData<EBCellFAB>& a_data)
{
  int nComp = a_data.nComp();

  ITER(a_data)
    EBCellFAB& data = a_data[d];

    const Box& dataBox = data.box();
    const EBISBox& dataEBISBox = data.getEBISBox();
    const EBGraph& dataEBGraph = dataEBISBox.getEBGraph();

    const IntVectSet& dataIrreg = dataEBISBox.getIrregIVS(dataBox);

    for (VoFIterator vofit(dataIrreg,dataEBGraph); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();

      for (int comp = 0; comp < nComp; comp++)
      {
        data(vof,comp) *= dataEBISBox.volFrac(vof);
      }
    }
  ENDFOR(a_data);
}

Real EBLevelDataOps::kappaNorm(Real&                       a_volume,
                               const LevelData<EBCellFAB>& a_data,
                               int                         a_which,
                               const ProblemDomain&        a_domain,
                               int                         a_p)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data)
    const EBCellFAB& data = a_data[d];

    Real cur;
    Real curVolume;

    cur = sumKappaPow(curVolume,data,a_which,a_domain,a_p);

    a_volume += curVolume;

    if (a_p == 0)
    {
      if (cur > accum)
      {
        accum = cur;
      }
    }
    else
    {
      accum += cur;
    }
  ENDFOR(a_data);

  if (a_p != 0)
  {
    if (a_volume > 0.0)
    {
      accum = accum / a_volume;
    }

    accum = pow(accum,1.0/a_p);
  }

  return accum;
}

Real EBLevelDataOps::noKappaNorm(Real&                       a_volume,
                                 const LevelData<EBCellFAB>& a_data,
                                 int                         a_which,
                                 const ProblemDomain&        a_domain,
                                 int                         a_p)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data)
    const EBCellFAB& data = a_data[d];

    Real cur;
    Real curVolume;

    cur = sumNoKappaPow(curVolume,data,a_which,a_domain,a_p);

    a_volume += curVolume;

    if (a_p == 0)
    {
      if (cur > accum)
      {
        accum = cur;
      }
    }
    else
    {
      accum += cur;
    }
  ENDFOR(a_data);

  if (a_p != 0)
  {
    if (a_volume > 0.0)
    {
      accum = accum / a_volume;
    }

    accum = pow(accum,1.0/a_p);
  }

  return accum;
}

Real EBLevelDataOps::kappaDotProduct(Real&                       a_volume,
                                     const LevelData<EBCellFAB>& a_data1,
                                     const LevelData<EBCellFAB>& a_data2,
                                     int                         a_which,
                                     const ProblemDomain&        a_domain)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data1)
    const EBCellFAB& data1 = a_data1[d];
    const EBCellFAB& data2 = a_data2[d];

    Real cur;
    Real curVolume;

    cur = sumKappaDotProduct(curVolume,data1,data2,a_which,a_domain);

    a_volume += curVolume;
    accum += cur;
  ENDFOR(a_data1);

  if (a_volume > 0.0)
  {
    accum = accum / a_volume;
  }

  return accum;
}

Real EBLevelDataOps::noKappaDotProduct(Real&                       a_volume,
                                       const LevelData<EBCellFAB>& a_data1,
                                       const LevelData<EBCellFAB>& a_data2,
                                       int                         a_which,
                                       const ProblemDomain&        a_domain)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data1)
    const EBCellFAB& data1 = a_data1[d];
    const EBCellFAB& data2 = a_data2[d];

    Real cur;
    Real curVolume;

    cur = sumNoKappaDotProduct(curVolume,data1,data2,a_which,a_domain);

    a_volume += curVolume;
    accum += cur;
  ENDFOR(a_data1);

  if (a_volume > 0.0)
  {
    accum = accum / a_volume;
  }

  return accum;
}

Real EBLevelDataOps::kappaSumLevel(Real&                       a_volume,
                                   const LevelData<EBCellFAB>& a_data,
                                   int                         a_which,
                                   const ProblemDomain&        a_domain)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data)
    const EBCellFAB& data = a_data[d];

    Real cur;
    Real curVolume;

    cur = sumKappa(curVolume,data,a_which,a_domain);

    a_volume += curVolume;
    accum += cur;
  ENDFOR(a_data);

  if (a_volume > 0.0)
  {
    accum = accum / a_volume;
  }

  return accum;
}

Real EBLevelDataOps::noKappaSumLevel(Real&                       a_volume,
                                     const LevelData<EBCellFAB>& a_data,
                                     int                         a_which,
                                     const ProblemDomain&        a_domain)
{
  Real accum = 0.0;

  a_volume = 0.0;

  ITER(a_data)
    const EBCellFAB& data = a_data[d];

    Real cur;
    Real curVolume;

    cur = sumNoKappa(curVolume,data,a_which,a_domain);

    a_volume += curVolume;
    accum += cur;
  ENDFOR(a_data);

  if (a_volume > 0.0)
  {
    accum = accum / a_volume;
  }

  return accum;
}

// Protected functions - for internal use

Real EBLevelDataOps::sumKappaPow(Real&                a_volume,
                                 const EBCellFAB&     a_data,
                                 int                  a_which,
                                 const ProblemDomain& a_domain,
                                 int                  a_p)
{
  assert(a_data.nComp() == 1);

  const Box& curBox = a_data.box();
  const EBISBox& curEBISBox = a_data.getEBISBox();
  int comp = 0;

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isRegular(iv))
    {
      Real val = Abs(a_data(vof,comp));

      a_volume += 1.0;

      if (a_p == 0)
      {
        sum = Max(sum,val);
      }
      else
      {
        Real p = a_p;
        Real integrand = pow(val,p);

        sum += integrand;
      }
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isIrregular(iv))
    {
      Real volfrac = curEBISBox.volFrac(vof);
      Real val     = volfrac * Abs(a_data(vof,comp));

      a_volume += volfrac;

      if (a_p == 0)
      {
        sum = Max(sum,val);
      }
      else
      {
        Real p = a_p;
        Real integrand = pow(val,p);

        sum += integrand;
      }
    }
  }

  return sum;
}

Real EBLevelDataOps::sumNoKappaPow(Real&                a_volume,
                                   const EBCellFAB&     a_data,
                                   int                  a_which,
                                   const ProblemDomain& a_domain,
                                   int                  a_p)
{
  assert(a_data.nComp() == 1);

  const Box& curBox = a_data.box();
  const EBISBox& curEBISBox = a_data.getEBISBox();
  int comp = 0;

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isRegular(iv))
    {
        Real val = Abs(a_data(vof,comp));

        a_volume += 1.0;

        if (a_p == 0)
        {
          sum = Max(sum,val);
        }
        else
        {
          Real p = a_p;
          Real integrand = pow(val,p);

          sum += integrand;
        }
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isIrregular(iv))
    {
      Real volfrac = curEBISBox.volFrac(vof);
      Real val     = Abs(a_data(vof,comp));

      a_volume += volfrac;

      if (a_p == 0)
      {
        sum = Max(sum,val);
      }
      else
      {
        Real p = a_p;
        Real integrand = pow(val,p);

        sum += integrand;
      }
    }
  }

  return sum;
}

Real EBLevelDataOps::sumKappaDotProduct(Real&                a_volume,
                                        const EBCellFAB&     a_data1,
                                        const EBCellFAB&     a_data2,
                                        int                  a_which,
                                        const ProblemDomain& a_domain)
{
  assert(a_data1.nComp() == 1);
  assert(a_data2.nComp() == 1);

  const Box& curBox1 = a_data1.box();
  const EBISBox& curEBISBox1 = a_data1.getEBISBox();
  int comp1 = 0;

  const Box& curBox2 = a_data2.box();
  int comp2 = 0;

  assert(curBox1 == curBox2);

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox1);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox1.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox1.isRegular(iv))
    {
      Real val1 = a_data1(vof,comp1);
      Real val2 = a_data2(vof,comp2);

      a_volume += 1.0;
      sum += val1*val2;
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox1);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox1.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox1.isIrregular(iv))
    {
      Real volfrac = curEBISBox1.volFrac(vof);
      Real val1    = volfrac * a_data1(vof,comp1);
      Real val2    = volfrac * a_data2(vof,comp2);

      a_volume += volfrac;
      sum += val1*val2;
    }
  }

  return sum;
}

Real EBLevelDataOps::sumNoKappaDotProduct(Real&                a_volume,
                                          const EBCellFAB&     a_data1,
                                          const EBCellFAB&     a_data2,
                                          int                  a_which,
                                          const ProblemDomain& a_domain)
{
  assert(a_data1.nComp() == 1);
  assert(a_data2.nComp() == 1);

  const Box& curBox1 = a_data1.box();
  const EBISBox& curEBISBox1 = a_data1.getEBISBox();
  int comp1 = 0;

  const Box& curBox2 = a_data2.box();
  int comp2 = 0;

  assert(curBox1 == curBox2);

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox1);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox1.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox1.isRegular(iv))
    {
      Real val1 = a_data1(vof,comp1);
      Real val2 = a_data2(vof,comp2);

      a_volume += 1.0;
      sum += val1*val2;
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox1;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox1);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox1.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox1.isIrregular(iv))
    {
      Real volfrac = curEBISBox1.volFrac(vof);
      Real val1    = a_data1(vof,comp1);
      Real val2    = a_data2(vof,comp2);

      a_volume += volfrac;
      sum += val1*val2;
    }
  }

  return sum;
}

Real EBLevelDataOps::sumKappa(Real&                a_volume,
                              const EBCellFAB&     a_data,
                              int                  a_which,
                              const ProblemDomain& a_domain)
{
  assert(a_data.nComp() == 1);

  const Box& curBox = a_data.box();
  const EBISBox& curEBISBox = a_data.getEBISBox();
  int comp = 0;

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isRegular(iv))
    {
      Real val = a_data(vof,comp);

      a_volume += 1.0;
      sum += val;
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isIrregular(iv))
    {
      Real volfrac = curEBISBox.volFrac(vof);
      Real val     = volfrac * a_data(vof,comp);

      a_volume += volfrac;
      sum += val;
    }
  }

  return sum;
}

Real EBLevelDataOps::sumNoKappa(Real&                a_volume,
                                const EBCellFAB&     a_data,
                                int                  a_which,
                                const ProblemDomain& a_domain)
{
  assert(a_data.nComp() == 1);

  const Box& curBox = a_data.box();
  const EBISBox& curEBISBox = a_data.getEBISBox();
  int comp = 0;

  Real sum = 0.0;

  a_volume = 0.0;

  IntVectSet regIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORREGVOFS) == EBLEVELDATAOPS_INTERIORREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    regIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYREGVOFS) == EBLEVELDATAOPS_BOUNDARYREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    regIvsRegion |= boundary;
  }

  for (VoFIterator vofit(regIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isRegular(iv))
    {
      Real val = a_data(vof,comp);

      a_volume += 1.0;
      sum += val;
    }
  }

  IntVectSet irregIvsRegion;

  if ((a_which & EBLEVELDATAOPS_INTERIORIRREGVOFS) == EBLEVELDATAOPS_INTERIORIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    irregIvsRegion |= interior;
  }

  if ((a_which & EBLEVELDATAOPS_BOUNDARYIRREGVOFS) == EBLEVELDATAOPS_BOUNDARYIRREGVOFS)
  {
    Box interior = curBox;

    interior.grow(1);
    interior &= a_domain;
    interior.grow(-1);

    IntVectSet boundary(curBox);
    boundary -= interior;

    irregIvsRegion |= boundary;
  }

  for (VoFIterator vofit(irregIvsRegion,curEBISBox.getEBGraph()); vofit.ok(); ++vofit)
  {
    const VolIndex& vof = vofit();
    const IntVect& iv = vof.gridIndex();

    if (curEBISBox.isIrregular(iv))
    {
      Real volfrac = curEBISBox.volFrac(vof);
      Real val     = a_data(vof,comp);

      a_volume += volfrac;
      sum += val;
    }
  }

  return sum;
}
