/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "BoxIterator.H"

#include "NeumannDomainBC.H"

NeumannDomainBC::NeumannDomainBC()
{
  m_value = 12345.6789;
  m_flux = NULL;

  m_onlyHomogeneous = true;
  m_isFunction = false;
}

NeumannDomainBC::~NeumannDomainBC()
{
}

void NeumannDomainBC::setValue(Real a_value)
{
  m_value = a_value;
  m_flux = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}

void NeumannDomainBC::setFunction(RefCountedPtr<BaseBCFlux> a_flux)
{
  m_value = 12345.6789;
  m_flux = a_flux;

  m_onlyHomogeneous = false;
  m_isFunction = true;
}

void NeumannDomainBC::getFaceFlux(FArrayBox&       a_faceFlux,
                                  const FArrayBox& a_phi,
                                  const RealVect&  a_probLo,
                                  const RealVect&  a_dx,
                                  int              a_idir,
                                  Side::LoHiSide   a_side,
                                  bool             a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);
  int comp = 0;

  const Box& box = a_faceFlux.box();

  int iside;

  if (a_side == Side::Lo)
  {
    iside = 1;
  }
  else
  {
    iside = -1;
  }

  if (a_useHomogeneous)
  {
    Real value = 0.0;

    a_faceFlux.setVal(iside * value);
  }
  else
  {
    if (m_isFunction)
    {
      BoxIterator bit(box);

      for (bit.begin(); bit.ok(); ++bit)
      {
        const IntVect& iv = bit();

        RealVect point = RealVect::Unit;

        point *= 0.5;
        point[a_idir] -= iside * 0.5;
        point += iv;

        point *= a_dx;
        point += a_probLo;

        RealVect normal = RealVect::Zero;
        normal[a_idir] = iside;

        a_faceFlux(iv,comp) = iside * m_flux->flux(point,normal);
      }
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("NeumannDomainBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      Real value = m_value;

      a_faceFlux.setVal(iside * value);
    }
  }
}

void NeumannDomainBC::getFaceFlux(Real&            a_faceFlux,
                                  const VolIndex&  a_vof,
                                  const EBCellFAB& a_phi,
                                  const RealVect&  a_probLo,
                                  const RealVect&  a_dx,
                                  int              a_idir,
                                  Side::LoHiSide   a_side,
                                  bool             a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);

  int iside;

  if (a_side == Side::Lo)
  {
    iside = 1;
  }
  else
  {
    iside = -1;
  }

  if (a_useHomogeneous)
  {
    Real value = 0.0;

    const EBISBox& ebisBox = a_phi.getEBISBox();

    Real faceAreaFrac = 0.0;

    Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
    for (int i = 0; i < faces.size(); i++)
    {
      faceAreaFrac += ebisBox.areaFrac(faces[i]);
    }

    a_faceFlux = iside * value * faceAreaFrac;
  }
  else
  {
    if (m_isFunction)
    {
      const EBISBox& ebisBox = a_phi.getEBISBox();

      Real totalMassFlux = 0.0;

      const IntVect& iv = a_vof.gridIndex();

      RealVect center = RealVect::Unit;

      center *= 0.5;
      center += iv;

      center *= a_dx;
      center += a_probLo;

      Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);

      for (int i = 0; i < faces.size(); i++)
      {
        const FaceIndex& curFace = faces[i];

        Real area = ebisBox.areaFrac(curFace);

        RealVect centroid = ebisBox.centroid(curFace);
        centroid *= a_dx;
        centroid += center;

        RealVect normal = RealVect::Zero;
        normal[a_idir] = iside;

        Real flux = m_flux->flux(centroid,normal);

        totalMassFlux += area*flux;
      }

      a_faceFlux = iside * totalMassFlux;
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("NeumannDomainBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      Real value = m_value;

      const EBISBox& ebisBox = a_phi.getEBISBox();

      Real faceAreaFrac = 0.0;

      Vector<FaceIndex> faces = ebisBox.getFaces(a_vof,a_idir,a_side);
      for (int i = 0; i < faces.size(); i++)
      {
        faceAreaFrac += ebisBox.areaFrac(faces[i]);
      }

      a_faceFlux = iside * value * faceAreaFrac;
    }
  }
}
