/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "BoxIterator.H"

#include "NeumannEBBC.H"

NeumannEBBC::NeumannEBBC(const ProblemDomain& a_domain,
                         const EBISLayout&    a_layout,
                         const RealVect&      a_dx)
{
  m_value = 12345.6789;
  m_flux = NULL;

  m_onlyHomogeneous = true;
  m_isFunction = false;
}

NeumannEBBC::~NeumannEBBC()
{
}

void NeumannEBBC::setValue(Real a_value)
{
  m_value = a_value;
  m_flux = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}

void NeumannEBBC::setFunction(RefCountedPtr<BaseBCFlux> a_flux)
{
  m_value = 12345.6789;
  m_flux = a_flux;

  m_onlyHomogeneous = false;
  m_isFunction = true;
}

void NeumannEBBC::getEBFlux(Real&                       a_flux,
                            const VolIndex&             a_vof,
                            const LevelData<EBCellFAB>& a_phi,
                            const DataIndex             a_dit,
                            const RealVect&             a_probLo,
                            const RealVect&             a_dx,
                            bool                        a_useHomogeneous)
{
  assert(a_phi.nComp() == 1);

  const EBCellFAB& curPhi     = a_phi[a_dit];
  const EBISBox&   curEBISBox = curPhi.getEBISBox();

  if (a_useHomogeneous)
  {
    a_flux = 0.0;
  }
  else
  {
    if (m_isFunction)
    {
      const IntVect& iv = a_vof.gridIndex();

      RealVect point = RealVect::Unit;
      point *= 0.5;
      point += iv;

      point += curEBISBox.bndryCentroid(a_vof);
      point *= a_dx;

      point += a_probLo;

      RealVect normal = curEBISBox.normal(a_vof);

      a_flux = -m_flux->flux(point,normal);
    }
    else
    {
      if (m_onlyHomogeneous)
      {
        MayDay::Error("NeumannEBBC::getFaceFlux called with undefined inhomogeneous BC");
      }

      a_flux = m_value;
    }
  }

  const Real& areaFrac = curEBISBox.bndryArea(a_vof);
  a_flux *= areaFrac;
}

NeumannEBBCFactory::NeumannEBBCFactory()
{
  m_value = 12345.6789;
  m_flux = NULL;

  m_onlyHomogeneous = true;
  m_isFunction = false;
}

NeumannEBBCFactory::~NeumannEBBCFactory()
{
}

void NeumannEBBCFactory::setValue(Real a_value)
{
  m_value = a_value;
  m_flux = NULL;

  m_onlyHomogeneous = false;
  m_isFunction = false;
}

void NeumannEBBCFactory::setFunction(RefCountedPtr<BaseBCFlux> a_flux)
{
  m_value = 12345.6789;
  m_flux = a_flux;

  m_onlyHomogeneous = false;
  m_isFunction = true;
}

NeumannEBBC* NeumannEBBCFactory::newEBBC(const ProblemDomain& a_domain,
                                         const EBISLayout&    a_layout,
                                         const RealVect&      a_dx)
{
  NeumannEBBC* fresh = new NeumannEBBC(a_domain,a_layout,a_dx);

  if (!m_onlyHomogeneous)
  {
    if (!m_isFunction)
    {
      fresh->setValue(m_value);
    }
    else
    {
      fresh->setFunction(m_flux);
    }
  }

  return fresh;
}
