/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtg 1-29-2004
#ifndef _EB_LEVEL_ADVECT_H
#define _EB_LEVEL_ADVECT_H

#include "EBCellFAB.H"
#include "EBQuadCFInterp.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "PiecewiseLinearFillPatch.H"
#include "EBPWLFillPatch.H"
#include "EBFluxRegister.H"
#include "ProblemDomain.H"
#include "EBPatchGodunovFactory.H"
#include "EBPatchAdvect.H"

/// 
/**
   Advection and diffusion on a level using 
 */
class EBLevelAdvect
{
public:
  /// Default constructor
  /**
     Object requires define(..) to be called before all other functions.
  */
  EBLevelAdvect();

  /// Destructor
  /**
     destroys all objects created by define(..). Passed in
     data references of define(..) are left alone.
  */
  ~EBLevelAdvect();

  /// Actual constructor.
  /**
     For the coarsest level, an empty DisjointBoxLayout
     is passed in for coaserDisjointBoxLayout.
     Inside the routine, we cast away const-ness on the data members
     for the assignment.  The arguments passed in are maintained const.
     (coding standards).  a_nRefine is the refinement ratio between
     this level and the next coarser level.  a_numGhosts is the number
     of ghost cells in each direction.
  */
  void define(const DisjointBoxLayout&           a_thisDBL,
              const DisjointBoxLayout&           a_coarDBL,
              const EBISLayout&                  a_thisEBISL,
              const EBISLayout&                  a_coarEBISL,
              const ProblemDomain&               a_DProblem,
              const int&                         a_nRefine,
              const Real&                        a_dx,
              const bool&                        a_hasCoarser,
              const bool&                        a_hasFiner,
              const EBPatchGodunovFactory* const a_patchGodunov,
              Real                               a_viscosity = 0.0,
              bool                               a_useDiffusion = false);


  /// 
  /**
     1.   Average advection velocity to cell centers to get unorm.
     2.   Do quadradic C/F interpolation on a_cellState   and compute
          the diffusion source term if necessary.
     3.   Do linear C/F interpolation in space and time
     4.   Extrapolate cellState to to faces using patchgodunov->extrapolatePrim
  */
  void 
  advectToFaces(LevelData< EBFluxFAB >&                         a_extrapState,
                LayoutData< Vector< BaseIVFAB<Real>* > >&       a_coveredPrimLo, 
                LayoutData< Vector< BaseIVFAB<Real>* > >&       a_coveredPrimHi, 
                const LayoutData< Vector< Vector<VolIndex> > >& a_coveredFaceLo, 
                const LayoutData< Vector< Vector<VolIndex> > >& a_coveredFaceHi, 
                const LayoutData< Vector< IntVectSet> >&        a_coveredSetsLo, 
                const LayoutData< Vector< IntVectSet> >&        a_coveredSetsHi, 
                const LevelData< EBCellFAB >&                   a_consState,
                const LevelData< EBCellFAB >&                   a_normalVel,
                const LevelData< EBFluxFAB >&                   a_advectionVel,
                const LevelData< EBCellFAB >*                   a_consStateCoarseOld,
                const LevelData< EBCellFAB >*                   a_consStateCoarseNew,
                const Real&                                     a_timeCoarseOld,
                const Real&                                     a_timeCoarseNew,
                const Real&                                     a_timeFine,
                const Real&                                     a_dt);

  void
  computeNormalVel(LevelData<EBCellFAB>&                          a_normalVel, 
                   const LevelData<EBFluxFAB>&                    a_advectionVel,
                   const LayoutData<Vector<BaseIVFAB<Real> * > >& a_coveredVeloLo, 
                   const LayoutData<Vector<BaseIVFAB<Real> * > >& a_coveredVeloHi, 
                   const LayoutData<Vector<Vector<VolIndex> > >&  a_coveredFaceLo, 
                   const LayoutData<Vector<Vector<VolIndex> > >&  a_coveredFaceHi) const;


  bool isDefined() const;
protected:

  bool              m_useDiffusion;
  bool              m_hasCoarser;
  bool              m_hasFiner;
  bool              m_isDefined;
  Real              m_dx;
  Real              m_viscosity;
  int               m_refRatCrse;
  int               m_nVar;
  int               m_nGhost;

  DisjointBoxLayout m_thisGrids;
  DisjointBoxLayout m_coarGrids;
  EBISLayout        m_thisEBISL;
  EBISLayout        m_coarEBISL;
  EBQuadCFInterp    m_quadCFInterp;
  EBPWLFillPatch    m_pwlFillPatch;
  EBPatchAdvect*    m_ebPatchAdvect;
  ProblemDomain     m_domain;

private:
  //disallowed for all the usual reasons
  void operator=(const EBLevelAdvect& a_input)
  {
    MayDay::Error("invalid operator");
  }
  EBLevelAdvect(const EBLevelAdvect& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
