/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "DebugOut.H"
#include "EBLevelAdvect.H"
#include "BaseIVFactory.H"
#include "BaseIFFactory.H"
#include "BaseIFFAB.H"
#include "EBFluxFAB.H"
#include "FaceIterator.H"
#include "REAL.H"
#include "EBCellFactory.H"
#include "FaceIterator.H"
#include "EBPatchAdvectF_F.H"
#include <cstdio>
#include "EBAMRIO.H"
/*****************************/
EBLevelAdvect::EBLevelAdvect()
{
  m_isDefined = false;
  m_ebPatchAdvect = NULL;
}
/*****************************/
bool
EBLevelAdvect::isDefined() const
{
  return m_isDefined;
}
/*****************************/
EBLevelAdvect::~EBLevelAdvect()
{
  if(m_ebPatchAdvect != NULL)
    {
      delete m_ebPatchAdvect;
    }
}
/*****************************/
void 
EBLevelAdvect::define(const DisjointBoxLayout&           a_thisDBL,
                      const DisjointBoxLayout&           a_coarDBL,
                      const EBISLayout&                  a_thisEBISL,
                      const EBISLayout&                  a_coarEBISL,
                      const ProblemDomain&               a_domain,
                      const int&                         a_nRefine,
                      const Real&                        a_dx,
                      const bool&                        a_hasCoarser,
                      const bool&                        a_hasFiner,
                      const EBPatchGodunovFactory* const a_patchGodunov,
                      Real                               a_viscosity,
                      bool                               a_useDiffusion)
{
  assert(a_dx > 0.0);
  assert(a_nRefine > 0);
  
  m_isDefined = true;
  m_useDiffusion = a_useDiffusion;
  m_viscosity    = a_viscosity;
  m_thisGrids    = a_thisDBL;
  m_thisEBISL    = a_thisEBISL;
  m_refRatCrse   = a_nRefine;
  m_dx           = a_dx;
  m_domain       = a_domain;
  m_hasCoarser   = a_hasCoarser;
  m_hasFiner     = a_hasFiner;
  if(m_hasCoarser)
    {
      m_coarGrids = a_coarDBL;
      m_coarEBISL = a_coarEBISL;
    }

  if(m_ebPatchAdvect != NULL)
    {
      delete m_ebPatchAdvect;
    }
  EBPatchGodunov* ebPatchGodunov = a_patchGodunov->create();
  m_ebPatchAdvect = dynamic_cast<EBPatchAdvect*>(ebPatchGodunov);
  m_nVar         =  m_ebPatchAdvect->numConserved();
  if(m_ebPatchAdvect == NULL)
    {
      MayDay::Error("problem in casting to patch advection class");
    }
  m_ebPatchAdvect->define(m_domain, m_dx);
  
  m_nGhost = 5;

  if(m_hasCoarser)
    {
      ProblemDomain domainCrse = coarsen(m_domain, m_refRatCrse);
      Box domainBoxCrse = domainCrse.domainBox();

      //patcher is defined with the number of conserved vars.
      m_pwlFillPatch.define(m_thisGrids, m_coarGrids,
                            m_thisEBISL, m_coarEBISL,
                            domainBoxCrse, m_refRatCrse, m_nVar, 
                            m_nGhost);
      if(m_useDiffusion)
        {
          m_quadCFInterp.define(m_thisGrids, m_coarGrids,
                                m_thisEBISL, m_coarEBISL,
                                domainBoxCrse, m_refRatCrse, m_nVar);
        }
    }

}
/*****************************/
void 
EBLevelAdvect::
advectToFaces(LevelData< EBFluxFAB >&                         a_extrapState,
              LayoutData< Vector< BaseIVFAB<Real>* > >&       a_coveredPrimLo, 
              LayoutData< Vector< BaseIVFAB<Real>* > >&       a_coveredPrimHi, 
              const LayoutData< Vector< Vector<VolIndex> > >& a_coveredFaceLo, 
              const LayoutData< Vector< Vector<VolIndex> > >& a_coveredFaceHi, 
              const LayoutData< Vector< IntVectSet> >&        a_coveredSetsLo, 
              const LayoutData< Vector< IntVectSet> >&        a_coveredSetsHi, 
              const LevelData< EBCellFAB >&                   a_consState,
              const LevelData< EBCellFAB >&                   a_normalVel,
              const LevelData< EBFluxFAB >&                   a_advectionVel,
              const LevelData< EBCellFAB >*                   a_consStateCoarseOld,
              const LevelData< EBCellFAB >*                   a_consStateCoarseNew,
              const Real&                                     a_timeCoarseOld,
              const Real&                                     a_timeCoarseNew,
              const Real&                                     a_timeFine,
              const Real&                                     a_dt)
{
  assert(isDefined());
  //for now we will not do the diffusion thing
  assert(!m_useDiffusion);

  //create temp data with the correct number of ghost cells
  IntVect ivGhost = m_nGhost*IntVect::Unit;
  Interval consInterv(0, m_nVar-1);

  EBCellFactory factory(m_thisEBISL);
  LevelData<EBCellFAB> consTemp(m_thisGrids, m_nVar, ivGhost, factory);
  a_consState.copyTo(consInterv, consTemp, consInterv);

  // Fill ghost cells using fillInterp, and copyTo.
  if (m_hasCoarser)
    {
      m_pwlFillPatch.interpolate(consTemp,
                                 *a_consStateCoarseOld,
                                 *a_consStateCoarseNew,
                                 a_timeCoarseOld,
                                 a_timeCoarseNew,
                                 a_timeFine,
                                 consInterv);
    }

  // Exchange all the data between grids
  consTemp.exchange(consInterv);
  Box domainBox = m_domain.domainBox();

  for (DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& cellBox = m_thisGrids.get(dit());
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      if(!ebisBox.isAllCovered())
        {
          EBCellFAB& consState = consTemp[dit()];
          IntVectSet cfivs; //not used here.  only used in flux interpolation
          m_ebPatchAdvect->setValidBox(cellBox, ebisBox, cfivs, a_timeFine, a_dt);
          Box grownBox = grow(cellBox, 4);
          grownBox &= m_domain;
          const EBCellFAB& uNorm   = a_normalVel[dit()];
          const EBFluxFAB& uAdvect = a_advectionVel[dit()];

          m_ebPatchAdvect->setVelocities(uNorm, uAdvect);

          //the viscous term goes into here once we have it
          EBCellFAB source;

          //placeholder (no flattening used here)
          EBCellFAB flattening;
          //not reused
          EBCellFAB  primState;
          EBCellFAB  slopesPrim[SpaceDim];
          EBCellFAB  slopesSeco[SpaceDim];
          bool verbose = false;
          m_ebPatchAdvect->extrapolatePrim(a_extrapState[dit()],
                                           primState,
                                           slopesPrim,
                                           slopesSeco,
                                           a_coveredPrimLo[dit()],
                                           a_coveredPrimHi[dit()],
                                           a_coveredSetsLo[dit()],
                                           a_coveredSetsHi[dit()],
                                           a_coveredFaceLo[dit()],
                                           a_coveredFaceHi[dit()],
                                           flattening,
                                           consState, 
                                           source, cellBox, 
                                           verbose);
        }
    }
}
/*****************************/
void
EBLevelAdvect::
computeNormalVel(LevelData<EBCellFAB>&                          a_normalVel, 
                 const LevelData<EBFluxFAB>&                    a_advectionVel,
                 const LayoutData<Vector<BaseIVFAB<Real> * > >& a_coveredVeloLo, 
                 const LayoutData<Vector<BaseIVFAB<Real> * > >& a_coveredVeloHi, 
                 const LayoutData<Vector<Vector<VolIndex> > >&  a_coveredFaceLo, 
                 const LayoutData<Vector<Vector<VolIndex> > >&  a_coveredFaceHi) const
{
  for(DataIterator dit = m_thisGrids.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = m_thisEBISL[dit()];
      const Box& cellBox = m_thisGrids.get(dit());
      Box  grownBox = grow(cellBox, 4);
      grownBox &= m_domain;
      IntVectSet cfivs;
      Real time = 0.;
      m_ebPatchAdvect->setValidBox(cellBox, ebisBox, cfivs, time, time);
      m_ebPatchAdvect->averageVelToCC(a_normalVel[dit()],
                                      a_advectionVel[dit()],
                                      a_coveredVeloLo[dit()],
                                      a_coveredVeloHi[dit()],
                                      a_coveredFaceLo[dit()],
                                      a_coveredFaceHi[dit()],
                                      grownBox);
    }

}
/*****************************/
