/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental, 
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
#include "EBPatchAdvect.H"
#include "EBPatchAdvectF_F.H"
#include <cstdio>
void 
EBPatchAdvect::
extrapolatePrim(EBFluxFAB&                       a_flux, 
                EBCellFAB&                       a_primState,
                EBCellFAB                        a_slopePrim[SpaceDim],
                EBCellFAB                        a_slopeNLim[SpaceDim],
                Vector<BaseIVFAB<Real> * >&      a_coveredFluxMinu, 
                Vector<BaseIVFAB<Real> * >&      a_coveredFluxPlus, 
                const Vector<IntVectSet >&       a_coveredSetsMinu, 
                const Vector<IntVectSet >&       a_coveredSetsPlus, 
                const Vector<Vector<VolIndex> >& a_coveredFaceMinu, 
                const Vector<Vector<VolIndex> >& a_coveredFacePlus, 
                const EBCellFAB&                 a_flattening,
                const EBCellFAB&                 a_consState, 
                const EBCellFAB&                 a_source, 
                const Box&                       a_box,
                bool                             a_verbose)
{
  assert(a_coveredFluxMinu.size() == SpaceDim);
  assert(a_coveredFluxPlus.size() == SpaceDim);
  assert(a_coveredFaceMinu.size() == SpaceDim);
  assert(a_coveredFacePlus.size() == SpaceDim);
  assert(a_coveredSetsMinu.size() == SpaceDim);
  assert(a_coveredSetsPlus.size() == SpaceDim);

  EBCellFAB primPlus[SpaceDim];
  EBCellFAB primMinu[SpaceDim];
#if CH_SPACEDIM==2

  extrapolatePrim2D(primMinu, primPlus,
                    a_primState,    a_slopePrim, a_slopeNLim, 
                    a_flattening, a_consState, a_source, 
                    a_box, a_verbose);

#elif CH_SPACEDIM==3

  extrapolatePrim3D(primMinu,   primPlus,
                    a_primState,  a_slopePrim, a_slopeNLim, 
                    a_flattening, a_consState, a_source, 
                    a_box, a_verbose);

#else
  MayDay::Error("bogus SpaceDim");
#endif


  int numPrim = numPrimitives();
  BaseIVFAB<Real> extendedStateMinu[SpaceDim];
  BaseIVFAB<Real> extendedStatePlus[SpaceDim];
  Box faceBox[SpaceDim];
  //this keeps the fluxes from being calculated
  //on boundaries
  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
    {
      faceBox[dir1] = a_box;
      faceBox[dir1].grow(dir1,1);
      faceBox[dir1] &= m_domain;
      faceBox[dir1].grow(dir1,-1);
      faceBox[dir1].surroundingNodes(dir1);
    }
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      extendedStatePlus[idir].define( a_coveredSetsPlus[idir], 
                                      m_ebisBox.getEBGraph(), numPrim);
      extendedStateMinu[idir].define( a_coveredSetsMinu[idir], 
                                      m_ebisBox.getEBGraph(), numPrim);

      extendedStatePlus[idir].setVal(0.);
      extendedStateMinu[idir].setVal(0.);
    }
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      //extrapolate to covered faces using updated extrapolated state
      extrapToCoveredFaces(extendedStateMinu[faceDir],
                           primMinu[faceDir],
                           primPlus[faceDir],
                           a_primState,
                           a_slopeNLim,
                           a_coveredFaceMinu[faceDir],
                           faceDir, Side::Lo, a_box);

      extrapToCoveredFaces(extendedStatePlus[faceDir],
                           primMinu[faceDir],
                           primPlus[faceDir],
                           a_primState,
                           a_slopeNLim,
                           a_coveredFacePlus[faceDir],
                           faceDir, Side::Hi, a_box);



      riemann(a_flux[faceDir], primPlus[faceDir], primMinu[faceDir], 
              faceDir, faceBox[faceDir]);

      // Use the user supplied PhysBC object to obtain boundary fluxes
      m_bc->fluxBC(a_flux[faceDir], a_primState, primMinu[faceDir], 
                   Side::Lo,  m_time, m_ebisBox, a_box, faceDir);
      m_bc->fluxBC(a_flux[faceDir], a_primState, primPlus[faceDir], 
                   Side::Hi,  m_time, m_ebisBox, a_box, faceDir);

      //solve riemann problem between extended state and 
      //value in vof to get covered flux.  
      riemann(*a_coveredFluxMinu[faceDir], 
              extendedStateMinu[faceDir], primMinu[faceDir], 
              a_coveredFaceMinu[faceDir], faceDir, Side::Lo);
      riemann(*a_coveredFluxPlus[faceDir], 
              extendedStatePlus[faceDir], primPlus[faceDir], 
              a_coveredFacePlus[faceDir], faceDir, Side::Hi);
    }
}
/****************/
void
EBPatchAdvect::
advectiveDerivative(EBCellFAB&                       a_uDotDelRho,
                    const EBFluxFAB&                 a_faceRho,
                    const EBFluxFAB&                 a_faceVel,
                    const Vector<BaseIVFAB<Real>*> & a_coveredRhoMinu,
                    const Vector<BaseIVFAB<Real>*> & a_coveredRhoPlus,
                    const Vector<BaseIVFAB<Real>*> & a_coveredVelMinu,
                    const Vector<BaseIVFAB<Real>*> & a_coveredVelPlus,
                    const Vector<Vector<VolIndex> >& a_coveredFaceMinu,
                    const Vector<Vector<VolIndex> >& a_coveredFacePlus,
                    const Box&                       a_box)
{
  assert(m_isDefined);
  assert(m_isBoxSet);
  assert(a_faceRho.getRegion().contains(a_box));
  assert(a_faceVel.nComp() == 1);
  assert(a_uDotDelRho.nComp() == a_faceRho.nComp());
  assert(a_faceVel.getRegion().contains(a_box));
  assert(a_uDotDelRho.getRegion().contains(a_box));
  assert(a_coveredRhoMinu.size() == SpaceDim);
  assert(a_coveredVelMinu.size() == SpaceDim);
  assert(a_coveredRhoPlus.size() == SpaceDim);
  assert(a_coveredVelPlus.size() == SpaceDim);
  assert(a_coveredFacePlus.size() == SpaceDim);
  assert(a_coveredFaceMinu.size() == SpaceDim);

  //average velocities to cell centers
  EBCellFAB normalVel(m_ebisBox, a_box, SpaceDim);
  averageVelToCC(normalVel,    
                 a_faceVel,
                 a_coveredVelMinu, 
                 a_coveredVelPlus, 
                 a_coveredFaceMinu, 
                 a_coveredFacePlus, 
                 a_box);

  //set udotdelrho to zero.  the fortran is additive
  a_uDotDelRho.setVal(0.);
  BaseFab<Real>& regUDotDelRho = a_uDotDelRho.getSingleValuedFAB();
  int ncons = numConserved();
  //compute udotdelu everywhere as regular.
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      const EBFaceFAB& faceRho = a_faceRho[faceDir];
      const EBFaceFAB& faceVel = a_faceVel[faceDir];

      const BaseFab<Real>& regFaceRho = faceRho.getSingleValuedFAB();
      const BaseFab<Real>& regFaceVel = faceVel.getSingleValuedFAB();
      
      //this does udotdelrho += 0.5*(uhigh+ulow)*(rhohigh-rholow)/dx
      //where high == ivec + half*e^facedir
      FORT_ADVECTIVEF(CHF_FRA(regUDotDelRho),
                      CHF_CONST_FRA(regFaceRho),
                      CHF_CONST_FRA1(regFaceVel, 0),
                      CHF_CONST_INT(faceDir),
                      CHF_CONST_INT(ncons),
                      CHF_CONST_REAL(m_dx),
                      CHF_BOX(a_box));
    }

  //update the irregular vofs
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //udelrho was set in regular update.  we reset it
      // to zero and recalc.
      for(int ivar = 0; ivar < ncons; ivar++)
        {
          Real uDelRhoPt = 0.0;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const EBFaceFAB& faceRho = a_faceRho[idir];
              Real rhoDiff = 0.;
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Real rhoDiffSide = 0.;
                  int isign = sign(sit());
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, idir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      Real rhoSide = faceRho(face, ivar);
                      rhoDiffSide += isign*rhoSide;
                    }
                  if(faces.size() > 1)
                    {
                      rhoDiffSide /= Real(faces.size());
                    }

                  rhoDiff += rhoDiffSide;
                } //end loop over sides

              Real velAve = normalVel(vof, idir);
              //numvel has to be > 1 by construction
              uDelRhoPt += velAve*rhoDiff/m_dx;
            }//end loop over directions

          a_uDotDelRho(vof, ivar) = uDelRhoPt;
        }//end loop over variables
    }//end loop over irreg vofs

  //now correct for the covered fluxes
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      for(SideIterator sit; sit.ok(); ++sit)
        {
          int isign = sign(sit());
          const BaseIVFAB<Real>*  coveredVelPtr;
          const BaseIVFAB<Real>*  coveredRhoPtr;
          const Vector<VolIndex>* coveredFacePtr;
          if(sit() == Side::Lo)
            {
              coveredVelPtr  = a_coveredVelMinu[faceDir];
              coveredRhoPtr  = a_coveredRhoMinu[faceDir];
              coveredFacePtr =&a_coveredFaceMinu[faceDir];
            }
          else
            {
              coveredVelPtr  = a_coveredVelPlus[faceDir];
              coveredRhoPtr  = a_coveredRhoPlus[faceDir];
              coveredFacePtr =&a_coveredFacePlus[faceDir];
            }
          const BaseIVFAB<Real>&  coveredRho = *coveredRhoPtr;
          const Vector<VolIndex>& coveredFace = *coveredFacePtr;
          for(int ivof = 0; ivof < coveredFace.size(); ivof++)
            {
              const VolIndex& vof = coveredFace[ivof];
              if(a_box.contains(vof.gridIndex()))
                {
                  //face on this side is covered.  use covered flux.
                  for(int ivar = 0; ivar < ncons; ivar++)
                    {
                      Real rhoCov = coveredRho(vof, ivar);
                      Real velAve = normalVel(vof, faceDir);
                      Real update = isign*rhoCov*velAve/m_dx;
                      a_uDotDelRho(vof, ivar) += update;
                    }
                }
            }
        }
    }
}
/****************/
void 
EBPatchAdvect::
setVelocities(const EBCellFAB& a_normalVel,
              const EBFluxFAB& a_advectionVel)

{
  m_normalVelPtr    = &a_normalVel;
  m_advectionVelPtr = &a_advectionVel;
  m_isVelSet = true;
}
/******/
void 
EBPatchAdvect::
updatePrim(EBCellFAB&              a_primMinu, 
           EBCellFAB&              a_primPlus, 
           const EBFaceFAB&        a_primFace, 
           const BaseIVFAB<Real>&  a_coveredPrimMinu, 
           const BaseIVFAB<Real>&  a_coveredPrimPlus, 
           const Vector<VolIndex>& a_coveredFaceMinu, 
           const Vector<VolIndex>& a_coveredFacePlus, 
           const int&              a_faceDir, 
           const Box&              a_box, 
           const Real&             a_scale)
{
  //uses flux as prims 
  //update the regular vofs.  
  //first we have to copy the original state 
  //so that the irregular stuff can be updated later.
  int numPrim = numPrimitives();
  EBCellFAB inputPrimMinu(m_ebisBox, a_primMinu.getRegion(), numPrim);
  EBCellFAB inputPrimPlus(m_ebisBox, a_primPlus.getRegion(), numPrim);
  inputPrimMinu.copy(a_primMinu);
  inputPrimPlus.copy(a_primPlus);

  const EBCellFAB& normalVel = *m_normalVelPtr;
  BaseFab<Real>& regPrimMinu = a_primMinu.getSingleValuedFAB();
  BaseFab<Real>& regPrimPlus = a_primPlus.getSingleValuedFAB();
  const BaseFab<Real>& regPrimFace = a_primFace.getSingleValuedFAB();
  
  const BaseFab<Real>& regNormVel = normalVel.getSingleValuedFAB();
  FORT_ADVECTUPDATE(CHF_BOX(a_box),
                    CHF_FRA(regPrimMinu), 
                    CHF_FRA(regPrimPlus), 
                    CHF_CONST_FRA(regNormVel), 
                    CHF_CONST_FRA(regPrimFace), 
                    CHF_CONST_INT(a_faceDir), 
                    CHF_CONST_INT(numPrim), 
                    CHF_CONST_REAL(a_scale));

  //update the irregular vofs
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //Real velMinu = Min(normalVel(vof, a_faceDir), 0.0);
      //Real velPlus = Max(normalVel(vof, a_faceDir), 0.0);
      Real velMinu = normalVel(vof, a_faceDir);
      Real velPlus = normalVel(vof, a_faceDir);

      for(int ivar = 0; ivar < numPrim; ivar++)
        {
          Real primDiff = 0.0;
          //state changed in regular update. 
          //need to overwrite this.
          Real origPrimMinu= inputPrimMinu(vof, ivar);
          Real origPrimPlus= inputPrimPlus(vof, ivar);

          for(SideIterator sit; sit.ok(); ++sit)
            {
              int isign = sign(sit());
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, a_faceDir, sit());
              //if there are faces, use the flux at them
              Real aggPrim = 0.0;
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  Real facePrim = a_primFace(face, ivar);
                  aggPrim += facePrim;
                }
              if(faces.size() > 1)
                aggPrim /= Real(faces.size());

              primDiff += isign*aggPrim;
            }
          //dx is already divided out (part of scale)
          primDiff *= a_scale;
          a_primMinu(vof, ivar) = origPrimMinu - velMinu*primDiff;
          a_primPlus(vof, ivar) = origPrimPlus - velPlus*primDiff;
        }
    }

  //add in covered face fluxes
  for(SideIterator sit; sit.ok(); ++sit)
    {
      int isign = sign(sit());
      const BaseIVFAB<Real>*  coveredPrimPtr;
      const Vector<VolIndex>* coveredFacePtr;
      if(sit() == Side::Lo)
        {
          coveredPrimPtr = &(a_coveredPrimMinu);
          coveredFacePtr = &(a_coveredFaceMinu);
        }                                     
      else                                    
        {                                     
          coveredPrimPtr = &(a_coveredPrimPlus);
          coveredFacePtr = &(a_coveredFacePlus);
        }
      const BaseIVFAB<Real>&  coveredPrim = *coveredPrimPtr;
      const Vector<VolIndex>& coveredFace = *coveredFacePtr;
      for(int ivof = 0; ivof < coveredFace.size(); ivof++)
        {
          const VolIndex& vof = coveredFace[ivof];
          //Real velMinu = Min(normalVel(vof, a_faceDir), 0.0);
          //Real velPlus = Max(normalVel(vof, a_faceDir), 0.0);
          Real velMinu = normalVel(vof, a_faceDir);
          Real velPlus = normalVel(vof, a_faceDir);

          if(a_box.contains(vof.gridIndex()))
            {
              for(int ivar = 0; ivar < numPrim; ivar++)
                {
                  Real covPrim = coveredPrim(vof, ivar);
                  Real primDiff =  isign*covPrim;

                  a_primMinu(vof, ivar) -=  a_scale*velMinu*primDiff;
                  a_primPlus(vof, ivar) -=  a_scale*velPlus*primDiff;
                }
            }
        }
    }

  floorPrimitives(a_primMinu, a_box);
  floorPrimitives(a_primPlus, a_box);

}
/*****/
bool 
EBPatchAdvect::
usesFlattening() const
{
  return false;
}
/******/
bool 
EBPatchAdvect::
usesArtificialViscosity() const
{
  return false;
}
/******/
bool 
EBPatchAdvect::
usesFourthOrderSlopes() const
{
  return false;
}
/******/
void 
EBPatchAdvect::
consToPrim(EBCellFAB&       a_primState, 
           const EBCellFAB& a_consState,
           const Box&       a_box,
           int              a_logflag)
{ 
  assert(numPrimitives() == numConserved());
  Interval interv(0, numConserved()-1);
  a_primState.copy(a_box, interv, a_box, a_consState, interv);
}
/******/
void 
EBPatchAdvect::
consToPrim(BaseIVFAB<Real>&  a_primState, 
           const EBCellFAB&  a_consState,
           const IntVectSet& a_ivs)
{ 
  assert(numPrimitives() == numConserved());
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      for(int ivar = 0; ivar < numPrimitives(); ivar++)
        {
          a_primState(vofit(), ivar) = a_consState(vofit(), ivar);
        }
    }
}
/******/
void 
EBPatchAdvect::
consToPrim(BaseIVFAB<Real>&        a_primState, 
           const BaseIVFAB<Real>&  a_consState,
           const IntVectSet&       a_ivs)
{ 
  assert(numPrimitives() == numConserved());
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      for(int ivar = 0; ivar < numPrimitives(); ivar++)
        {
          a_primState(vofit(), ivar) = a_consState(vofit(), ivar);
        }
    }
}
/******/
void 
EBPatchAdvect::
primToCons(EBCellFAB&       a_consState, 
           const EBCellFAB& a_primState,
           const Box&       a_box)
{ 
  assert(numPrimitives() == numConserved());
  Interval interv(0, numConserved()-1);
  a_consState.copy(a_box, interv, a_box, a_primState, interv);
}
/******/
void 
EBPatchAdvect::
primToCons(BaseIVFAB<Real>&       a_consState, 
           const BaseIVFAB<Real>& a_primState,
           const IntVectSet&      a_ivs)
{ 
  assert(numPrimitives() == numConserved());
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      for(int ivar = 0; ivar < numPrimitives(); ivar++)
        {
          a_consState(vofit(), ivar) = a_primState(vofit(), ivar);
        }
    }
}
void 
EBPatchAdvect::
setValidBox(const Box&        a_validBox, 
            const EBISBox&    a_ebisBox, 
            const IntVectSet& a_coarseFineIVS,
            const Real&       a_cumulativeTime, 
            const Real&       a_timeStep)
{
  EBPatchGodunov::setValidBox(a_validBox, a_ebisBox, a_coarseFineIVS, a_cumulativeTime, a_timeStep);
}
/******/
///returns primitives not fluxes
/***/
void 
EBPatchAdvect::
riemann(EBFaceFAB&       a_primGdnv, 
        const EBCellFAB& a_primLeft, 
        const EBCellFAB& a_primRigh, 
        const int&       a_faceDir, 
        const Box&       a_box)
{
  assert(m_isVelSet);
  const EBFluxFAB& fluxVel = *m_advectionVelPtr;
  const EBFaceFAB& advectVel = fluxVel[a_faceDir];
  assert(advectVel.nComp() == 1);

  int nPrim = numPrimitives();

  Box cellBox = enclosedCells(a_box);
  
  //explicitly cast the left and right states to modifiable references
  //because we need to shift them to faces and then shift them back.
  BaseFab<Real>& regPrimRigh = (BaseFab<Real>&)a_primRigh.getSingleValuedFAB();
  BaseFab<Real>& regPrimLeft = (BaseFab<Real>&)a_primLeft.getSingleValuedFAB();
  BaseFab<Real>& regPrimGdnv = a_primGdnv.getSingleValuedFAB();
  const BaseFab<Real>& regVeloGdnv = advectVel.getSingleValuedFAB();
  //shift the cell centered stuff to edges
  regPrimRigh.shiftHalf(a_faceDir, -1);
  regPrimLeft.shiftHalf(a_faceDir,  1);

  assert(regPrimRigh.box().contains(a_box));
  assert(regPrimLeft.box().contains(a_box));
  assert(regPrimGdnv.box().contains(a_box));
  
  //find the regular part of qgdnv.  
      
  FORT_ADVECTRIEMANN(CHF_BOX(a_box), 
                     CHF_FRA(regPrimGdnv), 
                     CHF_CONST_FRA(regPrimLeft), 
                     CHF_CONST_FRA(regPrimRigh), 
                     CHF_CONST_FRA1(regVeloGdnv, 0), 
                     CHF_CONST_INT(a_faceDir),
                     CHF_CONST_INT(nPrim));
      

  //shift the cell centered stuff back to cells
  regPrimRigh.shiftHalf(a_faceDir,  1);
  regPrimLeft.shiftHalf(a_faceDir, -1);

  //the box sent into this is face-centered.
  //we need to use the cell-centered one it surrounds.
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(cellBox);
  FaceStop::WhichFaces stopCrit = FaceStop::SurroundingNoBoundary;
  for(FaceIterator faceit(ivsMulti, m_ebisBox.getEBGraph(), a_faceDir, stopCrit);
      faceit.ok(); ++faceit)
    { 
      const FaceIndex& face = faceit();
      Real velFace = advectVel(face, 0);
      VolIndex vofl = face.getVoF(Side::Lo);
      VolIndex vofr = face.getVoF(Side::Hi);
      for(int ivar = 0; ivar < nPrim; ivar++)
        {
          if(velFace > 0)
            {
              a_primGdnv(face, ivar) = a_primLeft(vofl, ivar);
            }
          else
            {
              a_primGdnv(face, ivar) = a_primRigh(vofr, ivar);
            }
        }
    }
 
}
/***/
///returns primitives not fluxes
/*****************************/
void 
EBPatchAdvect::
riemann(BaseIVFAB<Real>&        a_coveredPrim, 
        const BaseIVFAB<Real>&  a_exteState, 
        const EBCellFAB&        a_primState, 
        const Vector<VolIndex>& a_vofset, 
        const int&              a_faceDir, 
        const Side::LoHiSide&   a_sd)
{
  const EBCellFAB& normalVel = *m_normalVelPtr;
  int nPrim = numPrimitives();
  for(int ivof = 0; ivof < a_vofset.size(); ivof++)
    {
      const VolIndex& vof = a_vofset[ivof];
      Real  vel = normalVel(vof, a_faceDir);

      for(int ivar = 0; ivar < nPrim; ivar++)
        {
          Real leftState, righState;
          if(a_sd == Side::Lo)
            {
              leftState = a_exteState(vof, ivar);
              righState = a_primState(vof, ivar);
            }
          else
            {
              leftState = a_primState(vof, ivar);
              righState = a_exteState(vof, ivar);
            }
          
          if(vel > 0.0)
            {
              a_coveredPrim(vof, ivar) = leftState;
            }
          else
            {
              a_coveredPrim(vof, ivar) = righState;
            }
        }
    }

}
/*****************************/
EBPatchAdvect::
EBPatchAdvect():EBPatchGodunov()
{
  m_isVelSet        = false;
  m_advectionVelPtr = NULL;
  m_normalVelPtr    = NULL;
  m_useLimiting            = true;
  m_useFourthOrderSlopes   = false;
  m_useFlattening          = false;
}
/*****************************/
void 
EBPatchAdvect::
useLimiting(bool a_limiting)
{
  //sets flattening and fourth order slopes to false
  setSlopeParameters(false, false, a_limiting);
}
/*****************************/
EBPatchAdvect::
~EBPatchAdvect()
{
}
/*****************************/
Vector<string> 
EBPatchAdvect::
stateNames()
{
  Vector<string> retval(numConserved());
  for(int ivar = 0; ivar < numConserved(); ivar++)
    {
      char strname[100];
      sprintf(strname, "consname%d",ivar);
      retval[ivar] = string(strname);
    }
  return retval;
}
/*****************************/
Vector<string> 
EBPatchAdvect::
primNames()
{
  Vector<string> retval(numConserved());
  for(int ivar = 0; ivar < numConserved(); ivar++)
    {
      char strname[100];
      sprintf(strname, "primname%d",ivar);
      retval[ivar] = string(strname);
    }
  return retval; 
}
/*****************************/
void
EBPatchAdvect::
computeEBIrregFlux(BaseIVFAB<Real>&  a_ebIrregFlux,
                   const EBCellFAB&  a_primState,
                   const EBCellFAB&  a_nonConsDivF,
                   const EBCellFAB   a_slopePrim[SpaceDim],
                   const IntVectSet& a_irregIVS,
                   const EBCellFAB&  a_source)
{
  a_ebIrregFlux.setVal(0.);
}
/*****************************/
void 
EBPatchAdvect::
averageVelToCC(EBCellFAB&                        a_normalVel, 
               const EBFluxFAB&                  a_advectionVel,
               const Vector<BaseIVFAB<Real> * >& a_coveredVeloLo, 
               const Vector<BaseIVFAB<Real> * >& a_coveredVeloHi, 
               const Vector<Vector<VolIndex> >&  a_coveredFaceLo, 
               const Vector<Vector<VolIndex> >&  a_coveredFaceHi, 
               const Box&                        a_box) const
{
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      const EBFaceFAB& faceNormVel = a_advectionVel[faceDir];

      //treat every cell as regular
      BaseFab<Real>&       regCellVel = a_normalVel.getSingleValuedFAB();
      const BaseFab<Real>& regFaceVel = faceNormVel.getSingleValuedFAB();
      FORT_AVEFACETOCELL(CHF_FRA(regCellVel),
                         CHF_CONST_FRA1(regFaceVel,0),
                         CHF_CONST_INT(faceDir),
                         CHF_BOX(a_box));
      
      //average face data to cell data on irregular cells.
      IntVectSet irregIVS = m_ebisBox.getIrregIVS(a_box);
      for(VoFIterator vofit(irregIVS, m_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          int nVel = 0;
          Real faceVelAve = 0.0;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              Vector<FaceIndex> faces = m_ebisBox.getFaces(vof, faceDir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  faceVelAve += faceNormVel(faces[iface], 0);
                }
              //if there are no faces, there should be a covered face
              //(added in later but need to get the factor correct)
              int facesize = Max(int(faces.size()), 1);
              nVel += facesize;
            }
          if(nVel > 1)
            {
              faceVelAve /= nVel;
            }
          a_normalVel(vof, faceDir) = faceVelAve;
        }

      //add in contribution of covered cells
      for(SideIterator sit; sit.ok(); ++sit)
        {
          const BaseIVFAB<Real>*   coveredVeloPtr = NULL;
          const Vector<VolIndex>*  coveredFacePtr = NULL;
          if(sit() == Side::Lo)
            {
              coveredVeloPtr =  a_coveredVeloLo[faceDir];
              coveredFacePtr = &a_coveredFaceLo[faceDir];
            }
          else
            {
              coveredVeloPtr =  a_coveredVeloHi[faceDir];
              coveredFacePtr = &a_coveredFaceHi[faceDir];
            }
          
          const BaseIVFAB<Real>&   coveredVelo = *coveredVeloPtr;
          const Vector<VolIndex>&  coveredFace = *coveredFacePtr;
          assert(coveredVelo.nComp() == 1);
          for(int ivof = 0; ivof < coveredFace.size(); ivof++)
            {
              const VolIndex& vof = coveredFace[ivof];
              if(a_box.contains(vof.gridIndex()))
                {

                  int nFaces = Max(1, m_ebisBox.numFaces(vof, faceDir, flip(sit())));
                  nFaces += 1;
                  
                  a_normalVel(vof, faceDir) += coveredVelo(vof,0)/Real(nFaces);
                }
            }
        }
    }
}
