/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "EBPatchGodunov.H"
#include "EBPatchGodunovF_F.H"
#include "EBArith.H"
#include "PolyGeom.H"
#include "EBDebugOut.H"
#include "DebugOut.H"
#include "DatasetClient.H"
#include "VoFIterator.H"
#include "FaceIterator.H"
#include "EBLoHiCenter.H"
#include "EBPatchGodunov.H"
#include "Stencils.H"
#include "EBArith.H"
#include "EBAMRIO.H"
#include "PolyGeom.H"
#include "ParmParse.H"

EBPatchGodunov::EBPatchGodunov()
{
  IntVect ivdebug1(D_DECL(67,11,10));
  IntVect ivdebug2(D_DECL(68,11,10));
  m_debugBox = Box(ivdebug1,ivdebug2);
  //  m_debugBox.grow(1);
  m_isDefined     = false;
  m_isBCSet       = false;
  m_isBoxSet      = false;
  m_bc = NULL;
}

EBPatchGodunov::~EBPatchGodunov()
{
  if (m_bc != NULL)
    {
      delete m_bc;
    }
}
/******/
void 
EBPatchGodunov::
incrementWithSource(EBCellFAB&       a_primState, 
                    const EBCellFAB& a_source, 
                    const Real&      a_scale, 
                    const Box&       a_box) 
{
  assert(isDefined());
  assert(a_source.nComp() == a_primState.nComp());
  assert(a_source.getRegion().contains(a_box));
  assert(a_primState.getRegion().contains(a_box));

  BaseFab<Real>&       primReg = a_primState.getSingleValuedFAB();
  const BaseFab<Real>& sourReg = a_source.getSingleValuedFAB();
      
  FORT_INCSOURCE(CHF_FRA(primReg), 
                 CHF_CONST_FRA(sourReg), 
                 CHF_CONST_REAL(a_scale), 
                 CHF_BOX(a_box));
      

  IntVectSet ivsMulti = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < a_source.nComp(); ivar++)
        {
          Real increm = a_scale*a_source(vof, ivar);
          a_primState(vof, ivar) +=  increm;
        }
    }
}
/******/
void 
EBPatchGodunov::
extrapToCoveredFaces(BaseIVFAB<Real>&        a_extendedPrim, 
                     const EBCellFAB&        a_primMinu,
                     const EBCellFAB&        a_primPlus,
                     const EBCellFAB&        a_primState,
                     const EBCellFAB         a_slopesSeco[SpaceDim],
                     const Vector<VolIndex>& a_coveredFaces,
                     const int&              a_faceDir, 
                     const Side::LoHiSide&   a_sd,
                     const Box&              a_box)
{
  for(int ivof = 0; ivof < a_coveredFaces.size(); ivof++)
    {
      const VolIndex& vof = a_coveredFaces[ivof];

      RealVect normal = m_ebisBox.normal(vof);
      if(a_box.contains(vof.gridIndex()))
        {
          const int numPrim = numPrimitives();
          Vector<Real> extPrim(numPrim, 0.0);
#if CH_SPACEDIM==2
          pointExtrapToCovered2D(extPrim,
                                 a_primMinu,
                                 a_primPlus,
                                 a_primState,
                                 a_slopesSeco,
                                 a_faceDir,
                                 vof,
                                 normal,
                                 a_sd,
                                 numPrim);
#elif CH_SPACEDIM==3
          pointExtrapToCovered3D(extPrim,
                                 a_primMinu,
                                 a_primPlus,
                                 a_primState,
                                 a_slopesSeco,
                                 a_faceDir,
                                 vof,
                                 normal,
                                 a_sd,
                                 numPrim);
#else
          bogus_ch_spacedim();
#endif          
          for(int ivar = 0; ivar < numPrim; ivar++)
            {
              a_extendedPrim(vof, ivar) = extPrim[ivar];
            }
        }
    }  
  //debug to keep me from going insane.
  //need to uncomment this for undebug
  const IntVectSet& ivs  = a_extendedPrim.getIVS();
  floorPrimitives(a_extendedPrim, ivs);
} 
/********/
void
EBPatchGodunov::
pointExtrapToCovered2D(Vector<Real>&           a_extrapVal,
                       const EBCellFAB&        a_primMinu,
                       const EBCellFAB&        a_primPlus,
                       const EBCellFAB&        a_primState,
                       const EBCellFAB         a_slopesSeco[SpaceDim],
                       const int&              a_faceDir, 
                       const VolIndex&         a_vof,
                       const RealVect&         a_normal,
                       const Side::LoHiSide&   a_sd,
                       const int&              a_numPrim)
{
  assert(SpaceDim==2);
  int tangenDir = 1 - a_faceDir;

  int signNorm = 1;
  int signTang = 1;
  if(a_normal[a_faceDir] < 0) signNorm = -1;
  if(a_normal[tangenDir]    < 0)  signTang = -1;

  //iv[0][0] is the vof. iv[1][0] is vofside.  iv[0][1] is vofup iv[1][1] is vofcorner
  IntVect   ivSten[2][2];
  bool      hasVoF[2][2];
  VolIndex     vof[2][2];
  Real         val[2][2];
  ivSten[0][0] = a_vof.gridIndex();
  ivSten[1][0] = ivSten[0][0] + signNorm*BASISV(a_faceDir);
  ivSten[0][1] = ivSten[0][0] - signNorm*BASISV(a_faceDir) + signTang*BASISV(tangenDir) ;;
  ivSten[1][1] = ivSten[0][0]                              + signTang*BASISV(tangenDir) ;
  
  int radius = 1;
  Vector<VolIndex> vofsStencil;
  EBArith::getAllVoFsInMonotonePath(vofsStencil, a_vof, m_ebisBox, radius);

  for(int ivar = 0; ivar < a_numPrim; ivar++)
    {
      for(int ix = 0; ix < 2; ix++)
        {
          for(int iy = 0; iy < 2; iy++)
            {
              hasVoF[ix][iy] = EBArith::isVoFHere(vof[ix][iy], vofsStencil, ivSten[ix][iy]);
              if(hasVoF[ix][iy])
                {
                  if(a_sd == Side::Hi)
                    {
                      val[ix][iy] = a_primMinu(vof[ix][iy], ivar);
                    }
                  else
                    {
                      val[ix][iy] = a_primPlus(vof[ix][iy], ivar);
                    }
                }
              else
                {
                  val[ix][iy] = 0.0;
                }
            }
        }

      int dirBigNorm, dirLitNorm;
      VolIndex vofBigNorm, vofLitNorm;
      bool hasVoFBigNorm,  hasVoFLitNorm;
      int  signBigNorm, signLitNorm;
      Real wBigNorm, wLitNorm;
      if(Abs(a_normal[a_faceDir]) <= Abs(a_normal[tangenDir]))
        {
          dirBigNorm = tangenDir;
          dirLitNorm = a_faceDir;

          signBigNorm   = signTang;
          signLitNorm   = signNorm;
          wBigNorm      =    val[0][1];
          vofBigNorm    =    vof[0][1];
          hasVoFBigNorm = hasVoF[0][1];
        }
      else
        {
          dirBigNorm = a_faceDir;
          dirLitNorm = tangenDir;

          signBigNorm   = signNorm;
          signLitNorm   = signTang;
      
          Real deltaW = a_slopesSeco[a_faceDir](vof[1][0]  , ivar);
          wBigNorm      =    val[1][0] - signBigNorm*deltaW;
          vofBigNorm    =    vof[1][0];
          hasVoFBigNorm = hasVoF[1][0];
        }
      wLitNorm      =    val[1][1];
      vofLitNorm    =    vof[1][1];
      hasVoFLitNorm = hasVoF[1][1];

      linearInterp( a_extrapVal[ivar]
                    ,     wBigNorm,      wLitNorm
                    ,   vofBigNorm,    vofLitNorm
                    ,   dirBigNorm,    dirLitNorm
                    ,  signBigNorm,   signLitNorm
                    ,hasVoFBigNorm, hasVoFLitNorm
                    ,a_primState  , a_slopesSeco
                    ,a_normal     , a_vof, ivar);
    }
}

/********/
void 
EBPatchGodunov::
linearInterp(Real&                   a_extrapVal,
             const Real&             a_wBigNorm,
             const Real&             a_wLitNorm,
             const VolIndex&         a_vofBigNorm,
             const VolIndex&         a_vofLitNorm,
             const int&              a_dirBigNorm,
             const int&              a_dirLitNorm,
             const int&              a_signBigNorm,
             const int&              a_signLitNorm,
             const bool&             a_hasVoFBigNorm,
             const bool&             a_hasVoFLitNorm,
             const EBCellFAB&        a_primState,
             const EBCellFAB         a_slopesSeco[SpaceDim],
             const RealVect&         a_normal,
             const VolIndex&         a_vof,
             const int&              a_ivar)
{
  assert(m_isSlopeSet);
  if(a_hasVoFBigNorm && a_hasVoFLitNorm)
    {
      Real normRat = ((Abs(a_normal[a_dirLitNorm]))/(Abs(a_normal[a_dirBigNorm])));
      Real deltaWLitNorm = a_slopesSeco[a_dirLitNorm](a_vofLitNorm, a_ivar);
      Real deltaWBigNorm = a_slopesSeco[a_dirBigNorm](a_vofBigNorm, a_ivar);
      a_extrapVal = normRat*a_wLitNorm + (1.0-normRat)*a_wBigNorm ;

      Real deltaW1, deltaW2;

      coveredExtrapSlopes(deltaW1, a_vofLitNorm, a_primState, a_dirLitNorm, a_ivar);
      coveredExtrapSlopes(deltaW2, a_vofBigNorm, a_primState, a_dirLitNorm, a_ivar);
      if(m_useLimiting)
        {
          deltaWLitNorm = pointLimiter(deltaW1, deltaW2);
        }
      else
        {
          deltaWLitNorm = deltaW1;
        }

      coveredExtrapSlopes(deltaW1, a_vofLitNorm, a_primState, a_dirBigNorm, a_ivar);
      coveredExtrapSlopes(deltaW2, a_vofBigNorm, a_primState, a_dirBigNorm, a_ivar);
      if(m_useLimiting)
        {
          deltaWBigNorm = pointLimiter(deltaW1, deltaW2);
        }
      else
        {
          deltaWBigNorm = deltaW2;
        }

      a_extrapVal -= (normRat*a_signLitNorm*deltaWLitNorm + a_signBigNorm*deltaWBigNorm);
    }
  else if(a_hasVoFLitNorm)
    {
      a_extrapVal = a_wLitNorm;
    }
  else if(a_hasVoFBigNorm)
    {
      a_extrapVal = a_wBigNorm;
    }
  else
    {
      a_extrapVal = a_primState(a_vof, a_ivar);
    }
}
/********/
Real 
EBPatchGodunov::pointLimiter(const Real& a_deltaW1,
                             const Real& a_deltaW2)
{
  Real retval;
  if(a_deltaW1*a_deltaW2 > 0.0)
    {
      retval = Min(Abs(a_deltaW1), Abs(a_deltaW2));
      if(a_deltaW1 < 0.0)
        {
          retval *= -1.0;
        }
    }
  else
    {
      retval = 0.0;
    }
  return retval;
}
/******/
Real 
EBPatchGodunov::
bilinearFunc(const Real  a_WVal[2][2], 
             const Real& a_xd1,
             const Real& a_xd2)
{
  Real D =                a_WVal[0][0];
  Real A =(a_WVal[1][0] - D);
  Real B =(a_WVal[0][1] - D);
  Real C =(a_WVal[1][1] - D - A - B); 
  Real retval = A*a_xd1 + B*a_xd2 + C*a_xd1*a_xd2 + D;
  return retval;
}
/********/
void
EBPatchGodunov::
pointExtrapToCovered3D(Vector<Real>&           a_extrapVal,
                       const EBCellFAB&        a_primMinu,
                       const EBCellFAB&        a_primPlus,
                       const EBCellFAB&        a_primState,
                       const EBCellFAB         a_slopesSeco[SpaceDim],
                       const int&              a_faceDir, 
                       const VolIndex&         a_vof,
                       const RealVect&         a_normal,
                       const Side::LoHiSide&   a_sd,
                       const int&              a_numPrim)
{
  assert(SpaceDim==3);
  Tuple<int,CH_SPACEDIM-1> tangenDir = PolyGeom::computeTanDirs(a_faceDir);
  Real anormNorm = Abs(a_normal[a_faceDir]);
  Real anormTan[2];
  int signNorm = 1;
  if(a_normal[a_faceDir] < 0) signNorm = -1;
  int signTang[2];
  for(int itan = 0; itan < 2; itan++)
    {
      int tanDir = tangenDir[itan];
      anormTan[itan] = Abs(a_normal[tanDir]);
      signTang[itan] =  1;
      if(a_normal[tanDir] < 0.) signTang[itan] = -1;
    }
    
  const IntVect& ivVoF= a_vof.gridIndex();

  //whice one of the tangential directions has the largest normal
  int d1, d2;
  if(anormTan[0] > anormTan[1])
    {
      d1 = 0;
      d2 = 1;
    }
  else
    {
      d1 = 1;
      d2 = 0;
    }

  // figure out in which plane we are extrapolating 
  bool faceDirOut = ((anormNorm > anormTan[0]) && (anormNorm > anormTan[1]));

  IntVect ivSten[2][2];
  bool hasVoF[2][2];
  VolIndex vofSten[2][2];
      
  if(faceDirOut)
    {// face direction has largest normal.
      
      ivSten[0][0] = ivVoF + signNorm*BASISV(a_faceDir);
      ivSten[1][0] = ivVoF + signTang[0]*BASISV(tangenDir[0]);
      ivSten[0][1] = ivVoF + signTang[1]*BASISV(tangenDir[1]);
      ivSten[1][1] = ivVoF 
        + signTang[0]*BASISV(tangenDir[0]) 
        + signTang[1]*BASISV(tangenDir[1]) ;
      
    }
  else
    { //tandir[d1] is the biggest

      ivSten[0][0] = ivVoF    + signTang[d1]*BASISV(tangenDir[d1])                                      - signNorm*BASISV(a_faceDir);
      ivSten[1][0] = ivVoF    + signTang[d1]*BASISV(tangenDir[d1])                                      ;
      ivSten[0][1] = ivVoF    + signTang[d1]*BASISV(tangenDir[d1]) + signTang[d2]*BASISV(tangenDir[d2]) - signNorm*BASISV(a_faceDir);
      ivSten[1][1] = ivVoF    + signTang[d1]*BASISV(tangenDir[d1]) + signTang[d2]*BASISV(tangenDir[d2]) ;
      
    }
  
  int radius = 1;
  Vector<VolIndex> vofsStencil;
  EBArith::getAllVoFsInMonotonePath(vofsStencil, a_vof, m_ebisBox, radius);
                                   
  for(int ix = 0; ix < 2; ix++)
    {
      for(int iy = 0; iy < 2; iy++)
        {
          hasVoF[ix][iy] =  EBArith::isVoFHere(vofSten[ix][iy], vofsStencil, ivSten[ix][iy]);
        }
    }
  bool hasAllVoFs = hasVoF[0][0] && hasVoF[1][0] && hasVoF[0][1] && hasVoF[1][1];
  if(!hasAllVoFs)
    {
      Box bgrow(ivVoF, ivVoF);
      bgrow.grow(1);
      if(m_domain.contains(bgrow))
        {
          pout() << "interior vof " << a_vof << " drops order " << endl;
        }
    }

  for(int ivar = 0; ivar < a_numPrim; ivar++)
    {
      if(hasAllVoFs)
        {
          Real WVal[2][2];
          for(int ix = 0; ix < 2; ix++)
            {
              for(int iy = 0; iy < 2; iy++)
                {
                  if(a_sd == Side::Hi)
                    {
                      WVal[ix][iy] = a_primMinu(vofSten[ix][iy], ivar);
                    }
                  else
                    {
                      WVal[ix][iy] = a_primPlus(vofSten[ix][iy], ivar);
                    }
                }
            }
          Real deltaWNorm;
          Real deltaWTan0;
          Real deltaWTan1;
          if(faceDirOut)
            {
              coveredExtrapSlopes(deltaWNorm, vofSten[0][0], a_primState, a_faceDir   , ivar);
              coveredExtrapSlopes(deltaWTan0, vofSten[1][0], a_primState, tangenDir[0], ivar);
              coveredExtrapSlopes(deltaWTan1, vofSten[0][1], a_primState, tangenDir[1], ivar);
            }
          else
            {
              coveredExtrapSlopes(deltaWNorm, vofSten[1][0], a_primState, a_faceDir    , ivar);
              coveredExtrapSlopes(deltaWTan0, vofSten[0][0], a_primState, tangenDir[d1], ivar);
              coveredExtrapSlopes(deltaWTan1, vofSten[0][1], a_primState, tangenDir[d2], ivar);
            }
          if(m_useLimiting)
            {
              for(int ix = 0; ix < 2; ix++)
                {
                  for(int iy = 0; iy < 2; iy++)
                    {
                      Real dWNorm, dWTan0, dWTan1;
                      coveredExtrapSlopes(dWNorm, vofSten[ix][iy], a_primState, a_faceDir   , ivar);
                      coveredExtrapSlopes(dWTan0, vofSten[ix][iy], a_primState, tangenDir[0], ivar);
                      coveredExtrapSlopes(dWTan1, vofSten[ix][iy], a_primState, tangenDir[1], ivar);
                      deltaWNorm = pointLimiter(deltaWNorm, dWNorm);
                      deltaWTan0 = pointLimiter(deltaWTan0, dWTan0);
                      deltaWTan1 = pointLimiter(deltaWTan1, dWTan1);
                    }
                }
            }

          if(faceDirOut)
            {
              // case 00 is special because we have to move it backwards in face direction to get it to plane
              Real slipSlope = a_slopesSeco[a_faceDir](vofSten[0][0], ivar);
              WVal[0][0] -= signNorm*slipSlope;
              //debug to make compatible with 2d
              //WVal[0][1] = WVal[0][0];
              //end debug
              
              Real x0Val = anormTan[0]/anormNorm;
              Real x1Val = anormTan[1]/anormNorm;
              Real funcVal = bilinearFunc(WVal,x0Val,x1Val);

              a_extrapVal[ivar] = funcVal - signNorm*deltaWNorm - signTang[0]*x0Val*deltaWTan0 - signTang[1]*x1Val*deltaWTan1;

            }
          else
            {
              //tanDir[d1] the biggest normal
              Real x0Val = anormNorm/anormTan[d1];
              Real x1Val = anormTan[d2]/anormTan[d1];
              Real funcVal = bilinearFunc(WVal,x0Val,x1Val);

              a_extrapVal[ivar] = funcVal -signNorm*x0Val*deltaWNorm - signTang[d2]*x1Val*deltaWTan1 - signTang[d1]*deltaWTan0;
            }
        } //end if(hasAllVoFs)
      else 
        {
          a_extrapVal[ivar] = a_primState(a_vof, ivar);
        }
    }
}
/******/
/*****/
void 
EBPatchGodunov::
updatePrim(EBCellFAB&              a_primMinu, 
           EBCellFAB&              a_primPlus, 
           const EBFaceFAB&        a_flux, 
           const BaseIVFAB<Real>&  a_coveredFluxMinu, 
           const BaseIVFAB<Real>&  a_coveredFluxPlus, 
           const Vector<VolIndex>& a_coveredFaceMinu, 
           const Vector<VolIndex>& a_coveredFacePlus, 
           const int&              a_dir, 
           const Box&              a_box, 
           const Real&             a_scale) 
{
  assert(isDefined());
  assert(a_primPlus.nComp() == numPrimitives());
  assert(a_primMinu.nComp() == numPrimitives());
  assert(a_flux.nComp() == numFluxes());
  assert((a_dir >= 0) && (a_dir < SpaceDim));
  assert(a_flux.direction() == a_dir);
  assert(a_flux.getRegion().contains(surroundingNodes(a_box, a_dir)));
  assert(a_primPlus.getRegion().contains(a_box));
  assert(a_primMinu.getRegion().contains(a_box));
  assert(a_primPlus.getRegion() == a_primMinu.getRegion());

  const Box& primBox = a_primPlus.getRegion();
  int nCons =  numConserved();
  EBCellFAB consTemp(m_ebisBox, primBox, nCons);

  primToCons(consTemp, a_primMinu, primBox);
  updateCons(consTemp, a_flux, 
             a_coveredFluxMinu, a_coveredFluxPlus, 
             a_coveredFaceMinu, a_coveredFacePlus, 
             a_dir, a_box, a_scale);
  int logflag = 0;
  consToPrim(a_primMinu, consTemp, primBox, logflag);

  primToCons(consTemp, a_primPlus, primBox);
  updateCons(consTemp, a_flux, 
             a_coveredFluxMinu, a_coveredFluxPlus, 
             a_coveredFaceMinu, a_coveredFacePlus, 
             a_dir, a_box, a_scale);
  consToPrim(a_primPlus, consTemp, primBox, logflag);

}
/*****/
void 
EBPatchGodunov::
updateCons(EBCellFAB&              a_consState, 
           const EBFaceFAB&        a_flux, 
           const BaseIVFAB<Real>&  a_coveredFluxMinu, 
           const BaseIVFAB<Real>&  a_coveredFluxPlus, 
           const Vector<VolIndex>& a_coveredFaceMinu, 
           const Vector<VolIndex>& a_coveredFacePlus, 
           const int&              a_dir, 
           const Box&              a_box, 
           const Real&             a_scale) 
{
  assert(isDefined());
  assert(a_consState.nComp() <= a_flux.nComp());
  assert(a_consState.nComp() == numConserved() );
  assert((a_dir >= 0) && (a_dir < SpaceDim));
  assert(a_flux.direction() == a_dir);
  assert(a_flux.getRegion().contains(surroundingNodes(a_box, a_dir)));
  assert(a_consState.getRegion().contains(a_box));

  //update the regular vofs.  
  //first we have to copy the original state 
  //so that the irregular stuff can be updated later.
  EBCellFAB inputConsState(m_ebisBox, 
                           a_consState.getRegion(), 
                           a_consState.nComp());
  //  setCoveredConsVals(a_consState);
  inputConsState.copy(a_consState);

  int ncons = numConserved();
  BaseFab<Real>&       consReg = a_consState.getSingleValuedFAB();
  const BaseFab<Real>& fluxReg = a_flux.getSingleValuedFAB();

      
  FORT_UPDATE( CHF_BOX(a_box), 
               CHF_FRA(consReg), 
               CHF_CONST_FRA(fluxReg), 
               CHF_CONST_INT(a_dir), 
               CHF_CONST_INT(ncons), 
               CHF_CONST_REAL(a_scale));
      

  //update the irregular vofs
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //      const IntVect& iv = vof.gridIndex();
      for(int ivar = 0; ivar < a_consState.nComp(); ivar++)
        {
          Real update = 0.0;
          //state changed in regular update. 
          //need to overwrite this.
          Real originalState= inputConsState(vof, ivar);
          for(SideIterator sit; sit.ok(); ++sit)
            {
              int isign = sign(sit());
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, a_dir, sit());
              //if there are faces, use the flux at them
              Real aggFlux = 0.0;
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  Real faceFlux = a_flux(face, ivar);
                  aggFlux += faceFlux;
                }
              if(faces.size() > 1)
                aggFlux /= Real(faces.size());

              update += isign*aggFlux;
            }
          //dx is already divided out (part of scale)
          update *= -a_scale;
          a_consState(vof, ivar) = originalState + update;
        }
    }
  //add in covered face fluxes
  for(SideIterator sit; sit.ok(); ++sit)
    {
      int isign = sign(sit());
      const BaseIVFAB<Real>*  coveredFluxPtr;
      const Vector<VolIndex>* coveredFacePtr;
      if(sit() == Side::Lo)
        {
          coveredFluxPtr = &(a_coveredFluxMinu);
          coveredFacePtr = &(a_coveredFaceMinu);
        }                                     
      else                                    
        {                                     
          coveredFluxPtr = &(a_coveredFluxPlus);
          coveredFacePtr = &(a_coveredFacePlus);
        }
      const BaseIVFAB<Real>&  coveredFlux = *coveredFluxPtr;
      const Vector<VolIndex>& coveredFace = *coveredFacePtr;
      for(int ivof = 0; ivof < coveredFace.size(); ivof++)
        {
          const VolIndex& vof = coveredFace[ivof];
          //          const IntVect& iv = vof.gridIndex();
          if(a_box.contains(vof.gridIndex()))
            {
              for(int ivar = 0; ivar < a_consState.nComp(); ivar++)
                {
                  Real covFlux = coveredFlux(vof, ivar);
                  Real update =   isign*covFlux;

                  //dx is already divided out (part of scale)
                  update *= -a_scale;
                  Real state = a_consState(vof, ivar);
                  a_consState(vof, ivar) = state + update;
                }
            }
        }
    }

  floorConserved(a_consState, a_box);
}
/******/
bool 
EBPatchGodunov::
isDefined() const
{
  return 
    m_isDefined && 
    m_isBCSet && 
    m_isBoxSet &&          
    m_isSlopeSet &&
    m_isArtViscSet;
}
/*****************************/
void 
EBPatchGodunov::
setSlopeParameters(bool a_useFourthOrderSlopes, 
                   bool a_useFlattening,
                   bool a_useLimiting)
{
  // Slope flattening is only allowed with 4th order slopes
  assert(a_useFourthOrderSlopes || !a_useFlattening);

  // Store the slope computation parameters
  m_useFourthOrderSlopes = a_useFourthOrderSlopes;
  m_useFlattening        = a_useFlattening;
  m_useLimiting          = a_useLimiting;

  m_isSlopeSet = true;
}
/*****************************/
const EBPhysIBC*
EBPatchGodunov::
getEBPhysIBC() const
{
  assert(m_isBCSet);
  return m_bc;
}
/*****************************/
void 
EBPatchGodunov::
setEBPhysIBC(const EBPhysIBCFactory& a_bcFact)
{
  // Delete old boundary condition object - if any
  if (m_bc != NULL)
    {
      delete m_bc;
    }
  m_bc = a_bcFact.create();
  if(m_isDefined)
    m_bc->define(m_domain, m_dx);
  m_isBCSet = true;
}
/*****************************/
void 
EBPatchGodunov::
setValidBox(const Box&        a_validBox, 
            const EBISBox&    a_ebisBox, 
            const IntVectSet& a_coarseFineIVS,
            const Real&       a_cumulativeTime, 
            const Real&       a_timestep)
{
  m_isBoxSet = true;
  m_dt       = a_timestep;
  m_time     = a_cumulativeTime;
  m_validBox = a_validBox;
  m_ebisBox  = a_ebisBox;

  //define the interpolation stencils
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(m_validBox);
  FaceStop::WhichFaces facestop = FaceStop::SurroundingWithBoundary;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      BaseIFFAB<FaceStencil>& stenDir = m_interpStencils[idir];
      stenDir.define(ivsIrreg, m_ebisBox.getEBGraph(), idir, 1);
      for(FaceIterator faceit(ivsIrreg, m_ebisBox.getEBGraph(), idir, facestop);
          faceit.ok(); ++faceit)
        {
          const  FaceIndex& face = faceit();
          FaceStencil sten = EBArith::getInterpStencil(face, a_coarseFineIVS, 
                                                       m_ebisBox, m_domain.domainBox());
          stenDir(face, 0) = sten;
        }
    }
}
/*****************************/
void
EBPatchGodunov::
define(ProblemDomain& a_domain, 
       const Real& a_dx)
{
  // Store the domain and grid spacing
  m_isDefined = true;
  m_domain = a_domain;
  // Set the domain and grid spacing in the boundary condition object
  m_dx = a_dx;
  if(m_isBCSet)
    m_bc->define(a_domain, a_dx);
}
/******/
void 
EBPatchGodunov::
applyLimiter(EBCellFAB&       a_slopePrim, 
             const EBCellFAB& a_slopeLeft, 
             const EBCellFAB& a_slopeRigh, 
             const int&       a_dir, 
             const Box&       a_box)
{
  assert(isDefined());
  assert(a_slopePrim.getRegion().contains(a_box));
  assert(a_slopeLeft.getRegion().contains(a_box));
  assert(a_slopeRigh.getRegion().contains(a_box));
  assert((a_dir >= 0) && (a_dir < SpaceDim));
  assert(a_slopePrim.nComp() == numSlopes());
  assert(a_slopeLeft.nComp() == numSlopes());
  assert(a_slopeRigh.nComp() == numSlopes());

  const BaseFab<Real>& regSlopeLeft = a_slopeLeft.getSingleValuedFAB();
  const BaseFab<Real>& regSlopeRigh = a_slopeRigh.getSingleValuedFAB();
  BaseFab<Real>& regSlopePrim       = a_slopePrim.getSingleValuedFAB();

  FORT_VLLIMITER(CHF_FRA(regSlopePrim), 
                 CHF_CONST_FRA(regSlopeLeft), 
                 CHF_CONST_FRA(regSlopeRigh), 
                 CHF_BOX(a_box));
      

  IntVectSet ivsMulti = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < numSlopes(); ivar++)
        {
          //the output array arrives with the centered diff value
          const Real& dql = a_slopeLeft(vof, ivar);
          const Real& dqr = a_slopeRigh(vof, ivar);
          Real& dqlim = a_slopePrim(vof, ivar);
              
          FORT_POINTVLLIMITER(CHF_REAL(dqlim), 
                              CHF_CONST_REAL(dql), 
                              CHF_CONST_REAL(dqr)); 
        }
    }
}
/*****************************/
void 
EBPatchGodunov::
slope(EBCellFAB&       a_slopePrim, 
      EBCellFAB&       a_slopeNLim,
      const EBCellFAB& a_primState, 
      const EBCellFAB& a_flattening, 
      const int&       a_dir, 
      const Box&       a_box)
{
  int numSlope = numSlopes(); 
  assert(a_slopePrim.nComp() == numSlopes());
  assert(a_primState.nComp() >= numSlopes());
  EBCellFAB delta2W, deltaWL, deltaWR, deltaWC;

  doSecondOrderSlopes(delta2W,
                      deltaWL,
                      deltaWR,
                      deltaWC,
                      a_primState, 
                      a_dir, 
                      a_box);

  a_slopeNLim.copy(deltaWC);

  if (usesFourthOrderSlopes())
    {
      EBCellFAB delta4W;
      doFourthOrderSlopes(delta4W, 
                          deltaWC,
                          delta2W,
                          deltaWL,
                          deltaWR,
                          a_primState,
                          a_dir,
                          a_box);

      //apply flattening coefficient
      //would be nice to just use *= operator here 
      //but they have different numbers of variables
      if (usesFlattening())
        {
          Box entireBox = delta4W.getRegion();
          const BaseFab<Real>& regFlattening = a_flattening.getSingleValuedFAB();
          BaseFab<Real>& regDelta4W = delta4W.getSingleValuedFAB();
          
          FORT_APPLYFLAT(CHF_FRA(regDelta4W), 
                         CHF_CONST_FRA1(regFlattening, 0), 
                         CHF_CONST_INT(numSlope), 
                         CHF_BOX(entireBox));

          // all single valued cells multiplied in fortran
          IntVectSet multiIVS  = m_ebisBox.getMultiCells(entireBox);
          for(VoFIterator vofit(multiIVS, m_ebisBox.getEBGraph()); 
              vofit.ok(); ++vofit)
            {
              const VolIndex& vof = vofit();
              for(int ivar = 0; ivar < numSlope; ivar++)
                {
                  delta4W(vof, ivar) *= a_flattening(vof, 0);
                }
            }
        }

      a_slopePrim.copy(delta4W);
    }//end if(usefourthorder slopes())
  else
    {
      a_slopePrim.copy(delta2W);
    }
}
/*****************************/
void 
EBPatchGodunov::
doSecondOrderSlopes(EBCellFAB&       a_delta2W,
                    EBCellFAB&       a_deltaWL,
                    EBCellFAB&       a_deltaWR,
                    EBCellFAB&       a_deltaWC,
                    const EBCellFAB& a_primState, 
                    const int&       a_dir, 
                    const Box&       a_box)
{
  assert(m_isSlopeSet);
  int numSlope = numSlopes(); 
  Box box1 = a_box;
  box1.grow(a_dir, 1);
  box1 &= m_domain;

  Box box2 = a_box;
  box2.grow(a_dir, 2);
  box2 &= m_domain;

  Box loBox, hiBox, centerBox, entireBox;
  int hasLo, hasHi;

  if (usesFourthOrderSlopes())
    {
      loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
                 box2, m_domain, a_dir);
    }
  else
    {
      loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
                 box1, m_domain, a_dir);
    }

  a_delta2W.define(m_ebisBox, entireBox, numSlope);
  a_deltaWL.define(m_ebisBox, entireBox, numSlope);
  a_deltaWR.define(m_ebisBox, entireBox, numSlope);
  a_deltaWC.define(m_ebisBox, entireBox, numSlope);
  a_delta2W.setVal(0.);
  a_deltaWL.setVal(0.);
  a_deltaWR.setVal(0.);
  a_deltaWC.setVal(0.);

  BaseFab<Real>& regDelta2W = a_delta2W.getSingleValuedFAB();
  BaseFab<Real>& regDeltaWC = a_deltaWC.getSingleValuedFAB();
  BaseFab<Real>& regDeltaWL = a_deltaWL.getSingleValuedFAB();
  BaseFab<Real>& regDeltaWR = a_deltaWR.getSingleValuedFAB();
  const BaseFab<Real>& regPrimState = a_primState.getSingleValuedFAB();
      
  FORT_SECONDSLOPEDIFFS(CHF_FRA(regDeltaWC), 
                        CHF_FRA(regDeltaWL), 
                        CHF_FRA(regDeltaWR), 
                        CHF_CONST_FRA(regPrimState), 
                        CHF_CONST_INT(numSlope), 
                        CHF_CONST_INT(a_dir), 
                        CHF_BOX(loBox), 
                        CHF_CONST_INT(hasLo), 
                        CHF_BOX(hiBox), 
                        CHF_CONST_INT(hasHi), 
                        CHF_BOX(centerBox));

      
  //apply van leer limiter to regular cells.
  //limiting for irregular cells is more complicated
  //now that we are doing higher order limited
  //one-sided diffs
  regDelta2W.copy(regDeltaWC);
  if(m_useLimiting)
    {
      FORT_VLLIMITER(CHF_FRA(regDelta2W), 
                     CHF_CONST_FRA(regDeltaWL), 
                     CHF_CONST_FRA(regDeltaWR), 
                     CHF_BOX(centerBox));
    }

      
  //at irregular cells, if one side does not exist, 
  //set all to one-sided diffs.  try to do higher order
  //limited diffs if possible.  
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(entireBox);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < numSlopes(); ivar++)
        {
          bool hasFacesLeft, hasFacesRigh;
          Real dql, dqr, dqc;
          bool verbose =false;
          pointGetSlopes(dql, dqr,dqc,
                         hasFacesLeft,
                         hasFacesRigh,
                         vof, a_primState, a_dir, ivar, verbose);
          Real dqSec;
          if(!m_useLimiting)
            {
              dqSec = dqc;
            }
          else
            {
              if(hasFacesLeft && hasFacesRigh)
                {
                  Real dqlim = dqc;
                  
                  FORT_POINTVLLIMITER(CHF_REAL(dqlim), 
                                      CHF_CONST_REAL(dql), 
                                      CHF_CONST_REAL(dqr));
                  dqSec = dqlim;
                }
              else if(!hasFacesLeft && !hasFacesRigh)
                {
                  dqSec = 0.0;
                }
              else if(hasFacesLeft && !hasFacesRigh)
                {
                  if(dqc*dql > 0.0)
                    {
                      Real rsign = 1.0;
                      if(dqc < 0.0)
                        {
                          rsign = -1.0;
                        }
                      dqSec = rsign*Min(Abs(dqc), Abs(dql));
                    }
                  else
                    {
                      dqSec = 0.0;
                    }
              
                }
              else if(hasFacesRigh && !hasFacesLeft)
                {
                  if(dqc*dqr > 0.0)
                    {
                      Real rsign = 1.0;
                      if(dqc < 0.0)
                        {
                          rsign = -1.0;
                        }
                      dqSec = rsign*Min(Abs(dqc), Abs(dqr));
                    }
                  else
                    {
                      dqSec = 0.0;
                    }
                }
              else
                {
                  MayDay::Error("doh! missed a case");
                }
            }//end if using limiting
          a_delta2W(vof, ivar) = dqSec;
          a_deltaWL(vof, ivar) = dql;
          a_deltaWR(vof, ivar) = dqr;
          a_deltaWC(vof, ivar) = dqc;
        } //end loop over variables
    }//end loop over vofs

  //want to be able to call this in test codes
  if(m_isBCSet)
    {
      m_bc->setBndrySlopes(a_delta2W, a_primState, m_ebisBox, entireBox, a_dir);
    }
}

/*****************************/
void EBPatchGodunov::
coveredExtrapSlopes(Real&            a_dq,
                    const VolIndex&  a_vof,
                    const EBCellFAB& a_primState, 
                    const int&       a_dir, 
                    const int&       a_ivar)
{
  bool hasFacesLeft, hasFacesRigh;
  Real dql, dqr;
  bool verbose = false;
  //debug
  //verbose = (m_debugBox.contains(a_vof.gridIndex()) && a_ivar==0);
  //end debug
  pointGetSlopes(dql, dqr,a_dq,
                 hasFacesLeft,
                 hasFacesRigh,
                 a_vof, a_primState, a_dir, a_ivar, verbose);
  if(usesFlattening())
    {
      int pindex = pressureIndex();
      Real press = Max(a_primState(a_vof, pindex), 1.0e-10);
      Real dqp;
      verbose = false;
      pointGetSlopes(dql, dqr, dqp,
                     hasFacesLeft,
                     hasFacesRigh,
                     a_vof, a_primState, a_dir, pindex,verbose);
      if(Abs(dqp/press) > 0.1)
        {
          a_dq = 0.0;
        }
    }
  //debug! set slopes to zero!
  // a_dq = 0;
  //end debug
}
/*****************************/
void EBPatchGodunov::
pointGetSlopes(Real&            a_dql,
               Real&            a_dqr,
               Real&            a_dqc,
               bool&            a_hasFacesLeft,
               bool&            a_hasFacesRigh,
               const VolIndex&  a_vof,
               const EBCellFAB& a_primState, 
               const int&       a_dir, 
               const int&       a_ivar,
               const bool&      a_verbose)
{
  const IntVect&  iv = a_vof.gridIndex();
  //one-sided diffs on domain bndry
  const Box& domainBox = m_domain.domainBox();
  bool onLeftDomain = (iv[a_dir] == domainBox.smallEnd(a_dir));
  bool onRighDomain = (iv[a_dir] == domainBox.bigEnd(a_dir)  );
  VolIndex vofLeft;
  VolIndex vofRigh;
  a_hasFacesLeft = (m_ebisBox.numFaces(a_vof, a_dir, Side::Lo)==1) && !onLeftDomain;
  a_hasFacesRigh = (m_ebisBox.numFaces(a_vof, a_dir, Side::Hi)==1) && !onRighDomain;
      
  if(a_hasFacesLeft)
    {
      Vector<FaceIndex> facesLeft =
        m_ebisBox.getFaces(a_vof, a_dir, Side::Lo);
      vofLeft = facesLeft[0].getVoF(Side::Lo);
    }
          
  if(a_hasFacesRigh)
    {
      Vector<FaceIndex> facesRigh =
        m_ebisBox.getFaces(a_vof, a_dir, Side::Hi);
      vofRigh = facesRigh[0].getVoF(Side::Hi);
    }
          
  Real valCent = a_primState(a_vof, a_ivar);
  Real valLeft = 0.0;
  Real valRigh = 0.0;
  if(a_hasFacesLeft)
    {
      valLeft = a_primState(vofLeft, a_ivar);
    }
  if(a_hasFacesRigh)
    {
      valRigh = a_primState(vofRigh, a_ivar);
    }

  //at irregular cells, if one side does not exist, 
  //set all to one-sided diffs.
  //if neither exists, set all slopes to zero
  //limiting sprinked in here to make higher order
  //one-sided diffs possible
  if(a_hasFacesLeft)
    {
      a_dql = valCent - valLeft;
    }
  if(a_hasFacesRigh)
    {
      a_dqr = valRigh - valCent;
    }
  if(a_hasFacesLeft && a_hasFacesRigh)
    {
      a_dqc = 0.5*(a_dql+a_dqr);
      if(m_useLimiting)
        {
          Real dqmin = Min(Abs(a_dql)*2,Abs(a_dqr)*2);
          if (a_dql*a_dqr < 0.) 
            {
              dqmin = 0.;
            }
          if (a_dqc < 0.) 
            {
              a_dqc = -Min(dqmin,Abs(a_dqc));
            } 
          else
            {
              a_dqc = Min(dqmin,Abs(a_dqc));
            }
        }
    }
  else if(!a_hasFacesLeft && !a_hasFacesRigh)
    {
      a_dql = 0.0;
      a_dqr = 0.0;
      a_dqc = 0.0;
    }
  else if(a_hasFacesLeft && !a_hasFacesRigh)
    {
      a_dqr = a_dql;
      //no higher-order one sided diffs
      a_dqc = a_dql;
    }
  else if(a_hasFacesRigh && !a_hasFacesLeft)
    {
      a_dql = a_dqr;
      //no higher-order one sided diffs
      a_dqc = a_dqr;
    }
  else
    {
      MayDay::Error("doh! missed a case");
    }
  if(a_verbose)
    {
      pout() << "  a_vof="    << a_vof 
             << ", vofLeft= " << vofLeft  
             << ", vofRigh= " << vofRigh  << endl;
      pout() << "  hasFacesLeft=" << a_hasFacesLeft
             << ", hasFacesRigh=" << a_hasFacesRigh;
      pout() << ", valLeft=" << valLeft
             << ", valRigh=" << valRigh
             << ", valCent=" << valCent << endl;
      pout() << "  a_dql=" << a_dql 
             << ", a_dqr=" << a_dqr 
             << ", a_dqc=" << a_dqc << endl;
    }
}
/*****************************/
void EBPatchGodunov::
doFourthOrderSlopes(EBCellFAB&       a_delta4W, 
                    EBCellFAB&       a_deltaWC,
                    const EBCellFAB& a_delta2W,
                    const EBCellFAB& a_deltaWL,
                    const EBCellFAB& a_deltaWR,
                    const EBCellFAB& a_primState,
                    const int&       a_dir,
                    const Box&       a_box)
{
  Box loBox, hiBox, centerBox, entireBox;
  int hasLo, hasHi;
  int numSlope = numSlopes(); 
  const Box& domainBox = m_domain.domainBox();
  Box box1 = a_box;
  box1.grow(a_dir, 1);
  box1 &= m_domain;
  loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
             box1, m_domain, a_dir);

  a_delta4W.define(m_ebisBox, entireBox, numSlope);

  BaseFab<Real>& regDelta4W = a_delta4W.getSingleValuedFAB();

  const BaseFab<Real>& regDelta2W = a_delta2W.getSingleValuedFAB();
  //const BaseFab<Real>& regDeltaWC = a_deltaWC.getSingleValuedFAB();
  //const BaseFab<Real>& regDeltaWL = a_deltaWL.getSingleValuedFAB();
  //const BaseFab<Real>& regDeltaWR = a_deltaWR.getSingleValuedFAB();
  const BaseFab<Real>& regPrimState = a_primState.getSingleValuedFAB();
      
  FORT_FORTHSLOPEDIFFS(CHF_FRA(regDelta4W), 
                       CHF_CONST_FRA(regPrimState), 
                       CHF_CONST_FRA(regDelta2W), 
                       CHF_CONST_INT(numSlope), 
                       CHF_CONST_INT(a_dir), 
                       CHF_BOX(loBox), 
                       CHF_CONST_INT(hasLo), 
                       CHF_BOX(hiBox), 
                       CHF_CONST_INT(hasHi), 
                       CHF_BOX(centerBox));
      
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    { 
      const VolIndex& vof = vofit();
      const IntVect&  iv = vof.gridIndex();
      //one-sided diffs on domain bndry
      bool onLeftDomain = iv[a_dir] == domainBox.smallEnd(a_dir);
      bool onRighDomain = iv[a_dir] == domainBox.bigEnd(a_dir);
      bool hasFacesLeft = (m_ebisBox.numFaces(vof, a_dir, Side::Lo) > 0) && !onLeftDomain;
      bool hasFacesRigh = (m_ebisBox.numFaces(vof, a_dir, Side::Hi) > 0) && !onRighDomain;
          
      for(int ivar = 0; ivar < numSlopes(); ivar++)
        {
          Real dq4;
          if(hasFacesRigh && hasFacesLeft)
            {
              Vector<FaceIndex> facesLeft = 
                m_ebisBox.getFaces(vof, a_dir, Side::Lo);
              Real valLeft = 0.0;
              Real sloLeft = 0.0;
              for(int iface = 0; iface < facesLeft.size(); iface++)
                {
                  VolIndex vofLeft = facesLeft[iface].getVoF(Side::Lo);
                  valLeft += a_primState(vofLeft, ivar);
                  sloLeft +=      a_delta2W(vofLeft, ivar);
                }
              valLeft /= Real(facesLeft.size());
              sloLeft /= Real(facesLeft.size());

              Vector<FaceIndex> facesRigh = 
                m_ebisBox.getFaces(vof, a_dir, Side::Hi);
              Real valRigh = 0.0;
              Real sloRigh = 0.0;
              for(int iface = 0; iface <facesRigh.size(); iface++)
                {
                  VolIndex vofRigh = facesRigh[iface].getVoF(Side::Hi);
                  valRigh += a_primState(vofRigh, ivar);
                  sloRigh +=     a_delta2W(vofRigh, ivar);
                }
              valRigh /= Real(facesRigh.size());
              sloRigh /= Real(facesRigh.size());
                  
              Real dwr = valRigh - 0.25*sloRigh;
              Real dwl = valLeft + 0.25*sloLeft;
              dq4 = (2.0/3.0)*(dwr - dwl);

            }
          else if(hasFacesRigh || hasFacesLeft)
            {
              dq4 = a_delta2W(vof, ivar);
            }
          else
            {
              //no faces on either side.  set the slope to zero
              dq4 = 0.0;
            }
          a_delta4W(vof, ivar) = dq4;
        } //end loop over variables
    }//end loop over vofs
  
  if(m_useLimiting)
    {
      applyLimiter(a_delta4W, a_deltaWL, a_deltaWR,  a_dir, centerBox);
    }

}
/*****************************/
void 
EBPatchGodunov::
regularUpdate(EBCellFAB&        a_consState, 
              EBFluxFAB&        a_flux, 
              BaseIVFAB<Real>&  a_ebIrregFlux,
              BaseIVFAB<Real>&  a_nonConservativeDivergence, 
              const EBCellFAB&  a_flattening, 
              const EBCellFAB&  a_source, 
              const Box&        a_box,
              const IntVectSet& a_ivsSmall,
              bool              a_verbose)
{
  assert(isDefined());
  int numCons = numConserved();

  assert(a_flux.nComp() >= numConserved());
  assert(a_consState.nComp() >= numConserved());
  setCoveredConsVals(a_consState);

  BaseIVFAB<Real>  coveredFluxMinu[SpaceDim];
  BaseIVFAB<Real>  coveredFluxPlus[SpaceDim];
  Vector<VolIndex> coveredFaceMinu[SpaceDim];
  Vector<VolIndex> coveredFacePlus[SpaceDim];
  EBCellFAB primState;
  EBCellFAB slopePrim[SpaceDim];
  EBCellFAB slopeNLim[SpaceDim];
  

  computeFluxes(a_flux,
                coveredFluxMinu, coveredFluxPlus, 
                coveredFaceMinu, coveredFacePlus, 
                primState,    slopePrim, slopeNLim, 
                a_flattening, a_consState, a_source, 
                a_box, a_verbose);

  //now that we have the fluxes, modify them with 
  //artificial viscosity if that is called for.
  //the artificial viscosity correction to the 
  //embedded boundary flux is computed inside
  //computeebirregflux
  if(usesArtificialViscosity())
    {
      EBFluxFAB 
        openDivU(m_ebisBox, a_box, 1);

      getFaceDivergence(openDivU, 
                        primState, slopeNLim, 
                        a_box, a_ivsSmall);
      
      applyArtificialViscosity(a_flux,
                               coveredFluxMinu,
                               coveredFluxPlus,
                               coveredFaceMinu,
                               coveredFacePlus,
                               a_consState,
                               openDivU,
                               a_box,
                               a_ivsSmall);
    }
                      
  //compute nonconservative divergences of fluxes
  EBCellFAB nonConsDivF(m_ebisBox, a_box, numConserved());

  //this is the stable, non-conservative estimate of the solution update
  nonconservativeDivergence(nonConsDivF, a_flux, 
                            coveredFluxMinu, coveredFluxPlus, 
                            coveredFaceMinu, coveredFacePlus, 
                            a_box);

  //copy nonconservative div f into sparse output thingy
  IntVectSet ivsIrreg = a_ivsSmall;
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < numCons; ivar++)
        {
          a_nonConservativeDivergence(vof, ivar) = nonConsDivF(vof, ivar);
        }
    }

  //compute irregular boundary flux.  this includes
  //an artificial viscosity modification if appropriate
  computeEBIrregFlux( a_ebIrregFlux, primState, nonConsDivF,
                      slopeNLim, ivsIrreg, a_source);

  BaseFab<Real>& regConsState   = 
    a_consState.getSingleValuedFAB();
  const BaseFab<Real>& regNonConsDivF = 
    nonConsDivF.getSingleValuedFAB();

  //this updates all single valued cells with the non-conservative divergence.
  FORT_REGUPDATE( CHF_BOX(a_box), 
                  CHF_CONST_FRA(regConsState), 
                  CHF_CONST_FRA(regNonConsDivF), 
                  CHF_CONST_INT(numCons), 
                  CHF_CONST_REAL(m_dt));

  //this does the multi-valued cells.
  IntVectSet ivsMulti = m_ebisBox.getMultiCells(a_box);
  for(VoFIterator vofit(ivsMulti, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      for(int ivar = 0; ivar < numCons; ivar++)
        {
          Real update = -m_dt*nonConsDivF(vof, ivar);
          a_consState(vof, ivar) += update;
        }
    }
  //irregular cells get updated in irregularUpdate
  //floors also happen there.
  // with the linear combination of 
  //divergences thing, taking into account that nonConsDivF has
  //already been included once.
}
void
EBPatchGodunov::
nonconservativeDivergence(EBCellFAB&             a_divF, 
                          const EBFluxFAB&       a_flux, 
                          const BaseIVFAB<Real>  a_coveredFluxMinu[SpaceDim], 
                          const BaseIVFAB<Real>  a_coveredFluxPlus[SpaceDim], 
                          const Vector<VolIndex> a_coveredFaceMinu[SpaceDim], 
                          const Vector<VolIndex> a_coveredFacePlus[SpaceDim], 
                          const Box&             a_box)
{
  assert(m_isDefined);
  assert(m_isBoxSet);
  assert(a_divF.nComp() >= numConserved());
  assert(a_flux.nComp() >= numConserved());
  assert(a_flux.getRegion().contains(a_box));
  assert(a_divF.getRegion().contains(a_box));

  //set the divergence initially to zero
  //then loop through directions and increment the divergence 
  //with each directions flux difference.
  int ncons = numConserved();
  a_divF.setVal(0.0);
  BaseFab<Real>&       regDivF = a_divF.getSingleValuedFAB();
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //update for the regular vofs in the nonconservative 
      //case  works for all single valued vofs.
      const EBFaceFAB& fluxDir = a_flux[idir];
      
      const BaseFab<Real>& regFluxDir = fluxDir.getSingleValuedFAB();

      /* do the regular vofs */
          
      FORT_DIVERGEF( CHF_BOX(a_box), 
                     CHF_FRA(regDivF), 
                     CHF_CONST_FRA(regFluxDir), 
                     CHF_CONST_INT(idir), 
                     CHF_CONST_INT(ncons),
                     CHF_CONST_REAL(m_dx)); 
            
    }
  //update the irregular vofsn
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //divergence was set in regular update.  we reset it
      // to zero and recalc.
      for(int ivar = 0; ivar < ncons; ivar++)
        {
          Real irregDiv = 0.0;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              const EBFaceFAB& fluxDir = a_flux[idir];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  int isign = sign(sit());
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, idir, sit());
                  Real update = 0.;
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      Real flux = fluxDir(face, ivar);
                      update += isign*flux;
                    }
                  if(faces.size() > 1)
                    update /= Real(faces.size());
                  irregDiv += update/m_dx;
                } //end loop over sides
            }//end loop over directions
          a_divF(vof, ivar) = irregDiv;
        }//end loop over variables
    }//end loop over irreg vofs

  //now correct for the covered fluxes
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      for(SideIterator sit; sit.ok(); ++sit)
        {
          int isign = sign(sit());
          const BaseIVFAB<Real>*  coveredFluxPtr;
          const Vector<VolIndex>* coveredFacePtr;
          if(sit() == Side::Lo)
            {
              coveredFluxPtr = &a_coveredFluxMinu[idir];
              coveredFacePtr = &a_coveredFaceMinu[idir];
            }
          else
            {
              coveredFluxPtr = &a_coveredFluxPlus[idir];
              coveredFacePtr = &a_coveredFacePlus[idir];
            }
          const BaseIVFAB<Real>&  coveredFlux = *coveredFluxPtr;
          const Vector<VolIndex>& coveredFace = *coveredFacePtr;
          for(int ivof = 0; ivof < coveredFace.size(); ivof++)
            {
              const VolIndex& vof = coveredFace[ivof];
              if(a_box.contains(vof.gridIndex()))
                {
                  //face on this side is covered.  use covered flux.
                  for(int ivar = 0; ivar < ncons; ivar++)
                    {
                      Real flux = coveredFlux(vof, ivar);
                      Real update = isign*flux/m_dx;
                      a_divF(vof, ivar) += update;
                    }
                }
            }
        }
    }
}
/*****************************/
void
EBPatchGodunov::
consUndividedDivergence(BaseIVFAB<Real>&       a_divF, 
                        const BaseIFFAB<Real>  a_centroidFlux[SpaceDim], 
                        const BaseIVFAB<Real>& a_ebIrregFlux,
                        const IntVectSet&      a_ivs)
{
  assert(m_isDefined);
  assert(m_isBoxSet);
  assert(a_divF.nComp() >= numConserved());
  int ncons = numConserved();
  a_divF.setVal(0.0);
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();

      Real bndryArea = m_ebisBox.bndryArea(vof);
      for(int ivar = 0; ivar < ncons ; ivar++)
        {
          Real bndryFlux= a_ebIrregFlux(vof,ivar);
          Real update = 0.;
          for( int idir = 0; idir < SpaceDim; idir++)
            {
              const BaseIFFAB<Real>& fluxDir = a_centroidFlux[idir];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  int isign = sign(sit());
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, idir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      Real areaFrac = m_ebisBox.areaFrac(face);
                      Real faceFlux =fluxDir(face, ivar);
                      update += isign*areaFrac*faceFlux;
                      
                    }
                }
            }

          //add EB boundary condtions in divergence
          update += bndryFlux*bndryArea; 
          update /= m_dx; //note NOT divided by volfrac

          a_divF(vof, ivar) = update;
        } //end loop over variables
    } //end loop over vofs
}
void
EBPatchGodunov::
interpolateFluxToCentroids(BaseIFFAB<Real>              a_centroidFlux[SpaceDim], 
                           const BaseIFFAB<Real>* const a_fluxInterpolant[SpaceDim], 
                           const IntVectSet&            a_irregIVS)
{
  FaceStop::WhichFaces stopCrit = FaceStop::SurroundingWithBoundary;

  //could just use numFluxes but this allows unit testing
  //with a single variable interpolant 
  int nflux = a_centroidFlux[0].nComp();
  //now loop through the irregular faces
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      BaseIFFAB<Real>& fluxDir = a_centroidFlux[faceDir];
      const BaseIFFAB<Real>& interpol = *(a_fluxInterpolant[faceDir]);
      const BaseIFFAB<FaceStencil>& stencils = m_interpStencils[faceDir];

      for(FaceIterator faceit(a_irregIVS, m_ebisBox.getEBGraph(), faceDir, stopCrit);
          faceit.ok(); ++faceit)
        {
          const FaceIndex& centFace = faceit();
          //          Real     areaFrac = m_ebisBox.areaFrac(centFace);
          //          RealVect centroid = m_ebisBox.centroid(centFace);
          const FaceStencil& sten = stencils(centFace, 0);
          for(int ivar = 0; ivar < nflux; ivar++)
            {
              //              Real centVal = interpol(centFace, ivar);
              Real newFlux = 0.0;
              for(int isten = 0; isten < sten.size(); isten++)
                {
                  const FaceIndex& stenFace= sten.face(isten);
                  Real weight = sten.weight(isten);
                  Real interpFlux = interpol(stenFace, ivar);
                  newFlux += weight*interpFlux;
                }
              fluxDir(centFace, ivar) = newFlux;
            }

          for(int ivar = 0; ivar < numFluxes(); ivar++)
            {
              //debug!  turn off interpolation
              //fluxDir(centFace, ivar) = interpol(centFace, ivar);
            }
        }
    }
}
/*****************************/
void 
EBPatchGodunov::
finalUpdate(EBCellFAB&              a_consState, 
            BaseIVFAB<Real>&        a_massDiff, 
            const BaseIVFAB<Real>&  a_nonConsDivF, 
            const BaseIVFAB<Real>&  a_conservDivF, 
            const IntVectSet&       a_ivs)
{
  assert(a_nonConsDivF.nComp() == numConserved());
  assert(a_conservDivF.nComp() == numConserved());
  assert(a_consState.nComp()   == numConserved());
  assert(a_massDiff.nComp()    == numConserved());

  //initialize mass difference to zero
  a_massDiff.setVal(0.0);
  int ncons = numConserved();
  //update the irreg vofs
  for(VoFIterator vofit(a_ivs, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //      const IntVect& iv = vof.gridIndex();
      Real volFrac = m_ebisBox.volFrac(vof);
      for(int ivar = 0; ivar < ncons ; ivar++)
        {
          //volume fraction not divided out when divergence
          //calculated.  so we do not multiply here.

          //conservative divergence already is already
          //multiplied by volfrac
          Real kapConsDiv = a_conservDivF(vof, ivar);
          Real ncDivergeF = a_nonConsDivF(vof, ivar);;

          //dt*DF^NC has already been added in regular update
          //subtract back off to get back to the old state
          Real updateIncrementNC =  -m_dt*ncDivergeF;
          Real oldState = a_consState(vof, ivar) - updateIncrementNC;

          //un+1 = un -dt(kappa*divfcons + (1-kapp)divfnc)
          Real newState = oldState 
            - m_dt*kapConsDiv
            - m_dt*(1.0-volFrac)*ncDivergeF;
          //dm =-(1-kappa)*kappa*dt*divfcons- kappa*dt*(1-kappa)*divfnc
          //dm =-(1-kappa)*dt*(kappa*divfcons- kappa*divfnc)
          Real deltaM = 
            -(1.0-volFrac)*m_dt*(kapConsDiv - volFrac*ncDivergeF);

          a_consState(vof, ivar) = newState;
          a_massDiff(vof, ivar) = deltaM;
        }
    }
}

void 
EBPatchGodunov::
computeFluxes(EBFluxFAB&       a_flux, 
              BaseIVFAB<Real>  a_coveredFluxMinu[SpaceDim], 
              BaseIVFAB<Real>  a_coveredFluxPlus[SpaceDim], 
              Vector<VolIndex> a_coveredFaceMinu[SpaceDim], 
              Vector<VolIndex> a_coveredFacePlus[SpaceDim], 
              EBCellFAB&       a_primState,
              EBCellFAB        a_slopePrim[SpaceDim],
              EBCellFAB        a_slopeNLim[SpaceDim],
              const EBCellFAB& a_flattening,
              const EBCellFAB& a_consState, 
              const EBCellFAB& a_source, 
              const Box&       a_box,
              bool             a_verbose)
{

  EBCellFAB primPlus[SpaceDim];
  EBCellFAB primMinu[SpaceDim];
#if CH_SPACEDIM==2

  extrapolatePrim2D(primMinu, primPlus,
                    a_primState,    a_slopePrim, a_slopeNLim, 
                    a_flattening, a_consState, a_source, 
                    a_box, a_verbose);

#elif CH_SPACEDIM==3

  extrapolatePrim3D(primMinu,   primPlus,
                    a_primState,  a_slopePrim, a_slopeNLim, 
                    a_flattening, a_consState, a_source, 
                    a_box, a_verbose);

#else
  MayDay::Error("bogus SpaceDim");
#endif


  int numFlux = numFluxes();
  int numPrim = numPrimitives();
  BaseIVFAB<Real> extendedStateMinu[SpaceDim];
  BaseIVFAB<Real> extendedStatePlus[SpaceDim];
  IntVectSet        coveredSetsPlus[SpaceDim];
  IntVectSet        coveredSetsMinu[SpaceDim];
  Box faceBox[SpaceDim];
  //this keeps the fluxes from being calculated
  //on boundaries
  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
    {
      faceBox[dir1] = a_box;
      faceBox[dir1].grow(dir1,1);
      faceBox[dir1] &= m_domain;
      faceBox[dir1].grow(dir1,-1);
      faceBox[dir1].surroundingNodes(dir1);
    }
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      IntVectSet irregIVSPlus, irregIVSMinu;
      computeCoveredFaces(a_coveredFacePlus[idir], 
                          coveredSetsPlus[idir], 
                          irregIVSPlus,
                          idir, Side::Hi, a_box);
      computeCoveredFaces(a_coveredFaceMinu[idir], 
                          coveredSetsMinu[idir], 
                          irregIVSMinu,
                          idir, Side::Lo, a_box);
      extendedStatePlus[idir].define(coveredSetsPlus[idir], 
                                     m_ebisBox.getEBGraph(), numPrim);
      extendedStateMinu[idir].define(coveredSetsMinu[idir], 
                                     m_ebisBox.getEBGraph(), numPrim);
      a_coveredFluxPlus[idir].define(coveredSetsPlus[idir],
                                     m_ebisBox.getEBGraph(), numFlux);
      a_coveredFluxMinu[idir].define(coveredSetsMinu[idir], 
                                     m_ebisBox.getEBGraph(), numFlux);


      extendedStatePlus[idir].setVal(0.);
      extendedStateMinu[idir].setVal(0.);
      a_coveredFluxPlus[idir].setVal(0.);
      a_coveredFluxMinu[idir].setVal(0.);
    }
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      //extrapolate to covered faces using updated extrapolated state
      extrapToCoveredFaces(extendedStateMinu[faceDir],
                           primMinu[faceDir],
                           primPlus[faceDir],
                           a_primState,
                           a_slopeNLim,
                           a_coveredFaceMinu[faceDir],
                           faceDir, Side::Lo, a_box);

      extrapToCoveredFaces(extendedStatePlus[faceDir],
                           primMinu[faceDir],
                           primPlus[faceDir],
                           a_primState,
                           a_slopeNLim,
                           a_coveredFacePlus[faceDir],
                           faceDir, Side::Hi, a_box);



      //eqyatuin 1.19 top
      riemann(a_flux[faceDir], primPlus[faceDir], primMinu[faceDir], 
              faceDir, faceBox[faceDir]);

      // Use the user supplied PhysBC object to obtain boundary fluxes
      m_bc->fluxBC(a_flux[faceDir], a_primState, primMinu[faceDir], 
                   Side::Lo,  m_time, m_ebisBox, a_box, faceDir);
      m_bc->fluxBC(a_flux[faceDir], a_primState, primPlus[faceDir], 
                   Side::Hi,  m_time, m_ebisBox, a_box, faceDir);

      //solve riemann problem between extended state and 
      //value in vof to get covered flux.  
      //Equation 1.19 bottom.
      riemann(a_coveredFluxMinu[faceDir], 
              extendedStateMinu[faceDir], primMinu[faceDir], 
              a_coveredFaceMinu[faceDir], faceDir, Side::Lo);
      riemann(a_coveredFluxPlus[faceDir], 
              extendedStatePlus[faceDir], primPlus[faceDir], 
              a_coveredFacePlus[faceDir], faceDir, Side::Hi);
    }
}
/******/
/******/
void 
EBPatchGodunov::
extrapolatePrim2D(EBCellFAB        a_primMinu[SpaceDim], 
                  EBCellFAB        a_primPlus[SpaceDim], 
                  EBCellFAB&       a_primState,
                  EBCellFAB        a_slopesPrim[SpaceDim],
                  EBCellFAB        a_slopesSeco[SpaceDim],
                  const EBCellFAB& a_flattening,
                  const EBCellFAB& a_consState, 
                  const EBCellFAB& a_source, 
                  const Box&       a_box,
                  bool             a_verbose)
{
  int numPrim = numPrimitives();

  //now define the plethora of data holders that I need
 
  //intermediate flux
  EBFaceFAB fluxOne[SpaceDim];


  //now do the actual computation.
  //1. transform to primitive state
  Box primBox = a_consState.getRegion();
  assert(m_domain.contains(primBox));
  a_primState.define(m_ebisBox, primBox, numPrim);
  int logflag = 0;
  consToPrim(a_primState, a_consState, primBox, logflag);

  //2.  compute slopes delta^d w
  //compute normal derivative stuff
  //const EBFaceFAB& fluxOne0 = fluxOne[0];
  //const EBFaceFAB& fluxOne1 = fluxOne[1];
  BaseIVFAB<Real>  coveredFluxNormMinu[SpaceDim]; 
  BaseIVFAB<Real>  coveredFluxNormPlus[SpaceDim]; 
  Vector<VolIndex> coveredFaceNormMinu[SpaceDim]; 
  Vector<VolIndex> coveredFaceNormPlus[SpaceDim]; 

  doNormalDerivativeExtr2D(a_primMinu,
                           a_primPlus,
                           fluxOne,
                           coveredFluxNormMinu, 
                           coveredFluxNormPlus, 
                           coveredFaceNormMinu, 
                           coveredFaceNormPlus, 
                           a_slopesPrim,
                           a_slopesSeco,
                           a_flattening,
                           a_primState,
                           a_source,
                           a_box);

  // Do the final corrections to the fluxes
  //this only happens on the non-ghosted box
  finalExtrap2D(a_primMinu,
                a_primPlus,
                coveredFluxNormMinu, 
                coveredFluxNormPlus, 
                coveredFaceNormMinu, 
                coveredFaceNormPlus, 
                fluxOne,
                a_primState,
                a_slopesPrim,
                a_slopesSeco,
                a_box
                );

}

void 
EBPatchGodunov::
extrapolatePrim3D(EBCellFAB        a_primMinu[SpaceDim], 
                  EBCellFAB        a_primPlus[SpaceDim], 
                  EBCellFAB&       a_primState,
                  EBCellFAB        a_slopesPrim[SpaceDim],
                  EBCellFAB        a_slopesSeco[SpaceDim],
                  const EBCellFAB& a_flattening,
                  const EBCellFAB& a_consState, 
                  const EBCellFAB& a_source, 
                  const Box&       a_box,
                  bool             a_verbose)
{
  //define the plethora of data holders that I need
 
  //intermediate flux
  EBFaceFAB fluxOne[SpaceDim];
  //data holders specific to 3d
  BaseIVFAB<Real>  coveredFlux3DMinu[SpaceDim][SpaceDim];
  BaseIVFAB<Real>  coveredFlux3DPlus[SpaceDim][SpaceDim];
  EBFaceFAB               fluxTwoVec[SpaceDim][SpaceDim];
  Vector<VolIndex> coveredFace3DMinu[SpaceDim][SpaceDim];
  Vector<VolIndex> coveredFace3DPlus[SpaceDim][SpaceDim];


  //now do the actual computation.
  //1. transform to primitive state
  int numPrim = numPrimitives();
  Box primBox = a_consState.getRegion();
  assert(m_domain.contains(primBox));
  a_primState.define(m_ebisBox, primBox, numPrim);
  int logflag = 0;
  consToPrim(a_primState, a_consState, primBox, logflag);

  //2.  compute slopes delta^d w
  //compute normal derivative stuff
  //const EBFaceFAB& fluxOne0 = fluxOne[0];
  //const EBFaceFAB& fluxOne1 = fluxOne[1];
  BaseIVFAB<Real>  coveredFluxNormMinu[SpaceDim]; 
  BaseIVFAB<Real>  coveredFluxNormPlus[SpaceDim]; 
  Vector<VolIndex> coveredFaceNormMinu[SpaceDim]; 
  Vector<VolIndex> coveredFaceNormPlus[SpaceDim]; 
  doNormalDerivativeExtr3D(a_primMinu,
                           a_primPlus,
                           fluxOne,
                           coveredFluxNormMinu, 
                           coveredFluxNormPlus, 
                           coveredFaceNormMinu, 
                           coveredFaceNormPlus, 
                           a_slopesPrim,
                           a_slopesSeco,
                           a_flattening,
                           a_primState,
                           a_source,
                           a_box);

  //6. In 3D compute corrections to u corresponding to one set
  //   of transverse derivatives appropriate to obtain (1, 1, 1)
  //   coupling.
  // In 3D, compute some additional intermediate fluxes
  // NOTE:  The diagonal entries of this array of fluxes are not
  // used and will not be defined.

  
  do111coupling(fluxTwoVec,
                coveredFlux3DMinu, 
                coveredFlux3DPlus, 
                coveredFace3DMinu, 
                coveredFace3DPlus, 
                a_primMinu,
                a_primPlus,
                coveredFluxNormMinu, 
                coveredFluxNormPlus, 
                coveredFaceNormMinu, 
                coveredFaceNormPlus, 
                fluxOne,
                a_primState,
                a_slopesPrim,
                a_slopesSeco,
                a_box);

  finalExtrap3D(a_primMinu,
                a_primPlus,
                coveredFlux3DMinu, 
                coveredFlux3DPlus, 
                coveredFace3DMinu, 
                coveredFace3DPlus, 
                fluxTwoVec,
                a_primState,
                a_slopesPrim,
                a_slopesSeco,
                a_box);
}
/******************/
void EBPatchGodunov::
finalExtrap2D(EBCellFAB              a_primMinu[SpaceDim],
              EBCellFAB              a_primPlus[SpaceDim],
              const BaseIVFAB<Real>  a_coveredFluxNormMinu[SpaceDim], 
              const BaseIVFAB<Real>  a_coveredFluxNormPlus[SpaceDim], 
              const Vector<VolIndex> a_coveredFaceNormMinu[SpaceDim], 
              const Vector<VolIndex> a_coveredFaceNormPlus[SpaceDim], 
              const EBFaceFAB        a_fluxOne[SpaceDim],
              const EBCellFAB&       a_primState,
              const EBCellFAB        a_slopesPrim[SpaceDim],
              const EBCellFAB        a_slopesSeco[SpaceDim],
              const Box&             a_box)
              
{
  Box slopeBoxG1 = grow(a_box, 1);
  slopeBoxG1 &= m_domain;


  //7. compute fnal corrections to W due to the 
  //final transverse derivatives.
  // Do the final corrections to the fluxes
  for (int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      // Correct the flux using fluxes in the remaining direction(s)
      for (int diffDir = 0; diffDir < SpaceDim; diffDir++)
        {
          // A different direction has been found
          if (diffDir != faceDir)
            {

              // In 2D,  
              //the current primitive state is updated by a flux in
              // the other direction 
              //a_coveredFlux holds the covered flux 
              //due to the normal derivative
              //equation 1.18 line 3.
              updatePrim(a_primMinu[faceDir], 
                         a_primPlus[faceDir], a_fluxOne[diffDir], 
                         a_coveredFluxNormMinu[diffDir], a_coveredFluxNormPlus[diffDir], 
                         a_coveredFaceNormMinu[diffDir], a_coveredFaceNormPlus[diffDir], 
                         diffDir,  slopeBoxG1, (0.5)*m_dt/m_dx);

            } //end if dir2 != faceDir
        } //loop over dir2
      

    } //loop over facedir

}
/***********/
void EBPatchGodunov::
finalExtrap3D(EBCellFAB              a_primMinu[SpaceDim],
              EBCellFAB              a_primPlus[SpaceDim],
              const BaseIVFAB<Real>  a_coveredFlux3DMinu[SpaceDim][SpaceDim], 
              const BaseIVFAB<Real>  a_coveredFlux3DPlus[SpaceDim][SpaceDim], 
              const Vector<VolIndex> a_coveredFace3DMinu[SpaceDim][SpaceDim], 
              const Vector<VolIndex> a_coveredFace3DPlus[SpaceDim][SpaceDim], 
              const EBFaceFAB        a_fluxTwoVec[SpaceDim][SpaceDim],
              const EBCellFAB&       a_primState,
              const EBCellFAB        a_slopesPrim[SpaceDim],
              const EBCellFAB        a_slopesSeco[SpaceDim],
              const Box&             a_box)
              
{
  Box slopeBoxG1 = grow(a_box, 1);
  slopeBoxG1 &= m_domain;

  //7. compute fnal corrections to W due to the 
  //final transverse derivatives.
  // Do the final corrections to the fluxes
  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
    {
      // Correct the flux using fluxes in the remaining direction(s)
      for (int dir2 = 0; dir2 < SpaceDim; dir2++)
        {
          // A different direction has been found
          if (dir2 != dir1)
            {
              // In 3D,  find a direction different from the two above
              // here we need to use the coveredFlux1D for the covered flux
              
              int dir3 = 3 - dir2  - dir1;
              // A different direction has been found
              //const EBFluxFAB& fluxTwo = a_fluxTwoVec[dir3];
              // Update the conservative state 
              //using both corrected fluxes in
              // the other two directions.  Equation 1.18 line 4
              updatePrim(a_primMinu[dir1], a_primPlus[dir1], 
                         a_fluxTwoVec[dir2][dir3], 
                         a_coveredFlux3DMinu[dir2][dir3], 
                         a_coveredFlux3DPlus[dir2][dir3], 
                         a_coveredFace3DMinu[dir2][dir3], 
                         a_coveredFace3DPlus[dir2][dir3], 
                         dir2,  slopeBoxG1,  (0.5)*m_dt/m_dx);

            }
        } //loop over dir2

    } //loop over dir1

}
/*******************/
void EBPatchGodunov::
do111coupling(EBFaceFAB              a_fluxTwoVec[SpaceDim][SpaceDim],
              BaseIVFAB<Real>        a_coveredFlux3DMinu[SpaceDim][SpaceDim], 
              BaseIVFAB<Real>        a_coveredFlux3DPlus[SpaceDim][SpaceDim], 
              Vector<VolIndex>       a_coveredFace3DMinu[SpaceDim][SpaceDim], 
              Vector<VolIndex>       a_coveredFace3DPlus[SpaceDim][SpaceDim], 
              const EBCellFAB        a_primMinu[SpaceDim],
              const EBCellFAB        a_primPlus[SpaceDim],
              const BaseIVFAB<Real>  a_coveredFluxNormMinu[SpaceDim], 
              const BaseIVFAB<Real>  a_coveredFluxNormPlus[SpaceDim], 
              const Vector<VolIndex> a_coveredFaceNormMinu[SpaceDim], 
              const Vector<VolIndex> a_coveredFaceNormPlus[SpaceDim], 
              const EBFaceFAB        a_fluxOne[SpaceDim],
              const EBCellFAB&       a_primState,
              const EBCellFAB        a_slopesPrim[SpaceDim],
              const EBCellFAB        a_slopesSeco[SpaceDim],
              const Box&             a_box)
{
  Box slopeBoxG1 = grow(a_box, 1);
  Box slopeBoxG2 = grow(a_box, 2);
  slopeBoxG1 &= m_domain;
  slopeBoxG2 &= m_domain;
  Box faceBox[SpaceDim][SpaceDim];
  Box modBoxCov[SpaceDim];
  Box modBoxOpen[SpaceDim][SpaceDim];
  for (int faceDir = 0; faceDir < SpaceDim; ++faceDir)
    {
      for (int diffDir = 0; diffDir < SpaceDim; ++diffDir)
        {

          modBoxOpen[faceDir][diffDir] = grow(a_box, 2);
          modBoxOpen[faceDir][diffDir].grow(diffDir, -1);
          modBoxOpen[faceDir][diffDir] &= m_domain;
        }
    }

  for (int modDir = 0; modDir < SpaceDim; ++modDir)
    {
      modBoxCov[modDir] = grow(a_box, 1);
      modBoxCov[modDir] &= m_domain;
      for (int faceDir = 0; faceDir < SpaceDim; ++faceDir)
        {
          faceBox[faceDir][modDir] = modBoxCov[modDir];
          faceBox[faceDir][modDir].grow(faceDir, 1);
          faceBox[faceDir][modDir] &= m_domain;
          faceBox[faceDir][modDir].grow(faceDir, -1);
          faceBox[faceDir][modDir].surroundingNodes(faceDir);
        }
    }
  int numPrim = numPrimitives();
  int numFlux = numFluxes();

  BaseIVFAB<Real> extendedState3DMinu[SpaceDim][SpaceDim];
  BaseIVFAB<Real> extendedState3DPlus[SpaceDim][SpaceDim];
  IntVectSet      coveredSets3DPlus[SpaceDim][SpaceDim];
  IntVectSet      coveredSets3DMinu[SpaceDim][SpaceDim];
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      for(int diffDir = 0; diffDir < SpaceDim; diffDir++)
        {
          if (diffDir != faceDir)
            {
              //first the definitions
              int modDir = 3 - faceDir - diffDir;
              IntVectSet irregIVSPlus, irregIVSMinu;
              computeCoveredFaces(a_coveredFace3DPlus[faceDir][diffDir], 
                                  coveredSets3DPlus[faceDir][diffDir], 
                                  irregIVSPlus, faceDir, Side::Hi, modBoxCov[modDir]);

              computeCoveredFaces(a_coveredFace3DMinu[faceDir][diffDir], 
                                  coveredSets3DMinu[faceDir][diffDir], 
                                  irregIVSMinu, faceDir, Side::Lo, modBoxCov[modDir]);

              a_coveredFlux3DPlus[faceDir][diffDir].define(coveredSets3DPlus[faceDir][diffDir], 
                                                           m_ebisBox.getEBGraph(), numFlux);
              a_coveredFlux3DMinu[faceDir][diffDir].define(coveredSets3DMinu[faceDir][diffDir], 
                                                           m_ebisBox.getEBGraph(), numFlux);

              extendedState3DMinu[faceDir][diffDir].define(coveredSets3DMinu[faceDir][diffDir],
                                                           m_ebisBox.getEBGraph(), numPrim);
              extendedState3DPlus[faceDir][diffDir].define(coveredSets3DPlus[faceDir][diffDir],
                                                           m_ebisBox.getEBGraph(), numPrim);

              a_fluxTwoVec[faceDir][diffDir].define(m_ebisBox, modBoxCov[modDir], 
                                                    faceDir, numFlux);

              a_coveredFlux3DPlus[faceDir][diffDir].setVal(0.);
              a_coveredFlux3DMinu[faceDir][diffDir].setVal(0.);
              a_fluxTwoVec[faceDir][diffDir].setVal(0.);
              extendedState3DMinu[faceDir][diffDir].setVal(0.);
              extendedState3DPlus[faceDir][diffDir].setVal(0.);

              //now for the actual computation
              // Temporary primitive variables
              EBCellFAB primMinuTemp(m_ebisBox, slopeBoxG2, numPrim);
              EBCellFAB primPlusTemp(m_ebisBox, slopeBoxG2, numPrim);

              // Copy data for in place modification
              primMinuTemp.copy(a_primMinu[faceDir]);
              primPlusTemp.copy(a_primPlus[faceDir]);

              // Update the current, extrapolated primitive 
              //variable using a flux
              // in a different direction.  uses covered fluxes calculated
              // above.  Equation 1.12
              updatePrim(primMinuTemp, primPlusTemp, a_fluxOne[diffDir],  
                         a_coveredFluxNormMinu[diffDir], a_coveredFluxNormPlus[diffDir], 
                         a_coveredFaceNormMinu[diffDir], a_coveredFaceNormPlus[diffDir], 
                         diffDir,  modBoxOpen[faceDir][diffDir],  (1.0/3.0)*m_dt/m_dx);


              extrapToCoveredFaces(extendedState3DMinu[faceDir][diffDir],
                                   primMinuTemp,
                                   primPlusTemp,
                                   a_primState,
                                   a_slopesSeco,
                                   a_coveredFace3DMinu[faceDir][diffDir],
                                   faceDir, Side::Lo, modBoxCov[modDir]);

              extrapToCoveredFaces(extendedState3DPlus[faceDir][diffDir],
                                   primMinuTemp,
                                   primPlusTemp,
                                   a_primState,
                                   a_slopesSeco,
                                   a_coveredFace3DPlus[faceDir][diffDir],
                                   faceDir, Side::Hi, modBoxCov[modDir]);

              // Solve the Riemann problem and get fluxes.  Eqution 1.15
              riemann(a_fluxTwoVec[faceDir][diffDir], primPlusTemp, primMinuTemp, 
                      faceDir, faceBox[faceDir][modDir]);

              m_bc->fluxBC(a_fluxTwoVec[faceDir][diffDir],  a_primState, primMinuTemp,  
                           Side::Lo,  m_time, m_ebisBox, a_box, faceDir);
              m_bc->fluxBC(a_fluxTwoVec[faceDir][diffDir],  a_primState, primPlusTemp,
                           Side::Hi,  m_time, m_ebisBox, a_box, faceDir);
              //solve riemann problem between extended state and 
              //value in vof to get covered flux.  Equation 1.14
              riemann(a_coveredFlux3DMinu[faceDir][diffDir], 
                      extendedState3DMinu[faceDir][diffDir], primMinuTemp, 
                      a_coveredFace3DMinu[faceDir][diffDir], faceDir, Side::Lo);
              riemann(a_coveredFlux3DPlus[faceDir][diffDir], 
                      extendedState3DPlus[faceDir][diffDir], primPlusTemp, 
                      a_coveredFace3DPlus[faceDir][diffDir], faceDir, Side::Hi);

            }
        }
    }
}
/*******************/
void EBPatchGodunov::
doNormalDerivativeExtr3D(EBCellFAB              a_primMinu[SpaceDim],
                         EBCellFAB              a_primPlus[SpaceDim],
                         EBFaceFAB              a_fluxOne[SpaceDim],
                         BaseIVFAB<Real>        a_coveredFluxNormMinu[SpaceDim], 
                         BaseIVFAB<Real>        a_coveredFluxNormPlus[SpaceDim], 
                         Vector<VolIndex>       a_coveredFaceNormMinu[SpaceDim], 
                         Vector<VolIndex>       a_coveredFaceNormPlus[SpaceDim], 
                         EBCellFAB              a_slopesPrim[SpaceDim],
                         EBCellFAB              a_slopesSeco[SpaceDim],
                         const EBCellFAB&       a_flattening,
                         const EBCellFAB&       a_primState,
                         const EBCellFAB&       a_source,
                         const Box&             a_box)

{
  Box modBoxCov[SpaceDim];
  Box modBoxOpen[SpaceDim];
  IntVectSet        coveredSetsNormMinu[SpaceDim];
  IntVectSet        coveredSetsNormPlus[SpaceDim];
  BaseIVFAB<Real> extendedStateNormMinu[SpaceDim];
  BaseIVFAB<Real> extendedStateNormPlus[SpaceDim];
  Box faceBox[SpaceDim];
  int numPrim = numPrimitives();
  int numFlux = numFluxes();
  int numSlop = numSlopes();
  //all one big loop since no cross terms here
  for (int idir = 0; idir < SpaceDim; ++idir)
    {

      modBoxOpen[idir] = a_box;
      modBoxOpen[idir].grow(3);
      modBoxOpen[idir].grow(idir, -1);
      modBoxOpen[idir] &= m_domain;

      modBoxCov[idir] = a_box;
      modBoxCov[idir].grow(2);
      modBoxCov[idir].grow(idir, -1);
      modBoxCov[idir] &= m_domain;

      
      faceBox[idir] = modBoxCov[idir];
      faceBox[idir].grow(idir, 1);
      faceBox[idir] &= m_domain;
      faceBox[idir].grow(idir, -1);
      faceBox[idir].surroundingNodes(idir);


      IntVectSet irregIVSPlus, irregIVSMinu;
      computeCoveredFaces(a_coveredFaceNormPlus[idir], 
                          coveredSetsNormPlus[idir], 
                          irregIVSPlus,
                          idir, Side::Hi, modBoxCov[idir]);
      computeCoveredFaces(a_coveredFaceNormMinu[idir], 
                          coveredSetsNormMinu[idir], 
                          irregIVSMinu,
                          idir, Side::Lo,  modBoxCov[idir]);
      extendedStateNormPlus[idir].define(coveredSetsNormPlus[idir], 
                                         m_ebisBox.getEBGraph(), numPrim);
      extendedStateNormMinu[idir].define(coveredSetsNormMinu[idir], 
                                         m_ebisBox.getEBGraph(), numPrim);
      a_coveredFluxNormPlus[idir].define(coveredSetsNormPlus[idir],
                                         m_ebisBox.getEBGraph(), numFlux);
      a_coveredFluxNormMinu[idir].define(coveredSetsNormMinu[idir], 
                                         m_ebisBox.getEBGraph(), numFlux);

      a_primPlus[idir].define(  m_ebisBox, modBoxOpen[idir], numPrim);
      a_primMinu[idir].define(  m_ebisBox, modBoxOpen[idir], numPrim);
      a_slopesPrim[idir].define(m_ebisBox, modBoxOpen[idir], numSlop);
      a_slopesSeco[idir].define(m_ebisBox, modBoxOpen[idir], numSlop);
      a_fluxOne[idir].define(m_ebisBox,    modBoxCov[idir], idir, numFlux);

      extendedStateNormPlus[idir].setVal(0.);
      extendedStateNormMinu[idir].setVal(0.);
      a_coveredFluxNormPlus[idir].setVal(0.);
      a_coveredFluxNormMinu[idir].setVal(0.);
      a_primPlus[idir].setVal(0.);
      a_primMinu[idir].setVal(0.);
      a_fluxOne[idir].setVal(0.);
      
  
      //check flattening coefficients have correct box and compute slopes
      if(usesFlattening()) 
        assert(a_flattening.getRegion().contains(modBoxOpen[idir]));

      //compute slopes
      slope(a_slopesPrim[idir], a_slopesSeco[idir], a_primState, 
            a_flattening, idir, modBoxOpen[idir]);


      //3. Compute the effect of the normal derivative terms and the 
      //source term on the extrapolation in space and time from cell 
      //centers to faces. Equation 1.7 (top).
      normalPred(a_primMinu[idir], a_primPlus[idir], 
                 a_primState, a_slopesPrim[idir], 
                 m_dt/m_dx, idir, modBoxOpen[idir]);

      // If the source term is valid add it to the primitive quantities
      //if source term is not defined, assume it is zero.  lots of
      //problems will have no sources.  not the most elegant method
      //but it beats passing a null pointer.
      if(a_source.isDefined())
        {
          incrementWithSource(a_primMinu[idir], a_source, 
                              0.5*m_dt, modBoxOpen[idir]);;
          incrementWithSource(a_primPlus[idir], a_source, 
                              0.5*m_dt, modBoxOpen[idir]);;
        }

      //4. compute estimates of the flux suitable for computing
      // 1D flux derivatives using a riemann solver for the interior R
      //and for the boundary RB.  Equation 1.8
      riemann(a_fluxOne[idir], a_primPlus[idir], a_primMinu[idir], 
              idir, faceBox[idir]);
      
      // Use the user supplied PhysBC object to obtain boundary fluxes
      m_bc->fluxBC(a_fluxOne[idir], a_primState, a_primMinu[idir], 
                   Side::Lo, m_time, m_ebisBox, a_box, idir);
      m_bc->fluxBC(a_fluxOne[idir], a_primState, a_primPlus[idir], 
                   Side::Hi, m_time, m_ebisBox, a_box, idir);

      //extrapolate to covered faces
      extrapToCoveredFaces(extendedStateNormMinu[idir],
                           a_primMinu[idir],
                           a_primPlus[idir],
                           a_primState,
                           a_slopesSeco,
                           a_coveredFaceNormMinu[idir],
                           idir, Side::Lo, modBoxCov[idir]);

      extrapToCoveredFaces(extendedStateNormPlus[idir],
                           a_primMinu[idir],
                           a_primPlus[idir],
                           a_primState,
                           a_slopesSeco,
                           a_coveredFaceNormPlus[idir],
                           idir, Side::Hi, modBoxCov[idir]);
      
      //solve riemann problem between extended state and 
      //value in vof to get covered flux  Equation 1.9
      riemann(a_coveredFluxNormMinu[idir], 
              extendedStateNormMinu[idir], a_primMinu[idir], 
              a_coveredFaceNormMinu[idir], idir, Side::Lo);
      riemann(a_coveredFluxNormPlus[idir], 
              extendedStateNormPlus[idir], a_primPlus[idir], 
              a_coveredFaceNormPlus[idir], idir, Side::Hi);


    }
}
/*******************/
void EBPatchGodunov::
doNormalDerivativeExtr2D(EBCellFAB              a_primMinu[SpaceDim],
                         EBCellFAB              a_primPlus[SpaceDim],
                         EBFaceFAB              a_fluxOne[SpaceDim],
                         BaseIVFAB<Real>        a_coveredFluxNormMinu[SpaceDim], 
                         BaseIVFAB<Real>        a_coveredFluxNormPlus[SpaceDim], 
                         Vector<VolIndex>       a_coveredFaceNormMinu[SpaceDim], 
                         Vector<VolIndex>       a_coveredFaceNormPlus[SpaceDim], 
                         EBCellFAB              a_slopesPrim[SpaceDim],
                         EBCellFAB              a_slopesSeco[SpaceDim],
                         const EBCellFAB&       a_flattening,
                         const EBCellFAB&       a_primState,
                         const EBCellFAB&       a_source,
                         const Box&             a_box)

{
  Box slopeBoxG1 = grow(a_box, 1);
  Box slopeBoxG2 = grow(a_box, 2);
  slopeBoxG1 &= m_domain;
  slopeBoxG2 &= m_domain;
  Box cellBoxG2[SpaceDim];
  Box cellBoxG1[SpaceDim];
  IntVectSet        coveredSetsNormMinu[SpaceDim];
  IntVectSet        coveredSetsNormPlus[SpaceDim];
  BaseIVFAB<Real> extendedStateNormMinu[SpaceDim];
  BaseIVFAB<Real> extendedStateNormPlus[SpaceDim];
  Box faceBox[SpaceDim];
  int numPrim = numPrimitives();
  int numFlux = numFluxes();
  int numSlop = numSlopes();
  for (int idir = 0; idir < SpaceDim; ++idir)
    {

      cellBoxG2[idir] = a_box;
      cellBoxG2[idir].grow(2);
      cellBoxG2[idir].grow(idir, -1);
      cellBoxG2[idir] &= m_domain;

      cellBoxG1[idir] = a_box;
      cellBoxG1[idir].grow(1);
      cellBoxG1[idir].grow(idir, -1);
      cellBoxG1[idir] &= m_domain;

      faceBox[idir] = a_box;
      faceBox[idir].grow(2);
      faceBox[idir] &= m_domain;
      faceBox[idir].grow(idir, -1);
      faceBox[idir].surroundingNodes(idir);
    }
  

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      IntVectSet irregIVSPlus, irregIVSMinu;
      computeCoveredFaces(a_coveredFaceNormPlus[idir], 
                          coveredSetsNormPlus[idir], 
                          irregIVSPlus,
                          idir, Side::Hi, slopeBoxG1);
      computeCoveredFaces(a_coveredFaceNormMinu[idir], 
                          coveredSetsNormMinu[idir], 
                          irregIVSMinu,
                          idir, Side::Lo, slopeBoxG1);
      extendedStateNormPlus[idir].define(coveredSetsNormPlus[idir], 
                                         m_ebisBox.getEBGraph(), numPrim);
      extendedStateNormMinu[idir].define(coveredSetsNormMinu[idir], 
                                         m_ebisBox.getEBGraph(), numPrim);
      a_coveredFluxNormPlus[idir].define(coveredSetsNormPlus[idir],
                                         m_ebisBox.getEBGraph(), numFlux);
      a_coveredFluxNormMinu[idir].define(coveredSetsNormMinu[idir], 
                                         m_ebisBox.getEBGraph(), numFlux);

      a_primPlus[idir].define(m_ebisBox,     slopeBoxG2, numPrim);
      a_primMinu[idir].define(m_ebisBox,     slopeBoxG2, numPrim);
      a_slopesPrim[idir].define(m_ebisBox, slopeBoxG2, numSlop);
      a_slopesSeco[idir].define(m_ebisBox, slopeBoxG2, numSlop);
      a_fluxOne[idir].define(m_ebisBox,cellBoxG2[idir], idir, numFlux);

      extendedStateNormPlus[idir].setVal(0.);
      extendedStateNormMinu[idir].setVal(0.);
      a_coveredFluxNormPlus[idir].setVal(0.);
      a_coveredFluxNormMinu[idir].setVal(0.);
      a_primPlus[idir].setVal(0.);
      a_primMinu[idir].setVal(0.);
      a_fluxOne[idir].setVal(0.);
    }

  
  //check flattening coefficients have correct box
  if(usesFlattening()) 
    assert(a_flattening.getRegion().contains(slopeBoxG2));

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      slope(a_slopesPrim[idir], a_slopesSeco[idir], a_primState, 
            a_flattening, idir, slopeBoxG2);
    }

  for(int idir = 0; idir < SpaceDim; idir++)
    {

      //3. Compute the effect of the normal derivative terms and the 
      //source term on the extrapolation in space and time from cell 
      //centers to faces. Equation 1.7 (top).
       
      normalPred(a_primMinu[idir], a_primPlus[idir], 
                 a_primState, a_slopesPrim[idir], 
                 m_dt/m_dx, idir, slopeBoxG2);

      // If the source term is valid add it to the primitive quantities
      //if source term is not defined, assume it is zero.  lots of
      //problems will have no sources.  not the most elegant method
      //but it beats passing a null pointer.
      if(a_source.isDefined())
        {
          incrementWithSource(a_primMinu[idir], a_source, 
                              0.5*m_dt, slopeBoxG2);;
          incrementWithSource(a_primPlus[idir], a_source, 
                              0.5*m_dt, slopeBoxG2);;
        }

      //4. compute estimates of the flux suitable for computing
      // 1D flux derivatives using a riemann solver for the interior R
      //and for the boundary RB.  Equation 1.8
      riemann(a_fluxOne[idir], a_primPlus[idir], a_primMinu[idir], 
              idir, faceBox[idir]);
      
      // Use the user supplied PhysBC object to obtain boundary fluxes
      m_bc->fluxBC(a_fluxOne[idir], a_primState, a_primMinu[idir], 
                   Side::Lo, m_time, m_ebisBox, a_box, idir);
      m_bc->fluxBC(a_fluxOne[idir], a_primState, a_primPlus[idir], 
                   Side::Hi, m_time, m_ebisBox, a_box, idir);

      //extrapolate to covered faces
      extrapToCoveredFaces(extendedStateNormMinu[idir],
                           a_primMinu[idir],
                           a_primPlus[idir],
                           a_primState,
                           a_slopesSeco,
                           a_coveredFaceNormMinu[idir],
                           idir, Side::Lo, slopeBoxG1);

      extrapToCoveredFaces(extendedStateNormPlus[idir],
                           a_primMinu[idir],
                           a_primPlus[idir],
                           a_primState,
                           a_slopesSeco,
                           a_coveredFaceNormPlus[idir],
                           idir, Side::Hi, slopeBoxG1);
      
      //solve riemann problem between extended state and 
      //value in vof to get covered flux  Equation 1.9
      riemann(a_coveredFluxNormMinu[idir], 
              extendedStateNormMinu[idir], a_primMinu[idir], 
              a_coveredFaceNormMinu[idir], idir, Side::Lo);
      riemann(a_coveredFluxNormPlus[idir], 
              extendedStateNormPlus[idir], a_primPlus[idir], 
              a_coveredFaceNormPlus[idir], idir, Side::Hi);


    }
}
/*****************************/
/*****************************/
void
EBPatchGodunov::
computeCoveredFaces(Vector<VolIndex>&     a_coveredFace, 
                    IntVectSet&           a_coveredSets,
                    IntVectSet&           a_irregIVS,
                    const int&            a_idir, 
                    const Side::LoHiSide& a_sd, 
                    const Box&            a_region)
{
  EBArith::computeCoveredFaces(a_coveredFace, a_coveredSets, 
                               a_irregIVS, a_idir, a_sd, 
                               m_ebisBox, a_region);
}
/*****************************/
/*****************************/
void EBPatchGodunov::
getFaceDivergence(EBFluxFAB&        a_openDivU, 
                  const EBCellFAB&  a_primState, 
                  const EBCellFAB   a_slopePrim[SpaceDim],
                  const Box&        a_box,
                  const IntVectSet& a_ivsIrreg)
{
  //first set open divergence to zero because
  //we will compute the divergence additively
  assert(a_openDivU.nComp() == 1);
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      a_openDivU[faceDir].setVal(0.);
    }

  //compute divergence on all cells as though they
  //were regular.  Because clarity is not the only virtue
  //in life, we will do this in fortran
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      //this box calc stuff poached directly from
      //the non-eb version.  The fortran used here
      //is slightly different since i am using the
      //existing slopes instead of recalculating the
      //gradients.

      
      // Now, we can calculate the divergence of the normal velocity
      // at the center normal-direction edges. To do this, we determine
      // which edges at which we have sufficient data to compute centered
      // estimates of h*(div(u)). At the remaining edges. i.e. those
      // corresponding to the physical boundaries, we use zeroth-order
      // extrapolation.
      EBFaceFAB& divUDir = a_openDivU[faceDir];

      Box divBox = divUDir.getRegion();
      divBox.enclosedCells(faceDir);
      divBox.grow(faceDir,1);
      divBox &= m_domain;

      Box loBox,hiBox,centerBox,entireBox;
      int hasLo,hasHi;

      loHiCenterFace(loBox,hasLo,hiBox,hasHi,centerBox,entireBox,
                     divBox,m_domain,faceDir);

      // All of the boxes computed above are shifted so as to be cell-centered,
      // with the index of the cell center being identified with the low edge.
      // We then shift a_divVel to be compatible with that convention on input,
      // and undo the shift on output.  Nutty though this may seem,
      // it still beats the crap out of rewriting it.

      loBox.shiftHalf(faceDir,1);
      centerBox.shiftHalf(faceDir,1);
      hiBox.shiftHalf(faceDir,1);
      
      BaseFab<Real>& regOpenDivU = divUDir.getSingleValuedFAB();
      const BaseFab<Real>& regPrimState = a_primState.getSingleValuedFAB();
      Interval velInt = velocityInterval();
      int inormVel = velInt.begin() + faceDir;

      //this does nothing 
      regOpenDivU.shiftHalf(faceDir,1);

      
      //put in normal direction velocity comp
      FORT_DIVUONED(CHF_FRA1(regOpenDivU,0),
                    CHF_CONST_FRA1(regPrimState,inormVel),
                    CHF_CONST_INT(faceDir),
                    CHF_BOX(centerBox));
      
      //add in slopes
      for(int tranDir = 0; tranDir < SpaceDim; tranDir++)
        {
          if(tranDir != faceDir)
            {
              const BaseFab<Real>& regSlopeDir =  a_slopePrim[tranDir].getSingleValuedFAB();
              Interval velInt = velocityInterval();
              int itranVel = velInt.begin() + tranDir;
              FORT_DIVUTRAN(CHF_FRA1(regOpenDivU,0),
                            CHF_CONST_FRA1(regSlopeDir,itranVel),
                            CHF_CONST_INT(faceDir),
                            CHF_BOX(centerBox));
            }
        }

      //enforce bcs
      FORT_DIVUEDGE(CHF_FRA1(regOpenDivU,0),
                    CHF_CONST_INT(faceDir),
                    CHF_BOX(loBox),
                    CHF_CONST_INT(hasLo),
                    CHF_BOX(hiBox),
                    CHF_CONST_INT(hasHi));

      //this does nothing.
      regOpenDivU.shiftHalf(faceDir,-1);

    }

  //now for the irregular cells
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      EBFaceFAB& openDiv = a_openDivU[faceDir];
      FaceIterator faceit(a_ivsIrreg, m_ebisBox.getEBGraph(), faceDir, 
                          FaceStop::SurroundingNoBoundary);

      //reset the divergence to zero
      for(faceit.reset(); faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          openDiv(face, 0) = 0.0;
        }
      //loop through divergence directions and add
      //in diffs
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          //use the normal velocity diff if the direction
          //is the same direction as the face normal.
          //use the ave of velocity slopes otherwise

          Interval velInt = velocityInterval();
          int velInd = velInt.begin() + idir;
          if(idir == faceDir) //use velocity diff
            {
              for(faceit.reset(); faceit.ok(); ++faceit)
                {
                  const FaceIndex& face = faceit();
                  Real normalDiv = 0.0;
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      VolIndex vof = face.getVoF(sit());
                      Real velDir = a_primState(vof, velInd);
                      normalDiv += sign(sit())*velDir;
                    }
                  openDiv(face, 0) += normalDiv;
                }
            }
          else //use ave of slopes
            {
              const EBCellFAB& slopeDir = a_slopePrim[idir];
              for(faceit.reset(); faceit.ok(); ++faceit)
                {
                  const FaceIndex& face = faceit();
                  Real tranDiv = 0.0;
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      VolIndex vof = face.getVoF(sit());
                      Real velSlop = slopeDir(vof, velInd);
                      //the 0.5 is there because we are doing an average
                      tranDiv += 0.5*velSlop;
                    }
                  openDiv(face, 0) += tranDiv;
                }
            }
          
        }
    }
}
/*****************************/
void EBPatchGodunov::
applyArtificialViscosity(EBFluxFAB&             a_openFlux,
                         BaseIVFAB<Real>        a_coveredFluxMinu[SpaceDim], 
                         BaseIVFAB<Real>        a_coveredFluxPlus[SpaceDim], 
                         const Vector<VolIndex> a_coveredFaceMinu[SpaceDim], 
                         const Vector<VolIndex> a_coveredFacePlus[SpaceDim], 
                         const EBCellFAB&       a_consState,
                         const EBFluxFAB&       a_divVel,
                         const Box&             a_box,
                         const IntVectSet&      a_ivsIrreg)
{
  // Get the artificial viscosity coefficient
  Real coeff = artificialViscosityCoefficient();

  Box faceBox[SpaceDim];
  for (int dir1 = 0; dir1 < SpaceDim; ++dir1)
    {
      faceBox[dir1] = a_box;
      faceBox[dir1].grow(dir1,1);
      faceBox[dir1] &= m_domain;
      faceBox[dir1].grow(dir1,-1);
      faceBox[dir1].surroundingNodes(dir1);
    }
  //copy the flux into a temp array so to
  //not screw things up at irregular cells
  //since this is an update in place
  int numCons = numConserved();
  assert(a_openFlux.getRegion() == a_box);
  EBFluxFAB fluxSave(m_ebisBox, a_box, numCons);

  Interval interv(0, numCons-1);
  fluxSave.copy(a_box, interv, a_box, a_openFlux, interv);

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      const Box& faceBoxDir = faceBox[idir];
      BaseFab<Real>& regFluxDir = a_openFlux[idir].getSingleValuedFAB();
      const BaseFab<Real>& regConsState = a_consState.getSingleValuedFAB();
      const BaseFab<Real>& regDivVel = a_divVel[idir].getSingleValuedFAB();
      FORT_ARTVISC(CHF_FRA(regFluxDir),
                   CHF_CONST_FRA(regConsState),
                   CHF_CONST_FRA1(regDivVel,0),
                   CHF_CONST_REAL(coeff),
                   CHF_CONST_INT(idir),
                   CHF_BOX(faceBoxDir),
                   CHF_CONST_INT(numCons),
                   CHF_CONST_REAL(m_dx));
    }

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      EBFaceFAB& fluxDir = a_openFlux[idir];
      const EBFaceFAB& fluxSaveDir = fluxSave[idir];
      const EBFaceFAB& divergeUDir = a_divVel[idir];
      //save the change in the flux so we can 
      //do 0th order extrapolation to the covered faces
      EBFaceFAB fluxChange(m_ebisBox, a_box, idir, numCons);
      fluxChange.setVal(0.);
      for(FaceIterator  faceit(a_ivsIrreg, m_ebisBox.getEBGraph(), idir,
                               FaceStop::SurroundingNoBoundary); 
          faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          VolIndex voflo = face.getVoF(Side::Lo);
          VolIndex vofhi = face.getVoF(Side::Hi);
          for(int ivar = 0; ivar < numCons; ivar++)
            {
              Real dv = divergeUDir(face, 0);
              Real uhi= a_consState(vofhi, ivar);
              Real ulo= a_consState(voflo, ivar);

              Real fsave = fluxSaveDir(face, ivar);
              Real deltaF = -coeff*abs(dv)*(uhi-ulo);

              fluxDir(face, ivar) = fsave + deltaF;
              fluxChange(face, ivar) = deltaF;
              
            }
        }
      //now do the covered face fluxes
      for(SideIterator sit; sit.ok(); ++sit)
        {
          BaseIVFAB<Real>*  coveredFluxPtr;
          const Vector<VolIndex>* coveredFacePtr;
          if(sit() == Side::Lo)
            {
              coveredFacePtr = &a_coveredFaceMinu[idir];
              coveredFluxPtr = &a_coveredFluxMinu[idir];
            }
          else
            {
              coveredFacePtr = &a_coveredFacePlus[idir];
              coveredFluxPtr = &a_coveredFluxPlus[idir];
            }
          BaseIVFAB<Real>&  coveredFlux = *coveredFluxPtr;
          const Vector<VolIndex>& coveredFace = *coveredFacePtr;
          for(int ivof = 0; ivof < coveredFace.size(); ivof++)
            {
              const VolIndex& vof = coveredFace[ivof];
              if(a_box.contains(vof.gridIndex()))
                {
                  Vector<FaceIndex> flipFaces = 
                    m_ebisBox.getFaces(vof, idir, flip(sit()));
                  for(int ivar = 0; ivar < numCons; ivar++)
                    {
                      Real areaTot = 0.;
                      Real diffTot = 0.;
                      for(int iface = 0; iface < flipFaces.size(); iface++)
                        {
                          const FaceIndex& face = flipFaces[iface];
                          areaTot += m_ebisBox.areaFrac(face);
                          diffTot += areaTot*fluxChange(face, ivar);
                        }
                      if(areaTot > 0.0)
                        {
                          diffTot /= areaTot;
                        }
                      coveredFlux(vof, ivar) += diffTot;
                    }
                }
            }
        }
    }
}
/*****************************/
void 
EBPatchGodunov::
irregularUpdate(EBCellFAB&             a_consState, 
                Real&                  a_maxWaveSpeed, 
                BaseIVFAB<Real>&       a_massDiff, 
                const BaseIFFAB<Real>  a_centroidFlux[SpaceDim], 
                const BaseIVFAB<Real>& a_ebIrregFlux,
                const BaseIVFAB<Real>& a_nonConservativeDivergence, 
                const Box&             a_box, 
                const IntVectSet&      a_ivs) 
{
  assert(isDefined());

  //compute conservative divergences of fluxes
  BaseIVFAB<Real> conservativeDivergence(a_ivs, m_ebisBox.getEBGraph(), numConserved());

  consUndividedDivergence(conservativeDivergence, 
                          a_centroidFlux, a_ebIrregFlux, a_ivs);
  
  // Update conserved variables to the next time step using the final
  // flux in each direction.  As directions progress, increment a_massDiff
  finalUpdate(a_consState, a_massDiff, a_nonConservativeDivergence, 
              conservativeDivergence,  a_ivs);

  floorConserved(a_consState, a_box);
  // Get and return the maximum wave speed on this patch/grid
  a_maxWaveSpeed = getMaxWaveSpeed(a_consState, a_box);
}
/*****************************/
void
EBPatchGodunov::
artificialViscosity(bool a_useArtificialVisc)
{
  m_useArtificialVisc = a_useArtificialVisc;
  m_isArtViscSet = true;
}
/*****************************/

void EBPatchGodunov::
computeFlattening(EBCellFAB&       a_flattening, 
                  const EBCellFAB& a_primState, 
                  const Box&       a_box)
{
  assert(isDefined());
  assert(isDefined());
  assert(a_primState.getRegion().contains(a_box));
  assert(usesFourthOrderSlopes());
  assert(a_flattening.nComp() == 1);
  assert(a_flattening.getRegion().contains(a_box));

  BaseFab<Real>& regFlattening = a_flattening.getSingleValuedFAB();
  const BaseFab<Real>& regPrimState  = a_primState.getSingleValuedFAB();

  const Box& domainBox = m_domain.domainBox();
  EBCellFAB zetaDir(m_ebisBox, a_box, SpaceDim);
  EBCellFAB delta1U(m_ebisBox, a_box, SpaceDim);
  BaseFab<Real>& regZetaDir = zetaDir.getSingleValuedFAB();
  BaseFab<Real>& regDelta1U = delta1U.getSingleValuedFAB();

  Interval velInterval= velocityInterval();
  int v0index = velInterval.begin();

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      Box box1 = a_box;
      box1.grow(idir, 1);
      box1 &= m_domain;

      Box box2 = a_box;
      box2.grow(idir, 2);
      box2 &= m_domain;

      Box box3 = a_box;
      box3.grow(idir, 3);
      box3 &= m_domain;

      assert(a_primState.getRegion().contains(box3));

      Box loBox, hiBox, centerBox, entireBox;
      int hasLo, hasHi;

      loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
                 box3, m_domain, idir);

      EBCellFAB delta1P(m_ebisBox, entireBox, 1);
      BaseFab<Real>& regDelta1P = delta1P.getSingleValuedFAB();

      int pressIndex = pressureIndex();
      //compute delta1P
          
      FORT_GETGRAD(CHF_FRA1(regDelta1P, 0), 
                   CHF_CONST_FRA1(regPrimState, pressIndex), 
                   CHF_CONST_INT(idir), 
                   CHF_BOX(loBox), 
                   CHF_CONST_INT(hasLo), 
                   CHF_BOX(hiBox), 
                   CHF_CONST_INT(hasHi), 
                   CHF_BOX(centerBox));
          

      IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(entireBox);
      for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
          vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect&  iv = vof.gridIndex();
          //one-sided diffs on domain bndry
          bool onLeftDomain = iv[idir] == domainBox.smallEnd(idir);
          bool onRighDomain = iv[idir] == domainBox.bigEnd(idir);
          bool hasFacesLeft = (m_ebisBox.numFaces(vof, idir, Side::Lo) > 0) && !onLeftDomain;
          bool hasFacesRigh = (m_ebisBox.numFaces(vof, idir, Side::Hi) > 0) && !onRighDomain;
          
          Real valCent = a_primState(vof, pressIndex);
          Real dpl, dpr, dpc;
          //compute one-sided diffs where you have them
          if(hasFacesLeft)
            {
              Vector<FaceIndex> facesLeft = 
                m_ebisBox.getFaces(vof, idir, Side::Lo);
              Real valLeft = 0.0;
              for(int iface = 0; iface <facesLeft.size(); iface++)
                {
                  VolIndex vofLeft = facesLeft[iface].getVoF(Side::Lo);
                  valLeft += a_primState(vofLeft, pressIndex);
                }
              valLeft /= Real(facesLeft.size());
              dpl = valCent - valLeft;
            }
          if(hasFacesRigh)
            {
              Vector<FaceIndex> facesRigh = 
                m_ebisBox.getFaces(vof, idir, Side::Hi);
              Real valRigh = 0.0;
              for(int iface = 0; iface <facesRigh.size(); iface++)
                {
                  VolIndex vofRigh = facesRigh[iface].getVoF(Side::Hi);
                  valRigh += a_primState(vofRigh, pressIndex);
                }
              valRigh /= Real(facesRigh.size());
              dpr = valRigh - valCent;
            }
          if(hasFacesLeft && hasFacesRigh)
            {
              dpc = 0.5*(dpl+dpr);
            }
          else if(!hasFacesLeft && !hasFacesRigh)
            {
              dpc = 0.0;
            }
          else if(hasFacesLeft && !hasFacesRigh)
            {
              dpc = dpl;
            }
          else if(hasFacesRigh && !hasFacesLeft)
            {
              dpc = dpr;
            }

          delta1P(vof, 0) = dpc;
        }

      loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
                 box2, m_domain, idir);
      ivsIrreg = m_ebisBox.getIrregIVS(entireBox);

      EBCellFAB delta2P(m_ebisBox, entireBox, 1);
      EBCellFAB bulkMin(m_ebisBox, entireBox, 1);
      EBCellFAB zetaTwi(m_ebisBox, entireBox, 1);
      BaseFab<Real>& regDelta2P = delta2P.getSingleValuedFAB();
      BaseFab<Real>& regBulkMin = bulkMin.getSingleValuedFAB();
      BaseFab<Real>& regZetaTwi = zetaTwi.getSingleValuedFAB();

      //compute delta2P
          
      FORT_GETDPTWO(CHF_FRA1(regDelta2P, 0), 
                    CHF_CONST_FRA1(regDelta1P, 0), 
                    CHF_CONST_INT(idir), 
                    CHF_BOX(loBox), 
                    CHF_CONST_INT(hasLo), 
                    CHF_BOX(hiBox), 
                    CHF_CONST_INT(hasHi), 
                    CHF_BOX(centerBox));
          
 
      //compute min3(press, d) = bulkMin
      int bulkIndex = bulkModulusIndex();
          
      FORT_MIN3PTS(CHF_FRA1(regBulkMin, 0), 
                   CHF_CONST_FRA1(regPrimState, bulkIndex), 
                   CHF_CONST_INT(idir), 
                   CHF_BOX(loBox), 
                   CHF_CONST_INT(hasLo), 
                   CHF_BOX(hiBox), 
                   CHF_CONST_INT(hasHi), 
                   CHF_BOX(centerBox));
          
      //compute preliminary flattening coeff
          
      FORT_GETFLAT(CHF_FRA1(regZetaTwi, 0), 
                   CHF_CONST_FRA1(regDelta1P, 0), 
                   CHF_CONST_FRA1(regDelta2P, 0), 
                   CHF_CONST_FRA1(regBulkMin, 0), 
                   CHF_BOX(entireBox));
          
      for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
          vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect&  iv = vof.gridIndex();
          //one-sided diffs on domain bndry
          bool onLeftDomain = iv[idir] == domainBox.smallEnd(idir);
          bool onRighDomain = iv[idir] == domainBox.bigEnd(idir);
          bool hasFacesLeft = (m_ebisBox.numFaces(vof, idir, Side::Lo) > 0) && !onLeftDomain;
          bool hasFacesRigh = (m_ebisBox.numFaces(vof, idir, Side::Hi) > 0) && !onRighDomain;
          
          Real dp2c;
          Real bulkMinL, bulkMinR, bulkMinC;
          
          Real dp1Cent = delta1P(vof, 0);
          Real bulkCent = a_primState(vof, bulkIndex);
          Real dp1Left, dp1Righ, bulkLeft, bulkRigh;
          //compute one-sided diffs where you have them
          if(hasFacesLeft)
            {
              Vector<FaceIndex> facesLeft = 
                m_ebisBox.getFaces(vof, idir, Side::Lo);
              dp1Left = 0.0;
              bulkLeft = 0.0;
              for(int iface = 0; iface <facesLeft.size(); iface++)
                {
                  VolIndex vofLeft = facesLeft[iface].getVoF(Side::Lo);
                  dp1Left += delta1P(vofLeft, 0);
                  bulkLeft +=a_primState(vofLeft, bulkIndex);
                }
              dp1Left /= Real(facesLeft.size());
              bulkLeft /= Real(facesLeft.size());

              bulkMinL = Min(bulkCent, bulkLeft);
            }
          if(hasFacesRigh)
            {
              Vector<FaceIndex> facesRigh = 
                m_ebisBox.getFaces(vof, idir, Side::Hi);
              dp1Righ = 0.0;
              bulkRigh = 0.0;
              for(int iface = 0; iface <facesRigh.size(); iface++)
                {
                  VolIndex vofRigh = facesRigh[iface].getVoF(Side::Hi);
                  dp1Righ += delta1P(vofRigh, 0);
                  bulkRigh += a_primState(vofRigh, bulkIndex);
                }
              dp1Righ /= Real(facesRigh.size());
              bulkRigh /= Real(facesRigh.size());
              bulkMinR = Min(bulkCent, bulkRigh);
            }

          if(hasFacesLeft && hasFacesRigh)
            {
              dp2c = (dp1Left+dp1Righ);
              bulkMinC = Min(bulkMinR, bulkMinL);
            }
          else if(!hasFacesLeft && !hasFacesRigh)
            {
              dp2c = dp1Cent;
              bulkMinC = bulkCent;
            }
          else if(hasFacesLeft && !hasFacesRigh)
            {
              dp2c = dp1Cent + dp1Left;
              bulkMinC = bulkMinL;
            }
          else if(hasFacesRigh && !hasFacesLeft)
            {
              dp2c = dp1Cent + dp1Righ;
              bulkMinC = bulkMinR;
            }
 
          delta2P(vof, 0) = dp2c;
          bulkMin(vof, 0) = bulkMinC;

          //this is that whole zeta(dp1, dp2, p0) func
          Real r0 = 0.75;  Real r1= 0.85; Real d = 0.33;
          Real d1pVoF = Abs(delta1P(vof, 0)) ;
          //
          //     bad idea among many
          Real smallp = 1.0e-7;
          Real d2pVoF = Max(Abs(delta2P(vof, 0)), smallp);

          Real strength = Abs(d1pVoF/bulkMinC);
          Real zetaFuncDP;
          if(strength >= d)
            {
              Real ratio =  d1pVoF/d2pVoF;
              if( ratio <= r0)
                {
                  zetaFuncDP = 1.0;
                }
              else if(ratio >= r1)
                {
                  zetaFuncDP = 0.0;
                }
              else
                {
                  zetaFuncDP= 1.0 - (ratio - r0)/(r1-r0);;
                }
            } 
          else //strength < d
            {
              zetaFuncDP = 1.0;
            }
          zetaTwi(vof, 0) = zetaFuncDP;
        }

      loHiCenter(loBox, hasLo, hiBox, hasHi, centerBox, entireBox, 
                 box1, m_domain, idir);
      ivsIrreg = m_ebisBox.getIrregIVS(entireBox);

      //compute zeta = min3(zetatwid) and delta1U
          
      FORT_MIN3PTS(CHF_FRA1(regZetaDir, idir), 
                   CHF_CONST_FRA1(regZetaTwi, 0), 
                   CHF_CONST_INT(idir), 
                   CHF_BOX(loBox), 
                   CHF_CONST_INT(hasLo), 
                   CHF_BOX(hiBox), 
                   CHF_CONST_INT(hasHi), 
                   CHF_BOX(centerBox));
          

      int velIndex = v0index + idir;
          
      FORT_GETGRAD(CHF_FRA1(regDelta1U, idir), 
                   CHF_CONST_FRA1(regPrimState, velIndex), 
                   CHF_CONST_INT(idir), 
                   CHF_BOX(loBox), 
                   CHF_CONST_INT(hasLo), 
                   CHF_BOX(hiBox), 
                   CHF_CONST_INT(hasHi), 
                   CHF_BOX(centerBox));
          
      for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
          vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect&  iv = vof.gridIndex();
          //one-sided diffs on domain bndry
          bool onLeftDomain = iv[idir] == domainBox.smallEnd(idir);
          bool onRighDomain = iv[idir] == domainBox.bigEnd(idir);
          bool hasFacesLeft = (m_ebisBox.numFaces(vof, idir, Side::Lo) > 0) && !onLeftDomain;
          bool hasFacesRigh = (m_ebisBox.numFaces(vof, idir, Side::Hi) > 0) && !onRighDomain;
          
          Real velLeft, velRigh, du1c;
          Real zetaMinL, zetaMinR, zetaMinC;
          Real zetaLeft, zetaRigh;
          Real velCent = a_primState(vof, velIndex);
          Real zetaCent = zetaTwi(vof, 0);
          //compute one-sided diffs where you have them
          if(hasFacesLeft)
            {
              Vector<FaceIndex> facesLeft = 
                m_ebisBox.getFaces(vof, idir, Side::Lo);
              velLeft  = 0.0;
              zetaLeft = 0.0;
              for(int iface = 0; iface <facesLeft.size(); iface++)
                {
                  VolIndex vofLeft = facesLeft[iface].getVoF(Side::Lo);
                  velLeft += a_primState(vofLeft, velIndex);
                  zetaLeft +=    zetaTwi(vofLeft, 0);
                }
              velLeft /= Real(facesLeft.size());
              zetaLeft /= Real(facesLeft.size());

              zetaMinL = Min(zetaCent, zetaLeft);
            }
          if(hasFacesRigh)
            {
              Vector<FaceIndex> facesRigh = 
                m_ebisBox.getFaces(vof, idir, Side::Hi);
              velRigh = 0.0;
              zetaRigh = 0.0;
              for(int iface = 0; iface <facesRigh.size(); iface++)
                {
                  VolIndex vofRigh = facesRigh[iface].getVoF(Side::Hi);
                  velRigh += a_primState(vofRigh, velIndex);
                  zetaRigh += zetaTwi(vofRigh, 0);
                }
              velRigh /= Real(facesRigh.size());
              zetaRigh /= Real(facesRigh.size());
              zetaMinR = Min(zetaCent, zetaRigh);
            }

          if(hasFacesLeft && hasFacesRigh)
            {
              du1c = 0.5*(velRigh - velLeft);
              zetaMinC = Min(zetaMinR, zetaMinL);
            }
          else if(!hasFacesLeft && !hasFacesRigh)
            {
              du1c = 0.0;
              zetaMinC = zetaCent;
            }
          else if(hasFacesLeft && !hasFacesRigh)
            {
              du1c = velCent-velLeft;
              zetaMinC = zetaMinL;
            }
          else if(hasFacesRigh && !hasFacesLeft)
            {
              du1c = velRigh-velCent;
              zetaMinC = zetaMinR;
            }
 
          delta1U(vof, idir) = du1c;
          zetaDir(vof, idir) = zetaMinC;
        }
    }
  // take the minimum of all directions if the sum of velocity diffs
  // are negative.  unity (no flattening) otherwise.
      
  FORT_MINFLAT(CHF_FRA1(regFlattening, 0), 
               CHF_CONST_FRA(regZetaDir), 
               CHF_CONST_FRA(regDelta1U), 
               CHF_BOX(a_box));
      
  IntVectSet ivsIrreg = m_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, m_ebisBox.getEBGraph()); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      Real velDiffSum = 0.0;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          velDiffSum += delta1U(vof, idir);
        }
      Real finalFlat;
      if(velDiffSum < 0)
        {
          finalFlat = zetaDir(vof, 0);
          for(int idir = 1; idir < SpaceDim; idir++)
            {
              finalFlat = Min(finalFlat, zetaDir(vof, idir));
            }
          assert(finalFlat >= 0.0);
        }
      else
        {
          finalFlat = 1.0;
        }
      a_flattening(vof, 0) = finalFlat;
    }
}
