/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTGraves, Thurs, Aug 16, 2001

#include <fstream>
#include <string>
using std::fstream;
using std::string;
#include <cstdio>
#include <cmath>
#include "CH_HDF5.H"
#include "AMRIO.H"
#include "EBAMRIO.H"
#include "BoxIterator.H"
#include "LayoutIterator.H"
#include "PolyGeom.H"


#ifdef HDF5

void
writeEBHDF5(const string& a_filename, 
            const Vector<DisjointBoxLayout>& a_vectGrids, 
            const Vector<LevelData<EBCellFAB>* > & a_vectData, 
            const Vector<string>& a_vectNames,
            const Box& a_domain,                               
            const Real& a_dx,
            const Real& a_dt,
            const Real& a_time,
            const Vector<int>& a_vectRatio,                       
            const int& a_numLevels,
            const bool& a_replaceCovered,
            const Vector<Real>& a_coveredValues)
{
  assert(a_numLevels > 0);
  assert(a_vectData.size() >= a_numLevels);
  assert(a_vectRatio.size() >= a_numLevels-1);

  IntVect ghostIV = IntVect::Zero;
  Vector<LevelData<FArrayBox>* > chomboData(a_numLevels, NULL);
  int ncomp = (*a_vectData[0]).nComp();
  int nnames = a_vectNames.size();
  Real dx = a_dx;
  int indexVolFrac = ncomp;
  int indexAreaFrac = indexVolFrac+1;
  int indexNormal = indexAreaFrac+2*SpaceDim;
  int indexDist = indexNormal+SpaceDim;
  int ncompTotal = indexDist+1;

  assert(nnames >= ncomp);

  Vector<string> names(ncompTotal);

  for (int i = 0; i < nnames; i++)
    {
      names[i] = a_vectNames[i];
    }

  string volFracName("fraction-0");
  Vector<string> normName(3);
  Vector<string> areaName(6);
  string distName("distance-0");

  normName[0] = "xnormal-0";
  normName[1] = "ynormal-0";
  normName[2] = "znormal-0";

  areaName[0] = "xAreafractionLo-0";
  areaName[1] = "xAreafractionHi-0";
  areaName[2] = "yAreafractionLo-0";
  areaName[3] = "yAreafractionHi-0";
  areaName[4] = "zAreafractionLo-0";
  areaName[5] = "zAreafractionHi-0";

  names[indexVolFrac] = volFracName;

  for (int i = 0; i < 2*SpaceDim; i++)
    {
      names[indexAreaFrac+i] = areaName[i];
    }

  for (int i = 0; i < SpaceDim; i++)
    {
      names[indexNormal+i] = normName[i];
    }

  names[indexDist] = distName;

  // set things up for each level
  for (int ilev = 0; ilev < a_numLevels; ilev++)
    {
      const DisjointBoxLayout& grids = a_vectGrids[ilev];
      const LevelData<EBCellFAB>& ebcfData = *a_vectData[ilev];

      if (a_replaceCovered)
        {
          assert(a_coveredValues.size() == ncomp);
        }

      //copy data into something writeAMRHierarchy can gronk
      chomboData[ilev] = 
        new LevelData<FArrayBox>(grids, ncompTotal, ghostIV);
      LevelData<FArrayBox>& fabData = *chomboData[ilev];

      // go through all the grids on this level
      for (DataIterator dit = grids.dataIterator(); dit.ok(); ++dit)
        {
          const EBCellFAB& ebcellfab = ebcfData[dit()];
          const EBISBox& ebisbox = ebcellfab.getEBISBox();
          FArrayBox& currentFab = fabData[dit()];

          // copy regular data
          currentFab.copy(ebcellfab.getSingleValuedFAB(),0,0,ncomp);

          // set default volume fraction
          currentFab.setVal(1.0,indexVolFrac);

          // set default area fractions
          for (int i = 0; i < 2*SpaceDim; i++)
            {
              currentFab.setVal(1.0,indexAreaFrac+i);
            }

          // set default normal
          for (int i = 0; i < SpaceDim; i++)
            {
              currentFab.setVal(0.0,indexNormal+i);
            }

          // set default distance of EB from corner
          currentFab.setVal(0.0,indexDist);

          // set special values
          // iterate through the current grid
          // NOTE:  this is probably an inefficient way to do this!
          //can be diff than ebcellfab because of ghost
          //Box bregion = ebcellfab.getRegion();
          Box bregion = currentFab.box();
          bregion &= ebisbox.getDomain();
          for (BoxIterator bit(bregion); bit.ok(); ++bit)
            {
              const IntVect& iv = bit();

              // set special values for covered cells
              if (ebisbox.isCovered(iv))
                {
                  // replace regular data if flagged
                  if (a_replaceCovered)
                    {
                      for (int icomp = 0; icomp < ncomp; icomp++)
                        {
                          Real cval = a_coveredValues[icomp];

                          currentFab(iv,icomp) = cval;
                        }
                    }

                  // volume fraction is zero
                  currentFab(iv,indexVolFrac) = 0.0;

                  // area fractions are zero
                  for (int i = 0; i < 2*SpaceDim; i++)
                    {
                      currentFab(iv,indexAreaFrac+i) = 0.0;
                    }
                }

              // set special values for irregular cells
              if (ebisbox.isIrregular(iv))
                {
                  Vector<VolIndex> vofs = ebisbox.getVoFs(iv);
                  Real volFrac = ebisbox.volFrac(vofs[0]);
                  RealVect normal = ebisbox.normal(vofs[0]);
                  Real bndryArea = ebisbox.bndryArea(vofs[0]);

                  if (bndryArea == 0.0)
                    {
                      if (volFrac > 0.5)
                        {
                          volFrac = 1.0;
                        }
                      else
                        {
                          volFrac = 0.0;
                        }

                      normal = RealVect::Zero;
                    }

                  // set volume fraction
                  currentFab(iv,indexVolFrac) = volFrac;

                  // set area fractions
                  for (int i = 0; i < SpaceDim; i++)
                    {
                      Vector<FaceIndex> faces;

                      faces = ebisbox.getFaces(vofs[0],i,Side::Lo);
                      if (faces.size() == 0)
                        {
                          currentFab(iv,indexAreaFrac+2*i) = 0.0;
                        }
                      else
                        {
                          currentFab(iv,indexAreaFrac+2*i) = ebisbox.areaFrac(faces[0]);
                        }

                      faces = ebisbox.getFaces(vofs[0],i,Side::Hi);
                      if (faces.size() == 0)
                        {
                          currentFab(iv,indexAreaFrac+2*i+1) = 0.0;
                        }
                      else
                        {
                          currentFab(iv,indexAreaFrac+2*i+1) = ebisbox.areaFrac(faces[0]);
                        }
                    }

                  // set normal
                  for (int i = 0; i < SpaceDim; i++)
                    {
                      currentFab(iv,indexNormal+i) = normal[i];
                    }

                  // set distance unless the length of the normal is zero
                  Real length = PolyGeom::dot(normal,normal);

                  if (length > 0)
                    {
                      Real dist = PolyGeom::computeAlpha(volFrac,normal) * dx;
                      currentFab(iv,indexDist) = -dist;
                    }
                }
            }
        }

        dx /= a_vectRatio[ilev];
    }

  // write the data
  WriteAMRHierarchyHDF5(a_filename,
                        a_vectGrids,
                        chomboData,
                        names,
                        a_domain,
                        a_dx,
                        a_dt,
                        a_time,
                        a_vectRatio,
                        a_numLevels);

  //clean up memory
  for (int ilev = 0; ilev < a_numLevels; ilev++)
    {
      delete chomboData[ilev];
    }
}

void
writeEBHDF5(const string& a_filename,
            const DisjointBoxLayout& a_grids, 
            const LevelData<EBCellFAB>& a_data, 
            const Vector<string>& a_vectNames,
            const Box& a_domain,                               
            const Real& a_dx,
            const Real& a_dt,
            const Real& a_time,
            const bool& a_replaceCovered,
            const Vector<Real>& a_coveredValues)
{
  int numLevels = 1;
  Vector<LevelData<EBCellFAB>* > vectData(numLevels, 
                                          (LevelData<EBCellFAB>*)&a_data);
  Vector<DisjointBoxLayout> vectGrids(numLevels, a_grids);
  Vector<int> refRat(numLevels,2);

  writeEBHDF5(a_filename,
              vectGrids,
              vectData,
              a_vectNames,
              a_domain, 
              a_dx,
              a_dt,
              a_time,
              refRat,
              numLevels,
              a_replaceCovered,
              a_coveredValues);
}

class LocalEBCellFactory
  : public DataFactory<EBCellFAB>
{
public:
  /// factory function.
  /** 
      Creates a new baseebcellfab object 
      and returns a pointer to it.  Responsiblitly
      for calling operator 'delete' on this pointer is passed to the user. 
  */
  virtual EBCellFAB* create(const Box& a_box, int a_ncomps,
                            const DataIndex& a_dit) const;

  ///
  /**
     create the factory with an ispace.  calls full define function 
  */
  LocalEBCellFactory(const EBCellFAB* a_unmanagedPtr)
  {
    assert (a_unmanagedPtr != NULL);
    m_nonmanagedPtr = a_unmanagedPtr;
  }

  ///
  virtual ~LocalEBCellFactory(){;}

private:

  const EBCellFAB* m_nonmanagedPtr;
};

EBCellFAB*
LocalEBCellFactory::create(const Box& a_box, int a_ncomps,
                           const DataIndex& a_dit) const
{

//  check is broken by EBCellFAB's BaseFab's no longer being 
//  intersected with the domain
//  if(a_box != m_nonmanagedPtr->getRegion())
//    {
//      MayDay::Error("Incorrect use of LocalEBCellFactory, try again");
//    }
  return new EBCellFAB(m_nonmanagedPtr->getEBISBox(), a_box, a_ncomps);
}


void
writeEBHDF5(const string& a_filename,
            const Box& a_grid, 
            const EBCellFAB& a_data, 
            const Vector<string>& a_vectNames,
            const Box& a_domain,                               
            const Real& a_dx,
            const Real& a_dt,
            const Real& a_time,
            const bool& a_replaceCovered,
            const Vector<Real>& a_coveredValues)
{
  Vector<int> procID(1, 0);
  Vector<Box> vecbox(1, a_grid);
  DisjointBoxLayout dbl(vecbox, procID);
  int ncomp = a_data.nComp();
  Interval interv(0, ncomp-1);
  LocalEBCellFactory factory(&a_data);
  LevelData<EBCellFAB> ldf(dbl, ncomp, IntVect::Zero, factory);

  for (DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
    {
      ldf[dit()].copy(a_grid, interv, a_grid, a_data, interv);
    }

  writeEBHDF5(a_filename,
              dbl,
              ldf,
              a_vectNames,
              a_domain,
              a_dx,
              a_dt,
              a_time,
              a_replaceCovered,
              a_coveredValues);
}

// put simple debugging functions here, at the end of the hdf5 stuff

void
writeEBFAB(const EBCellFAB* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "EBfab.hdf5";
  writeEBFABname(a_dataPtr, fname);
}

void
viewEBFAB(const EBCellFAB* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeEBFABname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void
writeEBFABname(const EBCellFAB* a_dataPtr,
               const char*      a_filename)
{
  if (a_dataPtr == NULL) return;

  const BaseFab<Real>& data = a_dataPtr->getSingleValuedFAB();

  int nComp = data.nComp();
  
  Vector<string> names(nComp);
  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      names[ivar] = label;
    }

  Box domain = data.box();

  // put bogus numbers here
  Real dx = 1.0;
  Real dt = 1.0;
  Real time = 1.0;

  bool replaceCovered = false;
  Vector<Real> coveredValues;

  string fileString(a_filename);
 
  writeEBHDF5(fileString,
              domain, 
              *a_dataPtr, 
              names,
              domain,                               
              dx,
              dt,
              time,
              replaceCovered,
              coveredValues);
}

void viewEBFace(const EBFaceFAB* a_face)
{
  const FArrayBox& reg = a_face->getFArrayBox();
  viewFAB(&reg);
}

void 
writeEBLevel(const LevelData<EBCellFAB>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = "EBLDF.hdf5";
  writeEBLevelname(a_dataPtr, fname);
}

void 
viewEBLevel(const LevelData<EBCellFAB>* a_dataPtr)
{
  if (a_dataPtr == NULL) return;

  char* fname = tempnam(NULL,NULL);
  writeEBLevelname(a_dataPtr, fname);

  char command[2000];
  sprintf(command,"$CHOMBOVIS_HOME/bin/chombovis %s &",fname);
  system(command);
}

void 
writeEBLevelname(const LevelData<EBCellFAB>* a_dataPtr,
                 const char*                 a_filename)
{
  if (a_dataPtr == NULL) return;

  const LevelData<EBCellFAB>& data = *a_dataPtr;

  int nComp = data.nComp();
  
  Vector<string> names(nComp);
  for(int ivar = 0; ivar < nComp; ivar++)
    {
      char labelChSt[80];
      sprintf(labelChSt, "component_%d", ivar);
      string label(labelChSt);
      names[ivar] = label;
    }

  // need to figure out what domain will contain this LevelData
  DataIterator dit = data.dataIterator();
  const DisjointBoxLayout& levelBoxes = data.getBoxes();
  dit.reset();
  Box domain = levelBoxes[dit()];
  for (dit.reset(); dit.ok(); ++dit)
    {
      const Box thisBox = levelBoxes[dit()];
      D_TERM(
             if (thisBox.smallEnd(0)<domain.smallEnd(0)) 
                domain.setSmall(0,thisBox.smallEnd(0)); , 
             if (thisBox.smallEnd(1)<domain.smallEnd(1)) 
                domain.setSmall(1,thisBox.smallEnd(1)); , 
             if (thisBox.smallEnd(2)<domain.smallEnd(2)) 
                domain.setSmall(2, thisBox.smallEnd(2)); );

      D_TERM(
             if (thisBox.bigEnd(0)>domain.bigEnd(0)) 
                domain.setBig(0,thisBox.bigEnd(0)); , 
             if (thisBox.bigEnd(1)>domain.bigEnd(1)) 
                domain.setBig(1,thisBox.bigEnd(1)); , 
             if (thisBox.bigEnd(2)>domain.bigEnd(2)) 
                domain.setBig(2, thisBox.bigEnd(2)); );
             
    } // end loop over levels to determine "domain"

  // put bogus numbers here
  Real dx = 1.0;
  Real dt = 1.0;
  Real time = 1.0;

  bool replaceCovered = false;
  Vector<Real> coveredValues;

  string fileString(a_filename);
 
  writeEBHDF5(fileString,
              levelBoxes, 
              *a_dataPtr, 
              names,
              domain,                            
              dx,
              dt,
              time,
              replaceCovered,
              coveredValues);
}

#endif   // HDF5
