/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves mon nov 19 2001

#ifndef _EBCoar_To_Coar_Redist_H_
#define _EBCoar_To_Coar_Redist_H_
#include "REAL.H"
#include "Vector.H"
#include "EBCellFAB.H"
#include "EBFaceFAB.H"
#include "EBISLayout.H"
#include "EBISBox.H"
#include "IntVectSet.H"
#include "CFStencil.H"
#include "LoHiSide.H"
#include "LevelData.H"
#include "LayoutData.H"
#include "RedistStencil.H"


///
/**
   A EBCoarToCoarRedist  handles all the data choreography 
   for coarse to coarse redistribution ("re-redistribution").
 */
class EBCoarToCoarRedist
{
public:
  
  friend class EBFluxRegister;

  ///
  /**
     Default constructor.  Leaves object undefined.
  */
  EBCoarToCoarRedist();

  ///
  ~EBCoarToCoarRedist();

  ///
  /**
     Full define function.   \\
     dblFine, dblCoar: The fine and coarse layouts
     of the data. \\
     ebislFine, ebislCoar: The fine and coarse layouts
     of the geometric description.\\
     nref: The refinement ratio between the two levels.\\
     nvar: The number of variables contained in the data
     at each VoF. \\
     coarStencil: The redistribution stencil for this level.
  */
  void define(const DisjointBoxLayout& a_dblFine,
              const DisjointBoxLayout& a_dblCoar,
              const EBISLayout& a_ebislFine,
              const EBISLayout& a_ebislCoar,
              const Box& a_domainCoar,
              const int& a_nref,
              const int& a_nvar,
              int a_redistRad = 1);

  ///
  /** 
      Initialize values of registers  to zero.
  */
  void setToZero(); 

  ///
  /**
     Increments the register with data from coarseMass.
     This is the full redistribution mass.  Internally
     the class figures out what actually gets re-redistributed.
  */
  void increment(const BaseIVFAB<Real>& a_coarMass,
                 const DataIndex& a_fineDataIndex,
                 const Interval&  a_variables);


  ///
  /**
   */
  void
  resetWeights(const LevelData<EBCellFAB>& a_modifier,
               const int& a_ivar);

  ///
  /**
     Reredistribute the data contained in the internal buffers.
  */
  void redistribute(LevelData<EBCellFAB>& a_coarSolution,
                    const Interval& a_variables);

  ///
  bool isDefined() const;

protected:

  //internal use only
  void setDefaultValues();

  bool m_isDefined;
  int  m_redistRad;
  int  m_nComp;
  int  m_refRat;
  Box  m_domainCoar;

  //input fine grid
  DisjointBoxLayout    m_gridsFine;
  //input coarse grid
  DisjointBoxLayout    m_gridsCoar;

  //buffers on refineed coarse layout
  LevelData<BaseIVFAB<Real> > m_regsCoar;

  //location of mass sources on refined coarse layout
  LayoutData<IntVectSet> m_setsCoar;

  //ebisl of input fine grid
  EBISLayout m_ebislFine;
  //ebisl of input coar grid
  EBISLayout m_ebislCoar;
 
  LayoutData<BaseIVFAB<VoFStencil> > m_stenCoar;
  //need both of these to facilitate mass-weighted 
  //redistribution
  LayoutData<BaseIVFAB<VoFStencil> > m_volumeStenc;
  LayoutData<BaseIVFAB<VoFStencil> > m_standardStenc;
  LevelData<EBCellFAB> m_densityCoar;
 
private:
  //For all the usual reasons,
  //there is no copy constructor for this class.
  //Neither is there an operator= for this class.
  void operator= (const EBCoarToCoarRedist&)
  {
    MayDay::Error("invalid operator");
  }
  EBCoarToCoarRedist(const EBCoarToCoarRedist&)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
