/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves fri aug 17, 2001

#include "EBCoarseAverage.H"
#include "EBAverageF_F.H"
#include "VoFIterator.H"
#include "DisjointBoxLayout.H"
#include "EBCellFactory.H"
#include "EBIndexSpace.H"
#include "EBArith.H"
/************************************/
void
EBCoarseAverage::setDefaultValues()
{
  m_isDefined = false;
  m_refRat = -1;
  m_nComp = -1;
}
/************************************/
EBCoarseAverage::EBCoarseAverage()
{
  setDefaultValues();
}
/************************************/
EBCoarseAverage::~EBCoarseAverage()
{
}
/************************************/
EBCoarseAverage::EBCoarseAverage(const DisjointBoxLayout& a_dblFine,
                                 const DisjointBoxLayout& a_dblCoar,
                                 const EBISLayout& a_ebislFine,
                                 const EBISLayout& a_ebislCoar,
                                 const Box& a_domainCoar,
                                 const int& a_nref,
                                 const int& a_nvar)
{
  setDefaultValues();

  define(a_dblFine, a_dblCoar, a_ebislFine, a_ebislCoar, 
         a_domainCoar, a_nref, a_nvar);
}
/************************************/
void 
EBCoarseAverage::define(const DisjointBoxLayout& a_dblFine,
                        const DisjointBoxLayout& a_dblCoar,
                        const EBISLayout& a_ebislFine,
                        const EBISLayout& a_ebislCoar,
                        const Box& a_domainCoar,
                        const int& a_nref,
                        const int& a_nvar)
{
  assert(a_nref > 0);
  assert(a_nvar > 0);

  m_isDefined = true;
  m_refRat = a_nref;
  m_nComp = a_nvar;
  m_coarGrids = a_dblCoar;
  m_fineGrids = a_dblFine;
  m_coarEBISL = a_ebislCoar;
  m_fineEBISL = a_ebislFine;
  m_coarDomain = a_domainCoar;
  
  IntVect ghostiv = IntVect::Zero;

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());
  int nghost = 0;
  m_coarsenedFineGrids = DisjointBoxLayout();
  coarsen(m_coarsenedFineGrids, m_fineGrids, m_refRat);
  ebisPtr->fillEBISLayout(m_coarsenedFineEBISL, 
                          m_coarsenedFineGrids, 
                          a_domainCoar, nghost);
  m_coarsenedFineEBISL.setMaxRefinementRatio(m_refRat);
  EBCellFactory ebcellfact(m_coarsenedFineEBISL);
  m_coarsenedFineData.define(m_coarsenedFineGrids, m_nComp, 
                             ghostiv, ebcellfact);
}
/************************************/
bool 
EBCoarseAverage::isDefined() const
{
  return m_isDefined;
}
/************************************/
void
EBCoarseAverage::average(LevelData<EBCellFAB>& a_coarData,
                         const LevelData<EBCellFAB>& a_fineData,
                         const Interval& a_variables)
{
  assert(isDefined());
  //first average the data onto the coarsenedFine data
  //then copy over
  for(DataIterator fineit = m_fineGrids.dataIterator(); 
      fineit.ok(); ++fineit)
    {
      averageFAB(m_coarsenedFineData[fineit()], 
                 a_fineData[fineit()], 
                 fineit(),
                 a_variables);
    }
  
  //this is the part that is blocking
  m_coarsenedFineData.copyTo(a_variables, a_coarData, a_variables);
}
/************************************/
void
EBCoarseAverage::averaRZ(LevelData<EBCellFAB>&       a_coarData,
                         const LevelData<EBCellFAB>& a_fineData,
                         const Interval&             a_variables,
                         const Real&                 a_dx)
{
  assert(isDefined());
  //first average the data onto the coarsenedFine data
  //then copy over
  for(DataIterator fineit = m_fineGrids.dataIterator(); 
      fineit.ok(); ++fineit)
    {
      averageFRZ(m_coarsenedFineData[fineit()], 
                 a_fineData[fineit()], 
                 fineit(),
                 a_variables,
                 a_dx);
    }
  
  //this is the part that is blocking
  m_coarsenedFineData.copyTo(a_variables, a_coarData, a_variables);
}
/************************************/
void
EBCoarseAverage::averageFAB(EBCellFAB&       a_coar,
                            const EBCellFAB& a_fine,
                            const DataIndex& a_datInd,
                            const Interval&  a_variables) const
{
  assert(isDefined());
  //do all cells as if they were regular
  BaseFab<Real>& coarRegFAB =  a_coar.getSingleValuedFAB();
  const BaseFab<Real>& fineRegFAB = a_fine.getSingleValuedFAB();
  Box refbox(IntVect::TheZeroVector(), 
             (m_refRat-1)*IntVect::TheUnitVector());
  const Box& coarBox = m_coarsenedFineGrids.get(a_datInd);

#ifndef NDEBUG
  Box fineBox = refine(coarBox, m_refRat);
  assert(coarRegFAB.box().contains(coarBox));
  assert(fineRegFAB.box().contains(fineBox));
#endif
  
  for(int ivar = a_variables.begin(); 
      ivar <= a_variables.end(); ivar++)
    {
      FORT_EBAVERAGE(CHF_FRA1(coarRegFAB,ivar),
                     CHF_CONST_FRA1(fineRegFAB,ivar),
                     CHF_BOX(coarBox),
                     CHF_CONST_INT(m_refRat),
                     CHF_BOX(refbox));          
    }
  
  //overwrite irregular cells

  //recall that datInd is from the fine layout.
  const EBISBox& ebisBoxCoar = m_coarsenedFineEBISL[a_datInd];
  const EBISBox& ebisBoxFine = m_fineEBISL[a_datInd];
  IntVectSet coarIrregIVS = ebisBoxCoar.getIrregIVS(coarBox);

  //fine cell volume is normalized to one.
  //compute coarse volume
  Real dxCoar = Real(m_refRat);
  Real cellVolCoar = 1.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    cellVolCoar *= dxCoar;
  
  for(VoFIterator vofitCoar(coarIrregIVS, ebisBoxCoar.getEBGraph()); 
      vofitCoar.ok(); ++vofitCoar)
    {
      const VolIndex& coarVoF = vofitCoar();
      Vector<VolIndex> fineVoFs =
        m_coarsenedFineEBISL.refine(coarVoF, m_refRat, a_datInd);

      Real volCoar = cellVolCoar*ebisBoxCoar.volFrac(coarVoF);
      for(int ivar = a_variables.begin(); 
          ivar <= a_variables.end(); ivar++)
        {
          Real dataVal = 0.0;
          if(volCoar > 0.)
            {
              for(int ifine = 0; ifine < fineVoFs.size(); ifine++)
                {
                  const VolIndex& fineVoF = fineVoFs[ifine];
                  //fine cell volume is normalized to one...
                  Real volFine  = ebisBoxFine.volFrac(fineVoF);
                  Real fineVal =  a_fine(fineVoF, ivar);
                  if(volFine > 0.)
                    {
                      dataVal += fineVal*volFine;
                    }
                }
              dataVal /= volCoar;
            }
          else
            {
              //if there is no real volume, just take the ave
              //of fine values
              for(int ifine = 0; ifine < fineVoFs.size(); ifine++)
                {
                  const VolIndex& fineVoF = fineVoFs[ifine];
                  Real fineVal =  a_fine(fineVoF, ivar);
                  dataVal += fineVal;
                }
              if(fineVoFs.size() > 0)
                {
                  dataVal /= Real(fineVoFs.size());
                }
            }
          a_coar(coarVoF, ivar) = dataVal;
        }
    }
}
/************************************/
void
EBCoarseAverage::averageFRZ(EBCellFAB&       a_coar,
                            const EBCellFAB& a_fine,
                            const DataIndex& a_datInd,
                            const Interval&  a_variables,
                            const Real&      a_dx) const
{
  assert(isDefined());
  //do all cells as if they were regular
  BaseFab<Real>& coarRegFAB =  a_coar.getSingleValuedFAB();
  const BaseFab<Real>& fineRegFAB = a_fine.getSingleValuedFAB();
  Box refbox(IntVect::TheZeroVector(), 
             (m_refRat-1)*IntVect::TheUnitVector());
  const Box& coarBox = m_coarsenedFineGrids.get(a_datInd);

#ifndef NDEBUG
  Box fineBox = refine(coarBox, m_refRat);
  assert(coarRegFAB.box().contains(coarBox));
  assert(fineRegFAB.box().contains(fineBox));
#endif
  Real dxFine = a_dx;
  Real dxCoar = Real(m_refRat)*a_dx;
  
  for(int ivar = a_variables.begin(); 
      ivar <= a_variables.end(); ivar++)
    {
      FORT_EBAVERARZ(CHF_FRA1(coarRegFAB,ivar),
                     CHF_CONST_FRA1(fineRegFAB,ivar),
                     CHF_BOX(coarBox),
                     CHF_CONST_INT(m_refRat),
                     CHF_BOX(refbox),
                     CHF_REAL(dxCoar),
                     CHF_REAL(dxFine));          
    }
  
  //overwrite irregular cells

  //recall that datInd is from the fine layout.
  const EBISBox& ebisBoxCoar = m_coarsenedFineEBISL[a_datInd];
  const EBISBox& ebisBoxFine = m_fineEBISL[a_datInd];
  IntVectSet coarIrregIVS = ebisBoxCoar.getIrregIVS(coarBox);

  //fine cell volume is normalized to one.
  
  for(VoFIterator vofitCoar(coarIrregIVS, ebisBoxCoar.getEBGraph()); 
      vofitCoar.ok(); ++vofitCoar)
    {
      const VolIndex& vofCoar = vofitCoar();
      Real cellVolCoar, kvolCoar;
      EBArith::getKVolRZ(kvolCoar, cellVolCoar, ebisBoxCoar, dxCoar, vofCoar);
      Real  volCoar = kvolCoar*cellVolCoar;

      Vector<VolIndex> vofsFine =
        m_coarsenedFineEBISL.refine(vofCoar, m_refRat, a_datInd);

      for(int ivar = a_variables.begin(); ivar <= a_variables.end(); ivar++)
        {
          Real dataVal = 0.0;
          if(volCoar > 0.)
            {
              for(int ifine = 0; ifine < vofsFine.size(); ifine++)
                {
                  const VolIndex& vofFine = vofsFine[ifine];
                  Real cellVolFine, kvolFine;
                  EBArith::getKVolRZ(kvolFine, cellVolFine, ebisBoxFine, dxFine, vofFine);
                  Real  volFine = kvolFine*cellVolFine;
                  Real fineVal =  a_fine(vofFine, ivar);
                  if(volFine > 0.)
                    {
                      dataVal += fineVal*volFine;
                    }
                }
              dataVal /= volCoar;
            }
          else
            {
              //if there is no real volume, just take the ave
              //of fine values
              for(int ifine = 0; ifine < vofsFine.size(); ifine++)
                {
                  const VolIndex& vofFine = vofsFine[ifine];
                  Real fineVal =  a_fine(vofFine, ivar);
                  dataVal += fineVal;
                }
              if(vofsFine.size() > 0)
                {
                  dataVal /= Real(vofsFine.size());
                }
            }
          a_coar(vofCoar, ivar) = dataVal;
        }
    }
}
/************************************/
