/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves  11-02-2001

#include "EBFineToCoarRedist.H"
#include "LayoutIterator.H"
#include "BaseIVFactory.H"
#include "VoFIterator.H"
#include "EBIndexSpace.H"
#include "EBCellFactory.H"
/***********************/
/***********************/
void 
EBFineToCoarRedist::
resetWeights(const LevelData<EBCellFAB>& a_modifierCoar,
             const int& a_ivar)
{
  //set the weights to mass weighting if the modifier
  //is the coarse density.   
  Interval srcInterv(a_ivar, a_ivar);
  Interval dstInterv(0,0);
  a_modifierCoar.copyTo(srcInterv, m_densityCoar, dstInterv);
  for(DataIterator dit = m_gridsCoar.dataIterator(); dit.ok(); ++dit)
    {
      const IntVectSet& fineSet = m_setsRefCoar[dit()];
      BaseIVFAB<VoFStencil>& massStenFAB = m_stenRefCoar[dit()];
      const BaseIVFAB<VoFStencil>& volStenFAB  = m_volumeStenc[dit()];
      const BaseIVFAB<VoFStencil>& stanStenFAB  = m_standardStenc[dit()];
      const EBISBox& ebisBoxFine = m_ebislRefCoar[dit()];
      const EBCellFAB& modFAB = m_densityCoar[dit()];

      for(VoFIterator vofit(fineSet, ebisBoxFine.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vofFine = vofit();

          const VoFStencil& oldSten = volStenFAB(vofFine, 0);
          const VoFStencil& stanSten = stanStenFAB(vofFine, 0);
          
          VoFStencil newSten;
          for(int isten = 0; isten < oldSten.size(); isten++)
            {
              const VolIndex& thatVoFFine = oldSten.vof(isten);
              VolIndex refCoarVoF = 
                m_ebislRefCoar.coarsen(thatVoFFine, m_refRat, dit());

              Real weight = modFAB(refCoarVoF, a_ivar);

              //it is weight*volfrac that is normalized
              newSten.add(thatVoFFine, weight);
            }
          //have to normalize using the whole stencil
          Real sum = 0.0;
          for(int isten = 0; isten < stanSten.size(); isten++)
            {
              const VolIndex& thatVoFFine = stanSten.vof(isten);
              VolIndex refCoarVoF = 
                m_ebislRefCoar.coarsen(thatVoFFine, m_refRat, dit());

              Real weight = modFAB(refCoarVoF, a_ivar);

              Real volfrac = ebisBoxFine.volFrac(thatVoFFine);
              //it is weight*volfrac that is normalized
              sum += weight*volfrac;
            }
          if(Abs(sum) > 0.0)
            {
              Real scaling = 1.0/sum;
              newSten *= scaling;
            }
          massStenFAB(vofFine, 0) = newSten;
          
        }
    }
}
/**********************/
/**********************/
EBFineToCoarRedist::EBFineToCoarRedist()
{
  m_isDefined = false;
}
/**********************/
/**********************/
EBFineToCoarRedist::~EBFineToCoarRedist()
{
}
/**********************/
/**********************/
void 
EBFineToCoarRedist::
define(const DisjointBoxLayout& a_dblFine,
       const DisjointBoxLayout& a_dblCoar,
       const EBISLayout& a_ebislFine,
       const EBISLayout& a_ebislCoar,
       const Box& a_domainCoar,
       const int& a_nref,
       const int& a_nvar,
       int a_redistRad)
{
  m_isDefined = true;
  m_nComp = a_nvar;
  m_refRat = a_nref;
  m_domainCoar = a_domainCoar;
  m_gridsFine = a_dblFine;
  m_gridsCoar = a_dblCoar;
  m_ebislFine = a_ebislFine;
  m_ebislCoar = a_ebislCoar;
  m_redistRad = a_redistRad;
  EBCellFactory ebcellfactcoar(m_ebislCoar);
  m_densityCoar.define(m_gridsCoar, 1, 2*m_redistRad*IntVect::Unit, ebcellfactcoar);
  //created the coarsened fine layout
  m_gridsRefCoar = DisjointBoxLayout();
  refine(m_gridsRefCoar, m_gridsCoar, m_refRat);

  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());
  int nghost = 3*m_redistRad;
  Box domainFine = refine(m_domainCoar, m_refRat);
  ebisPtr->fillEBISLayout(m_ebislRefCoar, m_gridsRefCoar, 
                          domainFine, nghost);
  m_ebislRefCoar.setMaxCoarseningRatio(m_refRat);

  //define the intvectsets over which the objects live
  m_setsFine.define(m_gridsFine);
  m_setsRefCoar.define(m_gridsCoar);

  //make sets
  //global set consists of redistrad on the fine side of the coarse-fine
  //interface 
  //the fine set is that within one fine box.
  //the refcoar set is that set within one refined coarse box.
  for(DataIterator dit = 
        m_gridsFine.dataIterator(); dit.ok(); ++dit)
    {
      const Box& fineBox = m_gridsFine.get(dit());
      IntVectSet& fineSet = m_setsFine[dit()];
      //add entire fine box
      fineSet = IntVectSet(fineBox);
      IntVectSet irregIVS = m_ebislFine[dit()].getIrregIVS(fineBox);
      fineSet &= irregIVS;
      //subtract fine box shrunk by the redistribution radius
      fineSet -= grow(fineBox, -m_redistRad);
      //subtract all other the fine boxes shifted in all 
      //directions
      for(LayoutIterator lit = 
            m_gridsFine.layoutIterator(); lit.ok(); ++lit)
        {
          const Box& otherBox = m_gridsFine.get(lit());
          if(otherBox != fineBox)
            {
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      Box shiftBox = otherBox;
                      shiftBox.shift(idir, sign(sit())*m_redistRad);
                      fineSet -= shiftBox;
                    }
                }
            }
        }
    }

  for(DataIterator dit = 
        m_gridsCoar.dataIterator(); dit.ok(); ++dit)
    {
      Box grownBox = grow(m_gridsRefCoar.get(dit()), m_redistRad);
      grownBox &= domainFine;
      //find the complement of what we really want
      IntVectSet ivsComplement(grownBox);
      for(LayoutIterator litFine =
            m_gridsFine.layoutIterator(); litFine.ok(); ++litFine)
        {
          const Box& fineBox = m_gridsFine.get(litFine());
          IntVectSet fineSet(fineBox);
          //add entire fine box
          fineSet = IntVectSet(fineBox);
          Box interiorBox = grow(fineBox, -m_redistRad);
          //subtract fine box shrunk by the redistribution radius
          fineSet -= interiorBox;
          //subtract all other the fine boxes shifted in all 
          //directions
          for(LayoutIterator lit = 
                m_gridsFine.layoutIterator(); lit.ok(); ++lit)
            {
              const Box& otherBox = m_gridsFine.get(lit());
              if(otherBox != fineBox)
                {
                  for(int idir = 0; idir < SpaceDim; idir++)
                    {
                      for(SideIterator sit; sit.ok(); ++sit)
                        {
                          Box shiftBox = otherBox;
                          shiftBox.shift(idir, sign(sit())*m_redistRad);
                          fineSet -= shiftBox;
                        }
                    }
                }
            }
          //subtract the fine set from the complement
          ivsComplement -= fineSet;
        }
      //now the set we want is the grownbox - complement
      IntVectSet& refCoarSet = m_setsRefCoar[dit()];
      refCoarSet = IntVectSet(grownBox);
      refCoarSet -= ivsComplement;
      IntVectSet irregIVS = m_ebislRefCoar[dit()].getIrregIVS(grownBox);
      refCoarSet &= irregIVS;
    }

  //make mass buffers
  BaseIVFactory<Real> factFine(m_ebislFine, m_setsFine);
  IntVect noghost = IntVect::Zero;
  m_regsFine.define(m_gridsFine, m_nComp, noghost, factFine);

  BaseIVFactory<Real> factRefCoar(m_ebislRefCoar, m_setsRefCoar);
  m_regsRefCoar.define(m_gridsRefCoar, m_nComp, m_redistRad*IntVect::Unit, factRefCoar);

  //define the stencils with volume weights
  //use resetWeights to do anything different
  m_stenRefCoar.define(m_gridsRefCoar);
  m_volumeStenc.define(m_gridsRefCoar);
  m_standardStenc.define(m_gridsRefCoar);
  RedistStencil rdstenRefCoar(m_gridsRefCoar, m_ebislRefCoar,
                              domainFine, m_redistRad);

  for(DataIterator dit = 
        m_gridsRefCoar.dataIterator(); dit.ok(); ++dit)
    {
      BaseIVFAB<VoFStencil>& stenFAB = m_stenRefCoar[dit()];
      BaseIVFAB<VoFStencil>& volStenFAB = m_volumeStenc[dit()];
      BaseIVFAB<VoFStencil>& stanStenFAB = m_standardStenc[dit()];
      const EBISBox& ebisBox = m_ebislRefCoar[dit()];
      const IntVectSet& ivs  = m_setsRefCoar[dit()];
      const BaseIVFAB<VoFStencil>& rdStenFAB = rdstenRefCoar[dit()];
      const Box& gridRefCoar = m_gridsRefCoar.get(dit());
      stenFAB.define(   ivs, ebisBox.getEBGraph(), 1);
      volStenFAB.define(ivs, ebisBox.getEBGraph(), 1);
      stanStenFAB.define(ivs, ebisBox.getEBGraph(), 1);
      
      for(VoFIterator vofit(ivs, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& srcVoF = vofit();
          VoFStencil newStencil;
          const VoFStencil& stanSten = rdStenFAB(srcVoF, 0);
          for(int istan = 0; istan < stanSten.size(); istan++)
            {
              const VolIndex& dstVoF = stanSten.vof(istan);
              const Real& weight = stanSten.weight(istan);
              if(gridRefCoar.contains(dstVoF.gridIndex()))
                {
                  newStencil.add(dstVoF, weight);
                }
            }
          stenFAB(srcVoF, 0)    = newStencil;
          volStenFAB(srcVoF, 0) = newStencil;
          stanStenFAB(srcVoF,0) =  stanSten;
        }
    }

  //initialize the buffers to zero
  setToZero();
}
/**********************/
/**********************/
void 
EBFineToCoarRedist::setToZero()
{
  for(DataIterator dit = m_gridsRefCoar.dataIterator(); 
      dit.ok(); ++dit)
    {
      m_regsRefCoar[dit()].setVal(0.0);
    }

  for(DataIterator dit = m_gridsFine.dataIterator(); 
      dit.ok(); ++dit)
    {
      m_regsFine[dit()].setVal(0.0);
    }
}
/**********************/
/**********************/
void 
EBFineToCoarRedist::increment(const BaseIVFAB<Real>& a_fineMass,
                              const DataIndex& a_fineDataIndex,
                              const Interval&  a_variables)
{
  BaseIVFAB<Real>& fineBuf =  m_regsFine[a_fineDataIndex];
  const EBISBox&   ebisBox = m_ebislFine[a_fineDataIndex];
  const IntVectSet& ivsLoc =  m_setsFine[a_fineDataIndex];

  assert(a_fineMass.getIVS().contains(ivsLoc));
  assert(fineBuf.getIVS().contains(ivsLoc));

  for(VoFIterator vofit(ivsLoc, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      for(int ivar = a_variables.begin(); 
          ivar <= a_variables.end();  ivar++)
        {
          fineBuf(vof, ivar) += a_fineMass(vof, ivar);
        }
    }
}
/**********************/
/**********************/
void 
EBFineToCoarRedist::redistribute(LevelData<EBCellFAB>& a_coarSolution,
                                 const Interval& a_variables)
{
  Real nrefD = 1.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    nrefD *= m_refRat;
  //copy the buffer to the coarse layout
  m_regsFine.copyTo(a_variables, m_regsRefCoar, a_variables);
  //redistribute the refined coarse registers to the coarse solution
  int ibox = 0;
  for(DataIterator dit = m_gridsCoar.dataIterator(); dit.ok(); ++dit)
    {
      const BaseIVFAB<Real>& regRefCoar = m_regsRefCoar[dit()];
      const IntVectSet& ivsRefCoar = m_setsRefCoar[dit()];
      const EBISBox& ebisBoxRefCoar = m_ebislRefCoar[dit()];
      const EBISBox& ebisBoxCoar = m_ebislCoar[dit()];
      const BaseIVFAB<VoFStencil>& stenFAB = m_stenRefCoar[dit()];

      EBCellFAB& solFAB = a_coarSolution[dit()];

      for(VoFIterator vofit(ivsRefCoar, ebisBoxRefCoar.getEBGraph()); 
          vofit.ok(); ++vofit)
        {
          const VolIndex& srcVoFFine = vofit();
          const VoFStencil& vofsten = stenFAB(srcVoFFine, 0);
          for(int isten = 0; isten < vofsten.size(); isten++)
            {
              const Real& weight = vofsten.weight(isten);
              const VolIndex& dstVoFFine = vofsten.vof(isten);
              VolIndex dstVoFCoar = 
                m_ebislRefCoar.coarsen(dstVoFFine,m_refRat, dit());
              //by construction...
              assert(m_gridsCoar.get(dit()).contains(dstVoFCoar.gridIndex()));
              Real dstVolFracFine = ebisBoxRefCoar.volFrac(dstVoFFine);
              Real dstVolFracCoar = ebisBoxCoar.volFrac(dstVoFCoar);
              Real denom = dstVolFracCoar*nrefD;
              for(int ivar = a_variables.begin(); 
                  ivar <= a_variables.end();  ivar++)
                {
                  Real dmFine = regRefCoar(srcVoFFine, ivar);
                  //ucoar+= massfine/volcoar, ie.
                  //ucoar+= (wcoar*dmCoar*volFracfine/volfraccoar)=massfine/volcoar
                  Real dUCoar = dmFine*weight*dstVolFracFine/denom;
                  solFAB(dstVoFCoar, ivar) += dUCoar;
                }
            }
        }
      ibox++;
    }
}
/**********************/
/**********************/
bool 
EBFineToCoarRedist::isDefined() const
{
  return m_isDefined;
}
/**********************/
/**********************/
