/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _EBLEVELFLUXREGISTER_H_
#define _EBLEVELFLUXREGISTER_H_
#include "REAL.H"
#include "Vector.H"
#include "EBCellFAB.H"
#include "EBFaceFAB.H"
#include "EBISLayout.H"
#include "EBISBox.H"
#include "IntVectSet.H"
#include "CFStencil.H"
#include "LoHiSide.H"
#include "LevelData.H"
#include "LayoutData.H"

class EBCoarToFineRedist;
class EBCoarToCoarRedist;

///EBFluxRegister-A class to encapsulate a levels worth of flux registers.
/**
   A EBFluxRegister  handles all the data choreography 
   necessary to create a levels worth of flux registers.
 */
class EBFluxRegister
{
public:

  ///
  /**
     Default constructor.  Leaves object undefined.
  */
  EBFluxRegister();

  ///
  /**
     Full constructor.  Calls the define function which creates 
     a levels worth of flux registers. 
  */
  EBFluxRegister(const DisjointBoxLayout& a_dblFine,
                 const DisjointBoxLayout& a_dblCoar,
                 const EBISLayout&        ebislFine,
                 const EBISLayout&        a_ebislCoar,
                 const Box&               a_domainCoar,
                 const int&               a_nref,
                 const int&               a_nvar);

  ///
  ~EBFluxRegister();

  ///
  /**
     Full define function.  Creates a levels worth of flux registers.
  */
  void 
  define(const DisjointBoxLayout& a_dblFine,
         const DisjointBoxLayout& a_dblCoar,
         const EBISLayout&        a_ebislFine,
         const EBISLayout&        a_ebislCoar,
         const Box&               a_domainCoar,
         const int&               a_nref,
         const int&               a_nvar);


  ///
  /** 
      Initialize values of registers  to zero.
  */
  void 
  setToZero(); 

  ///
  /**
     increments the register with data from coarseFlux, multiplied by scale.
     Increment coarse buffer with the area-weighted sum
     of the fluxes on the faces.
     buf += -scale*sum(areaFrac*flux)
     CoarseFlux must contain the coarse fluxes in the dir direction  
     for the grid   m_coarseLayout[coarseDataIndex].
  */
  void 
  incrementCoarseRegular(const EBFaceFAB&   a_coarseFlux,
                         const Real&        a_scale,
                         const DataIndex&   a_coarseDataIndex,
                         const Interval&    a_variables,
                         const int&         a_dir);

  
  ///
  void 
  incrementCoarseRegulRZ(const EBFaceFAB&  a_coarseFlux,
                         const Real&       a_scale,
                         const DataIndex&  a_coarseDataIndex,
                         const Interval&   a_variables,
                         const int&        a_dir,
                         const Real&       a_dx);

  ///
  /**
     increments the register with data from coarseFlux, multiplied by scale.
     Increment coarse buffer with the area-weighted sum
     of the fluxes on the faces.
     buf += -scale*sum(areaFrac*flux)
     CoarseFlux must contain the coarse fluxes in the dir direction  
     for the grid   m_coarseLayout[coarseDataIndex].
  */
  void 
  incrementCoarseIrregular(const BaseIFFAB<Real>& a_coarseFlux,
                           const Real&            a_scale,
                           const DataIndex&       a_coarseDataIndex,
                           const Interval&        a_variables,
                           const int&             a_dir);

  ///
  void 
  incrementCoarseIrregulRZ(const BaseIFFAB<Real>& a_coarseFlux,
                           const Real&            a_scale,
                           const DataIndex&       a_coarseDataIndex,
                           const Interval&        a_variables,
                           const int&             a_dir,
                           const Real&            a_dx);

  ///
  /**
     increments the register with data from fineFlux, multiplied by scale.
     fineFlux must contain the coarse fluxes in the dir direction  
     for the grid m_fineLayout[fineDataIndex]. 
     This operation is local.
  */
  void 
  incrementFineRegular(const EBFaceFAB&      a_fineFlux,
                       const Real&           a_scale,
                       const DataIndex&      a_fineDataIndex,
                       const Interval&       a_variables,
                       const int&            a_dir,
                       const Side::LoHiSide& a_sd);


  ///
  void 
  incrementFineRegulRZ(const EBFaceFAB&      a_fineFlux,
                       const Real&           a_scale,
                       const DataIndex&      a_fineDataIndex,
                       const Interval&       a_variables,
                       const int&            a_dir,
                       const Side::LoHiSide& a_sd,
                       const Real&           a_dx);


  ///
  /**
     increments the register with data from fineFlux, multiplied by scale.
     fineFlux must contain the coarse fluxes in the dir direction  
     for the grid m_fineLayout[fineDataIndex]. 
     This operation is local.
  */
  void 
  incrementFineIrregular(const BaseIFFAB<Real>& a_fineFlux,
                         const Real&            a_scale,
                         const DataIndex&       a_fineDataIndex,
                         const Interval&        a_variables,
                         const int&             a_dir,
                         const Side::LoHiSide&  a_sd);


  ///
  void 
  incrementFineIrregulRZ(const BaseIFFAB<Real>& a_fineFlux,
                         const Real&            a_scale,
                         const DataIndex&       a_fineDataIndex,
                         const Interval&        a_variables,
                         const int&             a_dir,
                         const Side::LoHiSide&  a_sd,
                         const Real&            a_dx);

  ///
  /**
     increments uCoarse with the reflux divergence of the 
     contents of the flux register.  
     This operation is global and blocking.
  */
  void reflux(LevelData<EBCellFAB>& a_uCoarse,
              const Interval&       a_variables,
              const Real&           a_scale);


  ///
  /**
     Increments redistribution register with left-over mass
     from reflux divergence.
  */
  void incrementRedistRegister(EBCoarToFineRedist& a_register,
                               const Interval&     a_variables,
                               const Real&         a_scale);

  ///
  /**
     Increments redistribution register with left-over mass
     from reflux divergence.
  */
  void incrementRedistRegister(EBCoarToCoarRedist& a_register,
                               const Interval&     a_variables,
                               const Real&         a_scale);

  ///
  void incrementRedistRegistRZ(EBCoarToCoarRedist& a_register,
                               const Interval&     a_variables,
                               const Real&         a_dx);

  ///
  /**
     Increments mass array  with left-over mass
     from reflux divergence.  
  */
  void incrementDensityArray(LevelData<EBCellFAB>& a_coarMass,
                             const Interval&       a_variables,
                             const Real&           a_scale);
  
  ///
  bool isDefined() const;

  /// 
  void dumpCoar(const int& a_idir,
                const Side::LoHiSide& a_sd);
  ///
  void dumpFine(const int& a_idir,
                const Side::LoHiSide& a_sd);
  
  ///
  void dumpBIVF(const BaseIVFAB<Real>& reg,
                const EBISBox& a_ebisBox);
protected:

  static int index(int a_dir, Side::LoHiSide a_side);

  //internal use only
  void setDefaultValues();

  // both fabCoarse and fabFine have fabs that are
  // defined in the coarse index space.  fabCoarse 
  // can ONLY be indexed into by the DataIndex stored in
  // coarToCoarMap.  fabFine is indexed by the same indexing
  // used in the fine DisjointBoxLayout.

  //input fine grid
  DisjointBoxLayout    m_gridsFine;
  //input coarse grid
  DisjointBoxLayout    m_gridsCoar;

  //grids for fine buffers
  DisjointBoxLayout m_bufGridsFine[SpaceDim*2];
  //grids for coarse buffers
  DisjointBoxLayout m_bufGridsCoar[SpaceDim*2];

  //fine buffers
  LevelData<BaseIVFAB<Real> > m_regsCoar[SpaceDim*2];
  //coarse buffers
  LevelData<BaseIVFAB<Real> > m_regsFine[SpaceDim*2];

  //scratch space for refluxing
  LevelData<BaseIVFAB<Real> > m_scratchc[SpaceDim*2];

  //sets for fine buffers
  LayoutData<IntVectSet> m_cfivsFine[SpaceDim*2];
  //sets for coarse buffers
  LayoutData<IntVectSet> m_cfivsCoar[SpaceDim*2];

  //ebisl of input fine grid
  EBISLayout m_ebislFine;
  //ebisl of input coar grid
  EBISLayout m_ebislCoar;

  //ebisl of coarse buffers 
  EBISLayout m_ebislBufCoar[2*SpaceDim];
  //ebisl of fine buffers 
  EBISLayout m_ebislBufFine[2*SpaceDim];

  //map from input coarse grid to coarse buffers
  LayoutData<Vector<DataIndex>  > m_coarIndexMap[SpaceDim*2];

  bool m_isDefined;
  int m_nComp;
  int m_refRat;
  Box m_domainCoar;

private:
  //for all the usual reasons,
  //there is no copy constructor for this class
  //and there is no operator= for this class
  //(there is no rule six).
  void operator= (const EBFluxRegister&)
  {
    MayDay::Error("invalid operator");
  }
  EBFluxRegister(const EBFluxRegister&)
  {
    MayDay::Error("invalid operator");
  }
};
#endif
