/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves mon oct 15 2001
#include "EBLevelRedist.H"
#include "BaseIVFactory.H"
#include "VoFIterator.H"
#include "EBIndexSpace.H"
/***********************/
/***********************/
EBLevelRedist::EBLevelRedist()
{
  m_isDefined = false;
}
/***********************/
/***********************/
EBLevelRedist::~EBLevelRedist()
{
}
/***********************/
/***********************/
EBLevelRedist::
EBLevelRedist(const DisjointBoxLayout& a_dbl,
              const EBISLayout& a_ebisl,
              const Box& a_domain,
              const int& a_ncomp,
              int   a_redistRad)
{
  define(a_dbl, a_ebisl, a_domain, a_ncomp, a_redistRad);
}
/***********************/
/***********************/
void 
EBLevelRedist::
resetWeights(const LevelData<EBCellFAB>& a_modifier,
             const int& a_ivar)
{
  m_stencil.resetWeights(a_modifier, a_ivar);
}
/***********************/
/***********************/
void 
EBLevelRedist::
define(const DisjointBoxLayout& a_dbl,
       const EBISLayout& a_ebisl,
       const Box& a_domain,
       const int& a_ncomp,
       int a_redistRad)
{
  m_isDefined = true;

  m_grids     = a_dbl;
  m_ebisl     = a_ebisl;
  m_domain    = a_domain;
  m_ncomp     = a_ncomp;
  m_redistRad = a_redistRad;

  //this sets the stencil to volume-weighted.
  //use resetWeights to set to mass weighted or whatever
  m_stencil.define(m_grids, m_ebisl, m_domain, m_redistRad); 

  //define the sets for iterating over 
  //to be the irregular cells over the
  //grid grown by the redistribution radius
  m_sets.define(m_grids);
  int redistRad = m_stencil.getRedistRadius();
  for(DataIterator dit = m_grids.dataIterator(); 
      dit.ok(); ++dit)
    {
      Box thisBox = m_grids.get(dit());
      thisBox.grow(redistRad);
      thisBox &= m_domain;
      m_sets[dit()] = m_ebisl[dit()].getIrregIVS(thisBox);
    }
  BaseIVFactory<Real> factory(m_ebisl, m_sets);
  IntVect ivghost = redistRad*IntVect::Unit;
  m_buffer.define(m_grids, m_ncomp, ivghost, factory);
  setToZero();
}
/***********************/
/***********************/
void
EBLevelRedist::setToZero() 
{
  assert(isDefined());
  for(DataIterator dit = m_grids.dataIterator(); 
      dit.ok(); ++dit)
    {
      m_buffer[dit()].setVal(0.0);
    }
}
/***********************/
/***********************/
void
EBLevelRedist::
increment(const BaseIVFAB<Real>& a_massDiff,
          const DataIndex& a_datInd,
          const Interval& a_variables)
{
  assert(isDefined());
  BaseIVFAB<Real>& bufFAB = m_buffer[a_datInd];
  const IntVectSet& fabIVS = a_massDiff.getIVS();
  const IntVectSet& bufIVS = m_sets[a_datInd];
  //the buffers is bigger because it has to be
  //to make life simpler
  assert(bufIVS.contains(fabIVS));
  IntVectSet ivs = m_ebisl[a_datInd].getIrregIVS(m_grids.get(a_datInd));;
  assert(fabIVS.contains(ivs));
  assert(bufIVS.contains(ivs));
  for(VoFIterator vofit(ivs, m_ebisl[a_datInd].getEBGraph());
      vofit.ok(); ++vofit)
    {
      for(int ivar = a_variables.begin(); 
          ivar <= a_variables.end(); ivar++)
        {
          bufFAB(vofit(), ivar) += a_massDiff(vofit(), ivar);
        }
    }
}
/***********************/
/***********************/
void
EBLevelRedist::
redistribute(LevelData<EBCellFAB>& a_solution,
             const Interval& a_variables)
{
  assert(isDefined());

  //pout() << "redistribute 0" << endl;
  //exchange ghost cell information of the buffer.
  //this way the redistribution from ghost cells will 
  //account for fine-fine interfaces.
  m_buffer.exchange(a_variables);
  //pout() << "redistribute 1" << endl;
  //loop over grids.
  for(DataIterator dit = m_grids.dataIterator(); 
      dit.ok(); ++dit)
    {
      const BaseIVFAB<Real>& bufFAB = m_buffer[dit()];
      const BaseIVFAB<VoFStencil>& stenFAB = m_stencil[dit()];
      const Box& grid = m_grids.get(dit());
      EBCellFAB& solFAB = a_solution[dit()];

      //pout() << "redistribute 2" << endl;
      //loop over the irregular vofs in the grown box.
      for(VoFIterator vofit(m_sets[dit()], m_ebisl[dit()].getEBGraph());
          vofit.ok(); ++vofit)
        {
          const VolIndex& srcVoF = vofit();
          const VoFStencil& vofsten = stenFAB(srcVoF, 0);
          for(int isten = 0; isten < vofsten.size(); isten++)
            {
              const Real& weight = vofsten.weight(isten);
              const VolIndex& dstVoF = vofsten.vof(isten);
              //only add the mass to the solution if it is in the
              //solutions valid region --- because the buffer is over
              //the grown box, this prevents redistribution off
              //the end of the world.
              if(grid.contains(dstVoF.gridIndex()))
                {
                  for(int ivar = a_variables.begin(); 
                      ivar <= a_variables.end(); ivar++)
                    {
                      const Real& mass = bufFAB(srcVoF, ivar);
                      const Real& solu = solFAB(dstVoF, ivar);
                      solFAB(dstVoF, ivar) = mass*weight + solu;
                    }
                } //end check if adding to valid soltuion
            } //end loop over vof stencil
        } //end loop over irregular vofs in grown box
    } //end loop over grids.
}
/***********************/
/***********************/
bool 
EBLevelRedist::isDefined() const
{
  return m_isDefined;
}
/***********************/
/***********************/

