/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves thurs, aug 23, 2001

#ifndef _EB_PWL_FINE_INTERP_H_
#define _EB_PWL_FINE_INTERP_H_

#include "REAL.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "DisjointBoxLayout.H"
#include "EBISLayout.H"
#include "EBCellFAB.H"
#include "Interval.H"
#include "Stencils.H"
#include "BaseIVFAB.H"
#include "EBIndexSpace.H"

///
/**
 */
class EBPWLFineInterp
{
public:
  ///
  /**
     Default constructor.  User must subsequently call define().
  */
  EBPWLFineInterp();

  ///
  ~EBPWLFineInterp();

  ///
  /**
     Defining constructor.  Constructs a valid object.
     Equivalent to default construction followed by define().

     {\bf Arguments:}\\
     dblFine, dblCoar: The fine and coarse layouts
     of the data.\\
     ebislFine, ebislCoar: The fine and coarse layouts
     of the geometric description.\\
     nref: The refinement ratio between the two levels. \\
     nvar: The number of variables contained in the data
     at each VoF.
  */
  EBPWLFineInterp(const DisjointBoxLayout& a_dblFine,
                  const DisjointBoxLayout& a_dblCoar,
                  const EBISLayout& a_ebislFine,
                  const EBISLayout& a_ebislCoar,
                  const Box& a_domainCoar,
                  const int& a_nref,
                  const int& a_nvar,
                  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

  ///
  /**
     Defines this object.  Existing information is overriden.

     {\bf Arguments:}\\
     dblFine, dblCoar: The fine and coarse layouts
     of the data.\\
     ebislFine, ebislCoar: The fine and coarse layouts
     of the geometric description.\\
     nref: The refinement ratio between the two levels. \\
     nvar: The number of variables contained in the data
     at each VoF.
  */
  void define(const DisjointBoxLayout& a_dblFine,
              const DisjointBoxLayout& a_dblCoar,
              const EBISLayout& a_ebislFine,
              const EBISLayout& a_ebislCoar,
              const Box& a_domainCoar,
              const int& a_nref,
              const int& a_nvar,
              const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

  ///
  /**
     Returns true if this object was created with the defining
     constructor or if define() has been called.
  */
  bool isDefined() const;

  ///
  /**
     Interpolate the fine data from the coarse data 
     over the intersection
     of the fine layout with the refined coarse layout. \\
     NOTE: This function sends these arrays
     to Fortran so you must set the covered 
     cell values to something.  Unitialized memory
     might cause IEEE problems with some compilers.
     {\bf Arguments:}\\
     coarData: The data over the coarse layout.\\
     fineData: The data over the fine layout.  \\
     Fine and coarse data must
     have the same number of variables.\\
     variables:  The variables to interpolate.  Those not
     in this range will be left alone.   This range of variables
     must be in both the coarse and fine data.
  */
  void
  interpolate(LevelData<EBCellFAB>& a_fineData,
              const LevelData<EBCellFAB>& a_coarseData,
              const Interval& a_variables);          

protected:
  void
  interpolateFAB(EBCellFAB& a_fine,
                 const EBCellFAB& a_coar,
                 const DataIndex& a_datInd,
                 const Interval& a_variables) const;

  void makeDerivStencils();

  void setDefaultValues();

  bool m_isDefined;

  DisjointBoxLayout m_coarsenedFineGrids;
  Box m_coarDomain;
  EBISLayout m_coarsenedFineEBISL;

  int m_refRat;
  int m_nComp;
  LevelData<EBCellFAB> m_coarsenedFineData;

  //irregular cells
  LayoutData<IntVectSet> m_irregRegions;

  //diffs high and low---need to make the choice
  //about which to use based on the data.
  LayoutData<BaseIVFAB<VoFStencil> >  m_hiStencils[SpaceDim];
  LayoutData<BaseIVFAB<VoFStencil> >  m_loStencils[SpaceDim];

private:  
  //disallowed for all the usual reasons
  EBPWLFineInterp(const EBPWLFineInterp& ebcin)
  {
    MayDay::Error("ebpwl 2 invalid operator");
  }
  void operator=(const EBPWLFineInterp& fabin)
  {
    MayDay::Error("ebpwl 3 invalid operator");
  }
};

#endif
