/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves fri, sept 14, 2001

#ifndef _EB_QUAD_CF_INTERP_H_
#define _EB_QUAD_CF_INTERP_H_

#include "REAL.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "DisjointBoxLayout.H"
#include "EBISLayout.H"
#include "EBCellFAB.H"
#include "Interval.H"
#include "Stencils.H"
#include "QuadCFInterp.H"
#include "BaseIVFAB.H"

///
/**
   Fills one layer of fine level ghost cells 
   by piecewise linear interpolation
   from the coarse level.
 */
class EBQuadCFInterp
{
public:
  ///
  /**
     Default constructor.  User must subsequently call define().
  */
  EBQuadCFInterp();

  ///
  ~EBQuadCFInterp();

  ///
  /**
     Defining constructor.  Constructs a valid object.
     Equivalent to default construction followed by define().

     {\bf Arguments:}\\
     dblFine, dblCoar: The fine and coarse layouts
     of the data.  \\
     ebislFine, ebislCoar: The fine and coarse layouts
     of the geometric description.\\
     nref: The refinement ratio between the two levels. \\
     nvar: The number of variables contained in the data
     at each VoF.
  */
  EBQuadCFInterp(const DisjointBoxLayout& a_dblFine,
                 const DisjointBoxLayout& a_dblCoar,
                 const EBISLayout& a_ebislFine,
                 const EBISLayout& a_ebislCoar,
                 const Box& a_domainCoar,
                 const int& a_nref,
                 const int& a_nvar);

  ///
  /**
     Defines this object.  Existing information is overriden.

     {\bf Arguments:}\\
     dblFine, dblCoar: The fine and coarse layouts
     of the data.\\
     ebislFine, ebislCoar: The fine and coarse layouts
     of the geometric description.\\
     nref: The refinement ratio between the two levels. \\
     nvar: The number of variables contained in the data
     at each VoF.
  */
  void define(const DisjointBoxLayout& a_dblFine,
              const DisjointBoxLayout& a_dblCoar,
              const EBISLayout& a_ebislFine,
              const EBISLayout& a_ebislCoar,
              const Box& a_domainCoar,
              const int& a_nref,
              const int& a_nvar);

  ///
  /**
     Interpolate the fine data from the coarse data 
     over the intersection
     of the fine layout with the refined coarse layout. \\
     {\bf Arguments:}\\
     coarDataOld: The data over the coarse layout.\\
     fineData: The data over the fine layout.  \\
     Fine and coarse data must
     have the same number of variables.\\
     variables:  The variables to interpolate.  Those not
     in this range will be left alone.   This range of variables
     must be in both the coarse and fine data. 
  */
  void
  interpolate(LevelData<EBCellFAB>&       a_fineData,
              const LevelData<EBCellFAB>& a_coarData,
              const Interval&             a_variables);          

  bool isDefined() const;
protected:
  
  EBISLayout        m_ebislCoar;
  EBISLayout        m_ebislFine;
  DisjointBoxLayout m_gridsCoar;
  DisjointBoxLayout m_gridsFine;
  QuadCFInterp      m_quadCFInterp;
  Box               m_domainFine;
  Box               m_domainCoar;
  bool              m_isDefined;
  int               m_nComp;
  int               m_refRat;

private:  

  //disallowed for all the usual reasons
  EBQuadCFInterp(const EBQuadCFInterp& ebcin)
  {
    MayDay::Error("ebqcfi 2 invalid operator");
  }
  void operator=(const EBQuadCFInterp& fabin)
  {
    MayDay::Error("ebqcfi 3 invalid operator");
  }
};

#endif
