/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves weds oct 3 2001

#ifndef _Redist_stencil_H_
#define _Redist_stencil_H_

#include "REAL.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "DisjointBoxLayout.H"
#include "EBISLayout.H"
#include "EBCellFAB.H"
#include "Interval.H"
#include "Stencils.H"
#include "BaseIVFAB.H"

/// stencil for mass redistribution
/**
   The  RedistStencil class holds the stencil for mass redistribution
   at every irregular VoF in a layout.  The default weights
   that the stencil holds are volume weights.  The class
   does allow the flexibility to redefine these weights.
  */
class RedistStencil
{
public:
  ///
  /**
     Default constructor.  User must subsequently call define().
  */
  RedistStencil();

  ///
  ~RedistStencil();

  ///
  /**
     Defining constructor.  Constructs a valid object.
     Equivalent to default construction followed by define().
     ebisl needs ghost cells out to 3*redistradius
  */
  RedistStencil(const DisjointBoxLayout& dbl,
                      const EBISLayout& ebisl,
                      const Box& domain,
                      const int& redistRadius);

  ///
  /**
     Defines this object.  Existing information is overriden.
     ebisl needs ghost cells out to 3*redistradius
  */
  void define(const DisjointBoxLayout& dbl,
              const EBISLayout& ebisl,
              const Box& domain,
              const int& redistRadius);

  ///
  /**
     Returns true if this object was created with the defining
     constructor or if define() has been called.
  */
  bool isDefined() const;


  ///
  /**
     Modify the weights in the stencil by multiplying by
     the inputs in a normalized way.
     If you want mass
     weighting, send in the density.
  */
  void resetWeights(const LevelData<EBCellFAB>& a_modifier,
                    const int& a_ivar);

  ///
  /**
     Returns the redistribution stencil at every irregular
     point in input  Box associated with  this DataIndex.
  */
  const BaseIVFAB<VoFStencil >& 
  operator[](const DataIndex& datInd) const;

  ///
  /**
     This copy operation detaches copier from copyee.
   */
  void deepCopy(const RedistStencil& a_stenin);

  ///
  /**
   */
  int getRedistRadius() const;

protected:

  LayoutData<BaseIVFAB<VoFStencil > > m_stencil;
  LayoutData<BaseIVFAB<VoFStencil > > m_volsten;
  DisjointBoxLayout  m_grids;
  Box m_domain;
  EBISLayout m_ebisl;
  bool m_isDefined;
  bool m_hasDefaultWeights;
  int m_redistRadius;

private:
  //internal use
  void computePointStencil(VoFStencil& a_stencil,
                           const VolIndex&   a_srcVoF,
                           const DataIndex&  a_datInd);
  
  //disallowed for all the usual reasons
  RedistStencil(const RedistStencil& ebcin)
  {
    MayDay::Error("rdsi 2 invalid operator");
  }
  void operator=(const RedistStencil& fabin)
  {
    MayDay::Error("rdsi 3 invalid operator");
  }
};



#endif
 
