/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _BaseEBCellFAB_H_
#define _BaseEBCellFAB_H_

#include <cmath>
#include <cstdlib>
#include "Vector.H"
#include "EBISBox.H"
#include "BaseIVFAB.H"
#include "IntVectSet.H"
#include "BaseFab.H"


///  Array defined at the VolIndexs of an Box in an EBIS.
/**
   An BaseEBCellFAB is a templated holder for cell-centered 
   data over the intersection of a box and an EbisBox.
   At every uncovered VolIndex in this intersection, 
   the BaseEBCellFAB contains a specified number of data values.
   At singly valued cells,
   the data is stored internally in a BaseFab.  At multiply-valued
   cells, the data is stored internally in an  BaseIVFAB.
   BaseEBCellFAB provides indexing by VolIndex.
   This class does not provide a copy constructor or assignment operator
   for all the usual  reasons.
*/
template <class T>
class BaseEBCellFAB
{
public:
  ///
  BaseEBCellFAB();

  ///
  BaseEBCellFAB(const EBISBox& a_ebisBox, const Box& a_region, int a_nVar);

  ///
  void  define(const EBISBox& a_ebisBox, const Box& a_region, int a_nVar);

  ///
  virtual ~BaseEBCellFAB();


  ///
  void clear();

  ///
  void  setVal(const T& value);

  ///
  void  setVal(int ivar,const T& value);

  ///
  bool isDefined() const;


  ///
  int nComp() const ;

  ///
  const Box& getRegion() const ;

  ///
  const Box& box() const ;

  ///
  const IntVectSet& getMultiCells() const ;

  ///
  const BaseIVFAB<T>& getMultiValuedFAB() const;

  ///
  BaseIVFAB<T>& getMultiValuedFAB() ;

  ///
  const BaseFab<T>& getSingleValuedFAB() const;

  ///
  BaseFab<T>& getSingleValuedFAB();

  ///
  const T& operator() (const VolIndex& a_ndin,int  a_nVarLoc) const;

  ///
  T& operator() (const VolIndex& a_ndin,int  a_nVarLoc);


  ///
  void copy(const Box& RegionFrom, 
            const Interval& destInt, 
            const Box& RegionTo,
            const BaseEBCellFAB<T>& source, 
            const Interval& srcInt);

  ///
  BaseEBCellFAB<T>& copy(const BaseEBCellFAB<T>& source);

  ///
  /**
     This stuff required by LevelData in parallel:
  */
  static int preAllocatable() 
  {
    return 1; // symmetric messaging 
  } 

  ///
  int size(const Box& R, const Interval& comps) const ;

  ///
  int regSize(const Box& R, const Interval& comps) const ;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const ;

  ///
  void regLinearOut(void* buf, const Box& R, const Interval& comps) const ;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  void regLinearIn(void* buf, const Box& R, const Interval& comps);


  ///invalid but necessary for leveldata to compile
  BaseEBCellFAB(const Box& a_region, int a_nVar)
  {
    MayDay::Error("invalid constructor called for baseebcellfab");
  }

  ///
  const EBISBox& getEBISBox() const;

  ///
  void
  setCoveredCellVal(const T& a_val,
                    const int& a_comp);

protected:

  ///data at multi-valued cells
  BaseIVFAB<T> m_irrFAB;

  ///data at single-valued cells
  BaseFab<T> m_regFAB;

  EBISBox m_ebisBox;
  //number of data values at each VolIndex
  int m_nComp;

  ///region over which the BaseEBCellFAB exists
  Box m_region;

  ///whether each given cell is multi-valued
  IntVectSet m_multiCells;

  ///has full define function been called?
  bool m_isDefined;

protected:
  void
  setDefaultValues();

private:
  //disallowed for perfomance reasons
  void operator= (const BaseEBCellFAB<T>& ebcin)
  {
    MayDay::Error("operator= for baseebcellfab undefined");
  }
  BaseEBCellFAB (const BaseEBCellFAB<T>& ebcin)
  {
    MayDay::Error("copy constructor for baseebcellfab undefined");
  }
};

#include "BaseEBCellFABI.H"

#endif
