/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _BaseEBFaceFAB_H_
#define _BaseEBFaceFAB_H_

#include <cmath>
#include <cstdlib>
#include <cassert>
#include "Vector.H"
#include "BaseIFFAB.H"
#include "IntVectSet.H"
#include "BaseFab.H"
#include "EBISBox.H"

///  Array defined at the Faces of an Box in an EBISBox.
/**
   An BaseEBFaceFAB is a templated holder for face-centered 
   data over a region which consists of the intersection
   of a face-centered box and an EBISBox.  
   At every uncovered VoF in this intersection, 
   the BaseEBFaceFAB contains a specified number of data values.
   At singly valued faces,
   the data is stored internally in a BaseFab.  At multiply-valued
   faces, the data is stored internally in an  INFab.
   A multiply-valued face is defined to be a face which
   abuts at least one multiply-valued Face.
   BaseEBFaceFAB provides indexing by Face.
   This class does not provide a copy constructor or assignment operator
   for all the usual reasons.
 */
template <class T>
class BaseEBFaceFAB
{
public:
  ///
  BaseEBFaceFAB();


  ///
  /**
     Box going into this needs to be cell-centered.
     Data will exist over the surrounding nodes of the box.
  */
  BaseEBFaceFAB(const EBISBox& a_ebisBox,
                const Box& a_region, 
                int a_iDir, int a_nVar);


  ///
  /**
     Box going into this needs to be cell-centered.
     Data will exist over the surrounding nodes of the box.
  */
  void 
  define(const EBISBox& a_ebisBox, 
         const Box& a_region, 
         int a_iDir, int a_nVar);


  ///
  virtual ~BaseEBFaceFAB();


  ///
  void  clear();

  ///
  void setVal(const T& value);


  ///
  void setVal(int ivar,const T& value);


  ///
  bool isDefined() const;


  ///
  int nComp() const ;

  ///
  const BaseIFFAB<T>&   getMultiValuedFAB() const;

  ///
  BaseIFFAB<T>&  getMultiValuedFAB() ;

  ///
  const BaseFab<T>& getSingleValuedFAB() const;

  ///
  const IntVectSet& getMultiCells() const;

  ///
  BaseFab<T>&   getSingleValuedFAB() ;

  ///
  int  direction() const;

  ///
  /**
     Returns the FACE-CENTERED region of the fab
  */
  const Box& getRegion() const ;

  ///
  /**
     Returns the CELL-CENTERED region of the fab
  */
  const Box& getCellRegion() const ;

  ///
  T& operator() (const FaceIndex& a_facein, int  a_nVarLoc);

  ///
  const T&
  operator() (const FaceIndex& a_facein, int  a_nVarLoc) const;

  ///
  static bool preAllocatable(){ return true;}

  ///
  /**
     Box going into this should be CELL CENTERED.
     copy done over surrounding nodes.
  */
  void copy(const Box& RegionFrom, 
            const Interval& destInt, 
            const Box& RegionTo,
            const BaseEBFaceFAB<T>& source, 
            const Interval& srcInt);

  BaseEBFaceFAB<T>& copy(const BaseEBFaceFAB<T>& source);

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  int size(const Box& R, const Interval& comps) const
  {
    MayDay::Error("size called for BaseEBFaceFAB");
    return(-1);
  }

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  int regSize(const Box& R, const Interval& comps) const
  {
    MayDay::Error("regSize called for BaseEBFaceFAB - unimplemented");
    return(-1);
  }

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  void linearOut(void* buf, const Box& R, const Interval& comps) const
  {
    MayDay::Error("linearOut called for BaseEBFaceFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  void regLinearOut(void* buf, const Box& R, const Interval& comps) const
  {
    MayDay::Error("regLinearOut called for BaseEBFaceFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  void linearIn(void* buf, const Box& R, const Interval& comps)
  {
    MayDay::Error("linearIn called for BaseEBFaceFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<BaseEBFaceFAB> to compile
  void regLinearIn(void* buf, const Box& R, const Interval& comps)
  {
    MayDay::Error("regLinearIn called for BaseEBFaceFAB - unimplemented");
  }

  /// Invalid but necessary for LevelData<T> to compile
  BaseEBFaceFAB(const Box& a_region, int a_nVar)
  {
    MayDay::Error("invalid constructor called for BaseEBFaceFAB");
  }

  ///
  const EBISBox& getEBISBox() const;
    
protected:
  ///data at faces which abut multi-valued cells
  BaseIFFAB<T> m_irrFAB;

  ///data at faces between two single-valued cells
  BaseFab<T> m_regFAB;

  ///number of data values at each face in BaseEBFaceFAB
  int m_nComp;

  //direction of the faces in the BaseEBFaceFAB
  int m_iDir;

  ///the face-centered region over which the BaseEBFaceFAB lives
  Box m_regionFace;

  ///the cell-centered region over which the BaseEBFaceFAB lives
  Box m_region;

  ///has the full define function been called for the BaseEBFaceFAB?
  bool m_isDefined;

  EBISBox m_ebisBox;
private:
  void
  setDefaultValues();
  //disallowed for performance reasons
  void operator= (const BaseEBFaceFAB<T>& ebcin)
  {
    MayDay::Error("BaseEBFaceFAB<T>::operator=  not defined");
  }
  BaseEBFaceFAB (const BaseEBFaceFAB<T>& ebcin)
  { 
    MayDay::Error("BaseEBFaceFAB<T> copy constructor  not defined");
  }
};

#include "BaseEBFaceFABI.H"

#endif

