/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTG, ANAG, LBNL

#ifndef _BaseIFFAB_H_
#define _BaseIFFAB_H_

#include <cmath>
#include <cstdlib>
#include <cassert>
#include "Vector.H"
#include "IntVectSet.H"
#include "FaceIndex.H"
#include "BaseFab.H"
#include "EBGraph.H"

///  
/**
   BaseIFFAB is a templated data holder 
   defined over the Faces of an irregular domain.
*/
template <class T>
class BaseIFFAB
{
public:
  ///
  /**
     Default constructor.  Constructs an uninitialized IFFAB.  You must
     subsequently call {\bf define} before using this IFFAB.
  */
  BaseIFFAB();


  ///
  /**
     Defining constructor.  Specifies the irregular domain
     and the number of data components per face.  The irregular domain
     must lie completely within the EBGraph.  The
     contents are uninitialized.  Calls full define function.
  */
  BaseIFFAB(const IntVectSet& a_region, 
            const EBGraph& a_ebgraph,
            const int& a_direction,
            const int& a_nvarin);

  ///
  ~BaseIFFAB();

  ///
  /**
     Full define function.  Specifies the irregular domain and the 
     number of data components per face.  The irregular domain
     must lie completely within the EBGraph.   The contents
     are uninitialized.  If it has previously been defined, the old definition 
     data is overwritten and lost. 
  */
  void define(const IntVectSet& a_region, 
              const EBGraph& a_ebgraph,
              const int& a_direction,
              const int& a_nvarin);


  ///
  /**
     Set a value at every data location.
  */
  void setVal(const T& value);

  ///
  /**
     Set a value at every data location.
  */
  void setVal(int ivar, const T& value);


  ///
  /**
     define over surrounding nodes of invectset (ivs is cell-centered).
   */
  void copy(const Box& a_intBox, 
            const Interval& a_destInterval,
            const Box& a_toBox,
            const BaseIFFAB<T>& a_src, 
            const Interval& a_srcInterval);
            


  ///
  /**
     Remove all data from this BaseIFFAB.  
     You must call {\bf define} again in order 
     to use it.
  */

  void clear();


  ///invalid but necessary for leveldata to compile
  BaseIFFAB(const Box& a_region, int a_nVar)
  {
    MayDay::Error("invalid constructor called for baseebcellfab");
  }

  ///
  /**
   
     Tells whether this BaseIFFAB has been defined, either with a constructor 
     or with {\bf define}.  It must be initialized in order to access its data.
  */        
  bool 
  isDefined() const;


  ///
  /**
     Return the number of face in this BaseIFFAB.
  */
  int numFaces() const; 

  ///
  /**
     Return the number of data components of this BaseIFFAB.
  */
  int nComp() const;

  ///

  ///
  /**
     Return the direction of the faces of this BaseIFFAB.
  */
  int direction() const;

  ///

  /**
     Return  the  irregular domain of the BaseIFFAB.
  */
  const IntVectSet& getIVS() const;


  ///
  /**
     Indexing operator.  Return a reference to the contents of this IFFAB,
     at the specified face and data component.  The first component is
     zero, the last is {\em nvar-1}.  The returned object is a modifiable
     lvalue.
  */
  T& operator() (const FaceIndex& a_face,const int& varlocin);
  const T& operator()(const FaceIndex& a_face,const int& varlocin) const;

  ///
  /**
     Return a const pointer to the internal data storage of this IFFAB. \\
  */
  const T* dataPtr(const int& a_comp) const;

  ///
  /**
     Return a pointer to the internal data storage of this IFFAB. 
  */
  T* dataPtr(const int& a_comp) ;

  ///
  /**
     Return a fake "lower left corner of the domain" for use with Fortran. 
  */
  const int* loVect() const;

  ///
  /**
     Return a fake "upper right corner of the domain" for use with Fortran. 
  */
  const int* hiVect() const;

  ///
  const EBGraph& getEBGraph() const;

  static int preAllocatable()
  {
    return 1; // symmetric allocatable.  
  }

  ///
  int size(const Box& R, const Interval& comps) const ;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  static void setVerbose(bool a_verbose);
private:
  void
  setDefaultValues();

protected:
  static bool s_verbose;
  //get index vector<int> at the face location
  //using lexographic ordering of vof's indicies
  int getLocalVecIndex(const FaceIndex& a_face) const;
  //get index into dataPtr
  int getIndex(const FaceIndex& a_face,const int& a_comp) const;
  //has to be a pointer because we are planning to ship
  //the damned  thing to fortran.  
  T* m_dataPtr;
  int m_nComp;
  int m_nFaces;
  int m_direction;
  IntVect m_loVect;
  IntVect m_hiVect;
  // This is how we map from face to vector of data:
  // The low  intvect (by convention)
  // tells you the vector<int>
  // To index into the vector<int> use the lexigraphic
  // ordering of the cellindicies of the VoF (low, high).
  // That tells you the offset into m_dataPtr.
  // you further have to account for the variable number.
  // the getIndex() function encapsulates the process
  BaseFab<Vector<int> > m_ivmap;
  IntVectSet m_ivs;
  EBGraph m_ebgraph;
  bool m_isDefined;
private:
  BaseIFFAB<T>& operator= (const BaseIFFAB<T>&);
  BaseIFFAB (const BaseIFFAB<T>&);
};

#include "BaseIFFABI.H"

#endif
