/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _BaseIVFAB_H_
#define _BaseIVFAB_H_

#include <cmath>
#include <cstdlib>
#include <cassert>
#include "Vector.H"
#include "IntVectSet.H"
#include "VolIndex.H"
#include "BaseFab.H"
#include "EBGraph.H"

///  
/**
   BaseIVFAB is a templated 
   data holder defined at the VoFs of an irregular domain.
*/
template <class T>
class BaseIVFAB
{
public:
  ///
  /**
     Default constructor.  Constructs an uninitialized IVFAB.  You must
     subsequently call {\bf define} before using this IVFAB.
  */
  BaseIVFAB();

  ///
  /**
     Defining constructor.  Specifies the irregular domain
     and the number of data components per VoF.  The irregular domain
     must lie completely within the EBGraph.  The
     contents are uninitialized.  Calls full define function.
  */
  BaseIVFAB(const IntVectSet& a_region, 
            const EBGraph&    a_ebgraph, 
            const int&        a_nvarin);

  ///
  ~BaseIVFAB();

  ///
  /**
     Full define function.  Specifies the irregular domain and the 
     number of data components per VoF.  The irregular domain
     must lie completely within the EBGraph.   The contents
     are uninitialized.  If it has previously been defined, the old definition 
     data is overwritten and lost. 
  */
  void define(const IntVectSet& a_region, 
              const EBGraph&    a_ebgraph,
              const int&        a_nvarin);

  ///
  /**
     Set a value at every data location.
  */
  void setVal(const T& value);


  ///
  /**
     Set a value at every data location.
  */
  void setVal(int ivar, const T& value);

  ///
  void copy(const Box& a_fromBox, 
            const Interval& a_destInterval,
            const Box& a_toBox,
            const BaseIVFAB<T>& a_src, 
            const Interval& a_srcInterval);

   ///
  /**
         This stuff required by LevelData in parallel:
  */

  static int preAllocatable()
  {
    return 1; // symmetric allocatable.  
  }

  ///
  int size(const Box& R, const Interval& comps) const ;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  /**
     Remove all data from this BaseIVFAB.  
     You must call {\bf define} again in order 
     to use it.
  */
  void clear();

  ///
  /**

     Tells whether this BaseIVFAB has been defined, either with a constructor 
     or with {\bf define}.  It must be initialized in order to access its data.
  */        
  bool 
  isDefined() const;

  ///
  /**
     Return the number of VoFs in this BaseIVFAB.
  */
  int numVoFs() const; 

  ///
  /**
     Return the number of data components of this BaseIVFAB.
  */
  int nComp() const;

  ///
  /**
     Return  the  irregular domain of the BaseIVFAB.
  */
  const IntVectSet& getIVS() const;

  ///
  /**
     Indexing operator.  Return a reference to the contents of this IVFAB,
     at the specified VoF and data component.  The first component is
     zero, the last is {\em nvar-1}.  The returned object is a modifiable
     lvalue.
  */
  T& operator() (const VolIndex& a_vof,const int& varlocin);
  const T& operator() (const VolIndex& a_vof,const int& varlocin) const;

  ///
  /**
     Return a const pointer to the internal data storage of this IVFAB. \\
  */
  const T* dataPtr(const int& a_comp) const;

  ///
  /**
     Return a pointer to the internal data storage of this IVFAB. 
  */
  T* dataPtr(const int& a_comp) ;

  ///invalid but necessary for leveldata to compile
  BaseIVFAB(const Box& a_region, int a_nVar)
  {
    MayDay::Error("invalid constructor called for baseebcellfab");
  }

  ///
  /**
     Return a fake "lower left corner of the domain" for use with Fortran. 
  */
  const int* loVect() const;

  ///
  /**
     Return a fake "upper right corner of the domain" for use with Fortran. 
  */
  const int* hiVect() const;

  ///
  const EBGraph& getEBGraph() const;

  ///
  static void setVerbose(bool a_verbose);
private:
  void
  setDefaultValues();

protected:
  //get index into dataPtr
  int getIndex(const VolIndex& a_vof,const int& a_comp) const;
  //has to be a pointer because we are planning to ship
  //the damned  thing to fortran.  
  T* m_dataPtr;
  int m_nComp;
  int m_nVoFs;
  IntVect m_loVect;
  IntVect m_hiVect;
  EBGraph m_ebgraph;

  //this is how we map from vof to vector of data.
  //grab the vof and its intvect tells you the vector<int>
  //the integer in the vof indexes into the vof.
  //that tells you the offset into m_dataPtr 
  //you further have to account for the variable number.
  //the getIndex() function encapsulates the process
  BaseFab<Vector<int> > m_ivmap;
  IntVectSet m_ivs;
  bool m_isDefined;
  
  static bool s_verbose;
private:
  //disallowed for all the usual reasons
  BaseIVFAB<T>& operator= (const BaseIVFAB<T>&);
  BaseIVFAB (const BaseIVFAB<T>&);
};

#include "BaseIVFABI.H"

#endif

