/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _BASEIVFABI_H_
#define _BASEIVFABI_H_
#include "MayDay.H"
#include "IntVectSet.H"
#include "VoFIterator.H"
#include "parstream.H"
#include "SPMD.H"
#include "DebugOut.H"
#include "EBDebugOut.H"
#include "SPMD.H"

template <class T> 
bool BaseIVFAB<T>::s_verbose = false;

template <class T> 
void 
BaseIVFAB<T>::setVerbose(bool a_verbose) 
{
  s_verbose = a_verbose;
}
/******************/
template <class T> inline 
const EBGraph& 
BaseIVFAB<T>::getEBGraph() const
{
  return m_ebgraph;
}
/*************/
template <class T> inline
const IntVectSet& 
BaseIVFAB<T>::getIVS() const
{
  return m_ivs;
}
/******************/
template <class T> inline 
BaseIVFAB<T>::BaseIVFAB()
{
  setDefaultValues();
}
/******************/
template <class T> inline 
BaseIVFAB<T>::~BaseIVFAB()
{
  clear();
}
/******************/
template <class T> inline 
BaseIVFAB<T>::BaseIVFAB(const IntVectSet& a_ivsin, 
                        const EBGraph&    a_ebgraph,
                        const int&        a_nvarin)
{
  setDefaultValues();
  define(a_ivsin, a_ebgraph, a_nvarin);
}
/******************/
template <class T> inline 
void
BaseIVFAB<T>::define(const IntVectSet& a_ivsin, 
                     const EBGraph& a_ebGraph,
                     const int& a_nvarin)
{
  clear();
  m_isDefined = true;
  assert(a_nvarin > 0);
  m_ivs = a_ivsin;
  m_nComp = a_nvarin;
  m_ebgraph = a_ebGraph;
  m_nVoFs = 0;
  if(!a_ivsin.isEmpty())
    {
      Box minbox = a_ivsin.minBox();
      m_ivmap.resize(minbox, 1);
      IVSIterator ivsit(a_ivsin);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebgraph.getVoFs(iv);
          int numvofs = vofs.size();
          Vector<int>& vIndex = m_ivmap(iv, 0);
          vIndex.resize(numvofs);
          for(int ivof = 0; ivof < numvofs; ivof++)
            {
              vIndex[ivof] = m_nVoFs;
              m_nVoFs++;
            }
        }

      if( (m_nVoFs > 0) && m_nComp > 0)
        {
          m_dataPtr = new T[m_nComp*m_nVoFs];
        }
      else
        {
          m_dataPtr = NULL;
        }
    }
  else
    {
      m_dataPtr = NULL;
    }
  if(m_nVoFs > 0)
    {
      //set up face low and high vectors
      for (int dir = 0; dir < CH_SPACEDIM; ++dir)
        {
          m_loVect[dir] = 1;
          m_hiVect[dir] = 1;
        }
      m_hiVect[0] = m_nVoFs;

    }
  else
    {
      m_loVect = IntVect::Unit;
      m_hiVect = IntVect::Zero;
      m_dataPtr = NULL;
    }

}
/******************/
template <class T> inline 
void
BaseIVFAB<T>::setVal(const T& a_value)
{
  assert(isDefined());
  for(int ivec = 0; ivec < m_nVoFs*m_nComp; ivec++)
    {
      m_dataPtr[ivec] = a_value;
    }
}

/******************/
template <class T> inline 
void
BaseIVFAB<T>::setVal(int ivar, const T& a_value)
{
  assert(isDefined());
  assert(ivar >= 0);
  assert(ivar < m_nComp);

  int startval =     ivar*m_nVoFs;
  int endval   = (ivar+1)*m_nVoFs;
  for(int ivec = startval; ivec < endval; ivec++)
    {
      m_dataPtr[ivec] = a_value;
    }
}
/******************/
template <class T> inline 
void
BaseIVFAB<T>::copy(const Box& a_fromBox, 
                   const Interval& a_dstInterval,
                   const Box& a_toBox,
                   const BaseIVFAB<T>& a_src, 
                   const Interval& a_srcInterval)
{
  assert(a_fromBox == a_toBox);
  assert(isDefined());
  assert(a_src.isDefined());
  assert(a_srcInterval.size() == a_dstInterval.size());
  assert(a_dstInterval.begin() >= 0);
  assert(a_srcInterval.begin() >= 0);
  assert(a_dstInterval.end()   < m_nComp);
  assert(a_srcInterval.end()   < a_src.m_nComp);

  if((!m_ivs.isEmpty()) && (!a_src.m_ivs.isEmpty()))
    {
      Box intBox = a_fromBox;
      IntVectSet ivsIntersect = m_ivs;
      ivsIntersect &= a_src.m_ivs; //how did this ever work without this line?
      ivsIntersect &= intBox;  // 
      BaseIVFAB<T>& thisFAB = *this;
      int compSize = a_srcInterval.size();
      for(VoFIterator vofit(ivsIntersect, m_ebgraph); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          for(int icomp = 0; icomp < compSize; icomp++)
            {
              int isrccomp = a_srcInterval.begin() + icomp;
              int idstcomp = a_dstInterval.begin() + icomp;
              thisFAB(vof, idstcomp) = a_src(vof, isrccomp);
            }
        }
    }
}
/******************/
template <class T> inline 
int BaseIVFAB<T>::size(const Box& a_region, 
                       const Interval& a_comps) const
{
  assert(isDefined());
  //create set of cells in fab that are also in the input region
  IntVectSet subIVS = m_ivs;
  subIVS &= a_region;

  VoFIterator vofit(subIVS, m_ebgraph);
  const Vector<VolIndex>& vofs = vofit.getVector();
  
  //account for vof list
  int vofsize  = linearListSize(vofs);
  
  //add for each data point
  int datasize= vofs.size()*a_comps.size()*sizeof(T);
  
  int retval = vofsize + datasize;
  return retval;
}
/********************/
template <class T> inline 
void BaseIVFAB<T>::linearOut(void* a_buf, 
                             const Box& a_region, 
                             const Interval& a_comps) const
{
  assert(isDefined());
  //create set of cells in fab that are also in the input region
  IntVectSet subIVS = m_ivs;
  subIVS &= a_region;
  
  VoFIterator vofit(subIVS, m_ebgraph);
  const Vector<VolIndex>& vofs = vofit.getVector();
  //output the vofs.
  unsigned char* charbuffer = (unsigned char *) a_buf;
  linearListOut(charbuffer, vofs);
  charbuffer += linearListSize(vofs);

  //cast buffer into a pointer of the correct type
  //this way incrementing the pointer goes by the correct amount
  T* tbuffer = (T*) charbuffer;
  const BaseIVFAB<T>& thisFAB = *this;
  for(int ivof = 0; ivof < vofs.size(); ivof++)
    {
      const VolIndex& vof = vofs[ivof];
      for(int icomp = a_comps.begin(); icomp <= a_comps.end(); icomp++)
        {
          *tbuffer = thisFAB(vof, icomp);
          //increment the buffer offset
          ++tbuffer;
        }
    }
}
/********************/
template <class T> inline 
void BaseIVFAB<T>::linearIn(void* a_buf, const Box& a_region, const Interval& a_comps)
{
  assert(isDefined());
  //input the vofs
  Vector<VolIndex> vofs;
  unsigned char* charbuffer = (unsigned char *) a_buf;
  linearListIn(vofs, charbuffer);
  charbuffer += linearListSize(vofs);

  //cast buffer into a pointer of the correct type
  //this way incrementing the pointer goes by the correct amount
  T* tbuffer = (T*) charbuffer;
  BaseIVFAB<T>& thisFAB = *this;
  for(int ivof = 0; ivof < vofs.size(); ivof++)
    {
      const VolIndex& vof = vofs[ivof];
      for(int icomp = a_comps.begin(); icomp <= a_comps.end(); icomp++)
        {
          thisFAB(vof, icomp) = *tbuffer ;
          //increment the buffer offset
          ++tbuffer;
        }
    }
  
}
/********************/
template <class T> inline 
int
BaseIVFAB<T>::getIndex(const VolIndex& a_vof, const int& a_comp) const
{
  assert(isDefined());
  assert(m_ivs.contains(a_vof.gridIndex()));
  assert((a_comp >= 0) && (a_comp < m_nComp));

  const Vector<int>& vecOffset = m_ivmap(a_vof.gridIndex(), 0);
  //this checks whether the volindex is defined 
  assert(a_vof.cellIndex() >= 0);
  //if this fails on a size() thing, then the
  //basefab<vector<int> > was constructed incorrectly
  //(the assert should catch wheter the vof is outside the orig ivs)
  int ioffset = vecOffset[a_vof.cellIndex()];
  //need these because we are indexing into a straight
  //pointer (no vector bounds checking).
  assert(ioffset >= 0);
  assert(ioffset < m_nVoFs);
  //now add offset from componentnitude
  ioffset += m_nVoFs*a_comp;
  return ioffset;
}
/********************/
template <class T> inline 
void
BaseIVFAB<T>::clear()
{
  m_nComp = 0;
  m_nVoFs = 0;
  m_ivs.makeEmpty();
  m_ivmap.clear();
  if(m_dataPtr != NULL)
    {
      delete[] m_dataPtr;
      m_dataPtr = NULL;
    }
  m_isDefined = false;
}
/*************************/
template <class T> inline 
bool 
BaseIVFAB<T>::isDefined() const
{
  return (m_isDefined);
}
/*************************/
template <class T> inline 
int
BaseIVFAB<T>::numVoFs() const 
{
  return m_nVoFs;
}
/*************************/
template <class T> inline 
int
BaseIVFAB<T>::nComp() const 
{
  return m_nComp;
}
/*************************/
template <class T> inline 
T&
BaseIVFAB<T>::operator() (const VolIndex& a_ndin,
                          const int& a_comp)
{ 
  assert(isDefined());
  assert(a_comp >= 0);
  assert(a_comp < m_nComp);
  int iloc = getIndex(a_ndin, a_comp);
  return(m_dataPtr[iloc]);
}
/**************************/
template <class T> inline 
const T&
BaseIVFAB<T>::operator() (const VolIndex& a_ndin,
                          const int& a_comp) const
{ 
  assert(isDefined());
  assert(a_comp >= 0);
  assert(a_comp < m_nComp);
  int iloc = getIndex(a_ndin, a_comp);
  return(m_dataPtr[iloc]);
}
/******************/
template <class T> inline 
const T*
BaseIVFAB<T>::dataPtr(const int& a_comp) const
{
  assert(isDefined());
  assert(a_comp >= 0);
  assert(a_comp < m_nComp);
  return m_dataPtr + a_comp*m_nVoFs;
}
/******************/
template <class T> inline 
T* 
BaseIVFAB<T>::dataPtr(const int& a_comp) 
{
  assert(isDefined());
  assert(a_comp >= 0);
  assert(a_comp < m_nComp);
  return m_dataPtr + a_comp*m_nVoFs;
}
/******************/
template <class T> inline 
const int*
BaseIVFAB<T>::loVect() const 
{
  return m_loVect.getVect();
}
/******************/
template <class T> inline 
const int*
BaseIVFAB<T>::hiVect() const 
{
  return m_hiVect.getVect();
}
/******************/
template <class T> inline 
void
BaseIVFAB<T>::setDefaultValues()
{
  m_isDefined = false;
  m_dataPtr = NULL;
  m_nVoFs = 0;
  m_nComp = 0;
  m_loVect = IntVect::Unit;
  m_hiVect = IntVect::Zero;
}
/******************/
template <class T> inline 
BaseIVFAB<T>& 
BaseIVFAB<T>::operator= (const BaseIVFAB<T>& a_input)
{
  MayDay::Error("BaseIVFAB operator = not defined");
  return *this;
}
/******************/
template <class T> inline 
BaseIVFAB<T>::BaseIVFAB (const BaseIVFAB<T>& a_input)
{
  MayDay::Error("BaseIVFAB copy constructor not defined");
}

#endif

