/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _EB_ARITH_H_
#define _EB_ARITH_H_

#include "EBCellFAB.H"
#include "EBISBox.H"
#include "EBISLayout.H"
#include "EBFaceFAB.H"
#include "BoxLayout.H"
#include "BoxLayoutData.H"
#include "Stencils.H"
#include "LayoutData.H"
#include "Tuple.H"
#include "EBFluxFAB.H"
#include "EBCellFAB.H"
///
/**
   enumeration class to ennumerate the different
   extents of norm covereage. sometimes one
   wants to only take a norm over the irregular 
   cells.  sometimes one only wants to take a norm
   over the regular cells, and sometimes one wants
   both.
 */
class EBNormType
{
public:
  ///
  /**
     OverOnlyRegular means take the norm only over
     regular cells. \\
     OverOnlyirregular means take the norm only over
     irregular cells. \\
     OverBoth means use all uncovered cells. \\
     All norms exclude covered cells.
   */
  enum  NormMode{OverOnlyRegular=0, OverOnlyIrregular, OverBoth};
};
///
/**
   class to encapsulate the common arithmatic operations 
   for ebfabs
 */
class EBArith
{
public:

  ///
  static void
  interpolateFluxToCentroids(LevelData<EBFluxFAB>&       a_centroidFlux,
                             const LevelData<EBFluxFAB>& a_faceCentFlux,
                             const DisjointBoxLayout&    a_grids,
                             const EBISLayout&           a_ebisl,
                             const Box&                  a_domain);


  ///
  static void
  interpolateFluxToCentroids(EBFaceFAB&                  a_centroidFlux,
                             const EBFaceFAB&            a_faceCentFlux,
                             const Box&                  a_box,
                             const EBISBox&              a_ebisBox,
                             const Box&                  a_domain,
                             const int&                  a_idir);

  ///
  static void
  irregNorm(Real& a_ebIrregNorm,
            const BaseIVFAB<Real>& a_ebiError,
            const IntVectSet& a_ivsIrreg,
            const EBISBox& a_ebisBox, 
            const int& a_comp,
            const int& a_normtype);


  ///
  static FaceStencil getInterpStencil(const FaceIndex&  a_face, 
                                      const IntVectSet& a_coarseFineIVS,
                                      const EBISBox&    a_ebisBox,
                                      const Box&        a_domainBox);

  ///
  static void getInterpStencil2D(FaceStencil&      a_sten,
                                 const FaceIndex&  a_face, 
                                 const IntVectSet& a_coarseFineIVS,
                                 const EBISBox&    a_ebisBox,
                                 const Box&        a_domainBox) ;

  ///
  static void getInterpStencil3D(FaceStencil& a_sten,
                                 const FaceIndex& a_face, 
                                 const IntVectSet& a_coarseFineIVS,
                                 const EBISBox&   a_ebisBox,
                                 const Box&       a_domainBox);

  ///
  /**
     Return true if there is unique vof associated with a_cell 
     that lies in the a_vofsStencil. If so, a_vof = vof. Else return false.
  */
  static bool
  isVoFHere(VolIndex& a_vof,
            const Vector<VolIndex>& a_vofsStencil,
            const IntVect& a_cell );

  /// 
  /** Returns all vofs that can be reached from a_vof via a 
      monotone path of length <= than radius
  */
  static void
  getAllVoFsInMonotonePath(Vector<VolIndex>& a_vofList,
                           const VolIndex&   a_vof,
                           const EBISBox&    a_ebisBox,
                           const int&        a_redistRad);

  ///get volume of a vof in rz coords
  static void
  getKVolRZ(Real&           a_kvol, 
            Real&           a_cellVol, 
            const EBISBox&  a_ebisBox,
            const Real&     a_dx,
            const VolIndex& a_vof);

  ///
  /**
     Return true if there is a unique adjacent face in the 
     given direction.  False if either covered or multivalued.
   */
  static
  bool
  getAdjacentFace(FaceIndex& a_adjacentFace,
                  const FaceIndex& a_face,
                  const EBISBox& a_ebisBox, 
                  const Box& a_domain,
                  const int& a_idir,
                  const Side::LoHiSide& a_side);



  ///
  /**
     compute a stencil for one of our fancy interpolated
     gradients.
   */
  static
  void computeGradFluxStencil(VoFStencil& a_thisStencil,
                              const FaceIndex& a_thisFace,
                              const EBISBox& a_ebisBox,
                              const Box& a_domain,
                              const int& a_dir);

  ///
  /**
     compute a stencil for one of our fancy interpolated 
     fluxes.
   */
  static
  void computeInterpStencil(FaceStencil& a_thisStencil,
                            const FaceIndex& a_thisFace,
                            const EBISBox& a_ebisBox,
                            const Box& a_domain,
                            const int& a_dir);

  ///
  /**
     return lp-norm of component comp of a_src,
     weighted by local volume fraction.
     not normalized by number of points or anything like that.
     if p==0, volume returned is one 
     and norm returned is Max(abs(a_src)) over
     uncovered regions.  otherwise,
     returns sum(volfrac*a_src(iv,comp)**p)  of component comp of a_src
     weighted by local volume fraction and
     also returns volume of uncovered regions. 
     Only uncovered regions count here.
   */
  static
  void
  volWeightedSum(Real& norm, Real& volume,
                 const EBCellFAB& a_src, 
                 const Box& a_region,
                 const EBISBox& a_ebisBox,
                 const int& comp,
                 const int& p,
                 EBNormType::NormMode = EBNormType::OverBoth);


  ///
  /** return l-p norm of a_src.  
      if p==0, v
      norm returned is Max(abs(a_src)) over
      uncovered regions.  otherwise,
      returns 1/vol(sum(volfrac*a_src(iv,comp)**p)^(1/p))  
      of component comp of a_src
      weighted by local volume fraction and
      also returns volume of uncovered regions. 
      Only uncovered regions count here.
      The data must have the same layout as a_layout with
      the possible exception of ghost cells.
   */
  static
  Real 
  norm(const BoxLayoutData<EBCellFAB >& a_dataOne, 
       const BoxLayout& a_layout, 
       const EBISLayout& a_ebisl, 
       const int& comp,
       const int&  p,
       EBNormType::NormMode = EBNormType::OverBoth);

  ///
  static 
  void
  computeCoveredFaces(Vector<VolIndex>&     a_coveredFace, 
                      IntVectSet&           a_coveredSets,
                      IntVectSet&           a_irregIVS,
                      const int&            a_idir, 
                      const Side::LoHiSide& a_sd, 
                      const EBISBox&        a_ebisBox,
                      const Box&            a_region);

  //testing interface
  static
  Real 
  norm(Real& volume, const BoxLayoutData<EBCellFAB >& a_dataOne, 
       const BoxLayout& a_layout, 
       const EBISLayout& a_ebisl, 
       const int& comp,
       const int&  p,
       EBNormType::NormMode = EBNormType::OverBoth);

  ///
  /** return l-p norm of a_src.  
      if p==0, v
      norm returned is Max(abs(a_src)) over
      uncovered regions.  otherwise,
      returns 1/vol(sum(volfrac*a_src(iv,comp)**p)^(1/p))  
      of component comp of a_src
      weighted by local volume fraction and
      also returns volume of uncovered regions. 
      Only uncovered regions count here.
      The data must have the same layout as a_layout with
      the possible exception of ghost cells.
   */
  static
  Real 
  norm(const EBCellFAB& a_dataOne, 
       const Box&       a_grid, 
       const EBISBox&   a_ebisl, 
       const int&       a_comp,
       const int&       a_p,
       EBNormType::NormMode = EBNormType::OverBoth);


  static
  Real 
  norm(Real& volume, const EBCellFAB& a_dataOne, 
       const Box&       a_grid, 
       const EBISBox&   a_ebisl, 
       const int&       a_comp,
       const int&       a_p,
       EBNormType::NormMode = EBNormType::OverBoth);

  ///
  /** return l-p norm of a_src.  
      if p==0, v
      norm returned is Max(abs(a_src)) over
      uncovered regions.  otherwise,
      returns sum(volfrac*a_src(iv,comp)**p)  of component comp of a_src
      weighted by local volume fraction and
      also returns volume of uncovered regions. 
      Only uncovered regions count here.
      The data must have the same layout as a_layout with
      the possible exception of ghost cells.
   */
  static
  void
  volWeightedSum(Real& sum, Real& volume, 
                 const BoxLayoutData<EBCellFAB >& a_dataOne, 
                 const BoxLayout& a_layout, 
                 const EBISLayout& a_ebisl, 
                 const int& comp,
                 const int& p,
                 EBNormType::NormMode = EBNormType::OverBoth);

  ///
  /** return the dotproduct of two leveldatas of ebfabs,
      Only uncovered regions count here.
   */
  static
  Real 
  dotProduct(const BoxLayoutData<EBCellFAB >& a_dataOne, 
             const BoxLayoutData<EBCellFAB >& a_dataTwo,
             const BoxLayout& a_layout, 
             const EBISLayout& a_ebisl, 
             const int& a_comp);

  ///
  /** return the dotproduct of two  ebfabs,
      Only uncovered regions count here.
   */
  static
  Real 
  dotProduct(const EBCellFAB& a_dataOne, 
             const EBCellFAB& a_dataTwo,
             const Box& a_layout, 
             const EBISBox& a_ebisBox,
             const int& a_comp);


  /// 
  /** Given a VoF, a_vof1, and a cell, a_cell2, determine if there is a single
      VoF in a_cell2 that connects to a_vof1 via a monotone path.  If there is
      one such VoF then TRUE is returned (and the VolIndex is returned,
      a_vof2).  If there is no such VoF or if there are more than one such
      VoF then FALSE is returned (and a_vof2 is unchanged).
   */
  static
  bool
  monotonePathVoFToCellVoF(VolIndex& a_vof2,
                           const VolIndex& a_vof1,
                           const IntVect& a_cell2,
                           const EBISBox& a_ebisBox);
  /// 
  /** Returns all vofs that can be reached from a_vof via a 
      monotone path of length <= than radius
  */
  static
  void
  getAllVoFsInMonotonePath(Vector<VolIndex>& a_vofs,
                           const IntVect&    a_timesMoved,
                           const IntVect&    a_pathSign,
                           const VolIndex&   a_vof,
                           const EBISBox&    a_ebisBox,
                           const int&        a_radius);
  

  static VolIndex& getVoFMax()
  {
    return s_vofMax;
  }
  static Real&     getValMax()
  {
    return s_valMax;
  }
  
private:
  static VolIndex s_vofMax;
  static Real     s_valMax;

};


extern IntVect ebcoarsen (const IntVect& b,
                          int        refinement_ratio);

extern
Box ebrefine (const Box& b,
              int        refinement_ratio);

extern Box ebcoarsen (const Box& b,
                      int        refinement_ratio);

extern void ebrefine(DisjointBoxLayout& output, 
                     const DisjointBoxLayout& input, 
                     int refinement);

extern void ebcoarsen(DisjointBoxLayout& output, 
                      const DisjointBoxLayout& input, 
                      int refinement);

#endif
