/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "EBArith.H"
#include "EBArithF_F.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "VoFIterator.H"
#include "FaceIterator.H"
#include "BaseFab.H"
#include "PolyGeom.H"
#include "BaseEBCellFactory.H"

Real     EBArith::s_valMax = 0.0;
VolIndex EBArith::s_vofMax = VolIndex(IntVect::Zero, 0);

/****/
void
EBArith::
interpolateFluxToCentroids(LevelData<EBFluxFAB>&       a_centroidFlux,
                           const LevelData<EBFluxFAB>& a_faceCentFlux,
                           const DisjointBoxLayout&    a_grids,
                           const EBISLayout&           a_ebisl,
                           const Box&                  a_domain)
{
  for(DataIterator dit = a_grids.dataIterator(); dit.ok(); ++dit)
    {
      EBFluxFAB&       centroidFluxFAB = a_centroidFlux[dit()];
      const EBFluxFAB& faceCentFluxFAB = a_faceCentFlux[dit()];
      const EBISBox&   ebisBox = a_ebisl[dit()];
      const Box& box = a_grids.get(dit());
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          EBFaceFAB&       centroidFaceFAB = centroidFluxFAB[idir];
          const EBFaceFAB& faceCentFaceFAB = faceCentFluxFAB[idir];
          interpolateFluxToCentroids(centroidFaceFAB, faceCentFaceFAB,  box, ebisBox, a_domain, idir);
        }
    }
}
/****/
void
EBArith::
interpolateFluxToCentroids(EBFaceFAB&                  a_centroidFlux,
                           const EBFaceFAB&            a_faceCentFlux,
                           const Box&                  a_box,
                           const EBISBox&              a_ebisBox,
                           const Box&                  a_domain,
                           const int&                  a_idir)
{
  FaceStop::WhichFaces stopCrit = FaceStop::SurroundingWithBoundary;
  IntVectSet cfivs;

  //first copy all the fluxes over then interpolate on irregular cells
  a_centroidFlux.copy(a_faceCentFlux);
  IntVectSet irregIVS = a_ebisBox.getIrregIVS(a_box);

  for(FaceIterator faceit(irregIVS, a_ebisBox.getEBGraph(), a_idir, stopCrit);
      faceit.ok(); ++faceit)
    {
      const FaceIndex& face = faceit();
      FaceStencil sten = getInterpStencil(face, cfivs, a_ebisBox, a_domain);
      for(int ivar = 0; ivar < a_centroidFlux.nComp(); ivar++)
        {
          Real newFlux = 0.0;
          for(int isten = 0; isten < sten.size(); isten++)
            {
              const FaceIndex& stenFace = sten.face(isten);
              Real weight     = sten.weight(isten);
              Real interpFlux = a_faceCentFlux(stenFace, ivar);
              newFlux += weight*interpFlux;
            }
          a_centroidFlux(face, ivar) = newFlux;
        }
    }
}
/****/
void EBArith::
irregNorm(Real& a_ebIrregNorm,
          const BaseIVFAB<Real>& a_ebiError,
          const IntVectSet& a_ivsIrreg,
          const EBISBox& a_ebisBox, 
          const int& a_comp,
          const int& a_normtype)
{
  assert(a_normtype >= 0);
  assert(a_normtype <= 2);
  a_ebIrregNorm = 0.0;
  if(a_normtype == 0)
    {
      for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          Real valVoF = a_ebiError(vof, a_comp);
          a_ebIrregNorm = Max(Abs(valVoF), a_ebIrregNorm);
        }
    }
  else
    {
      //integral norm
      Real areaTot = 0.0;
      Real normTot = 0.0;
      for(VoFIterator vofit(a_ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          Real valVoF = a_ebiError(vof, a_comp);
          Real bdarea = a_ebisBox.bndryArea(vof);
          areaTot += bdarea;
          if(a_normtype == 1)
            normTot += Abs(valVoF)*bdarea;
          else
            normTot += valVoF*valVoF*bdarea;
        }
      if(areaTot > 1.0e-8)
        normTot /= areaTot;
      if(a_normtype == 2)
        normTot = sqrt(normTot);
      a_ebIrregNorm = normTot;
    }

}
IntVect ebcoarsen (const IntVect& b,
                   int        refinement_ratio)
{
  return coarsen(b, refinement_ratio);
}


Box ebrefine (const Box& b,
              int        refinement_ratio)
{
  return refine(b, refinement_ratio);
}

Box ebcoarsen (const Box& b,
               int        refinement_ratio)
{
  return coarsen(b, refinement_ratio);
}

void ebrefine(DisjointBoxLayout& output, 
              const DisjointBoxLayout& input, 
              int refinement)
{
  return refine(output, input, refinement);
}

void ebcoarsen(DisjointBoxLayout& output, 
               const DisjointBoxLayout& input, 
               int refinement)
{
  return coarsen(output, input, refinement);
}
/*****************************/
void
EBArith::
getKVolRZ(Real&           a_kvol, 
          Real&           a_cellVol, 
          const EBISBox&  a_ebisBox,
          const Real&     a_dx,
          const VolIndex& a_vof) 
{
  int rdir = 0;
  int rindex   =  a_vof.gridIndex()[rdir];
  Real cellRad = a_dx*(Real(rindex) + 0.5);
  a_cellVol =cellRad*a_dx*a_dx;

  //kvol = (1/2*cellvol)*int(r^2 nr dl)
  Real openContrib = 0.;
  for(SideIterator sit; sit.ok(); ++sit)
    {
      Vector<FaceIndex> rfaces = a_ebisBox.getFaces(a_vof, rdir, sit());
      int isign = sign(sit());
      for(int iface = 0; iface < rfaces.size(); iface++)
        {
          const FaceIndex& face = rfaces[iface];
          Real faceRad = cellRad + 0.5*isign*a_dx;
          Real faceRad2 = a_dx*Real(face.gridIndex(Side::Hi)[rdir]);
          assert(Abs(faceRad - faceRad2) < 1.0e-9);
          Real areaFrac = a_ebisBox.areaFrac(face);
          openContrib += isign*areaFrac*faceRad*faceRad;
        }
    }
  //add in contribution from irregular face using midpoint rule
  Real bndryArea  = a_ebisBox.bndryArea(a_vof);
  Real rnormal    = a_ebisBox.normal(a_vof)[rdir];
  Real rbndryCent = a_ebisBox.bndryCentroid(a_vof)[rdir];
  Real radCent = cellRad + a_dx*rbndryCent;
  Real bdryContrib= -bndryArea*rnormal*radCent*radCent;

  Real factor = a_dx/(2.*a_cellVol);

  openContrib *=  factor;
  bdryContrib *=  factor;
  a_kvol =  openContrib + bdryContrib;
  //the two is from r^2/2
  //the dx comes from dl
}
/****/                        
Real 
EBArith::norm(Real&                a_volume,
              const EBCellFAB&     a_dataOne, 
              const Box&           a_grid, 
              const EBISBox&       a_ebisBox, 
              const int&           a_comp,
              const int&           a_pval,
              EBNormType::NormMode a_mode)
{
  Real normval, sum;
  volWeightedSum(sum, a_volume, a_dataOne, 
                 a_grid, a_ebisBox, a_comp, a_pval, a_mode);

  if(a_pval == 0)
    {
      normval = sum;
    }
  else if(a_pval == 1)
    {
      if(a_volume > 0.0)
        {
          normval = Abs(sum)/a_volume;
        }
      else
        {
          normval = 0.0;
        }
    }
  else
    {
      Real denom = a_pval;
      Real exponent = 1.0/denom;
      if(a_volume > 0.0)
        {
          sum /= a_volume;
          normval = pow(sum, exponent);
        }
      else
        {
          normval = 0.0;
        }
    }

  return normval;
}

Real 
EBArith::norm(const EBCellFAB&     a_dataOne, 
              const Box&           a_grid, 
              const EBISBox&       a_ebisBox, 
              const int&           a_comp,
              const int&           a_pval,
              EBNormType::NormMode a_mode)
{
  Real volume, normval, sum;
  volWeightedSum(sum, volume, a_dataOne, 
                 a_grid, a_ebisBox, a_comp, a_pval, a_mode);


  if(a_pval == 0)
    {
      normval = sum;
    }
  else if(a_pval == 1)
    {
      if(volume > 0.0)
        {
          normval = Abs(sum)/volume;
        }
      else
        {
          normval = 0.0;
        }
    }
  else
    {
      Real denom = a_pval;
      Real exponent = 1.0/denom;
      if(volume > 0.0)
        {
          sum /= volume;
          normval = pow(sum, exponent);
        }
      else
        {
          normval = 0.0;
        }
    }

  return normval;
}

bool
EBArith::getAdjacentFace(FaceIndex& a_adjacentFace,
                         const FaceIndex& a_face,
                         const EBISBox& a_ebisBox, 
                         const Box& a_domainBox,
                         const int& a_idir,
                         const Side::LoHiSide& a_side)
{
  bool uniqueFace = true;
  int faceDir = a_face.direction();
  VolIndex loFaceVoF = a_face.getVoF(Side::Lo);
  VolIndex hiFaceVoF = a_face.getVoF(Side::Hi);
  //figure out if we have the connectivity to find a face
  //with which to interpolate.  If so,
  //find the other face
  if(!a_face.isBoundary())
    {
      Vector<FaceIndex> loTanFaces = a_ebisBox.getFaces(loFaceVoF, a_idir, a_side);
      Vector<FaceIndex> hiTanFaces = a_ebisBox.getFaces(hiFaceVoF, a_idir, a_side);
      if((loTanFaces.size() != 1) || (hiTanFaces.size() != 1))
        {
          uniqueFace = false;
        }
      else if((loTanFaces[0].isBoundary()) && (hiTanFaces[0].isBoundary()))
        {
          uniqueFace = false;
        }
      else
        {
          const FaceIndex& loTanFace = loTanFaces[0];
          const FaceIndex& hiTanFace = hiTanFaces[0];
          VolIndex loOtherVoF = loTanFace.getVoF(a_side);
          VolIndex hiOtherVoF = hiTanFace.getVoF(a_side);
          if(!(a_ebisBox.isConnected(loOtherVoF, hiOtherVoF)))
            {
              uniqueFace = false;
            }
          if(uniqueFace)
            {
              Vector<FaceIndex> otherFaces = a_ebisBox.getFaces(loOtherVoF, faceDir, Side::Hi);
              if(otherFaces.size() != 1)
                {
                  uniqueFace = false;
                }
              else
                {
                  a_adjacentFace = otherFaces[0];
                }
            }
        }
    } //end if !face.isBoundary
  else
    {
      //boundary face.
      IntVect loVoFiv = loFaceVoF.gridIndex();
      IntVect hiVoFiv = hiFaceVoF.gridIndex();
      IntVect loDomiv = a_domainBox.smallEnd();
      IntVect hiDomiv = a_domainBox.bigEnd();
      Vector<FaceIndex> otherFaces;

      if(hiVoFiv[faceDir] == loDomiv[faceDir])
        {
          Vector<FaceIndex> hiTanFaces = a_ebisBox.getFaces(hiFaceVoF, a_idir, a_side);
          if(hiTanFaces.size() != 1)
            {
              uniqueFace = false;
            }
          else if(hiTanFaces[0].isBoundary())
            {
              //in 3d,  can be a boundary in two directions
              uniqueFace = false;
            }
          else
            {
              VolIndex hiOtherVoF = hiTanFaces[0].getVoF(a_side);
              Vector<FaceIndex> otherFaces = a_ebisBox.getFaces(hiOtherVoF, faceDir, Side::Lo);
              if(otherFaces.size() != 1)
                {
                  uniqueFace = false;
                }
              else
                {
                  a_adjacentFace = otherFaces[0];
                }
            }
          
        }
      else if(loVoFiv[faceDir] == hiDomiv[faceDir])
        {
          Vector<FaceIndex> loTanFaces = a_ebisBox.getFaces(loFaceVoF, a_idir, a_side);
          if(loTanFaces.size() != 1)
            {
              uniqueFace = false;
            }
          else if(loTanFaces[0].isBoundary())
            {
              //in 3d,  can be a boundary in two directions
              uniqueFace = false;
            }
          else
            {
              VolIndex loOtherVoF = loTanFaces[0].getVoF(a_side);
              Vector<FaceIndex> otherFaces = a_ebisBox.getFaces(loOtherVoF, faceDir, Side::Hi);
              if(otherFaces.size() != 1)
                {
                  uniqueFace = false;
                }
              else
                {
                  a_adjacentFace = otherFaces[0];
                }
            }
          
        }
      else
        {
          MayDay::Error("boundary face confusion");
        }
      
    }
  return uniqueFace;
}

/**************/
//If the face is covered, one of the vofs
//should have been covered which means
//i should not have gotten a face
//so I won't check that.
//If a vof is outside the domain, then
//we have a boundary face an those don't 
//get done here.
// check conditions for using a nontrivial O(h**2) flux stencil.
// The conditions are:
// (1) Is this face irregular ?
// (2) If so, are the components of the boundary normal in the directions 
//      tangent to the face of the same sign ?
// (3) If (1) or (2) are true, are the faces that will be used to
//      computed the centered flues regular, and are the vofs making up
//      those faces connected appropriately to the vofs making up 
//      face at which the flux is being evaluated ?
// If any of these conditions fail, then we use the centered difference
// flux. This corresponds to dropping order for some of the 
// irregular faces, but is still second order accurate for regular faces
// mumble set of codimension two mumble mumble
/**************/
void 
EBArith::computeInterpStencil(FaceStencil& a_thisStencil,
                              const FaceIndex& a_thisFace,
                              const EBISBox& a_ebisBox,
                              const Box& a_domain,
                              const int& a_dir)
{
  a_thisStencil.clear();
  const IntVect& ivlo = a_thisFace.gridIndex(Side::Lo);
  const IntVect& ivhi = a_thisFace.gridIndex(Side::Hi);

  //boundary faces are done separately
  if(!a_domain.contains(ivlo) || !a_domain.contains(ivhi))
    return;

  IntVect kperp = IntVect::Zero;
  for (int d = 0; d < SpaceDim-1; ++d) 
    {
      kperp[d] = (a_dir + d + 1) % SpaceDim;
    }

  bool MakeSimpleStencil = true;
  Real areaFrac = a_ebisBox.areaFrac(a_thisFace);
  if (areaFrac < 1.0)
    {
      // Condition (1) is met.
      VolIndex vofLo = a_thisFace.getVoF(Side::Lo);
      VolIndex vofHi = a_thisFace.getVoF(Side::Hi);

      RealVect rvnormLo = a_ebisBox.normal(vofLo);
      RealVect rvnormHi = a_ebisBox.normal(vofHi);
      //this cheecks that the normals are monotonic
      //in the non-facedir direction
      bool montonNorm = true;
      bool facesOk = true;
      for (int d = 0; d < SpaceDim-1; ++d) 
        {
          montonNorm = montonNorm 
            && ( rvnormLo[d]*rvnormHi[d] >= 0.0 );
        }
      if (montonNorm)
        {
          // Condition (2) is met.
          //find perpendicular directions and signs
          RealVect rvnorm = RealVect::Zero;
          IntVect  ivsign =  IntVect::Zero;
          Tuple<IntVect,SpaceDim-1> ivPerp;
          for (int d = 0; d < SpaceDim-1; ++d) 
            {
              rvnorm[d] = 0.5 * (rvnormLo[d] + rvnormHi[d]);
              if(rvnorm[d] > 0.0)
                ivsign[d] = 1;
              else
                ivsign[d] = -1;

              ivPerp[d] = IntVect::Zero;
              ivPerp[d].setVal(kperp[d],ivsign[d]);
              //by convention
              const IntVect& ivface = vofHi.gridIndex();
              ivPerp[d] += ivface;
              //this stuff  checks to see
              //if there are vofs in the perpendicular direction
              //that are regular.   if so, we can do the johansen
              //interpolation between faces thing without fear.
              //If there is no regular face nearby, this gets
              //godawful complicated and we just drop to the 
              //simple stencil.
              if (ivsign[d] != 0)
                {
                  const IntVect& perpIV = ivPerp[d];
                  IntVect othrIV = perpIV - BASISV(a_dir);
                  if(!a_ebisBox.getRegion().contains(perpIV) ||
                     !a_ebisBox.getRegion().contains(othrIV))
                    {
                      facesOk = false;
                    }
                  else if(a_ebisBox.isCovered(perpIV) ||
                          a_ebisBox.isCovered(othrIV))
                    {
                      facesOk = false;
                    }
                  else
                    {
                      facesOk = facesOk && (a_ebisBox.isRegular(perpIV) ||
                                            a_ebisBox.isRegular(othrIV));
                    }
                }
            }
          if (facesOk)
            {
              // Get vofs for faces at ivPerp to check connectivity.
              Tuple<FaceIndex,SpaceDim-1> perpFaces;
              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  if (ivsign[d] != 0)
                    {
                      perpFaces[d] =
                        a_ebisBox.getAllFaces(ivPerp[d],a_dir,Side::Lo)[0];
                      facesOk = facesOk &&
                        a_ebisBox.isConnected(perpFaces[d].getVoF(Side::Lo),vofLo);
                      facesOk = facesOk &&
                        a_ebisBox.isConnected(perpFaces[d].getVoF(Side::Hi),vofHi);
                    }
                }
              // Condition (3) is met - construct the stencil.
              // Compute the center of area for the face.
              RealVect faceCent = a_ebisBox.centroid(a_thisFace);
              Real thisCoef = 1.0;
              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  thisCoef -= Abs(faceCent[kperp[d]]);
                }
              a_thisStencil.add(a_thisFace, thisCoef);

              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  Real thisXC = Abs(faceCent[kperp[d]]);
                  if (ivsign[d] != 0)
                    {
                      a_thisStencil.add(perpFaces[d], thisXC);
                    }
                } //end loop over directions
              MakeSimpleStencil = false;
            } // end if(facesOK) (all conditons met
        } //end if(perp norms are monotonic) 
    } //end isFaceIrregular (if(areaFrac <1))

  if (MakeSimpleStencil)
    {
      a_thisStencil.add(a_thisFace, 1.0);
    } 
}
/**************/
//If the face is covered, one of the vofs
//should have been covered which means
//i should not have gotten a face
//so I won't check that.
//If a vof is outside the domain, then
//we have a boundary face an those don't 
//get done here.
// check conditions for using a nontrivial O(h**2) flux stencil.
// The conditions are:
// (1) Is this face irregular ?
// (2) If so, are the components of the boundary normal in the directions 
//      tangent to the face of the same sign ?
// (3) If (1) or (2) are true, are the faces that will be used to
//      computed the centered flues regular, and are the vofs making up
//      those faces connected appropriately to the vofs making up 
//      face at which the flux is being evaluated ?
// If any of these conditions fail, then we use the centered difference
// flux. This corresponds to dropping order for some of the 
// irregular faces, but is still second order accurate for regular faces
// mumble set of codimension two mumble mumble
/**************/
void 
EBArith::computeGradFluxStencil(VoFStencil& a_thisStencil,
                                const FaceIndex& a_thisFace,
                                const EBISBox& a_ebisBox,
                                const Box& a_domain,
                                const int& a_dir)
{

  a_thisStencil.clear();
  const IntVect& ivlo = a_thisFace.gridIndex(Side::Lo);
  const IntVect& ivhi = a_thisFace.gridIndex(Side::Hi);

  //boundary faces are done separately
  if(!a_domain.contains(ivlo) || !a_domain.contains(ivhi))
    return;

  IntVect kperp = IntVect::Zero;
  for (int d = 0; d < SpaceDim-1; ++d) 
    {
      kperp[d] = (a_dir + d + 1) % SpaceDim;
    }

  bool MakeSimpleStencil = true;
  Real areaFrac = a_ebisBox.areaFrac(a_thisFace);
  if (areaFrac < 1.0)
    {
      // Condition (1) is met.
      VolIndex vofLo = a_thisFace.getVoF(Side::Lo);
      VolIndex vofHi = a_thisFace.getVoF(Side::Hi);

      RealVect rvnormLo = a_ebisBox.normal(vofLo);
      RealVect rvnormHi = a_ebisBox.normal(vofHi);
      //this cheecks that the normals are monotonic
      //in the non-facedir direction
      bool montonNorm = true;
      bool facesOk = true;
      for (int d = 0; d < SpaceDim-1; ++d) 
        {
          montonNorm = montonNorm 
            && ( rvnormLo[d]*rvnormHi[d] >= 0.0 );
        }
      if (montonNorm)
        {
          // Condition (2) is met.
          //find perpendicular directions and signs
          RealVect rvnorm = RealVect::Zero;
          IntVect  ivsign =  IntVect::Zero;
          Tuple<IntVect,SpaceDim-1> ivPerp;
          for (int d = 0; d < SpaceDim-1; ++d) 
            {
              rvnorm[d] = 0.5 * (rvnormLo[d] + rvnormHi[d]);
              if(rvnorm[d] > 0.0)
                ivsign[d] = 1;
              else
                ivsign[d] = -1;

              ivPerp[d] = IntVect::Zero;
              ivPerp[d].setVal(kperp[d],ivsign[d]);
              //by convention
              const IntVect& ivface = vofHi.gridIndex();
              ivPerp[d] += ivface;
              //this stuff  checks to see
              //if there are vofs in the perpendicular direction
              //that are regular.   if so, we can do the johansen
              //interpolation between faces thing without fear.
              //If there is no regular face nearby, this gets
              //godawful complicated and we just drop to the 
              //simple stencil.
              if (ivsign[d] != 0)
                {
                  const IntVect& perpIV = ivPerp[d];
                  IntVect othrIV = perpIV - BASISV(a_dir);
                  if(!a_ebisBox.getRegion().contains(perpIV) ||
                     !a_ebisBox.getRegion().contains(othrIV))
                    {
                      facesOk = false;
                    }
                  else if(a_ebisBox.isCovered(perpIV) ||
                          a_ebisBox.isCovered(othrIV))
                    {
                      facesOk = false;
                    }
                  else
                    {
                      facesOk = facesOk && (a_ebisBox.isRegular(perpIV) ||
                                            a_ebisBox.isRegular(othrIV));
                    }
                }
            }
          if (facesOk)
            {
              // Get vofs for faces at ivPerp to check connectivity.
              Tuple<FaceIndex,SpaceDim-1> perpFaces;
              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  if (ivsign[d] != 0)
                    {
                      perpFaces[d] =
                        a_ebisBox.getAllFaces(ivPerp[d],a_dir,Side::Lo)[0];
                      facesOk = facesOk &&
                        a_ebisBox.isConnected(perpFaces[d].getVoF(Side::Lo),vofLo);
                      facesOk = facesOk &&
                        a_ebisBox.isConnected(perpFaces[d].getVoF(Side::Hi),vofHi);
                    }
                }
              // Condition (3) is met - construct the stencil.
              // Compute the center of area for the face.
              RealVect faceCent = a_ebisBox.centroid(a_thisFace);
              Real thisCoef = 1.0;
              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  thisCoef -= Abs(faceCent[kperp[d]]);
                }
              a_thisStencil.add(a_thisFace.getVoF(Side::Lo), -thisCoef);
              a_thisStencil.add(a_thisFace.getVoF(Side::Hi), thisCoef);
              for (int d = 0; d < SpaceDim-1; ++d) 
                {
                  if (ivsign[d] != 0)
                    {
                      Real thisXC = Abs(faceCent[kperp[d]]);

                      for(SideIterator jsit; jsit.ok(); ++jsit)
                        {
                          Real coef = Real(sign(jsit()))*thisXC;
                          const VolIndex& vof = perpFaces[d].getVoF(jsit());
                          a_thisStencil.add(vof, coef);
                        }
                    }
                } //end loop over directions
              MakeSimpleStencil = false;
            } // end if(facesOK) (all conditons met
        } //end if(perp norms are monotonic) 
    } //end isFaceIrregular (if(areaFrac <1))
  if (MakeSimpleStencil)
    {
      // We don't need / have the more elaborate stencil 
      // - instead use simple centered differences.
      for(SideIterator jsit; jsit.ok(); ++jsit)
        {
          Real weight = Real(sign(jsit()));
          VolIndex vof = a_thisFace.getVoF(jsit());
          a_thisStencil.add(vof, weight);
        }
    } // End stencil construction for simple face.
}

void
EBArith::volWeightedSum(Real& a_norm, Real& a_volume,
                        const EBCellFAB& a_src, 
                        const Box& a_region,
                        const EBISBox& a_ebisBox,
                        const int& a_comp,
                        const int& a_pval,
                        EBNormType::NormMode a_mode)
{
  assert(a_pval >= 0);
  assert(a_comp >= 0);
  assert(a_comp < a_src.nComp());
  assert(a_src.getRegion().contains(a_region));
  assert(a_ebisBox.getRegion().contains(a_region));
  int idoirr = 0;
  int idoreg = 0;
  if(a_mode == EBNormType::OverBoth)
    {
      idoirr = 1;
      idoreg = 1;
    }
  else if(a_mode == EBNormType::OverOnlyRegular)
    {
      idoirr = 0;
      idoreg = 1;
    }
  else if(a_mode == EBNormType::OverOnlyIrregular)
    {
      idoirr = 1;
      idoreg = 0;
    }

  a_norm = 0.;
  a_volume = 0.;


  IntVectSet ivsRegion(a_region);
  for(VoFIterator vofit(ivsRegion, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const IntVect& iv = vof.gridIndex();
      if(((idoirr == 1) && a_ebisBox.isIrregular(iv)) ||
         ((idoreg == 1) && a_ebisBox.isRegular(iv)))
        {
          Real val     = Abs(a_src(vof,a_comp));
          if(Abs(val) >s_valMax)
            {
              s_valMax = Abs(val);
              s_vofMax = vof;
            }
          Real volfrac =  a_ebisBox.volFrac(vof);

          a_volume += volfrac;
          if(a_pval == 0) // max a_norm
            {
              a_norm = Max(a_norm, val);
            }
          else if(a_pval == 1) //L1 norm
            {
              a_norm += Abs(volfrac*val);
            }
          else // a_pval > 1
            {
              Real rpval = a_pval;
              Real integrand= pow(val, rpval);
              a_norm += volfrac*integrand;
            }
        }
    }
}
Real 
EBArith::dotProduct(const EBCellFAB& a_dataOne, 
                    const EBCellFAB& a_dataTwo,
                    const Box& a_region,
                    const EBISBox& a_ebisBox,
                    const int& a_comp)
{
  //check everything in sight
  assert(a_comp >= 0);
  assert(a_comp < a_dataOne.nComp());
  assert(a_comp < a_dataTwo.nComp());
  assert(a_dataOne.getRegion().contains(a_region));
  assert(a_dataTwo.getRegion().contains(a_region));
  assert(a_ebisBox.getRegion().contains(a_region));

  Real dotgrid = 0.0;
  IntVectSet ivsregion(a_region);
  for(VoFIterator vofit(ivsregion, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      const Real& valone = a_dataOne(vof,a_comp);
      const Real& valtwo = a_dataTwo(vof,a_comp);
      dotgrid += valone*valtwo;
    }
  return dotgrid;
}
/***************/
/***************/
Real 
EBArith::dotProduct(const BoxLayoutData<EBCellFAB >& a_dataOne, 
                    const BoxLayoutData<EBCellFAB >& a_dataTwo,
                    const BoxLayout& a_dblIn,
                    const EBISLayout& a_ebisl,
                    const int& a_comp)
{
  BaseEBCellFactory<Real> testFactory(a_ebisl);
  Real rhodot = 0.0;
  //calculate the single-processor dot product
  //lots of assertions  have to hold 
  //(same number of vars, grid contained in all the fabs etc)
  DataIterator dit = a_dataOne.dataIterator();
  for(dit.reset(); dit.ok(); ++dit)
    {
      rhodot += dotProduct(a_dataOne[dit()], a_dataTwo[dit()], 
                           a_dblIn.get(dit()), a_ebisl[dit()], a_comp);
    }

  // now for the multi-processor fandango
  //gather all the rhodots onto a vector and add them up
  int baseProc = 0;
  Vector<Real> dotVec;
  gather(dotVec, rhodot, baseProc);

  Real rhodotTot = 0.0;
  if(procID() == baseProc)
    {
      assert(dotVec.size() == numProc());
      for(int ivec = 0; ivec < dotVec.size(); ivec++)
        {
          rhodotTot += dotVec[ivec];
        }
    }
  //broadcast the sum to all processors.
  broadcast(rhodotTot, baseProc);

  //return the total
  return rhodotTot;
}
/***************/
/*****************************/
void
EBArith::
getInterpStencil2D(FaceStencil&      a_sten,
                   const FaceIndex&  a_face, 
                   const IntVectSet& a_coarseFineIVS,
                   const EBISBox&    a_ebisBox,
                   const Box&        a_domainBox) 
{
  assert(SpaceDim == 2);
  a_sten.clear();
  RealVect faceCentroid = a_ebisBox.centroid(a_face);
  int faceDir = a_face.direction();
  int tanDir = 1 - faceDir;
  Side::LoHiSide tanFaceSide;
  if(faceCentroid[tanDir] > 0)
    tanFaceSide = Side::Hi;
  else
    tanFaceSide = Side::Lo;

  FaceIndex otherFace;
  bool uniqueFace = EBArith::getAdjacentFace(otherFace, a_face,
                                             a_ebisBox, a_domainBox,
                                             tanDir, tanFaceSide);

  //drop order of itnerpolation to zero if the other face is not unique
  //or if this face or the otherface is on the coarse fine interface
  bool dropOrder = false;
  dropOrder = !uniqueFace;
  if(!dropOrder)
    {
      for(SideIterator sit; sit.ok(); ++sit)
        {
          if(a_coarseFineIVS.contains(   a_face.gridIndex(sit())) ||
             a_coarseFineIVS.contains(otherFace.gridIndex(sit())))
            {
              dropOrder = true;
            }
        }
    }
 
  if(dropOrder)
    {
      a_sten.add(a_face, 1.0);
    }
  else
    {
      assert(a_face.isDefined());
      assert(otherFace.isDefined());
      Real dist = Abs(faceCentroid[tanDir]);
      Real thisWeight = 1 - dist;
      Real otherWeight =  dist;
      a_sten.add(a_face, thisWeight);
      a_sten.add(otherFace, otherWeight);
    }
}
/*****************************/
void
EBArith::
getInterpStencil3D(FaceStencil&      a_sten,
                   const FaceIndex&  a_face, 
                   const IntVectSet& a_coarseFineIVS,
                   const EBISBox&    a_ebisBox,
                   const Box&        a_domainBox) 
{
  assert(SpaceDim == 3);
  a_sten.clear();

  int faceDir = a_face.direction();
  RealVect faceCentroid = a_ebisBox.centroid(a_face);
  //first check to see if this is a genuinely 3d face
  //centroid > 0 in all directions
  bool really3D = true;
  int izerod = 7;
  Real tol = PolyGeom::getTolerance();
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(idir != faceDir)
        {
          if(Abs(faceCentroid[idir]) < tol)
            {
              really3D= false;
              izerod = idir;
            }
        }
    }
  if(!really3D)
    {
      //use 2D stencil because we have a centroid that is zero
      //in the given direction
      int tanDir = 3 - izerod - faceDir;
      Side::LoHiSide tanFaceSide;
      if(faceCentroid[tanDir] > 0)
        tanFaceSide = Side::Hi;
      else
        tanFaceSide = Side::Lo;

      bool dropOrder = false;
      FaceIndex otherFace;
      bool uniqueFace = EBArith::getAdjacentFace(otherFace, a_face,
                                                 a_ebisBox, a_domainBox,
                                                 tanDir, tanFaceSide);
      dropOrder = !uniqueFace;
 
      if(dropOrder)
        {
          a_sten.add(a_face, 1.0);
        }
      else
        {
          assert(a_face.isDefined());
          assert(otherFace.isDefined());
          Real dist = Abs(faceCentroid[tanDir]);
          Real thisWeight = 1 - dist;
          Real otherWeight =  dist;
          a_sten.add(a_face, thisWeight);
          a_sten.add(otherFace, otherWeight);
        }
    }
  else
    {
      int tanDirs[2];
      Side::LoHiSide tanFaceSides[2];
      {
        int itan = 0;
        for(int idir = 0; idir < 3; idir++)
          {
            if(idir != faceDir)
              {
                tanDirs[itan] = idir;
                if(faceCentroid[tanDirs[itan]] > 0)
                  tanFaceSides[itan] = Side::Hi;
                else
                  tanFaceSides[itan] = Side::Lo;
                itan++;
              }
          }
      }
  
      bool dropOrder = false;
      FaceIndex face01, face10, face11;
      int ixDir = tanDirs[0];
      int iyDir = tanDirs[1];
      Side::LoHiSide xSide = tanFaceSides[0];
      Side::LoHiSide ySide = tanFaceSides[1];
      //figure out whether to drop order.
      //if not,  get the faces involved
      VolIndex vofLo = a_face.getVoF(Side::Lo);
      VolIndex vofHi = a_face.getVoF(Side::Hi);
  
      //first get face10 which is the in 
      //the  ixdir direction from the input face
      if(!dropOrder)
        {
          bool uniqueFace10 
            = EBArith::getAdjacentFace(face10, a_face, a_ebisBox, a_domainBox, ixDir, xSide);
          dropOrder = !uniqueFace10;
          //now get face11 which is the in 
          //the  ixdir,iydir direction from the input face (diagonal)
          if(uniqueFace10)
            {
              bool uniqueFace11
                = EBArith::getAdjacentFace(face11, face10, a_ebisBox, a_domainBox, iyDir, ySide);
              dropOrder = !uniqueFace11;
            }
        }
      //the  ixdir,iydir direction from the input face (diagonal)
      //now get face01 which is the in 
      //the  iydir direction from the input face
      if(!dropOrder)
        {
          //first get face01 which is the in 
          //the  iydir direction from the input face
          bool uniqueFace01 
            = EBArith::getAdjacentFace(face01, a_face, a_ebisBox, a_domainBox, iyDir, ySide);
          dropOrder = !uniqueFace01;
          //now get face11 which is the in 
          //the  ixdir,iydir direction from the input face (diagonal)
          //compute temp face and see if it is the same as
          //the one computed from the other direction.
          //if not , drop order
          FaceIndex face11temp;
          if(uniqueFace01)
            {
              bool uniqueFace11
                = EBArith::getAdjacentFace(face11temp, face01, a_ebisBox, a_domainBox, ixDir, xSide);
              dropOrder = !uniqueFace11;
              if((!dropOrder) && !(face11temp == face11) )
                {
                  dropOrder = true;
                }
            }

          //finally if any of the stencil faces are in the coarse-fine interface, drop order
          if(!dropOrder)
            {
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  if((a_coarseFineIVS.contains(a_face.gridIndex(sit()))) ||
                     (a_coarseFineIVS.contains(face01.gridIndex(sit()))) ||
                     (a_coarseFineIVS.contains(face10.gridIndex(sit()))) ||
                     (a_coarseFineIVS.contains(face11.gridIndex(sit()))))
                    {
                      dropOrder = true;
                    }
                }
            }
        }
      ///////////////////////
      //construct the stencils
      ///////////////////////
      if(dropOrder)
        {
          a_sten.add(a_face, 1.0);
        }
      else
        {
          FaceIndex face00 = a_face;
          Real xbar = Abs(faceCentroid[ixDir]);
          Real ybar = Abs(faceCentroid[iyDir]);
          Real f00coef = 1.0 - xbar - ybar + xbar*ybar;
          Real f10coef = xbar - xbar*ybar;
          Real f01coef = ybar - xbar*ybar;
          Real f11coef = xbar*ybar;

          for(SideIterator sit; sit.ok(); ++sit)
            {
              if(!face00.isBoundary())
                assert(face00.cellIndex(sit()) >= 0);
              if(!face01.isBoundary())
                assert(face01.cellIndex(sit()) >= 0);
              if(!face10.isBoundary())
                assert(face10.cellIndex(sit()) >= 0);
              if(!face11.isBoundary())
                assert(face11.cellIndex(sit()) >= 0);
            }
          assert(face00.isDefined());
          assert(face01.isDefined());
          assert(face10.isDefined());
          assert(face11.isDefined());

          a_sten.add(face00, f00coef);
          a_sten.add(face01, f01coef);
          a_sten.add(face10, f10coef);
          a_sten.add(face11, f11coef);
        }
    }
}
/*****************************/
void
EBArith::
computeCoveredFaces(Vector<VolIndex>&     a_coveredFace, 
                    IntVectSet&           a_coveredSets,
                    IntVectSet&           a_irregIVS,
                    const int&            a_idir, 
                    const Side::LoHiSide& a_sd, 
                    const EBISBox&        a_ebisBox,
                    const Box&            a_region)
{
  //first compute the sets where where the covered faces exist
  //start with all irregular cells and subtract off  cells 
  //whose vofs all have faces in the given direction
  a_irregIVS =  a_ebisBox.getIrregIVS(a_region);
  a_coveredSets = a_irregIVS;
  a_coveredFace.resize(0);
  for(IVSIterator ivsit(a_irregIVS); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = a_ebisBox.getVoFs(iv);
      bool allVoFsHaveFaces = true;
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          Vector<FaceIndex> faces = a_ebisBox.getFaces(vof, a_idir, a_sd);
          if(faces.size() == 0)
            {
              allVoFsHaveFaces = false;
              a_coveredFace.push_back(vof);
            }
        }
      if(allVoFsHaveFaces)
        a_coveredSets -= iv;
    }
}
/*****************************/
FaceStencil
EBArith::
getInterpStencil(const FaceIndex&  a_face, 
                 const IntVectSet& a_cfivs, 
                 const EBISBox&    a_ebisBox,
                 const Box&        a_domainBox) 
{
  FaceStencil sten;
#if CH_SPACEDIM==2
  getInterpStencil2D(sten, a_face, a_cfivs, a_ebisBox, a_domainBox);

#elif CH_SPACEDIM==3
  getInterpStencil3D(sten, a_face, a_cfivs, a_ebisBox, a_domainBox);
#else
  bogus_ch_spacedim_macro();
#endif
  //EBArith::computeInterpStencil(sten, a_face, a_ebisBox, a_domainBox, a_face.direction());
  return sten;
}
/***************/
void
EBArith::volWeightedSum(Real& a_norm, Real& a_volume,
                        const BoxLayoutData<EBCellFAB >& a_src, 
                        const BoxLayout& a_region,
                        const EBISLayout& a_ebisl,
                        const int& a_comp,
                        const int& a_pval,
                        EBNormType::NormMode a_mode)
{
  assert(a_comp >= 0);
  assert(a_pval >= 0);
  assert(a_comp < a_src.nComp());

  Real normlocal = 0.;
  Real vollocal =  0.;
  DataIterator dit = a_src.dataIterator();
  for(dit.reset(); dit.ok(); ++dit)
    {
      Real volgrid = 0.;
      Real normgrid = 0.;
      assert(a_src[dit()].getRegion().contains(a_region.get(dit())));
      const EBCellFAB& srcfab = a_src[dit()];
      volWeightedSum(normgrid, volgrid, 
                     srcfab,
                     a_region.get(dit()), a_ebisl[dit()],
                     a_comp, a_pval, a_mode);

      if(volgrid > 0.)
        {
          vollocal += volgrid;
          if(a_pval == 0)
            {
              normlocal = Max(normlocal, normgrid);
            }
          else
            {
              normlocal += normgrid;
            }
        }
    }
  Real volTot = 0.0;
  Real normTot = 0.0;

  int baseProc = 0;
  Vector<Real> normVec;
  Vector<Real> volVec;
  gather(normVec, normlocal, baseProc);
  gather( volVec,  vollocal, baseProc);


  if(procID() == baseProc)
    {
      assert(volVec.size() == numProc());
      assert(normVec.size() == numProc());
      for(int ivec = 0; ivec < numProc(); ivec++)
        {
          if(volVec[ivec] > 0.)
            {
              volTot += volVec[ivec];
              if(a_pval == 0)
                {
                  normTot = Max(normTot, normVec[ivec]);
                }
              else
                {
                  normTot += normVec[ivec];
                }
            }
        }
    }

  //broadcast the sum to all processors.
  broadcast(normTot, baseProc);
  broadcast(volTot, baseProc);

  //return the totals
  a_norm = normTot;;
  a_volume = volTot;
}
/***************/
/***************/
Real
EBArith::norm(const BoxLayoutData<EBCellFAB >& a_src, 
              const BoxLayout& a_region,
              const EBISLayout& a_ebisl,
              const int& a_comp,
              const int& a_pval, 
              EBNormType::NormMode a_mode)
{
  assert(a_comp >= 0);
  assert(a_pval >= 0);
  assert(a_comp < a_src.nComp());

  Real sum, volume;
  volWeightedSum(sum, volume, a_src,  a_region, 
                 a_ebisl, a_comp, a_pval, a_mode);
  Real normval;
  if(a_pval == 0)
    {
      normval = sum;
    }
  else if(a_pval == 1)
    {
      if(volume > 0.0)
        {
          normval = Abs(sum)/volume;
        }
      else
        {
          normval = 0.0;
        }
    }
  else
    {
      Real denom = a_pval;
      Real exponent = 1.0/denom;
      if(volume > 0.0)
        {
          sum /= volume;
          normval = pow(sum, exponent);
        }
      else
        {
          normval = 0.0;
        }
    }
  return normval;
}

/***************/
/***************/
Real
EBArith::norm(Real& a_volume, const BoxLayoutData<EBCellFAB >& a_src, 
              const BoxLayout& a_region,
              const EBISLayout& a_ebisl,
              const int& a_comp,
              const int& a_pval, 
              EBNormType::NormMode a_mode)
{
  assert(a_comp >= 0);
  assert(a_pval >= 0);
  assert(a_comp < a_src.nComp());

  Real sum;
  volWeightedSum(sum, a_volume, a_src,  a_region, 
                 a_ebisl, a_comp, a_pval, a_mode);
  Real normval;
  if(a_pval == 0)
    {
      normval = sum;
    }
  else if(a_pval == 1)
    {
      if(a_volume > 0.0)
        {
          normval = Abs(sum)/a_volume;
        }
      else
        {
          normval = 0.0;
        }
    }
  else
    {
      Real denom = a_pval;
      Real exponent = 1.0/denom;
      if(a_volume > 0.0)
        {
          sum /= a_volume;
          normval = pow(sum, exponent);
        }
      else
        {
          normval = 0.0;
        }
    }
  return normval;
}
/***************/
bool
EBArith::isVoFHere(VolIndex& a_vof2,
                   const Vector<VolIndex>& a_vofsStencil,
                   const IntVect& a_cell2)
{
  bool found = false;
  for(int isten = 0; isten < a_vofsStencil.size(); isten++)
    {
      if(a_vofsStencil[isten].gridIndex() == a_cell2)
        {
          if(found == true)
            {
              //vof not unique.  return false;
              return false;
            }
          else
            {
              found = true;
              a_vof2 = a_vofsStencil[isten];
            }
        }
    }

  return found;
}
/***************/
bool
EBArith::monotonePathVoFToCellVoF(VolIndex& a_vof2,
                                  const VolIndex& a_vof1,
                                  const IntVect& a_cell2,
                                  const EBISBox& a_ebisBox)
{
  IntVect diffVect = a_cell2 - a_vof1.gridIndex();
  int imaxdiff = 0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(Abs(diffVect[idir]) > imaxdiff)
        imaxdiff = Abs(diffVect[idir]);
    }
  
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;
  Vector<VolIndex> vofsStencil;
  getAllVoFsInMonotonePath(vofsStencil, timesMoved, 
                           pathSign, a_vof1, a_ebisBox,
                           imaxdiff);

  

  bool found = isVoFHere(a_vof2,  vofsStencil, a_cell2);

  return found;
}
/*******/
void
EBArith::
getAllVoFsInMonotonePath(Vector<VolIndex>& a_vofList,
                         const VolIndex&   a_vof,
                         const EBISBox&    a_ebisBox,
                         const int&        a_redistRad)
{
  Vector<VolIndex> vofsStencil;
  IntVect timesMoved = IntVect::Zero;
  IntVect pathSign   = IntVect::Zero;
  EBArith::getAllVoFsInMonotonePath(vofsStencil, timesMoved, 
                                    pathSign, a_vof, a_ebisBox,
                                    a_redistRad);
  a_vofList = vofsStencil;
}
/*******/
void
EBArith::
getAllVoFsInMonotonePath(Vector<VolIndex>& a_vofList,
                         const IntVect&    a_timesMoved,
                         const IntVect&    a_pathSign,
                         const VolIndex&   a_vof,
                         const EBISBox&    a_ebisBox,
                         const int&        a_redistRad)
{
  const Box& domain = a_ebisBox.getDomain();
  if(domain.contains(a_vof.gridIndex()))
    {
      //check to see if we have already added it.
      //if not, add it
      bool found = false;
      for(int ivof = 0; ivof < a_vofList.size(); ivof++)
        {
          if(a_vofList[ivof] == a_vof)
            {
              found = true;
            }
        }
      if(!found)
        {
          a_vofList.push_back(a_vof);
        }
      for (int idir = 0; idir < SpaceDim; idir++)
        {
          //only move redist radius times in a direction
          if (a_timesMoved[idir] < a_redistRad)
            {
              IntVect newTimesMoved = a_timesMoved;
              newTimesMoved[idir]++;
              //pathSign preserves monotonicity
              //we set pathsign to -1 once you start in the low
              //direction, 1 in the high direction.  It starts at zero
          
              //if we started in the low direction or have not started
              //add vofs on low side
              if((a_pathSign[idir] == -1) || (a_pathSign[idir] == 0))
                {
                  IntVect newSign = a_pathSign;
                  newSign[idir] = -1;
                  Vector<FaceIndex> facesLo = 
                    a_ebisBox.getFaces(a_vof, idir, Side::Lo);
                  for(int iface = 0; iface < facesLo.size(); iface++)
                    {
                      VolIndex newVoF = facesLo[iface].getVoF(Side::Lo);
                      getAllVoFsInMonotonePath(a_vofList, newTimesMoved, newSign,
                                               newVoF, a_ebisBox, a_redistRad);
                    }
                }
              //if we started in the high direction or have not started
              //add vofs on high side
              if((a_pathSign[idir] == 1) || (a_pathSign[idir] == 0))
                {
                  IntVect newSign = a_pathSign;
                  newSign[idir] = 1;
                  Vector<FaceIndex> facesHi = 
                    a_ebisBox.getFaces(a_vof, idir, Side::Hi);
                  for(int iface = 0; iface < facesHi.size(); iface++)
                    {
                      VolIndex newVoF = facesHi[iface].getVoF(Side::Hi);
                      getAllVoFsInMonotonePath(a_vofList, newTimesMoved, newSign,
                                               newVoF, a_ebisBox, a_redistRad);
                    }
                }
            } //end if(we are less than redist radius away)
        }//end loop over directions
    }
}
