/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _EBCellFAB_H_
#define _EBCellFAB_H_

#include <cmath>
#include <cstdlib>
#include <cassert>
#include "BaseEBCellFAB.H"
#include "FArrayBox.H"
#include "EBISBox.H"
#include "Box.H"
/// 
/**
   EBCellFAB is a BaseEBCellFAB<Real> 
   which includes arithmetic functions.
 */
class EBCellFAB: public BaseEBCellFAB<Real>
{
public:
  ///
  EBCellFAB();


  ///
  /**
    Box going into this needs to be cell-centered.
    Data will exist over the surrounding nodes of the box.
   */
  EBCellFAB(const EBISBox& a_ebisBox,
            const Box& a_region, int a_nVar);


  ///
  void setInvalidData(const Real& a_val,
                      const int& a_comp);

  ///
  virtual ~EBCellFAB();

  ///
  const FArrayBox& getFArrayBox() const;

  ///
  /**
     Does a dynamic cast  of the BaseFab<Real>
     to a  FArrayBox for arithmetic purposes.
   */
  FArrayBox& getFArrayBox();

  ///
  /**
     Negate the values in all locations
   */
  EBCellFAB& negate(void);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator+=(const EBCellFAB& a_ebfab);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator-=(const EBCellFAB& a_ebfab);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator*=(const EBCellFAB& a_ebfab);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator/=(const EBCellFAB& a_ebfab);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator+=(const Real& a_scalar);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator-=(const Real& a_scalar);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator*=(const Real& a_scalar);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& mult(Real a_scalar);

  ///
  /**
     Both fabs need the same ebisBox and region.
   */
  EBCellFAB& operator/=(const Real& a_scalar);

  ///
  /**
     Current FAB += a_src FAB * a_scalar.  Both fabs need the same ebisBox
     and region.
   */
  EBCellFAB& plus(const EBCellFAB& a_src,
                  Real             a_scalar);

  /// (Not implemented) Returns the Lp-norm of this EBCellFAB
  /**
     (Not implemented) Returns the Lp-norm of this EBCellFAB using components
     (a_comp : a_comp + a_numcomp - 1).  a_power < 0 -> ERROR.
     a_power = 0  -> infinity norm (max norm).
     a_power = 1  -> L1-norm
     a_power > 1  -> Lp-norm
  */
  virtual Real norm(int a_power,
                    int a_comp,
                    int a_numComp) const;

  /// (Not implemented) Returns the Lp-norm of this EBCellFAB within a region 
  /**
     (Not implemented) Returns the Lp-norm of this EBCellFAB using components
     (a_comp : a_comp + a_numcomp - 1) and within the a_subbox.  a_power < 0
     -> ERROR.
     a_power = 0 -> infinity norm (max norm).
     a_power = 1 -> L1-norm
     a_power > 1 -> Lp-norm 
  */
  virtual Real norm(const Box& a_subbox,
                    int        a_power,
                    int        a_comp,
                    int        a_numComp) const;

  /// (Not implemented) Returns a sum of powers of a subset of this EBCellFAB
  /**  
     (Not implemented) Returns a sum of powers of a subset of this EBCellFAB,
     specifically components a_comp to a_comp+a_numcomp-1 within a_subbox.
     a_power >= 2 only.

  */
  virtual Real sumPow(const Box& a_subbox,
                      int        a_power,
                      int        a_comp,
                      int        a_numComp) const;


  /// (Not implemented) Return the dot product of this EBCellFAB with another
  /**
     (Not implemented) Return the dot product of this EBCellFAB and "ebfab2"
     over their overlap region and all components.
  */
  Real dotProduct(const EBCellFAB& ebfab2) const;

  //needed for leveldata to compile
  //but very invalid
  EBCellFAB(const Box& a_region, int a_nComp)
  {
    MayDay::Error("ebcelfab 1 invalid operator");
  }
private:
  //disallowed for all the usual reasons
  EBCellFAB(const EBCellFAB& ebcin)
  {
    MayDay::Error("ebcellfab 2 invalid operator");
  }
  void operator=(const EBCellFAB& fabin)
  {
    MayDay::Error("ebcellfab 3 invalid operator");
  }

};
#endif

