/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#include "EBCellFAB.H"
#include "EBArithF_F.H"
#include "FArrayBox.H"

void
EBCellFAB::setInvalidData(const Real& a_val,
                          const int& a_comp)
{
  assert(a_comp >= 0);
  assert(a_comp < m_nComp);

  if(m_ebisBox.isAllRegular())
    {
      return;
    }
  else if(m_ebisBox.isAllCovered())
    {
      m_regFAB.setVal(a_val, a_comp);
    }
  else
    {
      for(BoxIterator bit(m_region); bit.ok(); ++bit)
        {
          const IntVect& iv = bit();
          if(m_ebisBox.isCovered(iv))
            {
              m_regFAB(iv, a_comp) = a_val;
            }
        }
      //also set the multivalued cells
      for(IVSIterator ivsit(getMultiCells()); 
          ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          m_regFAB(iv, a_comp) = a_val;
        }
      //also set the multivalued cells
    }
}
/**********************/
/**********************/
EBCellFAB::EBCellFAB():BaseEBCellFAB<Real>()
{
}

/**********************/
/**********************/
EBCellFAB::EBCellFAB(const EBISBox& a_ebisBox, 
                     const Box& a_region, int a_nComp)
  :BaseEBCellFAB<Real>(a_ebisBox, a_region, a_nComp)
{
}

/**********************/
/**********************/
EBCellFAB::~EBCellFAB()
{
}

/**********************/
/**********************/
const FArrayBox& 
EBCellFAB::getFArrayBox() const
{
  assert(isDefined());
  return (const FArrayBox&)m_regFAB;
}

/**********************/
/**********************/
FArrayBox& 
EBCellFAB::getFArrayBox() 
{
  assert(isDefined());
  return (FArrayBox&)m_regFAB;
}

/**********************/
/**********************/
EBCellFAB&
EBCellFAB::negate(void)
{
  assert(isDefined());
  (*this) *= -1.0;
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator+=(const EBCellFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  // Dan G. feels strongly that the assert below should NOT be commented out
  // Brian feels that a weaker version of the assert (if possible) is needed
  // Terry is just trying to get his code to work
  // assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_ADDTWOFAB(CHF_FRA(m_regFAB), 
                     CHF_CONST_FRA(a_src.m_regFAB), 
                     CHF_BOX(locRegion));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for (int icomp = 0; icomp < m_nComp; ++icomp)
                {
                  m_irrFAB(vof, icomp) += a_src.m_irrFAB(vof, icomp);
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator-=(const EBCellFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  // Dan G. feels strongly that the assert below should NOT be commented out
  // Brian feels that a weaker version of the assert (if possible) is needed
  // Terry is just trying to get his code to work
  // assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_SUBTRACTTWOFAB(CHF_FRA(m_regFAB), 
                          CHF_CONST_FRA(a_src.m_regFAB), 
                          CHF_BOX(locRegion));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for (int icomp = 0; icomp < m_nComp; ++icomp)
                {
                  m_irrFAB(vof, icomp) -= a_src.m_irrFAB(vof, icomp);
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator*=(const EBCellFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  // Dan G. feels strongly that the assert below should NOT be commented out
  // Brian feels that a weaker version of the assert (if possible) is needed
  // Terry is just trying to get his code to work
  // assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_MULTIPLYTWOFAB(CHF_FRA(m_regFAB), 
                          CHF_CONST_FRA(a_src.m_regFAB), 
                          CHF_BOX(locRegion));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for (int icomp = 0; icomp < m_nComp; ++icomp)
                {
                  m_irrFAB(vof, icomp) *= a_src.m_irrFAB(vof, icomp);
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator/=(const EBCellFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  // Dan G. feels strongly that the assert below should NOT be commented out
  // Brian feels that a weaker version of the assert (if possible) is needed
  // Terry is just trying to get his code to work
  // assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_DIVIDETWOFAB(CHF_FRA(m_regFAB), 
                        CHF_CONST_FRA(a_src.m_regFAB), 
                        CHF_BOX(locRegion));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for (int icomp = 0; icomp < m_nComp; ++icomp)
                {
                  m_irrFAB(vof, icomp) /= a_src.m_irrFAB(vof, icomp);
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator+=(const Real& a_src)
{
  assert(isDefined());
  FORT_ADDFABR(CHF_FRA(m_regFAB), 
               CHF_CONST_REAL(a_src), 
               CHF_BOX(m_region));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for (int icomp = 0; icomp < m_nComp; ++icomp)
            {
              m_irrFAB(vof, icomp) += a_src;
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator-=(const Real& a_src)
{
  assert(isDefined());
  FORT_SUBTRACTFABR(CHF_FRA(m_regFAB), 
                    CHF_CONST_REAL(a_src), 
                    CHF_BOX(m_region));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for (int icomp = 0; icomp < m_nComp; ++icomp)
            {
              m_irrFAB(vof, icomp) -= a_src;
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator*=(const Real& a_src)
{
  assert(isDefined());
  FORT_MULTIPLYFABR(CHF_FRA(m_regFAB), 
                    CHF_CONST_REAL(a_src), 
                    CHF_BOX(m_region));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for (int icomp = 0; icomp < m_nComp; ++icomp)
            {
              m_irrFAB(vof, icomp) *= a_src;
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::mult(Real a_src)
{
  *this *= a_src;

  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::operator/=(const Real& a_src)
{
  assert(isDefined());
  FORT_DIVIDEFABR(CHF_FRA(m_regFAB), 
                  CHF_CONST_REAL(a_src), 
                  CHF_BOX(m_region));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for (int icomp = 0; icomp < m_nComp; ++icomp)
            {
              m_irrFAB(vof, icomp) /= a_src;
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBCellFAB& 
EBCellFAB::plus(const EBCellFAB& a_src, Real a_scale)
{
  assert(isDefined());
  assert(a_src.isDefined());
  // Dan G. feels strongly that the assert below should NOT be commented out
  // Brian feels that a weaker version of the assert (if possible) is needed
  // Terry is just trying to get his code to work
  // assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_SCALEADDTWOFAB(CHF_FRA(m_regFAB), 
                          CHF_CONST_FRA(a_src.m_regFAB), 
                          CHF_CONST_REAL(a_scale), 
                          CHF_BOX(locRegion));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for (int icomp = 0; icomp < m_nComp; ++icomp)
                {
                  m_irrFAB(vof, icomp) += a_src.m_irrFAB(vof, icomp) * a_scale;
                }
            }
        }
    }
  return *this;
}

// (Not implemented) Returns the Lp-norm of this EBCellFAB
Real
EBCellFAB::norm(int a_power,
                int a_comp,
                int a_numComp) const
{
  MayDay::Error("norm method 1 not implemented");
  
  return -1.0;
}

// (Not implemented) Returns the Lp-norm of this EBCellFAB within a region 
Real
EBCellFAB::norm(const Box& a_subbox,
                int        a_power,
                int        a_comp,
                int        a_numComp) const
{
  MayDay::Error("norm method 2 not implemented");
  
  return -1.0;
}

// (Not implemented) Returns a sum of powers of a subset of this EBCellFAB
Real
EBCellFAB::sumPow(const Box& a_subbox,
                  int        a_power,
                  int        a_comp,
                  int        a_numComp) const
{
  MayDay::Error("sumPow method 2 not implemented");
  
  return -1.0;
}

// (Not implemented) Return the dot product of this EBCellFAB with another
Real
EBCellFAB::dotProduct(const EBCellFAB& ebfab2) const
{
  MayDay::Error("dotProduct method 2 not implemented");
  
  return -1.0;
}
