/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _EBISData_H_
#define _EBISData_H_

#include "REAL.H"
#include "EBGraph.H"
#include "IrregNode.H"
#include "BaseIVFAB.H"
#include "BaseIFFAB.H"
#include "RealVect.H"

///
/**
 */
class VolData
{
public:
  VolData();
  VolData(const VolData& vol);
  ~VolData();
  void append(VolData* next);
  Real     m_volFrac;
  Real     m_bndryArea;
  RealVect m_normal;
  RealVect m_volCentroid;
  RealVect m_bndryCentroid;
  int      m_bndryPhase;
  int      m_volIndex;
  VolData* m_nextPhase;
};

std::ostream& operator<< (std::ostream&  a_os,
                          const VolData& a_iv);
///
/**
 */
class FaceData
{
public:
  Real     m_areaFrac;
  RealVect m_faceCentroid;     
};

///
/**
   This class contains all the geometric information
   for an ebisbox.
 */
class EBDataImplem
{

public:

  ///
  EBDataImplem();

  ///
  ~EBDataImplem();
  
  ///
  /**
     Noop.
  */
  void define(const Box& box, int comps);


  ///
  /**
     Noop
  */
  EBDataImplem(const Box& a_box, int a_comps);

  ///
  /**
     Copy the information from a_source to the over the intersection
     of the box a_region, the box of the current EBDataImplem and 
     the box of a_source.  The Interval arguments are ignored.
     This function is required by LevelData.
  */
  void copy(const Box&      a_regionFrom, 
            const Interval& a_Cd, 
            const Box&      a_regionto, 
            const EBDataImplem&   a_source, 
            const Interval& a_Cs);

  /// each data holder is defined over the irregular cells of the graph
  void 
  defineVoFData(const EBGraph& a_graph, const Box& a_region);

  /// each data holder is defined over the irregular cells of the graph
  void 
  defineFaceData(const EBGraph& a_graph, const Box& a_region);


  ///define the whole thing
  void
  define(const EBGraph&           a_graph,
         const Vector<IrregNode>& a_irregData,
         const Box&               a_validBox);

  ///
  void 
  coarsenVoFs(const EBDataImplem&   a_fineEBDataImplem,
              const EBGraph&        a_fineGraph,
              const EBGraph&        a_coarGraph,
               const Box&           a_validRegion);


  ///
  void 
  coarsenFaces(const EBDataImplem& a_fineEBDataImplem, 
               const EBGraph&      a_fineGraph,
               const EBGraph&      a_coarGraph,
               const Box&          a_validRegion);
  ///
  const Real& volFrac(const VolIndex& a_vof) const;
  
  ///
  const Real& areaFrac(const FaceIndex& a_face1) const;
  
  ///
  const RealVect& centroid(const FaceIndex& facein) const;
  
  ///
  const RealVect& centroid(const VolIndex& a_vof) const;
   
  ///
  const RealVect& bndryCentroid(const VolIndex& a_vof, int face=0) const;
  
  ///
  const Real& bndryArea(const VolIndex& a_vof, int face=0) const;
  
  ///
  const RealVect& normal(const VolIndex& a_vof, int face=0) const;

  //  New set of boundary inquiry functions to support multi-fluid applications
  /// used by multi-fluid applications

  /// used by multi-fluid code
  int facePhase(const VolIndex& a_vof, int face) const ;

  /// used by multi-fluid code
  int faceIndex(const VolIndex& a_vof, int face) const ;

  /// used by multi-fluid code
  void setFacePhase(const VolIndex& a_vof, int face, int phase);
  ///
  ///
  int numFacePhase(const VolIndex& a_vof) const ;

  /**
     This stuff required by LevelData in parallel:
  */
  int size(const Box& R, const Interval& comps) const;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  static int preAllocatable()
  {
    return 2; // dyanmic allocatable.  
  }

  BaseIVFAB<VolData>& getVolData()
  {
    return m_volData;
  }
  const BaseIVFAB<VolData>& getVolData() const
  {
    return m_volData;
  }

  ///
  void addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                            const EBGraph&    a_newGhostGraph,
                            const EBGraph&    a_oldGhostGraph);

  ///
  static void setVerbose(bool a_verbose);
private:

  static   bool s_verbose;
  ///
  BaseIVFAB<VolData>   m_volData;

  ///
  BaseIFFAB<FaceData>  m_faceData[SpaceDim];

  ///
  bool m_isFaceDataDefined;

  ///
  bool m_isVoFDataDefined;

  void operator=(const EBDataImplem& ebiin)
  {;}

  EBDataImplem(const EBDataImplem& ebiin)
  {;}

  void
  coarsenFaceCentroid(RealVect&                a_centroidCoar, 
                      const Vector<RealVect>&  a_centroidsFine,
                      const Vector<Real>&      a_areaFracFine,
                      const Vector<FaceIndex>& a_facesFine, 
                      const FaceIndex&         a_faceCoar);
  void 
  coarsenAreaFrac(Real& a_areaFracCoar,
                  const Vector<Real>& a_areaFracFine);

  void
  coarsenVolFracAndCentroid(Real&                   a_volFracCoar,
                            RealVect&               a_volCentroidCoar,
                            const Vector<Real>&     a_volFracFine,
                            const Vector<RealVect>& a_volCentroidFine,
                            const Vector<VolIndex>& a_fineVoFs,
                            const VolIndex&         a_coarVoF);


  void
  coarsenBoundaryAreaAndNormal(Real&                    a_bndryAreaCoar,
                               RealVect&                a_normalCoar,
                               const Vector<Real>&      a_bndryAreaFine,
                               const Vector<RealVect>&  a_normalFine);

  RealVect 
  fineToCoarseTransform(const RealVect& a_finePoint,
                        const IntVect&  a_coarCell,
                        const IntVect&  a_fineCell);

  void
  coarsenBndryCentroid(RealVect&               a_bndryCentroidCoar,
                       const Vector<RealVect>& a_bndryCentroidFine,
                       const Vector<Real>&     a_bndryAreaFine,
                       const Vector<VolIndex>& a_fineVoFs,
                       const VolIndex&         a_coarVoF);

};

///
/**
   Ref-counted version of EBDataImplem.
 */
class EBData
{
public:

  ///
  EBData();

  ///
  ~EBData();
  
  ///
  /**
     Noop.
  */
  void define(const Box& box, int comps);


  ///
  /**
     Noop
  */
  EBData(const Box& a_box, int a_comps);

  ///
  /**
     Copy the information from a_source to the over the intersection
     of the box a_region, the box of the current EBData and 
     the box of a_source.  The Interval arguments are ignored.
     This function is required by LevelData.
  */
  void copy(const Box&      a_regionFrom, 
            const Interval& a_Cd, 
            const Box&      a_regionto, 
            const EBData&   a_source, 
            const Interval& a_Cs);

  /// each data holder is defined over the irregular cells of the graph
  void 
  defineVoFData(const EBGraph& a_graph, const Box& a_region);

  /// each data holder is defined over the irregular cells of the graph
  void 
  defineFaceData(const EBGraph& a_graph, const Box& a_region);


  ///define the whole thing
  void
  define(const EBGraph&           a_graph,
         const Vector<IrregNode>& a_irregData,
         const Box&               a_validBox);

  ///
  void 
  coarsenVoFs(const EBData&   a_fineEBData,
              const EBGraph&  a_fineGraph,
              const EBGraph&  a_coarGraph,
              const Box&      a_validRegion);


  void 
  coarsenFaces(const EBData&  a_fineEBData, 
               const EBGraph& a_fineGraph,
               const EBGraph& a_coarGraph,
               const Box&     a_validRegion);
  ///
  Real volFrac(const VolIndex& a_vof) const;

  ///
  Real areaFrac(const FaceIndex& a_face1) const;

  ///
  RealVect centroid(const FaceIndex& facein) const;

  ///
  RealVect centroid(const VolIndex& a_vof) const;

  ///
  RealVect bndryCentroid(const VolIndex& a_vof, int face=0) const;
  
  ///
  Real bndryArea(const VolIndex& a_vof, int face=0) const;

  ///
  RealVect normal(const VolIndex& a_vof, int face=0) const;

  /// used by multi-fluid code
  int facePhase(const VolIndex& a_vof, int face) const ;

  /// used by multi-fluid code
  int faceIndex(const VolIndex& a_vof, int face) const ;

  /// used by multi-fluid code
  void setFacePhase(const VolIndex& a_vof, int face, int phase);

  ///
  int numFacePhase(const VolIndex& a_vof) const ;

  ///
  EBData(const EBData& a_ebiin);

  ///
  /**
     This is a pointer comparison.
  */
  bool operator==(const EBData& a_ebiin);

  ///
  EBData& operator=(const EBData& a_ebiin);

  ///
  /**
     This stuff required by LevelData in parallel:
  */
  int size(const Box& R, const Interval& comps) const;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  void addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                            const EBGraph&    a_newGhostGraph,
                            const EBGraph&    a_oldGhostGraph);

  ///
  void 
  computeNormalsAndBoundaryAreas(const EBGraph& a_graph, 
                                 const Box&     a_validRegion);

  ///
  static int preAllocatable()
  {
    return 2; // dyanmic allocatable.  
  }
private:

  ///
  RefCountedPtr<EBDataImplem> m_implem;

};



#endif
