/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL
#include "EBData.H"
#include "VoFIterator.H"
#include "FaceIterator.H"
#include "EBISBox.H"
#include "PolyGeom.H"

VolData::VolData()
  : m_volFrac(1.0),m_bndryArea(0.0), m_normal(RealVect::Zero),
    m_volCentroid(RealVect::Zero),m_bndryCentroid(RealVect::Zero),
    m_bndryPhase(-1), m_volIndex(0), m_nextPhase(NULL)
{}

VolData::VolData(const VolData& vol)
  :m_volFrac(vol.m_volFrac),
   m_bndryArea(vol.m_bndryArea),
   m_normal(vol.m_normal),
   m_volCentroid(vol.m_volCentroid),
   m_bndryCentroid(vol.m_bndryCentroid),
   m_bndryPhase(vol.m_bndryPhase),
   m_volIndex(vol.m_volIndex),
   m_nextPhase(NULL)
{
  if(vol.m_nextPhase != NULL)
    m_nextPhase = new VolData(*(vol.m_nextPhase));
}

void VolData::append(VolData* v)
{
  VolData* next = this;
  while(next->m_nextPhase != NULL)
    next = next->m_nextPhase;
  next->m_nextPhase = v;
}

VolData::~VolData()
{
  if(m_nextPhase != NULL) delete m_nextPhase;
}

bool EBDataImplem::s_verbose = false;
/************************/
void EBDataImplem::setVerbose(bool a_verbose)
{
  s_verbose = a_verbose;
}
/************************/
void EBData::
addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                     const EBGraph&    a_newGraph,
                     const EBGraph&    a_oldGraph)
{
  m_implem->addFullIrregularVoFs(a_vofsToChange, a_newGraph, a_oldGraph);
}
/************************/
void EBDataImplem::
addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                     const EBGraph&    a_newGraph,
                     const EBGraph&    a_oldGraph)
{
  if(!a_vofsToChange.isEmpty())
    {
      //calculate set by adding in new intvects
      const IntVectSet& ivsOld = m_volData.getIVS();
      IntVectSet ivsNew = ivsOld | a_vofsToChange;

      //for copying purposes
      Box minBox = ivsOld.minBox();
      Interval interv(0,0);
      
      //save old data into temporarys 
      BaseIVFAB<VolData>  oldVolData(ivsOld, a_oldGraph, 1);
      oldVolData.copy(minBox, interv, minBox, m_volData, interv);
      BaseIFFAB<FaceData> oldFaceData[SpaceDim];
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          oldFaceData[idir].define(ivsOld, a_oldGraph, idir, 1);
          oldFaceData[idir].copy(minBox, interv, minBox, m_faceData[idir], interv);
        }

      //redefine member data with new IVS and copy old data back
      //into it.
      m_volData.define(ivsNew, a_newGraph, 1);
      m_volData.copy(minBox, interv, minBox, oldVolData, interv);
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          m_faceData[idir].define(ivsNew, a_newGraph, idir, 1);
          m_faceData[idir].copy(minBox, interv, minBox, oldFaceData[idir], interv);
        }

      //now put correct data into new vofs
      VolData  fullVol;
 
      /*  now done in VolData null constructor
      fullVol.m_volFrac       = 1.0;
      fullVol.m_bndryArea     = 0.0;
      fullVol.m_normal        = BASISREALV(0);
      fullVol.m_volCentroid   = RealVect::Zero;
      fullVol.m_bndryCentroid = RealVect::Zero;
      */

      FaceData fullFace;
      fullFace.m_areaFrac     = 1.0;
      fullFace.m_faceCentroid = RealVect::Zero;
      for(VoFIterator vofit(a_vofsToChange, a_newGraph); vofit.ok(); ++vofit)
        {
          m_volData(vofit(), 0) = fullVol;
        }
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          FaceIterator faceit(a_vofsToChange, a_newGraph, idir, FaceStop::SurroundingWithBoundary);
          for(faceit.reset(); faceit.ok(); ++faceit)
            {
              m_faceData[idir](faceit(), 0) = fullFace;
            }
        }
    }
}
/************************/
/* sheesh, compiler writers can be lazy */
/************************/
ostream&
operator<< (ostream&       os,
            const VolData& iv)
{
  os << "(" 
     << iv.m_volFrac       << ", "
     << iv.m_bndryArea     << ", "
     << iv.m_normal        << ", "
     << iv.m_volCentroid   << ", "
     << iv.m_bndryCentroid << ", "
     << ")" << endl;;

  if (os.fail())
    MayDay::Error("operator<<(ostream&,VolData&) failed");
  return os;
}
/************************/
EBDataImplem::
EBDataImplem()
{
  m_isVoFDataDefined = false;
  m_isFaceDataDefined = false;
}
/************************/
EBDataImplem::
~EBDataImplem()
{
}
/************************/
void EBDataImplem::
define(const Box& box, int comps)
{
}
/************************/
EBDataImplem::
EBDataImplem(const Box& a_box, int a_comps)
{
}
/************************/
void EBDataImplem::
copy(const Box&      a_regionFrom, 
     const Interval& a_Cd, 
     const Box&      a_regionTo, 
     const EBDataImplem&   a_source, 
     const Interval& a_Cs)
{
  assert(m_isVoFDataDefined);
  assert(m_isFaceDataDefined);
  assert(a_source.m_isVoFDataDefined);
  assert(a_source.m_isFaceDataDefined);
  Interval ivsca(0,0);
  m_volData.copy(a_regionFrom, ivsca, a_regionTo, a_source.m_volData ,ivsca);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      m_faceData[idir].copy(a_regionFrom, ivsca, a_regionTo, a_source.m_faceData[idir], ivsca);
    }

}
/************************/
void EBDataImplem::
defineVoFData(const EBGraph& a_graph,
              const Box& a_validBox)
{
  m_isVoFDataDefined = true;

  IntVectSet ivsIrreg = a_graph.getIrregCells(a_validBox);
  m_volData.define(ivsIrreg, a_graph, 1);
}
/************************/
void EBDataImplem::
defineFaceData(const EBGraph& a_graph,
               const Box& a_validBox)
{
  m_isFaceDataDefined = true;
  IntVectSet ivsIrreg = a_graph.getIrregCells(a_validBox);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      m_faceData[idir].define(ivsIrreg, a_graph, idir, 1);
    }
}
/************************/
void
EBDataImplem::define(const EBGraph&           a_graph,
                     const Vector<IrregNode>& a_irregGraph,
                     const Box&               a_validBox)

{
  defineVoFData( a_graph, a_validBox);
  defineFaceData(a_graph, a_validBox);

  if(a_graph.hasIrregular())
    {
      for(int inode = 0; inode < a_irregGraph.size(); inode++)
        {
          const IrregNode& node = a_irregGraph[inode];
          const IntVect& iv = node.m_cell;
          const int&  cellInd = node.m_cellIndex;
          VolIndex vof(iv, cellInd);
          
          VolData& vol = m_volData(vof, 0);
          vol.m_volFrac       = node.m_volFrac;
          vol.m_volCentroid   = node.m_volCentroid;
          vol.m_bndryCentroid = node.m_bndryCentroid;
          vol.m_bndryPhase    = -1;  // EB
          vol.m_nextPhase     = NULL; // only one bndry per vol for now on creation.
 
          for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
            {
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Vector<FaceIndex> faces = a_graph.getFaces(vof, faceDir, sit());
                  int nodeind = node.index(faceDir, sit());
                  Vector<Real> areaFracs         = node.m_areaFrac[nodeind];
                  Vector<RealVect> faceCentroids = node.m_faceCentroid[nodeind];
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const Real&     areaFracNode     = areaFracs[iface];
                      const RealVect& faceCentroidNode = faceCentroids[iface];
                      const FaceIndex& face = faces[iface];

                      m_faceData[faceDir](face,0).m_areaFrac     = areaFracNode;
                      m_faceData[faceDir](face,0).m_faceCentroid = faceCentroidNode;
                    }
                }
            }
        }
    }
}
/*******************************/
const Real& EBDataImplem::volFrac(const VolIndex& a_vof) const
{
  return m_volData(a_vof, 0).m_volFrac;
}
/*******************************/
const Real& EBDataImplem::bndryArea(const VolIndex& a_vof, int face) const
{
  return m_volData(a_vof, 0).m_bndryArea;
}
/*******************************/
const RealVect& EBDataImplem::normal(const VolIndex& a_vof, int face) const
{
  return m_volData(a_vof, 0).m_normal;
}
/*******************************/
const RealVect& EBDataImplem::centroid(const VolIndex& a_vof) const
{
  return m_volData(a_vof, 0).m_volCentroid;
}
/*******************************/
const RealVect& EBDataImplem::bndryCentroid(const VolIndex& a_vof, int face) const
{
  return m_volData(a_vof, 0).m_bndryCentroid;
}
  /// used by multi-fluid code
int EBDataImplem::facePhase(const VolIndex& a_vof, int face) const 
{
  const VolData* v = &(m_volData(a_vof, 0));
  while(face > 0)
    {
      v=v->m_nextPhase;
      face++;
    }
  return v->m_bndryPhase;
}

  /// used by multi-fluid code
int EBDataImplem::faceIndex(const VolIndex& a_vof, int face) const 
{
  const VolData* v = &(m_volData(a_vof, 0));
  while(face > 0)
    {
      v=v->m_nextPhase;
      face++;
    }
  return v->m_volIndex;
}

  /// used by multi-fluid code
void EBDataImplem::setFacePhase(const VolIndex& a_vof, int face, int phase)
{
  VolData* v = &(m_volData(a_vof, 0));
  while(face > 0)
    {
      v=v->m_nextPhase;
      face++;
    }
  v->m_bndryPhase = phase;
}

///
int EBDataImplem::numFacePhase(const VolIndex& a_vof) const
{
  int rtn=1;
  const VolData* v = &(m_volData(a_vof, 0));
  while(v->m_nextPhase!= NULL)
    {
      v=v->m_nextPhase;
      rtn++;
    }
  return rtn;
}

/*******************************/
const RealVect& EBDataImplem::centroid(const FaceIndex& a_face) const
{
  int faceDir = a_face.direction();
  return m_faceData[faceDir](a_face, 0).m_faceCentroid;
}
/*******************************/
const Real& EBDataImplem::areaFrac(const FaceIndex& a_face) const 
{
  int faceDir = a_face.direction();
  return m_faceData[faceDir](a_face, 0).m_areaFrac;
}
/*******************************/
void EBData::
computeNormalsAndBoundaryAreas(const EBGraph& a_graph, 
                               const Box&     a_validRegion)
{

  const RealVect vectDx = PolyGeom::getVectDx();
  //We need maxDx to properly scale bndryArea
  Real maxDx = 0.;
  for (int idir = 0; idir <SpaceDim; ++idir)
    {
      assert( vectDx[idir] > 0);
      if (vectDx[idir] >maxDx)
        {
          maxDx = vectDx[idir];
        }
    }

  EBISBox ebisBox;
  ebisBox.define(a_graph, *this);
  BaseIVFAB<VolData>& volData = m_implem->getVolData();
  if(a_graph.hasIrregular())
    {
      IntVectSet ivsIrreg = a_graph.getIrregCells(a_validRegion);
      for(VoFIterator vofit(ivsIrreg, a_graph); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          //Real bndryArea = PolyGeom::bndryArea(vof, ebisBox);
          //RealVect normal =  PolyGeom::normal(vof, ebisBox, bndryArea);

          //Compute bndryArea (ignoring anisotropy if it exists)
          Real bndryArea;
          Real irregArea=0.0;
          for (int idir=0; idir < SpaceDim; idir++) 
            {
              Real hiArea = ebisBox.sumArea(vof, idir, Side::Hi);
              Real loArea = ebisBox.sumArea(vof, idir, Side::Lo);
              irregArea += (hiArea-loArea)*(hiArea-loArea);
            }
          bndryArea = sqrt(irregArea);

          //Compute normal (ignoring anisotropy if it exists)
          RealVect normal;
          const IntVect& iv = vof.gridIndex();
          if(ebisBox.isRegular(iv))
            { 
              normal = BASISREALV(0);
            }
          else if(ebisBox.isCovered(iv))
            {
              normal =  BASISREALV(0);
            }
          else
            {
              assert(ebisBox.isIrregular(iv));
              Real irregArea = bndryArea;
              if (irregArea> 0.0) 
                {
                  for(int idir = 0; idir < SpaceDim; idir++)
                    {
                      Real hiArea = ebisBox.sumArea(vof, idir, Side::Hi);
                      Real loArea = ebisBox.sumArea(vof, idir, Side::Lo);
                      normal[idir] = ((hiArea - loArea)/irregArea);
                    } //end loop over directions
                }
              else
                {
                  normal =  BASISREALV(0);
                }
              
            } //end else (vof is irregular)

          //fix normal for anisotropic case
          if (SpaceDim == 2)
            {
              for (int idir = 0;idir < SpaceDim;++idir)
                {
                  // (nx,ny)->(nxdy,nydx)
                  normal[idir] *= vectDx[1 - idir];
                }
            }
          if (SpaceDim==3)
            {
              for (int idir = 0;idir < SpaceDim;++idir)
                {
                  // (nx,ny,nz)->(nxdydz,nydxdz,nzdxdy)
                  normal[idir] *= vectDx[((idir-1)*(idir-2))/2]*vectDx[2 - ((idir-1)*idir)/2];
                }
            }
          Real anisBd = 0.;
          for (int idir = 0;idir < SpaceDim;++idir)
            {
              anisBd += normal[idir]*normal[idir];
            }
          anisBd = sqrt(anisBd);
          if (anisBd !=0.)
            {
              normal /= anisBd;
            }
          //fix bndryArea for anisotropic case
          bndryArea *= anisBd;
          bndryArea /= pow(maxDx,SpaceDim-1);

          //clipping ?? If someone wants to implement clipping, note that:
          //            sqrt(2) is no longer a magic clipping number for anisotropic bndryArea
          //            although it is a theoretical maximum for bndryArea (assuming we still
          //            scale bndryArea by maxDx)

          //copy results to volData
          volData(vof,0).m_bndryArea = bndryArea;
          volData(vof,0).m_normal    = normal;
        }
    }
  
}
/*******************************/
void
EBDataImplem::
coarsenBoundaryAreaAndNormal(Real&                    a_bndryAreaCoar,
                             RealVect&                a_normalCoar,
                             const Vector<Real>&      a_bndryAreaFine,
                             const Vector<RealVect>&  a_normalFine)
{
  //this is the factor by which the area of a fine
  //face is smaller than the area of a coarse face.
  Real faceCoarsenFactor = 1.0;
  for (int dir = 0; dir < SpaceDim-1; ++dir)
    {
      faceCoarsenFactor *= 0.5;
    }
  
  //A^B_xi, C = sum(A^B_xi, F/2^D-1)
  RealVect bndryAreaVec= RealVect::Zero;
  for(int ifine = 0; ifine < a_normalFine.size(); ifine++)
    {
      RealVect normalFine = a_normalFine[ifine];
      Real bndryAreaFine =  a_bndryAreaFine[ifine];
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          bndryAreaVec[idir] += normalFine[idir]*bndryAreaFine;
        }
    }
  bndryAreaVec *= faceCoarsenFactor;
  //A^B,C = ||A^B_xi,C||
  a_bndryAreaCoar = 0.;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      a_bndryAreaCoar += bndryAreaVec[idir]*bndryAreaVec[idir];
    }
  a_bndryAreaCoar = sqrt(a_bndryAreaCoar);
  
  //n_xi, C = A^B_xi,C/AB,c
  if(a_bndryAreaCoar > 0.)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          a_normalCoar[idir] = bndryAreaVec[idir]/a_bndryAreaCoar;
        }
    }
  else
    {
      a_normalCoar = BASISREALV(0);
    }
}
/*******************************/

void EBDataImplem::
coarsenVoFs(const EBDataImplem&  a_fineEBDataImplem,
            const EBGraph&       a_fineGraph,
            const EBGraph&       a_coarGraph,
            const Box&           a_validRegion)
{
  defineVoFData(a_coarGraph, a_validRegion);
  if(a_coarGraph.hasIrregular())
    {
      IntVectSet ivsIrreg = a_coarGraph.getIrregCells(a_validRegion);
      for(VoFIterator vofit(ivsIrreg, a_coarGraph); vofit.ok(); ++vofit)
        {
          const VolIndex& vofCoar = vofit();
          Vector<VolIndex> vofsFine = a_coarGraph.refine(vofCoar);
          int nFine = vofsFine.size();
          Vector<Real> bndryAreaFine(nFine);
          Vector<Real> volFracFine(nFine);
          Vector<RealVect> bndryCentroidFine(nFine);
          Vector<RealVect> volCentroidFine(nFine);
          Vector<RealVect> normalFine(nFine);
          int phase = -3;
          for(int ifine = 0; ifine < nFine; ifine++)
            {
              const VolIndex& vofFine =vofsFine[ifine];
              if(a_fineGraph.isIrregular(vofFine.gridIndex()))
                {
                  bndryAreaFine[ifine] = 
                    a_fineEBDataImplem.bndryArea(vofFine);
                  if(bndryAreaFine[ifine] != 0)
                    {
                      phase = a_fineEBDataImplem.facePhase(vofFine, 0);
                    }
                  volFracFine[ifine] = 
                    a_fineEBDataImplem.volFrac(vofFine);
                  bndryCentroidFine[ifine] =
                    a_fineEBDataImplem.bndryCentroid(vofFine);
                  volCentroidFine[ifine] = 
                    a_fineEBDataImplem.centroid(vofFine);
                  normalFine[ifine] = 
                    a_fineEBDataImplem.normal(vofFine);
                }
              else
                {
                  assert(a_fineGraph.isRegular(vofFine.gridIndex()));
                  bndryAreaFine[ifine] = 0.0;
                  volFracFine[ifine] = 1.0;
                  bndryCentroidFine[ifine] = RealVect::Zero;
                  volCentroidFine[ifine] = RealVect::Zero;
                  normalFine[ifine] = RealVect::Zero;
                }
            }
          if(phase == -3){
            // there was no boundary Surface for this coarse vof....so, 
          }
          Real volFracCoar, bndryAreaCoar;
          RealVect volCentroidCoar, bndryCentroidCoar, normalCoar;

          coarsenVolFracAndCentroid(volFracCoar,
                                    volCentroidCoar,
                                    volFracFine,
                                    volCentroidFine,
                                    vofsFine,
                                    vofCoar);

          coarsenBoundaryAreaAndNormal(bndryAreaCoar,
                                       normalCoar,
                                       bndryAreaFine,
                                       normalFine);

          coarsenBndryCentroid(bndryCentroidCoar,
                               bndryCentroidFine,
                               bndryAreaFine,
                               vofsFine,
                               vofCoar);
          VolData& vol = m_volData(vofCoar,0);
          vol.m_volFrac       = volFracCoar;
          vol.m_bndryArea     = bndryAreaCoar;
          vol.m_bndryCentroid = bndryCentroidCoar;
          vol.m_normal        = normalCoar;
          vol.m_volCentroid   = volCentroidCoar;
          vol.m_bndryPhase = phase;  // two-phase flow assumption
          vol.m_volIndex = 0;      // hack.  

        }
    }
}
/*
void EBDataImplem::
coarsenVoFs(const EBDataImplem&  a_fineEBDataImplem,
            const EBGraph&       a_fineGraph,
            const EBGraph&       a_coarGraph,
            const Box&           a_validRegion)
{
  defineVoFData(a_coarGraph, a_validRegion);
  if(a_coarGraph.hasIrregular())
    {
      IntVectSet ivsIrreg = a_coarGraph.getIrregCells(a_validRegion);
      for(VoFIterator vofit(ivsIrreg, a_coarGraph); vofit.ok(); ++vofit)
        {
          const VolIndex& vofCoar = vofit();
          Vector<VolIndex> vofsFine = a_coarGraph.refine(vofCoar);
          Vector<VolData*> fineVols(vofsFine.size(), NULL);
          for(int i=0; i<vofsFine.size(); ++i){
            VolIndex fine = vofsFine[i];
            fineVols[i] = new VolData(m_volData(fine, 0));
          }
          
          
        }
    }
}
*/             
          
/*******************************/
void EBDataImplem::
coarsenFaces(const EBDataImplem& a_fineEBDataImplem, 
             const EBGraph&      a_fineGraph,
             const EBGraph&      a_coarGraph, 
             const Box&          a_validRegion)
{
  defineFaceData(a_coarGraph, a_validRegion);
  IntVectSet ivsIrreg = a_coarGraph.getIrregCells(a_validRegion);
  Box fineRegion = a_fineGraph.getRegion();
  if(a_coarGraph.hasIrregular())
    {
      for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
        {
          FaceIterator faceit(ivsIrreg, a_coarGraph, faceDir, 
                              FaceStop::SurroundingWithBoundary);

          for(faceit.reset(); faceit.ok(); ++faceit)
            {
              const FaceIndex&  faceCoar  = faceit();
              Vector<FaceIndex> facesFine = a_coarGraph.refine(faceCoar, a_fineGraph);

              Vector<Real>     areaFracsFine(facesFine.size());
              Vector<RealVect> centroidsFine(facesFine.size());
              for(int ifine = 0; ifine < facesFine.size(); ifine++)
                {
                  const FaceIndex& faceFine = facesFine[ifine];
                  IntVect loiv = faceFine.gridIndex(Side::Lo);
                  IntVect hiiv = faceFine.gridIndex(Side::Hi);
                  if((fineRegion.contains(loiv) && a_fineGraph.isIrregular(loiv)) ||
                     (fineRegion.contains(hiiv) && a_fineGraph.isIrregular(hiiv)))
                    {
                      areaFracsFine[ifine] = a_fineEBDataImplem.areaFrac(faceFine);
                      centroidsFine[ifine] = a_fineEBDataImplem.centroid(faceFine);
                    }
                  else
                    {
                      areaFracsFine[ifine] = 1.0;
                      centroidsFine[ifine] = RealVect::Zero;
                    }
                }
              Real areaFracCoar;
              RealVect centroidCoar;
              coarsenAreaFrac(areaFracCoar, areaFracsFine);
              
              coarsenFaceCentroid(centroidCoar, centroidsFine,
                                  areaFracsFine, facesFine, 
                                  faceCoar);

              m_faceData[faceDir](faceCoar, 0).m_areaFrac     = areaFracCoar;
              m_faceData[faceDir](faceCoar, 0).m_faceCentroid = centroidCoar;
            } //end loop over faces
        } //end loop over face directions
    }
}
/*******************************/
void EBDataImplem::
coarsenFaceCentroid(RealVect&                a_centroidCoar, 
                    const Vector<RealVect>&  a_centroidsFine,
                    const Vector<Real>&      a_areaFracFine,
                    const Vector<FaceIndex>& a_facesFine, 
                    const FaceIndex&         a_faceCoar)
{
  Real totalFaceArea = 0.0;
  for(int ifineFace = 0; ifineFace < a_facesFine.size(); ifineFace++)
    {
      totalFaceArea += a_areaFracFine[ifineFace];
    }

  a_centroidCoar = RealVect::Zero;
  if(totalFaceArea > 0.0)
    {
      const IntVect& coarIV = a_faceCoar.gridIndex(Side::Lo);
      for(int ifineFace = 0; ifineFace < a_facesFine.size(); ifineFace++)
        {
          const FaceIndex& fineFace        = a_facesFine[ifineFace];
          IntVect fineIV = fineFace.gridIndex(Side::Lo);

          RealVect centroidFine = a_centroidsFine[ifineFace];
          centroidFine = fineToCoarseTransform(centroidFine,
                                               coarIV, fineIV);

          const Real&   areaFracFine= a_areaFracFine[ifineFace];
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              a_centroidCoar[idir] += 
                centroidFine[idir]*(areaFracFine/totalFaceArea);
            }
        }
    }
}
/*******************************/
void EBDataImplem::
coarsenAreaFrac(Real& a_areaFracCoar,
                const Vector<Real>& a_areaFracFine)
{
  //this is the factor by which the area of a fine
  //face is smaller than the area of a coarse face.
  Real faceCoarsenFactor = 1.0;
  for (int dir = 0; dir < SpaceDim-1; ++dir)
    {
      faceCoarsenFactor *= 0.5;
    }
  a_areaFracCoar = 0.0;
  for(int ifine = 0; ifine < a_areaFracFine.size(); ifine++)
    {
      a_areaFracCoar += a_areaFracFine[ifine];
    }
  a_areaFracCoar *= faceCoarsenFactor;
} 
/*******************************/
void
EBDataImplem::
coarsenVolFracAndCentroid(Real&                   a_volFracCoar,
                          RealVect&               a_volCentroidCoar,
                          const Vector<Real>&     a_volFracFine,
                          const Vector<RealVect>& a_volCentroidFine,
                          const Vector<VolIndex>& a_fineVoFs,
                          const VolIndex&         a_coarVoF)
{
  Real volCoarsenFactor = 1.0;
  for (int dir = 0; dir < SpaceDim; ++dir)
    {
      volCoarsenFactor *= 0.5;
    }

  Real totalVol = 0;
  for(int ifine = 0; ifine < a_fineVoFs.size(); ifine++)
    {
      totalVol  += a_volFracFine[ifine];
    }
  a_volFracCoar = volCoarsenFactor*totalVol;

  a_volCentroidCoar = RealVect::Zero;
  for(int ifine = 0; ifine < a_fineVoFs.size(); ifine++)
    {
      const VolIndex& fineVoF= a_fineVoFs[ifine];
      const IntVect& coarIV = a_coarVoF.gridIndex();
      const IntVect& fineIV = fineVoF.gridIndex();
      RealVect volCentroidFine = a_volCentroidFine[ifine];
      Real volFracFine=  a_volFracFine[ifine];
      volCentroidFine = fineToCoarseTransform(volCentroidFine,
                                              coarIV, fineIV);

      if(totalVol > 0.0)
        {
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              a_volCentroidCoar[idir] += 
                volCentroidFine[idir]*(volFracFine/totalVol);
            }
        }
    }
}
/*******************************/
RealVect 
EBDataImplem::
fineToCoarseTransform(const RealVect& a_finePoint,
                      const IntVect&  a_coarCell,
                      const IntVect&  a_fineCell)
{
  RealVect retval;
  //assuming nref = 2. make dxf = 1
  Real dxc = 2.0;
  RealVect fineCellLoc, coarCellLoc;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      fineCellLoc[idir] = Real(a_fineCell[idir]) + 0.5;
      coarCellLoc[idir] = dxc*(Real(a_coarCell[idir]) + 0.5);
    }
  retval = a_finePoint+ fineCellLoc;
  retval -= coarCellLoc;
  //to put it into a space where dxc = 1, the normalized answer
  retval /= dxc;
  return retval;
}
/*******************************/
void
EBDataImplem::
coarsenBndryCentroid(RealVect&               a_bndryCentroidCoar,
                     const Vector<RealVect>& a_bndryCentroidFine,
                     const Vector<Real>&     a_bndryAreaFine,
                     const Vector<VolIndex>& a_fineVoFs,
                     const VolIndex&         a_coarVoF)
{
  Real totalArea = 0;
  for(int ifine = 0; ifine < a_fineVoFs.size(); ifine++)
    {
      totalArea += a_bndryAreaFine[ifine];
    }

  a_bndryCentroidCoar = RealVect::Zero;
  for(int ifine = 0; ifine < a_fineVoFs.size(); ifine++)
    {
      const VolIndex& fineVoF= a_fineVoFs[ifine];
      const IntVect& fineIV=  fineVoF.gridIndex();
      const IntVect& coarIV=a_coarVoF.gridIndex();

      Real bndryAreaFine = a_bndryAreaFine[ifine];

      RealVect bndryCentroidFine = a_bndryCentroidFine[ifine];

      bndryCentroidFine = 
        fineToCoarseTransform(bndryCentroidFine,coarIV, fineIV);
      if(totalArea > 0.0)
        {
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              a_bndryCentroidCoar[idir] += 
                bndryCentroidFine[idir]*(bndryAreaFine/totalArea);
            }
        }
    }
}
/*******************************/
int 
EBDataImplem::size(const Box& a_region, const Interval& a_comps) const
{
  assert(m_isFaceDataDefined);
  assert(m_isVoFDataDefined);
  int linearSize = m_volData.size(a_region, a_comps);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      linearSize += m_faceData[idir].size(a_region, a_comps);
    }
  if(s_verbose)
    {
      pout() << " EBDataImplem::size returning " << linearSize << " for box " << a_region << endl; 
    }
  return linearSize;
}
/*******************************/
void 
EBDataImplem::linearOut(void* a_buf, 
                        const Box& a_region, 
                        const Interval& a_comps) const
{
  assert(m_isFaceDataDefined);
  assert(m_isVoFDataDefined);
  unsigned char* buffer = (unsigned char* ) a_buf;

  //int linearSize = 0;
  m_volData.linearOut(buffer, a_region, a_comps);
  buffer     +=   m_volData.size(a_region, a_comps);
  //linearSize += m_volData.size(a_region, a_comps);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      m_faceData[idir].linearOut(buffer, a_region, a_comps);
      buffer +=       m_faceData[idir].size(a_region, a_comps);
      //linearSize += m_faceData[idir].size(a_region, a_comps);
    }
}
/*******************************/
void 
EBDataImplem::linearIn(void*           a_buf, 
                       const Box&      a_region, 
                       const Interval& a_comps)
{
  assert(m_isFaceDataDefined);
  assert(m_isVoFDataDefined);
  //int linearSize = 0;
  unsigned char* buffer = (unsigned char* ) a_buf;
  m_volData.linearIn(buffer, a_region, a_comps);

  //linearSize += m_volData.size(a_region, a_comps);
  buffer     +=   m_volData.size(a_region, a_comps);

  for(int idir = 0; idir < SpaceDim; idir++)
    {
      m_faceData[idir].linearIn(buffer, a_region, a_comps);
      buffer +=       m_faceData[idir].size(a_region, a_comps);
      //linearSize += m_faceData[idir].size(a_region, a_comps);
    }
 
}
/*******************************/
/*******************************/
/*******************************/
/************************/
EBData::
EBData()
{
  m_implem = new EBDataImplem();
}
/************************/
EBData::
~EBData()
{
}
/************************/
void EBData::
define(const Box& a_box, int a_comps)
{
  m_implem->define(a_box, a_comps);
}
/************************/
EBData::
EBData(const Box& a_box, int a_comps)
{
  m_implem = new EBDataImplem(a_box, a_comps);
}
/************************/
void EBData::
copy(const Box&      a_regionFrom, 
     const Interval& a_Cd, 
     const Box&      a_regionTo, 
     const EBData&   a_source, 
     const Interval& a_Cs)
{
  m_implem->copy(a_regionFrom, a_Cd, a_regionTo, *a_source.m_implem, a_Cs);
}
/************************/
void EBData::
defineVoFData(const EBGraph& a_graph, 
              const Box& a_validBox)
{
  m_implem->defineVoFData(a_graph, a_validBox);
}
/************************/
void EBData::
defineFaceData(const EBGraph& a_graph,
               const Box& a_validBox)
{
  m_implem->defineFaceData(a_graph, a_validBox);
}
/************************/
void
EBData::define(const EBGraph&           a_graph,
               const Vector<IrregNode>& a_irregGraph,
               const Box&               a_validBox)
{
  m_implem->define(a_graph, a_irregGraph, a_validBox);
  computeNormalsAndBoundaryAreas(a_graph, a_validBox);
}
/*******************************/
Real EBData::volFrac(const VolIndex& a_vof) const
{
  return m_implem->volFrac(a_vof);
}
/*******************************/
Real EBData::bndryArea(const VolIndex& a_vof, int face) const
{
  return m_implem->bndryArea(a_vof, face);
}
/*******************************/
RealVect EBData::normal(const VolIndex& a_vof, int face) const
{
  return m_implem->normal(a_vof, face);
}
/*******************************/
RealVect EBData::centroid(const VolIndex& a_vof) const
{
  return m_implem->centroid(a_vof);
}
/*******************************/
RealVect EBData::bndryCentroid(const VolIndex& a_vof, int face) const
{
  return m_implem->bndryCentroid(a_vof, face);
}
/*******************************/
RealVect EBData::centroid(const FaceIndex& a_face) const
{
  return m_implem->centroid(a_face);
}
/*******************************/
Real EBData::areaFrac(const FaceIndex& a_face) const 
{
  return m_implem->areaFrac(a_face);
}
/*******************************/
void EBData::
coarsenVoFs(const EBData&  a_fineEBData,
            const EBGraph& a_fineGraph,
            const EBGraph& a_coarGraph,
            const Box&     a_validRegion)
{
  m_implem->coarsenVoFs(*a_fineEBData.m_implem, a_fineGraph, a_coarGraph, a_validRegion);
}
/*******************************/
void EBData::
coarsenFaces(const EBData& a_fineEBData, 
             const EBGraph& a_fineGraph,
             const EBGraph& a_coarGraph,
             const Box&     a_validRegion)
{
  m_implem->coarsenFaces(*a_fineEBData.m_implem, a_fineGraph, a_coarGraph, a_validRegion);
}
/*******************************/
EBData& EBData::operator=(const EBData& a_ebiin)
{
  if(&a_ebiin != this)
    {
      m_implem = a_ebiin.m_implem;
    }
  return *this;
}
/*******************************/
bool EBData::operator==(const EBData& a_ebiin)
{
  return (m_implem == a_ebiin.m_implem);
}
/*******************************/
EBData::EBData(const EBData& a_ebiin)
{
  m_implem = a_ebiin.m_implem;
}
/*******************************/
int 
EBData::size(const Box& R, const Interval& comps) const
{
  return  (m_implem->size(R, comps));
}
/*******************************/
void 
EBData::linearOut(void* buf, const Box& R, const Interval& comps) const
{
  m_implem->linearOut(buf, R, comps);
}
/*******************************/
void 
EBData::linearIn(void* buf, const Box& R, const Interval& comps)
{
  m_implem->linearIn(buf, R, comps);
}
/*******************************/


  /// used by multi-fluid code
int EBData::facePhase(const VolIndex& a_vof, int face) const 
{
  return m_implem->facePhase(a_vof, face);
}
  /// used by multi-fluid code
int EBData::faceIndex(const VolIndex& a_vof, int face) const 
{
  return m_implem->faceIndex(a_vof, face);
}
  /// used by multi-fluid code
void EBData::setFacePhase(const VolIndex& a_vof, int face, int phase)
{
  m_implem->setFacePhase(a_vof, face, phase);
}

int EBData::numFacePhase(const VolIndex& a_vof) const 
{
  return m_implem->numFacePhase(a_vof);
}
