/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#include "EBFaceFAB.H"
#include "EBArithF_F.H"
/**********************/
/**********************/
EBFaceFAB::EBFaceFAB():BaseEBFaceFAB<Real>()
{
}

/**********************/
/**********************/
EBFaceFAB::EBFaceFAB(const EBISBox& a_ebisBox, 
                     const Box& a_region, 
                     int a_iDir, int a_nComp)
  :BaseEBFaceFAB<Real>(a_ebisBox, a_region, a_iDir, a_nComp)
{
}

/**********************/
/**********************/
EBFaceFAB::~EBFaceFAB()
{
}

/**********************/
/**********************/
const FArrayBox& 
EBFaceFAB::getFArrayBox() const
{
  assert(isDefined());
  return (const FArrayBox&)m_regFAB;
}
/**********************/
/**********************/
FArrayBox& 
EBFaceFAB::getFArrayBox() 
{
  assert(isDefined());
  return (FArrayBox&)m_regFAB;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator+=(const EBFaceFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegionFace = a_src.m_regionFace & m_regionFace;
  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_ADDTWOFAB(CHF_FRA(m_regFAB), 
                     CHF_CONST_FRA(a_src.m_regFAB), 
                     CHF_BOX(locRegionFace));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, m_iDir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      for (int icomp = 0; icomp < m_nComp; ++icomp)
                        {
                          m_irrFAB(face, icomp) += 
                            a_src.m_irrFAB(face, icomp);
                        }
                    }
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator-=(const EBFaceFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegionFace = a_src.m_regionFace & m_regionFace;
  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_SUBTRACTTWOFAB(CHF_FRA(m_regFAB), 
                          CHF_CONST_FRA(a_src.m_regFAB), 
                          CHF_BOX(locRegionFace));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, m_iDir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      for (int icomp = 0; icomp < m_nComp; ++icomp)
                        {
                          m_irrFAB(face, icomp) -= 
                            a_src.m_irrFAB(face, icomp);
                        }
                    }
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator*=(const EBFaceFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegionFace = a_src.m_regionFace & m_regionFace;
  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_MULTIPLYTWOFAB(CHF_FRA(m_regFAB), 
                          CHF_CONST_FRA(a_src.m_regFAB), 
                          CHF_BOX(locRegionFace));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, m_iDir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      for (int icomp = 0; icomp < m_nComp; ++icomp)
                        {
                          m_irrFAB(face, icomp) *= 
                            a_src.m_irrFAB(face, icomp);
                        }
                    }
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator/=(const EBFaceFAB& a_src)
{
  assert(isDefined());
  assert(a_src.isDefined());
  assert(m_ebisBox == a_src.m_ebisBox);
  assert(a_src.m_nComp == m_nComp);

  Box locRegionFace = a_src.m_regionFace & m_regionFace;
  Box locRegion = a_src.m_region & m_region;
  if(!locRegion.isEmpty())
    {
      FORT_DIVIDETWOFAB(CHF_FRA(m_regFAB), 
                        CHF_CONST_FRA(a_src.m_regFAB), 
                        CHF_BOX(locRegionFace));

      IntVectSet ivsMulti = a_src.getMultiCells();
      ivsMulti &= getMultiCells();
      ivsMulti &= locRegion;
      IVSIterator ivsit(ivsMulti);
      for(ivsit.reset(); ivsit.ok(); ++ivsit)
        {
          const IntVect& iv = ivsit();
          Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  Vector<FaceIndex> faces = 
                    m_ebisBox.getFaces(vof, m_iDir, sit());
                  for(int iface = 0; iface < faces.size(); iface++)
                    {
                      const FaceIndex& face = faces[iface];
                      for (int icomp = 0; icomp < m_nComp; ++icomp)
                        {
                          m_irrFAB(face, icomp) /= 
                            a_src.m_irrFAB(face, icomp);
                        }
                    }
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator+=(const Real& a_src)
{
  assert(isDefined());

  FORT_ADDFABR(CHF_FRA(m_regFAB), 
               CHF_CONST_REAL(a_src), 
               CHF_BOX(m_regionFace));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, m_iDir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  for (int icomp = 0; icomp < m_nComp; ++icomp)
                    {
                      m_irrFAB(face, icomp) += a_src;
                    }
                }
            }
        }
    }
  return *this;
}

/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator-=(const Real& a_src)
{
  assert(isDefined());

  FORT_SUBTRACTFABR(CHF_FRA(m_regFAB), 
                    CHF_CONST_REAL(a_src), 
                    CHF_BOX(m_regionFace));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, m_iDir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  for (int icomp = 0; icomp < m_nComp; ++icomp)
                    {
                      m_irrFAB(face, icomp) -= a_src;
                    }
                }
            }
        }
    }
  return *this;
}
/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator*=(const Real& a_src)
{
  assert(isDefined());

  FORT_MULTIPLYFABR(CHF_FRA(m_regFAB), 
                    CHF_CONST_REAL(a_src), 
                    CHF_BOX(m_regionFace));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, m_iDir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  for (int icomp = 0; icomp < m_nComp; ++icomp)
                    {
                      m_irrFAB(face, icomp) *= a_src;
                    }
                }
            }
        }
    }
  return *this;
}
/**********************/
/**********************/
EBFaceFAB& 
EBFaceFAB::operator/=(const Real& a_src)
{
  assert(isDefined());

  FORT_DIVIDEFABR(CHF_FRA(m_regFAB), 
                  CHF_CONST_REAL(a_src), 
                  CHF_BOX(m_regionFace));

  const IntVectSet& ivsMulti = getMultiCells();
  IVSIterator ivsit(ivsMulti);
  for(ivsit.reset(); ivsit.ok(); ++ivsit)
    {
      const IntVect& iv = ivsit();
      Vector<VolIndex> vofs = m_ebisBox.getVoFs(iv);
      for(int ivof = 0; ivof < vofs.size(); ivof++)
        {
          const VolIndex& vof = vofs[ivof];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              Vector<FaceIndex> faces = 
                m_ebisBox.getFaces(vof, m_iDir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  for (int icomp = 0; icomp < m_nComp; ++icomp)
                    {
                      m_irrFAB(face, icomp) /= a_src;
                    }
                }
            }
        }
    }
  return *this;
}


