/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/

//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DTG  7-31-2001

#ifndef _EBFluxFAB_H_
#define _EBFluxFAB_H_

#include "Box.H"
#include "Vector.H"
#include "EBFaceFAB.H"

/// A EBFaceFAB-like container for edge-centered fluxes
/** 
    This is a class to contain face-centered fluxes on a box.
 */
class EBFluxFAB
{

public:

 ///
  /**
     This stuff required by LevelData in parallel:
  */
  static int preAllocatable() 
  {
    return 1; // symmetric messaging 
  } 

  /// default constructor
  EBFluxFAB();


  /// destructor
  ~EBFluxFAB();

  /// this function returns the EBFluxFAB to the undefined state
  void clear();

  /// {\bf access functions}

  /// number of components
  int nComp() const;

  /// returns cell-centered box which defines fluxBox
  const Box& getRegion() const;

  /// returns EBFaceFAB in direction dir
  EBFaceFAB& operator[] (const int dir);

  /// constant version 
  const EBFaceFAB& operator[] (const int dir) const;


  void setVal(const Real& val);

  ///
  /** Copies from a subsection of one box into another. 
      Assumes the boxes are both in the same index space, and that 
      box R is completely contained in both the src and destination 
      boxes.  This function equired by BoxLayoutData */
  void copy(const Box& Rfrom, 
            const Interval& Cdest,  
            const Box& Rto, 
            const EBFluxFAB& src,
            const Interval& Csrc);

  ///
  bool isDefined() const;

  ///
  EBFluxFAB(const EBISBox& a_ebisBox, const Box& bx, int n);

  ///
  void define(const EBISBox& a_ebisBox, const Box& bx, int n);

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  int size(const Box& R, const Interval& comps) const
  {
    MayDay::Error("size called for EBFluxFAB");
    return(-1);
  }

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  int regSize(const Box& R, const Interval& comps) const
  {
    MayDay::Error("regSize called for EBFluxFAB - unimplemented");
    return(-1);
  }

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  void linearOut(void* buf, const Box& R, const Interval& comps) const
  {
    MayDay::Error("linearOut called for EBFluxFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  void regLinearOut(void* buf, const Box& R, const Interval& comps) const
  {
    MayDay::Error("regLinearOut called for EBFluxFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  void linearIn(void* buf, const Box& R, const Interval& comps)
  {
    MayDay::Error("linearIn called for EBFluxFAB - unimplemented");
  }

  /// Unimplemented - necessary for LevelData<EBFluxFAB> to compile
  void regLinearIn(void* buf, const Box& R, const Interval& comps)
  {
    MayDay::Error("regLinearIn called for EBFluxFAB - unimplemented");
  }

  /// Invalid but necessary for LevelData<EBFluxFAB> to compile
  EBFluxFAB(const Box& bx, int n)
  {
    MayDay::Error("invalid constructor called for EBFluxFAB");
  }
protected:
  void setDefaultValues();

  ///
  Box m_region;

  ///
  EBISBox m_ebisBox;

  ///
  int m_nComp;

  bool m_isDefined;

  /// CH_SPACEDIM EBFaceFABes which hold fluxes
  Tuple<EBFaceFAB*, CH_SPACEDIM> m_fluxes;


private:
  /// these are disallowed
  
  EBFluxFAB (const EBFluxFAB&)
  {
    MayDay::Error("bogus constructor");
  }

  void operator = (const EBFluxFAB&)
  {
    MayDay::Error("bogus constructor");
  }

};

  

#endif


