/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _EBGRAPH_H_
#define _EBGRAPH_H_

#include "REAL.H"
#include "IntVect.H"
#include "IntVectSet.H"
#include "VolIndex.H"
#include "Interval.H"
#include "FaceIndex.H"
#include "IntVectSet.H"
#include "RealVect.H"
#include "BaseFab.H"
#include "SPMD.H"
#include "IrregNode.H"
#include "GraphNode.H"
#include "RefCountedPtr.H"

class EBGraph;
/// Geometric description within a box
/**
   EBGraphImplem represents the geometric information
   of the domain at a given refinement 
   within the boundaries of a particular box.
*/
class EBGraphImplem
{
public:
  ///
  /**
     The EBGraphImplem can be all regular or all
     covered or contain irregular cells.
  */
  enum TAG {AllRegular = 0, AllCovered, HasIrregular};

  ///
  /** 
      Makes an EBGraphImplem whose contents are undefined.
  */
  EBGraphImplem();

  ///
  ~EBGraphImplem();

  ///
  /**
     Return the VoFs in the cell.
  */
  Vector<VolIndex> getVoFs(const IntVect& a_iv) const;

  ///
  int numVoFs(const IntVect& a_iv) const;

  ///
  /**
     Return true if every cell in the EBGraphImplem is a regular cell.
  */
  bool isAllRegular() const;

  ///
  /**
     Return true neither all regular nor all covered
  */
  bool hasIrregular() const;

  ///
  /**
     Return true if  every cell in the EBGraphImplem is a covered cell.
  */
  bool isAllCovered() const;

  ///
  /**
     Return true if a_iv is an irregular cell.
  */
  bool isIrregular(const IntVect& a_iv) const;

  ///
  /**
     Return true if every cell in a_box is a 
     covered cell.
  */
  bool isCovered(const Box& a_box) const;

  ///
  /**
     Return true if every cell in a_box is a 
     regular cell.
  */
  bool isRegular(const Box& a_box) const;
  
  ///
  bool isRegular(const IntVect& a_iv) const;

  ///
  bool isCovered(const IntVect& a_iv) const;

  ///
  bool isConnected(const VolIndex& a_vof1,
                   const VolIndex& a_vof2) const;
  ///
  /**
     Return the faces on the side and direction
     of the input VoF. 
  */
  Vector<FaceIndex> getFaces(const VolIndex& a_vof, 
                             const int& a_idir, 
                             const Side::LoHiSide& a_sd) const;

  ///
  Vector<Vector<VolIndex> >  getVoFSets(const Box& a_region) const;

  ///
  /**
     Returns the corresponding set of VoFs from the next finer
     EBGraphImplem (factor of two refinement).  The result is only
     defined if this {\tt EBGraphImplem} was defined by coarsening.  
  */
  Vector<VolIndex> refine(const VolIndex& a_coarVoF) const;


  ///
  /**
     Returns the corresponding set of faces from the next finer
     EBGraphImplem (factor of two refinement).  The result is only
     defined if this {\tt EBGraphImplem} was defined by coarsening.  
  */
  Vector<FaceIndex> refine(const FaceIndex& a_coarFace, const EBGraphImplem& a_fineGraph) const;

  ///
  /**
     Returns the corresponding face from the next coarser
     EBGraphImplem (factor of two refinement).  This only makes
     sense if there is a coarser level.
  */
  FaceIndex coarsen(const FaceIndex& a_coarFace) const;

  ///
  /**
     Returns the corresponding  VoF from the next coarser 
     EBGraphImplem (same solution location, different index space, factor
     of two refinement ratio).
  */
  VolIndex coarsen(const VolIndex& a_fineVoF) const;


  ///
  /**
     Completely construct the graph from IrregNodes.  
     This makes a DENSE representation of the graph.  To make
     sparse representations of all regular or all covered graphs, 
     use setToAllRegular or setToAllCovered.\\
     irregGraph contains a complete list of irregular vofs. \\
     validRegion is the the region of the graph \\
     domain is the domain of compuation \\
     regIrregCovered = 1 for regular, 0 for irregular, -1 for covered.
     regIrregCovered must contain the valid region grown by 1. \\
     In chombospeak, \\
     regIrregCovered.box().contains(grow(validRegion, 1) & domain); \\
     otherwise an error will result
   */
  void
  buildGraph(const BaseFab<int>&      a_regIrregCovered,
             const Vector<IrregNode>& a_irregGraph,
             const Box&               a_validRegion,
             const Box&               a_domain);

  ///
  static void 
  checkGraph(const BaseFab<int>&      a_regIrregCovered,
             const Vector<IrregNode>& a_irregGraph,
             const Box&               a_validRegion,
             const Box&               a_domain)  ;

  ///
  /**
     Set the graph to all regular cells (sparse);
   */
  void setToAllRegular();

  ///
  /**
     Set the graph to all covered cells (sparse);
   */
  void setToAllCovered();

  ///
  /**
     Define an all-regular EBGraphImplem with the input box as its
     region. 
  */
  void define(const Box& box);


  EBGraphImplem(const Box& box);


  ///
  /**
     Copy the information from a_source to the over the intersection
     of the box a_region, the box of the current EBGraphImplem and 
     the box of a_source.  The Interval arguments are ignored.
     This function is required by LevelData.
  */
  void copy(const Box& a_regionFrom, 
            const Interval& Cd, 
            const Box& a_regionto, 
            const EBGraphImplem& a_source, 
            const Interval& Cs);


  ///
  const Box& getRegion() const;

  ///
  const Box& getDomain() const;

  ///
  /**
     This stuff required by LevelData in parallel:
  */
  int size(const Box& R, const Interval& comps) const;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  static int preAllocatable()
  {
    return 2; // dyanmic allocatable.  
  }

  ///
  bool isDefined() const;

  ///
  bool isDomainSet() const;

  //stuff below is not part of the public API
  
  //define vofs to be the coarsened vofs of the inputs
  //this must go first in coarsening.  defines happen here
  void coarsenVoFs(const EBGraphImplem&   a_fineGraph, 
                   const Box& a_coarRegion);
  
  //define the faces to be the coarsened faces of the input
  void coarsenFaces(const EBGraphImplem& a_coarGhostGraph,
                    const EBGraphImplem& a_fineEBIS);

  //fix fineVoF->coarseVoF information of the input
  void fixFineToCoarse(EBGraphImplem& a_fineEBIS) const;
 
  //don't ask
  Vector<int> 
  coarsenFaces(const VolIndex&       a_coarVoF,
               const EBGraphImplem&        a_coarGhostGraph,
               const EBGraphImplem&        a_fineGraph,
               const int&            a_idir,
               const Side::LoHiSide& a_sd);

  ///
  void setDomain(const Box& box);

  ///
  IntVectSet getIrregCells(const Box& a_subbox) const;
  
  ///
  IntVectSet getMultiCells(const Box& a_subbox) const;

  /// 
  Vector<FaceIndex> 
  getAllFaces(const IntVect& a_iv, 
              const int& a_idir, 
              const Side::LoHiSide& a_sd) const;

  ///
  void addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                            const EBGraph&    a_ghostGraph);

  ///
  void 
  getRegNextToMultiValued(IntVectSet&      a_vofsToChange,
                          const EBGraph&   a_ghostGraph) const;
private:

  ///box over which this graph is defined
  Box m_region;

  ///computational domain at this level of refinement
  Box m_domain;

  ///
  /**
     If this is allregular or allcovered, 
     the BaseFab below is undefined.
  */
  TAG m_tag;

  ///
  /**
     the graph description of the box.
     If the pointer GraphNode.m_thisNode == 0, the cell is covered.
     If the pointer GraphNode.m_thisNode == 1, the cell is regular.
     Otherwise, the pointer goes to the node of the graph.
  */
  BaseFab<GraphNode>  m_graph;

  ///
  bool m_isDefined;

  ///
  bool m_isDomainSet;

  ///
  IntVectSet* m_irregIVS;
  
  ///
  IntVectSet* m_multiIVS;
  
  static bool s_verbose;
private:
  
  void operator=(const EBGraphImplem& ebiin)
  {
    MayDay::Error("invalid operator");
  }
  EBGraphImplem(const EBGraphImplem& ebiin)
  {
    MayDay::Error("invalid operator");
  }
};


/// Geometric description within a box
/**
   EBGraph represents the geometric information
   of the domain at a given refinement 
   within the boundaries of a particular box.
   This is a refcounted object with all
   that implies.
*/
class EBGraph
{
public:
  ///
  /**
     The EBGraph can be all regular or all
     covered or contain irregular cells.
  */
  enum TAG {AllRegular = 0, AllCovered, HasIrregular};

  ///
  /** 
      Makes an EBGraph whose contents are undefined.
  */
  EBGraph();

  ///
  ~EBGraph();

  ///
  /**
     Return the VoFs in the cell.
  */
  Vector<VolIndex> getVoFs(const IntVect& a_iv) const;

  ///
  int numVoFs(const IntVect& a_iv) const;

  ///
  /**
     Return true if every cell in the EBGraph is a regular cell.
  */
  bool isAllRegular() const;

  ///
  /**
     Return true neither all regular nor all covered
  */
  bool hasIrregular() const;

  ///
  /**
     Return true if  every cell in the EBGraph is a covered cell.
  */
  bool isAllCovered() const;

  ///
  /**
     Return true if a_iv is an irregular cell.
  */
  bool isIrregular(const IntVect& a_iv) const;

  ///
  /**
     Return true if every cell in a_box is a 
     covered cell.
  */
  bool isCovered(const Box& a_box) const;

  ///
  /**
     Return true if every cell in a_box is a 
     regular cell.
  */
  bool isRegular(const Box& a_box) const;
  
  ///
  bool isRegular(const IntVect& a_iv) const;

  ///
  bool isCovered(const IntVect& a_iv) const;

  ///
  bool isConnected(const VolIndex& a_vof1,
                   const VolIndex& a_vof2) const;
  ///
  /**
     Return the faces on the side and direction
     of the input VoF. 
  */
  Vector<FaceIndex> getFaces(const VolIndex& a_vof, 
                             const int& a_idir, 
                             const Side::LoHiSide& a_sd) const;

  ///
  Vector<Vector<VolIndex> >  getVoFSets(const Box& a_region) const;
  ///
  /**
     Returns the corresponding set of VoFs from the next finer
     EBGraph (factor of two refinement).  The result is only
     defined if this {\tt EBGraph} was defined by coarsening.  
  */
  Vector<VolIndex> refine(const VolIndex& a_coarVoF) const;


  ///
  /**
     Returns the corresponding set of faces from the next finer
     EBGraph (factor of two refinement).  The result is only
     defined if this {\tt EBGraph} was defined by coarsening.  
  */
  Vector<FaceIndex> refine(const FaceIndex& a_coarFace, const EBGraph& a_fineGraph) const;

  ///
  /**
     Returns the corresponding face from the next coarser
     EBGraph (factor of two refinement).  This only makes
     sense if there is a coarser level.
  */
  FaceIndex coarsen(const FaceIndex& a_coarFace) const;

  ///
  /**
     Returns the corresponding  VoF from the next coarser 
     EBGraph (same solution location, different index space, factor
     of two refinement ratio).
  */
  VolIndex coarsen(const VolIndex& a_fineVoF) const;


  ///
  /**
     Define an all-regular EBGraph with the input box as its
     region.  The component argument is ignored.  
     This function is required by LevelData.
  */
  EBGraph(const Box& a_box, int a_comps);

  ///
  /**
     Completely construct the graph from IrregNodes.  
     This makes a DENSE representation of the graph.  To make
     sparse representations of all regular or all covered graphs, 
     use setToAllRegular or setToAllCovered.\\
     irregGraph contains a complete list of irregular vofs. \\
     validRegion is the the region of the graph \\
     domain is the domain of compuation \\
     regIrregCovered = 1 for regular, 0 for irregular, -1 for covered.
     regIrregCovered must contain the valid region grown by 1. \\
     In chombospeak, \\
     regIrregCovered.box().contains(grow(validRegion, 1) & domain); \\
     otherwise an error will result
   */
  void
  buildGraph(const BaseFab<int>&      a_regIrregCovered,
             const Vector<IrregNode>& a_irregGraph,
             const Box&               a_validRegion,
             const Box&               a_domain);

  ///
  void 
  checkGraph(const BaseFab<int>&      a_regIrregCovered,
             const Vector<IrregNode>& a_irregGraph,
             const Box&               a_validRegion,
             const Box&               a_domain)  ;

  ///
  /**
     Set the graph to all regular cells (sparse);
   */
  void setToAllRegular();

  ///
  /**
     Set the graph to all covered cells (sparse);
   */
  void setToAllCovered();

  ///
  /**
     Define an all-regular EBGraph with the input box as its
     region. 
  */
  void define(const Box& box);


  ///
  /**
     Copy the information from a_source to the over the intersection
     of the box a_region, the box of the current EBGraph and 
     the box of a_source.  The Interval arguments are ignored.
     This function is required by LevelData.
  */
  void copy(const Box& a_regionFrom, 
            const Interval& Cd, 
            const Box& a_regionto, 
            const EBGraph& a_source, 
            const Interval& Cs);


  ///
  const Box& getRegion() const;

  ///
  const Box& getDomain() const;

  ///
  bool isDefined() const;

  ///
  bool isDomainSet() const;

  //stuff below is not part of the public API
  
  //define vofs to be the coarsened vofs of the inputs
  //this must go first in coarsening.  defines happen here
  void coarsenVoFs(const EBGraph&   a_fineGraph, 
                   const Box& a_coarRegion);
  
  //define the faces to be the coarsened faces of the input
  void coarsenFaces(const EBGraph& a_coarGhostGraph,
                    const EBGraph& a_fineEBIS);

  //fix fineVoF->coarseVoF information of the input
  void fixFineToCoarse(EBGraph& a_fineEBIS) const;
 
  //don't ask
  Vector<int> 
  coarsenFaces(const VolIndex&       a_coarVoF,
               const EBGraph&        a_coarGhostGraph,
               const EBGraph&        a_fineGraph,
               const int&            a_idir,
               const Side::LoHiSide& a_sd);

  ///
  void setDomain(const Box& box);

  ///
  IntVectSet getIrregCells(const Box& a_subbox) const;
  
  ///
  IntVectSet getMultiCells(const Box& a_subbox) const;

  ///
  EBGraph& operator=(const EBGraph& ebiin);

  ///
  EBGraph(const EBGraph& a_ebiin);

  ///
  Vector<FaceIndex> 
  getAllFaces(const IntVect& a_iv, 
              const int& a_idir, 
              const Side::LoHiSide& a_sd) const;
  
  ///
  EBGraph(const Box& box);
  ///pointer comparison
  bool operator==(const EBGraph& a_ebiin);

  ///
  /**
     This stuff required by LevelData in parallel:
  */
  int size(const Box& R, const Interval& comps) const;

  ///
  void linearOut(void* buf, const Box& R, const Interval& comps) const;

  ///
  void linearIn(void* buf, const Box& R, const Interval& comps);

  ///
  static int preAllocatable()
  {
    return 2; // dyanmic allocatable.  
  }

  ///
  void addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                            const EBGraph&    a_ghostGraph);

  ///
  void 
  getRegNextToMultiValued(IntVectSet&    a_vofsToChange,
                          const EBGraph& a_ghostGraph) const;
private:

  ///
  RefCountedPtr<EBGraphImplem> m_implem;
};

#endif
