/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#include "EBGraph.H"
#include "DebugOut.H"
#include "EBDebugOut.H"
#include "BoxIterator.H"
#include "EBArith.H"

bool EBGraphImplem::s_verbose = false;
/*******************************/
void EBGraph::
getRegNextToMultiValued(IntVectSet&     a_vofsToChange,
                        const EBGraph&  a_ghostGraph) const
{
  m_implem->getRegNextToMultiValued(a_vofsToChange, a_ghostGraph);
}
/*******************************/
void EBGraphImplem::
getRegNextToMultiValued(IntVectSet&     a_vofsToChange,
                        const EBGraph&  a_ghostGraph) const
{
  Box ghostRegion = a_ghostGraph.getRegion();
  Box region = getRegion();
  assert(ghostRegion.contains(region));

  //loop through multiply-valued vofs in the grown region
  //if any of the vofs next to the multiply-valued vofs
  //are regular, collect them so they can be changed to 
  //irregular vofs with unit vol fracs and area fracs.
  a_vofsToChange = IntVectSet();
  IntVectSet multiIVS = a_ghostGraph.getMultiCells(ghostRegion);
  for(VoFIterator vofit(multiIVS, a_ghostGraph); 
      vofit.ok(); ++vofit)
    {
      const VolIndex& multiVoF = vofit();
      const IntVect& multiIV = multiVoF.gridIndex();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          for(SideIterator sit; sit.ok(); ++sit)
            {
              IntVect otherIV = multiIV + sign(sit())*BASISV(idir);
              if((region.contains(otherIV) && isRegular(otherIV)))
                {
                  a_vofsToChange |= otherIV;
                }
            }
        }
    }
}

/*******************************/
void EBGraph::
addFullIrregularVoFs(const IntVectSet& a_vofsToChange,
                     const EBGraph&    a_ghostGraph)
{
  m_implem->addFullIrregularVoFs(a_vofsToChange, a_ghostGraph);
}
/*******************************/
void EBGraphImplem::
addFullIrregularVoFs(const IntVectSet&  a_vofsToChange,
                     const EBGraph&     a_ghostGraph)
{
  if(!a_vofsToChange.isEmpty())
    {
      assert(isDefined());
      assert(isDomainSet());
      //this is all supposed to be called for regular vofs
      //to be changed to full irregular vofs
      assert(!isAllCovered());
      if(isAllRegular())
        {
          m_tag = HasIrregular;
          if(m_irregIVS != NULL) delete m_irregIVS;
          if(m_multiIVS != NULL) delete m_multiIVS;
          m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_graph.define(m_region, 1);
          for(BoxIterator bit(m_region); bit.ok(); ++bit)
            { 
              m_graph(bit(), 0).defineAsRegular();
            }
        }
  
      //  //now for changing vofs
      for(IVSIterator ivsit(a_vofsToChange); ivsit.ok(); ++ivsit)
        {
          const IntVect&  iv = ivsit();
          //needs to be a regular cell to start with
          assert(isRegular(iv));
          VolIndex vof(iv, 0);

          GraphNode regularNode;
          regularNode.defineAsRegular();

          //create node and its arcs
          //the coarse-fine info is created later.
          //this operation must be done before all that
          //the finer ones can be set trivially but not the 
          //coarse ones.
          GraphNodeImplem node;
          node.m_finerNodes = regularNode.refine(vof);
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              for(SideIterator sit; sit.ok(); ++sit)
                {
                  
                  int gNodeIndex = node.index(idir, sit());
                  Vector<int>& nodeArcsDir = node.m_arc[gNodeIndex];
                  nodeArcsDir.resize(1);
                  //find which vof node is connected to in each direction.
                  //cannot use isConnected here because it will always return
                  //true since one vof is still regular 
                  IntVect  otherIV = iv + sign(sit())*BASISV(idir);
                  if(m_domain.contains(otherIV))
                    {
                      Vector<VolIndex> otherVoFs = a_ghostGraph.getVoFs(otherIV);
                      bool found = false;
                      for(int iother = 0; iother < otherVoFs.size(); iother++)
                        {
                          const VolIndex& otherVoF = otherVoFs[iother];
                          Vector<FaceIndex> otherFaces = a_ghostGraph.getFaces(otherVoF, idir, flip(sit()));
                          //otherwise multiple faces on a regular vof
                          assert(otherFaces.size() <= 1);
                          if(otherFaces.size() == 1)
                            {
                              found = true;
                              nodeArcsDir[0] = otherVoF.cellIndex();
                            }
                        }
                      if(!found)
                        {
                          pout() << iv << " was specified as regular  but is not connected to anything in " << otherIV << endl;
                          MayDay::Error("former regular vof not connected to anything");
                        }
                    }
                  else
                    {
                      //boundary arc
                      nodeArcsDir[0] = -1;
                    }
                    
                }
            }
          //finally add node into graph.
          //again coarse and fine info have to be added on later.
          m_graph(iv, 0).addIrregularNode(node);
          (*m_irregIVS) |= iv;
          if(m_graph(iv, 0).size() > 1)
            {
              MayDay::Error("that vof was already irregular");
            }
        }
    }
}
/*******************************/
IntVectSet EBGraphImplem::
getIrregCells(const Box& a_subbox) const
{
  static IntVectSet emptySet;
  if(m_irregIVS == NULL) return emptySet;
  return (*m_irregIVS) & a_subbox;
}
/*******************************/
IntVectSet EBGraphImplem::
getMultiCells(const Box& a_subbox) const
{
  static IntVectSet emptySet;
  if(m_multiIVS == NULL) return emptySet;
  return (*m_multiIVS) & a_subbox;
}
/*******************************/
void EBGraphImplem::
setToAllRegular()
{
  assert(isDefined());
  assert(isDomainSet());
  m_tag = AllRegular;
  if(m_irregIVS != NULL) delete m_irregIVS;
  if(m_multiIVS != NULL) delete m_multiIVS;
  m_irregIVS = NULL;
  m_multiIVS = NULL;
}
/*******************************/
void EBGraphImplem::
setToAllCovered()
{
  assert(isDefined());
  assert(isDomainSet());
  m_tag = AllCovered;
  if(m_irregIVS != NULL) delete m_irregIVS;
  if(m_multiIVS != NULL) delete m_multiIVS;
  m_irregIVS = NULL;
  m_multiIVS = NULL;
}
/*******************************/
void EBGraphImplem::
checkGraph(const BaseFab<int>&      a_regIrregCovered,
           const Vector<IrregNode>& a_irregGraph,
           const Box&               a_validRegion,
           const Box&               a_domain)
{

  //this will check to see if the graph covers the region
  IntVectSet ivsUndefined(a_validRegion);
  ///account for regular and covered cells
  for(BoxIterator bit(a_validRegion); bit.ok(); ++bit)
    {
      //-1 is covered, 0 is irregular, and 1 is regular
      int fabval = a_regIrregCovered(bit(), 0);
      if(fabval != 0)
        {
          if(!(fabval == 1 || fabval == -1))
            {
              MayDay::Error("invalid flag in buildgraph");
            }
          ivsUndefined -= bit();
        }
    }
  //account for irregular vofs
  for(int ivec = 0; ivec < a_irregGraph.size(); ivec++)
    {
      const IrregNode& inputNode = a_irregGraph[ivec];
      //remove nodes points from undefined set
      ivsUndefined -= inputNode.m_cell;
    }
  //if the input graph does not cover the region, abort
  if(!ivsUndefined.isEmpty())
    {
      MayDay::Error("irreg graph does not cover region");
    }
}
/*******************************/
void EBGraphImplem::
buildGraph(const BaseFab<int>&      a_regIrregCovered,
           const Vector<IrregNode>& a_irregGraph,
           const Box&               a_validRegion,
           const Box&               a_domain)
{
#if 0
  // this is awful slow
  checkGraph(a_regIrregCovered, a_irregGraph, a_validRegion, a_domain);
#endif
  define(a_validRegion);
  setDomain(a_domain);

  //remember that this forces a dense representation.
  m_tag = HasIrregular;
  if(m_irregIVS != NULL) delete m_irregIVS;
  if(m_multiIVS != NULL) delete m_multiIVS;
  m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
  m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
  m_graph.define(m_region, 1);
  
  //set regular and covered cells
  for(BoxIterator bit(m_region); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();
      if(a_regIrregCovered(iv, 0) == 1) //regular cell
        { 
          m_graph(iv, 0).defineAsRegular();
        }
      else if(a_regIrregCovered(iv, 0) == -1) //covered cell
        { 
          m_graph(iv, 0).defineAsCovered();
        }
      else if(a_regIrregCovered(iv, 0) != 0) 
        {
          MayDay::Error("invalid flag");
        }
    }

  //now for irregular cells
  //add the vofs
  for(int ivecIrreg = 0; ivecIrreg < a_irregGraph.size(); ivecIrreg++)
    {
      const IrregNode& inputNode = a_irregGraph[ivecIrreg];
      const IntVect& iv =inputNode.m_cell;

      GraphNodeImplem newImplem;
      newImplem.m_nodeInd = ivecIrreg;

      m_graph(iv, 0).addIrregularNode(newImplem, inputNode.m_cellIndex);
      (*m_irregIVS) |= iv;
      if(m_graph(iv, 0).size() > 1)
        {
          (*m_multiIVS) |= iv;
        }
    }

  //add the faces
  for(int ivecIrreg = 0; ivecIrreg < a_irregGraph.size(); ivecIrreg++)
    {
      const IrregNode& inputNode = a_irregGraph[ivecIrreg];
      const IntVect& iv =inputNode.m_cell;
      Vector<GraphNodeImplem>& vecNodes = 
        *(m_graph(iv, 0).m_cellList);
      //pick out which node we are talking about 
      //by maching its nodeInd with ivecIrreg
      bool foundNode = false;
      GraphNodeImplem* nodePtr;
      for(int ivecGraph = 0; ivecGraph < vecNodes.size(); ivecGraph++)
        {
          if(vecNodes[ivecGraph].m_nodeInd == ivecIrreg)
            {
              foundNode = true;
              nodePtr = &(vecNodes[ivecGraph]);
            }
        }
      if(!foundNode)
        {
          MayDay::Error("internal error in construction");
        }
      //now add the arcs in the input to the node
      GraphNodeImplem& node = *nodePtr;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          for(SideIterator sit; sit.ok(); ++sit)
            {
              int irregIndex = inputNode.index(idir, sit());
              int gNodeIndex = node.index(idir, sit());
              const Vector<int>& irregArcs = inputNode.m_arc[irregIndex];
              Vector<int>& nodeArcs = node.m_arc[gNodeIndex];
              nodeArcs.resize(irregArcs.size());
              for(int iarc = 0; iarc < irregArcs.size(); iarc++)
                {
                  int otherNodeInd = irregArcs[iarc];
                  if(otherNodeInd == -1)
                    {
                      //if otherNodeInd == -1, boundary arc.
                      //just make the arc in our node = -1 
                      //to signify the same
                      nodeArcs[iarc] = -1;
                    }
                  else if(otherNodeInd == -2)
                    {
                      //this means that the vof is connected
                      //to a regular vof,
                      //which always have a cell index of 0
                      nodeArcs[iarc] =  0;
                    }
                  else
                    {
                      nodeArcs[iarc] =  otherNodeInd;
                    }
                }
            }
        }
      
    }
}
/*******************************/
const Box& EBGraphImplem::
getRegion() const
{
  return m_region; 
}
/*******************************/
const Box& EBGraphImplem::
getDomain() const
{
  return m_domain; 
}
/*******************************/
EBGraphImplem::
EBGraphImplem(const Box& a_box)
  :m_irregIVS(NULL),  m_multiIVS(NULL)
{

  define(a_box);
}
/*******************************/
void EBGraphImplem::
define(const Box& a_region)
{
  m_tag = AllRegular;
  m_region = a_region;
  if(m_irregIVS != NULL) delete m_irregIVS;
  if(m_multiIVS != NULL) delete m_multiIVS;
  m_irregIVS = NULL;
  m_multiIVS = NULL;
  m_isDefined= true;
}
/*******************************/
void EBGraphImplem::
setDomain(const Box& a_domain)
{
  m_isDomainSet = true;
  m_domain = a_domain;
}
/*******************************/
EBGraphImplem::
EBGraphImplem()
  :m_irregIVS(NULL),  m_multiIVS(NULL)
{
  m_isDefined = false;
  m_isDomainSet = false;
}
/*******************************/
EBGraphImplem::
~EBGraphImplem()
{
  if(m_irregIVS != NULL) delete m_irregIVS;
  if(m_multiIVS != NULL) delete m_multiIVS;
}
/*******************************/
bool EBGraphImplem::
isDefined() const
{
  return m_isDefined;
}
/*******************************/
bool EBGraphImplem::
isDomainSet() const
{
  return m_isDomainSet;
}
/*******************************/
Vector<VolIndex> EBGraphImplem::
getVoFs(const IntVect& a_iv) const
{
  assert(isDefined());
  assert(isDomainSet());
  Vector<VolIndex> retvec;
  if(m_tag == AllRegular)
    {
      retvec.push_back(VolIndex(a_iv, 0));
    }
  else if(m_tag == AllCovered)
    {
      //return an empty vector
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_iv));
      assert(m_domain.contains(a_iv));
      const GraphNode& node = m_graph(a_iv, 0);
      retvec = node.getVoFs(a_iv);
    }
  return retvec;
}
/*******************************/
bool EBGraphImplem::
isRegular(const IntVect& a_iv) const
{
  assert(isDefined());
  assert(isDomainSet());
  bool retval;
  if(m_tag == AllRegular)
    {
      retval = true;
    }
  else if(m_tag == AllCovered)
    {
      retval = false;
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_iv));
      assert(m_domain.contains(a_iv));
      const GraphNode& node = m_graph(a_iv, 0);
      retval = node.isRegular();
    }
  else
    {
      MayDay::Error("EBGraphImplem::isRegular:Bogus Tag");
    }
  return retval;
}
/*******************************/
Vector<FaceIndex> 
EBGraphImplem::getAllFaces(const IntVect& a_iv, 
                           const int& a_idir, 
                           const Side::LoHiSide& a_sd) const
{
  Vector<FaceIndex> retval(0);
  Vector<VolIndex> vofs = getVoFs(a_iv);
  for(int ivof= 0; ivof < vofs.size(); ivof++)
    {
      retval.append(getFaces(vofs[ivof], a_idir, a_sd));
    }
  return retval;
}
/*******************************/
bool EBGraphImplem::
isIrregular(const IntVect& a_iv) const
{
  assert(isDefined());
  assert(isDomainSet());
  bool retval;
  if(m_tag == AllRegular)
    {
      retval = false;
    }
  else if(m_tag == AllCovered)
    {
      retval = false;
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_iv));
      assert(m_domain.contains(a_iv));
      const GraphNode& node = m_graph(a_iv, 0);
      retval = node.isIrregular();
    }
  else
    {
      MayDay::Error("EBGraphImplem::isIrregular:Bogus Tag");
    }
  return retval;
}
/*******************************/
bool EBGraphImplem::
isAllCovered() const
{
  return (m_tag == AllCovered);
}
/*******************************/
bool EBGraphImplem::
isAllRegular() const
{
  return m_tag == AllRegular;
}
/*******************************/
int 
EBGraphImplem::size(const Box& a_region, const Interval& a_comps) const
{
  assert(isDefined());
  assert(isDomainSet());
  //regular irregular covered code
  int linearSize = sizeof(int);
  if(!isRegular(a_region) && !isCovered(a_region))
    {
      for(BoxIterator bit(a_region); bit.ok(); ++bit)
        {
          const GraphNode& node = m_graph(bit(), 0);
          int nodeSize = node.linearSize();
          linearSize += nodeSize;
        }
    }
  return linearSize;
}
/*******************************/
void 
EBGraphImplem::linearOut(void* a_buf, 
                         const Box& a_region, 
                         const Interval& a_comps) const
{
  assert(isDefined());
  assert(isDomainSet());
  int secretCode;
  if(isCovered(a_region))
    {
      secretCode = 0;
    }
  else if(isRegular(a_region))
    {
      secretCode = 1;
    }
  else
    {
      secretCode = 2;
    }
  //  int linearSize = sizeof(int);

  int* intbuf = (int*) a_buf;
  *intbuf = secretCode;
  intbuf++;

  if(!isRegular(a_region) && !isCovered(a_region))
    {
      unsigned char* buffer = (unsigned char*) intbuf;
      for(BoxIterator bit(a_region); bit.ok(); ++bit)
        {
          const GraphNode& node = m_graph(bit(), 0);
          int nodeSize = node.linearSize();
          node.linearOut(buffer);
          buffer += nodeSize;
          //          linearSize += nodeSize;
        }
    }
}
/*******************************/
void 
EBGraphImplem::linearIn(void*           a_buf, 
                        const Box&      a_region, 
                        const Interval& a_comps)
{
  assert(isDefined());
  assert(isDomainSet());
  assert(isDefined());
  assert(isDomainSet());
  int* intbuf = (int*) a_buf;
  int secretCode = *intbuf;
  intbuf++;
  //  int linearSize = sizeof(int);

  if(secretCode == 0)
    {
      //all covered input
      EBGraphImplem ebgraphSrc(a_region);
      ebgraphSrc.setDomain(m_domain);
      ebgraphSrc.setToAllCovered();
      copy(a_region, a_comps, a_region, ebgraphSrc, a_comps);
    }
  else if(secretCode == 1)
    {
      //all regular input
      EBGraphImplem ebgraphSrc(a_region);
      ebgraphSrc.setDomain(m_domain);
      ebgraphSrc.setToAllRegular();
      copy(a_region, a_comps, a_region, ebgraphSrc, a_comps);
    }
  else
    { 
      assert(secretCode==2);
      unsigned char* buffer = (unsigned char*) intbuf;

      if(isAllRegular() || isAllCovered())
        {
          m_tag = HasIrregular;
          if(m_irregIVS != NULL) delete m_irregIVS;
          if(m_multiIVS != NULL) delete m_multiIVS;
          m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_graph.define(m_region, 1);
        }
      for(BoxIterator bit(a_region); bit.ok(); ++bit)
        {
          GraphNode& node = m_graph(bit(), 0);
          node.linearIn(buffer);
          if(node.isIrregular()>0) (*m_irregIVS)|=bit();
          if(node.size()>1) (*m_multiIVS)|=bit();
          int nodeSize = node.linearSize();
          buffer += nodeSize;
          //          linearSize += nodeSize;
        }
    }
}
/*******************************/
bool EBGraph::
hasIrregular() const
{
  return m_implem->hasIrregular();
}
/*******************************/
bool EBGraphImplem::
hasIrregular() const
{
  return m_tag == HasIrregular;
}
/*******************************/
bool EBGraphImplem::
isCovered(const IntVect& a_iv) const
{
  assert(isDefined());
  assert(isDomainSet());
  bool retval;
  if(m_tag == AllRegular)
    {
      retval = false;
    }
  else if(m_tag == AllCovered)
    {
      retval = true;
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_iv));
      assert(m_domain.contains(a_iv));
      const GraphNode& node = m_graph(a_iv, 0);
      retval = node.isCovered();
    }
  else
    {
      MayDay::Error("EBGraphImplem::isIrregular:Bogus Tag");
    }
  return retval;
}
/*******************************/
bool EBGraphImplem::
isCovered(const Box& a_box) const
{
  assert(isDefined());
  assert(isDomainSet());
  bool retval;
  if(m_tag == AllRegular)
    {
      retval = false;
    }
  else if(m_tag == AllCovered)
    {
      retval = true;
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_box));
      assert(m_domain.contains(a_box));
      retval = true;
      BoxIterator bit(a_box);
      for(bit.reset(); bit.ok() && retval; ++bit)
        {
          if(!isCovered(bit()))
            {
              retval = false;
            }
        }
    }
  else
    {
      MayDay::Error("EBGraphImplem::isCovered:Bogus Tag");
    }
  return retval;
}
/*******************************/
bool EBGraphImplem::
isRegular(const Box& a_box) const
{
  assert(isDefined());
  assert(isDomainSet());
  bool retval;
  if(m_tag == AllRegular)
    {
      retval = true;
    }
  else if(m_tag == AllCovered)
    {
      retval = false;
    }
  else if(m_tag == HasIrregular)
    {
      assert(m_region.contains(a_box));
      assert(m_domain.contains(a_box));
      BoxIterator bit(a_box);
      retval = true;
      for(bit.reset(); bit.ok() && retval; ++bit)
        {
          if(!isRegular(bit()))
            {
              retval = false;
            }
        }
    }
  else
    {
      MayDay::Error("EBGraphImplem::isRegular:Bogus Tag");
    }
  return retval;
}
/*******************************/
Vector<FaceIndex> EBGraphImplem::
getFaces(const VolIndex&       a_vof, 
         const int&            a_idir, 
         const Side::LoHiSide& a_sd) const
{
  assert(isDefined());
  assert(isDomainSet());
  assert(m_region.contains(a_vof.gridIndex()));
  assert(m_domain.contains(a_vof.gridIndex()));
  assert((a_idir >= 0) && (a_idir < SpaceDim));
  assert((a_sd == Side::Lo) || (a_sd == Side::Hi));

  Vector<FaceIndex> retvec;
  if(m_tag == AllRegular)
    {
      IntVect otherIV = a_vof.gridIndex() 
        + sign(a_sd)*BASISV(a_idir);
      int otherCellInd = 0;
      if(!m_domain.contains(otherIV))
        {
          otherCellInd = -1;
        }
      VolIndex otherVoF(otherIV, otherCellInd);
      retvec.push_back(FaceIndex(a_vof, otherVoF, a_idir));  
    }
  else if(m_tag == AllCovered)
    {
      //return empty vector
    }
  else if(m_tag == HasIrregular)
    {
      const IntVect& iv = a_vof.gridIndex();
      const GraphNode& node = m_graph(iv, 0);
      retvec = node.getFaces(a_vof, a_idir, a_sd, m_domain);
    }

  return retvec;
}
/*******************************/
FaceIndex EBGraphImplem::
coarsen(const FaceIndex& a_fineFace) const
{
  VolIndex loVoFCoar = coarsen(a_fineFace.getVoF(Side::Lo));
  VolIndex hiVoFCoar = coarsen(a_fineFace.getVoF(Side::Hi));
  return FaceIndex(loVoFCoar, hiVoFCoar, a_fineFace.direction());
}
/*******************************/
Vector<FaceIndex> EBGraphImplem::
refine(const FaceIndex& a_coarFace, const EBGraphImplem& a_fineGraph) const
{
  Vector<FaceIndex> retval;
  IntVect ivLoCoar = a_coarFace.gridIndex(Side::Lo);
  IntVect ivHiCoar = a_coarFace.gridIndex(Side::Hi);
  int direction = a_coarFace.direction();
  if(m_region.contains(ivLoCoar) && m_region.contains(ivHiCoar))
    {
      //interior face
      Vector<VolIndex> loVoFFine = refine(a_coarFace.getVoF(Side::Lo));
      Vector<VolIndex> hiVoFFine = refine(a_coarFace.getVoF(Side::Hi));
      int idir = a_coarFace.direction();
      for(int ilo = 0; ilo < loVoFFine.size(); ilo++)
        {
          for(int ihi = 0; ihi < hiVoFFine.size(); ihi++)
            {
              if(a_fineGraph.isConnected(loVoFFine[ilo], hiVoFFine[ihi]))
                {
                  FaceIndex newFace(loVoFFine[ilo], hiVoFFine[ihi], idir);
                  retval.push_back(newFace);
                }
            }
        }
    }
  else if(m_region.contains(ivLoCoar))
    {
      //boundary face on the hi side of the domain
      Vector<VolIndex> loVoFsFine = refine(a_coarFace.getVoF(Side::Lo));
      Box fineRegion = ebrefine(m_region, 2);
      for(int ivof = 0; ivof < loVoFsFine.size(); ivof++)
        {
          VolIndex& loVoFFine = loVoFsFine[ivof];
          IntVect ivHiFine = loVoFFine.gridIndex() + BASISV(direction);
          if(!fineRegion.contains(ivHiFine))
            {
              Vector<FaceIndex> fineFaces = a_fineGraph.getFaces(loVoFFine, direction, Side::Hi);
              retval.append(fineFaces);
            }
        }
    }
  else if(m_region.contains(ivHiCoar))
    {
      //boundary face on the low side of the domain
      Vector<VolIndex> hiVoFsFine = refine(a_coarFace.getVoF(Side::Hi));
      Box fineRegion = ebrefine(m_region, 2);
      for(int ivof = 0; ivof < hiVoFsFine.size(); ivof++)
        {
          VolIndex& hiVoFFine = hiVoFsFine[ivof];
          IntVect ivLoFine = hiVoFFine.gridIndex() - BASISV(direction);
          if(!fineRegion.contains(ivLoFine) )
            {
              Vector<FaceIndex> fineFaces = a_fineGraph.getFaces(hiVoFFine, direction, Side::Lo);
              retval.append(fineFaces);
            }
        }
    }
  else
    {
      MayDay::Error("neither vof of the face is inside the domain");
    }
  return retval;
}
/*******************************/
bool 
EBGraphImplem::isConnected(const VolIndex& a_vof1,
                           const VolIndex& a_vof2) const
{
  assert(isDefined());
  const IntVect& iv1 = a_vof1.gridIndex();
  const IntVect& iv2 = a_vof2.gridIndex();

  assert(m_region.contains(iv1));
  assert(m_region.contains(iv2));
  VolIndex vofLo, vofHi;
  //use the lexical greaterThan because it will
  //always give the correct answer since only one component 
  //is different.
  if(iv1.lexGT(iv2))
    {
      vofLo = a_vof2;
      vofHi = a_vof1;
    }
  else
    {
      vofLo = a_vof1;
      vofHi = a_vof2;
    }
  int direction;
  bool dirfound;
  const IntVect& ivLo = vofLo.gridIndex();
  const IntVect& ivHi = vofHi.gridIndex();

  dirfound = false;
  for(int idir = 0; ((idir<SpaceDim) && !dirfound); idir++)
    {
      if((ivHi - ivLo) == BASISV(idir)) 
        {
          direction = idir;
          dirfound = true;
        }
    }
  //if not neigboring intvects, no way it can be connected
  if(!dirfound) return false;

  Vector<FaceIndex> faces = getFaces(vofLo, direction, Side::Hi);
  bool voffound = false;
  for(int iface = 0; iface < faces.size(); iface++)
    {
      const FaceIndex& face = faces[iface];
      if(face.getVoF(Side::Hi) == vofHi)
        {
          voffound = true;
        }
    }
  return voffound;
}
/*******************************/
VolIndex EBGraphImplem::
coarsen(const VolIndex& a_fineVoF) const
{
  VolIndex retval;
  IntVect ivfine = a_fineVoF.gridIndex();
  IntVect ivcoar = ebcoarsen(ivfine, 2);
  if(!m_domain.contains(ivfine))
    {
      //boundary vof
      retval = VolIndex(ivcoar, -1);
    }
  else if((m_tag == AllRegular ) || (m_tag == AllCovered ))
    {
      retval = VolIndex(ivcoar, 0);
    }
  else
    {
      assert(m_tag == HasIrregular);
      const IntVect& iv = a_fineVoF.gridIndex();
      const GraphNode& node = m_graph(iv, 0);
      retval = node.coarsen(a_fineVoF);
    }
  return retval;
}
/*******************************/
Vector<VolIndex> EBGraphImplem::
refine(const VolIndex& a_coarVoF) const
{
  assert(isDefined());
  assert(isDomainSet());

  Vector<VolIndex> retval(0);
  IntVect ivCoar = a_coarVoF.gridIndex();
  assert(m_domain.contains(ivCoar));

  if(m_tag == AllRegular)
    {
      if(s_verbose)
        {
          pout() << "tag is all regular" << endl;
        }
      const IntVect& iv = a_coarVoF.gridIndex();
      Box refbox = ebrefine(Box(iv,iv),2);
      BoxIterator bit(refbox);
      for(bit.reset(); bit.ok(); ++bit)
        {
          retval.push_back(VolIndex(bit(), 0));
        }
    }
  else if(m_tag == AllCovered)
    {
      //return empty vector
      if(s_verbose)
        {
          pout() << "tag is all covered" << endl;
        }
    }
  else 
    {
      assert(m_tag == HasIrregular);
      const IntVect& iv = a_coarVoF.gridIndex();
      const GraphNode& node = m_graph(iv, 0);
      if(s_verbose)
        {
          pout() << "tag is has irregular" << endl;
          if(node.isIrregular())
            {
              pout() << " node has " << node.m_cellList->size() << " elements" << endl;
              const Vector<GraphNodeImplem>& nodes = *node.m_cellList;
              for(int inode = 0; inode < nodes.size(); inode++)
                {
                  const GraphNodeImplem node = nodes[inode];
                  pout() << "finernodes size  = " <<  node.m_finerNodes.size() << endl;
                }
            }
        }
      retval = node.refine(a_coarVoF);
    }
  return retval;
}
/*******************************/
void EBGraphImplem::
copy(const Box& a_regionFrom,           
     const Interval& a_intDst, 
     const Box& a_regionTo,           
     const EBGraphImplem& a_source, 
     const Interval& a_intSrc)
{
  assert(isDefined());
  assert(isDomainSet());
  if(isAllRegular() && a_source.isAllRegular())
    {
      return;
    }
  else if(isAllCovered() && a_source.isAllCovered())
    {
      return;
    }
  if(a_source.isCovered(a_regionFrom) && a_regionTo.contains(m_region))
    {
      setToAllCovered();
      return;
    }
  if(a_source.isRegular(a_regionFrom) && a_regionTo.contains(m_region))
    {
      setToAllRegular();
      return;
    }
  if(isAllRegular() && a_source.isAllCovered())
    {
      //define the basefab as all regular and set the region to 
      //covered in the intersection
      m_tag = HasIrregular;
      if(m_irregIVS != NULL) delete m_irregIVS;
      if(m_multiIVS != NULL) delete m_multiIVS;
      m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_graph.define(m_region, 1);
      GraphNode regularNode;
      regularNode.defineAsRegular();
      m_graph.setVal(regularNode);
      Box interBox = m_region & a_regionTo;
      for(BoxIterator bit(interBox); bit.ok(); ++bit)
        {
          m_graph(bit(), 0).defineAsCovered();
        }
    }
  else if(isAllCovered() && a_source.isAllRegular())
    {
      //define the basefab as all covered and set the region to 
      //regular in the intersection
      m_tag = HasIrregular;
      if(m_irregIVS != NULL) delete m_irregIVS;
      if(m_multiIVS != NULL) delete m_multiIVS;
      m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_graph.define(m_region, 1);
      GraphNode  coveredNode;
      coveredNode.defineAsCovered();
      m_graph.setVal(coveredNode);
      Box interBox = m_region & a_regionTo;
      for(BoxIterator bit(interBox); bit.ok(); ++bit)
        {
          m_graph(bit(), 0).defineAsRegular();
        }
    }
  else 
    {
      //one or both has irregular cells.
      //because i am sick of combinatorics,
      //use basefab copy to transfer the data.

      //see if we need to generate a source fab
      BaseFab<GraphNode>* srcFabPtr;
      bool needToDelete;
      if(a_source.hasIrregular())
        { 
          srcFabPtr = (BaseFab<GraphNode>*)&a_source.m_graph;
          needToDelete = false;
        }
      else
        {
          needToDelete = true;
          srcFabPtr = new BaseFab<GraphNode>(a_regionFrom, 1);
          GraphNode srcVal;
          if(a_source.isAllRegular())
            {
              srcVal.defineAsRegular();
            }
          else
            {
              //this really has to be true
              assert(a_source.isAllCovered());
              srcVal.defineAsCovered();
            }
          srcFabPtr->setVal(srcVal);
        }
      //define our graph if i need to. leave alone otherwise
      if(isAllRegular() || isAllCovered())
        {
          if(m_irregIVS != NULL) delete m_irregIVS;
          if(m_multiIVS != NULL) delete m_multiIVS;
          m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
          m_graph.define(m_region, 1);
        }

      //copy the data
      m_tag = HasIrregular;
      m_graph.copy(a_regionFrom, a_intSrc, 
                   a_regionTo, *srcFabPtr, a_intDst);

      //if we needed to new the basefab, clean it up
      if(needToDelete)
        {
          delete srcFabPtr;
        }
    }
#if 1
  if(a_source.hasIrregular() && hasIrregular())
    {
      IntVectSet ivsInterIrreg = (*a_source.m_irregIVS) & m_region;
      IntVectSet ivsInterMulti = (*a_source.m_multiIVS) & m_region;
      (*m_irregIVS) |= ivsInterIrreg;
      (*m_multiIVS) |= ivsInterMulti;
    }
#else
            for(BoxIterator bit(m_region); bit.ok(); ++bit)
              {
                IntVect iv = bit();
                //      GraphNode& node = m_graph(iv, 0);
                if(isIrregular(iv))
                  {
                    (*m_irregIVS) |= iv;
                    if(numVoFs(iv) > 1)
                      {
                        (*m_multiIVS) |= iv;
                      }
                  }
              }
#endif    
  //pout() << " for box " << m_region << endl;
  //dumpIVS(&m_irregIVS);
}
/*******************************/
void EBGraphImplem::
coarsenVoFs(const EBGraphImplem& a_fineGraph,
            const Box&     a_coarRegion)
{
  //this also defines the boxes
  m_region = a_coarRegion;
  m_domain = ebcoarsen(a_fineGraph.getDomain(), 2);
  m_isDomainSet = true;
  m_isDefined = true;
  if(a_fineGraph.isCovered(ebrefine(m_region, 2)))
    {
      m_tag = AllCovered;
    }
  else if(a_fineGraph.isRegular(ebrefine(m_region, 2)))
    {
      m_tag = AllRegular;
    }
  else
    {
      m_tag = HasIrregular;
      m_region = a_coarRegion;
      m_domain = ebcoarsen(a_fineGraph.m_domain, 2);
      if(m_irregIVS != NULL) delete m_irregIVS;
      if(m_multiIVS != NULL) delete m_multiIVS;
      m_multiIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_irregIVS = new IntVectSet(DenseIntVectSet(m_region, false));
      m_graph.define(m_region, 1);
      for(BoxIterator bit(m_region); bit.ok(); ++bit)
        {
          Box fineBox = ebrefine(Box(bit(), bit()), 2);
          if(a_fineGraph.isRegular(fineBox))
            {
              m_graph(bit(), 0).defineAsRegular();
            }
          else if(a_fineGraph.isCovered(fineBox))
            {
              m_graph(bit(), 0).defineAsCovered();
            }
          else
            {
              //get sets of all connected vofs in the box
              //the number of sets will be the number of 
              //vofs in the coarse cell
              Vector<Vector<VolIndex> > fineVoFSets 
                = a_fineGraph.getVoFSets(fineBox);

              for(int iset = 0; iset < fineVoFSets.size(); iset++)
                {
                  GraphNodeImplem newImplem;
                  newImplem.m_finerNodes = fineVoFSets[iset];
                  m_graph(bit(), 0).addIrregularNode(newImplem);
                  (*m_irregIVS) |= bit();
                  if(m_graph(bit(), 0).size() > 1)
                    {
                      (*m_multiIVS) |= bit();
                    }
                }
            }
        }
    }
}
/*******************************/
Vector<Vector<VolIndex> >  EBGraphImplem::
getVoFSets(const Box& a_region) const
{
  Vector<Vector<VolIndex> > retval;
  //gather all  vofs
  Vector<VolIndex> allVoFs;
  for(BoxIterator bit(a_region); bit.ok(); ++bit)
    {
      allVoFs.append(getVoFs(bit()));
    }

  std::vector<bool> beenAdded(allVoFs.size(), false);
  for(int ivof = 0; ivof < allVoFs.size(); ivof++)
    {
      if(!beenAdded[ivof])
        {
          const VolIndex& thisVoF = allVoFs[ivof];
          //we have a vof to start with.
          //now find all vofs connected to this inside
          //this cell----not necessarily directly---
          //hence the inner loop of kvof.
          Vector<VolIndex> thisVoFSet(1, thisVoF);
          beenAdded[ivof] = true;
          bool doneAdding = false;
          while(!doneAdding)
            {
              int oldSetSize = thisVoFSet.size();
              //i can start at ivof+1 here because 0 to ivof
              //has always been added.
              for(int jvof = ivof+1; jvof < allVoFs.size(); jvof++)
                { 
                  const VolIndex& testVoF = allVoFs[jvof];
                  //this might be a tad confusing to people because
                  //the length of the vector can be changing inside
                  //the loop.  fortran would puke horribly here.
                  for(int kvof = 0; kvof < thisVoFSet.size(); kvof++)
                    {
                      //because of the nature of push_back, 
                      //the vector's pointer can change inside so beware.
                      //need to hold the vof by value here 
                      VolIndex innerVoF = thisVoFSet[kvof];
                      if(!beenAdded[jvof] && isConnected(innerVoF, testVoF))
                        {
                          thisVoFSet.push_back(testVoF);
                          beenAdded[jvof] = true;
                        }
                    }
                }
              doneAdding = (thisVoFSet.size() == oldSetSize);
            } //end while !doneadding
          //add the finished bunch to the list of  vofs.
          retval.push_back(thisVoFSet);
        } //end this bunch of connected vofs
    }

  //can't hurt to check in case some wacky case was missed.
  for(int ivof = 0; ivof < allVoFs.size(); ivof++)
    {
      if(!beenAdded[ivof])
        {
          MayDay::Error("We seem to have missed a vof in getVoFSets");
        }
    }

  return retval;
}
/*******************************/
Vector<int> EBGraphImplem::
coarsenFaces(const VolIndex&       a_coarVoF,
             const EBGraphImplem&  a_coarGhostGraph,
             const EBGraphImplem&  a_fineGraph,
             const int&            a_idir,
             const Side::LoHiSide& a_sd)
{
  Vector<int> retval;

  IntVect coarIV = a_coarVoF.gridIndex();
  IntVect otherIV= coarIV + sign(a_sd)*BASISV(a_idir); 
  Vector<VolIndex> theseFineVoFs = a_coarGhostGraph.refine(a_coarVoF);

  if(m_domain.contains(otherIV))
    {
      //interior faces.
      //just get all possible vofs to connect to and
      //check connectivity
      Vector<VolIndex> otherCoarVoFs = a_coarGhostGraph.getVoFs(otherIV);
      for(int iotherCoar = 0; iotherCoar < otherCoarVoFs.size(); iotherCoar++)
        {
          const VolIndex& otherCoarVoF = otherCoarVoFs[iotherCoar];
          Vector<VolIndex> otherFineVoFs = a_coarGhostGraph.refine(otherCoarVoF);
          bool addThisFace = false;
          for(int iotherFine = 0; iotherFine < otherFineVoFs.size(); iotherFine++)
            {
              const VolIndex& otherFineVoF = otherFineVoFs[iotherFine];
              for(int ithisFine = 0; ithisFine < theseFineVoFs.size(); ithisFine++)
                {
                  const VolIndex& thisFineVoF = theseFineVoFs[ithisFine];
                  int thesetwocon = -1;
                  if(a_fineGraph.isConnected(thisFineVoF, otherFineVoF))
                    {
                      addThisFace = true;
                      thesetwocon = 1;
                    }
                }
            }
          if(addThisFace)
            {
              retval.push_back(iotherCoar);
            }
        }
    }
  else
    {
      //boundary faces.  
      //if there are any boundary faces on the fine level,
      //make one here too
      bool hasBoundaryFaceFine = false;
      for(int ithis = 0; ithis < theseFineVoFs.size(); ithis++)
        {
          const VolIndex& thisVoF = theseFineVoFs[ithis];
          Vector<FaceIndex> fineFaces = 
            a_fineGraph.getFaces(thisVoF,a_idir, a_sd);
          for(int iface = 0; iface < fineFaces.size(); iface++)
            {
              if(fineFaces[iface].isBoundary())
                {
                  hasBoundaryFaceFine = true;
                }
            }
        }
      if(hasBoundaryFaceFine)
        {
          //remember that -1 is the secret code for boundary face arcs
          retval.push_back(-1);
        }
    }
  return retval; 
}
/*******************************/
void EBGraphImplem::
coarsenFaces(const EBGraphImplem& a_coarGhostGraph,
             const EBGraphImplem& a_fineGraph)
{
  if(hasIrregular())
    {
      assert(a_coarGhostGraph.getDomain() == m_domain);
      for(BoxIterator bit(m_region); bit.ok(); ++bit)
        {
          if(isIrregular(bit()))
            {
              Vector<VolIndex> vofsCoar = getVoFs(bit());
              Vector<GraphNodeImplem>& nodes = 
                *(m_graph(bit(), 0).m_cellList);
              for(int ivof = 0; ivof < vofsCoar.size(); ivof++)
                {
                  const VolIndex& vofCoar= vofsCoar[ivof];
                  GraphNodeImplem& node = nodes[vofCoar.cellIndex()];
                  for(int idir = 0; idir < SpaceDim; idir++)
                    {
                      for(SideIterator sit; sit.ok(); ++sit)
                        {
                          Vector<int> coarArcs = 
                            coarsenFaces(vofsCoar[ivof],
                                         a_coarGhostGraph,
                                         a_fineGraph,
                                         idir, sit());

                          int nodeind = node.index(idir, sit());
                          node.m_arc[nodeind] = coarArcs;
                        }
                    }
                }
            }
        }
    }
}
/*******************************/
void EBGraphImplem::
fixFineToCoarse(EBGraphImplem& a_fineGraph) const
{
  if(hasIrregular())
    {
      for(BoxIterator bit(m_region); bit.ok(); ++bit)
        {
          if(isIrregular(bit()))
            {
              const IntVect&     ivCoar = bit();
              const Vector<GraphNodeImplem>& 
                nodesCoar = *(m_graph(ivCoar, 0).m_cellList);
              for(int icoar = 0; icoar < nodesCoar.size(); icoar++)
                {
                  VolIndex vofCoar(ivCoar, icoar);
                  Vector<VolIndex> vofsFine = refine(vofCoar);
                  for(int ifine = 0; ifine < vofsFine.size(); ifine++)
                    {
                      const IntVect&   ivFine = vofsFine[ifine].gridIndex();
                      if(a_fineGraph.isIrregular(ivFine))
                        {
                          int cellIndexFine       = vofsFine[ifine].cellIndex();

                          Vector<GraphNodeImplem>& 
                            nodesFine = *(a_fineGraph.m_graph(ivFine, 0).m_cellList);
                      
                          nodesFine[cellIndexFine].m_coarserNode = icoar;
                        }
                    }
                }
              
            }
        }
    }
}
/*******************************/
/*******************************/
/*******************************/
void
EBGraph::buildGraph(const BaseFab<int>&      a_regIrregCovered,
                    const Vector<IrregNode>& a_irregGraph,
                    const Box&               a_validRegion,
                    const Box&               a_domain)
{
  m_implem->buildGraph(a_regIrregCovered, a_irregGraph, a_validRegion, a_domain);
}
/*******************************/
void EBGraph::copy(const Box& a_regionFrom,     
                   const Interval& Cd, 
                   const Box& a_regionTo,     
                   const EBGraph& a_source, 
                   const Interval& Cs)
{
  m_implem->copy(a_regionFrom, Cd, a_regionTo, 
                 *(a_source.m_implem), Cs);
}
/*******************************/
bool EBGraph::isDefined() const
{
  return m_implem->isDefined();
}
/*******************************/
bool EBGraph::isConnected(const VolIndex& a_vof1,
                          const VolIndex& a_vof2) const
{
  return m_implem->isConnected(a_vof1, a_vof2);
}
/*******************************/
const Box& EBGraph::getRegion() const
{
  return m_implem->getRegion();
}
/*******************************/
EBGraph& EBGraph::operator=(const EBGraph& a_ebiin)
{
  if(&a_ebiin != this)
    {
      m_implem = a_ebiin.m_implem;
    }
  return *this;
}
/*******************************/
EBGraph::EBGraph(const Box& a_box, int a_comps)
{
  m_implem = new EBGraphImplem(a_box);
}
/*******************************/
EBGraph::EBGraph(const Box& a_box)
{
  m_implem = new EBGraphImplem(a_box);
}
/*******************************/
EBGraph::EBGraph(const EBGraph& a_ebiin)
{
  m_implem = a_ebiin.m_implem;
}
/*******************************/
EBGraph::EBGraph()
{
  m_implem = new EBGraphImplem();
}
/*******************************/
EBGraph::~EBGraph()
{;}
/*******************************/
IntVectSet EBGraph::getMultiCells(const Box& a_subbox) const
{
  return m_implem->getMultiCells(a_subbox);
}
/*******************************/
IntVectSet EBGraph::getIrregCells(const Box& a_subbox) const
{
  return m_implem->getIrregCells(a_subbox);
}
/*******************************/
Vector<VolIndex> EBGraph::getVoFs(const IntVect& a_iv) const
{
  return m_implem->getVoFs(a_iv);
}
/*******************************/
bool EBGraph::isRegular(const IntVect& a_iv) const
{
  return m_implem->isRegular(a_iv);
}
/*******************************/
bool EBGraph::isAllRegular() const
{
  return m_implem->isAllRegular();
}
/*******************************/
bool EBGraph::isAllCovered() const
{
  return m_implem->isAllCovered();
}
/*******************************/
bool EBGraph::isIrregular(const IntVect& a_iv) const
{
  return m_implem->isIrregular(a_iv);
}
/*******************************/
bool EBGraph::isCovered(const IntVect& a_iv) const
{
  return m_implem->isCovered(a_iv);
}
/*******************************/
bool EBGraph::isCovered(const Box& a_box) const
{
  return m_implem->isCovered(a_box);
}
/*******************************/
bool EBGraph::isRegular(const Box& a_box) const
{
  return m_implem->isRegular(a_box);
}
/*******************************/
Vector<FaceIndex> 
EBGraph::getFaces(const VolIndex& a_vof, 
                  const int& a_idir, 
                  const Side::LoHiSide& a_sd) const
{
  return m_implem->getFaces(a_vof, a_idir, a_sd);
}
/*******************************/
Vector<VolIndex> 
EBGraph::refine(const VolIndex& a_coarVoF) const
{
  return m_implem->refine(a_coarVoF);
}
/*******************************/
Vector<FaceIndex> 
EBGraph::refine(const FaceIndex& a_coarFace, const EBGraph& a_fineGraph) const
{
  return m_implem->refine(a_coarFace, *a_fineGraph.m_implem);
}
/*******************************/
FaceIndex EBGraph::coarsen(const FaceIndex& a_fineFace) const
{
  return m_implem->coarsen(a_fineFace);
}
/*******************************/
VolIndex EBGraph::coarsen(const VolIndex& a_fineVoF) const
{
  return m_implem->coarsen(a_fineVoF);
}
/*******************************/
bool EBGraph::operator==(const EBGraph& a_ebiin)
{
  return (m_implem == a_ebiin.m_implem);
}
/*******************************/
const Box& EBGraph::getDomain() const
{
  return (m_implem->getDomain());
}
/*******************************/
void EBGraph::setDomain(const Box& box)
{
  m_implem->setDomain(box);
}
/*******************************/
void EBGraph::setToAllRegular()
{
  m_implem->setToAllRegular();
}
/*******************************/
void EBGraph::setToAllCovered()
{
  m_implem->setToAllCovered();
}
/*******************************/
void EBGraph::
coarsenVoFs(const EBGraph& a_fineEBIS,
            const Box& a_coarRegion)
{
  m_implem->coarsenVoFs(*a_fineEBIS.m_implem, a_coarRegion);
}
/*******************************/
void EBGraph::
coarsenFaces(const EBGraph& a_coarGhost, const EBGraph& a_fineEBIS)
{
  m_implem->coarsenFaces(*a_coarGhost.m_implem, *a_fineEBIS.m_implem);
}
/*******************************/
void EBGraph::
fixFineToCoarse(EBGraph& a_fineEBIS) const
{
  m_implem->fixFineToCoarse(*a_fineEBIS.m_implem);
}
/*******************************/
Vector<FaceIndex> 
EBGraph::getAllFaces(const IntVect& a_iv, 
                     const int& a_idir, 
                     const Side::LoHiSide& a_sd) const
{
  return m_implem->getAllFaces(a_iv, a_idir, a_sd);
}
/*******************************/
int 
EBGraphImplem::numVoFs(const IntVect& a_iv) const
{
  Vector<VolIndex> vofs = getVoFs(a_iv);
  return vofs.size();
}
/*******************************/
int 
EBGraph::numVoFs(const IntVect& a_iv) const
{
  return m_implem->numVoFs(a_iv);
}
/*******************************/
void 
EBGraph::define(const Box& a_box)  
{
  m_implem->define(a_box);
}
/*******************************/
int 
EBGraph::size(const Box& R, const Interval& comps) const
{
  return  (m_implem->size(R, comps));
}
/*******************************/
void 
EBGraph::linearOut(void* buf, const Box& R, const Interval& comps) const
{
  m_implem->linearOut(buf, R, comps);
}
/*******************************/
void 
EBGraph::linearIn(void* buf, const Box& R, const Interval& comps)
{
  m_implem->linearIn(buf, R, comps);
}
/*******************************/

