/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _EBISLAYOUT_H_
#define _EBISLAYOUT_H_

#include "REAL.H"
#include "IntVect.H"
#include "IntVectSet.H"
#include "VolIndex.H"
#include "FaceIndex.H"
#include "IntVectSet.H"
#include "EBISBox.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "EBGraph.H"

class EBISLayoutImplem;

/// 
/**
   EBISLayout represents the geometric information
   on a union of rectangles.    This is 
   another ref-counted pointer object so copying
   it is cheap but has the ref-counted pointer semantic.
*/
class EBISLayout
{
public:
  ///
  /**
     Access the {\tt EBISBox} associated with the input DataIndex.
     Only constant access is permitted.
  */

  const EBISBox& operator[](const DataIndex& a_index) const;

  ///
  /**
     Sets the maximum level of refinement that this {\tt EBISLayout}
     will have to perform.  Creates and holds new {\tt EBISLayout}s
     at intermediate levels of refinement.
     Default is one (no refinement done).
  */
  void setMaxRefinementRatio(const int& a_maxRefine);


  ///
  /**
     Sets the maximum level of coarsening that this {\tt EBISLayout}
     will have to perform.  Creates and holds new {\tt EBISLayout}s
     at intermediate levels of refinement.
     Default is one (no coarsening done).
  */
  void setMaxCoarseningRatio(const int& a_maxCoarsen);


  ///
  EBISLayout();


  ///
  /**
     Returns the ghosted layout associated with the ebisl.
   */
  const BoxLayout& getGrownLayout() const;

  ///
  /**
     Returns the disjoint layout that was sent in.
   */
  const DisjointBoxLayout& getDisjointLayout() const;

  ///
  ~EBISLayout();

  ///
  /**
     Returns the index of the VoF corresponding to coarsening
     the input VoF by the input ratio.  It is an error if
     the ratio is greater than the maximum coarsening ratio
     or if the vof does not exist at the input data index.
   */
  VolIndex coarsen(const VolIndex& a_vof,
                   const int& a_ratio,
                   const DataIndex& a_datInd) const;

  ///
  /**
     Returns the indices of the VoFs corresponding to refining
     the input VoF by the input ratio.  It is an error if
     the ratio is greater than the maximum refinement ratio
     or if the vof does not exist at the input data index.
  */
  Vector<VolIndex> refine(const VolIndex& a_vof,
                          const int& a_ratio,
                          const DataIndex& a_datInd) const;

  ///
  int getMaxCoarseningRatio() const;
  
  ///
  int getMaxRefinementRatio() const;

  ///
  void define(const Box& a_domain,
              const DisjointBoxLayout& a_grids,
              const int& a_nghost,
              const LevelData<EBGraph>& a_graph,
              const LevelData<EBData> & a_data);

protected:
  ///
  RefCountedPtr<EBISLayoutImplem> m_implem;
};


/// 
/**
   EBISLayoutImplem represents the geometric information
   on a union of rectangles.  This is not really part
   of the public interface.  Users should use EBISLayout.
*/
class EBISLayoutImplem
{
public:

  ///
  const EBISBox& operator[](const DataIndex&) const;

  ///
  EBISLayoutImplem();

  ///
  ~EBISLayoutImplem();

  ///
  const BoxLayout& getGrownLayout() const
  {
    return m_blGhostDom;
  }

  const DisjointBoxLayout& getDisjointLayout() const
  {
    return m_dblInputDom;
  }

  ///
  /**
     Returns the index of the VoF corresponding to coarsening
     the input VoF by the input ratio.  It is an error if
     the ratio is greater than the maximum coarsening ratio
     or if the vof does not exist at the input data index.
   */
  VolIndex coarsen(const VolIndex& a_vof,
                   const int& a_ratio,
                   const DataIndex& a_datInd) const;

  ///
  /**
     Returns the indices of the VoFs corresponding to refining
     the input VoF by the input ratio.  It is an error if
     the ratio is greater than the maximum refinement ratio
     or if the vof does not exist at the input data index.
  */
  Vector<VolIndex> refine(const VolIndex& a_vof,
                          const int& a_ratio,
                          const DataIndex& a_datInd) const;

  ///
  /**
     Sets the maximum level of refinement that this {\tt EBISLayout}
     will have to perform.  Creates and holds new {\tt EBISLayout}s
     at intermediate levels of refinement.
     Default is one (no refinement done).
  */
  void setMaxRefinementRatio(const int& a_maxRefine);


  ///
  /**
     Sets the maximum level of coarsening that this {\tt EBISLayout}
     will have to perform.  Creates and holds new {\tt EBISLayout}s
     at intermediate levels of refinement.
     Default is one (no coarsening done).
  */
  void setMaxCoarseningRatio(const int& a_maxCoarsen);

  ///
  int getMaxCoarseningRatio() const;
  
  ///
  int getMaxRefinementRatio() const;

  ///
  void define(const Box& a_domain,
              const DisjointBoxLayout& a_grids,
              const int& a_nghost,
              const LevelData<EBGraph>& a_graph,
              const LevelData<EBData> & a_data);


  ///
  static void setVerbose(bool a_verbose);
private:
  ///
  BoxLayout m_blGhostDom;
  
  ///
  DisjointBoxLayout m_dblInputDom;

  ///
  Vector<EBISLayout> m_fineLevels;

  ///
  Vector<EBISLayout> m_coarLevels;

  ///
  int m_nghost;

  ///
  int m_maxCoarseningRatio;

  ///
  int m_maxRefinementRatio;

  ///
  Box m_domain;

  ///
  LayoutData<EBISBox> m_ebisBoxes;

private:
  static bool s_verbose;
  void operator=(const EBISLayoutImplem& ebiin)
  {;}
  EBISLayoutImplem(const EBISLayoutImplem& ebiin)
  {;}

};



#endif
