/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _EBINDEXSPACE_H_
#define _EBINDEXSPACE_H_

#include "REAL.H"
#include "IntVect.H"
#include "IntVectSet.H"
#include "VolIndex.H"
#include "FaceIndex.H"
#include "IntVectSet.H"
#include "DisjointBoxLayout.H"
#include "LevelData.H"
#include "EBISBox.H"
#include "GeometryService.H"
#include "EBISLayout.H"
#include "CH_HDF5.H"

class EBIndexSpace;

///singleton version of EBIndexSpace.
class Chombo_EBIS
{
private:
  ///
  /**  
   This is NULL until the first time 
     instance() is called.
     On the first time it is called it
     is set to  new EBIndexSpace();
     After that it points to the same 
     thing.
   */
  static EBIndexSpace* s_instance;

  Chombo_EBIS()
  {;}

  ~Chombo_EBIS()
  {;}
public:
  ///
  /**
     Return a pointer to the singleton
     EBIndexSpace.  Until this is called
     once, the singleton is null.  See
     Design patterns for why this is done
     this way.
   */
  static EBIndexSpace* instance();

};


/// 
/**
   EBISLevel represents the geometric information
   of the domain at a particular level of refinement.
*/
class EBISLevel
{
public:


  void coarsenVoFs(EBISLevel& a_fineEBIS);

  void fixFineToCoarse(EBISLevel& a_fineEBIS);

  void coarsenFaces(EBISLevel& a_fineEBIS);

  void fineLevelStitch(EBISLevel& otherPhase); // MF Addition  bvs

  EBISLevel();
  ///
  /**
     Defines the level.  Calls the geoserver
     to fill the layout with geometric information.
   */
  EBISLevel(const Box& a_domain,
            const RealVect& a_origin,
            const Real& a_dx,
            const GeometryService& a_geoserver,
            const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

  ///
  /**
     Defines the level from a finer level
     (factor of two refinement ratio).
     Fills in all the coarse-fine information
     in both levels.
   */
  EBISLevel(EBISLevel& a_finerLevel,
            const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

#ifdef HDF5
  /// read from file
  EBISLevel(HDF5Handle& handle);

  ///
  void
  write(HDF5Handle& handle);
#endif

  ///
  /**
     Checks to see the vofs are in the correct cells.
     Checks to see that the faces are over the correct cells.
     Checks that volume fractions, area fractions are positive.
     Bail out with MayDay::Error if anything fails.
  */
  void sanityCheck(const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

  ///
  void fillEBISLayout(EBISLayout& a_ebis,
                      const DisjointBoxLayout& a_grids,
                      const int& nghost) const;
  ~EBISLevel();

  const Box& getDomain() const;

  const DisjointBoxLayout& getGrids() const;

  DisjointBoxLayout getIrregGrids() const;

  DisjointBoxLayout getFlowGrids() const;

  IntVectSet irregCells() const;

  const Real& getDX() const;

  const RealVect& getOrigin() const;

  ///
  void fixRegularNextToMultiValued();

  int m_phase;

private:

  ///
  DisjointBoxLayout m_grids;

  ///
  Box m_domain;

  ///
  RealVect m_origin;

  ///
  Real m_dx;

  ///
  LevelData<EBGraph> m_graph;
  
  static bool s_verbose;

  static Real s_tolerance;
  ///
  LevelData<EBData>  m_data;

  //make grids for this level.
  static void
  makeBoxes(Vector<Box>& a_boxes, 
            const Box& a_domain,
            const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance());

  //these are disallowed
  void operator=(const EBISLevel& ebiin)
  {
    MayDay::Error("EBISLevel::operator= not allowed");
  }

  EBISLevel(const EBISLevel& ebiin)
  {
    MayDay::Error("EBISLevel::copy constructor not allowed");
  }
};

/// 
/**
   EBIndexSpace represents the geometric information
   of the domain.    It should be generated through
   the Chombo_EBIS class.  This follows the singleton
   pattern. 
*/
class EBIndexSpace
{
public:

  EBIndexSpace();
  ///
  /**
      If a_ncellMax is set, that is the max width of
      an internal grid.  Otherwise use defaults
      of (16 in 3D, 64 in 2d)
   */
  void
  define(const Box& a_domain,
         const RealVect& a_origin,
         const Real& a_dx,
         const GeometryService& a_geoserver,
         int a_nCellMax = -1,
		 int maxCoarsenings = -1);

#ifdef HDF5
  ///
  void 
  define(HDF5Handle& handle, int maxCoarsenings = -1);
#endif

  void
  define(EBISLevel* level0, 
         int a_nCellMax = -1,
         int a_maxCoarsenings = -1);

  ///
  ~EBIndexSpace();

  ///
  int numLevels() const;

  ///
  /**
     return level index of domain.
     return -1 if a_domain does not
     correspond to any refinement of EBIS.
   */
  int getLevel(const Box& a_domain) const;

  ///
  /**
     returns the problem domain box at a level of refinement
  */
  const Box& getBox(int level)
  {
    return m_domainLevel[level];
  }

  ///
  void fillEBISLayout(EBISLayout& a_ebis,
                      const DisjointBoxLayout& a_grids,
                      const Box& a_domain,
                      const int & nghost) const;

  ///
  /**
     Return true if the define function has been called.
   */
  bool isDefined() const;

  ///
  void clear();

  int getNCellMax() const;
  ///
  /**
     This makes Chombo_EBIS be the only function
     that can call the private constructor.
   */
  friend class Chombo_EBIS;

#ifdef HDF5
  ///
  void
  write(HDF5Handle& handle);
#endif

  DisjointBoxLayout getIrregGrids(int level) const
  { return m_ebisLevel[level]->getIrregGrids();}

  DisjointBoxLayout getFlowGrids(int level) const
  { return m_ebisLevel[level]->getFlowGrids();}
  
  IntVectSet irregCells(int depth=0) const
  {
    return m_ebisLevel[depth]->irregCells();
  }

  void setCellMax(int max) {m_nCellMax = max;}

private:  
  //this is private to force the singleton thing.
  //EBIndexSpace();

  int  m_nCellMax;
  bool m_isDefined;

  Vector<EBISLevel*> m_ebisLevel;
  Vector<Box> m_domainLevel;
  int m_nlevels;
  static Real s_tolerance;
  static bool s_verbose;

  //disallowed for performance reasons
  void operator=(const EBIndexSpace& ebiin)
  {
    MayDay::Error("no assignment of EBIndexSpace allowed");
  }
  EBIndexSpace(const EBIndexSpace& ebiin)
  {
    MayDay::Error("no copy construction of EBIndexSpace allowed");
  }
};

#endif
