/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include "EBIndexSpace.H"
#include "parstream.H"
#include "LoadBalance.H"
#include "EBGraphFactory.H"
#include "EBDataFactory.H"
#include "BoxIterator.H"
#include "EBISLayout.H"
#include "LayoutIterator.H"
#include "BRMeshRefine.H"
#include "VoFIterator.H"
#include "IrregNode.H"
#include "memtrack.H"

Real EBIndexSpace::s_tolerance = 1.0e-12;
Real    EBISLevel::s_tolerance = 1.0e-12;
bool EBIndexSpace::s_verbose = false;
bool    EBISLevel::s_verbose = false;

/******************/
void
EBISLevel::makeBoxes(Vector<Box>& a_boxes, 
                     const Box& a_domain,
                     const EBIndexSpace* const ebisPtr)
{
  int ncellmax = ebisPtr->getNCellMax();
  int block_factor = 2;
  //need to be able to coarsen domain to break it up
  Box testdom = coarsen(a_domain, block_factor);
  testdom.refine(block_factor);
  bool dosplit = (a_domain == testdom);

  //domainsplit breaks unless all sides are big enough
  IntVect domsize = a_domain.size();
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(domsize[idir] <= ncellmax)
        dosplit = false;
    }
  
  if(dosplit)
    {
      domainSplit(a_domain, a_boxes, ncellmax, block_factor);
    }
  else
    {
      a_boxes.resize(1, a_domain);
    }
}
/******************/
#ifdef HDF5
EBISLevel::EBISLevel(HDF5Handle& a_handle)
{
  HDF5HeaderData header;
  header.readFromFile(a_handle);
  m_origin = header.m_realvect["EBIS_origin"];
  m_domain = header.m_box     ["EBIS_domain"];
  m_dx =     header.m_real    ["EBIS_dx"]    ;
  //read in the grids
  Vector<Box> boxes;
  read(a_handle,boxes);
  Vector<int> procAssign;
  LoadBalance(procAssign, boxes);
  m_grids.define(boxes, procAssign);
  EBGraphFactory graphfact(m_domain);
  m_graph.define(m_grids, 1, IntVect::Zero, graphfact);

  //read the graph  in from the file
  std::string graphName("EBIS_graph");
  int eekflag = read(a_handle, m_graph, graphName, m_grids, Interval(), false);
  if(eekflag != 0)
    {
      MayDay::Error("error in writing graph");
    }

  //need a ghosted layout so that the data can be defined properly
  LevelData<EBGraph> ghostGraph(m_grids, 1, IntVect::Unit, graphfact);
  Interval interv(0,0);
  m_graph.copyTo(interv, ghostGraph, interv);

  //now the data for the graph
  EBDataFactory dataFact;
  m_data.define(m_grids, 1, IntVect::Zero, dataFact);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    { 
      m_data[dit()].defineVoFData( ghostGraph[dit()], m_grids.get(dit()));
      m_data[dit()].defineFaceData(ghostGraph[dit()], m_grids.get(dit()));
    }
  //read the data  in from the file
  std::string  dataName("EBIS_data");
  eekflag = read(a_handle, m_data ,  dataName, m_grids, Interval(), false);
  if(eekflag != 0)
    {
      MayDay::Error("error in writing data");
    }
}
/******************/
void
EBISLevel::write(HDF5Handle& a_handle)
{
  HDF5HeaderData header;
  //this naming stuff kinda depends on the fact
  //that we are only outputting the finest level.
  //we could be slick and incorporate the
  //level number in there if we wanted.
  header.m_realvect["EBIS_origin"] = m_origin;
  header.m_box     ["EBIS_domain"] = m_domain;
  header.m_real    ["EBIS_dx"]     = m_dx;
  header.writeToFile(a_handle);
  //write the grids to the file
  ::write(a_handle, m_grids);

  std::string graphName("EBIS_graph");
  std::string  dataName("EBIS_data");
  int eekflag = ::write(a_handle, m_graph, graphName);
  if(eekflag != 0)
    {
      MayDay::Error("error in writing graph");
    }
  eekflag = ::write(a_handle, m_data ,  dataName);
  if(eekflag != 0)
    {
      MayDay::Error("error in writing data");
    }
}
#endif
/******************/
EBISLevel::EBISLevel(const Box& a_domain,
                     const RealVect& a_origin,
                     const Real& a_dx,
                     const GeometryService& a_geoserver,
                     const EBIndexSpace* const ebisPtr)
{

  m_domain = a_domain;
  m_dx = a_dx;
  m_origin = a_origin;
  //divide up the domain into a layout
  Vector<Box> vbox;
  makeBoxes(vbox, a_domain,ebisPtr);
  //load balance the boxes
  Vector<int> procAssign;
  LoadBalance(procAssign, vbox);
  m_grids.define(vbox, procAssign);

  //both domain and box set by the factory when leveldata construction is used
  EBGraphFactory graphfact(m_domain);
  m_graph.define(m_grids, 1, IntVect::Zero, graphfact);
  LayoutData<Vector<IrregNode> > allNodes(m_grids);

  //define the graph stuff
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& region = m_grids.get(dit());
      Box ghostRegion = grow(region,1);
      ghostRegion &= m_domain;

      EBGraph& ebgraph = m_graph[dit()];
      if(a_geoserver.isRegular(region, m_domain, m_origin, m_dx))
        {
          ebgraph.setToAllRegular();
        }
      else if(a_geoserver.isCovered(region, m_domain, m_origin, m_dx))
        {
          ebgraph.setToAllCovered();
        }
      else
        {
          BaseFab<int>       regIrregCovered(ghostRegion, 1);
          Vector<IrregNode>&  nodes = allNodes[dit()];
          a_geoserver.fillGraph(regIrregCovered, nodes, region, 
                                ghostRegion, m_domain,
                                m_origin, m_dx);

          ebgraph.buildGraph(regIrregCovered, nodes, region, m_domain);

        }
    }

  //need a ghosted graph to make data so we can define baseiffabs
  LevelData<EBGraph> ghostGraph(m_grids, 1, IntVect::Unit, graphfact);
  Interval interv(0,0);
  m_graph.copyTo(interv, ghostGraph, interv);
  //now the data for the graph
  EBDataFactory dataFact;
  m_data.define(m_grids, 1, IntVect::Zero, dataFact);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    { 
      m_data[dit()].define(ghostGraph[dit()], allNodes[dit()], m_grids.get(dit()));
    }

  //
  fixRegularNextToMultiValued();
}
/************/
//now fix the multivalued next to regular thing for the graph and the data
//the oldgraph/newgraph thing is necessary because the graphs are
//reference counted and they have to be kept consistent with the data
/***********/
void EBISLevel::
fixRegularNextToMultiValued()
{
  EBGraphFactory graphfact(m_domain);
  LayoutData<IntVectSet> vofsToChange(m_grids);
  LevelData<EBGraph> oldGhostGraph(m_grids, 1, IntVect::Unit, graphfact);
  Interval interv(0,0);
  m_graph.copyTo(interv, oldGhostGraph, interv);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      m_graph[dit()].getRegNextToMultiValued(vofsToChange[dit()], 
                                             oldGhostGraph[dit()]);

      m_graph[dit()].addFullIrregularVoFs(vofsToChange[ dit()], 
                                          oldGhostGraph[dit()]);
    }

  LevelData<EBGraph> newGhostGraph(m_grids, 1, IntVect::Unit, graphfact);
  m_graph.copyTo(interv, newGhostGraph, interv);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      m_data[dit() ].addFullIrregularVoFs(vofsToChange[dit()], 
                                          newGhostGraph[dit()],
                                          oldGhostGraph[dit()]);
    }
}
/******************/
//checks to see the vofs are in the correct cells.
//checks to see that the faces are over the correct cells
//checks that volume fractions, area fractions are positive
//bail out with MayDay::Error if anything fails
/******************/
void
EBISLevel::sanityCheck(  const EBIndexSpace* const ebisPtr)
{
#if 0
  EBISLayout ghostLayout;


  ebisPtr->fillEBISLayout(ghostLayout, m_grids, m_domain, 1);

  Real maxcentval = 0.5+ s_tolerance;
  for(DataIterator dit = m_grids.dataIterator();  dit.ok(); ++dit)
    {
      const Box& thisBox = m_grids.get(dit());
      const EBISBox& ebisBox = ghostLayout[dit()];
      for(BoxIterator bit(thisBox); bit.ok(); ++bit)
        {
          const IntVect& iv = bit();
          Vector<VolIndex> vofs = ebisBox.getVoFs(iv);
          for(int ivof = 0; ivof < vofs.size(); ivof++)
            {
              const VolIndex& vof = vofs[ivof];
              if(vof.gridIndex() != iv)
                {
                  pout() << "EBISLevel::sanityCheck: Error" << endl;
                  pout() << "VoF at Intvect = " << iv 
                         << "has grid index = " << vof.gridIndex() << endl;
                  MayDay::Error("EBISLevel::sanityCheck Error 1");
                }
              if(vof.cellIndex() < 0)
                {
                  pout() << "EBISLevel::sanityCheck: Error" << endl;
                  pout() << "VoF at Intvect = " << iv 
                         << "has negative cell index = " << vof.cellIndex() << endl;
                  MayDay::Error("EBISLevel::sanityCheck Error 2");
                }
              Real volFrac = ebisBox.volFrac(vof);
              if(volFrac < 0.0)
                {
                  pout() << "EBISLevel::sanityCheck: Error" << endl;
                  pout() << "VoF at Intvect = " << iv 
                         << "has invalid volume fraction = " << volFrac << endl;
                  MayDay::Error("EBISLevel::sanityCheck Error 5");
                }
              RealVect volCentroid = ebisBox.centroid(vof);
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  Real volcentdir = volCentroid[idir];
                  if(volFrac > s_tolerance)
                    {
                      if(volcentdir > maxcentval || volcentdir < -maxcentval)
                        {
                          pout() << "EBISLevel::sanityCheck: Error" << endl;
                          pout() << "VoF at Intvect = " << iv 
                                 << " has invalid vol centroid = " << volcentdir 
                                 << " at direction "<< idir << endl;
                          MayDay::Error("EBISLevel::sanityCheck Error 51");
                        }
                    }
                }
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  for(SideIterator sit; sit.ok(); ++sit)
                    {
                      Vector<FaceIndex> faces = ebisBox.getFaces(vof, idir, sit());
                      IntVect iv2 = iv + sign(sit())*BASISV(idir);
                      ////check for regular next to covered and multivalued next to covered
                      if(m_domain.contains(iv2))
                        {
                          if(ebisBox.isRegular(iv)) 
                            {
                              if(ebisBox.isCovered(iv2))
                                {
                                  pout() << iv << " is regular and " <<  iv2 << " is covered" << endl;
                                  MayDay::Error("EBISLevel::sanityCheck error 420 " );
                                }
                              else
                                {
                                  Vector<VolIndex> otherVoFs = ebisBox.getVoFs(iv2);
                                  if(otherVoFs.size() > 1)
                                    {
                                      pout() << iv << " is regular and " <<  iv2 << " is multivalued" << endl;
                                      MayDay::Error("EBISLevel::sanityCheck error 420.2 " );
                                    }
                                }
                            }
                        }
                      IntVect ivlo, ivhi;
                      if(sit() == Side::Lo)
                        {
                          ivlo = iv2;
                          ivhi = iv;
                        }
                      else
                        {
                          ivlo = iv;
                          ivhi = iv2;
                        }
                      for(int iface = 0; iface < faces.size(); iface++)
                        {
                          const FaceIndex& face = faces[iface];
                          if(face.gridIndex(Side::Lo) != ivlo)
                            {
                              pout() << "EBISLevel::sanityCheck: Error" << endl;
                              pout() << "face at IntVects = " << ivlo << "  " << ivhi
                                     << "has low IntVect  = " << face.gridIndex(Side::Lo) 
                                     << endl;
                              MayDay::Error("EBISLevel::sanityCheck Error 3");
                            }
                          if(face.gridIndex(Side::Hi) != ivhi)
                            {
                              pout() << "EBISLevel::sanityCheck: Error" << endl;
                              pout() << "face at IntVects = " << ivlo << "  " << ivhi
                                     << "has high IntVect = " << face.gridIndex(Side::Hi) 
                                     << endl;
                              MayDay::Error("EBISLevel::sanityCheck Error 4");
                            }
                          Real areaFrac = ebisBox.areaFrac(face);
                          if(areaFrac  < 0.0)
                            {
                              pout() << "EBISLevel::sanityCheck: Error" << endl;
                              pout() << "VoF at Intvect = " << iv 
                                     << "has invalid area fraction = " << areaFrac << endl;
                              MayDay::Error("EBISLevel::sanityCheck Error 51");
                            }
                          if(areaFrac  >  s_tolerance)
                            {
                              RealVect faceCentroid = ebisBox.centroid(face);
                              for(int idir = 0; idir < SpaceDim; idir++)
                                {
                                  Real facecentdir = faceCentroid[idir];
                                  if(facecentdir > maxcentval || facecentdir < -maxcentval)
                                    {
                                      pout() << "EBISLevel::sanityCheck: Error" << endl;
                                      pout() << "VoF at Intvect = " << iv 
                                             << " has invalid face centroid = " << facecentdir 
                                             << " at direction "<< idir << endl;
                                      MayDay::Error("EBISLevel::sanityCheck Error 51");
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
#endif
}

EBISLevel::EBISLevel()
{
}
/******************/
//steps to coarsen an ebislevel:
//1. coarsen vofs
//1a. make a layout over refine(mydbl,2) and copy
//    fine layout into it
//1b.do connectivity bizbaz to make my vofs, volfrac, vof->fineVofs
//2. make faces doing connectivity jive
//2.23 make coarse geometric stuff (centroids and all that) from fine
//3. make coarse layout from coarsen(finelayout). and copy my data into it
//   to make fine->coarserVoF 
// (so finer ebislevel does change in this function)
/******************/
void
EBISLevel::
coarsenVoFs(EBISLevel& a_fineEBIS)
{
  //so that i can do the vofs and faces of this level.
  DisjointBoxLayout fineFromCoarDBL;
  refine(fineFromCoarDBL, m_grids, 2);
  fineFromCoarDBL.close();
  
  //no need for ghost cells here except to define the face data
  //you need the graph to be one bigger
  EBGraphFactory ebgraphfactFine(a_fineEBIS.m_domain);
  LevelData<EBGraph> fineFromCoarEBGraph(fineFromCoarDBL,1, IntVect::Unit, ebgraphfactFine);

  Interval interv(0,0);
  a_fineEBIS.m_graph.copyTo(interv, fineFromCoarEBGraph, interv);

  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBGraph& fineEBGraph = fineFromCoarEBGraph[dit()];
      const Box& coarRegion      = m_grids.get(dit());
      EBGraph& coarEBGraph = m_graph[dit()];
      coarEBGraph.coarsenVoFs(fineEBGraph, coarRegion);
    }
  EBGraphFactory ebgraphfactCoar(m_domain);
  LevelData<EBGraph> coarGhostEBGraph(m_grids,1, IntVect::Unit, ebgraphfactCoar);
  m_graph.copyTo(interv, coarGhostEBGraph, interv);

  EBDataFactory ebdatafact;
  LevelData<EBData> fineFromCoarEBData(fineFromCoarDBL,1, IntVect::Zero, ebdatafact);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const Box& localBox = fineFromCoarDBL.get(dit());
      fineFromCoarEBData[dit()].defineVoFData(fineFromCoarEBGraph[dit()],  localBox);
      fineFromCoarEBData[dit()].defineFaceData(fineFromCoarEBGraph[dit()], localBox);
    }
  a_fineEBIS.m_data.copyTo(interv, fineFromCoarEBData, interv);
  
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBGraph& fineEBGraph =  fineFromCoarEBGraph[dit()];
      const EBData& fineEBData = fineFromCoarEBData[dit()];
      const EBGraph& coarEBGraph = coarGhostEBGraph[dit()];

      m_data[dit()].coarsenVoFs(fineEBData, fineEBGraph, coarEBGraph, m_grids.get(dit()));
    }
}
/******************/
void
EBISLevel::
fixFineToCoarse(EBISLevel& a_fineEBIS)
{
  // make a coarse layout from the fine layout so that we 
  //can fix the fine->coarseVoF thing.
  DisjointBoxLayout coarFromFineDBL;
  
  coarsen(coarFromFineDBL, a_fineEBIS.m_graph.getBoxes(), 2);
  coarFromFineDBL.close();
  EBGraphFactory ebgraphfact(m_domain);
  LevelData<EBGraph> coarFromFineEBGraph(coarFromFineDBL,1, IntVect::Zero, ebgraphfact);
  Interval interv(0,0);
  m_graph.copyTo(interv, coarFromFineEBGraph, interv);

  for(DataIterator dit = a_fineEBIS.m_grids.dataIterator(); dit.ok(); ++dit)
    {
      EBGraph& fineEBGraph       =  a_fineEBIS.m_graph[dit()];
      const EBGraph& coarEBGraph = coarFromFineEBGraph[dit()];

      coarEBGraph.fixFineToCoarse(fineEBGraph);
    }

}
/******************/
void
EBISLevel::
coarsenFaces(EBISLevel& a_fineEBIS)
{
  //now make a fine ebislayout with two ghost cell
  //on the same mapping as m_dbl
  //so that i can do the vofs and faces of this level.
  //this one will have the fine from coarse stuff fixed
  DisjointBoxLayout fineFromCoarDBL;
  refine(fineFromCoarDBL, m_grids, 2);
  fineFromCoarDBL.close();

  //no need for ghost cells here
  EBGraphFactory ebgraphfactfine(a_fineEBIS.m_domain);
  EBGraphFactory ebgraphfactcoar(m_domain);
  LevelData<EBGraph> fineEBGraphGhostLD(fineFromCoarDBL,1,3*IntVect::Unit, ebgraphfactfine);
  Interval interv(0,0);
  a_fineEBIS.m_graph.copyTo(interv, fineEBGraphGhostLD, interv);
  LevelData<EBGraph> coarEBGraphGhostLD(m_grids,        1,  IntVect::Unit, ebgraphfactcoar);
  m_graph.copyTo(           interv, coarEBGraphGhostLD, interv);

  for(DataIterator dit =  m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBGraph& fineEBGraphGhost = fineEBGraphGhostLD[dit()];
      const EBGraph& coarEBGraphGhost = coarEBGraphGhostLD[dit()];
      EBGraph& coarEBGraph = m_graph[dit()];
      coarEBGraph.coarsenFaces(coarEBGraphGhost, fineEBGraphGhost);
    }
  //redefine coarebghostgraphld so i can use the faces for the ebdata
  coarEBGraphGhostLD.define(m_grids, 1,  IntVect::Unit, ebgraphfactcoar);
  m_graph.copyTo( interv, coarEBGraphGhostLD, interv);

  EBDataFactory ebdatafact;
  LevelData<EBData> fineEBDataGhostLD(fineFromCoarDBL,1, 2*IntVect::Unit, ebdatafact);
  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      Box localBox = grow(fineFromCoarDBL.get(dit()), 2);
      localBox &= a_fineEBIS.m_domain;
      fineEBDataGhostLD[dit()].defineVoFData( fineEBGraphGhostLD[dit()], localBox);;
      fineEBDataGhostLD[dit()].defineFaceData(fineEBGraphGhostLD[dit()], localBox);
    }
  a_fineEBIS.m_data.copyTo(interv, fineEBDataGhostLD, interv);

  for(DataIterator dit = m_grids.dataIterator(); dit.ok(); ++dit)
    {
      const EBData&   fineEBData      = fineEBDataGhostLD[dit()];
      const EBGraph& fineEBGraphGhost = fineEBGraphGhostLD[dit()];
      const EBGraph& coarEBGraphGhost = coarEBGraphGhostLD[dit()];

      EBData& coarEBData   = m_data[dit()];
      coarEBData.coarsenFaces(fineEBData,  fineEBGraphGhost, coarEBGraphGhost, m_grids.get(dit()));
    }

}
/******************/
EBISLevel::EBISLevel(EBISLevel& a_fineEBIS,
                     const EBIndexSpace* const ebisPtr)
{
  m_domain = coarsen(a_fineEBIS.m_domain,2);
  m_dx = 2.*a_fineEBIS.m_dx;
  m_origin = a_fineEBIS.m_origin;
  Vector<Box> vbox;
  makeBoxes(vbox, m_domain, ebisPtr);
  //load balance the boxes
  Vector<int> procAssign;
  LoadBalance(procAssign, vbox);
  //define the layout.  this includes the domain and box stuff
  m_grids.define(vbox, procAssign);

  EBGraphFactory ebgraphfact(m_domain);
  m_graph.define(m_grids, 1, IntVect::Zero, ebgraphfact);

  EBDataFactory ebdatafact;
  m_data.define(m_grids, 1, IntVect::Zero, ebdatafact);

  //create coarsened vofs from fine.  
  coarsenVoFs(a_fineEBIS);

  //create coarse faces from fine
  coarsenFaces(a_fineEBIS);

  //fix the regular next to the multivalued cells
  //to be full irregular cells
  fixRegularNextToMultiValued();

  // fix the fine->coarseVoF thing.
  fixFineToCoarse(a_fineEBIS);
}
/******************/
EBISLevel::~EBISLevel()
{
}
/****************/
void 
EBISLevel::fillEBISLayout(EBISLayout& a_ebisLayout,
                          const DisjointBoxLayout& a_grids,
                          const int& a_nghost) const 
{
  assert(a_nghost >= 0);
  a_ebisLayout.define(m_domain, a_grids, a_nghost, m_graph, m_data);
}

void 
EBISLevel::fineLevelStitch(EBISLevel& otherPhase)
{
  DataIterator dit = m_grids.dataIterator();
  DataIterator dit2= otherPhase.m_grids.dataIterator();
  dit.begin(); dit2.begin();
  for(;dit.ok(); ++dit, ++dit2)
    {
      EBGraph& grapha = m_graph[dit];
      EBGraph& graphb = otherPhase.m_graph[dit2];
      EBData& a = m_data[dit];
      EBData& b = otherPhase.m_data[dit2];
      Box region = grapha.getRegion();
      assert(graphb.getRegion() == region);
      assert(grapha.getMultiCells(region).isEmpty());
      assert(graphb.getMultiCells(region).isEmpty());
      IntVectSet seta = grapha.getIrregCells(region);
      IntVectSet setb = graphb.getIrregCells(region);
      IntVectSet set = seta & setb;
      IVSIterator it(set);
      for(it.begin(); it.ok(); ++it)
        {
          VolIndex v;
          v.define(it(), 0);
          // in final code, the logic here would have to determine which
          // face of either volume match up.  in this case we can just use 0
          // since we are just doing two-phase flow at this point.
          a.setFacePhase(v,0,otherPhase.m_phase);
          b.setFacePhase(v,0,m_phase);
        }
          
    }

}


/******************/
bool 
EBIndexSpace::isDefined() const
{
  return m_isDefined;
}
/******************/
int 
EBIndexSpace::getNCellMax() const
{
  return m_nCellMax;
}
/******************/
#ifdef HDF5
void
EBIndexSpace::define(HDF5Handle& a_handle, int a_maxCoarsenings)
{
  clear();

  //read in ncellmax 
  HDF5HeaderData header;
  header.readFromFile(a_handle);
  int nCellMax = header.m_int["EBIS_ncellMax"];

  //read in finest level
  //coarser levels are derived from graph coarsening
  EBISLevel* level0 = new EBISLevel(a_handle);

  define(level0, nCellMax, a_maxCoarsenings);
}
/******************/
void
EBIndexSpace::write(HDF5Handle& a_handle)
{
  //read in ncellmax 
  HDF5HeaderData header;
  std::string filedescriptor("EBIndexSpace");
  header.m_string ["filetype"]      = filedescriptor;
  header.m_int["EBIS_ncellMax"] = m_nCellMax;
  header.writeToFile(a_handle);

  //write out finest level
  //coarser levels are derived from graph coarsening
  m_ebisLevel[0]->write(a_handle);
}


#endif

void
EBIndexSpace::define(EBISLevel* level0, int nCellMax, int a_maxCoarsenings)
{
  m_nCellMax = nCellMax;
  m_isDefined = true;
  const Box&        fineDomain = level0->getDomain();

  //figure out how deep we can go
  m_nlevels = 1;
  bool canref = (fineDomain == refine(coarsen(fineDomain,2), 2));
  assert(!fineDomain.isEmpty());
  Box refbox = fineDomain;
  while(canref)
    {
      Box refcoarbox = coarsen(refbox,2);
      refcoarbox.refine(2);
      if(refcoarbox != refbox)
        {
          canref = false;
        }
      else
        {
          m_nlevels++;
          refbox.coarsen(2);
        }
    }
  if(a_maxCoarsenings != -1)
    {
      assert(a_maxCoarsenings >= 0);
      if(m_nlevels > a_maxCoarsenings+1) m_nlevels = a_maxCoarsenings + 1;
    }
  
  m_ebisLevel.resize(m_nlevels, NULL);
  m_domainLevel.resize(m_nlevels);

  Box  domLevel = fineDomain;
  m_ebisLevel[0] = level0;
  m_domainLevel[0] = domLevel;
  for(int ilev = 1; ilev < m_nlevels; ilev++)
    {
      domLevel.coarsen(2);
      m_domainLevel[ilev] = domLevel;
      m_ebisLevel[ilev] = new EBISLevel(*m_ebisLevel[ilev-1], this);
    }

#ifndef NDEBUG
  for(int ilev = 0; ilev < m_nlevels; ilev++)
    {
      m_ebisLevel[ilev]->sanityCheck(this);
    }
#endif
}

/******************/
void
EBIndexSpace::define(const Box& a_domain,
                     const RealVect& a_origin,
                     const Real& a_dx,
                     const GeometryService& a_geoserver,
                     int a_nCellMax,
                     int a_maxCoarsenings)
{
  clear();
  m_isDefined = true;

  if(a_nCellMax > 0)
    {
      m_nCellMax = a_nCellMax;
    }
  else
    {
      if(SpaceDim == 2)
        {
          m_nCellMax = 64;
        }
      else
        {
          m_nCellMax = 16;
        }
    }
  m_nlevels = 1;
  bool canref = (a_domain == refine(coarsen(a_domain,2), 2));
 
  assert(!a_domain.isEmpty());
  Box refbox = a_domain;

  while(canref)
    {
      Box refcoarbox = coarsen(refbox,2);
      refcoarbox.refine(2);
      if(refcoarbox != refbox)
        {
          canref = false;
        }
      else
        {
          m_nlevels++;
          refbox.coarsen(2);
        }
    }
  if(a_maxCoarsenings != -1)
    {
      assert(a_maxCoarsenings >= 0);
      if(m_nlevels > a_maxCoarsenings+1) m_nlevels = a_maxCoarsenings + 1;
    }
  
  m_ebisLevel.resize(m_nlevels, NULL);
  m_domainLevel.resize(m_nlevels);

  Box  domLevel = a_domain;
  m_ebisLevel[0] = new EBISLevel(domLevel, a_origin,
                                 a_dx, a_geoserver, this);
  m_domainLevel[0] = domLevel;
  for(int ilev = 1; ilev < m_nlevels; ilev++)
    {
      domLevel.coarsen(2);
      m_domainLevel[ilev] = domLevel;
      m_ebisLevel[ilev] = new EBISLevel(*m_ebisLevel[ilev-1]);
    }

#ifndef NDEBUG
  m_ebisLevel[0]->sanityCheck();
  for(int ilev = 1; ilev < m_nlevels; ilev++)
    {
      m_ebisLevel[ilev]->sanityCheck();
    }
#endif

}
/******************/
const Box& 
EBISLevel::getDomain() const
{
  return m_domain;
}
/******************/
const Real& 
EBISLevel::getDX() const
{
  return m_dx;
}
/******************/
const RealVect& 
EBISLevel::getOrigin() const
{
  return m_origin;
}
/******************/
const DisjointBoxLayout& 
EBISLevel::getGrids() const
{
  return m_grids;
}

DisjointBoxLayout
EBISLevel::getIrregGrids() const
{
  DisjointBoxLayout irregGrids;
  DataIterator dit =   m_graph.dataIterator();
  Vector<Box> localBoxes;
 
  
  for(dit.begin(); dit.ok(); ++dit)
    {
      const EBGraph& graph = m_graph[dit()];
      const Box& b         = m_graph.box(dit());
      if(graph.hasIrregular())
        {
          localBoxes.push_back(b);

        }
    }
  Vector<Vector<Box> > allBoxes;

  gather(allBoxes, localBoxes, 0);
 
  broadcast(allBoxes, 0);
 

  for(int i=0; i<allBoxes.size(); i++)
    {
      int num=allBoxes[i].size();
      Vector<Box>& boxes = allBoxes[i];
      for(int j=0; j<num; j++) irregGrids.addBox(boxes[j], i);
    }

  irregGrids.close();
  return irregGrids; 
}


DisjointBoxLayout
EBISLevel::getFlowGrids() const
{
  DisjointBoxLayout flowGrids;
  DataIterator dit =   m_graph.dataIterator();
  Vector<Box> localBoxes;
 
  
  for(dit.begin(); dit.ok(); ++dit)
    {
      const EBGraph& graph = m_graph[dit()];
      const Box& b         = m_graph.box(dit());
      if(graph.hasIrregular() || graph.isAllRegular())
        {
          localBoxes.push_back(b);

        }
    }
  Vector<Vector<Box> > allBoxes;

  gather(allBoxes, localBoxes, 0);
 
  broadcast(allBoxes, 0);
 

  for(int i=0; i<allBoxes.size(); i++)
    {
      int num=allBoxes[i].size();
      Vector<Box>& boxes = allBoxes[i];
      for(int j=0; j<num; j++) flowGrids.addBox(boxes[j], i);
    }

  flowGrids.close();
  return flowGrids; 
}

IntVectSet
EBISLevel::irregCells() const
{
  DataIterator dit =   m_graph.dataIterator();
  IntVectSet rtn;

  for(dit.begin(); dit.ok(); ++dit)
    {
      const EBGraph& graph = m_graph[dit()];
      const Box& b         = m_graph.box(dit());
      if(graph.hasIrregular())
        {
          rtn |= graph.getIrregCells(b);
        }
    }
  return rtn;
}


/******************/
int 
EBIndexSpace::numLevels() const
{
  return m_nlevels;
}

void
EBIndexSpace::clear()
{
  for(int ilev = 0; ilev < m_ebisLevel.size(); ilev++)
    {
      delete m_ebisLevel[ilev];
      m_ebisLevel[ilev] = NULL;
    }
  m_ebisLevel.resize(0);
  m_domainLevel.resize(0);
  m_nlevels = 0;
  m_isDefined = false;
}
/******************/
EBIndexSpace::EBIndexSpace()
{
}
/******************/
EBIndexSpace::~EBIndexSpace()
{
  clear();
}
/******************/
int 
EBIndexSpace::getLevel(const Box& a_domain) const
{
  bool found = false;
  int whichlev = -1;
  for(int ilev = 0; ilev < m_domainLevel.size() && !found; ilev++)
    {
      if(m_domainLevel[ilev] == a_domain)
        {
          found = true;
          whichlev = ilev;
        }
    }
  return whichlev;
}
/****************/
void 
EBIndexSpace::fillEBISLayout(EBISLayout& a_ebisLayout,
                             const DisjointBoxLayout& a_grids,
                             const Box& a_domain,
                             const int& a_nghost) const 
{
  //figure out which level we are on
  int whichlev = getLevel(a_domain);
  if(whichlev < 0)
    {
      pout() << "a_domain = " << a_domain 
             << "does not correspond to any refinement of EBIS" << endl;
      MayDay::Error("Bad argument to EBIndexSpace::fillEBISLayout");
    }
  m_ebisLevel[whichlev]->fillEBISLayout(a_ebisLayout, a_grids, a_nghost);
}
/****************/
EBIndexSpace* Chombo_EBIS::s_instance = NULL;
/****************/
EBIndexSpace* 
Chombo_EBIS::instance()
{
  if(s_instance == NULL)
    s_instance = new EBIndexSpace();

  return  s_instance;
}
/****************/
