/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _FACEINDEX_H_
#define _FACEINDEX_H_

#include "LoHiSide.H"
#include "VolIndex.H"
#include "SPMD.H"

/// 
/**
   FaceIndex is a very lightweight object used
   to distinguish faces.  It has two VolIndex (s)
   and a direction and a way to access them.
*/
class FaceIndex
{
public:
  ///
  /**
     if both cellindexes of the vofs are >= 0, 
     then the face is interior.  otherwise the face is
     a boundary face.
   */
  FaceIndex(const VolIndex& a_vof1, 
            const VolIndex& a_vof2, 
            const int& a_direction); 

  ///
  /**
     if both cellindexes of the vofs are >= 0, 
     then the face is interior.  otherwise the face is
     a boundary face.
   */
  void define(const VolIndex& a_vof1, 
              const VolIndex& a_vof2, 
              const int& a_direction); 

  ///
  /**
   */
  FaceIndex();

  ///
  /**
   */
  FaceIndex(const FaceIndex& a_facein);

  ///
  /**
   */
  void define(const FaceIndex& a_facein);

  ///
  /**
   */
  ~FaceIndex();

  /// 
  /**
   */
  FaceIndex& operator= (const FaceIndex& a_facein);

  /// 
  /**
   */
  bool operator== (const FaceIndex& a_facein) const;

  /// 
  /**
   */
  bool operator!= (const FaceIndex& a_facein) const;

  ///
  /**
   */
  const int& direction() const;

  ///
  /**
   */
  const bool& isDefined() const;

  ///
  /**
   */
  const bool& isBoundary() const;

  ///
  /**
     return cell index of the vof on the a_sd side.
     returns -1 if that side of the face is outside
     the domain (boundary face case).
   */
  const int& cellIndex(const Side::LoHiSide& a_sd) const;

  ///
  /**
     return grid index of the vof on the a_sd side.
     may be inside or ouside domain.
   */
  const IntVect& gridIndex(const Side::LoHiSide& a_sd) const;

  ///
  /**
     manufactures the appropriate volindex.
   */
  VolIndex getVoF(const Side::LoHiSide& a_sd) const;

  // not user functions
  static size_t lo_offset, hi_offset, rest_offset;
  static int initializeOffsets();
private:

  ///the intvect of the low vof
  /**
   */
  IntVect m_loiv;

  ///the intvect of the high vof
  /**
   */
  IntVect m_hiiv;

  ///
  /**
     The index of the low vof.
     This = -1 if the face is set to be a boundary
     face on the low side of the domain.
  */
  int m_loIndex;

  ///
  /**
     The index of the high vof.
     This = -1 if the face is set to be a boundary
     face on the high side of the domain.
  */
  int m_hiIndex;

  ///direction of the face.
  int m_direction;

  ///
  /**
     true if one of the define functions has been called.
  */
  bool m_isDefined;

  ///
  /**
     true if the face is constructed
     using the boundary face constructor
  */
  bool m_isBoundary;

};

template < >
int linearSize(const FaceIndex& findex);

//FaceIndex specialization of linearIn
template < >
void linearIn(FaceIndex& a_outputT, const void* const inBuf); 

//FaceIndex specialization of linearOut
template < >
void linearOut(void* const a_outBuf, const FaceIndex& a_inputT); 


#endif
