/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#include "FaceIndex.H"
#include "MayDay.H"
#include "CH_HDF5.H"

using std::cerr;
using std::cout;
using std::endl;


size_t FaceIndex::lo_offset = HOFFSET(FaceIndex, m_loiv.vect);
size_t FaceIndex::hi_offset = HOFFSET(FaceIndex, m_hiiv.vect);
size_t FaceIndex::rest_offset = HOFFSET(FaceIndex, m_loIndex);

/*
int FaceIndex::initializeOffsets()
{
  lo_offset  = offsetof(FaceIndex, m_loiv.vect[0]);
  hi_offset  = offsetof(FaceIndex, m_hiiv.vect[0]);
  rest_offset= offsetof(FaceIndex, m_loIndex);
  return 1;
}


static int init = FaceIndex::initializeOffsets();
*/

/*****************************************/
FaceIndex::FaceIndex()
{
  m_loiv = IntVect::Zero;
  m_hiiv = IntVect::Zero;
  m_loIndex = -7;
  m_hiIndex = -7;
  m_direction = -7;
  m_isDefined = false;
  m_isBoundary = false;

}
/*****************************************/
const bool&
FaceIndex::isDefined() const
{
  return m_isDefined;
}
/******************************/
const int& 
FaceIndex::direction() const
{
  return m_direction;
}

/*****************************************/
const bool&
FaceIndex::isBoundary() const
{
  return m_isBoundary;
}
/*****************************************/
FaceIndex::FaceIndex(const FaceIndex& a_facein)
{
  define(a_facein);
}
/*****************************************/
void
FaceIndex::define(const FaceIndex& a_facein)
{
  //must be allowed to propogate undefinednitude
  //because of vector<vol>
  m_isDefined = a_facein.m_isDefined;
  m_isBoundary = a_facein.m_isBoundary;
  m_loiv = a_facein.m_loiv;
  m_hiiv = a_facein.m_hiiv;
  m_loIndex = a_facein.m_loIndex;
  m_hiIndex = a_facein.m_hiIndex;
  m_direction = a_facein.m_direction;
}

/*****************************************/
FaceIndex::FaceIndex(const VolIndex& a_vof1, 
                     const VolIndex& a_vof2,
                     const int& a_direction)
{
  define(a_vof1, a_vof2, a_direction);
}
	 
/*****************************************/
void
FaceIndex::define(const VolIndex& a_vof1, 
                  const VolIndex& a_vof2,
                  const int& a_direction)
{
  assert((a_direction >= 0) && 
         (a_direction < SpaceDim));
  //one vof or the other has to be in the domain.
  assert((a_vof1.cellIndex() >= 0) || (a_vof2.cellIndex() >= 0));
    
  
  //if either vof is outside the domain--shown by a negative
  //cell index, the face is a bounary face
  m_isBoundary = 
    ((a_vof1.cellIndex() < 0) || (a_vof2.cellIndex() < 0));
  m_isDefined  = true;

  const IntVect& iv1 = a_vof1.gridIndex();
  const IntVect& iv2 = a_vof2.gridIndex();

  int isign = 1;
  if(iv1[a_direction] < iv2[a_direction])
    {
      m_loiv = iv1;
      m_hiiv = iv2;
      m_loIndex = a_vof1.cellIndex();
      m_hiIndex = a_vof2.cellIndex();
      isign = -1;
    }
  else
    {
      m_loiv = iv2;
      m_hiiv = iv1;
      m_loIndex = a_vof2.cellIndex();
      m_hiIndex = a_vof1.cellIndex();
      isign =  1;
    }

  m_direction = a_direction;
  //this check needs to go away in the land
  //of periodic boundary conditions
  if((isign*(iv1-iv2)) != BASISV(a_direction))
    {
      cerr << "FaceIndex constructor--bad arguments" << endl;
      cerr << "iv1, iv2, direction =" << iv1 << iv2 << a_direction << endl;
      MayDay::Error("FaceIndex constructor--bad arguments");
    }
}


/*****************************************/
const int& 
FaceIndex::cellIndex(const Side::LoHiSide& a_sd) const
{
  assert(isDefined());
  const int* retval;
  if(a_sd == Side::Lo)
    retval=  &m_loIndex;
  else
    retval=  &m_hiIndex;

  return *retval;
}
/*****************************************/
const IntVect& 
FaceIndex::gridIndex(const Side::LoHiSide& a_sd) const
{
  assert(isDefined());
  const IntVect* retval;
  if(a_sd == Side::Lo)
    retval=  &m_loiv;
  else
    retval=  &m_hiiv;

  return *retval;
}
/*****************************************/
VolIndex
FaceIndex::getVoF(const Side::LoHiSide& a_sd) const
{
  assert(isDefined());
  VolIndex retval;
  if(a_sd == Side::Lo)
    retval=  VolIndex(m_loiv, m_loIndex);
  else
    retval=  VolIndex(m_hiiv, m_hiIndex);

  return retval;
}
/*****************************************/
FaceIndex::~FaceIndex()
{;}

/*****************************************/
FaceIndex& 
FaceIndex::operator= (const FaceIndex& a_facein)
{
  if(&a_facein != this)
    {
      //assert(a_facein.isDefined());
      //must be allowed to propogate undefinednitude
      //because of vector<vol>
      m_isDefined = a_facein.m_isDefined;
      m_isBoundary = a_facein.m_isBoundary;
      m_loiv = a_facein.m_loiv;
      m_hiiv = a_facein.m_hiiv;
      m_loIndex = a_facein.m_loIndex;
      m_hiIndex = a_facein.m_hiIndex;
      m_direction = a_facein.m_direction;
    }
  return *this;
}


/******************************/
bool FaceIndex::operator==(const FaceIndex& a_facein) const
{
  return((m_loiv == a_facein.m_loiv) &&
         (m_hiiv == a_facein.m_hiiv) &&
         (m_loIndex == a_facein.m_loIndex) &&
         (m_hiIndex == a_facein.m_hiIndex));
}

/******************************/
bool FaceIndex::operator!=(const FaceIndex& a_facein) const
{

  return (!(*this == a_facein));
}

template < >
int linearSize<FaceIndex>(const FaceIndex& findex)
{
  return sizeof(FaceIndex);
}
//FaceIndex specialization of linearIn
template < >
void linearIn<FaceIndex>(FaceIndex& a_outputT, const void* const inBuf)
{
  unsigned char* bob = (unsigned char*)inBuf;
  unsigned char* to = (unsigned char*)&a_outputT;

  memcpy(to + FaceIndex::lo_offset, bob, SpaceDim*sizeof(int));
  bob += SpaceDim*sizeof(int);
  memcpy(to + FaceIndex::hi_offset, bob, SpaceDim*sizeof(int));
  bob += SpaceDim*sizeof(int);
  memcpy(to + FaceIndex::rest_offset, bob, sizeof(FaceIndex)-FaceIndex::rest_offset);
}

//FaceIndex specialization of linearOut
template < >
void linearOut<FaceIndex>(void* const a_outBuf, const FaceIndex& a_inputT)
{
  unsigned char* bob = (unsigned char*)a_outBuf;
  const unsigned char* from = (const unsigned char*)&a_inputT;

  memcpy(bob, from + FaceIndex::lo_offset, SpaceDim*sizeof(int));
  bob += SpaceDim*sizeof(int);
  memcpy(bob, from + FaceIndex::hi_offset, SpaceDim*sizeof(int));
  bob += SpaceDim*sizeof(int);
  memcpy(bob,  from + FaceIndex::rest_offset, sizeof(FaceIndex)-FaceIndex::rest_offset);
}
