/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _EBGRAPHNODE_H_
#define _EBGRAPHNODE_H_

#include "REAL.H"
#include "IntVect.H"
#include "VolIndex.H"
#include "FaceIndex.H"
#include "Vector.H"

///
/**
   This is the graph data held for each irregular node.
 */
class GraphNodeImplem
{
public:
  ///
  inline 
  GraphNodeImplem();
  
  ///
  inline 
  ~GraphNodeImplem();

  ///
  GraphNodeImplem& operator=(const GraphNodeImplem& a_impin);

  ///
  GraphNodeImplem(const GraphNodeImplem& a_impin);

  
  ///we also use the default copy constructor and operator=

  ///  
  /**
     Cell indicies of neighboring cells.
     If the arc is a boundary face, the int = -1
  */
  Vector<int>      m_arc[2*SpaceDim];
  
  ///cell index of this vof is the index into the vector

  ///cell index of next coarser vofs
  int              m_coarserNode;
  
  // (for construction) index into node vector.
  int m_nodeInd;

  ///list of finer vofs
  Vector<VolIndex> m_finerNodes;
  
  ///return the index into the arc vector
  inline int index(int a_idir, Side::LoHiSide a_side) const;

  ///
  int linearSize() const;

  ///
  void linearOut(void* buffer ) const;

  ///
  void linearIn(void* buffer );

};

///
/**
   This is a list showing
   the connectivity of a given cell.  This 
   has the special property of also being able
   to set itself to regular or covered by setting the 
   current nodeimplem pointer to 1 or 0.
 */
class GraphNode
{
public:
  ///
  /**
     constructor sets node to regular as default.
   */
  inline GraphNode();

  ///
  /**
   */
  inline ~GraphNode();

  ///
  int size() const;

  ///
  /**
     deletes memory if it has been allocated and sets
     members to 0
   */
  inline void clear();
  ///
  /**
     Return true if the node is regular (m_thisNode==1).
   */
  inline bool isRegular() const;

  ///
  /**
     Return true if the node is covered (m_thisNode==0).
   */
  inline bool isCovered() const;
  
  ///
  /**
     Return true if the node is neither regular or covered.
   */
  inline bool isIrregular() const;
  
  ///
  /**
     Set Node to regular.  If previously set to irregular,
     deletes their memory.
   */
  inline void defineAsRegular();
  
  ///
  /**
     Set Node to covered.  If previously set to irregular,
     deletes their memory.
   */
  inline void defineAsCovered();
  
  ///
  /**
     Get the faces in the direction and side
     for the vof in the list.  if the vof's cell
     index is not found in the list, abort.
     Use the input vof's grid index for a the grid index
     of the list.
   */
  Vector<FaceIndex> 
  getFaces(const VolIndex&       a_vof, 
           const int&            a_idir, 
           const Side::LoHiSide& a_sd,
           const Box&            a_domain) const;

  ///
  /**
     Return all the vofs in the list, using the input
     intvect for the gridIndex
   */
  Vector<VolIndex> 
  getVoFs(const IntVect& a_iv) const;
  
  ///
  const GraphNode& operator=(const GraphNode& ebiin);

  ///
  GraphNode(const GraphNode& ebiin);

  ///
  /**
     Returns the corresponding set of VoFs from the next finer
     EBGraph (factor of two refinement).  The result is only
     defined if this {\tt EBGraph} was defined by coarsening.  
  */
  Vector<VolIndex> refine(const VolIndex& a_coarVoF) const;

  ///
  /**
     Returns the corresponding  VoF from the next coarser 
     EBGraph (same solution location, different index space, factor
     of two refinement ratio).
  */
  VolIndex coarsen(const VolIndex& a_fineVoF) const;

  ///
  /**
   */
  void addIrregularNode(const GraphNodeImplem& a_nodein, int cellIndex);
  ///
  /**
   */
  void addIrregularNode(const GraphNodeImplem& a_nodein);

  ///
  /**
     get all sets of connected vofs within the box.
     used in coarsening
  */
  Vector<Vector<VolIndex> >  getVoFSets(const Box& a_box) const;
  
  ///
  int linearSize() const;

  ///
  void linearOut(void* buffer ) const;

  ///
  void linearIn(void* buffer );

  /// internal use only
  inline void setDefaults();

  ///
  /**
     The connectivity data at this point.
     If m_thisNode == 0, node is covered.
     If m_thisNode == 1, node is regular.
     otherwise, it is a real list.
  */
  Vector<GraphNodeImplem>* m_cellList;

};


/*  inline functions */
/*******************************/
inline GraphNode::
~GraphNode()
{
  clear();
}
/*******************************/
inline bool GraphNode::
isRegular() const
{
  return (m_cellList==((Vector<GraphNodeImplem>*) 1));
}
/*******************************/
inline bool GraphNode::
isCovered() const
{
  return (m_cellList==((Vector<GraphNodeImplem>*) 0));
}
/*******************************/
inline bool GraphNode::
isIrregular() const
{
  return ((!isRegular()) && (!isCovered()));
}
/*******************************/
inline void GraphNode::
setDefaults()
{
  m_cellList = (Vector<GraphNodeImplem>*)1;
}

/*******************************/
inline GraphNode::
GraphNode()
{
  setDefaults();
}

/*******************************/
inline void GraphNode::
clear()
{
  if(isIrregular())
    {
      delete m_cellList;
    }
  setDefaults();
}

/*******************************/

inline void GraphNode::
defineAsRegular()
{
  clear();
  m_cellList = (Vector<GraphNodeImplem>*) 1;
}
/*******************************/

inline void GraphNode::
defineAsCovered()
{
  clear();
  m_cellList = (Vector<GraphNodeImplem>*) 0;
}

/*******************************/
inline GraphNodeImplem::
GraphNodeImplem()
  : m_coarserNode(-1), m_nodeInd(-1)
{
}
/*******************************/
inline GraphNodeImplem::
~GraphNodeImplem()
{
}

/*******************************/
inline int GraphNodeImplem::
index(int a_idir, Side::LoHiSide a_sd) const
{
  assert(a_idir >= 0 && a_idir < SpaceDim);
  return a_idir + a_sd*CH_SPACEDIM;
}
/*
  if(a_sd == Side::Lo)
    {
      retval = a_idir;
    }
  else
    {
      retval = a_idir + SpaceDim;
    }
  return retval;
}
*/
#endif
