/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#include "GraphNode.H"
#include "BoxIterator.H"
#include "parstream.H"
#include "EBArith.H"
/*******************************/
GraphNodeImplem::
GraphNodeImplem(const GraphNodeImplem& a_impin)
{
  for(int iarc = 0; iarc < 2*SpaceDim; iarc++)
    {
      m_arc[iarc] = a_impin.m_arc[iarc];
    }

  m_coarserNode = a_impin.m_coarserNode;
  m_nodeInd     = a_impin.m_nodeInd;
  m_finerNodes  = a_impin.m_finerNodes;
}
/*******************************/
GraphNodeImplem&
GraphNodeImplem::
operator=(const GraphNodeImplem& a_impin)
{
  if(&a_impin != this)
    {
      for(int iarc = 0; iarc < 2*SpaceDim; iarc++)
        {
          m_arc[iarc] = a_impin.m_arc[iarc];
        }

      m_coarserNode = a_impin.m_coarserNode;
      m_nodeInd     = a_impin.m_nodeInd;
      m_finerNodes  = a_impin.m_finerNodes;
    }
  return *this;
}



/*******************************/
int GraphNode::size() const
{
  int retval;
  if(isRegular())
    {
      retval =  1;
    }
  else if(isCovered())
    {
      retval =  0;
    }
  else
    {
      retval =  m_cellList->size();
    }
  return retval;
}

/*******************************/
void 
GraphNode::addIrregularNode(const GraphNodeImplem& a_nodein, int cellIndex)
{
  if(isRegular() || isCovered())
    {
      m_cellList = new Vector<GraphNodeImplem>();
    }
  if(m_cellList->size() < cellIndex+1)
    {
      m_cellList->resize(cellIndex+1);
    }
  (*m_cellList)[cellIndex] = a_nodein;
}
void 
GraphNode::addIrregularNode(const GraphNodeImplem& a_nodein)
{
  if(isRegular() || isCovered())
    {
      m_cellList = new Vector<GraphNodeImplem>();
    }
  m_cellList->push_back(a_nodein);
}


/*******************************/
Vector<VolIndex> GraphNode::
getVoFs(const IntVect& a_iv) const
{
  Vector<VolIndex> retvec;
  if(isCovered())
    {
      //return empty vector
    }
  else if(isRegular())
    {
      retvec.push_back(VolIndex(a_iv, 0));
    }
  else
    {
      const Vector<GraphNodeImplem>& vofVec = *m_cellList;
      for(int ivec = 0; ivec < vofVec.size(); ivec++)
        {
          VolIndex vof(a_iv, ivec);
          retvec.push_back(vof);
        }

    }
  return retvec;
}
/*******************************/
Vector<FaceIndex> GraphNode::
getFaces(const VolIndex&       a_vof, 
         const int&            a_idir, 
         const Side::LoHiSide& a_sd,
         const Box&            a_domain) const
{
  static Vector<FaceIndex> emptyVec;
  static Vector<FaceIndex> regularVec(1);



  IntVect otherIV = a_vof.gridIndex() +sign(a_sd)*BASISV(a_idir);

  if(isRegular())
    {
      FaceIndex& face = regularVec[0];
      // if node is regular, the other iv must be single valued
      int otherCellIndex = 0;
      if(!a_domain.contains(otherIV))
        {
          otherCellIndex = -1;
        }
      VolIndex otherVoF(otherIV, otherCellIndex);
      face.define(a_vof, otherVoF, a_idir);
      return regularVec;
    }
  else if(isCovered())
    {
      //return empty vector
      return emptyVec;
    }
  else
    {
      const Vector<GraphNodeImplem>& nodeVec = *m_cellList;
      const GraphNodeImplem& node =  nodeVec[a_vof.cellIndex()];
      const Vector<int>& arcs = node.m_arc[node.index(a_idir, a_sd)];
      if(arcs.size()==0) 
        {
          return emptyVec;
        }
      if(arcs.size()==1)
        {
          FaceIndex& face = regularVec[0];
          VolIndex otherVoF(otherIV, arcs[0]); 
          face.define(a_vof, otherVoF, a_idir);
          return regularVec;
        }

      Vector<FaceIndex> retvec;
      //cell index of the list is the same as the 
      //index into the vector. if the input cell
      //index is too big (or < 0), we can tell by Vector 
      //going out of bounds

      for(int ivec = 0; ivec < arcs.size(); ivec++)
        {
          VolIndex otherVoF(otherIV, arcs[ivec]); 
          retvec.push_back(FaceIndex(a_vof, otherVoF, a_idir));
        }
      return retvec;
    }
  return emptyVec;
}
/*******************************/
Vector<VolIndex> GraphNode::
refine(const VolIndex& a_coarVoF) const
{
  Vector<VolIndex> retvec;
  if(isCovered())
    {
      //return empty vector
    }
  else if(isRegular())
    {
      const IntVect& iv = a_coarVoF.gridIndex();
      Box refbox = ebrefine(Box(iv,iv),2);
      BoxIterator bit(refbox);
      for(bit.reset(); bit.ok(); ++bit)
        {
          retvec.push_back(VolIndex(bit(), 0));
        }
    }
  else 
    {
      //irregular node.
      //cell index of the list is the same as the 
      //index into the vector. if the input cell
      //index is too big (or < 0), we can tell by Vector 
      //going out of bounds
      const Vector<GraphNodeImplem>& nodeVec = *m_cellList;
      const GraphNodeImplem& node =  nodeVec[a_coarVoF.cellIndex()];
      retvec = node.m_finerNodes;
    }
  return retvec;
}
/*******************************/
const GraphNode& GraphNode::
operator=(const GraphNode& a_nodein)
{
  if(this != &a_nodein)
    {
      clear();
      setDefaults();
      //if node is regular or covered, just copy the pointer
      //otherwise, append the list of nodes
      if((a_nodein.isRegular()) || (a_nodein.isCovered()))
        {
          m_cellList = a_nodein.m_cellList;
        }
      else 
        {
          m_cellList = new Vector<GraphNodeImplem>();
          (*m_cellList) = (*a_nodein.m_cellList);
        }
    }
  return *this;
}
/*******************************/
GraphNode::
GraphNode(const GraphNode& a_nodein)
{
  //if node is regular or covered, just copy the pointer
  //otherwise, append the list of nodes
  if((a_nodein.isRegular()) || (a_nodein.isCovered()))
    {
      m_cellList = a_nodein.m_cellList;
    }
  else 
    {
      m_cellList = new Vector<GraphNodeImplem>();
      (*m_cellList) = *(a_nodein.m_cellList);
    }
}
/*******************************/
VolIndex 
GraphNode::coarsen(const VolIndex& a_fineVoF) const
{
  VolIndex retval;
  IntVect ivCoar = ebcoarsen(a_fineVoF.gridIndex(), 2);
  if(isRegular() || isCovered())
    {
      retval = VolIndex(ivCoar, 0);
    }
  else 
    {
      const Vector<GraphNodeImplem>& nodes = *m_cellList;
      int inode = a_fineVoF.cellIndex();
      int cellIndex = nodes[inode].m_coarserNode;
      retval = VolIndex(ivCoar, cellIndex);
    }
  return retval;
}
/*******************************/
int
GraphNode::linearSize() const
{
  int retval;
  if(isRegular() || isCovered())
    {
      //regular/irregular covered
      retval = sizeof(int);
    }
  else
    {
      //regular/irregular covered then
      //number of vofs
      retval = 2*sizeof(int);
      //node data
      const Vector<GraphNodeImplem>& nodes = *m_cellList;
      for(int inode = 0; inode < nodes.size(); inode++)
        {
          retval +=  nodes[inode].linearSize();
        }
    }
  return retval;
}
/*******************************/
void 
GraphNode::linearOut(void*  a_buf) const
{
  int secretCode;
  if(isRegular())
    {
      //secret code for regular.
      //trying to not have two separate secret codes in one class,
      //this matches the m_cellList val for regular
      secretCode = 1;
    }
  else if(isCovered())
    {
      //secret code for covered
      //trying to not have two separate secret codes in one class,
      //this matches the m_cellList val for covered
      secretCode =  0;
    }
  else
    {
      assert(isIrregular());
      //i had to pick something
      secretCode =  2;
    }

  int* intbuf = (int *) a_buf;
  //regular/irregular covered
  *intbuf = secretCode;
  intbuf++;
  if(isIrregular())
    {
      int nvofs = m_cellList->size();
      //number of vofs
      *intbuf =  nvofs;
      intbuf++;

      unsigned char* buffer=(unsigned char*) intbuf;

      //now put in the actual nodes
      const Vector<GraphNodeImplem>& nodes = *m_cellList;
      for(int inode = 0; inode < nodes.size(); inode++)
        {
          nodes[inode].linearOut(buffer);
          int nodeSize = nodes[inode].linearSize();
          buffer += nodeSize;
        }
    }
}
/*******************************/
void 
GraphNode::linearIn(void* a_buf)
{

  int* intbuf = (int *) a_buf;

  int secretCode = *intbuf;
  intbuf++;
  if(secretCode == 1)
    {
      //secret code for regular.
      //trying to not have two separate secret codes in one class,
      //this matches the m_cellList val for regular
      defineAsRegular();
    }
  else if(secretCode == 0)
    {
      //secret code for covered
      //trying to not have two separate secret codes in one class,
      //this matches the m_cellList val for covered
      defineAsCovered();
    }
  else
    {
      //secret code for irregular.  I had to pick something.
      assert(secretCode == 2);

      //regular/irregular covered
      //number of vofs
      int nvofs = *intbuf;
      intbuf++;

      //using intbuf for the two ints we just extracted
      unsigned char* buffer = (unsigned char*) intbuf;

      //now put in the actual nodes
      for(int inode = 0; inode < nvofs; inode++)
        {
          GraphNodeImplem newNode;
          newNode.linearIn(buffer);
          addIrregularNode(newNode);
          int nodeSize = newNode.linearSize();
          buffer   +=  nodeSize;
        }

    }
}
/*******************************/
/*******************************/
int
GraphNodeImplem::linearSize() const
{
  int linSize = 0;

  //arc sizes
  for(int iarc = 0; iarc < 2*SpaceDim; iarc++)
    {
      int thisArcSize = m_arc[iarc].size();
      //space for each int in each vector + 
      //the size of the vector
      linSize += sizeof(int)*(thisArcSize + 1);
    }

  //coarser node
  linSize += sizeof(int);

  //finer nodess size
  linSize += sizeof(int);

  //finer nodes
  for(int inode = 0; inode < m_finerNodes.size(); inode++)
    {
      linSize += m_finerNodes[inode].linearSize();
    }
  return linSize;
}
/*******************************/
void 
GraphNodeImplem::linearOut(void*  a_buf) const
{
  int* intbuf = (int*) a_buf;
  int linSize = 0;
  
  //space for each int in each vector + size of the vector
  for(int iarc = 0; iarc < 2*SpaceDim; iarc++)
    {
      const Vector<int>& thisArc = m_arc[iarc];
      *intbuf = thisArc.size();
      intbuf++;
      linSize += sizeof(int);
      for(int ivec = 0; ivec < thisArc.size(); ivec++)
        {
          *intbuf = thisArc[ivec];
 
          intbuf++;
          linSize += sizeof(int);
        }
    }

  //coarser node
  *intbuf = m_coarserNode;
  linSize += sizeof(int);
  intbuf++;

  //finer nodess size
  *intbuf = m_finerNodes.size();
  linSize += sizeof(int);
  intbuf++;

  //finer nodes
  char* charbuf = (char*) intbuf;
  for(int inode = 0; inode < m_finerNodes.size(); inode++)
    {
      m_finerNodes[inode].linearOut(charbuf);
      int thisSize = m_finerNodes[inode].linearSize();
      linSize += thisSize;
      charbuf += thisSize;
    }
}
/*******************************/
void 
GraphNodeImplem::linearIn(void* a_buf)
{
  int* intbuf = (int*) a_buf;
  int linSize = 0;
  //i am only outputting the arcs.

  //space for each int in each vector + size of the vector
  for(int iarc = 0; iarc < 2*SpaceDim; iarc++)
    {
      Vector<int>& thisArc = m_arc[iarc];
      int thisArcSize = *intbuf;
      intbuf++;
      linSize += sizeof(int);
      thisArc.resize(thisArcSize);
      for(int ivec = 0; ivec < thisArc.size(); ivec++)
        {
  
          thisArc[ivec] = *intbuf;
          intbuf++;
          linSize += sizeof(int);
        }
    }

  //coarser node
  m_coarserNode = *intbuf;
  linSize += sizeof(int);
  intbuf++;

  //finer nodess size
  int fineNodeSize = *intbuf;
  linSize += sizeof(int);
  intbuf++;

  //finer nodes
  m_finerNodes.resize(fineNodeSize);
  char* charbuf = (char*) intbuf;
  for(int inode = 0; inode < m_finerNodes.size(); inode++)
    {
      m_finerNodes[inode].linearIn(charbuf);
      int thisSize = m_finerNodes[inode].linearSize();
      linSize += thisSize;
      charbuf += thisSize;
    }
}
/*******************************/
