/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _STENCILS_H_
#define _STENCILS_H_

#include "Vector.H"
#include "VolIndex.H"
#include "FaceIndex.H"
#include "REAL.H"

/// VoF-centered stencil
/**
   This stencil is a fundamental tool for building
   eb applications by making the ability to cache
   stencil information. This object consists of 
   a vector of VoFs and a vector of weights.
 */
class VoFStencil 
{
public:
  /// 
  /**
     default constructor.  makes empty vectors.
  */
  VoFStencil();

  ///
  ~VoFStencil();

  ///
  void clear();

  ///
  /**
     add a VoF to the Stencil, with it's associated weight
     it is required that all VoFs are in the same EBIndexSpace
     If the vof is already in the 
     stencil, add the weights.  
  */
  void
  add(const VolIndex& vof,const Real& weight);

  ///
  /**
     number of VoFs in the Stencil
  */
  int
  size() const;

  ///
  /**
     access a VoF
  */
  const VolIndex&
  vof(int i) const;

  ///
  /** 
      access a weight
   */
  const Real&
  weight(int i) const;

  ///
  /**
     add all faces and weights of inputs
     to this.  If a vof is already in the 
     stencil, add the weights.
     only addition is well-defined here
     as far as arithmatic operations are
     concerned.
   */
  VoFStencil&
  operator+=(const VoFStencil& a_vofsten);


  ///
  /**
   */
  void operator*=(const Real& scaling);

  ///
  VoFStencil&
  operator=(const VoFStencil& a_vofsten);

  ///
  VoFStencil(const VoFStencil& a_vofstenin);

  ///
  bool isDefined() const;

protected:

  bool m_isDefined;
  /// the VoFs
  Vector<VolIndex> vofs;
  /// the weights
  Vector<Real> weights;
};


///
/**
   Face-centered Stencil for embedded boundary applications.
 */
class FaceStencil 
{
public:
  ///
  FaceStencil();

  ///
  ~FaceStencil();

  ///
  void clear();

  ///
  /**
     add an Face and it's weight
     If the face is already in the 
     stencil, add the weights.  
  */
  void
  add(const FaceIndex& face,const Real& weight);

  /// 
  /**
     number of Faces in the Stencil
  */
  int
  size() const;

  ///
  /** access an Face
   */
  const FaceIndex&
  face(int i) const;

  /// 
  /**
     access a weight
  */
  const Real&
  weight(int i) const;

  ///
  /**
     add all faces and weights of inputs
     to this.  If a face is already in the 
     stencil, add the weights.
     only addition is well-defined here
     as far as arithmatic operations are
     concerned.
   */
  FaceStencil&
  operator+=(const FaceStencil& a_facesten);

  ///
  FaceStencil&
  operator=(const FaceStencil& a_facesten);

  ///
  FaceStencil(const FaceStencil& a_facesten);

  ///
  /**
   */
  void operator*=(const Real& scaling);

private:
  /// the Faces
  Vector<FaceIndex> faces;
  /// the weights
  Vector<Real> weights;
};

#endif
