/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL

#ifndef _VolIndex_H_
#define _VolIndex_H_

#include "REAL.H"
#include "IntVect.H"
#include "SPMD.H"

/// Volume of Fluid Index 
/**
   VolIndex is a very lightweight object
   to be used to distinguish between volumes of
   fluid in a cell.  We use VolIndex for 
   array indexing.   Non-trivial constructors
   are all private because only EBISBox
   is supposed to make valid volume indicies.
*/
class VolIndex
{
public:
  ///
  inline
  VolIndex();

  ///
  //~VolIndex(); compiler generated is fine

  /// 
  //VolIndex& operator= (const VolIndex& a_vofin);

  /// 
  inline
  bool operator== (const VolIndex& a_vofin) const;

  ///
  inline
  const IntVect& gridIndex() const;

  ///
  inline
  int cellIndex() const;

  ///
  inline
  bool isDefined() const;

  ///
  // VolIndex(const VolIndex& a_vofin);

  ///
  void define(const VolIndex& a_vofin);

  ///
  inline
  VolIndex(const IntVect& a_ix,const  int& a_vofID);

  ///
  inline 
  void define(const IntVect& a_ix,const  int& a_vofID);

  inline
  bool operator<(const VolIndex& rhs) const;

  inline
  bool operator!=(const VolIndex& rhs) const;
  /**
	 /name linearization routines
  */
  /*@{*/
  int linearSize() const;
  void linearOut(void* const a_outBuf) const;
  void linearIn(const void* const inBuf);
  /*@}*/

  //  not for users
  static size_t iv_offset, index_offset;
  static int initializeOffsets();

private:


  //cell of the VolIndex
  IntVect m_iv;

  //which VoF in the cell (0 to nVoF-1)
  int m_cellIndex;

  bool m_isDefined;
};

std::ostream& operator<< (std::ostream&       os,
                          const VolIndex& iv);


inline bool VolIndex::operator<(const VolIndex& rhs) const
{
  if(m_iv.lexLT(rhs.m_iv)) return true;
  if(m_iv == rhs.m_iv) {
	return m_cellIndex < rhs.m_cellIndex;
  }
  return false;
}
/*****************************************/
inline
const IntVect& 
VolIndex::gridIndex() const
{
  return m_iv;
}

/*****************************************/
inline 
int
VolIndex::cellIndex() const
{
  return m_cellIndex;
}
/*****************************************/
inline void 
VolIndex::define(const IntVect& a_ix,const int& a_vofID)
{
  m_isDefined = true;
  m_iv = a_ix;
  m_cellIndex = a_vofID;
}
/*****************************************/
inline
VolIndex::VolIndex(const IntVect& a_ix,const int& a_vofID)
  :m_iv(a_ix), m_cellIndex(a_vofID), m_isDefined(true)
{
  ;
}
/*****************************************/
inline bool 
VolIndex::isDefined() const
{
  return m_isDefined;
}

/*****************************************/
inline
VolIndex::VolIndex()
  :m_iv(IntVect::Zero), m_cellIndex(-1), m_isDefined(false)
{
  //set index to bogus number to make undefined
  //ones catchable.

}

inline bool 
VolIndex::operator== (const VolIndex& a_vofin) const
{
  return ((m_iv == a_vofin.m_iv)&&
          (m_cellIndex == a_vofin.m_cellIndex));
}

inline bool
VolIndex::operator!=(const VolIndex& rhs) const
{
  return !(*this == rhs);
}

#endif
