/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// BVS, June 18, 2003

// We can assume that template class T has null construction.

#ifndef _AMRMULTIGRID_H_
#define _AMRMULTIGRID_H_

#include "MultiGrid.H"
#include "REAL.H"
#include "Box.H"


template <class T>
class AMRLevelOp : public MGLevelOp<T>
{
public:
  virtual ~AMRLevelOp(){;}
  /** returns 1 when there are no coarser AMRLevelOp objects */
  virtual int refToCoarser() = 0;

  /** generate approximate solution to L(a_correction) = a_res */
  virtual void AMRRelaxLevel(T& a_correction, const T& a_res) = 0;

  /** a_residual = a_rhs - L(a_phi, a_phiFine, a_phiCoarse) */
  virtual void AMRResidual(T& a_residual, const T& a_phiFine, const T& a_phi,
                           const T& a_phiCoarse, const T& a_rhs) = 0;

  /** a_residual = a_rhs - L(a_phi, a_phiCoarse)  */
  virtual void AMRResidualNF(T& a_residual, const T& a_phi, const T& a_phiCoarse,
                             const T& a_rhs) = 0;

  /** a_resCoarse = I[h-2h]( a_residual - L(a_correction, a_coarseCorrection)) */
  virtual void AMRRestrict(T& a_resCoarse, const T& a_residual, const T& a_correction,
                           const T& a_coarseCorrection) = 0;

  /** a_correction += I[2h->h](a_coarseCorrection) */
  virtual void AMRProlong(T& a_correction, const T& a_coarseCorrection) = 0;

  /** a_residual = a_residual - L(a_correction, a_coarseCorrection) */
  virtual void AMRUpdateResidual(T& a_residual, const T& a_correction,
                                 const T& a_coarseCorrection) = 0;

};


template <class T>
class AMRLevelOpFactory : public MGLevelOpFactory<T>
{
public:
  virtual ~AMRLevelOpFactory(){;}
  virtual AMRLevelOp<T>* AMRnewOp(const ProblemDomain& a_indexSpace)=0;
  //virtual void AMRreclaim(AMRLevelOp<T>* a_reclaim)=0;
};

template <class T>
class AMRLevelSolver
{
public:
  AMRLevelSolver();
  virtual ~AMRLevelSolver();
  virtual void define(const ProblemDomain& a_indexSpace,
                      AMRLevelOp<T>* a_operator,
                      LinearSolver<T>* a_bottomSolver,
                      int a_maxIter,
                      Real a_eps,
                      int a_minIter,
                      Real a_hangTolerance);

  virtual void lsolve(T& a_phi, const T* a_phiCoarse,
                      const T&  a_rhs);


protected:
  AMRLevelOp<T>* m_op;
  MultiGrid<T>   m_mg;
  int   m_iterMax;
  Real  m_eps;
  int   m_imin ;
  Real  m_hang ;

};

template <class T>
class AMRMultiGrid
{
public:
  AMRMultiGrid();
  virtual ~AMRMultiGrid();
  virtual void define(const ProblemDomain& a_fineDomain,
                      AMRLevelOpFactory<T>& a_factory,
                      LinearSolver<T>* a_bottomSolver,
                      int a_lmax,
                      int a_lbase);

  virtual void solve(Vector<T*> a_phi, Vector<T*> a_rhs, int a_max, int a_base);

  //virtual void AMRreclaim(AMRLevelOpFactory<T>& a_factory);

protected:
  void relax(T& phi, T& R, int depth);

  Vector<AMRLevelOp<T>*> m_op;
  Vector<T>  m_phiSave;
  Vector<T>  m_correction;
  Vector<T>  m_dcorrCorr;
  Vector<T>  m_residual;
  Vector<AMRLevelSolver<T>*>  m_levelSolvers;
  LinearSolver<T>* m_bottomSolver;
  
  int  m_depth, m_pre, m_post, m_bottom, m_cycle;
};




//*******************************************************
// AMRMultigrid Implementation
//*******************************************************



template <class T>
AMRLevelSolver<T>::AMRLevelSolver()
  :m_op(NULL)
{;}

template <class T>
AMRLevelSolver<T>::~AMRLevelSolver()
{
  assert(m_op == NULL);
}


// template <class T>
// void AMRLevelSolver<T>::lreclaim( MGLevelOpFactory<T>& a_factory)
// {
//   m_mg.MGreclaim(a_factory);
//   m_op=NULL;
// }

template <class T>
void AMRLevelSolver<T>::define(const ProblemDomain& a_indexSpace,
                               AMRLevelOp<T>* a_operator,
                               LinearSolver<T>* bottomSolver,
                               int a_maxIter,
                               Real a_eps,
                               int a_minIter,
                               Real a_hangTolerance)
{
  m_op = a_operator;
  int maxCoarse = a_operator->maxCoarsenings(a_indexSpace);
  m_mg.define(a_factory, maxCoarse, 2, 2, 1, bottomSolver, a_indexSpace);
  m_iterMax = a_maxIter;
  m_eps     = a_eps;
  m_imin    = a_minIter;
  m_hang    = a_hangTolerance;

}

template <class T>
void AMRLevelSolver<T>::lsolve(T& a_phi, const T* a_phiCoarse,
                               const T&  a_rhs)
{
  assert(m_op != NULL);

  T residual;
  T correction;
  Real norm, norm_0, norm_last;

  if(a_phiCoarse == NULL)
    {
      m_op->residual(residual, a_phi, a_rhs, false);
    }
  else
    {
      m_op->AMRResidualNF(residual, a_phi, *a_phiCoarse, a_rhs);
    }

  norm      = m_op->norm(residual);
  norm_0    = norm;
  norm_last = norm;

  int i=0;
  while (i<m_iterMax && norm > m_eps*norm_0 && (i < m_imin || norm <(1-m_hang)*norm_last))
    {
      m_mg->solve(correction, residual);
      m_op->incr(a_phi, correction, 1.0);
      if(a_phiCoarse == NULL)
        {
          m_op->residual(residual, a_phi, a_rhs, false);
        }
      else
        {
          m_op->AMRResidualNF(residual, a_phi, *a_phiCoarse, a_rhs);
        }
      norm_last  = norm;
      norm = m_op->norm(residual);

      i++;
    }

}
 

//===================================================================

template <class T>
AMRMultiGrid<T>::AMRMultiGrid():m_bottomSolver(NULL){}

template <class T>
void AMRMultiGrid<T>::relax(T& a_correction, T& a_residual, int depth)
{


  for(int i=0; i<m_pre; i++)
    {
      m_op[depth]->AMRRelaxLevel(a_correction, a_residual);  //numSmoothDown
    }

  if(m_op[depth]->refToCoarser() > 2)
    {
      //ok, there is an intermediate multigrid level that is 
      // not an AMR level.  We use regular MultiGrid smoothing algorithm for this.
      T cor, res;
      m_op[depth]->createCoarser(res, a_residual, false);
      m_op[depth]->createCoarser(cor, a_correction, true);
      m_op[depth]->restrictResidual(res, a_correction, a_residual);
      m_op[depth+1]->setToZero(cor);
      relax(cor, res, depth+1);
      m_op[depth]->prolongIncrement(a_correction, cor);
      for(int i=0; i<m_post; i++)
        {
          m_op[depth]->AMRRelaxLevel(a_correction, a_residual);  //numSmoothUp
        }
    }
}

template <class T>
AMRMultiGrid<T>::~AMRMultiGrid()
{
}

template <class T>
void AMRMultiGrid<T>::define(const ProblemDomain& a_fineDomain,
                      AMRLevelOpFactory<T>& a_factory,
                      LinearSolver<T>* a_bottomSolver,
                      int a_lmax,
                      int a_lbase)
{

}

template<class T>
void AMRMultiGrid<T>::solve(Vector<T*> a_phi, Vector<T*> a_rhs, int a_max, int a_base)
{
  
}
#endif
