/*   _______              __
     / ___/ /  ___  __ _  / /  ___
     / /__/ _ \/ _ \/  ' \/ _ \/ _ \
     \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "AMRPoissonOp.H"
#include "FORT_PROTO.H"
#include "AMRPoissonOpF_F.H"

void AMRPoissonOp::define(const DisjointBoxLayout& a_grids,
                          const DisjointBoxLayout& a_baseBAPtr,
                          const RealVect&          a_dxLevel,
                          int                      a_refRatio,
                          const ProblemDomain&     a_domain,
                          BCFunc                   a_bc,
                          int                      a_ncomp)
{
  this->define(a_grids, a_dxLevel, a_domain, a_bc);
  m_quadCFI.define(a_grids, &a_baseBAPtr, a_dxLevel[0],
                   a_refRatio, a_ncomp, a_domain);
}

  

void AMRPoissonOp::define(const DisjointBoxLayout& a_grids, const RealVect& a_dx,
                          const ProblemDomain& a_domain, BCFunc a_bc)
{
  m_bc     = a_bc;
  m_domain = a_domain;
  m_dx     = a_dx;
  m_exchangeCopier.define(a_grids, a_grids, IntVect::Unit, true);
  for(int i=0; i<CH_SPACEDIM; ++i)
    {
      LayoutData<NewCFIVS>& lo =  m_loCFIVS[i];
      LayoutData<NewCFIVS>& hi =  m_hiCFIVS[i];
      lo.define(a_grids);
      hi.define(a_grids);
      for(DataIterator dit(a_grids); dit.ok(); ++dit)
        {
          lo[dit].define(a_domain, a_grids.get(dit),a_grids, i, Side::Lo);
          hi[dit].define(a_domain, a_grids.get(dit),a_grids, i, Side::Hi);
        }
    }
}

void AMRPoissonOp::residual(LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_phi,
                            const LevelData<FArrayBox>& a_rhs, bool a_homogeneous)
{
  applyOp(a_lhs, a_phi, a_homogeneous);
  incr(a_lhs, a_rhs, -1);
  scale(a_lhs, -1.0);
}

/**************************/
// this preconditioner first initializes phihat to (IA)phihat = rhshat
// (diagonization of L -- A is the matrix version of L)
// then smooths with a couple of passes of levelGSRB
void AMRPoissonOp::preCond(LevelData<FArrayBox>& a_phi, const LevelData<FArrayBox>& a_rhs)
{
  // diagonal term of this operator is 4/h/h in 2D, 6/h/h in 3D,
  // so inverse of this is our initial multiplier
  Real mult = -m_dx[0]*m_dx[0]/(2.0*SpaceDim);
  Interval comps = a_phi.interval();
  assert(a_phi.nComp() == a_rhs.nComp());

  // don't need to use a Copier -- plain copy will do
  DataIterator dit = a_phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      a_phi[dit].copy(a_rhs[dit].box(), comps,
					  a_rhs[dit].box(), a_rhs[dit],
					  comps);
	
      a_phi[dit] *= mult;
    }

  relax(a_phi, a_rhs, 2);
}

void AMRPoissonOp::applyOp(LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_phi,
                           bool a_homogeneous )
{

  LevelData<FArrayBox>& phi = (LevelData<FArrayBox>&)a_phi;
  Real dx = m_dx[0];

  DataIterator dit = phi.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
    {
      m_bc(phi[dit], m_domain, dx, a_homogeneous);
    }

  phi.exchange(phi.interval(), m_exchangeCopier);


  for (dit.begin(); dit.ok(); ++dit)
    {
      const Box& region = a_lhs[dit].box();
      FORT_OPERATORLAP(CHF_FRA(a_lhs[dit]),
                       CHF_CONST_FRA(phi[dit]),
                       CHF_BOX(region),
                       CHF_CONST_REAL(dx));
    }
}

void AMRPoissonOp::create(    LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_rhs)
{
  m_levelOps.create(a_lhs, a_rhs);
}

void AMRPoissonOp::assign(    LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_rhs)
{
  m_levelOps.assign(a_lhs, a_rhs);
}

Real AMRPoissonOp::dotProduct(const LevelData<FArrayBox>& a_1, const LevelData<FArrayBox>& a_2)
{
  return m_levelOps.dotProduct(a_1, a_2);
}
void AMRPoissonOp::incr( LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_x, Real a_scale)
{
  m_levelOps.incr(a_lhs, a_x, a_scale);
}
void AMRPoissonOp::axby( LevelData<FArrayBox>& a_lhs, const LevelData<FArrayBox>& a_x,
                         const LevelData<FArrayBox>& a_y, Real a_a, Real a_b)
{
  m_levelOps.axby(a_lhs, a_x, a_y, a_a, a_b);
}
void AMRPoissonOp::scale(LevelData<FArrayBox>& a_lhs, const Real& a_scale)
{
  m_levelOps.scale(a_lhs, a_scale);
}
void AMRPoissonOp::setToZero(LevelData<FArrayBox>& a_lhs)
{
  m_levelOps.setToZero(a_lhs);
}

void AMRPoissonOp::relax(LevelData<FArrayBox>& a_e,
                         const LevelData<FArrayBox>& a_residual,
                         int a_iterations)
{
  for(int i=0; i<a_iterations; i++){
    levelGSRB(a_e, a_residual);
  }
}

void AMRPoissonOp::createCoarser(LevelData<FArrayBox>& a_coarse,
                                 const LevelData<FArrayBox>& a_fine,
                                 bool ghosted)
{
  // assert(!ghosted);
  IntVect ghost = a_fine.ghostVect();
  DisjointBoxLayout dbl;
  assert(dbl.coarsenable(2));
  coarsen(dbl, a_fine.disjointBoxLayout(), 2); //multigrid, so coarsen by 2
  a_coarse.define(dbl, a_fine.nComp(), ghost);
}

void AMRPoissonOp::restrictResidual(LevelData<FArrayBox>& a_resCoarse,
                                    LevelData<FArrayBox>& a_phiFine,
                                    const LevelData<FArrayBox>& a_rhsFine)
{
  homogeneousCFInterp(a_phiFine);
  Real dx = m_dx[0];
  for(DataIterator dit = a_phiFine.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi = a_phiFine[dit];
      m_bc(phi,  m_domain,  dx, true);
    }

  a_phiFine.exchange(a_phiFine.interval());

  for(DataIterator dit = a_phiFine.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi = a_phiFine[dit];
      const FArrayBox& rhs = a_rhsFine[dit];
      FArrayBox& res = a_resCoarse[dit];

      Box region = rhs.box();

 
      // This code doesn't work for negative indices!
      //assert(region.smallEnd() >= IntVect::Zero);
      IntVect shift = IntVect::Zero;
      bool shifted = false;
      if(!(region.smallEnd() >= IntVect::Zero))
        {
          shifted = true;
          for(int i=0; i<CH_SPACEDIM; ++i)
            {
              if(region.smallEnd()[i] < 0)
                shift[i] = - region.smallEnd()[i];
            }
          region.shift(shift);
          phi.shift(shift);
          res.shift(shift/2);
          FArrayBox& tmp = (FArrayBox&)rhs;
          tmp.shift(shift);
        }

      res.setVal(0.0);
      FORT_RESTRICTRES(CHF_CONST_FRA(phi),
                       CHF_CONST_FRA(rhs),
                       CHF_FRA(res),
                       CHF_BOX(region),
                       CHF_CONST_REAL(dx));
      if(shifted){
        shift.scale(-1);
        //region.shift(shift);
        phi.shift(shift);
        res.shift(shift/2);
        FArrayBox& tmp = (FArrayBox&)rhs;
        tmp.shift(shift);
      }
    }
}


Real AMRPoissonOp::norm(const LevelData<FArrayBox>& a_x, int a_ord)
{
  return ::norm(a_x, a_x.interval(), a_ord);
}

void AMRPoissonOp::prolongIncrement(LevelData<FArrayBox>& a_phiThisLevel,
                                    const LevelData<FArrayBox>& a_correctCoarse)
{
  for(DataIterator dit = a_phiThisLevel.dataIterator(); dit.ok(); ++dit)
    {
      FArrayBox& phi =  a_phiThisLevel[dit];
      const FArrayBox& coarse = a_correctCoarse[dit];
      FArrayBox& c = (FArrayBox&)coarse;
      Box region = a_phiThisLevel.box(dit);
      const Box& cBox = a_correctCoarse.box(dit);
      assert(cBox == coarsen(region, 2));

      // This code doesn't work for negative indices!
      //assert(region.smallEnd() >= IntVect::Zero);
      
      IntVect shift = IntVect::Zero;
      bool shifted = false;
      if(!(region.smallEnd() >= IntVect::Zero))
        {
          shifted = true;
          region.shift(shift);
          c.shift(shift/2);
          phi.shift(shift);
        }
          
      FORT_PROLONG(CHF_FRA(phi),
                   CHF_CONST_FRA(coarse),
                   CHF_BOX(region));
      if(shifted)
        {
          shift.scale(-1);
          phi.shift(shift);
          c.shift(shift/2);
        }
    }
}

void AMRPoissonOp::levelGSRB(LevelData<FArrayBox>& a_phi,
                             const LevelData<FArrayBox>& a_rhs)
{
  assert(a_phi.isDefined());
  assert(a_rhs.isDefined());
  assert(a_phi.ghostVect() >= IntVect::TheUnitVector());
  assert(a_phi.nComp() == a_rhs.nComp());

  // do first red, then black passes
  for (int whichPass =0; whichPass <= 1; whichPass++)
    {

      //homogeneousCFInterp(a_phi);
      // now step through grids...
      Real dx = m_dx[0];
      DataIterator dit = a_phi.dataIterator();
      for (dit.begin(); dit.ok(); ++dit)
        {
          // invoke physical BC's where necessary
          m_bc(a_phi[dit],  m_domain,  dx, true);
        }
      //fill in intersection of ghostcells and a_phi's boxes
      a_phi.exchange(a_phi.interval(), m_exchangeCopier);


      //fill in intersection of ghostcells and a_phi's boxes
      // dfm -- for a 5 point stencil, this should not be necessary
      //a_phi.exchange(a_phi.interval(), m_exchangeCopier);

      for (dit.begin(); dit.ok(); ++dit)
        {

#ifndef NDEBUG
//          FArrayBox& thisPhi = a_phi[dit];
//          const FArrayBox& thisRhs = a_rhs[dit];
//          const Box& thisBox =  a_phi.box(dit);
//
#endif

          FORT_GSRBLEVELLAP(CHF_FRA(a_phi[dit]),
                            CHF_CONST_FRA(a_rhs[dit]),
                            CHF_BOX(a_rhs.box(dit)),
                            CHF_CONST_REAL(dx),
                            CHF_CONST_INT(whichPass));

        } // end loop through grids

          //      ArrayViewLDFFormatLabel(&a_phi,"%7.5e","phismooth3");
    } // end loop through red-black

}

void
AMRPoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif)
{

  assert( a_phif.ghostVect() >= IntVect::TheUnitVector());

  DataIterator dit = a_phif.dataIterator();
  for(dit.begin(); dit.ok(); ++dit)
    {
      const DataIndex& datInd = dit();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          SideIterator sit;
          for(sit.begin(); sit.ok(); sit.next())
            {
              homogeneousCFInterp(a_phif,datInd,idir,sit());
            }
        }
    }
}

void
AMRPoissonOp::homogeneousCFInterp(LevelData<FArrayBox>& a_phif,
                                  const DataIndex& a_datInd,
                                  int a_idir,
                                  Side::LoHiSide a_hiorlo)
{

  assert((a_idir >= 0) && (a_idir  < SpaceDim));
  assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  assert( a_phif.ghostVect() >= IntVect::TheUnitVector());
  //  assert (m_ncomp == a_phif.nComp());

  const NewCFIVS* cfivs_ptr = NULL;
  if(a_hiorlo == Side::Lo)
    cfivs_ptr = &m_loCFIVS[a_idir][a_datInd];
  else
    cfivs_ptr = &m_hiCFIVS[a_idir][a_datInd];

  if(cfivs_ptr->isPacked())
    {
      int ihiorlo = sign(a_hiorlo);
      FORT_INTERPHOMO(CHF_FRA(a_phif[a_datInd]),
                      CHF_BOX(cfivs_ptr->packedBox()),
                      CHF_CONST_REAL(m_dx[a_idir]),
                      CHF_CONST_REAL(m_dxCrse[a_idir]),
                      CHF_CONST_INT(a_idir),
                      CHF_CONST_INT(ihiorlo));
    }
  else
    {
      const IntVectSet& interp_ivs = cfivs_ptr->getFineIVS();
      if(!interp_ivs.isEmpty())
        {
          // Assuming homogenous, interpolate on fine ivs
          interpOnIVSHomo(a_phif, a_datInd, a_idir,
                          a_hiorlo, interp_ivs);
        }
    }
}

void
AMRPoissonOp::interpOnIVSHomo(LevelData<FArrayBox>& a_phif,
                              const DataIndex& a_datInd,
                              const int a_idir,
                              const Side::LoHiSide a_hiorlo,
                              const IntVectSet& a_interpIVS)
{

  assert((a_idir >= 0) && (a_idir  < SpaceDim));
  assert((a_hiorlo == Side::Lo )||(a_hiorlo == Side::Hi ));
  assert( a_phif.ghostVect() >= IntVect::TheUnitVector());
  IVSIterator fine_ivsit(a_interpIVS);
  FArrayBox& a_phi = a_phif[a_datInd];


  // much of these scalar values can be precomputed and stored if
  // we ever need to speed-up this function (ndk)
  Real x1 = m_dx[a_idir];
  Real x2 = 0.5*(3. * x1 + m_dxCrse[a_idir]);
  Real denom = 1.0-((x1+x2)/x1);
  Real idenom = 1/(denom); // divide is more expensive usually
  Real x = 2.*x1;
  Real xsquared = x*x;

  Real m1 = 1/(x1*x1);
  Real m2 = 1/(x1*(x1-x2));

  Real q1 = 1/(x1-x2);
  Real q2 = x1+x2;

  int ihilo = sign(a_hiorlo);
  Real pa,pb,a,b;
  IntVect ivf;
  for(fine_ivsit.begin(); fine_ivsit.ok(); ++fine_ivsit)
    {
      ivf = fine_ivsit();
      // quadratic interpolation
      for(int ivar = 0; ivar < a_phif.nComp(); ivar++)
        {
          ivf[a_idir]-=2*ihilo;
          pa = a_phi(ivf, ivar);
          ivf[a_idir]+=ihilo;
          pb = a_phi(ivf, ivar);

          a = ((pb-pa)*m1 - (pb)*m2)*idenom;
          b = (pb)*q1 - a*q2;

          ivf[a_idir]+=ihilo;
          a_phi(fine_ivsit(), ivar) = a*xsquared + b*x + pa;

        } //end loop over components
    } //end loop over fine intvects
}

//==========================================================
//
//==========================================================


void AMRPoissonOpFactory::define(const ProblemDomain& a_domain,
                                 const DisjointBoxLayout& a_fine, 
                                 const RealVect& a_dx, 
                                 BCFunc a_bc,
                                 int maxDepth)
{
  m_bc = a_bc;
  m_boxes.resize(0);
  m_boxes.push_back(a_fine);
  m_dx.resize(0);
  m_dx.push_back(a_dx);
  m_domains.resize(0);
  m_domains.push_back(a_domain);
  int i=0;
  while(m_boxes[i].coarsenable(2))
    {
      DisjointBoxLayout dbl;
      coarsen_dbl(dbl, m_boxes[i], 2);
      m_boxes.push_back(dbl);
      RealVect dx = m_dx[i];
      dx*=2;
      m_dx.push_back(dx);
      ProblemDomain d = m_domains[i];
      d.coarsen(2);
      m_domains.push_back(d);
      i++;
    }
}

AMRPoissonOp*
AMRPoissonOpFactory::MGnewOp(const ProblemDomain& a_FineindexSpace,
                             int depth,
                             bool homoOnly)
{
  assert(m_boxes.size()>depth);
  AMRPoissonOp* newOp = new AMRPoissonOp;

  newOp->define(m_boxes[depth], m_dx[depth], m_domains[depth], m_bc);
  return newOp;
}

// void AMRPoissonOpFactory::reclaim(AMRPoissonOp* a_reclaim)
// {
//   delete a_reclaim;
// }

AMRPoissonOp* AMRPoissonOpFactory::AMRnewOp(const ProblemDomain& a_FineindexSpace)
{
  assert(false);
  AMRPoissonOp* newOp = new AMRPoissonOp;
  int depth=1;
  newOp->define(m_boxes[depth], m_dx[depth], m_domains[depth], m_bc);
  return newOp;
}


void AMRPoissonOpFactory::define(const ProblemDomain& a_domain,
                                 const Vector<DisjointBoxLayout>& a_grids,
                                 const Vector<int> a_refRatios,
                                 const RealVect& a_finedx, 
                                 BCFunc a_bc)
{
  
}
