/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "BCFunc.H"
#include "RealVect.H"
//#include "BCFuncF_F.H"


//compute value of a such that a parabola fits constant spaced a, b, c and passes
//through that value ab at a midpoint between a and b.


// solves for value of "a" given:
// u(x=-1)   = *a
// u(x=-1/2) = ab
// u(x=0)    = *b
// u(x=1)    = *c

inline
void quadraticInterp(Real* a, const Real* b, const Real* c, const Real& ab)
{
  *a = (8.0/3.0)*ab + (1.0/3.0)*(*c) -2*(*b);
}

void NeumBC(FArrayBox& a_state,
            const Box& a_valid,
            Real dx,
            bool a_homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side)
{
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  RealVect pos;
  int s = sign(a_side);
  if(a_side == Side::Lo)
    {
      Box to_region = adjCellLo(a_valid, a_dir, 1);
      Box fromRegion = to_region;
      fromRegion.shift(a_dir, 1);
      a_state.copy(a_state, fromRegion, 0, to_region, 0, a_state.nComp());
      if(!a_homogeneous){
        ForAllXBNN(Real,a_state,to_region,0, comps)
          {
            D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
            pos[a_dir]+=dx/2.0;
            a_value(&(pos[0]), &a_dir, &s, &(values[0]), &comps);
            a_stateR += dx*values[nR];
          } EndFor ;
      }

    }
  if(a_side == Side::Hi )
    {
      Box to_region = adjCellHi(a_valid, a_dir, 1);
      Box fromRegion = to_region;
      fromRegion.shift(a_dir, -1);
      a_state.copy(a_state, fromRegion, 0, to_region, 0, a_state.nComp());
      if(!a_homogeneous){
        ForAllXBNN(Real,a_state,to_region,0, comps)
          {
            D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
            pos[a_dir]-=dx/2.0;
            a_value(&(pos[0]), &a_dir, &s, &(values[0]), &comps);
            a_stateR += dx*values[nR];
          } EndFor ;
      }

    }
}
void GenBC(FArrayBox& a_state,
           const Box& a_valid,
           Real dx,
           bool a_homogeneous,
           GenBCFunc a_func,
           int a_dir,
           Side::LoHiSide a_side)
{
  const Box& box  = a_state.box();
  Real *l1, *l2, *lghost;
  RealVect pos1;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  Box edge ;
  int s = a_side;
  int hom = a_homogeneous;

  if(a_side == Side::Lo)
    { s=0; edge = adjCellLo(a_valid, a_dir, 1); edge.shift(a_dir, 1);}
  else
    { s=1; edge = adjCellHi(a_valid, a_dir, 1); edge.shift(a_dir,-1);}

  int incr;
  if(a_dir==0){
    incr     = -sign(a_side);
  }
  else if(a_dir==1){
    incr     = -sign(a_side)*box.size(0);
  }
  else {
    incr     = -sign(a_side)*box.size(0)*box.size(1);
  }
  ForAllXBNN(Real,a_state,edge,0, comps)
    {
      a_stateR += 0 * nR; // shuts up compiler "unused variable" warning
      l1  =_x_pp;
      l2 =_x_pp+incr;
      lghost = _x_pp-incr;
      D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
      pos1[a_dir]-=dx/2.0;
      a_func(&(pos1[0]), &a_dir, &s, &hom, l1, l2, lghost);
    } EndFor ;

}


void DirichletBC(FArrayBox& a_state,
                 const Box& valid,
                 Real dx,
                 bool homogeneous,
                 BCValueFunc a_value)
{
  const Box& box  = a_state.box();
  Real *h1, *h, *hghost, *l1, *l, *lghost;
  RealVect pos1, pos2;
  Real deltaPos;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();

  for(int i=0; i<CH_SPACEDIM; ++i)
    {
      Box lo = adjCellLo(valid, i, 1); lo.shift(i, 1);
      int hi, incr;
      if(i==0){
        incr     = 1;
        hi       = box.size(0)-2;
        deltaPos = box.size(0)*dx;
      }
      else if(i==1){
        incr     = box.size(0);
        hi       = box.size(0)*(box.size(1)-2);
        deltaPos = box.size(1)*dx;
      }
      else {
        incr     = box.size(0)*box.size(1);
        hi       = box.size(0)*box.size(1)*(box.size(2)-2);
        deltaPos = box.size(2)*dx;
      }
      if(!homogeneous){
        ForAllXBNN(Real,a_state,lo,0, comps)
          {
            a_stateR+=0;
            l  =_x_pp;
            l1 =_x_pp+incr;
            lghost = _x_pp-incr;
            h  = _x_pp+hi;
            h1 = h-incr;
            hghost = _x_pp+hi+incr;
            D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
            pos1[i]-=dx/2.0;
            pos2=pos1;
            pos2[i]+=deltaPos;
            int s = -1;
            a_value(&(pos1[0]), &i, &s, &(values[0]), &comps);
            quadraticInterp(lghost, l, l1, values[nR]);
            s=1;
            a_value(&(pos2[0]),  &i, &s, &(values[0]), &comps);
            quadraticInterp(hghost, h, h1, values[nR]);
          } EndFor ;
      }
      else {
        ForAllXBNN(Real,a_state,lo,0, comps)
          {
            a_stateR+=0 *nR * iR;//shuts up "unused variable" warning
            l  =_x_pp;
            l1 =_x_pp+incr;
            lghost = _x_pp-incr;
            h  = _x_pp+hi;
            h1 = h-incr;
            hghost = _x_pp+hi+incr;
            quadraticInterp(lghost, l, l1, 0);
            quadraticInterp(hghost, h, h1, 0);
          }EndFor ;
      }

    }
}

void DiriBC(FArrayBox& a_state,
            const Box& valid,
            Real dx,
            bool homogeneous,
            BCValueFunc a_value,
            int a_dir,
            Side::LoHiSide a_side)
{
  const Box& box  = a_state.box();
  Real *l1, *l, *lghost;
  RealVect pos1;
  Vector<Real> values(a_state.nComp());
  int comps = a_state.nComp();
  Box edge ;
  int s = sign(a_side);
  if(a_side == Side::Lo)
    { edge = adjCellLo(valid, a_dir, 1); edge.shift(a_dir, 1);}
  else
    { edge = adjCellHi(valid, a_dir, 1); edge.shift(a_dir,-1);}

  int incr;
  if(a_dir==0){
    incr     = -sign(a_side);
  }
  else if(a_dir==1){
     incr     = -sign(a_side)*box.size(0);
  }
  else {
    incr     = -sign(a_side)*box.size(0)*box.size(1);
  }
  if(!homogeneous){
    ForAllXBNN(Real,a_state,edge,0, comps)
      {
        a_stateR+=0;
        l  =_x_pp;
        l1 =_x_pp+incr;
        lghost = _x_pp-incr;
        D_TERM(pos1[0]=dx*(iR+0.5);, pos1[1]=dx*(jR+0.5);, pos1[2]=dx*(kR+0.5));
        pos1[a_dir]+=sign(a_side)*dx/2.0;
        a_value(&(pos1[0]), &a_dir, &s, &(values[0]), &comps);
        quadraticInterp(lghost, l, l1, values[nR]);
      } EndFor ;
  }
  else {
    ForAllXBNN(Real,a_state,edge,0, comps)
      {
        a_stateR += 0 *nR * iR; //shuts up "unused variable" warning
        l  =_x_pp;
        l1 =_x_pp+incr;
        lghost = _x_pp-incr;
        quadraticInterp(lghost, l, l1, 0);
      }EndFor ;
  }
}
