/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// BVS, June 18, 2003

#ifndef _CH_BICGSTABSOLVER_H_
#define _CH_BICGSTABSOLVER_H_

#include "LinearSolver.H"

template <class T>
class BiCGStabSolver : public LinearSolver<T>
{
public:
  BiCGStabSolver();
  virtual ~BiCGStabSolver();

  virtual void setHomogeneous(bool a_homogeneous)
     {m_homogeneous = a_homogeneous;}
  virtual void define(LinearOp<T>* a_factory, bool a_homogeneous);
  virtual void solve(T& a_phi, const T& a_rhs);


  bool m_homogeneous;
  LinearOp<T>* m_op;
  int m_imax;
  Real m_eps;
  Real m_hang;
  Real m_small;
  int m_numRestarts;

};




//*******************************************************
// BiCGStabSolver Implementation
//*******************************************************


template <class T>
BiCGStabSolver<T>::BiCGStabSolver()
  :m_homogeneous(false), m_op(NULL), m_imax(40), m_eps(1.0E-6), m_hang(1E-8), m_small(1.0E-30), m_numRestarts(5){;}

template <class T>
BiCGStabSolver<T>::~BiCGStabSolver(){;}

template <class T>
void BiCGStabSolver<T>::define(LinearOp<T>* a_operator, bool a_homogeneous)
{
  assert(m_op == NULL);
  m_homogeneous = a_homogeneous;
  m_op = a_operator;
}


template <class T>
void BiCGStabSolver<T>::solve(T& a_phi, const T& a_rhs)
{
  T r, r_tilde, e, p, p_tilde, s_tilde, t, v;

  m_op->create(r, a_rhs);
  m_op->create(r_tilde, a_rhs);
  m_op->create(e, a_phi);
  m_op->create(p, a_rhs);
  m_op->create(p_tilde, a_phi);
  m_op->create(s_tilde, a_phi);
  m_op->create(t, a_rhs);
  m_op->create(v, a_rhs);

  assert(m_op != NULL);

  m_op->residual(r, a_phi, a_rhs, m_homogeneous);

  m_op->assign(r_tilde, r);
  m_op->setToZero(e);
  int i=0;

  Real rho[4]={0,0,0,0}; // rho[0] = r_i , rho[1] = r_(i-1), etc.
  Real norm[2];
  norm[0] = m_op->norm(r, 2);
  Real initial_norm = norm[0];
  norm[1] = norm[0];

  Real alpha[2] = {0,0};
  Real beta[2] = {0,0};
  Real omega[2] ={0,0};

  int restarts = 0;

  while(i<m_imax && norm[0] > m_eps*norm[1] && norm[1] > 0)
    {
      i++;

      norm[1] = norm[0];
      alpha[1]= alpha[0];
      beta[1] = beta[0];
      omega[1]= omega[0];

      rho[3] = rho[2];
      rho[2] = rho[1];
      rho[1] = m_op->dotProduct(r_tilde, r);
      if(rho[1] == 0)
        {
          // we are finished, we will not converge anymore
          m_op->incr(a_phi, e, 1.0);
          return;
        }
      if(i==1)
        {
          m_op->assign(p, r);
        }
      else
        {
          beta[1] = (rho[1]/rho[2])*(alpha[1]/omega[1]);
          m_op->scale(p, beta[1]);
          m_op->incr(p, v, -beta[1]*omega[1]);
          m_op->incr(p, r, 1.0);
        }
      m_op->preCond(p_tilde, p);
      m_op->applyOp(v, p_tilde, true);
      Real m = m_op->dotProduct(r_tilde, v);
      alpha[0] = rho[1]/m;
      if(Abs(m) > m_small*Abs(rho[1]))
        {
          m_op->incr(r, v, -alpha[0]);
          norm[0] = m_op->norm(r, 2);
          m_op->incr(e, p_tilde, alpha[0]);
        }
      else
        {
          m_op->setToZero(r);
          norm[0] = 0.0;
        }
      if(norm[0] > m_eps*initial_norm)
        {
          m_op->preCond(s_tilde, r);
          m_op->applyOp(t, s_tilde, true);
          omega[0] = m_op->dotProduct(t, r)/m_op->dotProduct(t, t);
          m_op->incr(e, s_tilde, omega[0]);
          m_op->incr(r, t,      -omega[0]);
          norm[0] = m_op->norm(r, 2);
        }

      //pout()<<norm[0]<<"  ";

      if(norm[0] < m_eps*initial_norm ||
         omega[0] == 0.0              ||
         norm[0] > (1-m_hang)*norm[1])
        {

          m_op->incr(a_phi, e, 1.0);
          if(restarts == m_numRestarts)
            {
              return;
            }
          m_op->residual(r, a_phi, a_rhs, m_homogeneous);
          norm[0] = m_op->norm(r, 2);
          m_op->assign(r_tilde, r);
          m_op->setToZero(e);
          //m_op->assign(p, r);
          restarts ++;
        }
    }
  m_op->incr(a_phi, e, 1.0);
}




#endif /*_CH_BICGSTABSOLVER_H_*/
