/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// BVS, June 18, 2003

// We can assume that template class T has null construction.

#ifndef _MULTIGRID_H_
#define _MULTIGRID_H_

#include "LinearSolver.H"
#include "ProblemDomain.H"
#include "REAL.H"
#include "Box.H"


/**
   class MGLevelOp handles the additional tasks of coordinating operations
   between this level, and the level one coarser than this 'level'
*/
template <class T>
class MGLevelOp : public LinearOp<T>
{
public:
  virtual ~MGLevelOp(){;}

  virtual void createCoarser(T& a_coarse, const T& a_fine, bool ghosted) = 0;

  virtual void relax(T& a_correction, const T& a_residual, int a_iterations) = 0 ;
  /**
     calculate restricted residual
     a_resCoarse[2h] = I[h->2h] (rhsFine[h] - L[h](phiFine[h])
  */
  virtual void restrictResidual(T& a_resCoarse, T& a_phiFine, const T& a_rhsFine) = 0;

  /**
     correct the fine solution based on coarse correction
     a_phiThisLevel += I[2h->h](a_correctCoarse)
  */
  virtual void prolongIncrement(T& a_phiThisLevel, const T& a_correctCoarse) = 0;
};



template <class T>
class MGLevelOpFactory /* :  public LinearOpFactory<T> */
{
public:
  virtual ~MGLevelOpFactory(){;}
  virtual MGLevelOp<T>* MGnewOp(const ProblemDomain& a_FineindexSpace,
                                            int a_depth,
                                            bool a_homoOnly = true) = 0;
  //virtual void MGreclaim(MGLevelOp<T>* a_reclaim) = 0;
};




template <class T>
class MultiGrid
{
public:
  MultiGrid();
  virtual ~MultiGrid();

  virtual void define(MGLevelOpFactory<T>& a_factory, int a_depth,  
                      LinearSolver<T>* a_bottomSolver,
                      const ProblemDomain& a_domain);
  /** residual-correction form of solution
      L(a_e) = a_res
  */
  virtual void solve(T& a_e, const T& a_res);

  virtual void reclaim(MGLevelOpFactory<T>& a_factory);

  int  m_depth, m_pre, m_post, m_cycle;
  bool m_homogeneous;

protected:
  bool m_defined;
  LinearSolver<T>*  m_bottomSolver;
  ProblemDomain m_topLevelDomain;
  Vector<MGLevelOp<T>*> m_op;
  Vector< T* >   m_residual;
  Vector< T* >   m_correction;

  void cycle(int a_depth, T& a_correction, const T& a_residual);

};


//*******************************************************
// MultiGrid Implementation
//*******************************************************

template <class T>
MultiGrid<T>::MultiGrid():
  m_pre(2), m_post(2), m_cycle(2), m_homogeneous(true),
  m_defined(false){;}

template <class T>
MultiGrid<T>::~MultiGrid()
{
  assert(m_op.size() == 0);
}

template <class T> 
void MultiGrid<T>::define(MGLevelOpFactory<T>& factory, int a_depth, 
                          LinearSolver<T>* a_bottomSolver,
                          const ProblemDomain& a_domain)
{

  m_depth = a_depth;
  m_bottomSolver = a_bottomSolver;
  m_residual.resize(m_depth);   // the zero'th entry is left null constructed
  m_correction.resize(m_depth); // the zero'th entry is left null constructed
  m_op.resize(m_depth);
  m_topLevelDomain = a_domain;


  int i=0;
  for(; i<m_depth; i++)
    {
      m_residual[i] = new T();
      m_correction[i]= new T();
      m_op[i] = factory.MGnewOp(a_domain , i, true);

    }
  m_bottomSolver->define(m_op[m_depth-1], true);
  m_defined = true;
}


template <class T>
void MultiGrid<T>::solve(T& a_e, const T& a_residual)
{
  if (m_depth > 1)
  {
    m_op[0]->createCoarser(*(m_residual[1]), a_residual, false);
    m_op[0]->createCoarser(*(m_correction[1]), a_e, true);
  }

  for (int i = 2; i < m_depth; i++)
    {
      m_op[i-1]->createCoarser(*(m_residual[i]), *(m_residual[i-1]), false);
      m_op[i-1]->createCoarser(*(m_correction[i]), *(m_correction[i-1]), true);
    }

  if (m_homogeneous)
    {
      cycle(0, a_e, a_residual);
    }
  else
    {
      T correction, residual;
      m_op[0]->create(correction, a_e);
      m_op[0]->create(residual, a_residual);
      m_op[0]->residual(residual, a_e, a_residual);
      m_op[0]->setToZero(correction);
      cycle(0, correction, residual);
      m_op[0]->incr(a_e, correction, 1.0); 
    }
}

template <class T>
void MultiGrid<T>::cycle(int depth, T& correction, const T& residual)
{
  if(depth == m_depth-1)
    {
      m_bottomSolver->solve(correction, residual);
    }
  else
    {

      MGLevelOp<T>* op = m_op[depth];
      int cycles = m_cycle;
      if(depth == m_depth-2) cycles=1;
      for(int c=0; c<cycles; ++c){

        op->relax(correction, residual, m_pre);
        op->restrictResidual(*(m_residual[depth+1]), correction, residual);
        m_op[depth+1]->setToZero(*(m_correction[depth+1]));

        cycle(depth+1, *(m_correction[depth+1]), *(m_residual[depth+1]));

        op->prolongIncrement(correction, *(m_correction[depth+1]));
        op->relax(correction, residual, m_post);

      }
    }

}

template <class T>
void MultiGrid<T>::reclaim(MGLevelOpFactory<T>& factory)
{
  for(int i=0; i<m_op.size(); ++i)
    {
      //factory.reclaim(m_op[i]);
      delete m_op[i];
      delete m_correction[i];
      delete m_residual[i];
    }
  m_op.resize(0);
  m_residual.resize(0);
  m_correction.resize(0);
  m_defined = false;
}


#endif /*_LINEARSOLVER_H_*/
