/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _NEWCFIVS_H_
#define _NEWCFIVS_H_

#include <iostream>
#include <cmath>
#include <assert.h>
#include <cstdlib>
#include "REAL.H"
#include "IntVect.H"
#include "Box.H"
#include "LoHiSide.H"
#include "DisjointBoxLayout.H"
#include "IntVectSet.H"
#include "ProblemDomain.H"

/// Internal Chombo class to get fine intvectset at CF interface
/**
    Class to get fine intvectset at cf interface for one side of a given
    fine-level Box.
    This class should be considered internal to AMRSolver and should
    not be considered part of the Chombo API.
*/
class NewCFIVS
{

private:

protected:
  //for internal use
  void  setDefaultValues();

  //the following variables are mainly for self-identification

  /**
     fine intvects which need to be interpolated
  */
  IntVectSet fiinterp_ivs;

  /**
     Can the coarse-fine IntVectSet be represented as a Box?
  */
  bool m_packed;

  /**
      If m_packed is true, the Box with which the coarse-fine
      IntVectSet can be represented
  */
  Box  m_packedBox;

  /**
     Has this object been defined?
  */
  bool isdefined;

public:

  ///
  bool isDefined() const;

  /// Returns "true" if this CF IntVectSet can be represented as a simple Box.
  /**
      Returns 'true' if this coarse-fine IntVectSet can be represented
      as just a Box, ie.  a user can perform a dense computation instead of
      a pointwise calculation using IVSIterator.
  */
  bool isPacked() const;

  /// Returns packed box for when isPacked() is true.
  /** if isPacked() returns 'true', then the box returned by this function
      is a suitable substitute for this IntVectSet
  */
  const Box& packedBox() const;

  /// Get fine intvects which need to be interpolated.
  /**
      This will be empty if isEmpty() returns true
  */
  const IntVectSet& getFineIVS() const ;

  ///
  NewCFIVS();

  ///
  ~NewCFIVS();

  /// Deprecated constructor (The ProblemDomain class has replaced the Box domain)
  /**
     Replaces Box domain with a ProblemDomain and calls define fn.
  */
  NewCFIVS(
           const Box& Domain,
           const Box& box_in,
           const DisjointBoxLayout& Levboxes,
           int Direction,
           Side::LoHiSide hiorlo);


  /// Full constructor
  /**
     Calls corresponding define function.
   */
  NewCFIVS(
           const ProblemDomain& Domain,
           const Box& box_in,
           const DisjointBoxLayout& Levboxes,
           int Direction,
           Side::LoHiSide hiorlo);

  /// Deprecated define function
  /**
      Replaces Box domain with a ProblemDomain and calls define fn.
  */
  void define(
              const Box& Domain,
              const Box& box_in,
              const DisjointBoxLayout& fine_boxes,
              int Direction,
              Side::LoHiSide hiorlo);

  /// Full define function
  /**
   */
  void define(
              const ProblemDomain& Domain, /// Problem Domain at the fine level
              const Box& box_in,  /// fine-level grid Box
              const DisjointBoxLayout& fine_boxes, /// fine-level grids
              int Direction, /// direction of the face of interest
              Side::LoHiSide hiorlo // is the face on the hi or lo side of Box?
              );

  static long long m_packCount, m_sparseCount;

private:
  void operator= (const NewCFIVS& levcfs_in) {};
  NewCFIVS(const NewCFIVS& levcfs_in) {};

};

inline
bool NewCFIVS::isPacked() const
{
  return m_packed;
}


inline
const Box& NewCFIVS::packedBox() const
{
  return m_packedBox;
}

#endif
