/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "NewPoissonOp.H"
#include "FORT_PROTO.H"
#include "AMRPoissonOpF_F.H"

void NewPoissonOp::define( const RealVect& a_dx,
                           const ProblemDomain& a_domain,
                           BCFunc a_bc)
{
  m_bc     = a_bc;
  m_domain = a_domain;
  m_dx     = a_dx;
}

void NewPoissonOp::residual(  FArrayBox& a_lhs, const FArrayBox& a_phi,
                              const FArrayBox& a_rhs, bool a_homogeneous)
{
  applyOp(a_lhs, a_phi, a_homogeneous);
  a_lhs*=-1.0;
  incr(a_lhs, a_rhs, 1);
}

/**************************/
// this preconditioner first initializes phihat to (IA)phihat = rhshat
// (diagonization of L -- A is the matrix version of L)
// then smooths with a couple of passes of levelGSRB
void NewPoissonOp::preCond(   FArrayBox& a_phi, const FArrayBox& a_rhs)
{
    // diagonal term of this operator is 4/h/h in 2D, 6/h/h in 3D,
  // so inverse of this is our initial multiplier
  Real mult = -m_dx[0]*m_dx[0]/(2.0*SpaceDim);
  Interval comps = a_phi.interval();
  assert(a_phi.nComp() == a_rhs.nComp());


  a_phi.copy(a_rhs.box(), comps,
             a_rhs.box(), a_rhs,
             comps);
	
  a_phi *= mult;


  relax(a_phi, a_rhs, 2);
}

void NewPoissonOp::applyOp(   FArrayBox& a_lhs, const FArrayBox& a_phi,
                              bool a_homogeneous )
{

  FArrayBox& phi = (FArrayBox&)a_phi;
  Real dx = m_dx[0];


  m_bc(phi, m_domain, dx, a_homogeneous);

  const Box& region = a_lhs.box();
  FORT_OPERATORLAP(CHF_FRA(a_lhs),
                   CHF_CONST_FRA(phi),
                   CHF_BOX(region),
                   CHF_CONST_REAL(dx));

}

void NewPoissonOp::create(    FArrayBox& a_lhs, const FArrayBox& a_rhs)
{
  a_lhs.define(a_rhs.box(), a_rhs.nComp());
}

void NewPoissonOp::assign(    FArrayBox& a_lhs, const FArrayBox& a_rhs)
{
  a_lhs.copy(a_rhs);
}

Real NewPoissonOp::dotProduct(const FArrayBox& a_1, const FArrayBox& a_2)
{
  return a_1.dotProduct(a_2);
}
void NewPoissonOp::incr( FArrayBox& a_lhs, const FArrayBox& a_x, Real a_scale)
{
  a_lhs.plus(a_x, a_scale);
}
void NewPoissonOp::axby( FArrayBox& a_lhs, const FArrayBox& a_x,
                         const FArrayBox& a_y, Real a_a, Real a_b)
{
  a_lhs.copy(a_x);
  a_lhs*=a_a;
  a_lhs.plus(a_y, a_b);
}
void NewPoissonOp::scale(FArrayBox& a_lhs, const Real& a_scale)
{
  a_lhs*=a_scale;
}
void NewPoissonOp::setToZero(FArrayBox& a_lhs)
{
  a_lhs.setVal(0.0);
}


Real NewPoissonOp::norm(const FArrayBox& a_x, int a_ord)
{
  return a_x.norm(a_ord);
}

void NewPoissonOp::relax(FArrayBox& a_e,
                         const FArrayBox& a_residual,
                         int a_iterations)
{
  for(int i=0; i<a_iterations; i++){
    levelGSRB(a_e, a_residual);
  }
}

void NewPoissonOp::createCoarser(FArrayBox& a_coarse,
                                 const FArrayBox& a_fine,
                                 bool ghosted)
{
  Box c = a_fine.box();
  if(ghosted)
    {
      c.grow(-1);
    }
  c.coarsen(2);
  c.refine(2);
  if(ghosted) c.grow(1);
  assert(c == a_fine.box());
  if(ghosted) c.grow(-1);
  c.coarsen(2);
  if(ghosted)
    {
      c.grow(1);
    }
  a_coarse.define(c, a_fine.nComp());

}

void NewPoissonOp::restrictResidual(FArrayBox& a_resCoarse,
                                    FArrayBox& a_phiFine,
                                    const FArrayBox& a_rhsFine)
{

  Real dx = m_dx[0];

  FArrayBox& phi = a_phiFine;
  m_bc(phi,  m_domain,  dx, true);



  const FArrayBox& rhs = a_rhsFine;
  FArrayBox& res = a_resCoarse;

  Box region = rhs.box();

  // This code doesn't work for negative indices!
  //assert(region.smallEnd() >= IntVect::Zero);
  IntVect shift = IntVect::Zero;
  bool shifted = false;
  if(!(region.smallEnd() >= IntVect::Zero))
    {
      shifted = true;
      for(int i=0; i<CH_SPACEDIM; ++i)
        {
          if(region.smallEnd()[i] < 0)
            shift[i] = - region.smallEnd()[i];
        }
      region.shift(shift);
      phi.shift(shift);
      res.shift(shift/2);
      FArrayBox& tmp = (FArrayBox&)rhs;
      tmp.shift(shift);
    }

  res.setVal(0.0);
  FORT_RESTRICTRES(CHF_CONST_FRA(phi),
                   CHF_CONST_FRA(rhs),
                   CHF_FRA(res),
                   CHF_BOX(region),
                   CHF_CONST_REAL(dx));
  if(shifted){
    shift.scale(-1);
    //region.shift(shift);
    phi.shift(shift);
    res.shift(shift/2);
    FArrayBox& tmp = (FArrayBox&)rhs;
    tmp.shift(shift);
  }

}




void NewPoissonOp::prolongIncrement(FArrayBox& a_phiThisLevel,
                                    const FArrayBox& a_correctCoarse)
{

  FArrayBox& phi =  a_phiThisLevel;
  const FArrayBox& coarse = a_correctCoarse;
  FArrayBox& c = (FArrayBox&)coarse;
  Box region = a_phiThisLevel.box();
  const Box& cBox = a_correctCoarse.box();
  assert(cBox == coarsen(region, 2));

  // This code doesn't work for negative indices!
  //assert(region.smallEnd() >= IntVect::Zero);
  IntVect shift = IntVect::Zero;
  bool shifted = false;
  if(!(region.smallEnd() >= IntVect::Zero)){
    shifted = true;
    for(int i=0; i<CH_SPACEDIM; ++i)
      {
        if(region.smallEnd()[i] < 0)
          shift[i] = - region.smallEnd()[i];
      }
    region.shift(shift);
    phi.shift(shift);
    c.shift(shift/2);
  }
  
  FORT_PROLONG(CHF_FRA(phi),
               CHF_CONST_FRA(coarse),
               CHF_BOX(region));
  if(shifted){
    shift.scale(-1);
    phi.shift(shift);
    c.shift(shift/2);
  }
    

}

void NewPoissonOp::levelGSRB(FArrayBox& a_phi,
                             const FArrayBox& a_rhs)
{

  assert(a_phi.nComp() == a_rhs.nComp());

  Real dx = m_dx[0];
  //m_bc(a_phi,  m_domain,  dx, true);

  // do first red, then black passes
  for (int whichPass =0; whichPass <= 1; whichPass++)
    {

      // now step through grids...
      //fill in intersection of ghostcells and a_phi's boxes
      // dfm -- for a 5 point stencil, this should not be necessary
      //a_phi.exchange(a_phi.interval(), m_exchangeCopier);


          // invoke physical BC's where necessary
      m_bc(a_phi,  m_domain,  dx, true); //new approach to help with checker
#ifndef NDEBUG

#endif

          FORT_GSRBLEVELLAP(CHF_FRA(a_phi),
                            CHF_CONST_FRA(a_rhs),
                            CHF_BOX(a_rhs.box()),
                            CHF_CONST_REAL(dx),
                            CHF_CONST_INT(whichPass));




    } // end loop through red-black

}


NewPoissonOpFactory::NewPoissonOpFactory()
{
  m_dx  = RealVect::Unit;
  m_bc  = NULL;
}

NewPoissonOpFactory::NewPoissonOpFactory(RealVect& a_dx, BCFunc a_bc)
  :m_dx(a_dx), m_bc(a_bc)
{ }

void NewPoissonOpFactory::define(const RealVect& a_dx,  BCFunc a_bc)
{
  m_dx=a_dx;
  m_bc=a_bc;
}

NewPoissonOp* NewPoissonOpFactory::MGnewOp(const ProblemDomain& a_FineindexSpace,
                                           int   a_depth,
                                           bool  a_homoOnly)
{
  assert(a_depth >= 0 );
  assert(m_bc != NULL);
  NewPoissonOp* newOp = new NewPoissonOp();
  RealVect dx = m_dx;
  ProblemDomain domain = a_FineindexSpace;
  for(int i=0; i<a_depth; i++)
    {
      dx/=2;
      domain.coarsen(2);
    }
  newOp->define(dx, domain, m_bc);
  return newOp;
}

void NewPoissonOpFactory::MGreclaim(NewPoissonOp* a_reclaim)
{
  delete a_reclaim;
}
