/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _BINFAB_H_
#define _BINFAB_H_

#include "BaseFab.H"
#include "RealVect.H"
#include "FluxBox.H"
#include "List.H"
#include "Box.H"
#include "ProblemDomain.H"

/// Base class for particle data on a \p Box
/** This is a class for holding and sorting particle items.  Class <T>
    must have a \p RealVect \b <T>::position() const function which is used
    to place the item in an appropriate bin.  See class \p BinItem for a base
    class for particles.  All BinFabs are assumed to have 1 component.
*/

template<class T>
class BinFab : public BaseFab<List<T> >
{
public:
  /// Null constructor, copy constructor and operator= can be compiler defined.
  BinFab();

  /// Constructs an empty \p BinFab on \a a_domain.
  BinFab(const Box&           a_domain,
         const RealVect&      a_mesh_spacing,
         const RealVect&      a_origin,
         const ProblemDomain& a_probdomain);

  /// Copy constructor
  BinFab(const BinFab& a_binfab);

  /// Destructor
  virtual ~BinFab();

  /// Same as the constructor
  virtual void define(const Box&           a_domain,
                      const RealVect&      a_mesh_spacing,
                      const RealVect&      a_origin,
                      const ProblemDomain& a_probdomain);

  /// Sort all the items into the appropriate bins.
  virtual void reBin();

  /// Copy one item into the appropriate bin.
  virtual void addItem(const T& a_item);

  /// Copy a list of items into the appropriate bins.
  virtual void addItems(const List<T>& a_list);

  /// Move items off the argument list and into the appropriate bins.
  virtual void addItemsDestructive(List<T>& a_list);

  /// Move items into the appropriate bins and remove from the arg list only if the dest is in the Box
  virtual void addItemsDestructive(List<T>& a_list ,const Box& a_valid, bool a_in = true);

  /// Delete all the items in this \p BinFab and reset it to the null \p Box.
  virtual void clear();

  /// Count the number of items in this \p BinFab.
  /** The default is to do the whole Fab for all components.
      This is slow, so use it sparingly.
   */
  virtual int numItems(const Box&      a_box   = Box(),
                       const Interval& a_comps = Interval()) const;


  /// Calls corresponding BaseFab<> function
  /** this is supplied because the other size(...) function hides the 
      BaseFab one.
   */
  virtual const int* size() const;

  // Linearization functions

  /// The number of bytes used by \b linearIn()/linearOut().
  /** Returns the size, in number of bytes, of a flat linear
      representation of the items in the specified \p Box.
  */
  virtual int size(const Box& a_box, const Interval& a_comps) const;

  /// Write a serialized (packed) representation into \a a_buf.
  /** Write a linear representation of the items in the specified \p Box.
      Assumes that sufficient memory for the buffer has already been allocated by
      the caller.
  */
  virtual void linearOut(void* a_buf, const Box& a_box, const Interval& a_comps) const;

  /// Extract a serialized (packed) representation from \a a_buf.
  /** Given a linear representation of the class data previously made
      using linearIn() in \a a_buf, set the data for this class on the
      specified \p Box.
  */
  virtual void linearIn(void* a_buf, const Box& a_box, const Interval& a_comps);

  ///
  static int preAllocatable() { return 2; }

protected:
  /// location in physical space of the 0 index
  RealVect m_origin;

  /// cell size in physical space
  RealVect m_mesh_spacing;

  /// contains the \p Box on which this BinFab is defined
  ProblemDomain m_probdomain;

  /// compute the cell index containing the physical position of the item
  inline IntVect locateBin(const T& a_item) const;
};

// Implementation
#include "BinFabImplem.H"

#endif
