#include "AndyChannel.H"

#ifndef MIN
#define MIN(A,B)  (((A)<(B))?(A):(B))
#endif

#ifndef MAX
#define MAX(A,B)  (((A)>(B))?(A):(B))
#endif

// constructor
AndyChannel::AndyChannel(const Real& a_hW,
			 const Real& a_hw,
			 const Real& a_hd,
			 const Real& a_R,
			 const Real& a_r,
			 const RealVect& a_c)
{
  m_hW = a_hW; // half the wider channel width
  m_hw = a_hw; // half the narrow channel width
  m_hd = a_hd; // half z-depth
  m_R = a_R;   // large radius
  m_r = a_r;   // small radius
  m_c = a_c;   // coordinate for center of structure

  // error checking
  if(m_hW <= 0 || m_hw <= 0 || m_hd <= 0 || a_R <= 0 || a_r <= 0)
    {
      std::cerr << "AndyChannel has a non-positive input parameter." 
		<< std::endl;
      exit(0);
    }

  if(m_hw + m_r + m_R > m_hW)
    {
      std::cerr << "AndyChannel has inconsistent geometry inputs:"
		<< std::endl << "m_hw + m_r + m_R > m_hW." << std::endl;
      exit(0);
    }
}

// destructor  
AndyChannel::~AndyChannel()
{}

// return true if all cells are regular (fluid)
bool AndyChannel::isRegular(const Box&      region,
			    const Box&      domain,
			    const RealVect& origin,
			    const Real&     dx) const
{
  bool regular = true;

  // if 1D, no geometry -- must be regular
  if(SpaceDim == 1)
    {
      return regular;
    }

  BoxIterator bit(region);
  for(bit.reset(); bit.ok() && regular; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
      Vector<RealVect> boxloc = getBoxLocations(cellbox,origin,dx);

      int ncorner = boxloc.size();
      for(int icorner=0; icorner<ncorner && regular; icorner++)
	{
	  const RealVect& cornercoord = boxloc[icorner];

	  // if z-coord outside range, then at least partially covered
	  if(SpaceDim == 3)
	    {
	      if(ABS(cornercoord[2]-m_c[2]) > m_hd)
		{
		  regular = false;
		}
	    }

	  // if y-coord outside range, then at least partially covered
	  if(cornercoord[1] >= m_c[1])
	    {
	      if(cornercoord[1] > yofx(cornercoord))
		{
		  regular = false;
		}
	    }
	  else
	    {
	      if(cornercoord[1] < yofx(cornercoord))
		{
		  regular = false;
		}
	    }
	}
    }

  return regular;
}


// return true if all cells are covered (boundary)
bool AndyChannel::isCovered(const Box&      region,
			    const Box&      domain,
			    const RealVect& origin,
			    const Real&     dx) const
{
  bool covered = true;

  if(SpaceDim == 1)
    {
      return false;
    }

  BoxIterator bit(region);
  for(bit.reset(); bit.ok() && covered; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
      Vector<RealVect> boxloc = getBoxLocations(cellbox,origin,dx);

      bool HaveZAbove = false;
      bool HaveZBelow = false;
      bool HaveYAbove = false;
      bool HaveYBelow = false;

      int ncorner = boxloc.size();

      for(int icorner = 0; icorner < ncorner && covered; icorner++)
	{
	  const RealVect& cornercoord = boxloc[icorner];

	  bool zIn = (SpaceDim == 3) ? false:true;
	  bool yIn = false;

	  if(SpaceDim == 3)
	    {
	      if(ABS(cornercoord[2]-m_c[2]) <= m_hd)
		{
		  zIn = true;
		}
	      else if(cornercoord[2] > m_c[2])
		{
		  HaveZAbove = true;
		}
	      else
		{
		  HaveZBelow = true;
		}
	    }

	  if(cornercoord[1] >= m_c[1])
	    {
	      if(cornercoord[1] > yofx(cornercoord))
		{
		  HaveYAbove = true;
		}
	      else
		{
		  yIn = true;
		}
	    }
	  else
	    {
	      if(cornercoord[1] < yofx(cornercoord))
		{
		  HaveYBelow = true;
		}
	      else
		{
		  yIn = true;
		}
	    }

	  if(zIn && yIn)
	    {
	      covered = false;
	    }
	}

      if(covered && ((HaveZAbove && HaveZBelow) || (HaveYAbove && HaveYBelow)))
	{
	  covered = false;
	}
    }

  return covered;

}

pair<int, Side::LoHiSide> AndyChannel::upDirection(const RealVect& a_midpt,
						   const IntVect&  a_iv) const
{
  pair<int, Side::LoHiSide> retval;

  // find the distance to the boundary in each of the three directions.
  // upDirection returns which wall we're closest to
  Real dx,dy,dz;

  // if we are in 3-D, this deals with the z dimension
  if(SpaceDim == 3)
    {
      if(a_midpt[2] >= m_c[2])
	{
	  dz = (m_c[2] + m_hd) - a_midpt[2];
	}
      else
	{
	  dz = a_midpt[2] - (m_c[2] - m_hd);
	}
    }
  else
    {
      dz = 1.0e20;
    }

  // dy depends on where we are in the x direction
  dy = yofx(a_midpt) - a_midpt[1];


  // dx depends on where we are in the y direction
  // if there is no x boundary
  if(ABS(m_c[1] - a_midpt[1]) <= m_hw)
    {
      dx = 1.0e20;
    }
  // if dx will hit something
  else if(ABS(m_c[1] - a_midpt[1]) < m_hW)
    {
      dx = xofy(a_midpt) - a_midpt[0];
    }
  // if dx is out of fluid region
  else
    {
      dx = 1.0e20;
    }

  // determine the LoHiSide
  // if dx is the smallest
  if(ABS(dx) <= MIN(ABS(dy),ABS(dz)))
      {
	retval.first = 0;
	retval.second = Side::Lo;
      }
  // if dy is the smallest
  else if(ABS(dy) <= ABS(dz))
    {
      retval.first = 1;
      if(a_midpt[1] >= m_c[1])
	{
	  retval.second = Side::Lo;
	}
      else
	{
	  retval.second = Side::Hi;
	}
    }
  // if dz is the smallest
  else
    {
      retval.first = 2;
      if(a_midpt[2] >= m_c[2])
	{
	  retval.second = Side::Lo;
	}
      else
	{
	  retval.second = Side::Hi;
	}

    }

  return retval;
}

Real AndyChannel::localFuncValue(const RealVect& a_independentCoords,
			     const int&      a_updirection,
			     const IntVect&  a_iv,
			     const Box&      a_domain,
			     const RealVect& a_origin,
			     const Real&     a_dx) const
{
  Real retval;

  switch(a_updirection)
    {
    case 0:
      retval = xofy(a_independentCoords);
      break;
    case 1:
      retval = yofx(a_independentCoords);
      break;
    case 2:
      if(a_independentCoords[2] >= m_c[2])
	{
	  retval = m_c[2] + m_hd;
	}
      else
	{
	  retval = m_c[2] - m_hd;
	}
      break;
    }

  return retval;
}

BaseLevelSet* AndyChannel::new_baseLevelSet() const
{
  AndyChannel* tcPtr = new AndyChannel(m_hW,m_hw,m_hd,m_R,m_r,m_c);
  return static_cast<BaseLevelSet*>(tcPtr);
}


// returns the coordinate of the closest y value for a given (x,y)
Real AndyChannel::yofx(const RealVect& vec) const
{
  // translate coordinate around center, m_c
  Real x = vec[0]-m_c[0];
  Real y = vec[1]-m_c[1];

  Real retval = 0;
  Real temp = 0;

  if(x < -m_R) // large channel
    {
      retval = m_hW;
    }
  else if(x < 0) // big circle
    {
      x += m_R;
      temp = sqrt(m_R*m_R - x*x);
      retval = m_hW - m_R + temp;
    }
  else if(x < m_r) // small circle
    {
      x = m_r - x;
      temp = sqrt(m_r*m_r - x*x);
      retval = m_hw + m_r - temp;
    }
  else // small channel
    {
      retval = m_hw;
    }

  if(y >= 0)
    {
      return (m_c[1]+retval);
    }
  else
    {
      return (m_c[1]-retval);
    }
}

// returns the coordinate of the closest x value for a given (x,y)
Real AndyChannel::xofy(const RealVect& vec) const
{
  // translate coordinate around center, m_c
  //  Real x = vec[0]-m_c[0];
  Real absy = ABS(vec[1]-m_c[1]);

  Real retval = 0;
  Real temp;

  if(absy > m_hW || absy < m_hw) // no x value to return
    {
      std::cerr << "AndyChannel::xofy called with ABS(y) out of range."
		<< std::endl 
		<< "x= " << vec[0] << endl
		<< "y= " << vec[1] << endl;
      if(SpaceDim == 3)
	{
		std::cerr << "z = " << vec[2] << std::endl;
	}
      exit(0);
    }

  // wall
  if(absy >= m_hw+m_r && absy <= m_hW-m_R)
    {
      retval = m_c[0];
    }
  // large circle
  else if(absy > m_hW-m_R)
    {
      absy -= (m_hW - m_R);
      temp = sqrt(m_R*m_R - absy*absy);
      retval = m_c[0] - m_R + temp;
    }
  // small circle
  else
    {
      absy -= (m_hw + m_r);
      temp = sqrt(m_r*m_r - absy*absy);
      retval = m_c[0] +m_r - temp;
    }

  return retval;
}

Real AndyChannel::ABS(const Real& num) const
{
  if(num >= 0)
    return num;
  else
    return (-num);
}
