/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _BaseLevelSet_H_
#define _BaseLevelSet_H_

#include <cmath>
#include <cstdlib>
#include "REAL.H"
#include "LoHiSide.H" 
#include "IntVect.H"
#include "Vector.H"
#include "RealVect.H"
#include "Box.H"

using std::pair;
/// 
/**
   This is a pure base class that encapsulates
   the description of a surface as a locally-defined
   function.  This is the class that Workshop uses
   to create embedded boundary descriptions of geometry
   from functional descriptions.
*/
class BaseLevelSet
{
public:

  ///
  BaseLevelSet();


  ///
  virtual ~BaseLevelSet();


  /// 
  /** 
      Return true if every cell in region is regular at the
      refinement described by dx.
  */
  virtual bool isRegular(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const = 0;

  /// 
  /** 
      Return true if every cell in region is covered at the
      refinement described by dx.
  */
  virtual bool isCovered(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const = 0;


  ///
  /**
     This returns the signed integer which  most closely
     represents the normal direction.  The first integer
     of the pair is 
     of the interface at an irregular cell (which coordinate
     direction has the largest normal component). 
     This will only be called
     if the cell is irregular.
  */
  virtual pair<int, Side::LoHiSide> 
  upDirection(const RealVect& a_midpt, const IntVect& a_iv) const  = 0;

  ///
  /**
     Return the value at the dependent coordinate given the 
     independent coordinates.  
  */
  virtual Real localFuncValue(const RealVect& a_independentCoords, 
                              const int& a_upDirection, 
                              const IntVect& a_iv, 
                              const Box& a_domain, 
                              const RealVect& a_origin, 
                              const Real& a_dx) const = 0;  

  ///
  /**
     Return a newly allocated derived class.  The responsibility
     for deleting the memory is left to the calling function.
  */
  virtual
  BaseLevelSet* new_baseLevelSet() const = 0;       

  Vector<RealVect>
  getBoxLocations(const Box& a_region,
                  const RealVect& a_origin,
                  const Real& a_dx) const;
  ///
  Vector<IntVect>
  getBoxCornerCells(const Box& a_region,
                    const RealVect& a_origin,
                    const Real& a_dx) const;


  Vector<RealVect>
  getBoxLocations(const Box& a_region,
                  const RealVect& a_origin,
                  const RealVect& a_vectDx) const;

  void setUseLevelSet(bool a_useLevelSet);

  bool getUseLevelSet() const;
  ///
private:

  ///default is false.  set to true in your
  ///derived class if you want it to be true
  bool m_useLevelSet;

};
#endif
