/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG


#include "parstream.H"

#include "BaseLevelSet.H"
#include "Box.H"

bool BaseLevelSet::getUseLevelSet() const
{
  return m_useLevelSet;
}

void BaseLevelSet::setUseLevelSet(bool a_useLevelSet)
{
  m_useLevelSet = a_useLevelSet;
}

BaseLevelSet::BaseLevelSet()
{
  m_useLevelSet = false;
}

BaseLevelSet::~BaseLevelSet()
{
}
/*****/
/*****/
Vector<IntVect>
BaseLevelSet::getBoxCornerCells(const Box& a_region,
                                const RealVect& a_origin,
                                const Real& a_dx) const
{
#if CH_SPACEDIM==2
  int npoints = 4;
#elif CH_SPACEDIM==3
  int npoints = 8;
#else
  bogus_ch_spacedim();
#endif
  Vector<IntVect> retvec(npoints);
  //lowest point is boxOrigin;
  int ipt = 0;
  retvec[ipt] = a_region.smallEnd();
  //next two points are boxlen away from origin.
  //also get the offset for the last point
  IntVect diagOffset = IntVect::Zero;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      ipt++;
      retvec[ipt] = a_region.smallEnd()  + BASISV(idir);
      diagOffset += BASISV(idir);
    }
  ipt++;
  retvec[ipt] = a_region.smallEnd()  + diagOffset;
  //thats's it for 2d, 3d has 3 extra points.
  //just list the damned things.
#if CH_SPACEDIM ==3
  IntVect offset1, offset2, offset3;
  offset1 = BASISV(0) + BASISV(1);
  offset2 = BASISV(1) + BASISV(2);
  offset3 = BASISV(2) + BASISV(0);
  ipt++;
  IntVect origin= a_region.smallEnd();
  retvec[ipt] = origin  + offset1;
  ipt++;        
  retvec[ipt] = origin  + offset2;
  ipt++;        
  retvec[ipt] = origin  + offset3;
#endif
  return retvec;
}

/*****/
/*****/

Vector<RealVect>
BaseLevelSet::getBoxLocations(const Box& a_region,
                              const RealVect& a_origin,
                              const Real& a_dx) const
{
  const IntVect& loiv = a_region.smallEnd();
  const IntVect& hiiv = a_region.bigEnd();
  RealVect boxLength;
  RealVect boxOrigin;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      boxLength[idir] = a_dx*Real(hiiv[idir] - loiv[idir] + 1);
      boxOrigin[idir] = a_dx*(Real(loiv[idir])) + a_origin[idir];
    }
#if CH_SPACEDIM==2
  int npoints = 4;
#elif CH_SPACEDIM==3
  int npoints = 8;
#else
  bogus_ch_spacedim();
#endif

  Vector<RealVect> retvec(npoints);
  //lowest point is boxOrigin;
  int ipt = 0;
  retvec[ipt] = boxOrigin;
  //next two points are boxlen away from origin.
  //also get the offset for the last point
  RealVect diagOffset = RealVect::Zero;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      ipt++;
      RealVect offset = (boxLength[idir]*BASISREALV(idir));
      retvec[ipt] = boxOrigin  + offset;
      diagOffset += offset;
    }
  ipt++;
  retvec[ipt] = boxOrigin  + diagOffset;
  //thats's it for 2d, 3d has 3 extra points.
  //just list the damned things.
#if CH_SPACEDIM ==3
  RealVect offset1, offset2, offset3;
  offset1 = (boxLength[0]*BASISREALV(0)) + (boxLength[1]*BASISREALV(1));
  offset2 = (boxLength[1]*BASISREALV(1)) + (boxLength[2]*BASISREALV(2));
  offset3 = (boxLength[2]*BASISREALV(2)) + (boxLength[0]*BASISREALV(0));
  ipt++;
  retvec[ipt] = boxOrigin  + offset1;
  ipt++;
  retvec[ipt] = boxOrigin  + offset2;
  ipt++;
  retvec[ipt] = boxOrigin  + offset3;
#endif
  return retvec;
}
Vector<RealVect>
BaseLevelSet::getBoxLocations(const Box& a_region,
                              const RealVect& a_origin,
                              const RealVect& a_vectDx) const
{
  const IntVect& loiv = a_region.smallEnd();
  const IntVect& hiiv = a_region.bigEnd();
  RealVect boxLength;
  RealVect boxOrigin;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      boxLength[idir] = a_vectDx[idir]*Real(hiiv[idir] - loiv[idir] + 1);
      boxOrigin[idir] = a_vectDx[idir]*(Real(loiv[idir])) + a_origin[idir];
    }
#if CH_SPACEDIM==2
  int npoints = 4;
#elif CH_SPACEDIM==3
  int npoints = 8;
#else
  bogus_ch_spacedim();
#endif

  Vector<RealVect> retvec(npoints);
  //lowest point is boxOrigin;
  int ipt = 0;
  retvec[ipt] = boxOrigin;
  //next two points are boxlen away from origin.
  //also get the offset for the last point
  RealVect diagOffset = RealVect::Zero;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      ipt++;
      RealVect offset = (boxLength[idir]*BASISREALV(idir));
      retvec[ipt] = boxOrigin  + offset;
      diagOffset += offset;
    }
  ipt++;
  retvec[ipt] = boxOrigin  + diagOffset;
  //thats's it for 2d, 3d has 3 extra points.
  //just list the damned things.
#if CH_SPACEDIM ==3
  RealVect offset1, offset2, offset3;
  offset1 = (boxLength[0]*BASISREALV(0)) + (boxLength[1]*BASISREALV(1));
  offset2 = (boxLength[1]*BASISREALV(1)) + (boxLength[2]*BASISREALV(2));
  offset3 = (boxLength[2]*BASISREALV(2)) + (boxLength[0]*BASISREALV(0));
  ipt++;
  retvec[ipt] = boxOrigin  + offset1;
  ipt++;
  retvec[ipt] = boxOrigin  + offset2;
  ipt++;
  retvec[ipt] = boxOrigin  + offset3;
#endif
  return retvec;
}
