/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _ChannelBL_H_
#define _ChannelBL_H_

#include "MayDay.H"
#include "LoHiSide.H"
#include "IntVect.H"
#include "RealVect.H"
#include "BaseLevelSet.H"
#include "REAL.H"

using std::pair;

/// 
/**
   This is the workshop class for a channel whose center place
   has an alpha = 0.
   nx*x + ny*y + nz*z = 0
   It follows the BaseLevelSet base class.
 */
class ChannelBL: public BaseLevelSet
{
public:

  ///
  /**
   Returns nx*x + ny*y + nz*z = 0
  */
  ChannelBL(const RealVect& a_centerPlaneNormal,
         const Real&     a_radius);

  ///
  virtual ~ChannelBL();

  /// 
  /** 
      Return true if every cell in region is regular at the
      refinement described by dx.
  */
  virtual bool isRegular(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const;

  /// 
  /** 
      Return true if every cell in region is covered at the
      refinement described by dx.
  */
  virtual bool isCovered(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const;

  ///
  /**
     This returns the signed integer which  most closely
     represents the normal direction.  The first integer
     of the pair is 
     of the interface at an irregular cell (which coordinate
     direction has the largest normal component). 
     This will only be called
     if the cell is irregular.
   */
  
  pair<int, Side::LoHiSide> upDirection(const RealVect& a_midpt,const IntVect& a_iv) const; 

  ///
  /**
     Return the value at the dependent coordinate given the independent
     coordinates. \\
     Returns rampSlope*(independentCoord[indepVar] - startVal);
   */
  virtual Real localFuncValue(const RealVect& a_independentCoords, 
                              const int& a_upDirection,
                              const IntVect& a_iv, 
                              const Box& a_domain, 
                              const RealVect& a_origin, 
                              const Real& a_dx) const;  


  ///
  /**
     Return a newly allocated derived class.  The responsibility
     for deleting the memory is left to the calling function.
   */
  virtual
  BaseLevelSet* new_baseLevelSet() const;       
  virtual Real  LevelSurface(const RealVect& a_vec) const;
  Real getCenterPlanePt(const RealVect& a_indCoords) const;

protected:

  int m_upDir;

  RealVect m_centerPlaneNormal;
  Real     m_radius;
  Real     m_upDirDistance;

private:
  ///
  ChannelBL()
  {
    MayDay::Error("ChannelBL uses strong construction");
  }
  ChannelBL(const ChannelBL& a_ramplg)
  {
    MayDay::Error("ChannelBL disallows copy construction");
  }
  void operator=(const ChannelBL& a_ramplg)
  {
    MayDay::Error("ChannelBL disallows assignment");
  }

};
#endif
