/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG


#include "ChannelBL.H"
#include "Box.H"
#include "PolyGeom.H"
#include "BoxIterator.H"
/*****/
/*****/
ChannelBL::ChannelBL(const RealVect& a_centerPlaneNormal,
                     const Real&     a_radius)
{
  m_centerPlaneNormal = a_centerPlaneNormal;
  Real sumSquare;
  //make the normal a unit vector
  PolyGeom::unifyVector(m_centerPlaneNormal, sumSquare);
  m_radius = a_radius;

  //pick updir to be direction with biggest normal
  int ibigNorm = 0;
  int ismallNorm = 0;
  for(int idir = 1; idir < SpaceDim; idir++)
    {
      if(Abs(m_centerPlaneNormal[idir]) > Abs(m_centerPlaneNormal[idir-1]))
        {
          ibigNorm = idir;
        }
      if(Abs(m_centerPlaneNormal[idir]) < Abs(m_centerPlaneNormal[idir-1]))
        {
          ismallNorm = idir;
        }
    }

  m_upDir = ibigNorm;
  
  //  Real theta = acos(Abs(1.0/m_centerPlaneNormal[ibigNorm]));
  //  m_upDirDistance = m_radius/cos(theta);
  m_upDirDistance = m_radius/Abs(m_centerPlaneNormal[ibigNorm]);
}
/*****/
/*****/
ChannelBL::~ChannelBL()
{
}
/*****/
/*****/
bool 
ChannelBL::isRegular(const Box& a_region,
                     const Box& a_domain,
                     const RealVect& a_origin,
                     const Real& a_dx) const
{
  //create the locations of each corner of box
  //if every corner is above the face, the box is regular
  bool allRegular = true;
  Real boxLen = 0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real dirlen = a_dx*(a_region.size(idir));
      boxLen += dirlen*dirlen;
    }
  boxLen = sqrt(boxLen);
  if(boxLen > 2.*m_radius)
    {
      allRegular = false;
    }
  else
    {
      BoxIterator bit(a_region);
      for (bit.reset(); bit.ok() && allRegular; ++bit)
        {
          IntVect iv = bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

          // if every corner is above the face the box is covered
          for (int ilocs = 0; (ilocs < cellboxLocs.size()) && allRegular; ilocs++)
            {
              const RealVect& corner = cellboxLocs[ilocs];

              Real surfVal = LevelSurface(corner);
              // if every corner is interior, the box is regular
              if (surfVal <= 0.) // if levelSurface > 0, then a_vec is in the fluid
                {
                  allRegular = false;
                }
            }
        }
    }
  return allRegular;
}
/*****/
/*****/
bool 
ChannelBL::isCovered(const Box& a_region,
                     const Box& a_domain,
                     const RealVect& a_origin,
                     const Real& a_dx) const
{
  bool allCovered = true;
  Real boxLen = 0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real dirlen = a_dx*(a_region.size(idir));
      boxLen += dirlen*dirlen;
    }
  boxLen = sqrt(boxLen);
  if(boxLen > 2.*m_radius)
    {
      allCovered = false;
    }
  else
    {
      BoxIterator bit(a_region);
      for (bit.reset(); bit.ok() && allCovered; ++bit)
        {
          IntVect iv = bit();
          Box cellbox(iv,iv);
     
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

          // if every corner is above the face the box is covered
          for (int ilocs = 0; (ilocs<cellboxLocs.size())& allCovered;ilocs++)
            {
              const RealVect& corner = cellboxLocs[ilocs];
              Real surfVal = LevelSurface(corner);
              // if levelSurface > 0, then a_vec is in the fluid
              if (surfVal > 0.) 
                {
                  allCovered = false;
                }
            }
        }
    }
  return allCovered;
}
/*****/
/*****/
pair<int, Side::LoHiSide> 
ChannelBL::upDirection(const RealVect& a_midpt,const IntVect& a_iv) const 
{
  
  pair<int, Side::LoHiSide> retval;
  retval.first = m_upDir;
  Real centerval = getCenterPlanePt(a_midpt);
  if(a_midpt[m_upDir] > centerval)
    {
      retval.second = Side::Lo;
    }
  else
    {
      retval.second = Side::Hi;
    }
  return retval;
}
/*****/
/*****/
Real 
ChannelBL::localFuncValue(const RealVect& a_indCoords, 
                          const int& a_upDir,
                          const IntVect& a_iv, 
                          const Box& a_domain, 
                          const RealVect& a_origin, 
                          const Real& a_dx) const 
{
  assert(a_upDir == m_upDir);
  Real centerval = getCenterPlanePt(a_indCoords);

  Real retval;
  //if above axis, add updir distance.
  //if below axis, subtract
  if(a_indCoords[m_upDir] > centerval)
    {
      retval = centerval + m_upDirDistance;
    }
  else
    {
      retval = centerval - m_upDirDistance;
    }
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
ChannelBL::new_baseLevelSet() const       
{
  ChannelBL* rampPtr = new ChannelBL(m_centerPlaneNormal,
                                     m_radius);
  return static_cast<BaseLevelSet*>(rampPtr); 
}
/********************/
/********************/
Real 
ChannelBL::getCenterPlanePt(const RealVect& a_indCoords) const
{
  //alpha is zero for the center plane
  Real centerval = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(idir != m_upDir)
        {
          centerval -= m_centerPlaneNormal[idir]*a_indCoords[idir];
        }
    }
  centerval /= m_centerPlaneNormal[m_upDir];
  return centerval;
}
/********************/
/********************/
Real 
ChannelBL::LevelSurface(const RealVect& a_vec) const
{
  // if the return value is positive then a_vec is in the fluid 
  // if retval is negative, a_vec is covered.
  Real centerPt = getCenterPlanePt(a_vec);

  Real localval = a_vec[m_upDir];
  localval -= centerPt;

  Real retval =  m_upDirDistance - Abs(localval);
  return retval;
}
/********************/
/********************/
