#include "ChannelLS.H"

#ifndef MIN
#define MIN(A,B) (((A)<(B))?(A):(B))
#endif
#ifndef MAX
#define MAX(A,B) (((A)>(B))?(A):(B))
#endif

//                                      covered region
//   y-cy = halfbigwidth
//   --------------------------)  radius bigradius
//                              )
//                              |
//                              |
//                              ( radius smallradius
//    fluid                      \            y-cy = smallwidth
//                                (---------------------------
//
// -----------------------------c-----------------------------
//          c = center of problem, symmetric in y

// the function TrebChallenLS::yofxz() computes the height of the
// surface in center-relative coordinates, etc.


ChannelLS::ChannelLS(const Real& a_halfbigwidth,
                             const Real& a_halfsmallwidth,
                             const Real& a_halfdepth,
                             const Real& a_bigradius,
                             const Real& a_smallradius,
                             const Real& a_zradius,
                             const RealVect& a_center)
{
  m_halfbigwidth    = a_halfbigwidth;
  m_halfsmallwidth  = a_halfsmallwidth;
  m_halfdepth       = a_halfdepth;
  m_bigradius       = a_bigradius;
  m_smallradius     = a_smallradius;
  m_zradius         = a_zradius;
  m_center          = a_center;

  // check for valid input
  if( m_halfbigwidth <= 0 ) {
    cout << "ChannelLS bad input halfbigwidth " << m_halfbigwidth
      << endl;
    exit(0);
  }
  if( m_halfsmallwidth <= 0 ) {
    cout << "ChannelLS bad input halfsmallwidth " << m_halfsmallwidth
      << endl;
    exit(0);
  }
  if( m_halfdepth <= 0 ) {
    cout << "ChannelLS bad input halfdepth " << m_halfdepth
      << endl;
    exit(0);
  }
  if( m_bigradius <= 0 ) {
    cout << "ChannelLS bad input bigradius " << m_bigradius
      << endl;
    exit(0);
  }
  if( m_smallradius <= 0 ) {
    cout << "ChannelLS bad input smallradius " << m_smallradius
      << endl;
    exit(0);
  }
  if( m_bigradius <= m_zradius ) {
    cout << "ChannelLS bad input bigradius<zradius " << endl;
    exit(0);
  }
  if( m_zradius <= 0 ) {
    cout << "ChannelLS bad input zradius " << m_zradius
      << endl;
    exit(0);
  }
  if( m_halfsmallwidth + m_smallradius + m_bigradius > m_halfbigwidth ) {
    cout << "ChannelLS inconsistent geometry input" << endl;
    exit(0);
  }
}

ChannelLS::~ChannelLS()
{}

bool ChannelLS::isRegular(const Box&      region,
                              const Box&      domain,
                              const RealVect& origin,
                              const Real&     dx) const
{
  bool allIn = true;

  // if 1D, no geometry -- all regular
  if( SpaceDim == 1 ) return allIn;

  BoxIterator bit(region);
  for( bit.reset(); bit.ok() && allIn; ++bit ) {
    IntVect iv = bit();
    Box cellbox(iv,iv);
    Vector<RealVect> boxloc = getBoxLocations(cellbox,origin,dx);

    int ncorner = boxloc.size();
    for( int icorner=0; icorner<ncorner && allIn; icorner++ ) {
      const RealVect& cornercoord = boxloc[icorner];

      pair<Real,bool> zval = zofxy(cornercoord[0]-m_center[0],
                                   cornercoord[1]-m_center[1]);
      if( !zval.second  ) {
        allIn = false;
      } else if( SpaceDim == 3 ) {
        if( fabs(cornercoord[2]-m_center[2]) > zval.first ) allIn = false;
      }
    }
  }

#ifdef DEBG
  cout << "ChannelLS::isRegular " << region << " " << allIn << endl;
#endif

  return allIn;
}

bool ChannelLS::isCovered(const Box&      region,
                              const Box&      domain,
                              const RealVect& origin,
                              const Real&     dx) const
{
  // in 1D no geometry, so never covered
  if( SpaceDim == 1 ) return false;

  bool allOut = true;
  BoxIterator bit(region);
  for( bit.reset(); bit.ok() && allOut; ++bit ) {
    IntVect iv = bit();
    Box cellbox(iv,iv);
    Vector<RealVect> boxloc = getBoxLocations(cellbox,origin,dx);

    int ncorner = boxloc.size();
    for( int icorner=0; icorner<ncorner && allOut; icorner++ ) {
      const RealVect& cornercoord = boxloc[icorner];

      pair<Real,bool> zval = zofxy(cornercoord[0]-m_center[0],
                                   cornercoord[1]-m_center[1]);

      if( zval.second ) {
        if( SpaceDim==3 ) {
          if( fabs(cornercoord[2]-m_center[2]) < zval.first ) {
            allOut = false;
          }
        } else {
          allOut = false;
        }
      }
    }
  }

#ifdef DEBG
  cout << "ChannelLS::isCovered " << region << " " << allOut << endl;
#endif

  return allOut;
}

// up Direction points into fluid
pair<int, Side::LoHiSide> ChannelLS::upDirection(
           const RealVect& a_midpt,
           const IntVect&  a_iv) const
{
  pair<int, Side::LoHiSide> retval;

  if( SpaceDim == 3 ) {
    pair<Real,bool> zval = zofxy(a_midpt[0]-m_center[0],
                                 a_midpt[1]-m_center[1]);
    if( zval.second ) {
      if( zval.first > m_halfdepth + (M_SQRT1_2-1.0)*m_zradius ) {
	// z can be used as a dependent variable in this neighborhood
        retval.first = 2;
        retval.second = (a_midpt[2]-m_center[2]>0) ? Side::Lo : Side::Hi ;
  
#ifdef DEBG
  cout << "ChannelLS::upDirection cell " << a_iv <<
     " point " << a_midpt << endl;
  cout << "  return " << retval.first << " " << retval.second << endl;
#endif
	return retval;
      }
    }
  }

  Real dely = fabs(a_midpt[1]-m_center[1]);
  if( dely > m_halfsmallwidth-(M_SQRT1_2-1.0)*m_smallradius &&
      dely < m_halfbigwidth+(M_SQRT1_2-1.0)*m_bigradius ) {
    // x can be used as a dependent variable in this neighborhood
    retval.first = 0;
    retval.second = Side::Lo;
  }

  retval.first = 1;
  retval.second = (a_midpt[1]-m_center[1]>0) ? Side::Lo : Side::Hi ;
  
#ifdef DEBG
  cout << "ChannelLS::upDirection cell " << a_iv <<
     " point " << a_midpt << endl;
  cout << "  return " << retval.first << " " << retval.second << endl;
#endif

  return retval;
}


// all elements of a_independentCoords are valid.
// arguements iv,domain,origin,dx are not required
Real ChannelLS::localFuncValue(const RealVect& a_independentCoords,
                                   const int&      a_updirection,
                                   const IntVect&  a_iv,
                                   const Box&      a_domain,
                                   const RealVect& a_origin,
                                   const Real&     a_dx) const
{
#ifdef DEBG
  cout << "ChannelLS::localFuncValue " << a_independentCoords <<
  " " << a_updirection << endl;
#endif

  Real retval;
  Real zcoord = (SpaceDim==3) ? a_independentCoords[2]-m_center[2] : 0;
  pair<Real,bool> calccoord;
  switch( a_updirection ) {
  case 0:
    calccoord = xofyz(a_independentCoords[1]-m_center[1],zcoord);
    if( !calccoord.second ) {
      cout << "localFuncValue x input invalid" << endl;
      cout << "  independentCoords = " << a_independentCoords << endl;
      cout << " iv = " << a_iv << endl;
      exit(0);
    }
    retval = m_center[0] + calccoord.first;
    break;
  case 1:
    calccoord = yofxz(a_independentCoords[0]-m_center[0],zcoord);
    if( !calccoord.second ) {
      cout << "localFuncValue y input invalid" << endl;
      cout << " independentCoords = " << a_independentCoords << endl;
      cout << " iv = " << a_iv << endl;
      exit(0);
    }
    if( a_independentCoords[1] > m_center[1] ) {
      retval = m_center[1] + calccoord.first;
    } else {
      retval = m_center[1] - calccoord.first;
    }
    break;
  case 2:
    // the z direction is also kind of easy
    calccoord = zofxy(a_independentCoords[0]-m_center[0],
                      a_independentCoords[1]-m_center[1]);
    if( !calccoord.second ) {
      cout << "localFuncValue z input invalid" << endl;
      cout << "  independentCoords = " << a_independentCoords << endl;
      cout << " iv = " << a_iv << endl;
      exit(0);
    }
    if( a_independentCoords[2] > m_center[2] ) {
      retval = m_center[2]+calccoord.first;
    } else {
      retval = m_center[2]-calccoord.first;
    }
    break;
  }

#ifdef DEBG
  cout << " return " << retval << endl;
#endif

  return retval;
}

BaseLevelSet* ChannelLS::new_baseLevelSet() const
{
  ChannelLS* tcPtr = new ChannelLS(m_halfbigwidth,
                                           m_halfsmallwidth,
                                           m_halfdepth,
                                           m_bigradius,
                                           m_smallradius,
                                           m_zradius,
                                           m_center);
  return static_cast<BaseLevelSet*>(tcPtr);
}

// return abs(y(x,z)) describing the channel.
// all values assumed relative to center.  y value on the + side
// returned.
pair<Real,bool> ChannelLS::yofxz(const Real& x, const Real& z) const
{
  pair<Real,bool> retval;
  retval.second = true; // valid

  Real abz = fabs(z);
  Real margin = 0;
  if( SpaceDim==3 ) {
    if( m_halfdepth >= abz && abz >= m_halfdepth-m_zradius ) {
      Real delz = abz - ( m_halfdepth - m_zradius);
      margin = m_zradius-sqrt(MAX(m_zradius*m_zradius - delz*delz,0.0));
    } else if( abz > m_halfdepth ) {
      retval.second = false;
    }
  }

  // m_bigwidth -> m_bigwidth-margin
  // m_smallwidth -> m_smallwidth-margin
  // m_bigradius -> m_bigradius-margin
  // center point of big radius
  //  (-m_bigradius,m_bigwidth-m_bigradius) ->
  //  (-margin-(m_bigradius-margin),(m_bigwidth-margin)-(m_bigradius-margin))
  // = (-m_bigradius,m_bigwidth-m_bigradius) unchanged
  // m_smallradius -> m_smallradius+margin
  // center point of small radius
  //  (m_smallradius,m_smallwidth+m_smallradius) ->
  //  (-margin+(m_smallradius+margin),
  //         (m_smallwidth-margin)+(m_smallradius+margin) ) unchanged


  if( retval.second ) {
    if( x < -m_bigradius ) {

      // in the big width part
      retval.first = (m_halfbigwidth-margin);

    } else if( x < -margin ) {

      // in the big radius part
      Real dx = x + m_bigradius;
      Real y = (m_bigradius-margin)*(m_bigradius-margin) - dx*dx;
      y = m_halfbigwidth - m_bigradius + sqrt(MAX(y,0.0));
      retval.first = y;

    } else if( x < m_smallradius ) {

      // in the small radius part
      Real dx = x - m_smallradius;
      Real y = (m_smallradius+margin)*(m_smallradius+margin) - dx*dx;
      y = m_halfsmallwidth + m_smallradius - sqrt(MAX(y,0.0));
      retval.first = y;

    } else {

      // in the small width part
      retval.first = m_halfsmallwidth-margin;

    }
  }

  return retval;
}

pair<Real,bool> ChannelLS::xofyz(const Real& y, const Real& z) const
{
  pair<Real,bool> retval;
  retval.second = true; // valid

  Real aby = fabs(y);
  Real abz = fabs(z);
  Real margin = 0;
  if( SpaceDim==3 ){
    if( m_halfdepth >= abz && abz >= m_halfdepth-m_zradius ) {
      Real delz = abz - ( m_halfdepth - m_zradius);
      margin = m_zradius-sqrt(MAX(m_zradius*m_zradius - delz*delz,0.0));
    } else if( abz > m_halfdepth ) {
      retval.second = false;
    }
  }


  if( retval.second ) {
    if( aby > m_halfbigwidth-margin ) {
      retval.second = false;
    }
  }

  if( retval.second ) {
    if( aby < m_halfsmallwidth-margin ) {
      retval.second = false;
    }
  }

  if( retval.second ) {
    if( (aby >= m_halfsmallwidth + m_smallradius) &&
        (aby <= m_halfbigwidth - m_bigradius) ) {
      retval.first = -margin;
    } else if( aby >= m_halfbigwidth - m_bigradius ) {
      Real dy = aby - (m_halfbigwidth - m_bigradius);
      Real x = (m_bigradius-margin)*(m_bigradius-margin) - dy*dy;
      x = - m_bigradius + sqrt(MAX(x,0.0));
      retval.first = x;
    } else {
      Real dy = aby - (m_halfsmallwidth + m_smallradius);
      Real x = (m_smallradius+margin)*(m_smallradius+margin) - dy*dy;
      x = m_smallradius - sqrt(MAX(x,0.0));
      retval.first = x;
    }
  }

  return retval;
}

pair<Real,bool> ChannelLS::zofxy(const Real& x, const Real& y) const
{
  pair<Real,bool> retval;
  retval.first = m_halfdepth;
  retval.second = true; // valid
  Real aby = fabs(y);

  if( x < -m_bigradius ) {
    if( aby > m_halfbigwidth ) {
      retval.second = false;
    } else if( aby < m_halfbigwidth - m_zradius ) {
      retval.first = m_halfdepth;
    } else {
      Real margin = aby-(m_halfbigwidth-m_zradius);
      Real z = m_zradius*m_zradius - margin*margin;
      retval.first = m_halfdepth - m_zradius + sqrt(MAX(z,0.0));
    }
  } else if( x > m_smallradius ) {
    if( aby > m_halfsmallwidth ) {
      retval.second = false;
    } else if( aby < m_halfsmallwidth - m_zradius ) {
      retval.first = m_halfdepth;
    } else {
      Real margin = aby-(m_halfsmallwidth-m_zradius);
      Real z = m_zradius*m_zradius - margin*margin;
      retval.first = m_halfdepth - m_zradius + sqrt(MAX(z,0.0));
    }
  } else if( aby >= m_halfsmallwidth+m_smallradius &&
             aby <= m_halfbigwidth-m_bigradius ) {
    if( x > 0 ) {
      retval.second = false;
    } else if( x < -m_zradius ) {
      retval.first = m_halfdepth;
    } else {
      Real margin = x-(-m_zradius);
      Real z = m_zradius*m_zradius - margin*margin;
      retval.first = m_halfdepth - m_zradius + sqrt(MAX(z,0.0));
    }
  } else if( x <= m_smallradius && aby <= m_halfsmallwidth+m_smallradius ) {
    Real delx = x - (m_smallradius);
    Real dely = aby - (m_halfsmallwidth+m_smallradius);
    Real rad = sqrt(MAX(delx*delx+dely*dely,0.0));
    if( rad < m_smallradius ) {
      retval.second = false;
    } else if( rad >= m_smallradius + m_zradius ) {
      retval.first = m_halfdepth;
    } else {
      Real dr = m_smallradius + m_zradius - rad;
      Real z = m_zradius*m_zradius - dr*dr;
      retval.first = m_halfdepth - m_zradius + sqrt(MAX(z,0.0));
    }
  } else if( x >= -m_bigradius && aby >= m_halfbigwidth-m_bigradius ) {
    Real delx = x - (-m_bigradius);
    Real dely = aby - (m_halfbigwidth-m_bigradius);
    Real rad = sqrt(MAX(delx*delx+dely*dely,0.0));
    if( rad > m_bigradius ) {
      retval.second = false;
    } else if( rad <= m_bigradius - m_zradius ) {
      retval.first = m_halfdepth;
    } else {
      Real dr = rad + m_zradius - m_bigradius;
      Real z = m_zradius*m_zradius - dr*dr;
      retval.first = m_halfdepth - m_zradius + sqrt(MAX(z,0.0));
    }
  } else {
    retval.second = false;
  }

  return retval;
}
