/* _______              __ConcentSphere
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "ConcentricSphereBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
/*****/
ConcentricSphereBL::ConcentricSphereBL(const Real& a_bigRadius,
								 const Real& a_smallRadius,
								 const RealVect& a_center,
								 const Real& a_dx)
{
  //  assert(a_smallRadius >= 2*a_dx);
  // assert(dx>polygeom::get.tolerance());
 
  m_bigRadius       = a_bigRadius;
  m_smallRadius       = a_smallRadius;
  m_center       = a_center;
  m_dx           = a_dx;
}
/*****/
/*****/
ConcentricSphereBL::~ConcentricSphereBL()
{
}
/*****/
//finds the component of  a RealVect with largest magnitude
Real ConcentricSphereBL::max(const RealVect& vec)const
  
{
      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<abs(vec[idir]))retval=abs(vec[idir ]);
    }
  return retval;
}

//finds the eulclidean distance of a RealVect from the origin

Real ConcentricSphereBL::magnitude( const RealVect& vec)const
{ 
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }    
  retval=sqrt(retval);
  return retval;
}
 
bool 
ConcentricSphereBL::isRegular(const Box&           a_region,
							  const Box&           a_domain,
							  const RealVect&      a_origin,
							  const Real&          a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if regular
  bool inTheFluid = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok() && inTheFluid;++bit)
    {
      if (inTheFluid == true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);

          //if every corner is in the fluid the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && inTheFluid; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              RealVect translatedCorner = corner;
              translatedCorner -= m_center;
			 
                  if(magnitude(translatedCorner) > m_bigRadius)
					{
					  inTheFluid = false;
					}
                  else if(magnitude(translatedCorner) < m_smallRadius)
					{
					  inTheFluid = false;
					}
			}
        } 
    }
  
  return inTheFluid;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
ConcentricSphereBL::isCovered(const Box&      a_region,
						   const Box&      a_domain,
						   const RealVect& a_origin,
						   const Real&     a_dx)const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if covered

  bool outOfTheFluid = true;

  BoxIterator bit(a_region);
  for(bit.reset();bit.ok() && outOfTheFluid;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner is out of the fluid the cell is out of the fluid

      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && outOfTheFluid; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          RealVect translatedCorner = corner;
          translatedCorner -= m_center;
          
		  if(magnitude(translatedCorner) < m_bigRadius && 
			 magnitude(translatedCorner) > m_smallRadius)
			{
			  outOfTheFluid = false;
			}
		}
    }
  
  return outOfTheFluid;
}

pair<int, Side::LoHiSide> ConcentricSphereBL::upDirection(const RealVect& midPt,
													   const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;
  Real averageRadius = (m_bigRadius + m_smallRadius)/2.;
  RealVect gradient;
  //On a ConcentSphere, gradient porportional to position realative to the center
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      gradient[idir] = midPt[idir] - m_center[idir];
    }   
  
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (abs(gradient[idir]) >= max(gradient)) 
		{		
		  retval.first=idir;
		}
    }
  
  if (gradient[retval.first] >=0. && magnitude(gradient) > averageRadius)
    {
      retval.second=Side::Lo;
    }
  else  if (gradient[retval.first] < 0.&& magnitude(gradient) > averageRadius)
    {
      retval.second=Side::Hi;
    }
  else if (gradient[retval.first] >= 0.&& magnitude(gradient) <= averageRadius )
    {
      retval.second=Side::Hi;
    }
  else  if (gradient[retval.first] < 0.&& magnitude(gradient) <= averageRadius)
    {
      retval.second=Side::Lo;
    }
  return retval;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Real 
ConcentricSphereBL::localFuncValue(const RealVect& a_indCoords, 
								const int& a_upDir,
								const IntVect& a_iv, 
								const Box& a_domain, 
								const RealVect& a_origin, 
								const Real& a_dx) const
{
  Real averageRadius = (m_smallRadius + m_bigRadius)/2.;
  Real distToCenter = 0.;
  for(int idir = 0; idir< SpaceDim; idir++)
    {
	  distToCenter += pow(a_indCoords[idir] - m_center[idir],2);
	}
  distToCenter = sqrt(distToCenter);
  Real retval =0.;
  for(int idir = 0; idir< SpaceDim; idir++)
    {
      if (idir !=a_upDir)
        {
          retval += pow(a_indCoords[idir] - m_center[idir],2);
        }       
    }

  if (a_dx*a_iv[a_upDir] >= m_center[a_upDir])
    {
	  if (distToCenter <= averageRadius)
		{
		  retval = sqrt(Abs(pow(m_smallRadius,2) - retval)) + m_center[a_upDir];
		}
	  else	
		{
		   retval = sqrt(Abs(pow(m_bigRadius,2)-retval)) + m_center[a_upDir];
		}
    }
  else
	{
	  if (distToCenter <= averageRadius)

	  {
		retval = -sqrt(Abs(pow(m_smallRadius,2)-retval)) + m_center[a_upDir];
	  }
	  else
		{
		  retval = -sqrt(Abs(pow(m_bigRadius,2) -retval)) + m_center[a_upDir];
		}
	}
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
ConcentricSphereBL::new_baseLevelSet() const       
{
  ConcentricSphereBL* concentricSpherePtr = new ConcentricSphereBL(m_bigRadius,
																   m_smallRadius,
																   m_center,
																   m_dx);
    
  return static_cast<BaseLevelSet*>(concentricSpherePtr); 
}
