/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _ContractionChannelBL_H_
#define _ContractionChannelBL_H_

#include "MayDay.H"
#include "LoHiSide.H"
#include "IntVect.H"
#include "RealVect.H"
#include "BaseLevelSet.H"
#include "REAL.H"

using std::pair;

/// 
/**
   This is the workshop class for a ramp.
   It follows the BaseLevelSet base class.
 */
class ContractionChannelBL: public BaseLevelSet
{
public:

  ///
  /**
   Returns rampSlope*(independentCoord[indepVar] - startVal);
   in 3D this makes a 2D ramp.
  */
  ContractionChannelBL(const RealVect& a_origin,
		       const int&  a_upDir,
		       const int&  a_indepVar, 
		       const Real& a_start,
		       const Real& a_height);

  ///
  virtual ~ContractionChannelBL();

  /// 
  /** 
      Return true if every cell in region is regular at the
      refinement described by dx.
  */
  virtual bool isRegular(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const;

  /// 
  /** 
      Return true if every cell in region is covered at the
      refinement described by dx.
  */
  virtual bool isCovered(const Box& region,
                         const Box& domain,
                         const RealVect& origin,
                         const Real& dx) const;

  ///
  /**
     This returns the signed integer which  most closely
     represents the normal direction.  The first integer
     of the pair is 
     of the interface at an irregular cell (which coordinate
     direction has the largest normal component). 
     This will only be called
     if the cell is irregular.
   */
  
  pair<int, Side::LoHiSide> upDirection(const RealVect& a_midpt,const IntVect& a_iv) const; 

  ///
  /**
     Return the value at the dependent coordinate given the independent
     coordinates. \\
     Returns rampSlope*(independentCoord[indepVar] - startVal);
   */
  virtual Real localFuncValue(const RealVect& a_independentCoords, 
                              const int& a_upDirection,
                              const IntVect& a_iv, 
                              const Box& a_domain, 
                              const RealVect& a_origin, 
                              const Real& a_dx) const;  


  ///
  /**
     Return a newly allocated derived class.  The responsibility
     for deleting the memory is left to the calling function.
   */
  virtual
  BaseLevelSet* new_baseLevelSet() const;       

protected:
  //direction of largest normal
  pair<int, Side::LoHiSide>  m_upDir;
  int  m_indepVar;
  Real m_start;
  Real m_height;
  RealVect m_origin;

private:
  ///
  ContractionChannelBL()
  {
    MayDay::Error("ContractionChannelBL uses strong construction");
  }
  ContractionChannelBL(const ContractionChannelBL& a_ramplg)
  {
    MayDay::Error("ContractionChannelBL disallows copy construction");
  }
  void operator=(const ContractionChannelBL& a_ramplg)
  {
    MayDay::Error("ContractionChannelBL disallows assignment");
  }

};
#endif
