/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "Box.H"
#include "CylinderBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"

CylinderBL::CylinderBL(const Real&     a_radius,
                       const Real&     a_height,
                       const RealVect& a_center,
                       const int&      a_axis,
                       const bool&     a_interior,
                       const Real&     a_dx)
{
  assert(a_radius >= 2*a_dx);
 
  assert(a_dx > PolyGeom::getTolerance());
  assert(a_height >= 0.);

  assert(a_axis >= 0 && a_axis < SpaceDim);
 
  m_radius   = a_radius;
  m_height   = a_height;
  m_center   = a_center;
  m_axis     = a_axis;
  m_interior = a_interior;
  m_dx       = a_dx;
}

CylinderBL::~CylinderBL()
{
}

Real CylinderBL::LevelSurface(const RealVect& a_vec) const
{
  // if the return value is positive then a_vec is in the fluid (therefore
  // the answer is adjusted depending on whether the problem is interior
  // or exterior)

  Real retval = 0;

  RealVect v;

  v = a_vec;
  v -= m_center;

  if (Abs(v[m_axis]) <= m_height/2.)
    {
      retval = m_radius*m_radius;
      for (int i = 0; i < SpaceDim; i++)
      {
        if (i != m_axis)
        {
          retval -= v[i]*v[i];
        }
      }
    }
  else if (v[m_axis] > 0)
    {
      v[m_axis] -= m_height/2.;

      retval = m_radius*m_radius;
      for (int i = 0; i < SpaceDim; i++)
      {
        retval -= v[i]*v[i];
      }
    }  
  else if (v[m_axis] < 0)
    {
      v[m_axis] += m_height/2.;
      
      retval = m_radius*m_radius;
      for (int i = 0; i < SpaceDim; i++)
      {
        retval -= v[i]*v[i];
      }
    }
  else 
    {
      MayDay::Error("missed a case");
    }

  if (!m_interior)
    {
      retval = -retval;
    }

  return retval;
}

bool CylinderBL::isRegular(const Box&      a_region,
                           const Box&      a_domain,
                           const RealVect& a_origin,
                           const Real&     a_dx) const
{
  // iterate over each cell in the box to check if regular
  bool allRegular = true;
 

  BoxIterator bit(a_region);
  for (bit.reset(); bit.ok() && allRegular; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

      // if every corner is above the face the box is covered
      for (int ilocs = 0; (ilocs < cellboxLocs.size()) && allRegular; ilocs++)
        {
          const RealVect& corner = cellboxLocs[ilocs];

          // if every corner is interior, the box is regular
          if (LevelSurface(corner) <= 0.) // if levelSurface > 0, then a_vec is in the fluid
            {
              allRegular = false;
            }
        }
    }
  return allRegular;
}

bool CylinderBL::isCovered(const Box&      a_region,
                           const Box&      a_domain,
                           const RealVect& a_origin,
                           const Real&     a_dx) const
{
  // iterate over each cell in the box to check if covered
  bool allCovered = true;
  BoxIterator bit(a_region);
  for (bit.reset(); bit.ok() && allCovered; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
     
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

      // if every corner is above the face the box is covered
      for (int ilocs = 0; (ilocs < cellboxLocs.size()) && allCovered; ilocs++)
        {
          const RealVect& corner = cellboxLocs[ilocs];

          
         
          if (LevelSurface(corner) > 0.) // if levelSurface > 0, then a_vec is in the fluid
            {
              allCovered = false;
            }
        }
    }
  return allCovered;
}

pair<int, Side::LoHiSide> CylinderBL::upDirection(const RealVect& a_midPt, const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;

  // now we calculate a vector porportional to the gradient and equivalent 
  // to the gradient for our purposes
  RealVect gradient;
  if (Abs(a_midPt[m_axis] - m_center[m_axis]) <= m_height/2.)
    { 
      gradient = a_midPt;
      gradient -= m_center;

      gradient[m_axis] = 0.;
    }
  else if (a_midPt[m_axis] > m_center[m_axis])
    {
      RealVect capCenter = m_center;
      capCenter[m_axis] += m_height/2.;

      gradient = a_midPt;
      gradient -= capCenter;
    }
  else if (a_midPt[m_axis] < m_center[m_axis])
    {
      RealVect capCenter = m_center;
      capCenter[m_axis] -= m_height/2.;

      gradient = a_midPt;
      gradient -= capCenter;
    }
  else
    {
      MayDay::Error("Missed a case:CylinderBL::upDirection");
    }

  // We have the gradient. Now find the largest component.
  Real max = 0.;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      if (Abs(gradient[idir]) >= max) 
        {
          retval.first = idir;
          max = Abs(gradient[idir]);
        }
    }

  // Which way is up?
  if (gradient[retval.first] >= 0. && m_interior)
    {
      retval.second = Side::Lo;
    }
  else if (gradient[retval.first] < 0. && m_interior)
    {
      retval.second = Side::Hi;
    }
  else if (gradient[retval.first] >= 0. && !m_interior)
    {
      retval.second = Side::Hi;
    }
  else if (gradient[retval.first] < 0. && !m_interior)
    {
      retval.second = Side::Lo;
    }
  else
    {
      MayDay::Error("Missed a case:CylinderBL::upDirection");
    }

  return retval;
}

Real CylinderBL::localFuncValue(const RealVect& a_indCoords, 
                                const int&      a_upDir,
                                const IntVect&  a_iv, 
                                const Box&      a_domain, 
                                const RealVect& a_origin, 
                                const Real&     a_dx) const
{
  Real retval = 0.;
  RealVect v;
  v = a_indCoords;
  v -= m_center;

  if (Abs(v[m_axis]) <= m_height/2.)
    {
      Real arg;

#if (CH_SPACEDIM == 2)
      arg = 0.;
#elif (CH_SPACEDIM == 3)
      int indDir = 3 - m_axis - a_upDir;
      arg = a_indCoords[indDir] - m_center[indDir];
#else
      // Only 2D and 3D should be possible
      MayDay::Error("CH_SPACEDIM not 2 or 3!");
#endif

      if (a_indCoords[a_upDir] > m_center[a_upDir])
        {
          assert(m_radius*m_radius-arg*arg > 0.);
          retval = m_center[a_upDir] + sqrt(Abs(m_radius*m_radius-arg*arg));
        }
      else 
        {
          assert(m_radius*m_radius-arg*arg > 0.);
          retval = m_center[a_upDir] - sqrt(m_radius*m_radius-arg*arg);
        }
    }
  else if (v[m_axis] > m_height/2.)
    {
      RealVect capCenter;

      capCenter = m_center;
      capCenter[m_axis] += m_height/2.;

      RealVect pt = a_indCoords;
      pt -= capCenter;
      retval = m_radius*m_radius;

      for (int idir = 0; idir < SpaceDim; ++idir)
        {
          if (idir != a_upDir)
            {            
              retval -= pt[idir]*pt[idir];
            }
        }

      if (a_indCoords[a_upDir] > capCenter[a_upDir])
        {
          assert(retval > 0.);
          retval = sqrt(retval);
        }
      else
        {
          assert(retval > 0.);
          retval = -sqrt(retval);
        }
      retval += capCenter[a_upDir];
    }

  else if (v[m_axis] < -m_height/2.)
    {
      RealVect capCenter;
     
      capCenter = m_center;
      capCenter[m_axis] -= m_height/2.;

      RealVect pt = a_indCoords;
      pt -= capCenter;
      retval = m_radius*m_radius;

      for (int idir = 0; idir < SpaceDim; ++idir)
        {
          if (idir != a_upDir)
            {            
              retval -= pt[idir]*pt[idir];
            }
        }
      if (a_indCoords[a_upDir] > capCenter[a_upDir])
        {
          assert(retval > 0.);
          retval = sqrt(retval);
        }
      else
        {
          assert(retval > 0.);
          retval = -sqrt(retval);
        }
      retval += capCenter[a_upDir];
    }     
  return retval;
}

BaseLevelSet* CylinderBL::new_baseLevelSet() const       
{
  CylinderBL* cylinderPtr = new CylinderBL(m_radius,
                                           m_height,
                                           m_center,
                                           m_axis,
                                           m_interior,
                                           m_dx);
  
  return static_cast<BaseLevelSet*>(cylinderPtr); 
}
