/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include <iostream>
#include <fstream>
#include "Box.H"
#include "DEMBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"


/*****/
/*****/
DEMBL::DEMBL(const IntVect&     a_ncell,
			 const int&         a_interpType,
			 const Real&        a_dx,
			 const RealVect&    a_vectDx)
{
  assert(SpaceDim==3);

  m_ncell        = a_ncell;
  m_dx           = a_dx;
  m_vectDx       = a_vectDx;
  m_interpType   = a_interpType;
  
  // Read in the header info
  bool justhead = true;
  readDEM(justhead);
  
  //allocate memory for DEM matrix
  m_DEM = new Real *[m_ncols];
  for(int i = 0; i < m_ncols;i++)
	{
	  m_DEM[i] = new Real[m_nrows];
	}

  // Read in the whole file
  justhead = false;
  readDEM(justhead);

  //fill in stray "wet" cells, change nodata value to land
  // and convert DEM to meters.
  fixDEM();

 
}
/*****/
/*****/
DEMBL::~DEMBL()
{
  //reallocate memory for DEM matrix
  for(int i=0;i<m_ncols;i++)
	{
	  delete []m_DEM[i];
	}
  delete []m_DEM;
}
/*****/
//finds the component of  a RealVect with largest magnitude
Real DEMBL::max(const RealVect& vec)const
  
{
  pout()<<"check DEMBL::max"<<endl;      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<abs(vec[idir]))retval=abs(vec[idir ]);
    }
  return retval;
}
//finds the eulclidean distance of a RealVect from the origin
Real DEMBL::magnitude( const RealVect& vec)const
{ 
  pout()<<"check DEMBL::magnitude"<<endl;      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }    
  retval=sqrt(retval);
  return retval;
}
 
//Reads in a Digital Elevation Model file
bool DEMBL::readDEM(const bool& a_justhead)
{
  if(SpaceDim!=3)
    {
      MayDay::Error("DEMBL coded for 3D...check spacedim");
    }

  bool fileread;
  //cout<<"Reading DEM Header...\n";
  ifstream DEM_file;
  DEM_file.open("msl1k.asc"); //the input file
  if(DEM_file.bad())
    {
      MayDay::Error("Bad DEM_file in readDEM");
    }
  char astring[14];
  m_ncols = -99;

  DEM_file>>astring;
  DEM_file>>m_ncols;
   
  DEM_file>>astring;
  DEM_file>>m_nrows;

  DEM_file>>astring;
  DEM_file>>m_xllcorner;

  DEM_file>>astring;
  DEM_file>>m_yllcorner;

  DEM_file>>astring;
  DEM_file>>m_cellsize;

  DEM_file>>astring;
  DEM_file>>m_cellvalue;
   
  DEM_file>>astring;
  DEM_file>>m_NODATA;

  //make sure the grid is big enough
  if(m_ncols==-99)
    {
      MayDay::Error("Something wrong with DEM file");
    }

  if(!a_justhead)
    {	
      m_minDEM =  1e20;
      m_maxDEM = -1e20;
      //cout<<"Reading DEM Data...\n";	   
      int i,j;
      for (j=m_nrows-1;j>=0;j--)
        {
          for (i=0;i<m_ncols;i++)
            {
              DEM_file>>m_DEM[i][j];
              if (rint(m_DEM[i][j]) != m_NODATA)
                {
                  if (m_DEM[i][j] < m_minDEM)
                    {
                      m_minDEM = m_DEM[i][j]; //this is the highest spot so far...
                    }
                  if (m_DEM[i][j] > m_maxDEM)
                    {
                      m_maxDEM = m_DEM[i][j]; //this is the deepest spot so far...
                    }
                }
            }
        }
      m_minDEM *=m_cellvalue;
      m_maxDEM *=m_cellvalue;
      pout()<<"  DEM min = "<<m_minDEM<<endl;
      pout()<<"  DEM max = "<<m_maxDEM<<endl;
    }
  DEM_file.close();

  pout()<<"Warning: Setting (xllcorner,yllcorner) of DEM to (0.,0.)"<<endl;
  m_xllcorner=0.;
  m_yllcorner=0.;

  fileread = true;
  return fileread;
}

// change nodata value to land
// and convert DEM to meters.
bool DEMBL::fixDEM()
{
  bool DEMfixed = true;

  pout()<<"Changing datum and fixing land cells in DEM\n";

  //the following two should be added to the input file someday...FIXTHIS
  Real bottomBuff = 2.07;//this is the space added below the bathymetry, or the distance from the deepest spot to the box
  //Real topBuff    = 9.17;//this is the space added above the bathymetry, or the distance from the highest spot to the box

  //Real datumshift = 120.; //this is the distance from the DEM zero datum (mean sea level) to the domain origin (zero)
  // Scale is used for testing anisotropic vs isotropic geometry, if scale=1.0 then this is the true geometry;
  //   if scale is 100 then all elevations are multiplied by 100...
  Real scale = 1.0;
  //   ground is the elevation given for nodata points. 
  Real ground = (m_maxDEM+bottomBuff)*1.05*scale;
  //   highground is the elevation given for nodata points with all land neighbors (useful for higher order interpolation)
  m_highground = ground*1.05;



  int i,j,k=0;
  for (i=0;i<m_ncols;i++)
	{
	  for (j=0;j<m_nrows;j++)
		{
		  if(rint(m_DEM[i][j]) != m_NODATA)
			{
			  m_DEM[i][j] = (bottomBuff + m_maxDEM - m_DEM[i][j]*m_cellvalue)*scale;
			}
		  else
			{ //this could be done faster...
			  if((i!=0 & i!=m_ncols-1) & (j!=0 & j!=m_nrows-1)) {
				if( (rint(m_DEM[i+1][j]) == m_NODATA) || (rint(m_DEM[i+1][j]) == rint(m_highground))) {
				  if( (rint(m_DEM[i-1][j]) == m_NODATA) || (rint(m_DEM[i-1][j]) == rint(m_highground))) {
					if( (rint(m_DEM[i][j+1]) == m_NODATA) || (rint(m_DEM[i][j+1]) == rint(m_highground))) {
					  if( (rint(m_DEM[i][j-1]) == m_NODATA) || (rint(m_DEM[i][j-1]) == rint(m_highground))) {
						if( (rint(m_DEM[i+1][j+1]) == m_NODATA) || (rint(m_DEM[i+1][j+1]) == rint(m_highground))) {
						  if( (rint(m_DEM[i-1][j+1]) == m_NODATA) || (rint(m_DEM[i-1][j+1]) == rint(m_highground))) {
							if( (rint(m_DEM[i+1][j-1]) == m_NODATA) || (rint(m_DEM[i+1][j-1]) == rint(m_highground))) {
							  if( (rint(m_DEM[i-1][j-1]) == m_NODATA) || (rint(m_DEM[i-1][j-1]) == rint(m_highground))) {
								m_DEM[i][j]=m_highground;
								k++;
							  }
							}
						  }
						}
					  }
					}
				  }
				}
			  }
			  else
				{
				  m_DEM[i][j]=m_highground;
				  k++;
				}
			}
		}
	}

  for (i=0;i<m_ncols;i++)
	{
	  for (j=0;j<m_nrows;j++)
		{
		  if(rint(m_DEM[i][j]) == m_NODATA)
			{
			  m_DEM[i][j]=ground;
			  k++;
			}
		}
	}
  pout()<<"Filled: "<<k<<" cells."<<endl;
  return DEMfixed;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

bool DEMBL::isRegular(const Box&           a_region,
                      const Box&           a_domain,
                      const RealVect&      a_origin,
                      const Real&          a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, m_vectDx);
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, m_vectDx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs]; //corner is translated in localfuncvalue
                 
			  //if any corner is lower than or equal to the localfunval -> allbelow=false...
			  Real func;
			  int upDir = SpaceDim-1;
			  func = localFuncValue(corner, 
									upDir,
									iv, 
									a_domain, 
									a_origin, 
									a_dx);
			  if(corner[upDir]<=func) allAbove  = false;
			  //if any corner is higher than or equal to the water surface (air-water interface) -> allAbove=false
			  //Real surface;
			  //m_interface = 1;//0=soil/water, 1=air/water
			  //surface = localFuncValue(corner,
			  // 					     upDir,
			  // 					     iv, 
			  // 					     a_domain, 
			  // 						 a_origin, 
			  // 						 a_dx);
			  //  if(corner[upDir]>=surface) allAbove  = false;
			}
		}
	} 
  
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool DEMBL::isCovered(const Box&      a_region,
                      const Box&      a_domain,
                      const RealVect& a_origin,
                      const Real&     a_dx)const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, m_vectDx);
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, m_vectDx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs]; //corner is translated in localfuncvalue
          //if any corner is higher than the localfunval -> allbelow=false...
		  Real retval;
          int upDir=SpaceDim-1;
		  retval=localFuncValue(corner, 
                                upDir,
                                iv, 
                                a_domain, 
                                a_origin, 
                                a_dx);

          if(corner[upDir]>retval) 
            {
              allBelow  = false;
            }
        }
    }

  return allBelow;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

pair<int, Side::LoHiSide> DEMBL::upDirection(const RealVect& midPt,
                                             const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;
  retval.second=Side::Hi;
  retval.first=SpaceDim-1;
  return retval;
}

Real DEMBL::localFuncValue(const RealVect& a_indCoords, 
                           const int& a_upDir,
                           const IntVect& a_iv, 
                           const Box& a_domain, 
                           const RealVect& a_origin, 
                           const Real& a_dx) const
{

  Real retval =0.;
  //START DEBUG: Is the following correct???? ask Dan Graves...
  Real xin = a_indCoords[0]+a_origin[0];
  Real yin = a_indCoords[1]+a_origin[1];
  //END DEBUG

  if(m_interpType==1)
	{ //bilinear interpolation
	  //find nearest four cells in DEM (recall that the DEM values are cell centered)
	  int iw,ie,js,jn;

	  iw = (int)((xin-m_xllcorner)/m_cellsize-0.5);
	  js = (int)((yin-m_yllcorner)/m_cellsize-0.5); 
	  //where xllcorner is the x-lower-left corner of the cell-centered DEM
	  ie = iw + 1;
	  jn = js + 1;
	  
	  //the following checks to make sure we are inside of the DEM database (otherwise return a NODATA value)
	  if(iw<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(ie>m_ncols-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  if(js<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(jn>m_nrows-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  
	  //now let's do some bilinear interpolation...
	  Real a,b,xw,ys;
	  xw = m_cellsize*(float(iw)+0.5);
	  ys = m_cellsize*(float(js)+0.5);
	  a = (xin-m_xllcorner-xw)/m_cellsize; //this is a bilinear interp coeff
	  b = (yin-m_yllcorner-ys)/m_cellsize; //this is a bilinear interp coeff
  
	  retval = (1.-a)*( (1.-b)*m_DEM[iw][js] +
						(   b)*m_DEM[ie][js])+
        (   a)*( (1.-b)*m_DEM[iw][jn] +
                 (   b)*m_DEM[ie][jn]);
	}
  else if(m_interpType==2)
	{ //biquadratic interpolation

	  //find nearest cell in DEM (recall that the DEM values are cell centered)
	  int i,j;
	  i = (int)((xin-m_xllcorner)/m_cellsize);
	  j = (int)((yin-m_yllcorner)/m_cellsize);
	  //where xllcorner is the x-lower-left corner of the cell-centered DEM
	  	  
	  //the following checks to make sure we are inside of the DEM database (otherwise return a NODATA value)
	  if(i-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(i+1>m_ncols-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  if(j-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(j+1>m_nrows-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  
	  //now let's do some biquadratic interpolation...
	  Real dx     = m_cellsize;
	  Real dy     = m_cellsize;
	  Real dxdy   = dx*dy;
	  Real dx2    = dx*dx;
	  Real dy2    = dy*dy;
	  Real dx2dy  = dx2*dy;
	  Real dxdy2  = dx*dy2;
	  Real dx2dy2 = dx2*dy2;
	  Real f1 = m_DEM[i  ][j  ];
	  Real f2 = m_DEM[i+1][j  ];
	  Real f3 = m_DEM[i-1][j  ];
	  Real f4 = m_DEM[i  ][j+1];
	  Real f5 = m_DEM[i  ][j-1];
	  Real f6 = m_DEM[i+1][j+1];
	  Real f7 = m_DEM[i-1][j+1];
	  Real f8 = m_DEM[i+1][j-1];
	  Real f9 = m_DEM[i-1][j-1];

	  Real a1 = f1;
	  Real a2 = (f2-f3)/(2.*dx);
	  Real a3 = (f4-f5)/(2.*dy);
	  Real a4 = (f6+f9-f7-f8)/(4.*dxdy);
	  Real a5 = (f2+f3-f1*2.)/(2.*dx2);
	  Real a6 = (f4+f5-f1*2.)/(2.*dy2);
	  Real a7 = (2.*(-f4+f5)+f6+f7-f8-f9)/(4.*dx2dy);
	  Real a8 = (2.*(-f2+f3)+f6-f7+f8-f9)/(4.*dxdy2);
	  Real a9 = (f1-(f2+f3+f4+f5)/2.+(f6+f7+f8+f9)/4.)/dx2dy2;
	  
	  Real x=xin-m_xllcorner-dx*(float(i)+0.5); //this is x-distance from the cell-center of (i,j)
	  Real y=yin-m_yllcorner-dy*(float(j)+0.5); //this is y-distance from the cell-center of (i,j)
	  Real x2 = x*x;
	  Real y2 = y*y;

	  retval = a1 + a2*x + a3*y + a4*x*y + a5*x2 + a6*y2 + a7*x2*y + a8*x*y2 + a9*x2*y2;
	}
  else if(m_interpType==3)
	{ //bicubic interpolation type1 (this performs cubic interpolation in 1D, matching all values)

	  //find nearest cell to the south-west in DEM (recall that the DEM values are cell centered)
	  int i,j;
	  i = (int)((xin-m_xllcorner)/m_cellsize-0.5);
	  j = (int)((yin-m_yllcorner)/m_cellsize-0.5); 

	  //where xllcorner is the x-lower-left corner of the cell-centered DEM
	  	  
	  //the following checks to make sure we are inside of the DEM database (otherwise return a NODATA value)
	  if(i-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(i+2>m_ncols-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  if(j-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(j+2>m_nrows-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  
	  //now let's do some bicubic interpolation...
	  Real dx  = m_cellsize;
	  Real dx2 = dx*dx;
	  Real dx3 = dx2*dx;
	  Real dy  = m_cellsize;
	  Real dy2 = dy*dy;
	  Real dy3 = dy2*dy;

	  Real f1 = m_DEM[i-1][j-1];
	  Real f2 = m_DEM[i  ][j-1];
	  Real f3 = m_DEM[i+1][j-1];
	  Real f4 = m_DEM[i+2][j-1];
	  Real f5 = m_DEM[i-1][j  ];
	  Real f6 = m_DEM[i  ][j  ];
	  Real f7 = m_DEM[i+1][j  ];
	  Real f8 = m_DEM[i+2][j  ];
	  Real f9 = m_DEM[i-1][j+1];
	  Real f10= m_DEM[i  ][j+1];
	  Real f11= m_DEM[i+1][j+1];
	  Real f12= m_DEM[i+2][j+1];
	  Real f13= m_DEM[i-1][j+2];
	  Real f14= m_DEM[i  ][j+2];
	  Real f15= m_DEM[i+1][j+2];
	  Real f16= m_DEM[i+2][j+2];

	  Real x=xin-m_xllcorner-dx*(float(i+1)); //this is the x-distance from (i+1/2,j+1/2)
	  Real y=yin-m_yllcorner-dy*(float(j+1)); //this is the y-distance from (i+1/2,j+1/2)
	  Real x2 = pow(x,2.);
	  Real x3 = pow(x,3.);
	  Real y2 = pow(y,2.);
	  Real y3 = pow(y,3.);

	  //cubic type1 in x direction: for row j-1
	  Real r1 = (-3.*dx3*( f1 - 9.*(f2 + f3) + f4)    + 
                 2.*dx2*( f1 -27.*(f2 - f3) - f4)*x  + 
                 12.*dx*( f1 -    (f2 + f3) + f4)*x2 + 
                 8.*(-f1 + 3.*(f2 - f3) + f4)*x3)/(48.*dx3);
	  
	  //cubic type1 in x direction: for row j
	  Real r2 = (-3.*dx3*( f5 - 9.*(f6 + f7) + f8)    + 
                 2.*dx2*( f5 -27.*(f6 - f7) - f8)*x  + 
                 12.*dx*( f5 -    (f6 + f7) + f8)*x2 + 
                 8.*(-f5 + 3.*(f6 - f7) + f8)*x3)/(48.*dx3);
	  
	  //cubic type1 in x direction: for row j+1
	  Real r3 = (-3.*dx3*( f9 - 9.*(f10 + f11) + f12)    + 
                 2.*dx2*( f9 -27.*(f10 - f11) - f12)*x  + 
                 12.*dx*( f9 -    (f10 + f11) + f12)*x2 + 
                 8.*(-f9 + 3.*(f10 - f11) + f12)*x3)/(48.*dx3);
	  
	  //cubic type1 in x direction: for row j+2
	  Real r4 = (-3.*dx3*( f13 - 9.*(f14 + f15) + f16)    + 
                 2.*dx2*( f13 -27.*(f14 - f15) - f16)*x  + 
                 12.*dx*( f13 -    (f14 + f15) + f16)*x2 + 
                 8.*(-f13 + 3.*(f14 - f15) + f16)*x3)/(48.*dx3);
	  
	  //cubic type1 in y direction!!!
	  retval  = (-3.*dy3*( r1 - 9.*(r2 + r3) + r4)    + 
                 2.*dy2*( r1 -27.*(r2 - r3) - r4)*y  + 
                 12.*dy*( r1 -    (r2 + r3) + r4)*y2 + 
                 8.*(-r1 + 3.*(r2 - r3) + r4)*y3)/(48.*dy3);
	}

  else if(m_interpType==4)
	{ //bicubic interpolation type2 (this performs cubic interpolation in 1D, matching values and derivatives)

	  //find nearest cell to the south-west in DEM (recall that the DEM values are cell centered)
	  int i,j;
	  i = (int)((xin-m_xllcorner)/m_cellsize-0.5);
	  j = (int)((yin-m_yllcorner)/m_cellsize-0.5); 

	  //where xllcorner is the x-lower-left corner of the cell-centered DEM
	  	  
	  //the following checks to make sure we are inside of the DEM database (otherwise return a NODATA value)
	  if(i-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(i+2>m_ncols-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  if(j-1<0)
		{
		  retval=m_highground;
		  return retval;
		}
	  else if(j+2>m_nrows-1)
		{
		  retval=m_highground;
		  return retval;
		}
	  
	  //now let's do some bicubic interpolation...
	  Real dx  = m_cellsize;
	  Real dx2 = dx*dx;
	  Real dx3 = dx2*dx;
	  Real dy  = m_cellsize;
	  Real dy2 = dy*dy;
	  Real dy3 = dy2*dy;

	  Real f1 = m_DEM[i-1][j-1];
	  Real f2 = m_DEM[i  ][j-1];
	  Real f3 = m_DEM[i+1][j-1];
	  Real f4 = m_DEM[i+2][j-1];
	  Real f5 = m_DEM[i-1][j  ];
	  Real f6 = m_DEM[i  ][j  ];
	  Real f7 = m_DEM[i+1][j  ];
	  Real f8 = m_DEM[i+2][j  ];
	  Real f9 = m_DEM[i-1][j+1];
	  Real f10= m_DEM[i  ][j+1];
	  Real f11= m_DEM[i+1][j+1];
	  Real f12= m_DEM[i+2][j+1];
	  Real f13= m_DEM[i-1][j+2];
	  Real f14= m_DEM[i  ][j+2];
	  Real f15= m_DEM[i+1][j+2];
	  Real f16= m_DEM[i+2][j+2];

	  Real x=xin-m_xllcorner-dx*(float(i+1)); //this is the x-distance from (i+1/2,j+1/2)
	  Real y=yin-m_yllcorner-dy*(float(j+1)); //this is the y-distance from (i+1/2,j+1/2)
	  Real x2 = pow(x,2.);
	  Real x3 = pow(x,3.);
	  Real y2 = pow(y,2.);
	  Real y3 = pow(y,3.);
	  
	  //cubic type2 in x direction: for row j-1
	  Real r1 = (-f1 + 9.*(f2 + f3) - f4)   /    16.  + 
        ( f1 -11.*(f2 - f3) - f4)*x /(dx *8.) + 
        ( f1 -    (f2 + f3) + f4)*x2/(dx2*4.) + 
        (-f1 + 3.*(f2 - f3) + f4)*x3/(dx3*2.);
	  
	  //cubic type2 in x direction: for row j
	  Real r2 = (-f5 + 9.*(f6 + f7) - f8)   /    16.  + 
        ( f5 -11.*(f6 - f7) - f8)*x /(dx *8.) + 
        ( f5 -    (f6 + f7) + f8)*x2/(dx2*4.) + 
        (-f5 + 3.*(f6 - f7) + f8)*x3/(dx3*2.);
	  
	  //cubic type2 in x direction: for row j+1	  
	  Real r3 = (-f9 + 9.*(f10 + f11) - f12)   /    16.  + 
        ( f9 -11.*(f10 - f11) - f12)*x /(dx *8.) + 
        ( f9 -    (f10 + f11) + f12)*x2/(dx2*4.) + 
        (-f9 + 3.*(f10 - f11) + f12)*x3/(dx3*2.);
	  
	  //cubic type2 in x direction: for row j+2
	  Real r4 = (-f13 + 9.*(f14 + f15) - f16)   /    16.  + 
        ( f13 -11.*(f14 - f15) - f16)*x /(dx *8.) + 
        ( f13 -    (f14 + f15) + f16)*x2/(dx2*4.) + 
        (-f13 + 3.*(f14 - f15) + f16)*x3/(dx3*2.);
	  
	  //cubic type2 in y direction!!!
	  retval  = (-r1 + 9.*(r2 + r3) - r4)   /    16.  + 
        ( r1 -11.*(r2 - r3) - r4)*y /(dy *8.) + 
        ( r1 -    (r2 + r3) + r4)*y2/(dy2*4.) + 
        (-r1 + 3.*(r2 - r3) + r4)*y3/(dy3*2.);
	}
  else
	{
	  MayDay::Error("Bad InterpType in DEMBL");
	}

  
  return retval;
}
/*****/
/*****/
BaseLevelSet* DEMBL::new_baseLevelSet() const       
{
  DEMBL* DEMPtr = new DEMBL(m_ncell,
							m_interpType,
							m_dx,
							m_vectDx);
    
  return static_cast<BaseLevelSet*>(DEMPtr); 
}
