

/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "Box.H"
#include "DoubleCylBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"

DoubleCylBL::DoubleCylBL(const Real      a_radius[2],
                         const Real      a_height[2],
                         const RealVect  a_center[2],
                         const RealVect& a_origin,
                         const int&      a_axis,
                         const bool&     a_interior,
                         const Real&     a_dx)
{
  assert(a_dx > PolyGeom::getTolerance());
  
  assert(a_radius[0] >= 2*a_dx);
  assert(a_radius[1] >= 2*a_dx);
  
  assert(a_height[0] >= 0.);
  assert(a_height[1] >= 0.);
  
  assert(a_axis >= 0 && a_axis < SpaceDim);
 
  m_radius[0] = a_radius[0];
  m_height[0] = a_height[0];
  m_center[0] = a_center[0];
  
  m_radius[1] = a_radius[1];
  m_height[1] = a_height[1];
  m_center[1] = a_center[1];


  m_origin = a_origin;
  m_axis     = a_axis;
  m_interior = a_interior;
  m_dx       = a_dx;
  m_dividingLine = .5;
}

DoubleCylBL::~DoubleCylBL()
{
}

Real DoubleCylBL::LevelSurface(const RealVect& a_vec) const
{
  // if the return value is positive then a_vec is in the fluid (therefore
  // the answer is adjusted depending on whether the problem is interior
  // or exterior)

  //a_vec has already been translated w.r.t. m_origin

  Real retval = 0;

  RealVect vec = a_vec;
  int hilo;
  if (vec[m_axis] < m_dividingLine)
    {
      hilo = 0;
    }
  else
    {
      hilo = 1;
    }
  
  vec -= m_center[hilo];

  if (Abs(vec[m_axis]) <= m_height[hilo]/2.)
    {
      retval = m_radius[hilo]*m_radius[hilo];
      for (int i = 0; i < SpaceDim; i++)
      {
        if (i != m_axis)
        {
          retval -= vec[i]*vec[i];
        }
      }
    }
  else if (vec[m_axis] > 0)
    {
      vec[m_axis] -= m_height[hilo]/2.;

      retval = m_radius[hilo]*m_radius[hilo];
      for (int i = 0; i < SpaceDim; i++)
      {
        retval -= vec[i]*vec[i];
      }
    }  
  else if (vec[m_axis] < 0)
    {
      vec[m_axis] += m_height[hilo]/2.;
      
      retval = m_radius[hilo]*m_radius[hilo];
      for (int i = 0; i < SpaceDim; i++)
      {
        retval -= vec[i]*vec[i];
      }
    }
  else 
    {
      MayDay::Error("missed a case");
    }

  if (!m_interior)
    {
      retval = -retval;
    }

  return retval;
}

bool DoubleCylBL::isRegular(const Box&      a_region,
                            const Box&      a_domain,
                            const RealVect& a_origin,
                            const Real&     a_dx) const
{
  // iterate over each cell in the box to check if regular
  bool allRegular = true;
 

  BoxIterator bit(a_region);
  for (bit.reset(); bit.ok() && allRegular; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

      // if every corner is above the face the box is covered
      for (int ilocs = 0; (ilocs < cellboxLocs.size()) && allRegular; ilocs++)
        {
          RealVect& corner = cellboxLocs[ilocs];
          corner += m_origin;
          // if every corner is interior, the box is regular
          if (LevelSurface(corner) <= 0.) // if levelSurface > 0, then a_vec is in the fluid
            {
              allRegular = false;
            }
        }
    }
  return allRegular;
}

bool DoubleCylBL::isCovered(const Box&      a_region,
                           const Box&      a_domain,
                           const RealVect& a_origin,
                           const Real&     a_dx) const
{
  // iterate over each cell in the box to check if covered
  bool allCovered = true;
  BoxIterator bit(a_region);
  for (bit.reset(); bit.ok() && allCovered; ++bit)
    {
      IntVect iv = bit();
      Box cellbox(iv,iv);
     
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox,a_origin,a_dx);

      // if every corner is above the face the box is covered
      for (int ilocs = 0; (ilocs < cellboxLocs.size()) && allCovered; ilocs++)
        {
          RealVect& corner = cellboxLocs[ilocs];
          corner += m_origin;
    
          if (LevelSurface(corner) > 0.) // if levelSurface > 0, then a_vec is in the fluid
            {
              allCovered = false;
            }
        }
    }
  return allCovered;
}

pair<int, Side::LoHiSide> DoubleCylBL::upDirection(const RealVect& a_midPt,const IntVect& a_iv) const
{
  RealVect vec = a_midPt;
  vec += m_origin;

 
  int hilo;
  if (vec[m_axis] < m_dividingLine)
    {
      hilo = 0;
    }
  else
    {
      hilo = 1;
    }
  pair<int, Side::LoHiSide> retval;

  // now we calculate a vector porportional to the gradient and equivalent 
  // to the gradient for our purposes
  RealVect gradient;

  //if we're close to the intersection plane(3D)/line(2D) then updirection is special
  if(Abs(vec[m_axis] - m_dividingLine) <= m_dx)
    {
      RealVect divisionCenter = m_center[hilo];
      divisionCenter[m_axis] = m_dividingLine;
      gradient = vec;
      gradient -= divisionCenter; 
      gradient[m_axis] = 0.;
    }
   else if (Abs(vec[m_axis] - m_center[hilo][m_axis]) <= m_height[hilo]/2.)
     { 
       gradient = vec;
       gradient -= m_center[hilo];
       
       gradient[m_axis] = 0.;
     }
  else if (vec[m_axis] > m_center[hilo][m_axis])
    {
      RealVect capCenter = m_center[hilo];
      capCenter[m_axis] += m_height[hilo]/2.;
      
      gradient = vec;
      gradient -= capCenter;
    }
  else if (vec[m_axis] < m_center[hilo][m_axis])
    {
      RealVect capCenter = m_center[hilo];
      capCenter[m_axis] -= m_height[hilo]/2.;
      
      gradient = vec;
      gradient -= capCenter;
    }
  else
    {
      MayDay::Error("Missed a case:DoubleCylBL::upDirection");
    }
  
  // We have the gradient. Now find the largest component.
  Real max = 0.;
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      if (Abs(gradient[idir]) >= max) 
        {
          retval.first = idir;
          max = Abs(gradient[idir]);
        }
    }
  
  // Which way is up?
  if (gradient[retval.first] >= 0. && m_interior)
    {
      retval.second = Side::Lo;
    }
  else if (gradient[retval.first] < 0. && m_interior)
    {
      retval.second = Side::Hi;
    }
  else if (gradient[retval.first] >= 0. && !m_interior)
    {
      retval.second = Side::Hi;
    }
  else if (gradient[retval.first] < 0. && !m_interior)
    {
      retval.second = Side::Lo;
    }
  else
    {
      MayDay::Error("Missed a case");
    }

  return retval;
}

Real DoubleCylBL::localFuncValue(const RealVect& a_indCoords, 
                                const int&      a_upDir,
                                const IntVect&  a_iv, 
                                const Box&      a_domain, 
                                const RealVect& a_origin, 
                                const Real&     a_dx) const
{
 
  Real retval = 0.;
  
  RealVect vec;
  vec = a_indCoords;
  vec += m_origin;
  
  int hilo;

  if (vec[m_axis] < m_dividingLine)
    {
      hilo = 0;
    }
  else
    {
      hilo = 1;
    }
  
  vec -= m_center[hilo];

  if (Abs(vec[m_axis]) <= m_height[hilo]/2.)
    {
      Real arg;

#if (CH_SPACEDIM == 2)
      arg = 0.;
#elif (CH_SPACEDIM == 3)
      int indDir = 3 - m_axis - a_upDir;
      arg = a_indCoords[indDir] - m_center[hilo][indDir];
#else
      // Only 2D and 3D should be possible
      MayDay::Error("CH_SPACEDIM not 2 or 3!");
#endif

      if (a_indCoords[a_upDir] > m_center[hilo][a_upDir])
        {
          assert(m_radius[hilo]*m_radius[hilo]-arg*arg > 0.);
          retval = m_center[hilo][a_upDir] + sqrt(Abs(m_radius[hilo]*m_radius[hilo]-arg*arg));
        }
      else 
        {
          assert(m_radius[hilo]*m_radius[hilo]-arg*arg > 0.);
          retval = m_center[hilo][a_upDir] - sqrt(m_radius[hilo]*m_radius[hilo]-arg*arg);
        }
    }
  else if (vec[m_axis] > m_height[hilo]/2.)
    {
      RealVect capCenter;

      capCenter = m_center[hilo];
      capCenter[m_axis] += m_height[hilo]/2.;

      RealVect pt = a_indCoords;
      pt -= capCenter;
      retval = m_radius[hilo]*m_radius[hilo];

      for (int idir = 0; idir < SpaceDim; ++idir)
        {
          if (idir != a_upDir)
            {            
              retval -= pt[idir]*pt[idir];
            }
        }

      if (a_indCoords[a_upDir] > capCenter[a_upDir])
        {
          assert(retval > 0.);
          retval = sqrt(retval);
        }
      else
        {
          assert(retval > 0.);
          retval = -sqrt(retval);
        }
      retval += capCenter[a_upDir];
    }

  else if (vec[m_axis] < -m_height[hilo]/2.)
    {
      RealVect capCenter;
     
      capCenter = m_center[hilo];
      capCenter[m_axis] -= m_height[hilo]/2.;

      RealVect pt = a_indCoords;
      pt -= capCenter;
      retval = m_radius[hilo]*m_radius[hilo];

      for (int idir = 0; idir < SpaceDim; ++idir)
        {
          if (idir != a_upDir)
            {            
              retval -= pt[idir]*pt[idir];
            }
        }
      if (a_indCoords[a_upDir] > capCenter[a_upDir])
        {
          assert(retval > 0.);
          retval = sqrt(retval);
        }
      else
        {
          assert(retval > 0.);
          retval = -sqrt(retval);
        }
      retval += capCenter[a_upDir];
    }     
  return retval;
}

BaseLevelSet* DoubleCylBL::new_baseLevelSet() const       
{
  DoubleCylBL* doubleCylPtr = new DoubleCylBL(m_radius,
                                              m_height,
                                              m_center,
                                              m_origin,
                                              m_axis,
                                              m_interior,
                                              m_dx);
  
  return static_cast<BaseLevelSet*>(doubleCylPtr); 
}
