/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "DuellingParabolas.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
DuellingParabolas::DuellingParabolas(const Vector<Real>&      a_amplitude,
                                     const Vector<RealVect>&  a_center,
                                     const Real&              a_dx,
                                     const int&               a_updir)
{
  assert(a_amplitude.size() > 0);
  assert(a_amplitude.size() == a_center.size());
  m_numParabolas   = a_amplitude.size();
  m_amplitude      = a_amplitude;
  m_center         = a_center;
  m_dx             = a_dx;
  m_updir          = a_updir;
  if(m_updir == 0)
    {
      m_tandir = 1;
    }
  else
    {
      m_tandir = 0;
    }
}
/*****/
Real 
DuellingParabolas::parabFunction(const Real&     a_x,
                                 const Real&     a_amplitude,
                                 const RealVect& a_center) const
{
  Real xdiff = (a_x - a_center[m_tandir]);
  Real y = a_amplitude*xdiff*xdiff + a_center[m_updir];
  return y;
}
/*****/
Real 
DuellingParabolas::distFunction(const Real&         a_x,
                                const RealVect&     a_ab,
                                const Real&         a_amplitude,
                                const RealVect&     a_center) const
{
  Real x  = a_x;
  Real a  = a_ab[m_tandir];
  Real b  = a_ab[m_updir];
  Real c  = a_amplitude;
  Real x0 = a_center[m_tandir];
  Real y0 = a_center[m_updir];
  
  Real dfdx = 2.0*(x - a) + 2.0*(c*(x-x0)*(x-x0)+y0-b)*(2.0*c*(x-x0));
  return dfdx;
}
/*****/
Real 
DuellingParabolas::minimumDistToParabola(const RealVect& a_vec,
                                         const Real&     a_amplitude,
                                         const RealVect  a_center) const
{
  //solve for which x is closest
  Real xMin = -2.0;
  Real xMax = 2.0;
  Real xMid = 0.5;
  Real funcLo = distFunction(xMin, a_vec, a_amplitude, a_center);
  Real funcHi = distFunction(xMax, a_vec, a_amplitude, a_center);
  if(funcLo*funcHi > 0.0)
    {
      MayDay::Error("bisectiter: root not bracketed");
    }
  Real err = 1.0; //= xmax - xmin
  Real funcMid= distFunction(xMid, a_vec, a_amplitude, a_center);
    
  int itermax = 1000;
  int iteration = 0;
  Real xLo = xMin;
  Real xHi = xMax;
  Real tolerance = 1.0e-8;
  while((iteration < itermax) && (err > tolerance))
    {
      Real xOld = xMid;
      funcMid  = distFunction(xMid, a_vec, a_amplitude, a_center);
      if(funcLo*funcMid > 0.0)
        {
          xLo = xMid;
          funcLo = funcMid;
        }
      else
        {
          assert(funcHi*funcMid >= 0.0);
          xHi = xMid;
          funcHi = funcMid;
        }
      iteration++;
      xMid = 0.5*(xLo+xHi);
      err = Abs(xMid - xOld);
    }
  if((xMid < xMin) || (xMid > xMax))
    MayDay::Error("bisectiter: solution out of bounds");

  Real x = xMid;
  Real y = parabFunction(x, a_amplitude, a_center);

  Real dist = 
    (x - a_vec[m_tandir])*(x - a_vec[m_tandir]) + 
    (y - a_vec[m_updir])*(y - a_vec[m_updir]);

  return dist;
}
/*****/

int 
DuellingParabolas::closestParabola(const RealVect& a_vec) const
{
  assert(m_numParabolas > 0);
  //first check to see if you are inside either parabola
  for(int iparabola = 0; iparabola < m_numParabolas; iparabola++)
    {
      Real x = a_vec[m_tandir];
      Real y = a_vec[m_updir];
      Real yparab = parabFunction(x, m_amplitude[iparabola], m_center[iparabola]);
      Real monot = m_amplitude[iparabola]*(y-yparab);
      if(monot > 0.0)
        {
          return iparabola;
        }
    }
  //find the distance to each parabola
  Vector<Real> sqDistToParabola(m_numParabolas, 0.0);
  
  for(int iparabola = 0; iparabola < m_numParabolas; iparabola++)
    {
      Real thisDist = minimumDistToParabola(a_vec, 
                                            m_amplitude[iparabola], 
                                            m_center[iparabola]);
      sqDistToParabola[iparabola] = thisDist;
    }
  //return the index of the closest parabola 
  int iclose = 0;
  Real minDist = sqDistToParabola[0];
  for(int iparabola = 1; iparabola < m_numParabolas; iparabola++)
    {
      if(sqDistToParabola[iparabola] < minDist)
        {
          iclose = iparabola;
          minDist = sqDistToParabola[iparabola];
        }
    }
  return iclose;
}
/*****/
DuellingParabolas::~DuellingParabolas()
{
}
/*****/
bool 
DuellingParabolas::isPointCovered(const RealVect& a_vec,
                                  const RealVect& a_center,
                                  const Real&     a_amplitude) const
{
  Real x = a_vec[m_tandir];
  Real y = a_vec[m_updir];
  Real yparab = parabFunction(x, a_amplitude, a_center);
  bool retval;
  if(a_amplitude > 0) 
    {
      retval = (y > yparab);
    }
  else
    {
      retval = (y < yparab);
    }
  return retval;
}
/*****/
/*****/
bool 
DuellingParabolas::isRegular(const Box&           a_region,
                             const Box&           a_domain,
                             const RealVect&      a_origin,
                             const Real&          a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              int iparabola = closestParabola(corner);
              RealVect center = m_center[iparabola];
              Real     amplitude = m_amplitude[iparabola];
              
              bool thisPtCovered = isPointCovered(corner, center, amplitude);
              if(thisPtCovered)
                {
                  allAbove = false;
                }
            }
        } 
    }
  
  // if ( allAbove)   pout()<< "region "<<a_region<<" is regular"<<endl;
  // if (!allAbove)   pout()<< "region "<<a_region<<" is not regular"<<endl;  
 
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
DuellingParabolas::isCovered(const Box&      a_region,
                             const Box&      a_domain,
                             const RealVect& a_origin,
                             const Real&     a_dx)const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          int iparabola = closestParabola(corner);
          RealVect center = m_center[iparabola];
          Real     amplitude = m_amplitude[iparabola];

          bool thisPtCovered = isPointCovered(corner, center, amplitude);
          if(!thisPtCovered)
            {
              allBelow = false;
            }
        }
    }
  // if ( allBelow)   pout()<< "region "<<a_region<<" is covered"<<endl;
  // if (!allBelow)   pout()<< "region "<<a_region<<" is not covered"<<endl;

  return allBelow;
}
/*****/
pair<int, Side::LoHiSide> 
DuellingParabolas::upDirection(const RealVect& midPt,const IntVect& a_iv) const
{
  int iparabola = closestParabola(midPt);
  Real     amplitude = m_amplitude[iparabola];
  pair<int, Side::LoHiSide> retval;
  retval.first = m_updir;
  if(amplitude > 0.0)
    retval.second = Side::Lo;
  else
    retval.second = Side::Hi;
  return retval;
}
/*****/
Real 
DuellingParabolas::localFuncValue(const RealVect& a_indCoords, 
                                  const int& a_upDir,
                                  const IntVect& a_iv, 
                                  const Box& a_domain, 
                                  const RealVect& a_origin, 
                                  const Real& a_dx) const

{
  int iparabola = closestParabola(a_indCoords);
  RealVect center = m_center[iparabola];
  Real     amplitude = m_amplitude[iparabola];
 
  Real x = a_indCoords[m_tandir];
  Real yparab = parabFunction(x, amplitude, center);
  return yparab;
}
/*****/
/*****/
BaseLevelSet* 
DuellingParabolas::new_baseLevelSet() const       
{
  DuellingParabolas* parabolaPtr = new DuellingParabolas(m_amplitude,
                                                         m_center,
                                                         m_dx,
                                                         m_updir);
    
  return static_cast<BaseLevelSet*>(parabolaPtr); 
}
