/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "RealVect.H"

#include "EllipsoidIF.H"

EllipsoidIF::EllipsoidIF(const RealVect& a_radii,
                   const RealVect& a_center,
                   const bool&     a_inside)
{
  // Remember the parameters
  m_radii = a_radii;
  m_center = a_center;
  m_inside = a_inside;

  // Precompute the radii squared
  m_radii2 = m_radii;
  m_radii2 *= m_radii;
}

EllipsoidIF::EllipsoidIF(const EllipsoidIF& a_inputIF)
{
  // Remember the parameters
  m_radii = a_inputIF.m_radii;
  m_center = a_inputIF.m_center;
  m_inside = a_inputIF.m_inside;

  // Precompute the radii squared
  m_radii2 = m_radii;
  m_radii2 *= m_radii;
}

EllipsoidIF::~EllipsoidIF()
{
}

void EllipsoidIF::GetParams(RealVect& a_radii,
                         RealVect& a_center,
                         bool&     a_inside) const
{
  // Copy parameter information over
  a_radii = m_radii;
  a_center = m_center;
  a_inside = m_inside;
}

void EllipsoidIF::SetParams(const RealVect& a_radii,
                         const RealVect& a_center,
                         const bool&     a_inside)
{
  // Set parameter information
  m_radii = a_radii;
  m_center = a_center;
  m_inside = a_inside;

  // Precompute the radii squared
  m_radii2 = m_radii;
  m_radii2 *= m_radii;
}

Real EllipsoidIF::value(const RealVect& a_point) const
{
  Real retval;

  // Compute the equation of the ellipsoid
  Real sum;

  sum = 0.0;
  for (int idir = 0; idir < SpaceDim; idir++)
  {
    Real cur;
    cur = a_point[idir] - m_center[idir];

    sum += cur*cur / m_radii2[idir];
  }

  // The sum should be 1.0 on the surface of the ellipsoid
  retval = sum - 1.0;

  // Change the sign to change inside to outside
  if (!m_inside)
  {
    retval = -retval;
  }

  return retval;
}

BaseImplicitFunction* EllipsoidIF::newImplicitFunction() const
{
  EllipsoidIF* ellipsoidPtr = new EllipsoidIF(m_radii,
                                     m_center,
                                     m_inside);

  return static_cast<BaseImplicitFunction*>(ellipsoidPtr);
}
