/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG


#include "ExpansionChannelBL.H"
#include "Box.H"
#include "PolyGeom.H"
/*****/
/*****/
ExpansionChannelBL::ExpansionChannelBL(const RealVect& a_rampNormal,
			     const int&  a_indepVar, 
			     const Real&     a_alpha,
			     const Real&     a_rampEnd)
{
  m_rampNormal = a_rampNormal;
  m_indepVar = a_indepVar;
  m_rampEnd = a_rampEnd;
  Real sumSquare;
  //make the normal a unit vector
  PolyGeom::unifyVector(m_rampNormal, sumSquare);
  m_alpha = a_alpha/sumSquare;

  //pick updir to be direction with biggest normal
  int ibigNorm = 0;
  for(int idir = 1; idir < SpaceDim; idir++)
    {
      if(Abs(m_rampNormal[idir]) > Abs(m_rampNormal[idir-1]))
        {
          ibigNorm = idir;
        }
    }

  m_upDir.first = ibigNorm;
  if(m_rampNormal[ibigNorm] > 0.0)
    {
      m_upDir.second = Side::Hi;
    }
  else
    {
      m_upDir.second = Side::Lo;
    }
}
/*****/
/*****/
ExpansionChannelBL::~ExpansionChannelBL()
{
}
/*****/
/*****/
bool 
ExpansionChannelBL::isRegular(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is above the face, the box is regular
  bool allAbove = true;
  //  int count=0;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allAbove; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
      if(m_upDir.second == Side::Hi)
        {
	  if(corner[m_upDir.first] < funcVal)
	    {
//  	    count+=1;
//  	    if(count>1)allAbove = false;
	    allAbove = false;
	    }
        }
      else
        {
          if(corner[m_upDir.first] > funcVal)
	    {
//  	    count+=1;
//              if(count>1)allAbove = false;
	    allAbove = false;
	    }
        }
    }
  return allAbove;
}
/*****/
/*****/
bool 
ExpansionChannelBL::isCovered(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is below the face, the box is covered
  bool allBelow = true;
  //  int count=0;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allBelow; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
      if(m_upDir.second == Side::Hi)
        {
          if(corner[m_upDir.first] > funcVal)
	    {
//  	    count+=1;
//              if(count>1)allBelow = false;
	    allBelow = false;
	    }
        }
      else
        {
          if(corner[m_upDir.first] < funcVal)
	    {
//  	    count+=1;
//              if(count>1)allBelow = false;
	    allBelow = false;
	    }
        }
    }
  return allBelow;
}
/*****/
/*****/
pair<int, Side::LoHiSide> ExpansionChannelBL::upDirection(const RealVect& a_midpt,const IntVect& a_iv) const 
{
  return m_upDir;
}
/*****/
/*****/
Real 
ExpansionChannelBL::localFuncValue(const RealVect& a_indCoords, 
                       const int& a_upDir,
                       const IntVect& a_iv, 
                       const Box& a_domain, 
                       const RealVect& a_origin, 
                       const Real& a_dx) const 
{
  assert(a_upDir == m_upDir.first);
  Real retval = m_alpha;
  if(a_indCoords[m_indepVar] > m_rampEnd)
    {
      for(int idir = 0; idir < SpaceDim; idir++)
	{
	  if(idir != m_upDir.first)
	    {
	      retval -= m_rampNormal[idir]*a_indCoords[idir];
	    }
	}
      retval /= m_rampNormal[m_upDir.first];
    }
  else
    {
      for(int idir = 0; idir < SpaceDim; idir++)
	{
	  if(idir != m_upDir.first)
	    {
	      retval -= m_rampNormal[idir]*m_rampEnd;
	    }
	}
      retval /= m_rampNormal[m_upDir.first];
    }
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
ExpansionChannelBL::new_baseLevelSet() const       
{
  ExpansionChannelBL* rampPtr = new ExpansionChannelBL(m_rampNormal,m_indepVar,
                               m_alpha,m_rampEnd);
  return static_cast<BaseLevelSet*>(rampPtr); 
}
