/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG

#ifndef _GeometryShopBase_H_
#define _GeometryShopBase_H_

#include "REAL.H"
#include "RealVect.H"
#include "Box.H"
#include "IntVect.H"
#include "EBISBox.H"
#include "BaseLevelSet.H"
#include "GeometryService.H"
#include "Moments.H"
#include "CylinderBL.H"
/// 
/**
   This is the base class for the workshop algorithm.
   It forms the interface between the workshop classes
   and the geometryservice class.
 */
class GeometryShop: public GeometryService
{
public:
  ///
  /**
     Define the workshop using the local geometry description
  */
  //  GeometryShop(const BaseLevelSet& a_localGeom,int verbosity = 0);
  GeometryShop(const BaseLevelSet&  a_localGeom,
                           int      a_verbosity = 0,
                           RealVect a_vectDx = RealVect::Zero);

  ///
  ~GeometryShop();

  /// 
  bool twoEdgeIntersections(edgeMo a_edges[4]) const;

  ///
  bool tooManyUpDir(const IntVect& a_iv, 
                    const Real&    a_dx,
                    const int&     a_faceNormal) const;
  

  /** 
      Return true if every cell in region is regular at the
      refinement described by dx. 
  */
  bool isRegular(const Box&      a_region,
                 const Box&      a_domain,
                 const RealVect& a_origin,
                 const Real&     a_dx) const;

  /// 
  /** 
      Return true if every cell in region is covered at the
      refinement described by dx.
  */
  bool isCovered(const Box&      a_region,
                 const Box&      a_domain,
                 const RealVect& a_origin,
                 const Real&     a_dx) const;

  ///
  /**
     Define the internals of the input ebisRegion.
  */
  virtual void fillGraph(BaseFab<int>&      a_regIrregCovered,
                         Vector<IrregNode>& a_nodes,
                         const Box&         a_validRegion,
                         const Box&         a_ghostRegion,
                         const Box&         a_domain,
                         const RealVect&    a_origin,
                         const Real&        a_dx) const;

  /**
  */
  void computeVoFInternals(Real&             a_volFrac,
                           Vector<int>       a_loArc[SpaceDim],
                           Vector<int>       a_hiArc[SpaceDim],
                           Vector<Real>      a_loAreaFrac[SpaceDim],
                           Vector<Real>      a_hiAreaFrac[SpaceDim],
                           Real&             a_bndryArea,
                           RealVect&         a_normal,
                           RealVect&         a_volCentroid,
                           RealVect&         a_bndryCentroid,
                           Vector<RealVect>  a_loFaceCentroid[SpaceDim],
                           Vector<RealVect>  a_hiFaceCentroid[SpaceDim],
                           const IntVectSet& a_ivsIrreg,
                           const VolIndex&   a_vof,
                           const Box&        a_domain, 
                           const RealVect&   a_origin, 
                           const Real&       a_dx,
                           const RealVect&   a_vectDx,
                           const IntVect&    a_iv) const;                

  int m_phase;
  
private:
  int                 m_numCellsClipped;
  int                 m_verbosity;
  Real                m_threshold; 
  RealVect            m_vectDx;  
  const BaseLevelSet* m_localGeomPtr;


  static bool s_verbose;

  // local geometry description
  // const BaseLocalGeometry* m_localGeomPtr;
 
  void edgeData3D(
                  edgeMo          a_edges[4],
                  bool&           a_faceCovered,
                  bool&           a_faceRegular,
                  bool&           a_faceDontKnow,
                  const int       a_hiLoFace,
                  const int       a_faceNormal,
                  const Real&     a_dx,
                  const RealVect& a_vectDx,           
                  const IntVect&  a_coord, 
                  const Box&      a_domain,
                  const RealVect& a_origin) const;

  void edgeData2D(edgeMo          a_edges[4],
                  bool&           a_faceCovered,
                  bool&           a_faceRegular,
                  bool&           a_faceDontKnow,
                  const Real&     a_dx,
                  const RealVect& a_vectDx,
                  const IntVect&  a_coord, 
                  const Box&      a_domain,
                  const RealVect& a_origin) const;

  void edgeType(bool&                           a_regular,
                bool&                           a_covered, 
                bool&                           a_dontKnow,
                Real&                           a_signHi,
                Real&                           a_signLo,
                const pair<int,Side::LoHiSide>& a_upDir) const;

  Real BrentRootFinder(const RealVect& a_x1,
                       const RealVect& a_x2,
                       const int&      a_upDir,
                       const IntVect&  a_iv,
                       const Box&      a_domain,
                       const RealVect& a_origin,
                       const Real&     a_dx,
                       const int&      a_range) const;

#if 1
  Real PrismoidalAreaCalc(RealVect& a_xVec,
                          RealVect& a_yVec) const;
#endif

  int getNumCellsClipped();

  Real Min(const Real x, const Real y) const;

  //stuff disallowed for all the usual reasons.
  GeometryShop()
  {
    MayDay::Error("GeometryShop uses strong construction only");
  }
  GeometryShop(const GeometryShop& a_workshopin)
  {
    MayDay::Error("GeometryShop disallows copy contruction");
  }
  void operator=(const GeometryShop& a_workshopin)
  {
    MayDay::Error("GeometryShop disallows the assignment operator");
  }

};
#endif
