/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG


#include "HumpChannelBL.H"
#include "Box.H"
#include "PolyGeom.H"
/*****/
/*****/
HumpChannelBL::HumpChannelBL(const RealVect& a_origin,
					   const int&  a_upDir,
					   const int&  a_indepVar, 
					   const Real& a_start, 
					   const Real& a_height, 
					   const Real& a_tight)
{
  for (int idir=0;idir<SpaceDim;idir++)
    {
      m_origin[idir]=a_origin[idir];
    }
  assert((a_upDir >= 0) && (a_upDir < SpaceDim));
  m_upDir.first = a_upDir;
  //always point up
  m_upDir.second = Side::Hi;
  m_indepVar = a_indepVar;
  m_start = a_start;
  m_height = a_height;
  m_tight = a_tight;
}
/*****/
/*****/
HumpChannelBL::~HumpChannelBL()
{
}
/*****/
/*****/
bool 
HumpChannelBL::isRegular(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is above the face, the box is regular
  bool allAbove = true;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allAbove; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
//        cout<<corner[m_indepVar]<<" < "<<m_start<<" and "<<corner[m_upDir.first]<<" < "<<m_height<<endl;
      if(m_upDir.second == Side::Hi)
        {
	  if(corner[m_upDir.first] < funcVal)
	    allAbove = false;
        }
      else
        {
          if(corner[m_upDir.first] > funcVal)
            allAbove = false;
        }
    }
  return allAbove;
}
/*****/
/*****/
bool 
HumpChannelBL::isCovered(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is below the face, the box is covered
  bool allBelow = true;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allBelow; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
      if(m_upDir.second == Side::Hi)
        {
          if(corner[m_upDir.first] > funcVal)
            allBelow = false;
        }
      else
        {
          if(corner[m_upDir.first] < funcVal)
            allBelow = false;
        }
    }
  return allBelow;
}
/*****/
/*****/
pair<int, Side::LoHiSide> HumpChannelBL::upDirection(const RealVect& a_midpt,const IntVect& a_iv) const 
{
  return m_upDir;
}
/*****/
/*****/
Real 
HumpChannelBL::localFuncValue(const RealVect& a_indCoords, 
                       const int& a_upDir,
                       const IntVect& a_iv, 
                       const Box& a_domain, 
                       const RealVect& a_origin, 
                       const Real& a_dx) const 
{
  assert(a_upDir == m_upDir.first);
  Real retval;
  Real offset = a_indCoords[m_indepVar]-m_start;

  retval = m_height/2.*exp(-(offset*offset)/(m_tight*m_height/2000.));

  return retval;
}
/*****/
/*****/
BaseLevelSet* 
HumpChannelBL::new_baseLevelSet() const       
{
  HumpChannelBL* rampPtr = new HumpChannelBL(m_origin,m_upDir.first,m_indepVar,
                               m_start,m_height,m_tight);
  return static_cast<BaseLevelSet*>(rampPtr); 
}
