/*   _______              __
    / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "Box.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "Vector.H"

#include "PolyGeom.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"

#include "ImplicitFunctionBL.H"
#include "BaseImplicitFunction.H"

ImplicitFunctionBL::ImplicitFunctionBL(const BaseImplicitFunction& a_implicitFunction)
{
  // Level sets are being used - flag to GeometryShop
  bool useLevelSet = true;
  setUseLevelSet(useLevelSet);

  // Copy the implicit function passed in
  m_implicitFunction = a_implicitFunction.newImplicitFunction();
}

ImplicitFunctionBL::~ImplicitFunctionBL()
{
  // Delete the copied implicit function
  delete m_implicitFunction;
}

bool ImplicitFunctionBL::isRegular(const Box&      a_region,
                                   const Box&      a_domain,
                                   const RealVect& a_origin,
                                   const Real&     a_dx) const
{
  // Iterate over each cell in the box to check if regular
  bool inTheFluid = true;

  // All corner indices for the current box
  Box allCorners(a_region);
  allCorners.surroundingNodes();

  // If every corner is inside, the box is regular
  BoxIterator bit(allCorners);
  for (bit.reset(); bit.ok() && inTheFluid; ++bit)
  {
    // Current corner
    IntVect corner = bit();

    // Compute physical coordinate of corner
    RealVect physCorner;
    for (int idir = 0; idir < SpaceDim; ++idir)
    {
      physCorner[idir] = a_dx*(corner[idir]) + a_origin[idir];
    }

    // If the implicit function value is positive then the current corner is
    // outside the domain
    Real functionValue = m_implicitFunction->value(physCorner);

    if (functionValue > 0.0 )
    {
      inTheFluid = false;
    }
  }

  return inTheFluid;
}

bool ImplicitFunctionBL::isCovered(const Box&      a_region,
                                   const Box&      a_domain,
                                   const RealVect& a_origin,
                                   const Real&     a_dx) const
{
  // Iterate over each cell in the box to check if covered
  bool outOfTheFluid = true;

  // All corner indices for the current box
  Box allCorners(a_region);
  allCorners.surroundingNodes();

  // If every corner is outside, the box is covered
  BoxIterator bit(allCorners);
  for (bit.reset();bit.ok() && outOfTheFluid;++bit)
  {
    // Current corner
    IntVect corner = bit();

    // Compute physical coordinate of corner
    RealVect physCorner;
    for (int idir = 0; idir < SpaceDim; ++idir)
    {
      physCorner[idir] = a_dx*(corner[idir]) + a_origin[idir];
    }

    // If the implicit function value is negative then the current corner is
    // inside the domain
    Real functionValue = m_implicitFunction->value(physCorner);

    if (functionValue < 0.0 )
    {
      outOfTheFluid = false;
    }
  }

  return outOfTheFluid;
}

pair<int, Side::LoHiSide> ImplicitFunctionBL::upDirection(const RealVect& a_midPt,
                                                          const IntVect&  a_iv) const
{
  pair<int, Side::LoHiSide> retval;

  // Choose arbitrary value - not used by GeometryShop for Level sets
  retval.first = 0;
  retval.second=Side::Lo;

 return retval;
}

Real ImplicitFunctionBL::localFuncValue(const RealVect& a_indCoords,
                                        const int&      a_upDir,
                                        const IntVect&  a_iv,
                                        const Box&      a_domain,
                                        const RealVect& a_origin,
                                        const Real&     a_dx) const

{
  // Return the implicit function value
  Real retval = m_implicitFunction->value(a_indCoords);

  return retval;
}


BaseLevelSet* ImplicitFunctionBL::new_baseLevelSet() const
{
  ImplicitFunctionBL* ImplicitFunctionPtr;
  ImplicitFunctionPtr = new ImplicitFunctionBL(*m_implicitFunction);

  return static_cast<BaseLevelSet*>(ImplicitFunctionPtr);
}
