/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "RealVect.H"

#include "IntersectionIF.H"

IntersectionIF::IntersectionIF(const BaseImplicitFunction& a_impFunc1,
                         const BaseImplicitFunction& a_impFunc2)
{
  // Number of implicit function in intersection
  m_numFuncs = 2;

  // Vector of implicit function pointers
  m_impFuncs.resize(m_numFuncs);

  // Make copies of the implicit functions
  m_impFuncs[0] = a_impFunc1.newImplicitFunction();
  m_impFuncs[1] = a_impFunc2.newImplicitFunction();
}

IntersectionIF::IntersectionIF(const Vector<BaseImplicitFunction *>& a_impFuncs)
{
  // Number of implicit function in intersection
  m_numFuncs = a_impFuncs.size();

  // Vector of implicit function pointers
  m_impFuncs.resize(m_numFuncs);

  // Make copies of the implicit functions
  for (int ifunc = 0; ifunc < m_numFuncs; ifunc++)
  {
    if (a_impFuncs[ifunc] == NULL)
    {
      m_impFuncs[ifunc] = NULL;
    }
    else
    {
      m_impFuncs[ifunc] = a_impFuncs[ifunc]->newImplicitFunction();
    }
  }
}

IntersectionIF::IntersectionIF(const IntersectionIF& a_inputIF)
{
  // Number of implicit function in intersection
  m_numFuncs = a_inputIF.m_impFuncs.size();

  // Vector of implicit function pointers
  m_impFuncs.resize(m_numFuncs);

  // Make copies of the implicit functions
  for (int ifunc = 0; ifunc < m_numFuncs; ifunc++)
  {
    if (a_inputIF.m_impFuncs[ifunc] == NULL)
    {
      m_impFuncs[ifunc] = NULL;
    }
    else
    {
      m_impFuncs[ifunc] = a_inputIF.m_impFuncs[ifunc]->newImplicitFunction();
    }
  }
}

IntersectionIF::~IntersectionIF()
{
  // Delete all the copies
  for (int ifunc = 0; ifunc < m_numFuncs; ifunc++)
  {
    if (m_impFuncs[ifunc] != NULL)
    {
      delete m_impFuncs[ifunc];
    }
  }
}

Real IntersectionIF::value(const RealVect& a_point) const
{
  // Maximum of the implicit functions values
  Real retval;

  retval = 0.0;

  // Find the maximum value and return it
  if (m_numFuncs > 0)
  {
    retval = m_impFuncs[0]->value(a_point);

    for (int ifunc = 1; ifunc < m_numFuncs; ifunc++)
    {
      Real cur;

      cur = m_impFuncs[ifunc]->value(a_point);
      if (retval < cur)
      {
        retval = cur;
      }
    }
  }

  return retval;
}

BaseImplicitFunction* IntersectionIF::newImplicitFunction() const
{
  IntersectionIF* intersectionPtr = new IntersectionIF(m_impFuncs);

  return static_cast<BaseImplicitFunction*>(intersectionPtr);
}
