/* _______              __LSet
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "LSetBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
#include "Vector.H"
/*****/
/*****/
LSetBL::LSetBL(const FArrayBox& a_distFunction,
               const RealVect&  a_distFuncDx,
               const RealVect&  a_distFuncOrigin)
{
  bool useLevelSet = true;
  setUseLevelSet(useLevelSet);

  m_distFunction.define(a_distFunction.box(),1);
  m_distFunction.copy(a_distFunction);

  m_distFuncDx     = a_distFuncDx;
  m_distFuncOrigin = a_distFuncOrigin;

  m_dataBox = m_distFunction.box();

  for (int idir = 0; idir <SpaceDim; ++idir)
    {
      m_distFuncHiEnd[idir] = m_distFuncOrigin[idir] + m_dataBox.bigEnd(idir)*m_distFuncDx[idir];
    }
   
  //m_numCorners and pows[spacedim] are for 
  //looping over intersections of grid lines 
  m_numCorners = 1;
  int pows[SpaceDim];
  for (int idir = 0; idir <SpaceDim; ++idir)
    {
      pows[idir]     = m_numCorners;
      m_numCorners   *= 2;
    }
  m_corners.resize(m_numCorners);
  //m_corners orders the points by:
  //000,001,010,011,100,101,110,111 in 3D
  //and 00,01,10,11 in 2D
  
  for(int iCorner = 0; iCorner < m_numCorners; ++iCorner)
    {
      for (int idir = 0; idir <SpaceDim; ++idir)
        {
          if ((iCorner & pows[SpaceDim - idir -1]) == 0)
            {
              m_corners[iCorner][idir] = 0;
            }
          else
            {
              m_corners[iCorner][idir] = 1;
            }

        }
    }

}

LSetBL::~LSetBL()
{
}
 
bool LSetBL::isRegular(const Box&           a_region,
                       const Box&           a_domain,
                       const RealVect&      a_origin,
                       const Real&          a_dx) const
{
  //iterate over each cell in the box to check if regular
  bool inTheFluid = true;

  BoxIterator bit(a_region);
  for(bit.reset();bit.ok() && inTheFluid;++bit)
    {
      IntVect iv=bit();
      //if every corner is regular, the box is regular
      for(int iCorner = 0; (iCorner < m_numCorners) && inTheFluid; iCorner++)
        {
          IntVect corner = iv; 
          corner += m_corners[iCorner];
          RealVect physCorner;

          //a_origin is for the computational domain
          for (int idir = 0; idir < SpaceDim; ++idir)
            {
              physCorner[idir] = a_dx*(corner[idir]) + a_origin[idir]; 
            } 
          int bogusUpDir = 0;
          Real signedDistance = localFuncValue(physCorner,bogusUpDir,corner,a_domain,a_origin,a_dx);
          if(signedDistance > 0.0 )
            {
              inTheFluid = false;
            }
        }
    }
  return inTheFluid;
}

bool LSetBL::isCovered(const Box&      a_region,
                       const Box&      a_domain,
                       const RealVect& a_origin,
                       const Real&     a_dx)const
{
  //iterate over each cell in the box to check if regular
  bool outOfTheFluid = true;

  BoxIterator bit(a_region);
  for(bit.reset();bit.ok() && outOfTheFluid;++bit)
    {
      IntVect iv=bit();
      //if every corner is covered, the box is covered
      for(int iCorner = 0; (iCorner < m_numCorners) && outOfTheFluid; iCorner++)
        {
          IntVect corner = iv; 
          corner += m_corners[iCorner];
          RealVect physCorner;
          
          //a_origin is for the computational domain
          for (int idir = 0; idir < SpaceDim; ++idir)
            {
              physCorner[idir] = a_dx*(corner[idir]) + a_origin[idir]; 
            } 
          int bogusUpDir = 0;
          Real signedDistance = localFuncValue(physCorner,bogusUpDir,corner,a_domain,a_origin,a_dx);
          
          if(signedDistance < 0.0 )
            {
              outOfTheFluid = false;
            }
        }
    }
  return outOfTheFluid; 
}

pair<int, Side::LoHiSide> LSetBL::upDirection(const RealVect& midPt,
                                              const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;
  //choose arbitrary value
  retval.first = 0;
  retval.second=Side::Lo;
    
 return retval;
}

Real LSetBL::localFuncValue(const RealVect& a_indCoords, 
                            const int& a_upDir,
                            const IntVect& a_iv, 
                            const Box& a_domain, 
                            const RealVect& a_origin, 
                            const Real& a_dx) const

{
  Real retval = 0.0;
  //m_corners orders the points by:
  //000,001,010,011,100,101,110,111 in 3D
  //and 00,01,10,11 in 2D

#if 0
  Real max = m_distFunction(a_iv,0);
  Real min = m_distFunction(a_iv,0);
#endif
  Vector<Real> distances(m_numCorners);
  IntVect distFuncIv;
  for(int idir = 0; idir < SpaceDim; ++idir)
    {
      if(m_distFuncOrigin[idir] <= a_indCoords[idir] && a_indCoords[idir] <= m_distFuncHiEnd[idir])
        {
          distFuncIv[idir] = (int)floor((a_indCoords[idir] - m_distFuncOrigin[idir])/m_distFuncDx[idir]);
          if (distFuncIv[idir] == m_dataBox.bigEnd(idir))
            {
              distFuncIv[idir] -= 1;
            }
        }
      else
        {
          MayDay::Error("computational domain outside physical domain");
        }
    }

  for (int iCorner = 0; iCorner < m_numCorners; ++iCorner)
    {
      IntVect temp = distFuncIv;
      temp += m_corners[iCorner];
      distances[iCorner] = m_distFunction(temp,0);
#if 0
      if (distances[iCorner] > max)
        {
          max = distances[iCorner];
        }
      else if (distances[iCorner] < min)
        {
          min = distances[iCorner];
        }
#endif
    }
#if 0
  if (max < 0.0 || min > 0.0)
    {
        MayDay::Error("localfuncValue problem");
    }

#endif
  RealVect point;
  //make point relative to the cell center;
  for(int idir = 0; idir < SpaceDim; ++idir)
    {
      //first make point relative to (0,0,0) in the physical coordinates, 
      //then relative to cell corner
      
      point[idir] = (a_indCoords[idir] - m_distFuncOrigin[idir])/m_distFuncDx[idir];
      point[idir] -= distFuncIv[idir];
      point[idir] -= 0.5;
    }
  
  Vector<Real> interpValue1D(m_numCorners/2);
  
  for (int iValue = 0; iValue < m_numCorners/2; ++iValue)
    {
      interpValue1D[iValue] = (0.5 - point[SpaceDim - 1])*distances[2*iValue]
        + (0.5 + point[SpaceDim - 1])*distances[1+2*iValue];
    }
  
  Vector<Real> interpValue2D(m_numCorners/4);
  
  for (int iValue = 0; iValue < m_numCorners/4; ++iValue)
    {
      interpValue2D[iValue] = (0.5 - point[SpaceDim - 2])*interpValue1D[2*iValue]
        + (0.5 + point[SpaceDim - 2])*interpValue1D[1+2*iValue];
    }

#if CH_SPACEDIM == 2

  retval = interpValue2D[0];

#elif CH_SPACEDIM == 3

  Real interpValue3D;
  interpValue3D = (0.5 - point[0])*interpValue2D[0]
    + (0.5 + point[0])*interpValue2D[1];
  retval = interpValue3D;

#endif
 
  return retval;
}


BaseLevelSet* LSetBL::new_baseLevelSet() const       
{
  LSetBL* LSetPtr = new LSetBL(m_distFunction,
                               m_distFuncDx,
                               m_distFuncOrigin);
    
  return static_cast<BaseLevelSet*>(LSetPtr); 
}
