/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _LATHEIF_H_
#define _LATHEIF_H_

#include "MayDay.H"
#include "RealVect.H"
#include "Vector.H"

#include "BaseImplicitFunction.H"

///
/**
    This implicit function takes one or two implicit functions and uses them
    to produce a generalized surface of revolution.
    
    In 3D, if one function is given, then it is restricted to x >= 0 in the
    x-y plane (z = 0), revolved around the y axis, and the y axis is rotated
    to the z axis.  In 2D, the function is restricted to x >= 0 and the x axis
    (y = 0) and revolved around the origin.
    
    In 3D, if two functions are given then the first is restricted to x >= 0
    in the x-y plane (z = 0), revolved around the y axis, and the y axis is
    rotated to the z axis.  At each angle between 0 and 2*Pi, the second
    function restricted to the x axis (y = z = 0) is used to get an angle to
    rotate the first function in the x-y plane with respect to a given point
    before it is rotated about the y axis.  In 2D, the first function is used
    as above and the second function is ignored.

    Note: the second function (if given) should have the same value at x = 0
    and x = 2*Pi (y = z = 0) or a values that differs by some integer multiple
    of 2*Pi.
 */
class LatheIF: public BaseImplicitFunction
{
public:
  ///
  /**
      Constructor specifying one implicit function to be rotated and whether
      the domain is on the inside (a_inside), i.e. where the function is
      negative.
   */
  LatheIF(const BaseImplicitFunction& a_impFunc1,
          const bool&                 a_inside);

  ///
  /**
      Constructor specifying two implicit functions, one to be rotated
      and the other to cause an additional rotation about a_point, and
      whether the domain is on the inside (a_inside), i.e. where the
      function is negative.
   */
  LatheIF(const BaseImplicitFunction& a_impFunc1,
          const BaseImplicitFunction& a_impFunc2,
          const RealVect&             a_point,
          const bool&                 a_inside);

  /// Copy constructor
  LatheIF(const LatheIF& a_inputIF);

  /// Destructor
  virtual ~LatheIF();

  ///
  /**
      Return the value of the function at a_point.
   */
  virtual Real value(const RealVect& a_point) const;

  virtual BaseImplicitFunction* newImplicitFunction() const;

protected:
  BaseImplicitFunction* m_impFunc1; // implicit function to rotate
  BaseImplicitFunction* m_impFunc2; // implicit function to use for rotation

  RealVect              m_point;    // point for center of second rotation

  bool                  m_inside;   // inside flag

private:
  LatheIF()
  {
    MayDay::Error("LatheIF uses strong construction");
  }

  void operator=(const LatheIF& a_inputIF)
  {
    MayDay::Error("LatheIF doesn't allow assignment");
  }
};

#endif
