/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "RealVect.H"

#include "LatheIF.H"

LatheIF::LatheIF(const BaseImplicitFunction& a_impFunc1,
                 const bool&                 a_inside)
{
  // Make a copy of the implicit function and note only one was given
  m_impFunc1 = a_impFunc1.newImplicitFunction();
  m_impFunc2 = NULL;

  // Save inside flag
  m_inside = a_inside;
}

LatheIF::LatheIF(const BaseImplicitFunction& a_impFunc1,
                 const BaseImplicitFunction& a_impFunc2,
                 const RealVect&             a_point,
                 const bool&                 a_inside)
{
  // Make a copy of both implicit functions
  m_impFunc1 = a_impFunc1.newImplicitFunction();
  m_impFunc2 = a_impFunc2.newImplicitFunction();

  // Save the point of rotation
  m_point = a_point;

  // Save inside flag
  m_inside = a_inside;
}

LatheIF::LatheIF(const LatheIF& a_inputIF)
{
  // Make a copy of both implicit functions (the second may be NULL)
  m_impFunc1 = a_inputIF.m_impFunc1->newImplicitFunction();
  m_impFunc2 = a_inputIF.m_impFunc2->newImplicitFunction();

  // Save the point of rotation
  m_point = a_inputIF.m_point;

  // Save inside flag
  m_inside = a_inputIF.m_inside;
}

LatheIF::~LatheIF()
{
  delete m_impFunc1;
  if (m_impFunc2 != NULL)
  {
    delete m_impFunc2;
  }
}

Real LatheIF::value(const RealVect& a_point) const
{
  Real retval;

  // Get r value
  Real r = 0.0;
  for (int i = 0; i < 2; i++)
  {
    r += a_point[i] * a_point[i];
  }
  r = sqrt(r);

#if CH_SPACEDIM == 2
  RealVect coord(r,0.0);

  retval = m_impFunc1->value(coord);
#elif CH_SPACEDIM == 3
  Real x = a_point[0];
  Real y = a_point[1];
  Real z = a_point[2];

  Real r1,z1;

  if (m_impFunc2 == NULL)
  {
    r1 = r;
    z1 = z;
  }
  else
  {
    Real theta = atan2(y,x);
    RealVect coord(theta,0.0,0.0);

    Real angle = -m_impFunc2->value(coord);

    r -= m_point[0];
    z -= m_point[1];

    r1 = cos(angle)*r - sin(angle)*z;
    z1 = sin(angle)*r + cos(angle)*z;

    r1 += m_point[0];
    z1 += m_point[1];
  }

  RealVect coord(r1,z1,0.0);

  retval = m_impFunc1->value(coord);
#else
  this is bogus;
#endif

  // Change the sign to change inside to outside
  if (!m_inside)
  {
    retval = -retval;
  }
  
  return retval;
}

BaseImplicitFunction* LatheIF::newImplicitFunction() const
{
  LatheIF* lathePtr;

  if (m_impFunc2 == NULL)
  {
   lathePtr = new LatheIF(*m_impFunc1,m_inside);
  }
  else
  {
   lathePtr = new LatheIF(*m_impFunc1,*m_impFunc2,m_point,m_inside);
  }

  return static_cast<BaseImplicitFunction*>(lathePtr);
}
