/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "MultiSphereBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
/*****/
MultiSphereBL::MultiSphereBL(const Vector<Real>&      a_radius,
                             const Vector<RealVect>&  a_center,
                             const Real&              a_dx)
{
  assert(a_radius.size() > 0);
  assert(a_radius.size() == a_center.size());
  m_numSpheres   = a_radius.size();
  m_radius       = a_radius;
  m_center       = a_center;
  m_dx           = a_dx;
}
/*****/
/*****/
int 
MultiSphereBL::closestSphere(const RealVect& a_vec) const
{
  assert(m_numSpheres > 0);
  //find the distance to each sphere
  Vector<Real> sqDistToSphere(m_numSpheres, 0.0);
  for(int isphere = 0; isphere < m_numSpheres; isphere++)
    {
      Real thisDist = 0.0;
      const RealVect& thisCent = m_center[isphere];
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          Real diff = thisCent[idir] - a_vec[idir];
          thisDist += diff*diff;
        }
      sqDistToSphere[isphere] = thisDist;
    }
  //return the index of the closest sphere 
  int iclose = 0;
  Real minDist = sqDistToSphere[0];
  for(int isphere = 1; isphere < m_numSpheres; isphere++)
    {
      if(sqDistToSphere[isphere] < minDist)
        {
          iclose = isphere;
          minDist = sqDistToSphere[isphere];
        }
    }
  return iclose;
}
/*****/
/*****/
MultiSphereBL::~MultiSphereBL()
{
}
/*****/
//finds the component of  a RealVect with largest magnitude
Real MultiSphereBL::max(const RealVect& vec)const
  
{
      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<abs(vec[idir]))retval=abs(vec[idir ]);
    }
  return retval;
}
//finds the eulclidean distance of a RealVect from the origin
Real MultiSphereBL::magnitude( const RealVect& vec)const
{ 
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }    
  retval=sqrt(retval);
  return retval;
}
 
bool 
MultiSphereBL::isRegular(const Box&           a_region,
                         const Box&           a_domain,
                         const RealVect&      a_origin,
                         const Real&          a_dx) const
{
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              int isphere = closestSphere(corner);
              RealVect center = m_center[isphere];
              Real     radius = m_radius[isphere];

              RealVect translatedCorner=corner;
              translatedCorner-=center;
              //inside reg always false for this geom
              //for domain=outside of the sphere do...
              if(magnitude(translatedCorner) < radius)allAbove = false;
            }
        } 
    }
  
  // if ( allAbove)   pout()<< "region "<<a_region<<" is regular"<<endl;
  // if (!allAbove)   pout()<< "region "<<a_region<<" is not regular"<<endl;  
 
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
MultiSphereBL::isCovered(const Box&      a_region,
                         const Box&      a_domain,
                         const RealVect& a_origin,
                         const Real&     a_dx)const
{
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          RealVect translatedCorner=corner;
          int isphere = closestSphere(corner);
          RealVect center = m_center[isphere];
          Real     radius = m_radius[isphere];

          translatedCorner-=center;
          //For domain=the inside of the sphere do...
          //never inside sphere
          if (magnitude(translatedCorner) > radius)allBelow = false;
        }
    }
  // if ( allBelow)   pout()<< "region "<<a_region<<" is covered"<<endl;
  // if (!allBelow)   pout()<< "region "<<a_region<<" is not covered"<<endl;

  return allBelow;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

pair<int, Side::LoHiSide> 
MultiSphereBL::upDirection(const RealVect& midPt,const IntVect& a_iv) const
{
  int isphere = closestSphere(midPt);
  RealVect center = m_center[isphere];
  //Real     radius = m_radius[isphere];

  pair<int, Side::LoHiSide> retval;
  RealVect gradient;
  //On a Sphere, gradient porportional to position realative to the center
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      gradient[idir]=midPt[idir]-center[idir];
    }   
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (abs(gradient[idir])>=max(gradient)) retval.first=idir;
    }
  if (gradient[retval.first]>=0.)
    {
      retval.second=Side::Hi;
    }
  else  if (gradient[retval.first]< 0.)
    {
      retval.second=Side::Lo;
    }
  return retval;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Real 
MultiSphereBL::localFuncValue(const RealVect& a_indCoords, 
                              const int& a_upDir,
                              const IntVect& a_iv, 
                              const Box& a_domain, 
                              const RealVect& a_origin, 
                              const Real& a_dx) const

{
  int isphere = closestSphere(a_indCoords);
  RealVect center = m_center[isphere];
  Real     radius = m_radius[isphere];
 
  Real retval =0.;
  for(int idir = 0; idir< SpaceDim; idir++)
    {
      if (idir !=a_upDir)
        {
          retval+=pow(a_indCoords[idir]-center[idir],2);
        }       
    }
  if (a_dx*a_iv[a_upDir]>= center[a_upDir])
    {
      retval=sqrt(Abs(pow(radius,2)-retval))+center[a_upDir];
    }
  else
    {
      retval=-sqrt(Abs(pow(radius,2)-retval))+center[a_upDir];
    }


  
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
MultiSphereBL::new_baseLevelSet() const       
{
  MultiSphereBL* spherePtr = new MultiSphereBL(m_radius,
                                               m_center,
                                               m_dx);
    
  return static_cast<BaseLevelSet*>(spherePtr); 
}
