/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "OvoidBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
OvoidBL::OvoidBL(const Real&      a_radius,
                 const RealVect&  a_centerLo,
                 const Real&      a_axialDist,
                 const int&       a_axis,
                 const Real&      a_dx)
{
  assert(a_radius    > 0.);
  assert(a_dx        > 0.);
  assert(a_axialDist > 0.);

  assert(a_axis >= 0);
  assert(a_axis < SpaceDim);

  m_radius      = a_radius;
  m_centerLo    = a_centerLo;
  m_axialDist   = a_axialDist;
  m_axis        = a_axis;
  m_dx          = a_dx;

  m_centerHi = m_centerLo;
  m_centerHi[m_axis] += a_axialDist;
}
/*****/
void 
OvoidBL::distToAxis(Real& a_distToAxis,
                    const RealVect& a_vec) const
{
  a_distToAxis = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(idir != m_axis)
        {
          Real diff = m_centerLo[idir] - a_vec[idir];
          a_distToAxis += diff*diff;
        }
    }
  a_distToAxis = sqrt(a_distToAxis);
}
/*****/
void 
OvoidBL::distToSpheres(Real& a_distToLo,
                       Real& a_distToHi,
                       const RealVect& a_vec) const
{
  a_distToLo = 0.0;
  a_distToHi = 0.0;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Real diffLo = m_centerLo[idir] - a_vec[idir];
      Real diffHi = m_centerHi[idir] - a_vec[idir];

      a_distToLo += diffLo*diffLo;
      a_distToHi += diffHi*diffHi;
    }
  a_distToLo = sqrt(a_distToLo);
  a_distToHi = sqrt(a_distToHi);
}
/*****/
OvoidBL::~OvoidBL()
{
}
/*****/
Real OvoidBL::max(const RealVect& vec)const
{
      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<abs(vec[idir]))retval=abs(vec[idir ]);
    }
  return retval;
}
/****/
Real OvoidBL::magnitude( const RealVect& vec)const
{ 
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }    
  retval=sqrt(retval);
  return retval;
}
/*****/
bool 
OvoidBL::isRegular(const Box&           a_region,
                   const Box&           a_domain,
                   const RealVect&      a_origin,
                   const Real&          a_dx) const
{
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              Real localRadius;
              if((corner[m_axis] > m_centerHi[m_axis]) ||
                 (corner[m_axis] < m_centerLo[m_axis]))
                {
                  Real distToLo, distToHi;
                  distToSpheres(distToLo, distToHi, corner);
                  localRadius = Min(distToLo, distToHi);
                }
              else
                {
                  distToAxis(localRadius, corner);
                }
              //inside reg always false for this geom
              //for domain=outside of the sphere do...
              if(localRadius < m_radius)
                {
                  allAbove = false;
                }
            }
        } 
    }
  
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
OvoidBL::isCovered(const Box&      a_region,
                   const Box&      a_domain,
                   const RealVect& a_origin,
                   const Real&     a_dx)const
{
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          Real localRadius;
          if((corner[m_axis] > m_centerHi[m_axis]) ||
             (corner[m_axis] < m_centerLo[m_axis]))
            {
              Real distToLo, distToHi;
              distToSpheres(distToLo, distToHi, corner);
              localRadius = Min(distToLo, distToHi);
            }
          else
            {
              distToAxis(localRadius, corner);
            }
          //inside reg always false for this geom
          //for domain=outside of the sphere do...
          if(localRadius >= m_radius)
            {
              allBelow = false;
            }
        }
    }

  return allBelow;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

pair<int, Side::LoHiSide> 
OvoidBL::upDirection(const RealVect& a_midPt,const IntVect& a_iv) const
{
  RealVect gradient;
  if((a_midPt[m_axis] > m_centerHi[m_axis]) ||
     (a_midPt[m_axis] < m_centerLo[m_axis]))
    {
      Real distToLo, distToHi;
      distToSpheres(distToLo, distToHi, a_midPt);
      if(distToLo < distToHi)
        {
          gradient = a_midPt - m_centerLo;
        }
      else
        {
          gradient = a_midPt - m_centerHi;
        }
    }
  else
    {
      gradient = a_midPt - m_centerHi;
      gradient[m_axis] = 0.0;
    }

  pair<int, Side::LoHiSide> retval;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (abs(gradient[idir])>=max(gradient)) retval.first=idir;
    }

  if (gradient[retval.first]>=0.)
    {
      retval.second=Side::Hi;
    }
  else
    {
      retval.second=Side::Lo;
    }
  return retval;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Real 
OvoidBL::localFuncValue(const RealVect& a_indCoords, 
                        const int& a_upDir,
                        const IntVect& a_iv, 
                        const Box& a_domain, 
                        const RealVect& a_origin, 
                        const Real& a_dx) const

{
  RealVect center;
  if((a_indCoords[m_axis] > m_centerHi[m_axis]) ||
     (a_indCoords[m_axis] < m_centerLo[m_axis]))
    {
      Real distToLo, distToHi;
      distToSpheres(distToLo, distToHi, a_indCoords);
      if(distToLo < distToHi)
        {
          center = m_centerLo;
        }
      else
        {
          center= m_centerHi;
        }
    }
  else
    {
      center = m_centerLo;
      center[m_axis] = a_indCoords[m_axis];
    }

  Real retval =0.;
  for(int idir = 0; idir< SpaceDim; idir++)
    {
      if (idir !=a_upDir)
        {
          retval+=pow(a_indCoords[idir]-center[idir],2);
        }       
    }
  if (a_dx*a_iv[a_upDir]>= center[a_upDir])
    {
      retval=sqrt(Abs(pow(m_radius,2)-retval))+center[a_upDir];
    }
  else
    {
      retval=-sqrt(Abs(pow(m_radius,2)-retval))+center[a_upDir];
    }

  return retval;
}
/*****/
BaseLevelSet* 
OvoidBL::new_baseLevelSet() const       
{
  OvoidBL* ovoidPtr = new OvoidBL(m_radius,
                                  m_centerLo,
                                  m_axialDist,
                                  m_axis,
                                  m_dx);
    
  return static_cast<BaseLevelSet*>(ovoidPtr); 
}
