/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "ParabolicMirror.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
ParabolicMirror::ParabolicMirror(const Real&     a_amplitude,
                                 const RealVect& a_center,
                                 const int&      a_updir,
                                 const Real&     a_dx)
{
  m_amplitude      = a_amplitude;
  m_center         = a_center;
  m_dx             = a_dx;
  m_updir          = a_updir;
  if(m_updir == 0)
    m_tandir = 1;
  else
    m_tandir = 0;
}
/*****/
Real 
ParabolicMirror::parabFunction(const Real&     a_y) const
{
  Real ydiff = (a_y - m_center[m_tandir]);
  Real x = m_amplitude*ydiff*ydiff + m_center[m_updir];
  return x;
}
/*****/
ParabolicMirror::~ParabolicMirror()
{
}
/*****/
bool 
ParabolicMirror::isPointCovered(const RealVect& a_vec) const
{
  Real x = a_vec[m_updir];
  Real y = a_vec[m_tandir];
  Real xparab = parabFunction(y);
  bool retval;
  if(m_amplitude > 0) 
    {
      retval = (x < xparab);
    }
  else
    {
      retval = (x > xparab);
    }
  return retval;
}
/*****/
/*****/
bool 
ParabolicMirror::isRegular(const Box&           a_region,
                           const Box&           a_domain,
                           const RealVect&      a_origin,
                           const Real&          a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              bool thisPtCovered = isPointCovered(corner);
              if(thisPtCovered)
                {
                  allAbove = false;
                }
            }
        } 
    }
  
  // if ( allAbove)   pout()<< "region "<<a_region<<" is regular"<<endl;
  // if (!allAbove)   pout()<< "region "<<a_region<<" is not regular"<<endl;  
 
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
ParabolicMirror::isCovered(const Box&      a_region,
                           const Box&      a_domain,
                           const RealVect& a_origin,
                           const Real&     a_dx)const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          bool thisPtCovered = isPointCovered(corner);
          if(!thisPtCovered)
            {
              allBelow = false;
            }
        }
    }
  // if ( allBelow)   pout()<< "region "<<a_region<<" is covered"<<endl;
  // if (!allBelow)   pout()<< "region "<<a_region<<" is not covered"<<endl;

  return allBelow;
}
/*****/
pair<int, Side::LoHiSide> 
ParabolicMirror::upDirection(const RealVect& midPt,const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;
  retval.first = m_updir;
  if(m_amplitude > 0.0)
    retval.second = Side::Hi;
  else
    retval.second = Side::Lo;
  return retval;
}
/*****/
Real 
ParabolicMirror::localFuncValue(const RealVect& a_indCoords, 
                                const int& a_upDir,
                                const IntVect& a_iv, 
                                const Box& a_domain, 
                                const RealVect& a_origin, 
                                const Real& a_dx) const

{
  Real y = a_indCoords[m_tandir];
  Real xparab = parabFunction(y);
  return xparab;
}
/*****/
/*****/
BaseLevelSet* 
ParabolicMirror::new_baseLevelSet() const       
{
  ParabolicMirror* parabolaPtr = new ParabolicMirror(m_amplitude,
                                                     m_center,
                                                     m_updir,
                                                     m_dx);
    
  return static_cast<BaseLevelSet*>(parabolaPtr); 
}
