/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "PolarBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
/*****/
PolarBL::PolarBL(const Real&          a_coeffA,
                 const Real&          a_coeffB,
                 const Real&          a_coeffC,
                 const RealVect&      a_center,
                 const bool&          a_insideRegular,
                 const Real&          a_dx)
{
  assert(a_coeffA >= 2*a_dx);
  // assert(dx>polygeom::get.tolerance());
  
 
  m_coeffA      = a_coeffA;
  m_coeffB     = a_coeffB;
  m_coeffC      = a_coeffC;
  m_center       = a_center;
  m_insideRegular= a_insideRegular;
  m_dx           = a_dx;
}
/*****/
/*****/
PolarBL::~PolarBL()
{
}
/*****/
//finds the component of  a RealVect with largest magnitude
Real PolarBL::maxabs(const RealVect& vec)const
  
{
      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<Abs(vec[idir]))retval=Abs(vec[idir]);
    }
  return retval;
}
//finds the eulclidean distance of a RealVect from the origin
Real PolarBL::magnitude( const RealVect& vec)const
{ 
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }    
  retval=sqrt(retval);
  return retval;
}
 
Real PolarBL::star(const RealVect& vec)const
{
  Real retval=0.;
  Real radius=sqrt(vec[0]*vec[0]+vec[1]*vec[1]);
  Real theta=atan2(vec[1],vec[0]);
  retval=radius-(m_coeffA+m_coeffB*cos(m_coeffC*theta));
  return retval;
}   
Real PolarBL::star(const Real x, Real y)const
{
  Real retval=0.;
  Real radius=sqrt(x*x+y*y);
  Real theta=atan2(y,x);
  retval=radius-(m_coeffA+m_coeffB*cos(m_coeffC*theta));
  return retval;
}   
Real PolarBL::starX(const RealVect& vec)const
{
  Real retval=0.;
  Real radius=sqrt(vec[0]*vec[0]+vec[1]*vec[1]);
  Real theta=atan2(vec[1],vec[0]);
  //  Real temp=m_coeffC*theta;
  //  Real temp2= sin(m_coeffC*theta);
  //  Real temp3=1/(radius*radius);
  retval=(vec[0]/radius)-(m_coeffB*sin(m_coeffC*theta)*m_coeffC*vec[1]/
						  (radius*radius));
  return retval;
}   
Real PolarBL::starY(const RealVect& vec)const
{
  Real retval=0.;
  Real radius=sqrt(vec[0]*vec[0]+vec[1]*vec[1]);
  Real theta=atan2(vec[1],vec[0]);
  retval=(vec[1]/radius)+(m_coeffB*sin(m_coeffC*theta)*m_coeffC*vec[0]/
						  (radius*radius));
  return retval;
}   



bool 
PolarBL::isRegular(const Box&           a_region,
				   const Box&           a_domain,
				   const RealVect&      a_origin,
				   const Real&          a_dx) const
{
  

  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
		 
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              RealVect translatedCorner=corner;
              translatedCorner-=m_center;
              //For domain=the inside of the polar do...
              if(m_insideRegular==true)
                {
                  if(star(translatedCorner) > 0.)allAbove = false;
                
                  
                }
              //for domain=outside of the polar do...
              else 
                {
                  if(star(translatedCorner) <0.)allAbove = false;
                  // pout()<<star(translatedCorner)<<endl;
                  // pout()<<translatedCorner[0]<<"x coord"<<endl;
                  // pout()<<translatedCorner[1]<<"y coord"<<endl;
                }
            }
        } 
    }
  
  //  if ( allAbove)   pout()<< "region "<<a_region<<" is regular"<<endl;
  // if (!allAbove)   pout()<< "region "<<a_region<<" is not regular"<<endl;  
 
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
PolarBL::isCovered(const Box&      a_region,
				   const Box&      a_domain,
				   const RealVect& a_origin,
				   const Real&     a_dx)const
{
  
 
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
	 
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          RealVect translatedCorner=corner;
          translatedCorner-=m_center;
          //For domain=the inside of the polar do...
          if(m_insideRegular==true)
            {
              if(star(translatedCorner) < 0.)allBelow = false;
          
                  
            }
          //for domain=outside of the polar do...
          else 
            {
              if (star(translatedCorner) >0.)allBelow = false;
            }
        }
    }
  //  if ( allBelow)   pout()<< "region "<<a_region<<" is covered"<<endl;
  // if (!allBelow)   pout()<< "region "<<a_region<<" is not covered"<<endl;

  return allBelow;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


pair<int, Side::LoHiSide> 
PolarBL::upDirection(const RealVect& midPt,const IntVect& a_iv) const
{
 
  pair<int, Side::LoHiSide> retval;
  RealVect gradient;
  RealVect edgeCenter;
  //Use starX and starY to calculate the gradient
  
  {
	edgeCenter-=m_center;
	gradient[0]=starX(edgeCenter);
	gradient[1]=starY(edgeCenter);
	// pout() <<gradient[idir]<<"=gradient("<<idir<<")"<<endl;
  }   
  Real maxgrad = 0.0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
    
      if (Abs(gradient[idir])>=maxgrad) 
		{
		  retval.first=idir;
		  maxgrad = Abs(gradient[idir]);
		}
    }
  
  if (gradient[retval.first]>=0.&& m_insideRegular)retval.second=Side::Lo;
  if (gradient[retval.first]< 0.&& m_insideRegular)retval.second=Side::Hi;
  if (gradient[retval.first]>=0.&&!m_insideRegular)retval.second=Side::Hi;
  if (gradient[retval.first]< 0.&&!m_insideRegular)retval.second=Side::Lo;
  

  return retval;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Real 
PolarBL::localFuncValue(const RealVect& a_indCoords, 
						const int& a_upDir,
						const IntVect& a_iv, 
						const Box& a_domain, 
						const RealVect& a_origin, 
						const Real& a_dx) const

{
  RealVect cellCenter;
  Real min,max;
  for (int idir=0;idir<SpaceDim;++idir)
	{
	   cellCenter[idir]=(a_iv[idir]+.5)*a_dx;
	}
 
   Real retval2 =0.;
  Real tolerance=.00000000001;
  
  //case: y is dependent variable
  if (a_upDir==1)  
    {
	  min=cellCenter[1]-2*a_dx-m_center[1];
	  max=cellCenter[1]+2*a_dx-m_center[1];
	  Real x=a_indCoords[0]-m_center[0];
	  // Real min=pow(Abs(m_coeffA)-Abs(m_coeffB),2)-(x*x);
	  // if (min<=0.)min=0.;
	  //  Real max=pow(Abs(m_coeffA)+Abs(m_coeffB),2)-(x*x);
	  //  assert(max>=0.);
	  // if ((a_iv[1]+.5)*a_dx-m_center[1]>0.)
	  //	{
	  //	  min=sqrt((min));
	  //	  max=sqrt((max));
	  //	}
	  // else
	  //	{ 
	  //  Real temp=max;
	  //	  max=-sqrt(min);
	  //	  min=-sqrt(temp);
	  //	}
	  // Real stretch=Abs(min-max);
	  // min=min-.2*stretch;
	  // max=max+.2*stretch;
      retval2=(min+max)/2.;
      if (star(x,min)*star(x,max)> 0)
		{
		  MayDay::Error("not solvable in polarlg::localfuncval");
		}

	  int maxiter = 1000;
	  int iiter = 1;
	  while((Abs(min-max)>tolerance) && (iiter < maxiter))
		{ 
		  if (star(x,min)*star(x,retval2)>0) min=retval2;
		  else max=retval2;
		  retval2=(min+max)/2.;
		  iiter++;
		} 
	  if(iiter >= maxiter)
		{
		  MayDay::Error("did not converge");
		}

     //  if (Abs(star(x,retval2))>tolerance)
//         {
//           pout() <<a_iv <<" is a_iv"<<endl;
//           pout() <<fabs(star(x,retval2))<<"star(x,retval2)"<<endl;
//           pout() <<x<<"=x"<<endl;
//           pout()<<star(x,min)<<"=star(x,min)"<<endl;
//           pout()<<star(x,max)<<"=star(x,max)"<<endl;
//           pout() <<a_upDir<<"=a_upDir"<<endl;
//         }

    }

  //case x is dependent variable
  else if(a_upDir==0)  
    {
	   min=cellCenter[0]-2*a_dx-m_center[0];
	   max=cellCenter[0]+2*a_dx-m_center[0];
	   Real y=a_indCoords[1]-m_center[1];
	 //   Real min=(m_coeffA-m_coeffB)*(m_coeffA-m_coeffB)-(y*y);
// 	   if (min<=0.)min=0.;
// 	   Real max=(m_coeffA+m_coeffB)*(m_coeffA+m_coeffB)-(y*y);
	   
// 	   if ((a_iv[0]+.5)*a_dx-m_center[0]>0.)
		 
//         {
// 		  assert(max>0.);
//           min=sqrt(min);
//           max=sqrt(max);
//         }
//       else
//         { 
//           Real temp=max;
//           max=-sqrt(Abs(min));
//           min=-sqrt(Abs(temp));
//         }
//       Real stretch=abs(min-max);
//       min=min-.2*stretch;
//       max=max+.2*stretch;
      retval2=(min+max)/2.;
      // pout() <<star(min,y)<<"=star(min,y)"<<endl;
      //  pout() <<star(max,y)<<"=star(max,y)"<<endl;
      
      
      if (star(min,y)*star(max,y)> 0)
		{
		  MayDay::Error("not solvable in polarlg::localfuncval 2");
		}

	  int maxiter = 1000;
	  int iiter = 1;
	  while((Abs(min-max)>tolerance) && (iiter < maxiter))
		{ 
		  if (star(min,y)*star(retval2,y)>0) min=retval2;
		  else max=retval2;
		  retval2=(min+max)/2.;
		  iiter++;
		} 
	  if(iiter >= maxiter)
		{
		  MayDay::Error("did not converge 2");
		}

    }
  else
    {
      pout() << "a_upDir must be 0 or 1"<<endl;
    }
 
  {

  }  
 
  return retval2+m_center[a_upDir];
}
/*****/
/*****/
BaseLevelSet* 
PolarBL::new_baseLevelSet() const       
{
  PolarBL* polarPtr = new PolarBL(m_coeffA,
                                  m_coeffB,
                                  m_coeffC,
                                  m_center,
                                  m_insideRegular,
                                  m_dx);
    
  return static_cast<BaseLevelSet*>(polarPtr); 
}
