/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG


#include "RampBL.H"
#include "Box.H"
#include "PolyGeom.H"
/*****/
/*****/
RampBL::RampBL(const int&  a_upDir, 
               const int&  a_indepVar, 
               const Real& a_start,
               const Real& a_rampSlope)
{
  assert((a_upDir >= 0) && (a_upDir < SpaceDim));
  assert((a_indepVar >= 0) && (a_indepVar < SpaceDim));
  m_upDir.first = a_upDir;
  //always point up
  m_upDir.second = Side::Hi;

  //normal involved in 3D set to zero
  m_rampNormal = RealVect::Zero;
  m_rampNormal[a_upDir] = 1.0;
  m_rampNormal[a_indepVar] = -a_rampSlope;
  Real sumSquare;
  //make the normal a unit vector
  PolyGeom::unifyVector(m_rampNormal, sumSquare);

  m_alpha = m_rampNormal[a_indepVar]*a_start;
}
/*****/
/*****/
RampBL::RampBL(const RealVect& a_rampNormal,
               const Real&     a_alpha)
{
  m_rampNormal = a_rampNormal;
  Real sumSquare;
  //make the normal a unit vector
  PolyGeom::unifyVector(m_rampNormal, sumSquare);
  m_alpha = a_alpha/sumSquare;

  //pick updir to be direction with biggest normal
  int ibigNorm = 0;
  for(int idir = 1; idir < SpaceDim; idir++)
    {
      if(Abs(m_rampNormal[idir]) > Abs(m_rampNormal[idir-1]))
        {
          ibigNorm = idir;
        }
    }

  m_upDir.first = ibigNorm;
  if(m_rampNormal[ibigNorm] > 0.0)
    {
      m_upDir.second = Side::Hi;
    }
  else
    {
      m_upDir.second = Side::Lo;
    }
}
/*****/
/*****/
RampBL::~RampBL()
{
}
/*****/
/*****/
bool 
RampBL::isRegular(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is above the face, the box is regular
  bool allAbove = true;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allAbove; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
      if(m_upDir.second == Side::Hi)
        {
          if(corner[m_upDir.first] < funcVal)
            allAbove = false;
        }
      else
        {
          if(corner[m_upDir.first] > funcVal)
            allAbove = false;
        }
    }
  return allAbove;
}
/*****/
/*****/
bool 
RampBL::isCovered(const Box& a_region,
                  const Box& a_domain,
                  const RealVect& a_origin,
                  const Real& a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  Vector<IntVect> cornerIVs = getBoxCornerCells(a_region, a_origin, a_dx);
  //if every corner is below the face, the box is covered
  bool allBelow = true;
  for(int ilocs = 0; (ilocs < cornerLocs.size()) && allBelow; ilocs++)
    {
      const RealVect&  corner = cornerLocs[ilocs];
      const IntVect& iv = cornerIVs[ilocs];
      Real funcVal = localFuncValue(corner, m_upDir.first, iv, 
                                    a_domain, a_origin, a_dx);
      if(m_upDir.second == Side::Hi)
        {
          if(corner[m_upDir.first] > funcVal)
            allBelow = false;
        }
      else
        {
          if(corner[m_upDir.first] < funcVal)
            allBelow = false;
        }
    }
  return allBelow;
}
/*****/
/*****/
pair<int, Side::LoHiSide> RampBL::upDirection(const RealVect& a_midpt,const IntVect& a_iv) const 
{
  return m_upDir;
}
/*****/
/*****/
Real 
RampBL::localFuncValue(const RealVect& a_indCoords, 
                       const int& a_upDir,
                       const IntVect& a_iv, 
                       const Box& a_domain, 
                       const RealVect& a_origin, 
                       const Real& a_dx) const 
{
  assert(a_upDir == m_upDir.first);
  Real retval = m_alpha;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(idir != m_upDir.first)
        {
          retval -= m_rampNormal[idir]*a_indCoords[idir];
        }
    }
  retval /= m_rampNormal[m_upDir.first];
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
RampBL::new_baseLevelSet() const       
{
  RampBL* rampPtr = new RampBL(m_rampNormal,
                               m_alpha);
  return static_cast<BaseLevelSet*>(rampPtr); 
}
