/* _______              __Sphere
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/  
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
//  ANAG, LBNL, DTG
#include <cmath>
#include "Box.H"
#include "SphereBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "LoadBalance.H"
#include "PolyGeom.H"
/*****/
/*****/
SphereBL::SphereBL(const Real&          a_radius,
                   const RealVect&      a_center,
                   const bool&          a_insideRegular,
                   const Real&          a_dx)
{
  assert(a_radius >= 2*a_dx);
  // assert(dx>polygeom::get.tolerance());
  
 
  m_radius       = a_radius;
  m_center       = a_center;
  m_insideRegular= a_insideRegular;
  m_dx           = a_dx;
}
/*****/
/*****/
SphereBL::~SphereBL()
{
}
/*****/
//finds the component of  a RealVect with largest magnitude
Real SphereBL::max(const RealVect& vec)const
  
{
      
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (retval<abs(vec[idir]))
        {
          retval=abs(vec[idir ]);
        }
    }
  return retval;
}
//finds the eulclidean distance of a RealVect from the origin
Real SphereBL::magnitude( const RealVect& vec)const
{ 
  Real retval=0;
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      retval+=vec[idir]*vec[idir];
    }   
  // Real temp = pow(retval,0.5);
  retval=sqrt(retval);
  return retval;
}
 
bool 
SphereBL::isRegular(const Box&           a_region,
                    const Box&           a_domain,
                    const RealVect&      a_origin,
                    const Real&          a_dx) const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if regular
  bool allAbove = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allAbove;++bit)
    {
      if (allAbove==true)
        {
          IntVect iv=bit();
          Box cellbox(iv,iv);
          Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
          //if every corner is above the face the box is covered
          for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allAbove; ilocs++)
            {
              const RealVect&  corner = cellboxLocs[ilocs];
              RealVect translatedCorner = corner;
              translatedCorner -= m_center;
              //For domain=the inside of the sphere do...
              if(m_insideRegular==true)
                {
                  if(magnitude(translatedCorner) > m_radius)allAbove = false;
                 
                  
                }
              //for domain=outside of the sphere do...
              else 
                {
                  if(magnitude(translatedCorner) < m_radius)allAbove = false;
                }
            }
        } 
    }
  
  // if ( allAbove)   pout()<< "region "<<a_region<<" is regular"<<endl;
  // if (!allAbove)   pout()<< "region "<<a_region<<" is not regular"<<endl;  
 
  return allAbove;
}
/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
bool 
SphereBL::isCovered(const Box&      a_region,
                    const Box&      a_domain,
                    const RealVect& a_origin,
                    const Real&     a_dx)const
{
  //create the locations of each corner of box
  Vector<RealVect> cornerLocs = getBoxLocations(a_region, a_origin, a_dx);
  //iterate over each cell in the box to check if covered
  bool allBelow = true;
  BoxIterator bit(a_region);
  for(bit.reset();bit.ok()&&allBelow;++bit)
    {
      IntVect iv=bit();
      Box cellbox(iv,iv);
      Vector<RealVect> cellboxLocs = getBoxLocations(cellbox, a_origin, a_dx);
      //if every corner satisfies a condition on distance from the center
      //the cell is regular
      for(int ilocs = 0; (ilocs < cellboxLocs.size()) && allBelow; ilocs++)
        {
          const RealVect&  corner = cellboxLocs[ilocs];
          RealVect translatedCorner = corner;
          translatedCorner -= m_center;
          
          //For domain=the inside of the sphere do...
          if(m_insideRegular==true)
            {
              if(magnitude(translatedCorner) < m_radius)allBelow = false;
          
                  
            }
          //for domain=outside of the sphere do...
          else 
            {
              if (magnitude(translatedCorner) > m_radius)allBelow = false;
            }
        }
    }
  // if ( allBelow)   pout()<< "region "<<a_region<<" is covered"<<endl;
  // if (!allBelow)   pout()<< "region "<<a_region<<" is not covered"<<endl;

  return allBelow;
}
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

pair<int, Side::LoHiSide> SphereBL::upDirection(const RealVect& midPt,
                                                const IntVect& a_iv) const
{
  pair<int, Side::LoHiSide> retval;
  RealVect gradient;
  //On a Sphere, gradient porportional to position realative to the center
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      gradient[idir] = midPt[idir] - m_center[idir];
    }   
  for(int idir = 0; idir <SpaceDim; idir++)
    {
      if (abs(gradient[idir]) >= max(gradient)) 
        {
          retval.first=idir;
        }
    }
  if (gradient[retval.first] >= 0.&& m_insideRegular)
    {
      retval.second=Side::Lo;
    }
  else  if (gradient[retval.first] < 0.&& m_insideRegular)
    {
      retval.second=Side::Hi;
    }
  else if (gradient[retval.first] >= 0.&&!m_insideRegular)
    {
      retval.second=Side::Hi;
    }
  else  if (gradient[retval.first] < 0.&&!m_insideRegular)
    {
      retval.second=Side::Lo;
    }
  
 return retval;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
Real 
SphereBL::localFuncValue(const RealVect& a_indCoords, 
                         const int& a_upDir,
                         const IntVect& a_iv, 
                         const Box& a_domain, 
                         const RealVect& a_origin, 
                         const Real& a_dx) const

{
  Real retval = 0.;
  for(int idir = 0; idir< SpaceDim; idir++)
    {
      if (idir != a_upDir)
        {
          retval += pow(a_indCoords[idir] - m_center[idir],2);
        }       
    }
  if (a_origin[a_upDir] + a_dx*a_iv[a_upDir] >= m_center[a_upDir])
    {
      retval = sqrt(Abs(pow(m_radius,2) - retval)) + m_center[a_upDir];
    }
  else
    {
      retval = -sqrt(Abs(pow(m_radius,2)-retval)) + m_center[a_upDir];
    }
  return retval;
}
/*****/
/*****/
BaseLevelSet* 
SphereBL::new_baseLevelSet() const       
{
  SphereBL* spherePtr = new SphereBL(m_radius,
                                     m_center,
                                     m_insideRegular,
                                     m_dx);
    
  return static_cast<BaseLevelSet*>(spherePtr); 
}
