/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "RealVect.H"

#include "PolyGeom.H"

#include "TiltedCylinderIF.H"

TiltedCylinderIF::TiltedCylinderIF(const Real&     a_radius,
                                   const RealVect& a_direction,
                                   const RealVect& a_point,
                                   const bool&     a_inside)
{
  // Remember the parameters
  m_radius = a_radius;
  m_direction = a_direction;
  m_point = a_point;
  m_inside = a_inside;

  // Precompute the radius squared
  m_radius2 = m_radius * m_radius;

  // Precompute the length squared of the direction vector
  m_length2 = PolyGeom::dot(m_direction,m_direction);
}

TiltedCylinderIF::TiltedCylinderIF(const TiltedCylinderIF& a_inputIF)
{
  // Remember the parameters
  m_radius = a_inputIF.m_radius;
  m_direction = a_inputIF.m_direction;
  m_point = a_inputIF.m_point;
  m_inside = a_inputIF.m_inside;

  // Precompute the radius squared
  m_radius2 = m_radius * m_radius;

  // Precompute the length squared of the direction vector
  m_length2 = PolyGeom::dot(m_direction,m_direction);
}

TiltedCylinderIF::~TiltedCylinderIF()
{
}

void TiltedCylinderIF::GetParams(Real&     a_radius,
                                 RealVect& a_direction,
                                 RealVect& a_point,
                                 bool&     a_inside) const
{
  // Copy parameter information over
  a_radius = m_radius;
  a_direction = m_direction;
  a_point = m_point;
  a_inside = m_inside;
}

void TiltedCylinderIF::SetParams(const Real&     a_radius,
                                 const RealVect& a_direction,
                                 const RealVect& a_point,
                                 const bool&     a_inside)
{
  // Set parameter information
  m_radius = a_radius;
  m_direction = a_direction;
  m_point = a_point;
  m_inside = a_inside;

  // Precompute the radius squared
  m_radius2 = m_radius * m_radius;

  // Precompute the length squared of the direction vector
  m_length2 = PolyGeom::dot(m_direction,m_direction);
}

Real TiltedCylinderIF::value(const RealVect& a_point) const
{
  Real retval;

  // Get a_point relative to a point on the cylinder's axis, m_point
  RealVect delta;

  delta = a_point;
  delta -= m_point;

  // Get the dot product of the relative vector (above) and the cylinder's
  // axis vector and normalize by the square of the length of the direction
  // vector
  Real dot = PolyGeom::dot(m_direction,delta);
  dot /= m_length2;

  // Find the vector from a_point to the point on the cylinder's axis closest
  // to a_point
  RealVect closest;

  closest = m_direction;
  closest *= dot;
  closest -= delta;

  // Find the length squared of this vector and compare it to the radius
  // squared
  Real length2 = PolyGeom::dot(closest,closest);

  // Return the difference between the sqaures (zero on the cylinder)
  retval = length2 - m_radius2;

  // Change the sign to change inside to outside
  if (!m_inside)
  {
    retval = -retval;
  }

  return retval;
}

BaseImplicitFunction* TiltedCylinderIF::newImplicitFunction() const
{
  TiltedCylinderIF* tiltedCylinderPtr = new TiltedCylinderIF(m_radius,
                                                             m_direction,
                                                             m_point,
                                                             m_inside);

  return static_cast<BaseImplicitFunction*>(tiltedCylinderPtr);
}
