/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <cmath>

#include "RealVect.H"

#include "PolyGeom.H"

#include "TorusIF.H"

TorusIF::TorusIF(const Real&     a_majorRadius,
                 const Real&     a_minorRadius,
                 const RealVect& a_center,
                 const bool&     a_inside)
{
  // Remember parameters
  m_majorRadius = a_majorRadius;
  m_minorRadius = a_minorRadius;
  m_center = a_center;
  m_inside = a_inside;

  // Precompute the minor radius squared
  m_minorRadius2 = m_minorRadius * m_minorRadius;
}

TorusIF::TorusIF(const TorusIF& a_inputIF)
{
  // Remember parameters
  m_majorRadius = a_inputIF.m_majorRadius;
  m_minorRadius = a_inputIF.m_minorRadius;
  m_center = a_inputIF.m_center;
  m_inside = a_inputIF.m_inside;

  // Precompute the minor radius squared
  m_minorRadius2 = m_minorRadius * m_minorRadius;
}

TorusIF::~TorusIF()
{
}

void TorusIF::GetParams(Real&     a_majorRadius,
                        Real&     a_minorRadius,
                        RealVect& a_center,
                        bool&     a_inside) const
{
  // Copy parameter information over
  a_majorRadius = m_majorRadius;
  a_minorRadius = m_minorRadius;
  a_center = m_center;
  a_inside = m_inside;
}

void TorusIF::SetParams(const Real&     a_majorRadius,
                        const Real&     a_minorRadius,
                        const RealVect& a_center,
                        const bool&     a_inside)
{
  // Set parameter information
  m_majorRadius = a_majorRadius;
  m_minorRadius = a_minorRadius;
  m_center = a_center;
  m_inside = a_inside;

  // Precompute the minor radius squared
  m_minorRadius2 = m_minorRadius * m_minorRadius;
}

Real TorusIF::value(const RealVect& a_point) const
{
  Real retval;

  // Get the radius of a_point (excluding the z coordinate in 3D)
  Real radius1;

  radius1 = 0.0;
  for (int idir = 0; idir < 2; idir++)
  {
    Real cur;
    cur = a_point[idir] - m_center[idir];

    radius1 += cur*cur;
  }

  // Move to the center of the minor circle
  radius1 = sqrt(radius1) - m_majorRadius;

  // Get the radius squared in this coordinate system
  Real radius2;

  radius2 = radius1*radius1;
  for (int idir = 2; idir < SpaceDim; idir++)
  {
    Real cur;
    cur = a_point[idir] - m_center[idir];

    radius2 += cur*cur;
  }

  // Return the difference between the squares (zero on the torus)
  retval = radius2 - m_minorRadius2;

  // Change the sign to change inside to outside
  if (!m_inside)
  {
    retval = -retval;
  }

  return retval;
}

BaseImplicitFunction* TorusIF::newImplicitFunction() const
{
  TorusIF* torusPtr = new TorusIF(m_majorRadius,
                                  m_minorRadius,
                                  m_center,
                                  m_inside);

  return static_cast<BaseImplicitFunction*>(torusPtr);
}
