/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _TRANSFORMIF_H_
#define _TRANSFORMIF_H_

#include "MayDay.H"
#include "RealVect.H"
#include "Vector.H"

#include "BaseImplicitFunction.H"

///
/**
    This implicit function is a tranformed version of the implicit function
    passed in the constructor.  Various transformations (in 2D and 3D) are
    allowed including translating, scaling, and rotations.
 */
class TransformIF: public BaseImplicitFunction
{
public:
  ///
  /**
      Constructor specifying the implicit function
   */
  TransformIF(const BaseImplicitFunction& a_impFunc);

  ///
  /**
      Constructor for the factory
   */
  TransformIF(const BaseImplicitFunction& a_impFunc,
              const Real                  a_transform[SpaceDim+1][SpaceDim+1],
              const Real                  a_invTransform[SpaceDim+1][SpaceDim+1]);

  /// Copy constructor
  TransformIF(const TransformIF& a_inputIF);

  /// Destructor
  virtual ~TransformIF();

  ///
  /**
      Translate by a_trans
   */
  void translate(const RealVect& a_trans);

  ///
  /**
      Scale (uniformly) by a_scale
   */
  void scale(const Real& a_scale);

  ///
  /**
      Scale in each direction (idir) by a_scale[idir]
   */
  void scale(const RealVect& a_scale);

  ///
  /**
      In 2D - rotate by a_angle about a_point, a_axis is ignored.
      In 3D - rotate by a_angle about the line in space specified by
      a_point and and a_axis.
   */
  void rotate(const Real&     a_angle,
              const RealVect& a_point = RealVect::Zero,
              const RealVect& a_axis  = BASISREALV(0));

  ///
  /**
      About a_point, Rotate vector a_axis1 to align with vector a_axis2
   */
  void rotate(const RealVect& a_axis1,
              const RealVect& a_axis2,
              const RealVect& a_point = RealVect::Zero);

  ///
  /**
      Return the value of the function at a_point.
   */
  virtual Real value(const RealVect& a_point) const;

  virtual BaseImplicitFunction* newImplicitFunction() const;

protected:
  // transformation (in homogeneous coordinates)
  Real m_transform[SpaceDim+1][SpaceDim+1];

  // inverse transformation (in homogeneous coordinates)
  Real m_invTransform[SpaceDim+1][SpaceDim+1];

  BaseImplicitFunction* m_impFunc; // implicit function to transform

  void vectorMultiply(RealVect&       m_outPoint,
                      const Real      m_intrans[SpaceDim+1][SpaceDim+1],
                      const RealVect& m_inPoint) const;

  void matrixIdentity(Real m_trans[SpaceDim+1][SpaceDim+1]) const;

  void matrixMultiply(Real       m_outTrans[SpaceDim+1][SpaceDim+1],
                      const Real m_intrans1[SpaceDim+1][SpaceDim+1],
                      const Real m_intrans2[SpaceDim+1][SpaceDim+1]) const;

  void matrixTranslate(Real            m_trans[SpaceDim+1][SpaceDim+1],
                       const RealVect& m_translate) const;

  void matrixScale(Real            m_trans[SpaceDim+1][SpaceDim+1],
                   const RealVect& m_scale) const;

private:
  TransformIF()
  {
    MayDay::Error("TransformIF uses strong construction");
  }

  void operator=(const TransformIF& a_inputIF)
  {
    MayDay::Error("TransformIF doesn't allow assignment");
  }
};

#endif
