/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// DatasetClient.H

#ifndef _DATASETCLIENT_H_
#define _DATASETCLIENT_H_

#include "LayoutData.H"
#include "LevelData.H"
#include "FArrayBox.H"
#include "BaseFab.H"

/// global functions for using the DatasetServer program
/** These are C functions that can be called from the debugger.
 *  Note that the array arg. is always a pointer.
 *  The separate DatasetServer program must already by runnig
 *  for these to work correctly.
 */

extern "C"
{
  /// View a single BaseFab\<Real\> or FAB
  /**
     Sends single fab information to DatasetServer port.
     For use inside debuggers.
   */
  bool
  ArrayView(BaseFab<Real>* fab);

  /// View a single BaseFab\<int\>
  /**
     Sends single fab of ints information to DatasetServer port.
     For use inside debuggers.
   */
  bool
  ArrayViewInt(BaseFab<int>* fab);

  /// View a single FAB with custom format and window title
  /**
     Sends a single FAB to the DatasetServer port
     using the given label and given format.
   @param debugFab is the FAB to view
   @param format is the printf format for numerical display ( ex:  "%5.3f" ).
   @param label  is the first part of the window title ( ex: "myfabname" ).
     For use inside debuggers.
   */
  bool
  ArrayViewFabFormatLabel(BaseFab<Real>* debugFab,
                          const char *format,
                          const char *label);

  /// View a LayoutData of FABs
  /**
     Send layoutdata information to the DatasetServer port.
     For use inside debuggers.
   */
  bool
  MultiArrayView(LayoutData<FArrayBox>* layoutdata);

  /// View a LevelData of FABs
  /**
     Send leveldata information to the DatasetServer port.
     This is for parallel debugging and can only be used
     inside the code.  If you use this inside the debugger
     in parallel, it will hang on a leveldata::copyTo call.
   */
  bool
  ArrayViewLDF(LevelData<FArrayBox>* data);

  /// View a LevelData of FABs with custom format and window title
  /**
     Send leveldata information to the DatasetServer port.
     @param data points to the LevelData to view
     @param format is the printf format for numerical display ( ex:  "%5.3f" ).
     @param label  is the first part of the window title ( ex: "myfabname" ).
     This is for parallel debugging and can only be used
     inside the code.  If you use this inside the debugger
     in parallel, it will hang on a leveldata::copyTo call.
   */
  bool
  ArrayViewLDFFormatLabel(LevelData<FArrayBox>* data,
                          const char* format,
                          const char* label);

  /// View a LayoutData of FABs
  /**
      Send layoutdata information to the DatasetServer port.
      For use inside debuggers.
   */
  bool
  MultiArrayViewFab(LayoutData<FArrayBox>* layoutdata);

  /// View a LayoutData of BaseFab\<Real\>s
  /**
     Send LayoutData information to the DatasetServer port.
     For use inside debuggers.
   */
  bool
  MultiArrayViewRealBaseFab(LayoutData<BaseFab<Real> >* layoutdata);

  /// View a LayoutData of BaseFab\<Real\>s with custom format and title
  /**
     Send LayoutData information to the DatasetServer port.
     @param layoutdata points to the LayoutData to view
     @param format is the printf format for numerical display ( ex:  "%5.3f" ).
     @param label  is the first part of the window title ( ex: "myfabname" ).
     For use inside debuggers.
   */
  bool
  MultiArrayViewFormatLabel(LayoutData<BaseFab<Real> >* layoutdata,
                            const char* format,
                            const char* label);

  //
  /*
     don't use this.
   */
  bool
  ArrayViewRealPtrArrayNVarDims(Real *data[], int nvar,
                                const int *lodim, const int *hidim,
                                const char *format, const char *label);
}

//
/*
arrayview_user_port_offset:

allows the user to connect to a different port.  this is meant to
be used when you kill your DatasheetServer and want to statrt
another one immediately (primarily during debugging), since the
socket it opens will not be closed promptly by the system.  on a
multiuser system, you could accidentally connect to someone else's
server by doing this.

usage: set it in user code prior to calling an ArrayView function.
it is not necessary to declare it in user code.
*/
extern int arrayview_user_port_offset;

//
/*
 */
extern bool
CreateSocket(int& newsocket) ;

/*
 */
extern bool
SendRealArray(int sockfd, Real *data[], int nvar,
              const int *lodim, const int *hidim);

//
/*
 */
extern bool
SendString(int sockfd, const char *sendstring);

#endif
