/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Purpose:
//  Test the infrastucture for periodic boundary conditions

// Usage:
//  <program-name> [-q|-v] ...
//
//  where:
//    -q means run quietly (only pass/fail messages printed)
//    -v means run verbosely (all messages printed)
//    -write means write out the boxes; otherwise will read in 
//           boxes from a file and compare with boxes generated
//           in tests.
//    ... all non-option arguments are ignored (Chombo convention)
//
//  Default is `-v'
//
//  Unknown options are treated as errors.


////////////////////////////////////////////////////////////////


// Include files:
#include <cstdio>

#include "parstream.H"
#include "LoadBalance.H"
#include "FArrayBox.H"
#include "LevelData.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"
#include "BRMeshRefine.H"
#include "AMRIO.H"

#if 0
#include "GhostBC.H"
#include "PoissonBC.H"
#endif


#ifdef MPI
#include <mpi.h>
#endif

//////////////////////////////////////////////////////////////
using std::cout;
using std::endl;

void parseTestOptions(int argc, char* argv[]);

void initData(LevelData<FArrayBox>& a_data, const Real a_dx);

/// Global variables for handling output

static const char *pgmname = "testPeriodic";
static const char *indent = "   ";
//static const char *indent2 = "      " ;

static bool verbose = true ;
// if true, use arrayview to display results
static bool display = false;

/// Code:

int
main(int argc ,char *argv[] )
{

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  parseTestOptions(argc, argv);
  
  // establish periodic domain -- first multiply periodic in all directions
  int status = 0;
  int baseDomainSize = 8;
  int numGhost = 2;
  IntVect ghostVect(numGhost*IntVect::Unit);
  
  Box baseDomBox(IntVect::Zero, (baseDomainSize-1)*IntVect::Unit);
  ProblemDomain baseDomain(baseDomBox);

  // set periodic in all directions
  for (int dir=0; dir<SpaceDim; dir++) {
    baseDomain.setPeriodic(dir, true);
  }

#if 0
  // set up physical BC class for partially periodic tests
  DomainGhostBC physBC;
  for (int dir=0; dir<SpaceDim; dir++) {
    SideIterator sit;
    for (sit.begin(); sit.ok(); ++sit) 
      {
        // it should be easy to see if Neumann bcs are doing the right thing
        NeumannBC thisBC(dir,sit());
        physBC.setBoxGhostBC(thisBC);
      }
  }
#endif
  
  {
    // first try a single box on this level
    Vector<Box> levelBoxes(1, baseDomBox);
    Vector<int> procAssign(1, 0);
    DisjointBoxLayout levelGrids(levelBoxes,procAssign, baseDomain);
      
    LevelData<FArrayBox> tester(levelGrids, 1, ghostVect);

    Real dx = 1.0/baseDomainSize;

    initData(tester, dx);

    // test exchange in this case -- should fill ghost cells with
    // periodically copied data
    Interval comps = tester.interval();
    tester.exchange(comps);

  
  }

  // now try refining things, this time look at 2 boxes
  baseDomain.refine(2);
  {  
    const Box domainBox = baseDomain.domainBox();
    Vector<Box> levelBoxes1(1, domainBox);
    Vector<int> procAssign1(1,0);
    DisjointBoxLayout levelGrids1(levelBoxes1, procAssign1, baseDomain);

    LevelData<FArrayBox> tester0(levelGrids1, 1, IntVect::Zero);    
    LevelData<FArrayBox> tester1(levelGrids1, 1, ghostVect);

    Real dx = 1.0/(2.0*baseDomainSize);
    initData(tester1, dx);
    initData(tester0, dx);

    // first test exchange for single box
    Interval comps = tester1.interval();
    tester1.exchange(comps);

 

    // now try multiple boxes
    Vector<Box> levelBoxes2(3);
    Vector<int> procAssign2(3);

    levelBoxes2[0] = Box(IntVect::Zero, 9*IntVect::Unit);
    levelBoxes2[1] = Box(IntVect(D_DECL(7,10,10)), 15*IntVect::Unit);
    // this box should fail the disjointness test
    //levelBoxes2[1] = Box(IntVect(D_DECL(7,10,10)), 17*IntVect::Unit);
    levelBoxes2[2] = Box(IntVect(D_DECL(11,0,0)), IntVect(D_DECL(15,7,7)));

    int loadbalancestatus = LoadBalance(procAssign2, levelBoxes2);
    assert (loadbalancestatus == 0);

    DisjointBoxLayout levelGrids2;
    levelGrids2.define(levelBoxes2, procAssign2, baseDomain);
    LevelData<FArrayBox> tester2(levelGrids2, 1, ghostVect);

    tester0.copyTo(comps, tester2, comps);


  }

  { 
    // now test partially periodic case (also coarsen back to original domain
    baseDomain.coarsen(2);
    baseDomain.setPeriodic(0,false);

    // try similar tests to the second case, only coarsened
    const Box domainBox = baseDomain.domainBox();
    Vector<Box> levelBoxes1(1, domainBox);
    Vector<int> procAssign1(1,0);
    DisjointBoxLayout levelGrids1(levelBoxes1, procAssign1, baseDomain);

    LevelData<FArrayBox> tester0(levelGrids1, 1, IntVect::Zero);    
    LevelData<FArrayBox> tester1(levelGrids1, 1, ghostVect);

    Real dx = 1.0/(baseDomainSize);
    initData(tester1, dx);
    initData(tester0, dx);

    // first test exchange for single box
    Interval comps = tester1.interval();
    tester1.exchange(comps);

#if 0
    // now set physcial bcs
    DataIterator dit1 = tester1.dataIterator();
    for (dit1.begin(); dit1.ok(); ++dit1) 
      {
        FArrayBox& thisData = tester1[dit1()];
        physBC.applyHomogeneousBCs(thisData, baseDomain, dx);
      }
#endif


    
    // now try multiple boxes
    Vector<Box> levelBoxes2(3);
    Vector<int> procAssign2(3);

    levelBoxes2[0] = Box(IntVect::Zero, 4*IntVect::Unit);
    levelBoxes2[1] = Box(IntVect(D_DECL(3,5,5)), 7*IntVect::Unit);
    // this box should fail the disjointness test
    //levelBoxes2[1] = Box(IntVect(D_DECL(3,5,5)), 10*IntVect::Unit);
    levelBoxes2[2] = Box(IntVect(D_DECL(6,0,0)), IntVect(D_DECL(7,1,1)));

    int loadbalancestatus = LoadBalance(procAssign2, levelBoxes2);
    assert (loadbalancestatus == 0);

    DisjointBoxLayout levelGrids2(levelBoxes2, procAssign2, baseDomain);
    LevelData<FArrayBox> tester2(levelGrids2, 1, ghostVect);

    tester0.copyTo(comps, tester2, comps);

#if 0
    DataIterator dit2 = tester2.dataIterator();
    for (dit2.begin(); dit2.ok(); ++dit2) 
      {
        FArrayBox& thisData = tester2[dit2()];
        physBC.applyInhomogeneousBCs(thisData, baseDomain, dx);
      }
#endif
    

  }

  // i'm thinking all pout()'s should be before MPI_Finalize... (ndk)  
  pout() << indent << pgmname << ": " 
	 << ( (status == 0) ? "passed all tests" : "failed at least one test,")
	 << " status = " << status 
	 << endl;
  
#ifdef MPI
  MPI_Finalize();
#endif

  return status ;
}

///
// Parse the standard test options (-v -q) out of the command line.
// Stop parsing when a non-option argument is found.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-d" ,3 ) == 0 )
            {
              display = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-p" ,3 ) == 0 )
            {
              display = false ;
              argv[i] = "" ;
            }
          else
            {
              break ;
            }
        }
    }
  return ;
}

void
initData(LevelData<FArrayBox>& a_data, const Real a_dx)
{

  DataIterator dit = a_data.dataIterator();
  const DisjointBoxLayout& interiorBoxes = a_data.getBoxes();

  for (dit.begin(); dit.ok(); ++dit) 
    {
      // first set to a bogus value which will persist in ghost cells 
      // after initialization
      a_data[dit()].setVal(1.0e9);

      // this will be slow, but who cares?
      FArrayBox& localData = a_data[dit()];
      BoxIterator boxIt(interiorBoxes[dit()]);
      Real localVal;
      for (boxIt.begin(); boxIt.ok(); ++boxIt) 
        {
          const IntVect& loc = boxIt();
          localVal = a_dx*(D_TERM(loc[0]+0.5, 
                                + loc[1]+0.5,
                                + loc[2]+0.5));
          localData(loc) = localVal;
        }

    }
}
