/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
// dtgraves mon oct 29, 2001

#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "VoFIterator.H"
#include "EBArith.H"
#include "AllRegularService.H"
#include "DebugDump.H"
#include "SlabService.H"
#include "EBArrayView.H"
#include "EBCoarseAverage.H"
#include "EBFluxRegister.H"

/*****************/
/*****************/
int fluxRegTest(void);
/******************/
/******************/
int 
makeLayouts(DisjointBoxLayout& a_dblCoar, 
            DisjointBoxLayout& a_dblFine,
            const Box& a_domainCoar,
            const Box& a_domainFine);
/***************/
/***************/
int makeGeometry(Box& domain);
/***************/
/***************/
int makeEBISL(EBISLayout& a_ebisl, 
              const DisjointBoxLayout& a_grids, 
              const Box& a_domain, 
              const int& nghost);
/***************/
/***************/
void 
setDefaults( int& nref, 
             Vector<Box>& coarboxes, 
             Vector<Box>& fineboxes, 
             Box& domc, Box& domf);
/***************/
/***************/


int
main(int argc, char* argv[])
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  int eekflag = 0;
  {//begin forever present scoping trick
    
    eekflag = fluxRegTest();
    if(eekflag == 0)
      cout << "ebfluxregister test passed." << endl;
    else
      cout << "ebfluxregister test FAILED with error code " 
           << eekflag << endl;

  }//end omnipresent scoping trick
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->clear();
#ifdef MPI
  MPI_Finalize();
#endif
  return eekflag;
}

/*****************/
/*****************/

int fluxRegTest()
{
  int nref = 2;
  int eekflag = 0;
  bool verbose = false;

  Box domainCoar, domainFine;
  eekflag = makeGeometry(domainFine);
  if(eekflag != 0)
    return eekflag;
  domainCoar = coarsen(domainFine, nref);

  DisjointBoxLayout dblFine,dblCoar;
  eekflag = makeLayouts(dblCoar, dblFine, domainCoar, domainFine);

  Interval interv(0,0);
  if(verbose)
    {
      pout()  << "dblCoarCell is" << endl;
      dumpDBL(&dblCoar);
      pout()    << "dblFineCell is" << endl;
      dumpDBL(&dblFine);
    }
  EBISLayout ebislFine, ebislCoar;
  int nghost = 3;
  makeEBISL(ebislFine, dblFine, domainFine, nghost);
  makeEBISL(ebislCoar, dblCoar, domainCoar, nghost);

  EBCellFactory factFine(ebislFine); 
  EBCellFactory factCoar(ebislCoar); 
  IntVect ivghost = IntVect::Unit;
 
      
  LevelData<EBCellFAB> fineData(dblFine, 1,ivghost, factFine);
  LevelData<EBCellFAB> coarData(dblCoar, 1,ivghost, factCoar);

  LevelData<EBCellFAB> extraDense(dblCoar, 1,ivghost, factCoar);

  //set data and flux registers to zero
  for(DataIterator coarIt = coarData.dataIterator(); 
      coarIt.ok(); ++coarIt)
      coarData[coarIt()].setVal(0.);
  for(DataIterator fineIt = fineData.dataIterator(); 
      fineIt.ok(); ++fineIt)
    fineData[fineIt()].setVal(0.);

  EBFluxRegister fluxReg(dblFine, 
                         dblCoar, 
                         ebislFine,
                         ebislCoar,
                         domainCoar,
                         nref, 1);

  fluxReg.setToZero();
  //loop through directions
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      if(verbose)
        {
          pout() << "After setting to zero, the fluxReg in dir " << idir << " is" << endl;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              fluxReg.dumpCoar(idir, sit());
              fluxReg.dumpFine(idir, sit());
            }
        }

      //increment and decrement 
      //flux registers with equal size fluxes
      Real scale = 1.0;
      Real fluxVal = 4.77;
      for(DataIterator coarIt = coarData.dataIterator(); 
          coarIt.ok(); ++coarIt)
        {
          const Box&  boxCoar = dblCoar.get(coarIt());
          const EBISBox& ebisBox = ebislCoar[coarIt()];
          IntVectSet ivsBC(boxCoar);
          EBFaceFAB edgeFluxReg(ebisBox, boxCoar, idir, 1);
          BaseIFFAB<Real> edgeFluxIrr(ivsBC, ebisBox.getEBGraph(), idir, 1);
          edgeFluxReg.setVal(fluxVal);
          edgeFluxIrr.setVal(fluxVal);
          fluxReg.incrementCoarseRegular(edgeFluxReg, scale,
                                         coarIt(), interv, idir);
          fluxReg.incrementCoarseIrregular(edgeFluxIrr, scale,
                                           coarIt(), interv, idir);
        }
      if(verbose)
        {
          pout() << "After coarse increment, the fluxReg in dir " << idir << " is" << endl;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              fluxReg.dumpCoar(idir, sit());
              fluxReg.dumpFine(idir, sit());
            }
        }

      for(DataIterator fineIt = fineData.dataIterator(); 
          fineIt.ok(); ++fineIt)
        {
          const Box&  boxFine = dblFine.get(fineIt());
          const EBISBox& ebisBox = ebislFine[fineIt()];
          IntVectSet ivsBF(boxFine);
          EBFaceFAB edgeFluxReg(ebisBox, boxFine, idir, 1);
          BaseIFFAB<Real> edgeFluxIrr(ivsBF, ebisBox.getEBGraph(), idir, 1);
          edgeFluxReg.setVal(fluxVal);
          edgeFluxIrr.setVal(fluxVal);
          
          for(SideIterator sit; sit.ok(); ++sit)
            {
              fluxReg.incrementFineRegular(edgeFluxReg, scale,
                                           fineIt(),  interv, idir, sit());

              fluxReg.incrementFineIrregular(edgeFluxIrr, scale,
                                             fineIt(),  interv, idir, sit());

              if(verbose)
                fluxReg.dumpFine(idir, sit());
            }
        }
      if(verbose)
        {
          pout() << "After fine increment, the fluxReg in dir " << idir << " is" << endl;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              fluxReg.dumpCoar(idir, sit());
              fluxReg.dumpFine(idir, sit());
            }
        }


      //reflux what ought to be zero into zero and the result should be zero
      //except where the coarse-fine boundary gets crossed by the embedded
      //boundary.  That should get fixed by the extramass thing.
      fluxReg.reflux(coarData, interv, scale);
      if(verbose)
        {
          pout() << "After reflux, the fluxReg in dir " << idir << " is " << endl;
          for(SideIterator sit; sit.ok(); ++sit)
            {
              fluxReg.dumpCoar(idir, sit());
              fluxReg.dumpFine(idir, sit());
            }
        }

      for(DataIterator coarIt = coarData.dataIterator(); 
          coarIt.ok(); ++coarIt)
        extraDense[coarIt()].setVal(0.);

      // now add extra density to soltuion
      //in the end the solution should return to zero
      fluxReg.incrementDensityArray(extraDense, interv, scale);
      for(DataIterator coarIt = coarData.dataIterator(); 
          coarIt.ok(); ++coarIt)
        coarData[coarIt()] += extraDense[coarIt()];

      DataIterator datIt = coarData.dataIterator();
      for(datIt.reset(); datIt.ok(); ++datIt)
        {
          const EBCellFAB& data = coarData[datIt()];
          IntVectSet ivsBox(dblCoar.get(datIt()));
          const EBISBox& ebisBox = ebislCoar[datIt()];
          Real rmax = 0.;
          Real rmin = 0.;
          for(VoFIterator vofit(ivsBox, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
            {
              const VolIndex& vof = vofit();
              rmax = Max(rmax, data(vof, 0));
              rmin = Min(rmax, data(vof, 0));
              if((rmax > 1.0e-10)||(rmin > 1.0e-10))
                {
                  pout() << "EBFluxRegister failed the test at idir = "
                         << idir << ", vof = " << vof <<endl;
                  eekflag = 42;
                  return eekflag;
                }
            }
        }

    }      
  return eekflag;
}
/***************/
/***************/
int makeEBISL(EBISLayout& a_ebisl, 
              const DisjointBoxLayout& a_grids, 
              const Box& a_domain, 
              const int& a_nghost)
{
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());
  ebisPtr->fillEBISLayout(a_ebisl, a_grids, a_domain, a_nghost);
  return 0;
}
/***************/
/***************/
int 
makeLayouts(DisjointBoxLayout& a_dblCoar, 
            DisjointBoxLayout& a_dblFine,
            const Box& a_domainCoar,
            const Box& a_domainFine)
{
  //set up mesh refine object
  int eekflag= 0;
  Vector<Box> coarboxes;
  Vector<Box> fineboxes;
  Box domc; 
  Box domf;
  int nref;
  setDefaults( nref,  coarboxes, fineboxes, domc, domf);

  Vector<Box> vboxCoar = coarboxes;
  Vector<Box> vboxFine = fineboxes;

  Vector<int> procAssignFine;
  Vector<int> procAssignCoar;
  eekflag = LoadBalance(procAssignFine,vboxFine);
  if(eekflag != 0) return eekflag;
  eekflag = LoadBalance(procAssignCoar,vboxCoar);
  if(eekflag != 0) return eekflag;

  a_dblFine.define(vboxFine, procAssignFine);
  a_dblCoar.define(vboxCoar, procAssignCoar);
  return eekflag;
}
/**********/
/**********/
void 
setDefaults( int& nref, 
             Vector<Box>& coarboxes, 
             Vector<Box>& fineboxes, 
             Box& domc, Box& domf)
{
  nref = 2;
  int nc = 16;
  int nc16 = nc/2;
  int nc8 = nc/4; 
  int nc24 = 3*nc/4; 
  int nc12 = 3*nc/8; 
  int nc20 = 5*nc/8; 

  IntVect ivclo = IntVect::TheZeroVector();

  fineboxes.resize(0);
#if(CH_SPACEDIM ==2)

  IntVect ivchi(nc-1, nc-1);
  Box boxf1(IntVect(0 ,nc8), IntVect(nc16-1,nc24));
  Box boxf2(IntVect(nc16,nc12), IntVect(nc24-1,nc20-1));
  Box boxf3(IntVect(nc8 ,0 ), IntVect(nc20-1,nc8-1));
  Box boxf4(IntVect(nc20,nc8 ), IntVect(nc-1,nc12-1));

#elif(CH_SPACEDIM==3)
  IntVect ivchi(nc-1, nc-1, nc-1);
  Box boxf1(IntVect(0 ,nc8,0), IntVect(nc16-1,nc24,nc-1));
  Box boxf2(IntVect(nc16,nc12,0), IntVect(nc24-1,nc20-1,nc-1));
  Box boxf3(IntVect(nc8 ,0 ,0), IntVect(nc20-1,nc8-1,nc-1));
  Box boxf4(IntVect(nc20,nc8,0 ), IntVect(nc-1,nc12-1,nc-1));
#else
  error bogus spacedim;
#endif
  boxf1.refine(2);
  boxf2.refine(2);
  boxf3.refine(2);
  boxf4.refine(2);

  domc = Box(ivclo, ivchi);
  int ichop = domc.size(0)/2;
  Box bc1 = domc;
  Box bc2 = bc1.chop(0, ichop);
  Box bc3 = bc2.chop(1, ichop);
  coarboxes.push_back(bc1);
  coarboxes.push_back(bc2);
  coarboxes.push_back(bc3);

  fineboxes.push_back(boxf1);
  fineboxes.push_back(boxf2);
  fineboxes.push_back(boxf3);
  fineboxes.push_back(boxf4);
  domf = refine(domc,nref);
}
int makeGeometry(Box& a_domain)
{
  int eekflag =  0;
  Real dx = 1;
  //parse input file
  RealVect origin = RealVect::Zero;
  int nref;
  Vector<Box> coarboxes;
  Vector<Box> fineboxes;
  Box domc; 
  Box domf;
  setDefaults( nref,  coarboxes, fineboxes, domc, domf);
  a_domain = domf;
               
  //allregular
  pout() << "all regular geometry" << endl;
  AllRegularService regserv;
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(a_domain, origin, dx, regserv);
  return eekflag;
}
/*****************/
/*****************/

