/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//
#include "SphereBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "LevelData.H"
#include "EBCellFAB.H"
#include "EBCellFactory.H"
#include "VoFIterator.H"
#include "EBArith.H"
#include "AllRegularService.H"
#include "EBQuadCFInterp.H"
#include "DebugDump.H"
#include "EBArrayView.H"
/***************/
void compareError(const BaseIVFAB<Real>& a_errorFine,
                  const EBISBox&         a_ebislFine, 
                  const IntVectSet&      a_ivsFine,
                  const BaseIVFAB<Real>& a_errorCoar,
                  const EBISBox&         a_ebislCoar, 
                  const IntVectSet&      a_ivsCoar)
{
  int comp = 0;
  int inorm;
  pout() << endl << " Using max norm." << endl;
  Real coarnorm, finenorm;

  EBArith::irregNorm(coarnorm, a_errorCoar, 
                     a_ivsCoar, a_ebislCoar, 
                     comp, inorm);
  EBArith::irregNorm(finenorm, a_errorFine, 
                     a_ivsFine, a_ebislFine, 
                     comp, inorm);

  pout() << "Coarse Error Norm = " << coarnorm << endl;
  pout() << "Fine   Error Norm = " << finenorm << endl;
  if(Abs(finenorm) > 1.0e-8)
    {
      Real order = log(coarnorm/finenorm)/log(2.0);
      pout() << "Order of scheme = " << order << endl;
    }
}
/***************/
Real exactFunc(const IntVect& a_iv,
               const Real&    a_dx)
{
  Real retval;
  RealVect xloc;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      xloc[idir] = a_dx*(Real(a_iv[idir]) + 0.5);
    }
  Real x = xloc[0];
  Real y = xloc[1];
  
  retval = (1.0 + x*x + y*y + x*x*x + y*y*y);
  //retval = x*x;
  //retval = x;
  return retval;
}
/***************/
void makeEBISL(EBISLayout&              a_ebisl, 
               const DisjointBoxLayout& a_grids, 
               const Box&               a_domain, 
               const int&               a_nghost)
{
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());
  ebisPtr->fillEBISLayout(a_ebisl, a_grids, a_domain, a_nghost);
}
/**********/
void makeDomains(Box&  a_domainFine,
                 Box&  a_domainCoar,
                 Box&  a_domainRC,
                 Real& a_dxFine, 
                 Real& a_dxCoar,
                 Real& a_dxRC)
{
  //parse input file
  ParmParse pp;
  Vector<int> n_cell(SpaceDim);
  Real domainSize;
  pp.getarr("n_cell",n_cell,0,SpaceDim);
  pp.get("domain_length", domainSize);

  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          MayDay::Error();
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domainFine.setSmall(lo);
  a_domainFine.setBig(hi);
  a_domainCoar = coarsen(a_domainFine, 2);
  a_domainRC  =  coarsen(a_domainCoar, 2);
  
  a_dxFine = domainSize/Real(a_domainFine.size(0));
  a_dxCoar = 2.*a_dxFine;
  a_dxRC   = 2.*a_dxCoar;
}
/**********/
void makeGeometry(const Box&  a_domain, 
                  const Real& a_dx)
{
  ParmParse pp;

  int ebMaxCoarsen = 3;
  int ebMaxSize = 2056;
  pout() << "sphere geometry" << endl;
  vector<Real> sphere_center(SpaceDim);
  pp.getarr("sphere_center",sphere_center, 0, SpaceDim);
  RealVect sphereCenter;
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      sphereCenter[idir] = sphere_center[idir];
    }
  Real sphereRadius;
  pp.get("sphere_radius", sphereRadius);
  SphereBL sphere(sphereRadius, sphereCenter, false, a_dx);
  GeometryShop workshop(sphere);
  //this generates the new EBIS
  EBIndexSpace*  ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(a_domain, RealVect::Zero, a_dx, workshop, ebMaxSize, ebMaxCoarsen);
}
/***************/
void makeLayouts(DisjointBoxLayout& a_dblFine,
                 DisjointBoxLayout& a_dblCoar,
                 EBISLayout&        a_ebislFine,
                 EBISLayout&        a_ebislCoar,
                 IntVectSet&        a_ivsFine,
                 const Box&         a_domainFine,
                 const Box&         a_domainCoar)
{
  int nghost = 1;
  Vector<int> procs(1, 0);
  Vector<Box> coarBoxes(1, a_domainCoar);
  a_dblCoar = DisjointBoxLayout(coarBoxes, procs);
  makeEBISL(a_ebislCoar, a_dblCoar, a_domainCoar, nghost);

  IntVectSet ivsIrregCoar;
  for(DataIterator dit = a_dblCoar.dataIterator(); dit.ok(); ++dit)
    {
      ivsIrregCoar |=  a_ebislCoar[dit()].getIrregIVS(a_dblCoar.get(dit()));
    }
  Box minBoxCoar = ivsIrregCoar.minBox();
  minBoxCoar.grow(1);
  minBoxCoar.coarsen(2);
  minBoxCoar.refine(2);
  minBoxCoar &= a_domainCoar;

  Box boxFine = refine(minBoxCoar, 2);
  Vector<Box> fineBoxes(1, boxFine);
  a_dblFine = DisjointBoxLayout(fineBoxes, procs);
  makeEBISL(a_ebislFine, a_dblFine, a_domainFine, nghost);
  
  a_ivsFine = IntVectSet();
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      Box dirBox = boxFine;
      dirBox.grow(idir, 1);
      a_ivsFine |= IntVectSet(dirBox);
    }
  a_ivsFine -= boxFine;
  a_ivsFine &= a_domainFine;
}
/***************/
void getError(BaseIVFAB<Real>&   a_error,
              const EBCellFAB&   a_data,
              const EBISBox&     a_ebisBox, 
              const IntVectSet&  a_ivs, 
              const Box&         a_domain,
              const Real&        a_dx)
{
  for(VoFIterator vofit(a_ivs, a_ebisBox.getEBGraph());
      vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      Real righAns = exactFunc(vof.gridIndex(), a_dx);
      Real calcAns = a_data(vof, 0);
      a_error(vof, 0) = righAns - calcAns;
    }
}
/***************/
void setToExact(LevelData<EBCellFAB>&    a_data,
                const DisjointBoxLayout& a_dbl,
                const EBISLayout&        a_ebisl,
                const Box&               a_domain,
                const Real&              a_dx)
{
  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBGraph& ebgraph =   a_ebisl[dit()].getEBGraph();
      const Box&  grid       = a_dbl.get(dit());
      const IntVectSet ivs(grid);
      a_data[dit()].setVal(0.);
      for(VoFIterator vofit(ivs, ebgraph); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          Real righAns = exactFunc(vof.gridIndex(), a_dx);
          a_data[dit()](vof, 0) = righAns;
        }
    }
}
/***************/
void getData(LevelData<EBCellFAB>&     a_dataFine,
             LevelData<EBCellFAB>&     a_dataCoar,
             const DisjointBoxLayout&  a_dblFine,
             const DisjointBoxLayout&  a_dblCoar,
             const EBISLayout&         a_ebislFine,
             const EBISLayout&         a_ebislCoar,
             const Box&                a_domainFine,
             const Box&                a_domainCoar,
             const Real&               a_dxFine,
             const Real&               a_dxCoar)
{
  int nref = 2;
  int nvar = 1;
  EBCellFactory ebcellfactFine(a_ebislFine);
  EBCellFactory ebcellfactCoar(a_ebislCoar);
  a_dataFine.define(a_dblFine, nvar, IntVect::Unit, ebcellfactFine);
  a_dataCoar.define(a_dblCoar, nvar, IntVect::Unit, ebcellfactCoar);
  
  setToExact(a_dataCoar, a_dblCoar, a_ebislCoar,  a_domainCoar, a_dxCoar);
  setToExact(a_dataFine, a_dblFine, a_ebislFine,  a_domainFine, a_dxFine);
  //interpolate phiC onto phiF
  EBQuadCFInterp interpOp(a_dblFine,    a_dblCoar,
                          a_ebislFine,  a_ebislCoar,
                          a_domainCoar, nref, nvar);

  Interval interv(0,0);
  interpOp.interpolate(a_dataFine, a_dataCoar, interv);
}
/******/
void quadCFITest()
{
  int nghost =1;
  int nvar   =1;
  //make all three domains and dx
  Real dxRC, dxCoar, dxFine;
  Box domainRC, domainCoar, domainFine;

  makeDomains(domainFine, domainCoar, domainRC, 
              dxFine, dxCoar, dxRC);
  
  //need a full domain layout at coar and RC levels,
  //need a partial domain layout and fine and coar
  DisjointBoxLayout baseDBLCoar, topDBLCoar;
  DisjointBoxLayout baseDBLFine, topDBLFine; 
  EBISLayout baseEBISLCoar, topEBISLCoar;
  EBISLayout baseEBISLFine, topEBISLFine; 
  IntVectSet ivsFine, ivsCoar;

  pout() << "making fine geometry description" << endl;
  makeGeometry(domainFine, dxFine);
  makeLayouts(topDBLFine, baseDBLFine, topEBISLFine, baseEBISLFine, ivsFine, domainFine, domainCoar);
  coarsen(topDBLCoar, topDBLFine,  2);
  coarsen(baseDBLCoar,baseDBLFine, 2);
  ivsCoar = coarsen(ivsFine, 2);
  makeEBISL(topEBISLCoar,  topDBLCoar,  domainCoar, nghost);
  makeEBISL(baseEBISLCoar, baseDBLCoar, domainRC,   nghost);

  LevelData<EBCellFAB> baseDataFine, topDataFine;
  LevelData<EBCellFAB> baseDataCoar, topDataCoar;

  getData(topDataFine, baseDataFine, topDBLFine, baseDBLFine, topEBISLFine, baseEBISLFine, domainFine, domainCoar, dxFine, dxCoar);
  getData(topDataCoar, baseDataCoar, topDBLCoar, baseDBLCoar, topEBISLCoar, baseEBISLCoar, domainCoar, domainRC,   dxCoar, dxRC);
  for(DataIterator dit = topDBLFine.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBoxCoar = topEBISLCoar[dit()];
      const EBISBox& ebisBoxFine = topEBISLFine[dit()];
      BaseIVFAB<Real> errorCoar(ivsCoar, ebisBoxCoar.getEBGraph(), nvar);
      BaseIVFAB<Real> errorFine(ivsFine, ebisBoxFine.getEBGraph(), nvar);
      const EBCellFAB& dataFABFine = topDataFine[dit()];
      const EBCellFAB& dataFABCoar = topDataCoar[dit()];
      getError(errorFine, dataFABFine, ebisBoxFine, ivsFine, domainFine, dxFine);
      getError(errorCoar, dataFABCoar, ebisBoxCoar, ivsCoar, domainCoar, dxCoar);

      compareError(errorFine, ebisBoxFine, ivsFine, 
                   errorCoar, ebisBoxCoar, ivsCoar);
    }
}
/***************/
int
main(int argc, char** argv)

{

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  int eekflag = 0;
  {//begin forever present scoping trick
    const char* in_file = "sphere.inputs";
    //parse input file
    ParmParse pp(0,NULL,NULL,in_file);

    quadCFITest();
  }//end omnipresent scoping trick
#ifdef MPI
  MPI_Finalize();
#endif
  return eekflag;
}
/***************/
