
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cout;
using std::endl;

#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "CH_HDF5.H"
#include "parstream.H"
#include "BoxIterator.H"
#include "FABView.H"

#include "NewPoissonOp.H"
#include "AMRPoissonOp.H"
#include "BCFunc.H"

/// Global variables for handling output:
static const char* pgmname = "testAMRPoissonOp" ;
static const char* indent = "   ";
static const char* indent2 = "      " ;
static bool verbose = true ;

///
// Parse the standard test options (-v -q) out of the command line.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
        }
    }
  return ;
}

int
testPoissonOp();

int
testAMRPoissonOp();

int
main(int argc ,char* argv[])
{
#ifdef MPI
  MPI_Init (&argc, &argv);
#endif
  parseTestOptions( argc ,argv ) ;
  if( verbose )
    pout () << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int status = testPoissonOp();

  if( status == 0 )
    cout << indent << "PoissonOp" << " passed." << endl ;
  else
    cout << indent << "PoissonOp" << " failed with return code " << status << endl ;

  status = testAMRPoissonOp();

  if( status == 0 )
    cout << indent << pgmname << " passed." << endl ;
  else
    cout << indent << pgmname << " failed with return code " << status << endl ;
#ifdef MPI
  MPI_Finalize ();
#endif
  return status ;
}


//   u = x*x + y*y + z*z
//   du/dx = 2*x
//   du/dy = 2*y
//   du/dz = 2*z
//   Laplace(u) = 2*CH_SPACEDIM


static Box domain = Box(-64*IntVect::TheUnitVector(), 63*IntVect::TheUnitVector());
static Real dx = 0.0125;
static int blockingFactor = 8;

extern "C" {
  void Parabola_neum(Real* pos, 
                     int* dir, 
                     int* side, 
                     Real* a_values,
                     int*  a_numValues)
  {
    switch (*dir) {
    case 0:
      a_values[0]=2*(*side)*pos[0];
      return;
    case 1:
      a_values[0]=2*(*side)*pos[1];
    return;
    case 2:
      a_values[0]=2*(*side)*pos[2];
      return;
    default:
      MayDay::Error("no such dimension");
    };
  } 
  
  void Parabola_diri(Real* pos, 
                     int* dir, 
                     int* side, 
                     Real* a_values,
                     int*  a_numValues)
  {
    a_values[0] = D_TERM(pos[0]*pos[0],+pos[1]*pos[1],+pos[2]*pos[2]);
  }



  void DirParabolaBC(FArrayBox& a_state, 
                     const ProblemDomain& a_domain,
                     Real a_dx,
                     bool a_homogeneous)
  {
    Box valid = a_state.box();
    valid.grow(-1);
    for(int i=0; i<CH_SPACEDIM; ++i)
      {
        
        DiriBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_diri,
               i,
               Side::Lo);
        DiriBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_diri,
               i,
               Side::Hi);
      }
  }

  void NeumParabolaBC(FArrayBox& a_state, 
                      const ProblemDomain& a_domain,
                      Real a_dx,
                      bool a_homogeneous)
  {
    Box valid = a_state.box();
    valid.grow(-1);
    for(int i=0; i<CH_SPACEDIM; ++i)
      {
        
        NeumBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_neum,
               i,
               Side::Lo);
        NeumBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_neum,
               i,
               Side::Hi);
      }
  }
}

static BCValueFunc pointFunc = Parabola_diri;

void parabola(const Box& box, int comps, FArrayBox& t)
{
  RealVect pos;
  int side, dir;
  int num = 1;
  ForAllXBNN(Real,t, box, 0, comps)
    { 
      num=nR;
      D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
      pointFunc(&(pos[0]), &dir, &side, &tR, &num);
    }EndFor;
}

void makeGrids(DisjointBoxLayout& a_dbl, const Box& a_domain)
{
  
  BRMeshRefine br;
  Box domain = a_domain;
  domain.coarsen(blockingFactor);
  domain.refine(blockingFactor);
  assert(domain == a_domain);
  domain.coarsen(blockingFactor);

  ProblemDomain junk(domain);
  IntVectSet pnd(domain);
  IntVectSet tags;
  for(BoxIterator bit(domain); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();
      if(D_TERM(true, && iv[1]< 2*iv[0] && iv[1]>iv[0]/2, && iv[2] < domain.bigEnd(2)/2))
        {
          tags|=iv;
        }
    }
  Vector<Box> boxes;
  br.makeBoxes(boxes, tags, pnd, junk, 32/blockingFactor);
  Vector<int> procs;
  LoadBalance(procs, boxes);
  for(int i=0; i<boxes.size(); ++i) boxes[i].refine(blockingFactor);
  a_dbl.define(boxes, procs);
}


int
testAMRPoissonOp()
{
#ifdef HDF5
  writeLevel(NULL);
#endif
  ProblemDomain regularDomain(domain);

  DisjointBoxLayout  dbl;

  makeGrids(dbl, domain);
  // make interesting grids here.

  dbl.close();
  
  DataIterator dit(dbl);
  LevelData<FArrayBox> phi(dbl, 1, IntVect::TheUnitVector());
  LevelData<FArrayBox> rhs(dbl, 1);
  LevelData<FArrayBox> lofphi(dbl, 1);

  struct setvalue{
    static void setFunc(const Box& box, 
                        int comps, FArrayBox& t) {t.setVal(4.0);}
  };

  rhs.apply(setvalue::setFunc);
  phi.apply(parabola);
  
  RealVect pos(IntVect::TheUnitVector());
  pos*=dx;



  AMRPoissonOp amrop;

  amrop.define(dbl, pos, regularDomain, DirParabolaBC);

  amrop.applyOp(lofphi, phi);

  for(dit.begin(); dit.ok(); ++dit)
    {
      lofphi[dit]-=2*CH_SPACEDIM;
      Real norm = lofphi[dit].norm(0);
      if(norm > 1E-6)
        {
          pout()<< "applyOp for parabolic dirichlet function seems busted"
                <<std::endl;
          abort();
        }
    }

  amrop.define(dbl, pos, regularDomain, NeumParabolaBC);

  amrop.applyOp(lofphi, phi);

  for(dit.begin(); dit.ok(); ++dit)
    {
      lofphi[dit]-=2*CH_SPACEDIM;
      Real norm = lofphi[dit].norm(0);
      if(norm > 1E-6)
        {
          pout()<< "applyOp for parabolic neumann function seems busted"
                <<std::endl;
          abort();
        }
    }
  
  
  return 0; 
}

int
testPoissonOp()
{
#ifdef HDF5
  writeLevel(NULL);
#endif
  ProblemDomain regularDomain(domain);

  Box phiBox = domain;
  phiBox.grow(1);

  FArrayBox phi(phiBox, 1);
  FArrayBox rhs(domain, 1);
  FArrayBox lofphi(domain, 1);

 

  rhs.setVal(2*CH_SPACEDIM);
  parabola(domain, 1, phi);
  
  RealVect pos(IntVect::TheUnitVector());
  pos*=dx;



  NewPoissonOp op;

  op.define(pos, regularDomain, DirParabolaBC);

  op.applyOp(lofphi, phi);

  lofphi-=2*CH_SPACEDIM;
  Real norm = lofphi.norm(0);
  if(norm > 1E-6)
    {
      pout()<< "applyOp for parabolic dirichlet function seems busted"
            <<std::endl;
      abort();
    }
  

  op.define(pos, regularDomain, NeumParabolaBC);

  op.applyOp(lofphi, phi);

  lofphi-=2*CH_SPACEDIM;
  norm = lofphi.norm(0);
  if(norm > 1E-6)
    {
      pout()<< "applyOp for parabolic neumann function seems busted"
            <<std::endl;
      abort();
    }
  
  
  return 0; 
}
