
/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include <iostream>
using std::cout;
using std::endl;

#include "BRMeshRefine.H"
#include "CH_HDF5.H"
#include "parstream.H"
#include "BoxIterator.H"
#include "FABView.H"

#include "NewPoissonOp.H"
#include "AMRPoissonOp.H"
#include "BCFunc.H"
#include "BiCGStabSolver.H"
#include "LoadBalance.H"


/// Global variables for handling output:
static const char* pgmname = "testMultiGrid" ;
static const char* indent = "   ";
static const char* indent2 = "      " ;
static bool verbose = true ;

///
// Parse the standard test options (-v -q) out of the command line.
///
void
parseTestOptions( int argc ,char* argv[] ) 
{
  for( int i = 1 ; i < argc ; ++i )
    {
      if( argv[i][0] == '-' ) //if it is an option
        {
          // compare 3 chars to differentiate -x from -xx
          if( strncmp( argv[i] ,"-v" ,3 ) == 0 )
            {
              verbose = true ;
              argv[i] = "" ;
            }
          else if( strncmp( argv[i] ,"-q" ,3 ) == 0 )
            {
              verbose = false ;
              argv[i] = "" ;
            }
        }
    }
  return ;
}

//   u = x*x + y*y + z*z
//   du/dx = 2*x
//   du/dy = 2*y
//   du/dz = 2*z
//   Laplace(u) = 2*CH_SPACEDIM
#ifndef __USE_GNU
#define __USE_GNU
#endif
#include <fenv.h>
#undef  __USE_GNU


static Box domain = Box(-64*IntVect::TheUnitVector(), 63*IntVect::TheUnitVector());
static Real dx = 0.0125;
static int blockingFactor = 8;
static Real xshift = 0.0;


int
testMultiGrid();

int
main(int argc ,char* argv[])
{
#ifdef MPI
  MPI_Init (&argc, &argv);
#endif 
  //int except =  FE_INVALID | FE_UNDERFLOW ;
  int except =  FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW ;
  feenableexcept(except);

  parseTestOptions( argc ,argv ) ;
  if( verbose )
    pout () << indent2 << "Beginning " << pgmname << " ..." << endl ;

  int status = testMultiGrid();

  if( status == 0 )
    cout << indent << pgmname << " passed." << endl ;
  else
    cout << indent << pgmname << " failed with return code " << status << endl ;

  xshift = 0.2;
  blockingFactor = 4;
  status = testMultiGrid();

  if( status == 0 )
    cout << indent << pgmname << " passed." << endl ;
  else
    cout << indent << pgmname << " failed with return code " << status << endl ;

#ifdef MPI
  MPI_Finalize ();
#endif
  return status ;
}



extern "C" {
  void Parabola_neum(Real* pos, 
                     int* dir, 
                     int* side, 
                     Real* a_values,
                     int*  a_numValues)
  {
    switch (*dir) {
    case 0:
      a_values[0]=2*(*side)*(pos[0]-xshift);
      return;
    case 1:
      a_values[0]=2*(*side)*pos[1];
    return;
    case 2:
      a_values[0]=2*(*side)*pos[2];
      return;
    default:
      MayDay::Error("no such dimension");
    };
  } 
  
  void Parabola_diri(Real* pos, 
                     int* dir, 
                     int* side, 
                     Real* a_values,
                     int*  a_numValues)
  {
    a_values[0] = D_TERM((pos[0]-xshift)*(pos[0]-xshift),+pos[1]*pos[1],+pos[2]*pos[2]);
  }



  void DirParabolaBC(FArrayBox& a_state, 
                     const ProblemDomain& a_domain,
                     Real a_dx,
                     bool a_homogeneous)
  {

    for(int i=0; i<CH_SPACEDIM; ++i)
      {
        Box valid = a_state.box();
        valid.grow(i, -1);
        DiriBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_diri,
               i,
               Side::Lo);
        DiriBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_diri,
               i,
               Side::Hi);
      }
  }

  void NeumParabolaBC(FArrayBox& a_state, 
                      const ProblemDomain& a_domain,
                      Real a_dx,
                      bool a_homogeneous)
  {

    for(int i=0; i<CH_SPACEDIM; ++i)
      {
        Box valid = a_state.box();
        valid.grow(i, -1);
        NeumBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_neum,
               i,
               Side::Lo);
        NeumBC(a_state,
               valid,
               dx,
               a_homogeneous,
               Parabola_neum,
               i,
               Side::Hi);
      }
  }
}

static BCValueFunc pointFunc = Parabola_diri;

void parabola(const Box& box, int comps, FArrayBox& t)
{
  RealVect pos;
  int side, dir;
  int num = 1;
  ForAllXBNN(Real,t, box, 0, comps)
    { 
      num=nR;
      D_TERM(pos[0]=dx*(iR+0.5);, pos[1]=dx*(jR+0.5);, pos[2]=dx*(kR+0.5));
      pointFunc(&(pos[0]), &dir, &side, &tR, &num);
    }EndFor;
}

void makeGrids(DisjointBoxLayout& a_dbl, const Box& a_domain)
{
  
  BRMeshRefine br;
  Box domain = a_domain;
  domain.coarsen(blockingFactor);
  domain.refine(blockingFactor);
  assert(domain == a_domain);
  domain.coarsen(blockingFactor);

  ProblemDomain junk(domain);
  IntVectSet pnd(domain);
  IntVectSet tags;
  for(BoxIterator bit(domain); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();
      if(D_TERM(true, && iv[1]< 2*iv[0] && iv[1]>iv[0]/2, && iv[2] < domain.bigEnd(2)/2))
        {
          tags|=iv;
        }
    }
  Vector<Box> boxes;
  br.makeBoxes(boxes, tags, pnd, junk, 32/blockingFactor);
  Vector<int> procs;
  LoadBalance(procs, boxes);
  for(int i=0; i<boxes.size(); ++i) boxes[i].refine(blockingFactor);
  a_dbl.define(boxes, procs);
}

struct setvalue{
  static Real val;
  static void setFunc(const Box& box, 
                      int comps, FArrayBox& t) {t.setVal(val);}
};

Real setvalue::val = 0;

int
testMultiGrid()
{

  ProblemDomain regularDomain(domain);

  pout()<<"\n level solver \n";
  //  Level solve
  {
    DisjointBoxLayout  dbl;

    makeGrids(dbl, domain);
  
    dbl.close();
  
    DataIterator dit(dbl);
    LevelData<FArrayBox> phi(dbl, 1, IntVect::TheUnitVector());
    LevelData<FArrayBox> correction(dbl, 1, IntVect::TheUnitVector());
    LevelData<FArrayBox> phi_exact(dbl, 1);
    LevelData<FArrayBox> error(dbl, 1);
    LevelData<FArrayBox> rhs(dbl, 1);
    LevelData<FArrayBox> residual(dbl, 1);



    setvalue::val = 2*CH_SPACEDIM;
    rhs.apply(setvalue::setFunc);
    setvalue::val = 0;
    phi.apply(setvalue::setFunc);
    phi_exact.apply(parabola);

  
    RealVect pos(IntVect::TheUnitVector());
    pos*=dx;

    AMRPoissonOpFactory opFactory;
 
    opFactory.define(regularDomain, dbl, pos, DirParabolaBC);
    AMRMultiGrid<LevelData<FArrayBox> > solver;
    BiCGStabSolver<LevelData<FArrayBox> > bottomSolver;

    MGLevelOp<LevelData<FArrayBox> >* op = opFactory.MGnewOp(regularDomain,0);

    solver.define(regularDomain, opFactory, bottomSolver, 3, 0);
 

    int iter = 1;
    op->scale(phi, 0.0);
    op->axby(error, phi, phi_exact, 1, -1);
    op->residual(residual, phi, rhs, false);
    Real rnorm = op->norm(residual, 2);
    Real enorm = op->norm(error, 0);

    pout()<< "homogeneous solver mode : solver.solve(correction, residual)\n";
         
    pout()<<"\nInitial residual norm "<<rnorm<<" Error max norm "<<enorm<<"\n\n";

    for(int i=0; i<iter; ++i)
      {
        op->scale(correction, 0.0);
        solver.solve(correction, residual);
        op->incr(phi, correction, 1.0);
        op->axby(error, phi, phi_exact, 1, -1);
        op->residual(residual, phi, rhs, false);
        rnorm = op->norm(residual, 2);
        enorm = op->norm(error, 0);
        pout()<<indent<<"residual norm "<<rnorm<<"   Error max norm = "<<enorm<<std::endl;
      }

    solver.reclaim(opFactory);
    delete op;
  }
 

  return 0; 
}
