/*
   __  ______________              __      
  /  |/  / __/ ___/ /  ___  __ _  / /  ___ 
 / /|_/ / _// /__/ _ \/ _ \/  ' \/ _ \/ _ \
/_/  /_/_/  \___/_//_/\___/_/_/_/_.__/\___/
                                           
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 

// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "SphereBL.H"
#include "MFIndexSpace.H"
#include "MFCellFAB.H"
#include "MFRemapper.H"
#include "MFAliasFactory.H"

#include "EBISLayout.H"
#include "EBCellFactory.H"
#include "EBFluxFactory.H"
#include "EBAMRIO.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "EBArith.H"
#include "EBPatchGodunovF_F.H"
#include "PolyGeom.H"
#include "DebugDump.H"

Real pi = 2*atan2((Real)1,(Real)0);

int readGeometryInfo(Box& a_domain, 
                     Real& a_dx,
                     RealVect& a_origin,
                     RealVect& a_center,
                     Real& a_radius);


void makeHierarchy(Vector<DisjointBoxLayout>& dbl,
                   const ProblemDomain& baseDomain,
                   const IntVectSet&    baseTags);

void swapvector(Vector<LevelData<MFCellFAB>* >& left,
                Vector<LevelData<MFCellFAB>* >& right);

void
compareError(const LevelData<EBCellFAB>& a_errorFine,
             const EBISLayout& a_ebislFine, 
             const DisjointBoxLayout& a_gridsFine, 
             const Box& a_domainFine,
             const LevelData<EBCellFAB>& a_errorMed,
             const EBISLayout& a_ebislMed, 
             const DisjointBoxLayout& a_gridMed, 
             const Box& a_domainMed,
             const LevelData<EBCellFAB>& a_errorCoar,
             const EBISLayout& a_ebislCoar, 
             const DisjointBoxLayout& a_gridsCoar, 
             const Box& a_domainCoar);

void 
getError(LevelData<EBCellFAB>&       a_error,
         const EBISLayout&           a_ebisl,
         const DisjointBoxLayout&    a_dbl,
         const Real&                 a_dx);
                   
/***************/
// define a sphere EBIS.
/***************/
int makeGeometry(MFIndexSpace& mfIndexSpace,
                 const Box& domain, 
                 const Real& dx,
                 const RealVect& origin,
                 const RealVect& center,
                 const Real& radius);
/***************/
//make the corresponding layout
/***************/
int makeEBISL(EBISLayout& a_ebisl, 
              const DisjointBoxLayout& a_grids, 
              const Box& a_domain, 
              const int& nghost);
/***************/
/***************/
/*
int checkEBISL(const EBISLayout& a_ebisl, 
               const DisjointBoxLayout& a_grids, 
               const Box& a_domain, 
               const Real& dx,
               const RealVect& origin,
               const Real& radius,
               const RealVect& center,
               const bool& insideRegular);
*/

/***************/
/***************/
void dumpmemoryatexit();
/***************/
/***************/

char iter_str[80];

int 
main(int argc,char **argv)
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  // begin forever present scoping trick
  {


    const char* in_file = "sphere.inputs";
    //parse input file
    ParmParse pp(0,NULL,NULL,in_file);
    RealVect center;
    Real radius;
    RealVect origin;
    Real dx;
    Box domain;
    int eekflag = 0;
    MFIndexSpace a, b;
    MFIndexSpace *mfIndexSpace, *oldMFIndexSpace;
    mfIndexSpace = &a;
    oldMFIndexSpace = &b;
 
    LevelData<MFCellFAB> fine, med, coarse;
    Vector<LevelData<MFCellFAB>* > mfvector(3,NULL);
    mfvector[0]=&coarse;
    mfvector[1]=&med;
    mfvector[2]=&fine;

    readGeometryInfo(domain, 
                     dx,
                     origin,
                     center,
                     radius);

    Box domainFine(domain), domainMedi, domainCoar;
    ProblemDomain pFine(domain);
    Real dxFine(dx), dxMedi, dxCoar;

    assert(eekflag == 0);

    domainMedi = coarsen(domainFine, 2);
    domainCoar = coarsen(domainMedi, 2);
    dxMedi = 2.0*dxFine;
    dxCoar = 2.0*dxMedi;
    Vector<DisjointBoxLayout> grids(3);
    ProblemDomain baseDomain(domainCoar);
    ProblemDomain pMed(domainMedi);

    int maxsize = 16;
    Vector<Box> boxCoarse;
    domainSplit(domainCoar, boxCoarse, maxsize);
 
    Vector<int> procs(boxCoarse.size(), 0);
    LoadBalance(procs, boxCoarse);
    grids[0] = DisjointBoxLayout(boxCoarse, procs);
    refine(grids[1], grids[0], 2);
    refine(grids[2], grids[1], 2);

    //make data holders
    Vector<int> comps(2,1);
    int ghost = 1;

    int steps= 15;
    int step = 0;

    Vector<LevelData<EBCellFAB>* > phase(3);
    for(int i=0; i<3; i++) phase[i] = new LevelData<EBCellFAB>();

    while(step < steps){
      
      eekflag = makeGeometry(*mfIndexSpace,
                             domain,
                             dx,
                             origin,
                             center,
                             radius);
      if (eekflag != 0)
        {
          pout() << "non zero eek detected = " << eekflag << endl;
          // MayDay::Error("problem in makeGeometry");
        }

      MFCellFactory fineFactory(*mfIndexSpace, grids[2],
                                domainFine,
                                comps, ghost);
      MFCellFactory medFactory(*mfIndexSpace, grids[1], 
                               domainMedi,
                               comps, ghost);
      MFCellFactory coarseFactory(*mfIndexSpace, grids[0], 
                                  domainCoar,
                                  comps, ghost);
      
      mfvector[2]->define(grids[2], 1, IntVect::Unit, 
                          fineFactory);

      mfvector[1]->define(grids[1], 1, IntVect::Unit, 
                          medFactory); 
          
      mfvector[0]->define(grids[0], 1, IntVect::Unit, 
                          coarseFactory); 
     
    
          //initialize data holders
      LevelData<EBCellFAB> phase0_fine, phase1_fine;
      LevelData<EBCellFAB> phase0_med, phase1_med;
      LevelData<EBCellFAB> phase0_coarse, phase1_coarse;
      aliasMF(phase0_fine,   0, *mfvector[2]);
      aliasMF(phase1_fine,   1, *mfvector[2]);
      aliasMF(phase0_med ,   0, *mfvector[1]);
      aliasMF(phase1_med ,   1, *mfvector[1]);
      aliasMF(phase0_coarse, 0, *mfvector[0]);
      aliasMF(phase1_coarse, 1, *mfvector[0]);
 
      getError(phase0_fine, fineFactory.getEBISLayout(0),    grids[2], dxFine);
      getError(phase0_med , medFactory.getEBISLayout(0),     grids[1], dxMedi);
      getError(phase0_coarse , coarseFactory.getEBISLayout(0),  grids[0], dxCoar);
      
      
      compareError(phase0_fine, fineFactory.getEBISLayout(0), grids[2], domainFine,
                   phase0_med , medFactory.getEBISLayout(0),  grids[1], domainMedi,
                phase0_coarse , coarseFactory.getEBISLayout(0),  grids[0], domainCoar);

      getError(phase1_fine, fineFactory.getEBISLayout(1),    grids[2], dxFine);
      getError(phase1_med , medFactory.getEBISLayout(1),     grids[1], dxMedi);
      getError(phase1_coarse , coarseFactory.getEBISLayout(1),  grids[0], dxCoar);
      
      
      compareError(phase1_fine, fineFactory.getEBISLayout(1), grids[2], domainFine,
                   phase1_med , medFactory.getEBISLayout(1),  grids[1], domainMedi,
                phase1_coarse , coarseFactory.getEBISLayout(1),  grids[0], domainCoar);

      step++;
   
      center[0]+= dx/2.0;
      center[1]+= dx/2.0;
      radius -= dx/4.0;
      {
        MFIndexSpace* swap = oldMFIndexSpace;
        oldMFIndexSpace = mfIndexSpace;
        mfIndexSpace = swap;
      }
      pout()<<step<<std::endl;
        }
    
      pout() << "\n levelDivTest test passed" << endl;
    } // end scoping trick

    
#ifdef MPI
    MPI_Finalize();
#endif
    
    return 0;
}


void swapvector(Vector<LevelData<MFCellFAB>* >& left,
                Vector<LevelData<MFCellFAB>* >& right)

{

  LevelData<MFCellFAB>* swap;
  for(int i=0; i<left.size(); ++i)
    {
      swap = left[i]; 
      left[i]=right[i];
      right[i]=swap;
    }
}

int readGeometryInfo(Box& a_domain, 
                     Real& a_dx,
                     RealVect& a_origin,
                     RealVect& a_center,
                     Real& a_radius)
{

  // parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);

  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;

  for (int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if (n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          return(-1);
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain.setSmall(lo);
  a_domain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim, 1.0);
  Real prob_hi;

  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.get("prob_hi",prob_hi);

  a_dx = (prob_hi-prob_lo[0])/n_cell[0];
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      a_origin[idir] = prob_lo[idir];
    }

  pp.get("radius",a_radius);

  // ParmParse doesn't get RealVects, so work-around with Vector<Real>
  Vector<Real> vectorCenter;
  pp.getarr("center",vectorCenter,0,SpaceDim);
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      a_center[idir] = vectorCenter[idir];
    }

  return 0;
} // end read from file

int makeGeometry(MFIndexSpace& mfIndexSpace, 
                 const Box& a_domain, 
                 const Real& a_dx,
                 const RealVect& a_origin,
                 const RealVect& a_center,
                 const Real& a_radius)
{

  int eekflag = 0;
  Vector<GeometryService*> geometry(2, NULL);
  

  SphereBL insidesphere(a_radius,a_center,true,a_dx);
  GeometryShop workshop0(insidesphere);
  workshop0.m_phase = 0;
  geometry[0] = &workshop0;
  
  SphereBL outsidesphere(a_radius,a_center,false,a_dx);
  GeometryShop workshop1(outsidesphere);
  workshop1.m_phase=1;
  geometry[1] = &workshop1;

  mfIndexSpace.define(a_domain,a_origin,a_dx,geometry);
  
  return eekflag;
}




void makeHierarchy(Vector<DisjointBoxLayout>& dbl,
                   const ProblemDomain& baseDomain,
                   const IntVectSet&    baseTags)
{
  
  dbl.resize(3);
  Vector<int> refRatios(2,2);
  Real fillRatio = 0.85;
  int blockingFactor= 4;
  int bufferSize =   2;
  int maxSize    =   32;
  BRMeshRefine regridder(baseDomain, refRatios, fillRatio, blockingFactor,
                         bufferSize, maxSize);

  Vector<Vector<Box> > oldGrids(3,1), newGrids(3);
  oldGrids[0][0]=baseDomain.domainBox();
  oldGrids[1][0]=coarsen(oldGrids[0][0], 2);
  oldGrids[2][0]=coarsen(oldGrids[1][0], 2);

  regridder.regrid(newGrids, baseTags, 0, 1, oldGrids);

  Vector<int> procs;
  for(int i=0; i<3; i++){
    LoadBalance(procs, newGrids[i]);
    dbl[i] = DisjointBoxLayout(newGrids[i], procs);
  }

}
/************/
Real
exactDivergence(const RealVect& a_xval)
{
  Real retval;
  Real x = a_xval[0];
  Real y = a_xval[1];
  retval = 3*x*x + pi*cos(y*pi);

  return retval;
}
/************/
RealVect 
exactFlux(const RealVect& a_xval)
{
  RealVect retval = RealVect::Zero;
  Real x = a_xval[0];
  Real y = a_xval[1];
  retval[0] = x*x*x;
  retval[1] = sin(y*pi);

  return retval;
}
/************/
void
setToExactDivF(EBCellFAB&     a_exactDivF, 
               const EBISBox& a_ebisBox, 
               const Box&     a_region,
               const Real&    a_dx)
{
  a_exactDivF.setVal(0.);
  IntVectSet ivsregion(a_region);
  for(VoFIterator vofit(ivsregion, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      RealVect xval;
      IntVect iv = vof.gridIndex();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
        }
      Real solnrv = exactDivergence(xval);
      Real kappa = a_ebisBox.volFrac(vof);
      a_exactDivF(vof,0) = kappa*solnrv;
    }
}
/************/
void
setToExactFlux(EBCellFAB&             a_flux,
               const EBISBox&         a_ebisBox, 
               const Box&             a_region,
               const Real&            a_dx)
{
  IntVectSet ivsregion(a_region);
  for(VoFIterator vofit(ivsregion, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      RealVect xval;
      IntVect iv = vof.gridIndex();
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
        }
      RealVect fluxrv = exactFlux(xval);
      for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
        {
          a_flux(vof,faceDir) = fluxrv[faceDir];
        }
    } //end loop over face directions
}
/************/
void
kappaDivergence(EBCellFAB&             a_divF, 
                const EBFluxFAB&       a_flux, 
                const EBISBox&         a_ebisBox,
                const Box&             a_box,
                const Real&            a_dx)
{
  //set the divergence initially to zero
  //then loop through directions and increment the divergence 
  //with each directions flux difference.
  a_divF.setVal(0.0);
  BaseFab<Real>&       regDivF = a_divF.getSingleValuedFAB();
  regDivF.setVal(0.);
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      //update for the regular vofs in the nonconservative 
      //case  works for all single valued vofs.
      /* do the regular vofs */
      /**/
      const EBFaceFAB& fluxDir = a_flux[idir];
      const BaseFab<Real>& regFluxDir = fluxDir.getSingleValuedFAB();
      int ncons = 1;
      FORT_DIVERGEF( CHF_BOX(a_box), 
                     CHF_FRA(regDivF), 
                     CHF_CONST_FRA(regFluxDir), 
                     CHF_CONST_INT(idir), 
                     CHF_CONST_INT(ncons), 
                     CHF_CONST_REAL(a_dx));
      /**/
    }
  //update the irregular vofs using conservative diff
  IntVectSet ivsIrreg = a_ebisBox.getIrregIVS(a_box);
  for(VoFIterator vofit(ivsIrreg, a_ebisBox.getEBGraph()); vofit.ok(); ++vofit)
    {
      const VolIndex& vof = vofit();
      //divergence was set in regular update.  we reset it
      // to zero and recalc.
      Real update = 0.;
      for( int idir = 0; idir < SpaceDim; idir++)
        {
          const EBFaceFAB& fluxDir = a_flux[idir];
          for(SideIterator sit; sit.ok(); ++sit)
            {
              int isign = sign(sit());
              Vector<FaceIndex> faces = 
                a_ebisBox.getFaces(vof, idir, sit());
              for(int iface = 0; iface < faces.size(); iface++)
                {
                  const FaceIndex& face = faces[iface];
                  Real areaFrac = a_ebisBox.areaFrac(face);
                  Real faceFlux =fluxDir(face, 0);
                  update += isign*areaFrac*faceFlux;
                      
                }
            }
        }
      //add EB boundary condtions in divergence
      const IntVect& iv = vof.gridIndex();
      Real bndryArea = a_ebisBox.bndryArea(vof);
      RealVect bndryCent = a_ebisBox.bndryCentroid(vof);
      RealVect normal = a_ebisBox.normal(vof);
      RealVect bndryLoc;
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          bndryLoc[idir] = a_dx*(iv[idir] + 0.5 + bndryCent[idir]);
        }
      RealVect exactF = exactFlux(bndryLoc);
      Real     bndryFlux = PolyGeom::dot(exactF, normal);

      update -= bndryFlux*bndryArea; 
      update /= a_dx; //note NOT divided by volfrac

      a_divF(vof, 0) = update;
    }
}
/***************/
void 
kappaDivergenceLD(LevelData<EBCellFAB>&       a_divF,
                  const LevelData<EBFluxFAB>& a_flux,
                  const EBISLayout&           a_ebisl,
                  const DisjointBoxLayout&    a_dbl,
                  const Real&                 a_dx)
{
  for(DataIterator dit= a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      kappaDivergence(a_divF[dit()], 
                      a_flux[dit()], 
                      a_ebisl[dit()],
                      a_dbl.get(dit()),
                      a_dx);
    }
  
}

/**************************/
void
setToExactFluxLD(LevelData<EBCellFAB>&       a_flux,
                 const EBISLayout&           a_ebisl, 
                 const DisjointBoxLayout&    a_dbl,
                 const Real&                 a_dx)
{
  for(DataIterator dit= a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      setToExactFlux(a_flux[dit()], 
                     a_ebisl[dit()],
                     a_dbl.get(dit()),
                     a_dx);
    }
}
/**************************/
void
setToExactDivFLD(LevelData<EBCellFAB>&       a_soln,
                 const EBISLayout&           a_ebisl, 
                 const DisjointBoxLayout&    a_dbl,
                 const Real&                 a_dx)
{
  for(DataIterator dit= a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      setToExactDivF(a_soln[dit()], 
                     a_ebisl[dit()],
                     a_dbl.get(dit()),
                     a_dx);
    }
}
/**************************/
void averageFluxToFaces(EBFluxFAB&            a_outputFlux,
                        const EBCellFAB&      a_inputFlux,
                        const EBISBox&        a_ebisBox,
                        const Box&            a_box,
                        const Real&           a_dx)
{
  assert(a_outputFlux.nComp() == 1);
  assert(a_inputFlux.nComp() == SpaceDim);
  assert(a_outputFlux.getRegion().contains(a_box));
  Box domain = a_ebisBox.getDomain();
  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
    {
      Box interiorBox = a_box;
      interiorBox.grow(faceDir, 1);
      interiorBox &= domain;
      interiorBox.grow(faceDir, -1);

      Box faceBox = surroundingNodes(interiorBox, faceDir);
      EBFaceFAB& faceFlux = a_outputFlux[faceDir];
      BaseFab<Real>& regFaceFlux =          faceFlux.getSingleValuedFAB();
      regFaceFlux.setVal(0.);
      /**/
      const BaseFab<Real>& regCellFlux = a_inputFlux.getSingleValuedFAB();
      FORT_AVEFLUXTOFACE(CHF_FRA1(regFaceFlux, 0),
                         CHF_CONST_FRA1(regCellFlux, faceDir),
                         CHF_CONST_INT(faceDir),
                         CHF_BOX(faceBox));
      /**/

      //empty because no coarse-fine interface here
      IntVectSet cfivs;
      IntVectSet ivsIrreg = a_ebisBox.getIrregIVS(interiorBox);
      for(FaceIterator faceit(ivsIrreg, a_ebisBox.getEBGraph(), faceDir, FaceStop::SurroundingNoBoundary); 
          faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          FaceStencil sten = EBArith::getInterpStencil(face, cfivs, 
                                                       a_ebisBox, domain);
          
          Real newFlux = 0.0;
          for(int isten = 0; isten < sten.size(); isten++)
            {
              const FaceIndex& stenFace= sten.face(isten);
              const VolIndex& vofLo = stenFace.getVoF(Side::Lo);
              const VolIndex& vofHi = stenFace.getVoF(Side::Hi);

              Real valLo = a_inputFlux(vofLo, faceDir);
              Real valHi = a_inputFlux(vofHi, faceDir);
              Real weight = sten.weight(isten);
              Real increment = 0.5*weight*(valHi + valLo);
              newFlux += increment;
            }
        }
      
      //set the boundary fluxes to exact value
      IntVectSet ivsAll(a_box);
      for(FaceIterator faceit(ivsAll, a_ebisBox.getEBGraph(), faceDir, FaceStop::AllBoundaryOnly); 
          faceit.ok(); ++faceit)
        {
          const FaceIndex& face = faceit();
          const IntVect& iv = face.gridIndex(Side::Hi);
          RealVect xval;
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              if(idir == faceDir)
                {
                  xval[idir] = (Real(iv[idir]))*a_dx;
                }
              else
                {
                  xval[idir] = (Real(iv[idir]) + 0.5)*a_dx;
                }
            }
          RealVect exactF = exactFlux(xval); 
          faceFlux(face, 0) = exactF[faceDir];
        }
    }
}
/**************************/
void averageFluxToFacesLD(LevelData<EBFluxFAB>&       a_outputFlux,
                          const LevelData<EBCellFAB>& a_inputFlux,
                          const EBISLayout&           a_ebisl,
                          const DisjointBoxLayout&    a_dbl, 
                          const Real&                 a_dx)
{
  for(DataIterator dit= a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      averageFluxToFaces(a_outputFlux[dit()], 
                         a_inputFlux[dit()], 
                         a_ebisl[dit()],
                         a_dbl.get(dit()),
                         a_dx);
    }
}
/**************************/
void
compareError(const LevelData<EBCellFAB>& a_errorFine,
             const EBISLayout& a_ebislFine, 
             const DisjointBoxLayout& a_gridsFine, 
             const Box& a_domainFine,
             const LevelData<EBCellFAB>& a_errorMed,
             const EBISLayout& a_ebislMed, 
             const DisjointBoxLayout& a_gridsMed, 
             const Box& a_domainMed,
             const LevelData<EBCellFAB>& a_errorCoar,
             const EBISLayout& a_ebislCoar, 
             const DisjointBoxLayout& a_gridsCoar, 
             const Box& a_domainCoar)
{
  assert(a_errorFine.nComp() == 1);
  assert(a_errorCoar.nComp() == 1);

  pout() << "==============================================" << endl;

  EBNormType::NormMode normtype = EBNormType::OverBoth;
  pout() << endl << "Using all uncovered cells." << endl  ;
          
  for(int inorm = 0; inorm <= 2; inorm++)
    {

      if(inorm == 0)
        {
          pout() << endl << "Using max norm." << endl;
        }
      else
        {
          pout() << endl << "Using L-" << inorm << "norm." << endl;
        }
      int comp = 0;
      Real coarnorm = EBArith::norm(a_errorCoar, 
                                    a_gridsCoar, a_ebislCoar, 
                                    comp, inorm, normtype);
      Real mednorm  = EBArith::norm(a_errorMed, 
                                    a_gridsMed, a_ebislMed, 
                                    comp, inorm, normtype);

      Real finenorm = EBArith::norm(a_errorFine, 
                                    a_gridsFine, a_ebislFine, 
                                    comp, inorm, normtype);

      pout() << "Coarse Error Norm = " << coarnorm << endl;
      pout() << "Medium Error Norm = " << mednorm << endl;
      pout() << "Fine   Error Norm = " << finenorm << endl;

      if((Abs(finenorm) > 1.0e-12) && (Abs(coarnorm) > 1.0e-12)) 
        {
          Real order = log(Abs(coarnorm/mednorm))/log(2.0);
          pout() << "Order of scheme (coar to med)= " << order << endl;

          order = log(Abs(mednorm/finenorm))/log(2.0);
          pout() << "Order of scheme (med to fine)= " << order << endl;
        }
    }
  pout() << "==============================================" << endl ;;
}
/**************************/
void 
getError(LevelData<EBCellFAB>&       a_error,
         const EBISLayout&           a_ebisl,
         const DisjointBoxLayout&    a_dbl,
         const Real&                 a_dx)
{
  EBCellFactory ebcellfact(a_ebisl);
  EBFluxFactory ebfluxfact(a_ebisl);
  a_error.define(a_dbl, 1, IntVect::Zero,   ebcellfact);
  LevelData<EBCellFAB> divFCalc(a_dbl, 1, IntVect::Zero,   ebcellfact);
  LevelData<EBCellFAB> divFExac(a_dbl, 1, IntVect::Zero,   ebcellfact);
  LevelData<EBFluxFAB> faceFlux(a_dbl, 1, IntVect::Zero,   ebfluxfact);
  LevelData<EBCellFAB> cellFlux(a_dbl, SpaceDim, 2*IntVect::Unit, ebcellfact);

  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      a_error[dit()].setVal(0.);
      divFCalc[dit()].setVal(0.);
      divFExac[dit()].setVal(0.);
    }

  setToExactDivFLD(divFExac,  a_ebisl, a_dbl, a_dx);
  setToExactFluxLD(cellFlux,  a_ebisl, a_dbl, a_dx);

  Interval interv(0, SpaceDim-1);
  cellFlux.exchange(interv);

  averageFluxToFacesLD(faceFlux, cellFlux, a_ebisl, a_dbl, a_dx);
  kappaDivergenceLD(divFCalc, faceFlux, a_ebisl, a_dbl, a_dx);

  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      EBCellFAB& errorFAB = a_error[dit()];
      EBCellFAB& exactFAB = divFExac[dit()];
      EBCellFAB& calcuFAB = divFCalc[dit()];
      
      errorFAB += calcuFAB;
      errorFAB -= exactFAB;
    }
}
