/*
   __  ______________              __      
  /  |/  / __/ ___/ /  ___  __ _  / /  ___ 
 / /|_/ / _// /__/ _ \/ _ \/  ' \/ _ \/ _ \
/_/  /_/_/  \___/_//_/\___/_/_/_/_.__/\___/
                                           
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 

// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "SphereBL.H"
#include "MFIndexSpace.H"
#include "MFCellFAB.H"
#include "MFRemapper.H"
#include "MFAliasFactory.H"

#include "EBISLayout.H"
#include "EBAMRIO.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "DebugDump.H"

int readGeometryInfo(Box& a_domain, 
                     Real& a_dx,
                     RealVect& a_origin,
                     RealVect& a_center,
                     Real& a_radius);


void makeHierarchy(Vector<DisjointBoxLayout>& dbl,
                   const ProblemDomain& baseDomain,
                   const IntVectSet&    baseTags);

void swapvector(Vector<LevelData<MFCellFAB>* >& left,
                Vector<LevelData<MFCellFAB>* >& right);

                   
/***************/
// define a sphere EBIS.
/***************/
int makeGeometry(MFIndexSpace& mfIndexSpace,
                 const Box& domain, 
                 const Real& dx,
                 const RealVect& origin,
                 const RealVect& center,
                 const Real& radius);
/***************/
//make the corresponding layout
/***************/
int makeEBISL(EBISLayout& a_ebisl, 
              const DisjointBoxLayout& a_grids, 
              const Box& a_domain, 
              const int& nghost);
/***************/
/***************/
/*
int checkEBISL(const EBISLayout& a_ebisl, 
               const DisjointBoxLayout& a_grids, 
               const Box& a_domain, 
               const Real& dx,
               const RealVect& origin,
               const Real& radius,
               const RealVect& center,
               const bool& insideRegular);
*/

/***************/
/***************/
void dumpmemoryatexit();
/***************/
/***************/

char iter_str[80];

int 
main(int argc,char **argv)
{
#ifdef MPI
  MPI_Init(&argc, &argv);
#endif

  // begin forever present scoping trick
  {
    pout()<<std::endl;

    Vector<std::string> names0(1, "phi0");
    Vector<std::string> names1(1, "phi1");
    Vector<int> refRatio(3,2);
    Vector<Real> coveredVal(1,3.0);

    const char* in_file = "sphere.inputs";
    //parse input file
    ParmParse pp(0,NULL,NULL,in_file);
    RealVect center;
    Real radius;
    RealVect origin;
    Real dx;
    Box domain;
    int eekflag = 0;
    MFIndexSpace a, b;
    MFIndexSpace *mfIndexSpace, *oldMFIndexSpace;
    mfIndexSpace = &a;
    oldMFIndexSpace = &b;
 
    LevelData<MFCellFAB> fine, med, coarse;
    LevelData<MFCellFAB> fineOld, medOld, coarseOld;
    Vector<LevelData<MFCellFAB>* > mfvector(3,NULL);
    mfvector[0]=&coarse;
    mfvector[1]=&med;
    mfvector[2]=&fine;
    Vector<LevelData<MFCellFAB>* > mfvector_old(3,NULL);
    mfvector_old[0]=&coarseOld;
    mfvector_old[1]=&medOld;
    mfvector_old[2]=&fineOld;

    readGeometryInfo(domain, 
                     dx,
                     origin,
                     center,
                     radius);

    Box domainFine(domain), domainMedi, domainCoar;
    ProblemDomain pFine(domain);
    Real dxFine(dx), dxMedi, dxCoar;

    assert(eekflag == 0);

    domainMedi = coarsen(domainFine, 2);
    domainCoar = coarsen(domainMedi, 2);
    dxMedi = 2.0*dxFine;
    dxCoar = 2.0*dxMedi;
    Vector<DisjointBoxLayout> grids(3);
    ProblemDomain baseDomain(domainCoar);
    ProblemDomain pMed(domainMedi);

    MFRemapper remapper;

    //make data holders
    Vector<int> comps(2,1);
    int ghost = 1;

    int steps= 25;
    int step = 0;

    Vector<LevelData<EBCellFAB>* > phase(3);
    for(int i=0; i<3; i++) phase[i] = new LevelData<EBCellFAB>();

    while(step < steps){
      
      eekflag = makeGeometry(*mfIndexSpace,
                             domain,
                             dx,
                             origin,
                             center,
                             radius);
      if (eekflag != 0)
        {
          pout() << "non zero eek detected = " << eekflag << endl;
          // MayDay::Error("problem in makeGeometry");
        }
      if(step > 0)
        {
          // same grids, but new topology
          mfvector[2]->define(grids[2], 1, IntVect::Unit, 
                              MFCellFactory(*mfIndexSpace, grids[2],
                                            domainFine,
                                            comps, ghost));

          mfvector[1]->define(grids[1], 1, IntVect::Unit, 
                              MFCellFactory(*mfIndexSpace, grids[1], 
                                            domainMedi,
                                            comps, ghost)); 
          
          mfvector[0]->define(grids[0], 1, IntVect::Unit, 
                              MFCellFactory(*mfIndexSpace, grids[0],
                                            domainCoar,
                                            comps, ghost)); 
          DisjointBoxLayout d1 = mfvector_old[0]->getBoxes();
          assert(d1 == grids[0]);
          for(int i=0; i<3; ++i){
            remapper.remap(*oldMFIndexSpace,
                           *mfvector_old[i],
                           *mfIndexSpace,
                           *mfvector[i]);
          }
          
          swapvector(mfvector, mfvector_old);
        }
 
       /*
       if(step > 0){
         for(int i=0; i<3; i++) phase[i] = new LevelData<EBCellFAB>();
         aliasMF(phase, 0, mfvector_old);
         sprintf(iter_str, "phase0_old.%03d.%dd.hdf5",step, SpaceDim);
         writeEBHDF5(iter_str, grids, phase, names0, domainCoar,
                     dxCoar, 1, step, refRatio, 3, true, coveredVal);
         
         aliasMF(phase, 1, mfvector_old);
         sprintf(iter_str, "phase1_old.%03d.%dd.hdf5",step, SpaceDim);
         writeEBHDF5(iter_str, grids, phase, names1, domainCoar,
                     dxCoar, 1, step, refRatio, 3, true, coveredVal);
       }
       */

      //make grids 
      IntVectSet tags = mfIndexSpace->interfaceRegion(2);
      //tags.grow(1);
      makeHierarchy(grids, baseDomain, tags);

     
        MFCellFactory fineFactory(*mfIndexSpace, grids[2],domainFine,
                                  comps, ghost);
        mfvector[2]->define(grids[2], 1, IntVect::Unit, fineFactory); 
                          
        MFCellFactory  medFactory(*mfIndexSpace, grids[1], domainMedi,
                                  comps, ghost);
        mfvector[1]->define(grids[1], 1, IntVect::Unit, medFactory);
                         
        MFCellFactory coarseFactory(*mfIndexSpace, grids[0],domainCoar,
                                    comps, ghost);
        mfvector[0]->define(grids[0], 1, IntVect::Unit, coarseFactory); 
     
    
      if(step == 0)
        {
          //initialize data holders
          LevelData<EBCellFAB> phase0, phase1;
          for(int lev=0; lev<3; lev++)
            {
              aliasMF(phase0, 0, *mfvector[lev]);
              aliasMF(phase1, 1, *mfvector[lev]);
              DataIterator dit = phase0.dataIterator();
              for(dit.begin(); dit.ok(); ++dit)
                {
                  phase0[dit].setVal(0.0);
                  phase1[dit].setVal(1.0);
                }
            }

        }
      else
        {
          // perform second type of remap.  constant MFIndexSpace, but
          // different grids.
          //
          // coarse grids in this case remain identical.
          mfvector_old[0] ->copyTo(mfvector[0]->interval(),*mfvector[0],
                               mfvector_old[0]->interval());
  

          remapper.remap(*mfIndexSpace, baseDomain, pMed,
                           *mfvector_old[1], *mfvector[0],
                           2, 1, *mfvector[1]);
          remapper.remap(*mfIndexSpace, pMed, pFine,
                           *mfvector_old[2], *mfvector[1],
                           2, 1, *mfvector[2]);
                          

        }



      //  now, output results to an hdf5 file for each phase.
      for(int i=0; i<3; i++){
        phase[i] = new LevelData<EBCellFAB>();
      }
      aliasMF(phase, 0, mfvector);
      sprintf(iter_str, "phase0.%03d.%dd.hdf5",step, SpaceDim);
      writeEBHDF5(iter_str, grids, phase, names0, domainCoar,
                  dxCoar, 1, step, refRatio, 3, true, coveredVal);

      aliasMF(phase, 1, mfvector);
      sprintf(iter_str, "phase1.%03d.%dd.hdf5",step, SpaceDim);
      writeEBHDF5(iter_str, grids, phase, names1, domainCoar,
                  dxCoar, 1, step, refRatio, 3, true, coveredVal);

      for(int i=0; i<3; i++) delete phase[i];
    
      step++;
   
      center[0]+= dx/3.0;
      center[1]+= dx/3.0;
      radius -= dx/10.0;
      {
        MFIndexSpace* swap = oldMFIndexSpace;
        oldMFIndexSpace = mfIndexSpace;
        mfIndexSpace = swap;
      }
      swapvector(mfvector, mfvector_old);
      pout()<<step<<std::endl;
        }
    
      pout() << "\n mapper test passed" << endl;
    } // end scoping trick

    
#ifdef MPI
    MPI_Finalize();
#endif
    
    return 0;
}


void swapvector(Vector<LevelData<MFCellFAB>* >& left,
                Vector<LevelData<MFCellFAB>* >& right)

{

  LevelData<MFCellFAB>* swap;
  for(int i=0; i<left.size(); ++i)
    {
      swap = left[i]; 
      left[i]=right[i];
      right[i]=swap;
    }
}

int readGeometryInfo(Box& a_domain, 
                     Real& a_dx,
                     RealVect& a_origin,
                     RealVect& a_center,
                     Real& a_radius)
{

  // parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);

  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;

  for (int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if (n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          return(-1);
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain.setSmall(lo);
  a_domain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim, 1.0);
  Real prob_hi;

  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.get("prob_hi",prob_hi);

  a_dx = (prob_hi-prob_lo[0])/n_cell[0];
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      a_origin[idir] = prob_lo[idir];
    }

  pp.get("radius",a_radius);

  // ParmParse doesn't get RealVects, so work-around with Vector<Real>
  Vector<Real> vectorCenter;
  pp.getarr("center",vectorCenter,0,SpaceDim);
  for (int idir = 0; idir < SpaceDim; idir++)
    {
      a_center[idir] = vectorCenter[idir];
    }

  return 0;
} // end read from file

int makeGeometry(MFIndexSpace& mfIndexSpace, 
                 const Box& a_domain, 
                 const Real& a_dx,
                 const RealVect& a_origin,
                 const RealVect& a_center,
                 const Real& a_radius)
{

  int eekflag = 0;
  Vector<GeometryService*> geometry(2, NULL);
  

  SphereBL insidesphere(a_radius,a_center,true,a_dx);
  GeometryShop workshop0(insidesphere);
  workshop0.m_phase = 0;
  geometry[0] = &workshop0;
  
  SphereBL outsidesphere(a_radius,a_center,false,a_dx);
  GeometryShop workshop1(outsidesphere);
  workshop1.m_phase=1;
  geometry[1] = &workshop1;

  mfIndexSpace.define(a_domain,a_origin,a_dx,geometry);
  
  return eekflag;
}




void makeHierarchy(Vector<DisjointBoxLayout>& dbl,
                   const ProblemDomain& baseDomain,
                   const IntVectSet&    baseTags)
{
  
  dbl.resize(3);
  Vector<int> refRatios(2,2);
  Real fillRatio = 0.85;
  int blockingFactor= 4;
  int bufferSize =   2;
  int maxSize    =   32;
  BRMeshRefine regridder(baseDomain, refRatios, fillRatio, blockingFactor,
                         bufferSize, maxSize);

  Vector<Vector<Box> > oldGrids(3,1), newGrids(3);
  oldGrids[0][0]=baseDomain.domainBox();
  oldGrids[1][0]=coarsen(oldGrids[0][0], 2);
  oldGrids[2][0]=coarsen(oldGrids[1][0], 2);

  regridder.regrid(newGrids, baseTags, 0, 1, oldGrids);

  Vector<int> procs;
  for(int i=0; i<3; i++){
    LoadBalance(procs, newGrids[i]);
    dbl[i] = DisjointBoxLayout(newGrids[i], procs);
  }

}
