/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents ofthe University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "VoFIterator.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "DebugDump.H"
#include "EBDebugDump.H"
const bool g_verbose = true;
/***************/
// define a ramp EBIS.
/***************/
int makeGeometry(Box& domain);
/***************/
int checkCoarseAssortment(const Box& a_domain);
/***************/
/***************/
void dumpmemoryatexit();
/***************/
/***************/
RealVect getVolumeCentroid(const Box& a_domain,
                           const DisjointBoxLayout& a_dbl,
                           const EBISLayout&  a_ebisl);
/***************/
/***************/
RealVect getBoundaryCentroid(const Box& a_domain,
                             const DisjointBoxLayout& a_dbl,
                             const EBISLayout&  a_ebisl);
/***************/
/***************/
Real getUncoveredVolume(const Box& a_domain,
                        const DisjointBoxLayout& a_dbl,
                        const EBISLayout&  a_ebisl);
/***************/
/***************/
int
main(int argc, char** argv)
{

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  //begin forever present scoping trick
  {
    const char* in_file = "ramp.inputs";
    //parse input file
    ParmParse pp(0,NULL,NULL,in_file);
    int eekflag = 0;
    //define the geometry object.
    Box domain;
    eekflag =  makeGeometry(domain);
    if(eekflag != 0)
      {
        pout() << "non zero eek detected = " << eekflag << endl;
        MayDay::Error("problem in makeGeometry");
      }

    //check that total coarse vol = total fine vol
    //check that total coarse vol centroid = total fine vol centroid
    eekflag = checkCoarseAssortment(domain);
    if(eekflag != 0)
      {
        pout() << "checkCoarse: eek = " << eekflag << endl;
        MayDay::Error("problem in checkEBISL");
      }

    pout() << "coarsening test passed" << endl;
  }//end scoping trick
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->clear();
#ifdef MPI
  MPI_Finalize();
#endif
  return 0;
}
/**********/
int makeGeometry(Box& a_domain)
{
  Real dx;
  RealVect origin;
  int upDir;
  int indepVar;
  Real startPt;
  Real slope;
  int eekflag =  0;
  //parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          return(-1);
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain.setSmall(lo);
  a_domain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim, 1.0);
  Real prob_hi;
  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.get("prob_hi",prob_hi);
  dx = (prob_hi-prob_lo[0])/n_cell[0];
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      origin[idir] = prob_lo[idir];
    }
  pp.get("up_dir",upDir);
  pp.get("indep_var",indepVar);
  pp.get("start_pt", startPt);
  pp.get("ramp_slope", slope);

  RampBL ramp(upDir, indepVar, startPt, slope);
  GeometryShop workshop(ramp);
  //this generates the new EBIS
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(a_domain, origin, dx, workshop);

  return eekflag;
}
/***************/
/***************/
Real getUncoveredVolume(const Box& a_domain,
                        const DisjointBoxLayout& a_dbl,
                        const EBISLayout&  a_ebisl)
{
  assert(!a_domain.isEmpty());
  Real totalVol = 0;
  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = a_ebisl[dit()];
      IntVectSet domainIVS(a_domain);
      for(VoFIterator vofit(domainIVS, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          Real volVoF = ebisBox.volFrac(vof);
          totalVol += volVoF;
        }
    }
  Real domainSize = Real(a_domain.numPts());
  assert(domainSize >= 1.0);
  Real normedVol = totalVol/domainSize;
  return normedVol;
}
/***************/
/***************/
RealVect getVolumeCentroid(const Box& a_domain,
                           const DisjointBoxLayout& a_dbl,
                           const EBISLayout&  a_ebisl)
{
  assert(!a_domain.isEmpty());
  RealVect totalVolCent = RealVect::Zero;
  Real totalVol = 0.;
  Real dx = 1.0/Real(a_domain.size(0));
  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = a_ebisl[dit()];
      IntVectSet domainIVS(a_domain);
      for(VoFIterator vofit(domainIVS, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect& iv = vof.gridIndex();
          Real volVoF = ebisBox.volFrac(vof);
          totalVol += volVoF;

          RealVect volCentVoF = ebisBox.centroid(vof);
          //make centroid relative to global coordinate system
          for(int idir = 0; idir < SpaceDim; idir++)
            {
              volCentVoF[idir] += (Real(iv[idir])+ 0.5);
              volCentVoF[idir] *= dx;
              volCentVoF[idir] *= volVoF;
              totalVolCent[idir] += volCentVoF[idir];
            }
        }
    }
  RealVect normedVolCent = totalVolCent;
  assert(totalVol > PolyGeom::getTolerance());
  normedVolCent /= totalVol;
  return normedVolCent;
}
/***************/
/***************/
RealVect getBoundaryCentroid(const Box& a_domain,
                             const DisjointBoxLayout& a_dbl,
                             const EBISLayout&  a_ebisl)
{
  assert(!a_domain.isEmpty());
  RealVect totalBndryCent = RealVect::Zero;
  Real totalBndryArea = 0.0;
  int  numCellsWithBoundary = 0;
  Real dx = 1.0/Real(a_domain.size(0));
  for(DataIterator dit = a_dbl.dataIterator(); dit.ok(); ++dit)
    {
      const EBISBox& ebisBox = a_ebisl[dit()];
      IntVectSet domainIVS(a_domain);
      for(VoFIterator vofit(domainIVS, ebisBox.getEBGraph()); vofit.ok(); ++vofit)
        {
          const VolIndex& vof = vofit();
          const IntVect& iv = vof.gridIndex();
          Real bndryAreaVoF = ebisBox.bndryArea(vof);
          if(bndryAreaVoF > 0.01*PolyGeom::getTolerance())
            {
              numCellsWithBoundary++;
              totalBndryArea += bndryAreaVoF;

              RealVect bndryCentVoF = ebisBox.bndryCentroid(vof);
              //make centroid relative to global coordinate system
              for(int idir = 0; idir < SpaceDim; idir++)
                {
                  bndryCentVoF[idir] += (Real(iv[idir])+ 0.5);
                  bndryCentVoF[idir] *= dx;
                  bndryCentVoF[idir] *= bndryAreaVoF;
                  totalBndryCent[idir] += bndryCentVoF[idir];
                }
            }
        }
    }
  RealVect normedBndryCent = totalBndryCent;
  if(totalBndryArea > PolyGeom::getTolerance())
    normedBndryCent /= totalBndryArea;
  return normedBndryCent;
}
/***************/
/***************/
int checkCoarseAssortment(const Box& a_domain)
{
  Real tolerance  = PolyGeom::getTolerance();
  const EBIndexSpace* const ebisPtr = Chombo_EBIS::instance();
  assert(ebisPtr->isDefined());
  Box levelDomain = a_domain;
  EBISLayout ebisl;
  Vector<Box> vbox(1, levelDomain);
  Vector<int>  procAssign(1, 0);
  DisjointBoxLayout dbl(vbox, procAssign);
  int nghost = 0;
  ebisPtr->fillEBISLayout(ebisl, dbl, levelDomain, nghost);

  Real fineVolume = getUncoveredVolume(levelDomain, dbl, ebisl);
  RealVect fineVolCentroid = getVolumeCentroid(levelDomain, dbl, ebisl);
  RealVect fineBndryCentroid = getBoundaryCentroid(levelDomain, dbl, ebisl);
  if(g_verbose)
    {
      pout() << "finest level:" <<  endl;
      pout() << "volume       =\t" << fineVolume << endl;
      for(int idir=0; idir < SpaceDim; idir++)
        {
          pout() << "volCentroid("<<idir <<  ") =\t" 
                 << fineVolCentroid[idir] << endl;
          pout() << "bndryCentroid("<<idir <<") =\t" << 
            fineBndryCentroid[idir] << endl << endl;;
        }
    }
  int numLevels = ebisPtr->numLevels();
  for(int ilev = 1; ilev < numLevels; ilev++)
    {
      levelDomain.coarsen(2);
      assert(!levelDomain.isEmpty());
      vbox[0] = levelDomain;
      DisjointBoxLayout levelDBL(vbox, procAssign);
      EBISLayout levelEBISL;
      ebisPtr->fillEBISLayout(levelEBISL, levelDBL, levelDomain, nghost);

      Real volume = 
        getUncoveredVolume(levelDomain, levelDBL, levelEBISL);
      RealVect volCentroid = 
        getVolumeCentroid(levelDomain, levelDBL, levelEBISL);
      RealVect bndryCentroid =
        getBoundaryCentroid(levelDomain, levelDBL, levelEBISL);
      if(g_verbose)
        {
          pout() << "ilev         =\t" << ilev  << endl;
          pout() << "volume       =\t" << volume << endl;
          for(int idir=0; idir < SpaceDim; idir++)
            {
              pout() << "volCentroid("<<idir <<  ") =\t" 
                     << volCentroid[idir] << endl;
              pout() << "bndryCentroid("<<idir <<") =\t" << 
                bndryCentroid[idir] << endl << endl;;
            }
        }
      if(Abs(volume - fineVolume) > tolerance)
        {
          pout() << "volumes don't match at level = " << ilev << endl;
          return -1;
        }
      for(int idir = 0; idir < SpaceDim; idir++)
        {
          if(Abs(volCentroid[idir] - fineVolCentroid[idir]) > tolerance)
            {
              pout() << "vol centroids don't match at level = " 
                     << ilev << "  idir = " << idir <<  endl;
              return -2;
            }
          if(Abs(bndryCentroid[idir] - fineBndryCentroid[idir]) > tolerance)
            {
              pout() << "bndry centroids don't match at level = " 
                     << ilev << "  idir = " << idir <<  endl;
              return -3;
            }
        }
    }
  return 0;
}
/***************/
/***************/
