/* _______              __
   / ___/ /  ___  __ _  / /  ___
   / /__/ _ \/ _ \/  ' \/ _ \/ _ \
   \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents ofthe University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "RampBL.H"
#include "EBIndexSpace.H"
#include "EBISLayout.H"
#include "BoxIterator.H"
#include "ParmParse.H"
#include "BRMeshRefine.H"
#include "LoadBalance.H"
#include "VoFIterator.H"
#include "GeometryShop.H"
#include "PolyGeom.H"
#include "DebugDump.H"
#include "EBDebugDump.H"
#include "EBArith.H"
const bool g_verbose = true;
/***************/
// define a ramp EBIS.
/***************/
int makeGeometry(Box& domain);
/***************/
int checkIO(const Box& a_domain);
/***************/
void dumpmemoryatexit();
/***************/
int
main(int argc, char** argv)
{

#ifdef MPI
  MPI_Init(&argc, &argv);
#endif
  //begin forever present scoping trick
  {
    const char* in_file = "ramp.inputs";
    //parse input file
    ParmParse pp(0,NULL,NULL,in_file);
    int eekflag = 0;
    //define the geometry object.
    Box domain;
    eekflag =  makeGeometry(domain);
    if(eekflag != 0)
      {
        pout() << "non zero eek detected = " << eekflag << endl;
        MayDay::Error("problem in makeGeometry");
      }

    //check that total coarse vol = total fine vol
    //check that total coarse vol centroid = total fine vol centroid
    eekflag = checkIO(domain);
    if(eekflag != 0)
      {
        pout() << "checkCoarse: eek = " << eekflag << endl;
        MayDay::Error("problem in checkEBISL");
      }

    pout() << "eieio test passed" << endl;
  }//end scoping trick
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->clear();
#ifdef MPI
  MPI_Finalize();
#endif
  return 0;
}
/**********/
int makeGeometry(Box& a_domain)
{
  Real dx;
  RealVect origin;
  int upDir;
  int indepVar;
  Real startPt;
  Real slope;
  int eekflag =  0;
  //parse input file
  ParmParse pp;

  Vector<int> n_cell(SpaceDim);
  pp.getarr("n_cell",n_cell,0,SpaceDim);

  assert(n_cell.size() == SpaceDim);
  IntVect lo = IntVect::TheZeroVector();
  IntVect hi;
  for(int ivec = 0; ivec < SpaceDim; ivec++)
    {
      if(n_cell[ivec] <= 0)
        {
          pout() << " bogus number of cells input = " << n_cell[ivec];
          return(-1);
        }
      hi[ivec] = n_cell[ivec] - 1;
    }

  a_domain.setSmall(lo);
  a_domain.setBig(hi);

  Vector<Real> prob_lo(SpaceDim, 1.0);
  Real prob_hi;
  pp.getarr("prob_lo",prob_lo,0,SpaceDim);
  pp.get("prob_hi",prob_hi);
  dx = (prob_hi-prob_lo[0])/n_cell[0];
  for(int idir = 0; idir < SpaceDim; idir++)
    {
      origin[idir] = prob_lo[idir];
    }
  pp.get("up_dir",upDir);
  pp.get("indep_var",indepVar);
  pp.get("start_pt", startPt);
  pp.get("ramp_slope", slope);

  RampBL ramp(upDir, indepVar, startPt, slope);
  GeometryShop workshop(ramp);
  //this generates the new EBIS
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  ebisPtr->define(a_domain, origin, dx, workshop);

  return eekflag;
}
/***************/
int checkIO(const Box& a_domain)
{
#ifdef HDF5
  Real tolerance  = PolyGeom::getTolerance();
  EBIndexSpace* ebisPtr = Chombo_EBIS::instance();
  Vector<Box> vbox(1, a_domain);
  Vector<int>  procAssign(1, 0);
  DisjointBoxLayout dblFinest(vbox, procAssign);
  int numLevels = ebisPtr->numLevels();
  //create all the layouts
  Vector<DisjointBoxLayout> dblayouts(numLevels);
  Vector<Box>               domains(numLevels);
  dblayouts[0] = dblFinest;
  domains[0] = a_domain;
  for(int ilev = 1; ilev < numLevels; ilev++)
    {
      ebcoarsen(dblayouts[ilev], dblayouts[ilev-1], 2);
      domains[ilev] = ebcoarsen(domains[ilev-1], 2);
    }

  Vector<EBISLayout> ebislOutput(numLevels);
  Vector<EBISLayout> ebislInput (numLevels);
  //define a bunch of ebisls to represent ebis before output
  for(int ilev = 0; ilev < numLevels; ilev++)
    {
      ebisPtr->fillEBISLayout(ebislOutput[ilev], dblayouts[ilev], domains[ilev], 0);
    }
  //output original ebis to a file.
  const char* const  filename = "eieio.hdf5";
  HDF5Handle handleOut(filename, HDF5Handle::CREATE);
  ebisPtr->write(handleOut);
  handleOut.close();

  //define ebis anew from file output
  HDF5Handle handleIn(filename, HDF5Handle::OPEN_RDONLY);
  ebisPtr->define(handleIn);
  handleIn.close();

  //define a bunch of ebisls to represent ebis after input
  for(int ilev = 0; ilev < numLevels; ilev++)
    {
      ebisPtr->fillEBISLayout(ebislInput[ilev], dblayouts[ilev], domains[ilev], 0);
    }

  //now compare the damn things.  this test will break horribly if 
  //we ever do the "data base of EBISLayouts" idea
  for(int ilev = 0; ilev < numLevels; ilev++)
    {
      const EBISLayout& ebislOld = ebislOutput[ilev];
      const EBISLayout& ebislNew = ebislInput[ilev];
      const DisjointBoxLayout& dbl = dblayouts[ilev];
      for(DataIterator dit = dbl.dataIterator(); dit.ok(); ++dit)
        {
          const EBISBox& ebisBoxOld = ebislOld[dit()];
          const EBISBox& ebisBoxNew = ebislNew[dit()];
          const Box& box = dbl.get(dit());
          IntVectSet ivsOld = ebisBoxOld.getIrregIVS(box);
          IntVectSet ivsNew = ebisBoxNew.getIrregIVS(box);
          IntVectSet ivsSumOfDiffs = ((ivsOld - ivsNew) |
                                      (ivsNew - ivsOld));
          if(!ivsSumOfDiffs.isEmpty())
            {
              pout() << "irreg IVS does not match "  << endl;
              return -113;
            }
          ivsOld = ebisBoxOld.getMultiCells(box);
          ivsNew = ebisBoxNew.getMultiCells(box);
          ivsSumOfDiffs = ((ivsOld - ivsNew) |
                           (ivsNew - ivsOld));
          if(!ivsSumOfDiffs.isEmpty())
            {
              pout() << "multi IVS does not match "  << endl;
              return -115;
            }
          for(BoxIterator bit(box); bit.ok(); ++bit)
            {
              Vector<VolIndex> vofsOld = ebisBoxOld.getVoFs(bit());
              Vector<VolIndex> vofsNew = ebisBoxNew.getVoFs(bit());
              if(vofsOld.size() != vofsNew.size())
                {
                  pout() << "vof size does not match at " << bit() << endl;
                  return -1;
                }
              for(int ivof = 0; ivof < vofsOld.size(); ivof++)
                {
                  //now check all the cell-centered stuff
                  const VolIndex& vofOld = vofsOld[ivof];
                  const VolIndex& vofNew = vofsNew[ivof];
                  if(vofOld != vofNew)
                    {
                      pout() << "vofs do not match at " << bit() << endl;
                      return -2;
                    }
                  Real volFracOld = ebisBoxOld.volFrac(vofOld);
                  Real volFracNew = ebisBoxNew.volFrac(vofNew);
                  if(Abs(volFracOld - volFracNew) > tolerance)
                    {
                      pout() << "vol fracs do not match at " << bit() << endl;
                      return -3;
                    }
                  Real bndryAreaOld = ebisBoxOld.bndryArea(vofOld);
                  Real bndryAreaNew = ebisBoxNew.bndryArea(vofNew);
                  if(Abs(bndryAreaOld - bndryAreaNew) > tolerance)
                    {
                      pout() << "bndry areas do not match at " << bit() << endl;
                      return -4;
                    }
                  RealVect normalOld = ebisBoxOld.normal(vofOld);
                  RealVect normalNew = ebisBoxNew.normal(vofNew);
                  RealVect volCentroidOld = ebisBoxOld.centroid(vofOld);
                  RealVect volCentroidNew = ebisBoxNew.centroid(vofNew);
                  RealVect bndryCentroidOld = ebisBoxOld.bndryCentroid(vofOld);
                  RealVect bndryCentroidNew = ebisBoxNew.bndryCentroid(vofNew);
                  for(int idir = 0; idir < SpaceDim; idir++)
                    {
                      if(Abs(bndryCentroidOld[idir] - bndryCentroidNew[idir]) > tolerance)
                        {
                          pout() << "bndry centroids do not match at iv dir " << bit() 
                                 << " " << idir << endl;
                          return -4;
                        }
                      if(Abs(volCentroidOld[idir] - volCentroidNew[idir]) > tolerance)
                        {
                          pout() << "vol centroids do not match at iv dir " << bit() 
                                 << " " << idir << endl;
                          return -5;
                        }
                      if(Abs(normalOld[idir] - normalNew[idir]) > tolerance)
                        {
                          pout() << "normals do not match at iv dir " << bit() 
                                 << " " << idir << endl;
                          return -6;
                        }
                    }
                  
                  if(ilev > 0)
                    {
                      VolIndex coarVoFOld = ebisBoxOld.coarsen(vofOld);
                      VolIndex coarVoFNew = ebisBoxNew.coarsen(vofNew);
                      if(coarVoFOld != coarVoFNew)
                        {
                          pout() << "coarse vofs do not match at " << bit() << endl;
                          return -7;
                        }
                    }
                  if(ilev < numLevels - 1)
                    {
                      Vector<VolIndex> fineVoFsOld = ebisBoxOld.refine(vofOld);
                      Vector<VolIndex> fineVoFsNew = ebisBoxNew.refine(vofNew);
                      if(fineVoFsOld.size() != fineVoFsNew.size())
                        {
                          pout() << "fine vof sizes do not match at " << bit() << endl;
                          return -8;
                        }
                      for(int ifine = 0; ifine < fineVoFsOld.size();  ifine++)
                        {
                          const VolIndex& fineVoFOld = fineVoFsOld[ifine];
                          const VolIndex& fineVoFNew = fineVoFsNew[ifine];
                          if(fineVoFOld != fineVoFNew)
                            {
                              pout() << "fine vofs do not match at " << bit() << endl;
                              return -9;
                            }            
                        }
                    }
                  //now check all the face-centered stuff
                  for(int faceDir = 0; faceDir < SpaceDim; faceDir++)
                    {
                      for(SideIterator sit; sit.ok(); ++sit)
                        {
                          Vector<FaceIndex> facesOld = ebisBoxOld.getFaces(vofOld, faceDir, sit());
                          Vector<FaceIndex> facesNew = ebisBoxNew.getFaces(vofNew, faceDir, sit());
                          if(facesOld.size() != facesNew.size())
                            {
                              pout() << "fine face sizes do not match at " << bit() << endl;
                              pout() << "facedir, side = " << faceDir << "  " << sign(sit()) << endl;
                              return -10;
                            }
                          for(int iface = 0; iface < facesOld.size(); iface++)
                            {
                              const FaceIndex& faceOld = facesOld[iface];
                              const FaceIndex& faceNew = facesNew[iface];
                              if(faceOld != faceNew)
                                {
                                  pout() << "faces do not match at " << bit() << endl;
                                  pout() << "facedir, side = " << faceDir << "  " << sign(sit()) << endl;
                                  return -11;
                                } 
                              Real areaFracOld = ebisBoxOld.areaFrac(faceOld);
                              Real areaFracNew = ebisBoxNew.areaFrac(faceNew);
                              if(Abs(areaFracOld - areaFracNew) > tolerance)
                                {
                                  pout() << "area fracs do not match at " << bit() << endl;
                                  pout() << "facedir, side = " << faceDir << "  " << sign(sit()) << endl;
                                  return -12;
                                }
                              
                              RealVect faceCentroidOld = ebisBoxOld.centroid(faceOld);
                              RealVect faceCentroidNew = ebisBoxNew.centroid(faceNew);
                              for(int idir = 0; idir < SpaceDim; idir++)
                                {
                                  if(Abs(faceCentroidOld[idir] - faceCentroidNew[idir]) > tolerance)
                                    {
                                      pout() << "face centroids do not match at " << bit() << endl;
                                      pout() << "facedir, side = " << faceDir << "  " << sign(sit()) << endl;
                                      return -13;
                                    }
                                }
                            } //end loop over faces
                        } //end loop over sides for faces
                    } //end loop over face directions
                } //end loop over vofs at a cell
            } //end loop over cells in a box
        } //end loop over boxes 
    }//end loop over refinement levels
#endif // HDF5
  return 0;
}
/***************/
/***************/
