/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/ 
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
// 
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// Dan Martin, Fri, Jan 14, 2000

#include "computeSum.H"
#include "LayoutIterator.H"

Real computeSum(const Vector<LevelData<FArrayBox>* >& a_phi, 
                const Vector<int>&                    a_nRefFine,
                const Real&                           a_dxCrse,
                const Interval&                       a_comps, 
                const int&                            a_lBase)
{
  Real dxLevel = a_dxCrse;
  int numLevels = a_phi.size();
  Real sum, sumLevel;

  sum = 0.0;

  // loop over levels  
  for (int lev = a_lBase; lev < numLevels; lev++) {
    //in case there are extra levels which are not defined
    if (a_phi[lev] != NULL)
    {
      assert(a_phi[lev]->isDefined());
      
      LevelData<FArrayBox>& thisPhi = *(a_phi[lev]);
      const DisjointBoxLayout* finerGridsPtr;
      
      if (lev < numLevels-1)
      {
        finerGridsPtr = &(a_phi[lev+1]->getBoxes());
      }
      else
      {
        finerGridsPtr = NULL;
      }

      sumLevel = computeSum(thisPhi,finerGridsPtr,a_nRefFine[lev],
                            dxLevel,a_comps);
      sum += sumLevel;

      dxLevel = dxLevel/a_nRefFine[lev];      
    }
  }
  
  // do broadcast/gather thing here
  return sum;
}

Real computeSum(const LevelData<FArrayBox>& a_phi, 
                const DisjointBoxLayout*    a_finerGridsPtr,
                const int&                  a_nRefFine,
                const Real&                 a_dx, 
                const Interval&             a_comps)
{
  Real sum, sumLevel;
  sum = 0.0;

  const DisjointBoxLayout& levelGrids = a_phi.getBoxes();

  LevelData<FArrayBox> temp(levelGrids,a_comps.size());

  Interval tempComps(0,a_comps.size()-1);
  
  DataIterator dit = a_phi.dataIterator();

  for (dit.reset(); dit.ok(); ++dit) 
  {
    Box copyBox = temp[dit()].box();
    temp[dit()].copy(copyBox,tempComps,copyBox,a_phi[dit()],a_comps);

    if (a_finerGridsPtr != NULL)
    {  
      LayoutIterator litFine = a_finerGridsPtr->layoutIterator();
    
      // now loop over fine boxes and set covered regions to 0
      for (litFine.reset(); litFine.ok(); ++litFine)
      {
        Box coveredBox(a_finerGridsPtr->get(litFine()));
        coveredBox.coarsen(a_nRefFine);
        coveredBox &= temp[dit()].box();
      
        if (!coveredBox.isEmpty())
        {
          temp[dit()].setVal(0.0,coveredBox,0,tempComps.size());
        }
      }  // end loop over fine-grid boxes
    } // end if there is a finer level
    
    sumLevel = temp[dit()].sum(0,tempComps.size());
    sum += sumLevel;
  } // end loop over this level's grids

  for (int i = 0; i < SpaceDim; i++)
  {
    sum *= a_dx;
  }

  // do broadcast/gather thing here
#ifdef MPI
  Real recv;
  int result = MPI_Allreduce(&sum, &recv, 1, MPI_CH_REAL,
                             MPI_SUM, Chombo_MPI::comm);
  if (result != MPI_SUCCESS) 
    {
      MayDay::Error("Sorry, but I had a communication error in computeSum");
    }

  sum = recv;
#endif

  return sum;
}
