/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#ifndef _AMR_ADVECT_DIFFUSE_H_
#define _AMR_ADVECT_DIFFUSE_H_

#include "FArrayBox.H"
#include "LevelData.H"
#include "AMRLevel.H"
#include "CoarseAverage.H"
#include "FineInterp.H"
#include "LevelFluxRegister.H"

#include "LevelGodunov.H"
#include "PhysIBC.H"
#include "Box.H"
#include "IntVectSet.H"
#include "Vector.H"
#include "RealVect.H"
#include "DisjointBoxLayout.H"
#include "BaseHeatSolver.H"
#include "HelmholtzOp.H"

/// AMR Godunov
/**
 */
class AMRAdvectDiffuse : public AMRLevel
{
public:
  /// Constructor
  /**
   */
  AMRAdvectDiffuse();

  /// Destructor
  /**
   */
  virtual ~AMRAdvectDiffuse();

  /// This instance should never get called - historical
  /**
   */
  virtual void define(AMRLevel*  a_coarserLevelPtr,
                      const Box& a_problemDomain,
                      int        a_level,
                      int        a_refRatio);

  /// Define new AMR level
  /**
   */
  virtual void define(AMRLevel*            a_coarserLevelPtr,
                      const ProblemDomain& a_problemDomain,
                      int                  a_level,
                      int                  a_refRatio);


  /// Advance by one timestep
  /**
   */
  virtual Real advance();

  /// Things to do after a timestep
  /**
   */
  virtual void postTimeStep();

  /// Create tags for regridding
  /**
   */
  virtual void tagCells(IntVectSet& a_tags) ;

  /// Create tags at initialization
  /**
   */
  virtual void tagCellsInit(IntVectSet& a_tags) ;

  /// Set up data on this level after regridding
  /**
   */
  virtual void regrid(const Vector<Box>& a_newGrids);

  /// Initialize grids
  /**
   */
  virtual void initialGrid(const Vector<Box>& a_newGrids);

  /// Initialize data
  /**
   */
  virtual void initialData();

  /// Things to do after initialization
  /**
   */
  virtual void postInitialize();

  /// Create a local PatchGodunov factory using the argument as a factory
  /**
   */
  virtual void patchGodunov(const PatchGodunov* const a_patchGodunov);

  /// pass in a pointer to the heat solver we will be using
  /**
   */
  virtual void diffusionSolver(const BaseHeatSolver* const a_heatSolver);

#ifdef HDF5
  /// Write checkpoint header
  /**
   */
  virtual void writeCheckpointHeader(HDF5Handle& a_handle) const;

  /// Write checkpoint data for this level
  /**
   */
  virtual void writeCheckpointLevel(HDF5Handle& a_handle) const;

  /// Read checkpoint header
  /**
   */
  virtual void readCheckpointHeader(HDF5Handle& a_handle);

  /// Read checkpoint data for this level
  /**
   */
  virtual void readCheckpointLevel(HDF5Handle& a_handle);

  /// Write plotfile header
  /**
   */
  virtual void writePlotHeader(HDF5Handle& a_handle) const;

  /// Write plotfile data for this level
  /**
   */
  virtual void writePlotLevel(HDF5Handle& a_handle) const;
#endif

  /// Returns the dt computed earlier for this level
  /**
   */
  virtual Real computeDt();

  /// Compute dt using initial data
  /**
   */
  virtual Real computeInitialDt();

  /// Set the CFL number
  /**
   */
  virtual void CFL(Real a_cfl);

  /// set the diffusion coefficient nu
  virtual void diffusionCoeff(Real a_nu);

  /// set whether to do implicit refluxing (default is false)
  virtual void doImplicitReflux(bool a_doImplicitReflux);

  /// set tolerance for diffusion solvers (default is 1.0e-7)
  virtual void diffusionSolverTolerance(Real a_solverTol);

  /// set the background velocity for the test problem
  virtual void backgroundVelocity(const RealVect a_velocity);

  /// Set the physical dimension of the longest side of the domain
  /**
   */
  virtual void domainLength(Real a_domainLength);

  /// Set the refinement threshold
  /**
   */
  virtual void refinementThreshold(Real a_refineThresh);

  /// Set the tag buffer size
  /**
   */
  void tagBufferSize(int a_tagBufferSize);

protected:
  // Create a load-balanced DisjointBoxLayout from a collection of Boxes
  DisjointBoxLayout loadBalance(const Vector<Box>& a_grids);

  // Setup menagerie of data structures
  void levelSetup();

  // fill in advection velocity components into state container
  /** advection velocities are placed in a_state in the components 
      denoted by a_vel_interval.  it is an error for a_vel_interval to 
      not have SpaceDim components.
  */
  void fillAdvectionVelocities(LevelData<FArrayBox>& a_state, 
                               const Interval a_vel_interval);

  /// compute exact solution
  /** computes exact solution for advected scalar(s) if possible, 
      and places it in a_exact, starting at a_startComp
  */
  void computeExact(LevelData<FArrayBox>& a_exact, int a_startComp) const;

  // Get the next coarser level
  AMRAdvectDiffuse* getCoarserLevel() const;

  // Get the next finer level
  AMRAdvectDiffuse* getFinerLevel() const;

  // Conserverd state, U, at old and new time
  LevelData<FArrayBox> m_UOld,m_UNew;

  // CFL number
  Real m_cfl;

  // diffusion coefficient
  Real m_nu;

  // do implicit refluxing?
  bool m_doImplicitReflux;

  // Grid spacing
  Real m_dx;

  // Interpolation from fine to coarse level
  FineInterp m_fineInterp;

  // Averaging from coarse to fine level
  CoarseAverage m_coarseAverage;

  // New time step
  Real m_dtNew;

  // Number of converved states
  int m_numStates;

  // Names of conserved states
  Vector<string> m_stateNames;

  // Number of ghost cells (in each direction)
  int m_numGhost;

  // Physical dimension of the longest side of the domain
  Real m_domainLength;

  RealVect m_backgroundVelocity;

  // Level integrator
  LevelGodunov m_levelGodunov;

  // Flux register
  LevelFluxRegister m_fluxRegister;

  // Patch integrator (factory and object)
  PatchGodunov* m_patchGodunovFactory;
  PatchGodunov* m_patchGodunov;

  // helmholtzOp encapsulating diffusion operator
  HelmholtzOp m_helmop;

  // diffusion solver
  BaseHeatSolver* m_diffusionSolverPtr;

  // contains physical BCs for diffusion solver
  DomainGhostBC m_diffusionBC;

  // tolerance for diffusion solvers
  Real m_diffusion_solver_tol;

  // Refinement threshold for regridding
  Real m_refineThresh;

  // Tag buffer size
  int m_tagBufferSize;

  // Flag coarser and finer levels
  bool m_hasCoarser;
  bool m_hasFiner;

  DisjointBoxLayout m_grids;

private:
  // Disallowed for all the usual reasons
  void operator=(const AMRAdvectDiffuse& a_input)
  {
    MayDay::Error("invalid operator");
  }

  // Disallowed for all the usual reasons
  AMRAdvectDiffuse(const AMRAdvectDiffuse& a_input)
  {
    MayDay::Error("invalid operator");
  }
};

#endif
