/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

// #include <iostream>
// #include <cmath>
// #include <cstdio>
#include <iomanip>
#include "parstream.H"

#include "Box.H"
#include "Vector.H"
#include "IntVectSet.H"
#include "FArrayBox.H"
#include "DisjointBoxLayout.H"
#include "LayoutIterator.H"
#include "HDF5.H"
#include "SPMD.H"
#include "LoadBalance.H"
#include "LevelFluxRegister.H"
#include "ProblemDomain.H"
#include "BoxIterator.H"

#include "AMRIO.H"

#include "AMRLevel.H"
#include "AMRLevelIdealMHD.H"

#include "computeSum.H"

#include "PatchGodunovF_F.H"
#include "PatchIdealMHDF_F.H"

// Constructor
AMRLevelIdealMHD::AMRLevelIdealMHD()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD default constructor" << endl;
  }

  m_cfl = 0.3;
  m_domainLength = 1.0;
  m_refineThresh = 0.2;
  m_initial_dt_multiplier = 0.1;
  m_patchGodunovFactory = NULL;
  m_patchGodunov = NULL;
  m_poissonOpFactory = NULL;
  m_poissonOp = NULL;
}

// Destructor
AMRLevelIdealMHD::~AMRLevelIdealMHD()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD destructor" << endl;
  }

  // Get rid of the patch integrators (factory and object)
  if (m_patchGodunovFactory != NULL)
  {
    delete m_patchGodunovFactory;
  }

  if (m_patchGodunov != NULL)
  {
    delete m_patchGodunov;
  }

  // Get rid of the poisson operators (factory and object)
  if (m_poissonOpFactory != NULL)
  {
    delete m_poissonOpFactory;
  }

  if (m_poissonOp != NULL)
  {
    delete m_poissonOp;
  }
}

// This instance should never get called - historical
void AMRLevelIdealMHD::define(AMRLevel*  a_coarserLevelPtr,
                             const Box& a_problemDomain,
                             int        a_level,
                             int        a_refRatio)
{
  ProblemDomain physdomain(a_problemDomain);

  MayDay::Error("AMRLevelIdealMHD::define -\n\tShould never be called with a Box for a problem domain");
}

// Define new AMR level
void AMRLevelIdealMHD::define(AMRLevel*            a_coarserLevelPtr,
                             const ProblemDomain& a_problemDomain,
                             int                  a_level,
                             int                  a_refRatio)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::define " << a_level << endl;
  }

  // Call inherited define
  AMRLevel::define(a_coarserLevelPtr,
                   a_problemDomain,
                   a_level,
                   a_refRatio);

  // Get setup information from the next coarser level
  if (a_coarserLevelPtr != NULL)
  {
    AMRLevelIdealMHD* amrGodPtr = dynamic_cast<AMRLevelIdealMHD*>(a_coarserLevelPtr);

    if (amrGodPtr != NULL)
    {
      m_cfl = amrGodPtr->m_cfl;
      m_domainLength = amrGodPtr->m_domainLength;
      m_refineThresh = amrGodPtr->m_refineThresh;
      m_tagBufferSize = amrGodPtr->m_tagBufferSize;
    }
    else
    {
      MayDay::Error("AMRLevelIdealMHD::define: a_coarserLevelPtr is not castable to AMRLevelIdealMHD*");
    }
  }

  // Compute the grid spacing
  m_dx = m_domainLength / a_problemDomain.domainBox().longside();
  //  cout << "Domain Length" << m_domainLength << m_dx << endl;

  // Nominally, one layer of ghost cells is maintained permanently and
  // individual computations may create local data with more
  m_numGhost = 1;

  // Remove old patch integrator (if any), create a new one, and initialize
  if (m_patchGodunov != NULL)
  {
    delete m_patchGodunov;
  }

  m_patchGodunov = m_patchGodunovFactory->new_patchGodunov();
  m_patchGodunov->define(m_problem_domain,m_dx);

  // Remove old poisson operator (if any), create a new one, and initialize
  if (m_poissonOp != NULL)
  {
    delete m_poissonOp;
  }
  m_poissonOp = m_poissonOpFactory->new_levelop();
  m_poissonOp->setProblemDomain(m_problem_domain);
  m_poissonOp->setGridSpacing(m_dx);
  //  m_poissonOp->getPhysIBC()->->define(m_domain,m_dxLevel);


//    cout << "AMRLevelIdealMHD: define: Patch Godunov Defined:"<< isDefined()<<endl;
//    cout << "AMRLevelIdealMHD: define: m_patchGodunovFactory ptr address:"<< m_patchGodunovFactory<<endl;
//    cout << "AMRLevelIdealMHD: define: m_patchGodunov ptr address:"<< m_patchGodunov<<endl;

  // Get additional information from the patch integrator
  m_numStates = m_patchGodunov->numConserved();
  m_numFluxes = m_patchGodunov->numFluxes();
  m_stateNames = m_patchGodunov->stateNames();
  m_plotNames = m_patchGodunov->plotNames();
}

// Advance by one timestep
Real AMRLevelIdealMHD::advance()
{
  //  bool initializePhiToZero=false;
  bool initializePhiToZero=true;
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::advance level " << m_level << " to time " << m_time << endl;
  }

  // Copy the new to the old
  m_UNew.copyTo(m_UNew.interval(),
                m_UOld,
                m_UOld.interval());

  // Copy the new phi to the old phi
  m_phiNew.copyTo(m_phiNew.interval(),
                m_phiOld,
                m_phiOld.interval());

  Real newDt = 0.0;

  // Set up arguments to LevelGodunov::step based on whether there are
  // coarser and finer levels

  // Undefined flux register in case we need it
  LevelFluxRegister dummyFR;

  // Undefined leveldata in case we need it
  const LevelData<FArrayBox> dummyData;

  // Set arguments to dummy values and then fix if real values are available
  LevelFluxRegister* coarserFR = &dummyFR;
  LevelFluxRegister* finerFR   = &dummyFR;

  const LevelData<FArrayBox>* coarserDataOld = &dummyData;
  const LevelData<FArrayBox>* coarserDataNew = &dummyData;
    
  Real tCoarserOld = 0.0;
  Real tCoarserNew = 0.0;

  // A coarser level exists
  if (m_hasCoarser)
  {
    AMRLevelIdealMHD* coarserPtr = getCoarserLevel();

    // Recall that my flux register goes between my level and the next
    // finer level
    coarserFR = &coarserPtr->m_fluxRegister;

    coarserDataOld = &coarserPtr->m_UOld;
    coarserDataNew = &coarserPtr->m_UNew;
    const LevelData<FArrayBox>& coarserPhiNew = coarserPtr->m_phiNew;
    const LevelData<FArrayBox>& coarserPhiOld = coarserPtr->m_phiOld;
    //    coarserPhiNew = &coarserPtr->m_phiNew;
    //    coarserPhiOld = &coarserPtr->m_phiOld;

    tCoarserNew = coarserPtr->m_time;
    tCoarserOld = tCoarserNew - coarserPtr->m_dt;
  }

  // A finer level exists
  if (m_hasFiner)
  {
    // Recall that my flux register goes between my level and the next
    // finer level
    finerFR = &m_fluxRegister;
  }

  // Advance the solve one timestep
  newDt = m_levelGodunov.step(m_UNew,
                              m_rhs,
                              m_flux,
			      //                              m_yflux,
                              *finerFR,
                              *coarserFR,
                              dummyData,
                              *coarserDataOld,
                              tCoarserOld,
                              *coarserDataNew,
                              tCoarserNew,
                              m_time,
                              m_dt);


//    cout << "BEFORE LEVEL SOLVE m_level="<<m_level<<endl;
//    cout << "BEFORE LEVEL SOLVE m_time="<<m_time<<endl;
//    cout << "BEFORE LEVEL SOLVE m_dt="<<m_dt<<endl;

  if (m_hasCoarser)
  {
    // Fraction "m_time" falls between the old and the new coarse times
    AMRLevelIdealMHD* coarserPtr = getCoarserLevel();
    const LevelData<FArrayBox>& coarserPhiNew = coarserPtr->m_phiNew;
    const LevelData<FArrayBox>& coarserPhiOld = coarserPtr->m_phiOld;
    Real alpha = (m_time + m_dt - tCoarserOld) 
      / (tCoarserNew - tCoarserOld);
//      cout << "ALPHA="<<alpha<<endl;
//      cout << "tCoarserOld="<<tCoarserOld<<endl;
//      cout << "tCoarserNew="<<tCoarserNew<<endl;
    if (Abs(1.0-alpha) < 0.004* m_dt / m_ref_ratio)
      {
	alpha = 1.0;
      }
    if (Abs(alpha) < 0.004* m_dt / m_ref_ratio)
      {
	alpha = 0.0;
      }

    if (alpha < 0.0)
    {
      MayDay::Error( "AMRLevelIdealMHD::advance: alpha < 0.0");
    }

    if (alpha > 1.0)
    {
      MayDay::Error( "AMRLevelIdealMHD::advance: alpha > 1.0");
    }

    IntVect ivGhost = m_numGhost*IntVect::Unit;
    DisjointBoxLayout &coarseGrids=getCoarserLevel()->m_grids;
    LevelData<FArrayBox> coarserPhi(coarseGrids,1,ivGhost);

    InterpolatePhiInTime(coarserPhi,coarserPhiNew,coarserPhiOld,
    			 coarseGrids,alpha);
    //    if(alpha == 1.0)
    //      {
    //    	cout << "ALPHA="<<alpha<<endl;
    m_levelSolver.levelSolve(m_phiNew,&coarserPhi,m_rhs,initializePhiToZero);
    //        m_levelSolver.levelSolve(m_phiNew,&coarserPhiNew,m_rhs,initializePhiToZero);
    //    	m_levelSolver.levelSolveH(m_phiNew,m_rhs,initializePhiToZero);
       m_levelGodunov.BCorrection(m_UNew,
                                  m_phiNew,
                                  m_rhs,
                                  m_flux,
				  m_time,
                                  m_dt);
       //      }
  }
  else
    {
      //      cout << "NOCOARSER"<<endl;
      m_levelSolver.levelSolve(m_phiNew,NULL,m_rhs,initializePhiToZero);
       m_levelGodunov.BCorrection(m_UNew,
                                  m_phiNew,
                                  m_rhs,
                                  m_flux,
				  m_time,
                                  m_dt);
    }

  // Update the time and store the new timestep
  m_time += m_dt;
  Real returnDt = m_cfl * newDt;

  m_dtNew = returnDt;

  //  cout << "returnDt="<<m_dtNew<<endl;
  return returnDt;
}

// Things to do after a timestep
void AMRLevelIdealMHD::postTimeStep()
{
  static Real orig_integral = 0.0;
  static Real last_integral = 0.0;
  static bool first = true;

  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::postTimeStep " << m_level << endl;
  }

  if (m_hasFiner)
  {
    // Reflux
    Real scale = -1.0/m_dx;
    m_fluxRegister.reflux(m_UNew,scale);

    // Average from finer level data
    AMRLevelIdealMHD* amrGodFinerPtr = getFinerLevel();

    amrGodFinerPtr->m_coarseAverage.averageToCoarse(m_UNew,
                                                    amrGodFinerPtr->m_UNew);
    amrGodFinerPtr->m_coarseAveragePhi.averageToCoarse(m_phiNew,
           					    amrGodFinerPtr->m_phiNew);
  }

  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::postTimeStep " << m_level << " finished" << endl;
  }

  if (s_verbosity >= 2 && m_level == 0)
  {
    int nRefFine = 1;

    pout() << "AMRLevelIdealMHD::postTimeStep:" << endl;
    pout() << "  Sums:" << endl;
    for (int comp = 0; comp < m_numStates; comp++)
    {
      Interval curComp(comp,comp);
      Real integral = computeSum(m_UNew,NULL,nRefFine,m_dx,curComp);

      pout() << "    " << setw(23)
                       << setprecision(16)
                       << setiosflags(ios::showpoint)
                       << setiosflags(ios::scientific)
                       << integral
             << " --- " << m_stateNames[comp];

      if (comp == 0 && !first) {
        pout() << " (" << setw(23)
                       << setprecision(16)
                       << setiosflags(ios::showpoint)
                       << setiosflags(ios::scientific)
                       << (integral-last_integral)
               << " " << setw(23)
                      << setprecision(16)
                      << setiosflags(ios::showpoint)
                      << setiosflags(ios::scientific)
                      << (integral-orig_integral)
               << ")";
      }

      pout() << endl;

      if (comp == 0)
      {
        if (first)
        {
          orig_integral = integral;
          first = false;
        }

        last_integral = integral;
      }
    }
  }


}

// Create tags for regridding
void AMRLevelIdealMHD::tagCells(IntVectSet& a_tags)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::tagCells " << m_level << endl;
  }

  //  cout << "AMRLevelIdealMHD::tagCells m_problem_domain" <<  m_problem_domain << endl;
  // Since tags are calculated using only current time step data, use
  // the same tagging function for initialization and for regridding.

  // Create tags based on undivided gradient of density
  IntVectSet localTags;
  const DisjointBoxLayout& levelDomain = m_UNew.disjointBoxLayout();
  LevelData<FArrayBox> UWithGhost(levelDomain,
                                  1,
                                  IntVect::Unit);

  // Copy density
  m_UNew.copyTo(Interval(0,0),UWithGhost,Interval(0,0));
  // Copy x-magnetic field
  //    m_UNew.copyTo(Interval(4,4),UWithGhost,Interval(0,0));

  // Fill ghost cells
  UWithGhost.exchange(Interval(0,0));

  // If there is a coarser level interpolate undefined ghost cells
  if (m_hasCoarser)
  {
    const AMRLevelIdealMHD* amrGodCoarserPtr = getCoarserLevel();

    PiecewiseLinearFillPatch pwl(levelDomain,
                                 amrGodCoarserPtr->m_UNew.disjointBoxLayout(),
				 m_numStates,
                                 amrGodCoarserPtr->m_problem_domain,
                                 amrGodCoarserPtr->m_ref_ratio,
                                 1);

        pwl.fillInterp(UWithGhost,
                       amrGodCoarserPtr->m_UNew,
                       amrGodCoarserPtr->m_UNew,
                       1.0,
                       0,
                       0,
                       1);
//          pwl.fillInterp(UWithGhost,
//                         amrGodCoarserPtr->m_UNew,
//                         amrGodCoarserPtr->m_UNew,
//                         1.0,
//                         4,
//                         0,
//                         1);
  }

  // Compute undivided gradient
  DataIterator dit = levelDomain.dataIterator();
  for (dit.begin(); dit.ok(); ++dit)
  {
    const Box& b = levelDomain[dit()];
    FArrayBox gradFab(b,SpaceDim);
    const FArrayBox& UFab = UWithGhost[dit()];

    for (int dir = 0; dir < SpaceDim; ++dir)
    {
      const Box bCenter = b & grow(m_problem_domain,-BASISV(dir));
      const Box bLo     = b & adjCellLo(bCenter,dir);
      const int hasLo = ! bLo.isEmpty();
      const Box bHi     = b & adjCellHi(bCenter,dir);
      const int hasHi = ! bHi.isEmpty();
      FORT_GETGRAD(CHF_FRA1(gradFab,dir),
                   CHF_CONST_FRA1(UFab,0),
                   CHF_CONST_INT(dir),
                   CHF_BOX(bLo),
                   CHF_CONST_INT(hasLo),
                   CHF_BOX(bHi),
                   CHF_CONST_INT(hasHi),
                   CHF_BOX(bCenter));
    }

    FArrayBox gradMagFab(b,1);
    FORT_MAGNITUDE(CHF_FRA1(gradMagFab,0),
                   CHF_CONST_FRA(gradFab),
                   CHF_BOX(b));

    // Tag where gradient exceeds threshold
    BoxIterator bit(b);
    for (bit.begin(); bit.ok(); ++bit)
    {
      const IntVect& iv = bit();

      if (gradMagFab(iv) >= m_refineThresh)
      {
        localTags |= iv;
      }
    }
  }

  localTags.grow(m_tagBufferSize);

  // Need to do this in two steps unless a IntVectSet::operator &=
  // (ProblemDomain) operator is defined
  Box localTagsBox = localTags.minBox();
  localTagsBox &= m_problem_domain;
  localTags &= localTagsBox;

  a_tags = localTags;
  /*
  Vector<IntVectSet> allTags;

  const int destProc = uniqueProc(SerialTask::compute);

  gather(allTags,localTags,destProc);

  if (procID() == uniqueProc(SerialTask::compute))
  {
    for (int i = 0; i < allTags.size(); ++i)
    {
      a_tags |= allTags[i];
    }
  }
  */
  //  cout << "AMRLevelIdealMHD: Tags" << localTagsBox << endl;
  //  cout << "AMRLevelIdealMHD: Tags" << localTags << endl;
  //  cout << "AMRLevelIdealMHD: Tags" << allTags << endl;
}

// Create tags at initialization
void AMRLevelIdealMHD::tagCellsInit(IntVectSet& a_tags) 
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::tagCellsInit " << m_level << endl;
  }
  tagCells(a_tags);
}

// Set up data on this level after regridding
void AMRLevelIdealMHD::regrid(const Vector<Box>& a_newGrids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::regrid " << m_level << endl;
  }

  // Save original grids and load balance
  m_level_grids = a_newGrids;
  m_grids = loadBalance(a_newGrids);

  if (s_verbosity >= 4)
  {
    // Indicate/guarantee that the indexing below is only for reading
    // otherwise an error/assertion failure occurs
    const DisjointBoxLayout& constGrids = m_grids;

    pout() << "new grids: " << endl;

    for (LayoutIterator lit = constGrids.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << constGrids[lit()] << endl;
    }
  }

  // Save data for later
  LevelData<FArrayBox> UOld;
  UOld.define(m_UNew);
  m_UNew.copyTo(m_UNew.interval(),
                UOld,
                UOld.interval());

  LevelData<FArrayBox> PhiOld;
  PhiOld.define(m_phiNew);
  m_phiNew.copyTo(m_phiNew.interval(),
                PhiOld,
                PhiOld.interval());

//    LevelData<FArrayBox> xfluxOld;
//    xfluxOld.define(m_flux[0]);
//    m_flux[0].copyTo(m_flux[0].interval(),
//                  xfluxOld,
//                  xfluxOld.interval());

//    LevelData<FArrayBox> yfluxOld;
//    xfluxOld.define(m_flux[1]);
//    m_flux[1].copyTo(m_flux[1].interval(),
//                  yfluxOld,
//                  yfluxOld.interval());

  // Reshape state with new grids
  IntVect ivGhost = m_numGhost*IntVect::Unit;
  //  cout << "AMRLevelIdealMHD::regrid redefining" << ivGhost << endl;
  
  m_UNew.define(m_grids,m_numStates,ivGhost);
  m_UOld.define(m_grids,m_numStates,ivGhost);
  m_flux[0].define(m_grids,m_numFluxes,ivGhost);
  m_flux[1].define(m_grids,m_numFluxes,ivGhost);
  //  m_xflux.define(m_grids,m_numFluxes,ivGhost);
  //  m_yflux.define(m_grids,m_numFluxes,ivGhost);
  //  cout << "AMRLevelIdealMHD::regrid U done" << endl;
  m_phiNew.define(m_grids,1,ivGhost);
  m_phiOld.define(m_grids,1,ivGhost);
  m_rhs.define(m_grids,1,ivGhost);
  //  cout << "AMRLevelIdealMHD::regrid phi and rhs done" << endl;
  //  m_UNew.define(m_grids,m_numStates);
  //   m_UOld.define(m_grids,m_numStates);

  // Set up data structures
  levelSetup();

  // Interpolate from coarser level
  if (m_hasCoarser)
  {
    AMRLevelIdealMHD* amrGodCoarserPtr = getCoarserLevel();
    m_fineInterp.interpToFine(m_UNew,amrGodCoarserPtr->m_UNew);
    //    m_fineInterp.interpToFine(m_flux[0],amrGodCoarserPtr->m_flux[0]);
    //    m_fineInterp.interpToFine(m_flux[1],amrGodCoarserPtr->m_flux[1]);
    m_fineInterpPhi.interpToFine(m_phiNew,amrGodCoarserPtr->m_phiNew);
    m_fineInterpPhi.interpToFine(m_rhs,amrGodCoarserPtr->m_rhs);
  }

  // Copy from old state
  UOld.copyTo(UOld.interval(),
                  m_UNew,
                  m_UNew.interval());
  PhiOld.copyTo(PhiOld.interval(),
                  m_phiNew,
                  m_phiNew.interval());
//    xfluxOld.copyTo(xfluxOld.interval(),
//                    m_flux[0],
//                    m_flux[0].interval());
//    yfluxOld.copyTo(yfluxOld.interval(),
//                    m_flux[1],
//                    m_flux[1].interval());
}

// Initialize grids
void AMRLevelIdealMHD::initialGrid(const Vector<Box>& a_newGrids)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::initialGrid " << m_level << endl;
  }

  // Save original grids and load balance
  m_level_grids = a_newGrids;
  m_grids = loadBalance(a_newGrids);

  if (s_verbosity >= 4)
  {
    // Indicate/guarantee that the indexing below is only for reading
    // otherwise an error/assertion failure occurs
    const DisjointBoxLayout& constGrids = m_grids;

    pout() << "new grids: " << endl;
    for (LayoutIterator lit = constGrids.layoutIterator(); lit.ok(); ++lit)
    {
      pout() << constGrids[lit()] << endl;
    }
  }

  // Define old and new state data structures
  IntVect ivGhost = m_numGhost*IntVect::Unit;
  //  cout << "AMRLevelIdealMHD::initialgrid defines" << ivGhost << endl;
  m_UNew.define(m_grids,m_numStates,ivGhost);
  m_UOld.define(m_grids,m_numStates,ivGhost);
  m_flux[0].define(m_grids,m_numFluxes,ivGhost);
  m_flux[1].define(m_grids,m_numFluxes,ivGhost);
  //  m_xflux.define(m_grids,m_numFluxes,ivGhost);
  //  m_yflux.define(m_grids,m_numFluxes,ivGhost);
  m_phiNew.define(m_grids,1,ivGhost);
  m_phiOld.define(m_grids,1,ivGhost);
  m_rhs.define(m_grids,1,ivGhost);

  // Set up data structures
  levelSetup();
}

// Initialize data
void AMRLevelIdealMHD::initialData()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::initialData " << m_level << endl;
  }

  PhysIBC* physIBCPtr = m_patchGodunov->getPhysIBC();
  physIBCPtr->initialize(m_UNew);
  //  physIBCPtr = m_poissonOp->getPhysIBC();
  physIBCPtr->initializePhi(m_phiNew);
  //Temporary hack
  physIBCPtr->initializePhi(m_rhs);
}

// Things to do after initialization
void AMRLevelIdealMHD::postInitialize()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::postInitialize " << m_level << endl;
  }
}

// Create a local PatchGodunov factory using the argument as a factory
void AMRLevelIdealMHD::patchGodunov(const PatchGodunov* const a_patchGodunov)
{
  if (m_patchGodunovFactory != NULL)
  {
    delete m_patchGodunovFactory;
  }
  //  cout << "AMRLevelIdealMHD:new_patchGodunov will be called setting m_patchGodunovFactory" << a_patchGodunov << endl;
  m_patchGodunovFactory = a_patchGodunov->new_patchGodunov();
  //  cout << "AMRLevelIdealMHD:m_patchGodunovFactory set to" <<  m_patchGodunovFactory<< endl;
}

// Create a local PoissonOp factory using the argument as a factory
void AMRLevelIdealMHD::poissonOp(const PoissonOp* const a_poissonOp)
{
  if (m_poissonOpFactory != NULL)
  {
    delete m_poissonOpFactory;
  }
  //  cout << "AMRLevelIdealMHD:new_poissonOp will be called setting m_poissonOpFactory" << a_poissonOp << endl;
  m_poissonOpFactory = a_poissonOp->new_levelop();
  //  cout << "AMRLevelIdealMHD:m_poissonOpFactory set to" <<  m_poissonOpFactory<< endl;
}

#ifdef HDF5

// Write checkpoint header
void AMRLevelIdealMHD::writeCheckpointHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::writeCheckpointHeader" << endl;
  }

  // Setup the number of components
  HDF5HeaderData header;
  header.m_int["num_components"] = m_numStates+1;

  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    header.m_string[compStr] = m_stateNames[comp];
  }
  sprintf(compStr,"component_%d",8);
  header.m_string[compStr] = "Phi";

  // Write the header
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// Write checkpoint data for this level
void AMRLevelIdealMHD::writeCheckpointLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::writeCheckpointLevel" << endl;
  }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]       = m_ref_ratio;
  header.m_int ["tag_buffer_size"] = m_tagBufferSize;
  header.m_real["dx"]              = m_dx;
  header.m_real["dt"]              = m_dt;
  header.m_box ["prob_domain"]     = m_problem_domain.domainBox();

  // Setup the periodicity info
  D_TERM(
         if (m_problem_domain.isPeriodic(0))
           header.m_int ["is_periodic_0"] = 1;
         else
           header.m_int ["is_periodic_0"] = 0; ,
    
         if (m_problem_domain.isPeriodic(1))
           header.m_int ["is_periodic_1"] = 1;
         else
           header.m_int ["is_periodic_1"] = 0; ,
         
         if (m_problem_domain.isPeriodic(2))
           header.m_int ["is_periodic_2"] = 1;
         else 
           header.m_int ["is_periodic_2"] = 0; );

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  IntVect ivGhost = m_numGhost*IntVect::Unit;
    LevelData<FArrayBox> U( m_UNew.disjointBoxLayout(),
    			    m_patchGodunov->numConserved()+1,ivGhost);

  // Copy the current conserved variables into the temporary storage
       m_UNew.copyTo(m_UNew.interval(),U,m_UNew.interval());
       m_phiNew.copyTo(Interval(0,0),U,Interval(8,8));
  // Write the data for this level
  write(a_handle,U.boxLayout());
  write(a_handle,U,"data");
}

// Read checkpoint header
void AMRLevelIdealMHD::readCheckpointHeader(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::readCheckpointHeader" << endl;
  }

  // Reader the header
  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // Get the number of components
  if (header.m_int.find("num_components") == header.m_int.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointHeader: checkpoint file does not have num_components");
  }

  int numStates = header.m_int["num_components"];
  if (numStates != m_numStates+1)
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointHeader: num_components in checkpoint file does not match solver");
  }

  // Get the component names
  std::string stateName;
  char compStr[60];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    if (header.m_string.find(compStr) == header.m_string.end())
    {
      MayDay::Error("AMRLevelIdealMHD::readCheckpointHeader: checkpoint file does not have enough component names");
    }

    stateName = header.m_string[compStr];
    if (stateName != m_stateNames[comp])
    {
      MayDay::Error("AMRLevelIdealMHD::readCheckpointHeader: state_name in checkpoint does not match solver");
    }
  }
}

// Read checkpoint data for this level
void AMRLevelIdealMHD::readCheckpointLevel(HDF5Handle& a_handle)
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::readCheckpointLevel" << endl;
  }

  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  // Read the header for this level
  a_handle.setGroup(label);

  HDF5HeaderData header;
  header.readFromFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << "hdf5 header data:" << endl;
    pout() << header << endl;
  }

  // Get the refinement ratio
  if (header.m_int.find("ref_ratio") == header.m_int.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain ref_ratio");
  }
  m_ref_ratio = header.m_int["ref_ratio"];

  if (s_verbosity >= 2)
  {
    pout() << "read ref_ratio = " << m_ref_ratio << endl;
  }

  // Get the tag buffer size
  if (header.m_int.find("tag_buffer_size") == header.m_int.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain tag_buffer_size");
  }
  m_tagBufferSize=  header.m_int["tag_buffer_size"];

  if (s_verbosity >= 2)
  {
    pout() << "read tag_buffer_size = " << m_tagBufferSize << endl;
  }

  // Get dx
  if (header.m_real.find("dx") == header.m_real.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain dx");
  }
  m_dx = header.m_real["dx"];

  if (s_verbosity >= 2)
  {
    pout() << "read dx = " << m_dx << endl;
  }

  // Get dt
  if (header.m_real.find("dt") == header.m_real.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain dt");
  }
  m_dt = header.m_real["dt"];

  if (s_verbosity >= 2)
  {
    pout() << "read dt = " << m_dt << endl;
  }

  // Get the problem domain
  if (header.m_box.find("prob_domain") == header.m_box.end())
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain prob_domain");
  }

  Box domainBox = header.m_box["prob_domain"];
      
  // Get the periodicity info -- this is more complicated than it really
  // needs to be in order to preserve backward compatibility 
  bool isPeriodic[SpaceDim];
  D_TERM(if (!(header.m_int.find("is_periodic_0") == header.m_int.end()))
           isPeriodic[0] =  (header.m_int["is_periodic_0"] == 1);
         else
           isPeriodic[0] = false; ,

         if (!(header.m_int.find("is_periodic_1") == header.m_int.end()))
           isPeriodic[1] =  (header.m_int["is_periodic_1"] == 1);
         else
           isPeriodic[1] = false; ,

         if (!(header.m_int.find("is_periodic_2") == header.m_int.end()))
           isPeriodic[2] =  (header.m_int["is_periodic_2"] == 1);
         else
           isPeriodic[2] = false;);

  m_problem_domain = ProblemDomain(domainBox,isPeriodic);

  // Get the grids
  Vector<Box> grids;
  const int gridStatus = read(a_handle,grids);

  if (gridStatus != 0)
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain a Vector<Box>");
  }

  // Create level domain
  m_grids = loadBalance(grids);

  // Indicate/guarantee that the indexing below is only for reading
  // otherwise an error/assertion failure occurs
  const DisjointBoxLayout& constGrids = m_grids;

  LayoutIterator lit = constGrids.layoutIterator();
  for (lit.begin(); lit.ok(); ++lit)
  {
    const Box& b = constGrids[lit()];
    m_level_grids.push_back(b);
  }

  if (s_verbosity >= 4)
  {
    pout() << "read level domain: " << endl;
    LayoutIterator lit = m_grids.layoutIterator();
    for (lit.begin(); lit.ok(); ++lit)
    {
      const Box& b = m_grids[lit()];
      pout() << lit().intCode() << ": " << b << endl;
    }
    pout() << endl;
  }

  // Reshape state with new grids
  m_UNew.define(m_grids,m_numStates);
  const int dataStatus = read<FArrayBox>(a_handle,
                                         m_UNew,
                                         "data",
                                         m_grids);
  m_phiNew.define(m_grids,1);
  const int dataStatus2 = read<FArrayBox>(a_handle,
                                         m_phiNew,
                                         "data",
                                         m_grids);

  if (dataStatus != 0)
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain state data");
  }
  if (dataStatus2 != 0)
  {
    MayDay::Error("AMRLevelIdealMHD::readCheckpointLevel: file does not contain phi data");
  }
  m_UOld.define(m_grids,m_numStates);
  m_phiOld.define(m_grids,1);

  // Set up data structures
  levelSetup();
}

// Write plotfile header
void AMRLevelIdealMHD::writePlotHeader(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::writePlotHeader" << endl;
  }

  // Setup the number of components
  HDF5HeaderData header;
  header.m_int["num_components"] = m_numStates+2;

  // Setup the component names
  char compStr[30];
  for (int comp = 0; comp < m_numStates; ++comp)
  {
    sprintf(compStr,"component_%d",comp);
    header.m_string[compStr] = m_plotNames[comp];
  }
    sprintf(compStr,"component_%d",8);
    header.m_string[compStr] = "Phi";
    sprintf(compStr,"component_%d",9);
    header.m_string[compStr] = "divB";

  // Write the header
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }
}

// Write plotfile data for this level
void AMRLevelIdealMHD::writePlotLevel(HDF5Handle& a_handle) const
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::writePlotLevel" << endl;
  }

 // New stuff begins here
  // Setup an interval corresponding to the conserved variables
  //  Interval UInterval(0,m_patchGodunov->numConserved()-1);
  //  cout << "UInterval=" << UInterval << endl;
//      cout << "m_grids=" << m_grids << endl;
//      cout << "numConserved=" << m_patchGodunov->numConserved() << endl;

  // Create temporary storage with a layer of "m_numGhost" ghost cells
    IntVect ivGhost = m_numGhost*IntVect::Unit;
    LevelData<FArrayBox> U( m_UNew.disjointBoxLayout(),
    			    m_patchGodunov->numConserved()+2,ivGhost);

  // Copy the current conserved variables into the temporary storage
    //    m_UNew.copyTo(Interval(0,7),U,Interval(0,7));
       m_UNew.copyTo(m_UNew.interval(),U,m_UNew.interval());
       m_phiNew.copyTo(Interval(0,0),U,Interval(8,8));
       m_rhs.copyTo(Interval(0,0),U,Interval(9,9));

  // New stuff ends here
  // Setup the level string
  char levelStr[20];
  sprintf(levelStr,"%d",m_level);
  const std::string label = std::string("level_") + levelStr;

  a_handle.setGroup(label);

  // Setup the level header information
  HDF5HeaderData header;

  header.m_int ["ref_ratio"]   = m_ref_ratio;
  header.m_real["dx"]          = m_dx;
  header.m_real["dt"]          = m_dt;
  header.m_box ["prob_domain"] = m_problem_domain.domainBox();

  // Write the header for this level
  header.writeToFile(a_handle);

  if (s_verbosity >= 3)
  {
    pout() << header << endl;
  }

  //  cout << "Before Plotting check defines:"<< isDefined()<<endl;

    m_levelGodunov.plottingVars(m_UNew, U);

  // Write the data for this level
  //    write(a_handle,m_UNew.boxLayout());
  //    write(a_handle,m_UNew,"data");
      write(a_handle,U.boxLayout());
      write(a_handle,U,"data");
}

#endif

// Returns the dt computed earlier for this level
Real AMRLevelIdealMHD::computeDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::computeDt " << m_level << endl;
  }

  Real newDt;
  newDt = m_dtNew;

  return newDt;
}

// Compute dt using initial data
Real AMRLevelIdealMHD::computeInitialDt()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::computeInitialDt " << m_level << endl;
  }

  Real newDT = m_initial_dt_multiplier * m_dx
             / m_levelGodunov.getMaxWaveSpeed(m_UNew);

  return newDT;
}

// Set the CFL number
void AMRLevelIdealMHD::CFL(Real a_cfl)
{
  m_cfl = a_cfl;
}

// Set the physical dimension of the longest side of the domain
void AMRLevelIdealMHD::domainLength(Real a_domainLength)
{
  m_domainLength = a_domainLength;
}

// Set the refinement threshold
void AMRLevelIdealMHD::refinementThreshold(Real a_refineThresh)
{
  m_refineThresh = a_refineThresh;
}

// Set the tag buffer size
void AMRLevelIdealMHD::tagBufferSize(int a_tagBufferSize)
{
  m_tagBufferSize = a_tagBufferSize;
}

// Create a load-balanced DisjointBoxLayout from a collection of Boxes
DisjointBoxLayout AMRLevelIdealMHD::loadBalance(const Vector<Box>& a_grids)
{
  // Load balance and create boxlayout
  Vector<int> procMap;
  //if (procID() == uniqueProc(SerialTask::compute))
  //{
  //  LoadBalance(procMap,a_grids);
  //}
  //broadcast(procMap,uniqueProc(SerialTask::compute));

  // appears to be faster for all procs to do the loadbalance (ndk)
  LoadBalance(procMap,a_grids);

  if (s_verbosity >= 4)
  {
    pout() << "AMRLevelIdealMHD::loadBalance: procesor map: " << endl;
    for (int igrid = 0; igrid < a_grids.size(); ++igrid)
    {
      pout() << igrid << ": " << procMap[igrid] << "  " << endl;
    }
    pout() << endl;
  }

  DisjointBoxLayout dbl(a_grids,procMap,m_problem_domain);
  dbl.close();

  return dbl;
}

// Setup menagerie of data structures
void AMRLevelIdealMHD::levelSetup()
{
  if (s_verbosity >= 3)
  {
    pout() << "AMRLevelIdealMHD::levelSetup " << m_level << endl;
    pout() << "AMRLevelIdealMHD::levelSetup:m_patchGodunov factory:" << m_patchGodunovFactory << endl;
  }

  AMRLevelIdealMHD* amrGodCoarserPtr = getCoarserLevel();
  AMRLevelIdealMHD* amrGodFinerPtr   = getFinerLevel();

  m_hasCoarser = (amrGodCoarserPtr != NULL);
  m_hasFiner   = (amrGodFinerPtr   != NULL);

  if (m_hasCoarser)
  {
    int nvars=1;
    bool bcHomogeneous=true;
    int nRefCrse = m_coarser_level_ptr->refRatio();

    m_coarseAverage.define(m_grids,
                           m_numStates,
                           nRefCrse);

    m_coarseAveragePhi.define(m_grids,
                           1,
                           nRefCrse);

    m_fineInterp.define(m_grids,
                        m_numStates,
                        nRefCrse,
                        m_problem_domain);

    m_fineInterpPhi.define(m_grids,
                        1,
                        nRefCrse,
                        m_problem_domain);

    const DisjointBoxLayout& coarserLevelDomain = amrGodCoarserPtr->m_grids;

    // Maintain levelGodunov
    cout << "AMRLevelIdealMHD: Calling m_levelGodunov.define " << m_level<<m_patchGodunovFactory << endl;
    //    cout << "Before m_levelGodunov.define" << m_grids << endl;
    m_levelGodunov.define(m_grids,
                          coarserLevelDomain,
                          m_problem_domain,
                          nRefCrse,
                          m_dx,
                          m_patchGodunovFactory,
                          m_hasCoarser,
                          m_hasFiner);

    // Level OP
//        m_levelop.define(m_grids,
//                              &coarserLevelDomain,
//                              (double)m_dx,
//                              nRefCrse,
//                              m_problem_domain,
//                              bcHomogeneous,
//    			 nvars);

      // Maintain levelSolver
    //     cout << "Calling m_levelsolver.define2" << endl;
//       cout << "Before m_levelsolver.define2" << m_grids << endl;
//       cout << "Before m_levelsolver.define2" << m_problem_domain << endl;
//       cout << "Before m_levelsolver.define2" << m_dx << endl;
//       cout << "Before m_levelsolver.define2" << nRefCrse << endl;
//       cout << "Before m_levelsolver.define2" << m_poissonOp << endl;

     m_levelSolver.define(m_grids,
			  &coarserLevelDomain,
			  m_problem_domain,
			  (double)m_dx,
			  nRefCrse,
			  m_poissonOp,
			  nvars);

     cout << "Calling m_levelsolver. set max iter" << endl;
      m_levelSolver.setMaxIter(200);
    // This may look twisted but you have to do this this way because the
    // coarser levels get setup before the finer levels so, since a flux
    // register lives between this level and the next FINER level, the finer
    // level has to do the setup because it is the only one with the
    // information at the time of construction.
    
    // Maintain flux registers
    amrGodCoarserPtr->m_fluxRegister.define(m_grids,
                                            amrGodCoarserPtr->m_grids,
                                            m_problem_domain,
                                            amrGodCoarserPtr->m_ref_ratio,
                                            m_numStates);
    amrGodCoarserPtr->m_fluxRegister.setToZero();
  }
  else
  {
    int nvars=1;
    //    bool bcHomogeneous=true;
    //    cout << "AMRLevelIdealMHD: Calling m_levelGodunov.define2 " << m_level<<m_patchGodunovFactory << endl;
    m_levelGodunov.define(m_grids,
                          DisjointBoxLayout(),
                          m_problem_domain,
                          m_ref_ratio,
                          m_dx,
                          m_patchGodunovFactory,
                          m_hasCoarser,
                          m_hasFiner);
    //LevelOp
    //    m_levelop.define(m_grids,
    //                          &DisjointBoxLayout(),
    //                          (double)m_dx,
    //  			  m_ref_ratio,
    //                          m_problem_domain,
    //                          bcHomogeneous,
    //			 nvars);
    // Maintain levelSolver
    //    cout << "Calling m_levelsolver.define" << endl;
    m_levelSolver.define(m_grids,
			 NULL,
			 m_problem_domain,
			 (double)m_dx,
			 m_ref_ratio,
			 m_poissonOp,
			 nvars);
  }
  m_levelSolver.setVerbose(true);
  //  cout << "End of levelSetup"<< endl;
}

// Get the next coarser level
AMRLevelIdealMHD* AMRLevelIdealMHD::getCoarserLevel() const
{
  AMRLevelIdealMHD* amrGodCoarserPtr = NULL;

  if (m_coarser_level_ptr != NULL)
  {
    amrGodCoarserPtr = dynamic_cast<AMRLevelIdealMHD*>(m_coarser_level_ptr);

    if (amrGodCoarserPtr == NULL)
    {
      MayDay::Error("AMRLevelIdealMHD::getCoarserLevel: dynamic cast failed");
    }
  }

  return amrGodCoarserPtr;
}

// Get the next finer level
AMRLevelIdealMHD* AMRLevelIdealMHD::getFinerLevel() const
{
  AMRLevelIdealMHD* amrGodFinerPtr = NULL;

  if (m_finer_level_ptr != NULL)
  {
    amrGodFinerPtr = dynamic_cast<AMRLevelIdealMHD*>(m_finer_level_ptr);

    if (amrGodFinerPtr == NULL)
    {
      MayDay::Error("AMRLevelIdealMHD::getFinerLevel: dynamic cast failed");
    }
  }

  return amrGodFinerPtr;
}

void
AMRLevelIdealMHD::InterpolatePhiInTime(LevelData<FArrayBox> &a_Phi,
				       const LevelData<FArrayBox> &a_PhiNew,
				       const LevelData<FArrayBox> &a_PhiOld,
				       const DisjointBoxLayout & a_grids,
				       const Real& alpha)
{
  for (DataIterator dit = a_grids.dataIterator(); dit.ok(); ++dit)
  {
    // The current box
    Box curBox = a_grids.get(dit());

    // The current grid of conserved variables
    const FArrayBox& phiNew = a_PhiNew[dit()];
    const FArrayBox& phiOld = a_PhiOld[dit()];
    FArrayBox& phi = a_Phi[dit()];
    FORT_INTERPOLATEINTIME(CHF_FRA (phi),
			   CHF_CONST_FRA(phiNew),
			   CHF_CONST_FRA(phiOld),
			   CHF_BOX(curBox),
			   CHF_CONST_REAL(alpha)
			   );
  }
}
