/* _______              __
  / ___/ /  ___  __ _  / /  ___
 / /__/ _ \/ _ \/  ' \/ _ \/ _ \
 \___/_//_/\___/_/_/_/_.__/\___/
*/
//
// This software is copyright (C) by the Lawrence Berkeley
// National Laboratory.  Permission is granted to reproduce
// this software for non-commercial purposes provided that
// this notice is left intact.
//
// It is acknowledged that the U.S. Government has rights to
// this software under Contract DE-AC03-765F00098 between
// the U.S.  Department of Energy and the University of
// California.
//
// This software is provided as a professional and academic
// contribution for joint exchange. Thus it is experimental,
// is provided ``as is'', with no warranties of any kind
// whatsoever, no support, no promise of updates, or printed
// documentation. By using this software, you acknowledge
// that the Lawrence Berkeley National Laboratory and
// Regents of the University of California shall have no
// liability with respect to the infringement of other
// copyrights by any part of this software.
//

#include "PatchIdealMHD.H"
#include "PatchIdealMHDF_F.H"

#include "LoHiCenter.H"
#include "LoHiSide.H"

#include "LGintegrator.H"

PatchIdealMHD::PatchIdealMHD():PatchGodunov()
{
}

// Factory method - this object is its own factory.  It returns a pointer
// to new PatchGodunov object with its initial and boundary condtions, slope
// parameters, and artificial viscosity information defined.
PatchGodunov* PatchIdealMHD::new_patchGodunov() const
{
  //  cout << "PatchIdealMHD new_PatchGodunov:" << this <<endl;
  // Make the new object
  PatchGodunov* retval = 
    static_cast<PatchGodunov*>(new PatchIdealMHD());

  // Create a new initial and boundary condition (IBC) object for this new
  // patch integrator so it can define the IBC as it needs to (i.e. new
  // domain and grid spacing if necessary
  retval->setPhysIBC(getPhysIBC()->new_physIBC());

  // Set the slope and artificial viscosity parameters
  retval->setSlopeParameters(m_useFourthOrderSlopes,
                             m_useFlattening);
  retval->setArtificialViscosity(m_useArtificialViscosity,
                                 m_artificialViscosity);

  // Return the new object
  //  cout << "Patch Ideal MHD address:"<< retval << endl;
  return retval;
}

// Compute the maximum wave speed
Real PatchIdealMHD::getMaxWaveSpeed(const FArrayBox& a_U,
                                         const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.contains(a_box));

  Real speed = 0.0;

  FORT_MAXWAVESPEED(CHF_REAL(speed),
                    CHF_CONST_FRA(a_U),
                    CHF_BOX(a_box));

  return speed;
}

// Number of conserved variables
int PatchIdealMHD::numConserved()
{
  return UNUM;
}

// Names of the conserved variables
Vector<string> PatchIdealMHD::stateNames()
{
  Vector<string> retval;

  retval.push_back("density");
  retval.push_back("x-momentum");
  retval.push_back("y-momentum");
  retval.push_back("z-momentum");
  retval.push_back("x-magnetic");
  retval.push_back("y-magnetic");
  retval.push_back("z-magnetic");
  retval.push_back("energy-density");
  return retval; 
}

// Names of the plotting variables
Vector<string> PatchIdealMHD::plotNames()
{
  Vector<string> retval;

  retval.push_back("density");
  retval.push_back("x-velocity");
  retval.push_back("y-velocity");
  retval.push_back("z-velocity");
  retval.push_back("x-magnetic");
  retval.push_back("y-magnetic");
  retval.push_back("z-magnetic");
  retval.push_back("pressure");
  return retval; 
}

// Number of flux variables
int PatchIdealMHD::numFluxes()
{
  // In some computations there may be more fluxes than conserved variables
  return UFLUX;
}

// Number of primitive variables
int PatchIdealMHD::numPrimitives()
{
  // This doesn't equal the number of conserved variables because
  // auxiliary/redundant variable may be computed and stored
  return WNUM;
}

//  Number of primitive variables for which slopes are computed
int PatchIdealMHD::numSlopes()
{
  // This may be less than the number of primitive variables for the
  // reason given in numPrimitives() comments
  return WSLOPE;
}

// Compute the primitive variables from the conserved variables
void PatchIdealMHD::plottingVars(FArrayBox&       a_W,
                                    const FArrayBox& a_U,
                                    const Box&       a_box)
{
//     cout << "PatchIdealMHD plotting vars" << a_box << this << endl;
//     cout << "plottingVars" << isDefined()<<endl;
   assert(isDefined());
   assert(a_U.box().contains(a_box));
   assert(a_W.box().contains(a_box));

   FORT_CONSTOPLOT(CHF_FRA(a_W),
                   CHF_CONST_FRA(a_U),
                   CHF_BOX(a_box));
}

// Compute the primitive variables from the conserved variables
void PatchIdealMHD::consToPrim(FArrayBox&       a_W,
                                    const FArrayBox& a_U,
                                    const Box&       a_box)
{
  //  cout << "consToPrim in PatchIdealMHD:" << isDefined()<<endl;
  assert(isDefined());
  assert(a_U.box().contains(a_box));
  assert(a_W.box().contains(a_box));

  FORT_CONSTOPRIM(CHF_FRA(a_W),
                  CHF_CONST_FRA(a_U),
                  CHF_BOX(a_box));
}

// Compute the conserved variables from the primitive variables
void PatchIdealMHD::primToCons(FArrayBox&       a_U,
                                    const FArrayBox& a_W,
                                    const Box&       a_box)
{
  assert(isDefined());
  assert(a_U.box().contains(a_box));
  assert(a_W.box().contains(a_box));

  FORT_PRIMTOCONS(CHF_FRA(a_U),
                  CHF_CONST_FRA(a_W),
                  CHF_BOX(a_box));
}

// Extrapolate the primitive variables to the cell faces
void PatchIdealMHD::normalPred(FArrayBox&       a_WMinus,
                                    FArrayBox&       a_WPlus,
                                    const FArrayBox& a_W,
                                    const FArrayBox& a_dW,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  assert(isDefined());

  FORT_NORMALPRED(CHF_FRA(a_WMinus),
                  CHF_FRA(a_WPlus),
                  CHF_CONST_FRA(a_W),
                  CHF_CONST_FRA(a_dW),
                  CHF_CONST_REAL(a_scale),
                  CHF_CONST_INT(a_dir),
                  CHF_BOX(a_box));
}

// Compute a Riemann problem and generate fluxes at the faces
void PatchIdealMHD::riemann(FArrayBox&       a_F,
                                 const FArrayBox& a_WLeft,
                                 const FArrayBox& a_WRight,
                                 const int&       a_dir,
                                 const Box&       a_box)
{
  assert(isDefined());

  assert(a_F.box().contains(a_box));

  // Get the numbers of relevant variables
  int numPrim = numPrimitives();
  int numFlux = numFluxes();

  assert(a_F.nComp() == numFlux);
  assert(a_WLeft.nComp() == numPrim);
  assert(a_WRight.nComp() == numPrim);

  // Cast away "const" inputs so their boxes can be shifted left or right
  // 1/2 cell and then back again (no net change is made!)
  FArrayBox& shiftWLeft  = (FArrayBox&)a_WLeft;
  FArrayBox& shiftWRight = (FArrayBox&)a_WRight;

  // Solution to the Riemann problem
  FArrayBox Wgdnv(a_box,numPrim);

  // Shift the left and right primitive variable boxes 1/2 cell so they are
  // face centered
  shiftWLeft.shiftHalf(a_dir, 1);
  shiftWRight.shiftHalf(a_dir,-1);
  
  assert(shiftWLeft.box().contains(a_box));
  assert(shiftWRight.box().contains(a_box));

  // Riemann solver computes Wgdnv all edges that are not on the physical
  // boundary.
  FORT_RIEMANN(CHF_FRA(Wgdnv),
               CHF_CONST_FRA(shiftWLeft),
               CHF_CONST_FRA(shiftWRight),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
 
  // Shift the left and right primitive variable boxes back to their original
  // position
  shiftWLeft .shiftHalf(a_dir,-1);
  shiftWRight.shiftHalf(a_dir, 1);
  
  // Compute fluxes from the solution to the Riemann problem
  FORT_GETFLUX(CHF_FRA(a_F),
               CHF_CONST_FRA(Wgdnv),
               CHF_CONST_INT(a_dir),
               CHF_BOX(a_box));
}

// Update the primitive variables using fluxes (of the conserved variables)
void PatchIdealMHD::updatePrim(FArrayBox&       a_WMinus,
                                    FArrayBox&       a_WPlus,
                                    const FArrayBox& a_F,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  assert(isDefined());
  assert(a_WPlus.box().contains(a_box));
  assert(a_WMinus.box().contains(a_box));

  // Number of conserved variables
  int numCons = numConserved();

  // Temporary storage for converting primitive variables to conserved
  // variables and back again (after update with the fluxes)
  FArrayBox U(a_box,numCons);

  // for W+-, convert primitive quantities to conservative quantities,
  // update with the fluxes, and convert back
  primToCons(U,a_WMinus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WMinus,U,a_box);

  primToCons(U,a_WPlus,a_box);
  updateCons(U,a_F,a_scale,a_dir,a_box);
  consToPrim(a_WPlus,U,a_box);
}

// Update the conserved variable using fluxes and a scaling factor
void PatchIdealMHD::updateCons(FArrayBox&       a_U,
                                    const FArrayBox& a_F,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  assert(isDefined());

  //  cout << "PatchIdealMHD updateCons This is " << this << endl;
  FORT_UPDATECONS(CHF_FRA(a_U),
                  CHF_CONST_FRA(a_F),
                  CHF_CONST_REAL(a_scale),
                  CHF_CONST_INT(a_dir),
                  CHF_BOX(a_box));
}

// Final Update of the conserved variable using fluxes and a scaling factor
void PatchIdealMHD::updateConsFinal(FArrayBox&       a_U,
                                    const FArrayBox& a_F,
                                    const Real&      a_scale,
                                    const int&       a_dir,
                                    const Box&       a_box)
{
  assert(isDefined());

  //  cout << "PatchIdealMHD updateCons This is " << this << endl;
  FORT_UPDATECONSFINAL(CHF_FRA(a_U),
                  CHF_CONST_FRA(a_F),
                  CHF_CONST_REAL(a_scale),
                  CHF_CONST_INT(a_dir),
                  CHF_BOX(a_box));
}

// Update the conserved variable using fluxes and a scaling factor
void PatchIdealMHD::postUpdateCons(FArrayBox&       a_U,
				   FArrayBox&       a_rhs,
				   const FArrayBox& a_Uold,
				   const FArrayBox a_F[CH_SPACEDIM],
				   const Real&      a_dt,
				   const Real&      a_dx,
				   const Box&       a_box)
{
  assert(isDefined());
  // Compute divergence and store in rhs
  //Initialize RHS to be zero
  a_rhs.setVal(0.0);

  for (int dir1 = 0; dir1 < SpaceDim; dir1++)
  {
    const FArrayBox& a_Flux = a_F[dir1];

        FORT_COMPUTEDIVERGENCE(CHF_FRA(a_rhs),
    			    CHF_CONST_FRA(a_Flux),
    			    CHF_CONST_REAL(a_dx),
    			    CHF_CONST_INT(dir1),
    			    CHF_BOX(a_box));

  }

      FORT_FILTERCONS(CHF_FRA(a_U),
                      CHF_CONST_FRA(a_Uold),
                      CHF_CONST_REAL(a_dt),
                      CHF_CONST_REAL(a_dx),
                      CHF_BOX(a_box));
}

// Interval within the primitive variables corresponding to the velocities
Interval PatchIdealMHD::velocityInterval()
{
#if CH_SPACEDIM==2
  Interval retval(WVELX,WVELY);
#elif CH_SPACEDIM==3
  Interval retval(WVELX,WVELZ);
#else 
  bogus_spacedim();
#endif

  return retval;
}

// Component index within the primitive variables of the pressure
int PatchIdealMHD::pressureIndex()
{
  return WPRES;
}

// Component index within the primitive variables of the bulk modulus
int PatchIdealMHD::bulkModulusIndex()
{
  return WPRES;
}

//Correct State
void 
PatchIdealMHD::correctState(FArrayBox&       a_U,
			    FArrayBox&       a_phi,
			    FArrayBox&       a_rhs,
			    FArrayBox       a_flux[CH_SPACEDIM],
			    const Real&     a_dt,
			    const Box&     a_Box)
{
  //  return;
  assert(isDefined());
  //  cout << "PatchIdealMHD correctState This is " << this << endl;
  Real scale = a_dt/m_dx;
  Box fluxBox;

  for (int idir = 0; idir < SpaceDim; idir++)
    {
      fluxBox = a_Box;
      fluxBox.surroundingNodes(idir);

      FArrayBox &F=a_flux[idir];

      FORT_CORRECTFLUX(CHF_FRA(a_U),
		       CHF_CONST_FRA(a_phi),
		       CHF_FRA(F),
		       CHF_CONST_REAL(m_dx),
		       CHF_CONST_REAL(scale),
		       CHF_CONST_INT(idir),
		       CHF_BOX(a_Box),
		       CHF_BOX(fluxBox));
  }
  FArrayBox &Fx=a_flux[0];
  FArrayBox &Fy=a_flux[1];
  //  return;
    {
      FORT_CHECKDIVB(CHF_FRA(a_U),
		 CHF_CONST_FRA(a_rhs),
 		 CHF_FRA(Fx),
 		 CHF_FRA(Fy),
 		 CHF_CONST_REAL(m_dx),
 		 CHF_BOX(a_Box));
    }
}


